

#include "AMR-HDF5.h"
#include "hdf5.h"

#include <iostream>
#include <map>
#include <cstdio>

#include <cassert>
#include <cstring>
#include <string>
#include <vector>
#include <list>
#include <set>
#include <cmath>
#include <typeinfo>
#include <fstream>
#include <algorithm>
#include <cstdlib>
#include <iomanip>
#include <csignal>
#include <cstring>
#ifdef MPI
#include "mpi.h"
#endif

#define HDF5
#ifndef CH_SPACEDIM
#define CH_SPACEDIM 3 // available options are 1, 2, or 3
#endif

#ifndef CH_USE_FLOAT
#define CH_USE_DOUBLE // options are CH_USE_DOULBE, or CH_USE_FLOAT  for representing 
                      //  floating point values. Currently the code has to be compiled
                      //  for one or the other, as a side effect of using Chombo for the
                      //  I/O
#endif

#ifdef HAVE_CHOMBO
#include "CFile.h"    // large header file containing all the definitions of Chombo functions
                      // used in this translation unit.  It is assumed that the user will
                      // subsequently be linking this object file with the rest of the Chombo
                      // library
#else
#include "CFile.cpp"   //  One big C++ file holding all the Chombo source code needed for 
                       //  this functionality.  Later this code will have it's own independent
                       //  implementation.  People who have no intention of linking to the
                       //  Chombo libraries can use this version of the include.
#endif

typedef struct HandleNode {
  HDF5Handle* handle;
  int ID;
  bool operator==(const HandleNode& n){ return ID==n.ID;}
  int mpicomm;
} HandleNode ;

HandleNode emptyHandle;
int nextHandleID=17;
std::list<HandleNode> HandleList;
std::list<HandleNode>::iterator HIT;

HandleNode& getHandle(int ID)
{
  for(HIT=HandleList.begin(); HIT != HandleList.end(); HIT++)
	{
	  if((*HIT).ID == ID) return *HIT;
	}
  emptyHandle.ID=-1;
  emptyHandle.handle = NULL;
  return emptyHandle;
}  


void buildAliasData(int levelhandle, DisjointBoxLayout& layout, LevelData<FArrayBox>& data,
					int& nComponents)
{
  int MPI_communicator, dimension, centering, type, ordering, localcount;
  AMR_GetInfo(levelhandle,
			  &MPI_communicator,
			  &nComponents,
			  &dimension,
			  &centering,
			  &type,
			  &ordering,
			  &localcount);
  int loRange[3], hiRange[3], loValid[3], hiValid[3];
  IntVect ghost;
  for(int b=0; b<localcount; b++)
	{
	  AMR_GetPatchInfo(levelhandle, b, loRange, hiRange,loValid, hiValid);
	  IntVect lo(loRange), hi(hiRange), loV(loValid), hiV(hiValid);
	  if(b==0){
		ghost = loV - lo;
		if(D_TERM(ghost[0]<0, || ghost[1]<0, || ghost[2]<0))
		  MayDay::Error("ghost vector with negative size");
	  }
	  else
		{
		  if((loV - lo) != ghost || (hi-hiV) != ghost)
			MayDay::Error("inconsistent use of ghosting for AMR-HDF5 capabilities");
		}
	  Box next(loV, hiV);
	  layout.aliasAddBox(next);			
	}
	layout.aliasClose();
	LayoutData<Real*> ptrs(layout);
	DataIterator dit  = layout.dataIterator();
	int b=0;
	for(dit.begin(); dit.ok(); ++b, ++dit)
	  {
		Real* dataPtr;
#ifdef CH_USE_DOUBLE
		AMR_GetPatchDouble(levelhandle, b, &dataPtr);
#else
		AMR_GetPatchFloat(levelhandle, b, &dataPtr);
#endif
		ptrs[dit()] = dataPtr;
	  }
	assert(b == localcount);
	FABAliasDataFactory factory(ptrs);
	data.define(layout, nComponents, ghost, factory);
}

extern "C" {
  void AMR_OpenHDF5(char* filename, int MPIComm, int m, int* fileHandle)
  {
	HandleNode h;
	h.handle  = new HDF5Handle();
	h.ID = nextHandleID;
	h.mpicomm =  MPIComm;
	nextHandleID++;
#ifdef MPI
	MPI_Comm Ccomm = Chombo_MPI::comm;
	Chombo_MPI::comm = MPI_Comm_f2c(MPIComm);
#endif
	HDF5Handle::mode b = (HDF5Handle::mode)m;//enum assign.  behave warily.
	h.handle->open(filename, b);
	HandleList.insert(HandleList.begin(), h);
    *fileHandle = h.ID;
#ifdef MPI
	Chombo_MPI::comm = Ccomm;
#endif
  }
 void amropenhdf5_(char *filename, long int length, 
					int* MPI_Communicator, int* m, int *hdf5)
 {
   if(filename[length-1] != (char)0)
	 MayDay::Error("amropenhdf5_ requires null-terminated string");

   AMR_OpenHDF5(filename,*MPI_Communicator, *m, hdf5);
 }
   



  void AMR_CloseHDF5(int fileHandle)
  {
	HandleNode& n = getHandle(fileHandle);
	if(n.ID==-1) MayDay::Error("closeHDF5 called on fileHandle, but handle not found");
	n.handle->close();
	HandleList.remove(n);
  }

  void amrclosehdf5_( int* hdf5 ) { AMR_Close(*hdf5);}

	  
  void amrwritedata_(int* hdf5, int* numLevels, int *levelData, int* ncomp,
					 int *refineFactor, double *spacing,
					 double* time, double* dt, int *domain_lo, int *domain_hi)
  {

	char** cmpNames = new char* [*ncomp];
	char*  buffer = new char [*ncomp * 4];
	char* b = buffer;
	for(int i=0; i<*ncomp; i++, b+=4)
	  {
		cmpNames[i] = b;
		sprintf(b, "%i",i);
	  }
		
		
	AMR_WriteData(*hdf5, *numLevels, levelData, *ncomp, cmpNames, refineFactor,
				  spacing, *time, *dt, domain_lo, domain_hi);
	delete buffer;
	delete cmpNames;
  }

  void AMR_WriteData(int hdf5, int numLevels, int* levelData, 
					int ncomp, char** compNames , int* refineFactor, double* spacing,
					double time, double dt, const int* domain_lo, const int* domain_hi)
  {
	HandleNode& n = getHandle(hdf5);
	if(n.ID==-1) MayDay::Error("writeAMRData called on fileHandle, but handle not found");
	  
	assert(numLevels > 0);

	Vector<DisjointBoxLayout> grids(numLevels);
	Vector<LevelData<FArrayBox>* > fabData(numLevels);
	Vector<int> refRatios(numLevels-1);
	for(int i=0; i<numLevels-1; i++) refRatios[i] = refineFactor[i];
	for(int level=0; level<numLevels; level++)
	  {
		int levelhandle = levelData[level];
		DisjointBoxLayout& layout = grids[level];
		fabData[level] = new LevelData<FArrayBox>;

		int nComponents;
		buildAliasData(levelhandle, layout, *(fabData[level]),
					   nComponents);
	  }
	 
   
	IntVect lo(D_DECL(domain_lo[0], domain_lo[1], domain_lo[2]));
	IntVect hi(D_DECL(domain_hi[0], domain_hi[1], domain_hi[2]));
	Box domain(lo, hi);
	Vector<string> vectNames(ncomp);
	for(int i=0; i<ncomp; i++)
	  {
		vectNames[i] = compNames[i];
	  }
	
	WriteAMRHierarchyHDF5(*(n.handle),                             
                      grids, 
                      fabData, 
                      vectNames,                 
                      domain,                               
                      spacing[0],                                         
                      dt,                                         
                      time,
                      refRatios,                           
                      numLevels);


  }
  


 

  void AMR_WriteStringAttribute(int hdf5, char* name, char* value)
  {
	HandleNode& n = getHandle(hdf5);
	if(n.ID==-1) MayDay::Error("writeAMRString called on fileHandle, but handle not found");

	HDF5HeaderData att;
	att.m_string[name] = value;
	att.writeToFile(*(n.handle));

  }



  void  AMR_ReadMetaData(int hdf5, int *numLevels, int *ncomp, 
			int *ndim, double *space, 
			int *lo, int *hi,
			double *time, double* dtCoarse) 
  {

	HandleNode& n = getHandle(hdf5);
	if(n.ID==-1) MayDay::Error("readAMRString called on fileHandle, but handle not found");

	HDF5HeaderData att;
	att.readFromFile(*(n.handle));

	*numLevels = att.m_int["num_levels"];
	*ncomp = att.m_int["num_components"];

	n.handle->setGroup("/Chombo_global");
	att.clear();
	att.readFromFile(*(n.handle));
	*ndim = att.m_int["SpaceDim"];
	
	n.handle->setGroupToLevel(0);
	att.clear();
	att.readFromFile(*(n.handle));
	*dtCoarse = att.m_real["dt"];
	Box domain;  Real delta;
	*time = att.m_real["time"];
	domain = att.m_box["prob_domain"];
	const IntVect& l = domain.smallEnd();
	const IntVect& h = domain.bigEnd();
	delta  = att.m_real["dx"];
	D_TERM(space[0] = delta;, space[1]=delta;,space[2]=delta;);
	D_TERM(lo[0] = l[0];, lo[1]=l[1];, lo[2]=l[2]);
	D_TERM(hi[0] = h[0];, hi[1]=h[1];, hi[2]=h[2]);

	n.handle->setGroup("/");
  }

  void amrwritecompname_(int* hdf5, int* compIndex, char* compname, long int length)
  {
	char labelChSt[20];
	sprintf(labelChSt, "component_%d", *compIndex);
	AMR_WriteStringAttribute(*hdf5, labelChSt , compname);
  }

   void amrreadcompname_(int* hdf5, int* compIndex, char *names, long int length)
   {
	 char labelChSt[20];
	 sprintf(labelChSt, "component_%d", *compIndex);
	 char* value;
	 AMR_ReadStringAttribute(*hdf5, labelChSt , &value);
	 strcpy(names, value);
	 if(value != NULL)
	   free(value);

   }

  void AMR_ReadCompName(int hdf5, char **names) 
  {
	HandleNode& n = getHandle(hdf5);
	if(n.ID==-1) MayDay::Error("readAMRCompName called on fileHandle, but handle not found");

	HDF5HeaderData att;
	att.readFromFile(*(n.handle));
	int ncomp = att.m_int["num_components"];
	if(ncomp <= 0)
	  {
		MayDay::Warning("ReadAMRHierarchyHDF5: Bogus number of Components");
	  }
   
	for(int ivar = 0; ivar < ncomp; ivar++)
	  {
		char labelChSt[80];
		sprintf(labelChSt, "component_%d", ivar);
		string label(labelChSt);
		std::string str = att.m_string[label];
		names[ivar] = (char*)malloc(str.size()+1);
		strcpy(names[ivar], str.c_str());
	  }
  }





  void AMR_ReadNumBoxes(int hdf5, int level, int *ntotalBoxes) 
  {
	HandleNode& n = getHandle(hdf5);
	if(n.ID==-1) MayDay::Error("readAMRNumBoxes called on fileHandle, but handle not found");
	
	n.handle->setGroupToLevel(level);

	hid_t  boxdataset = H5Dopen(n.handle->groupID(), "boxes");
	assert(boxdataset >= 0);
	hid_t boxdataspace =  H5Dget_space(boxdataset);
	assert(boxdataspace >= 0);
	hsize_t dims[1], maxdims[1];
	H5Sget_simple_extent_dims(boxdataspace, dims, maxdims);
	*ntotalBoxes = dims[0];
	n.handle->setGroup("/");
	H5Sclose(boxdataspace);
	H5Dclose(boxdataset);
  }



  void AMR_ReadBoxes(int hdf5, int level, 
					int *loValid, int* hiValid, 
					int* loRange, int* hiRange) 
  {
	HandleNode& n = getHandle(hdf5);
	if(n.ID==-1) MayDay::Error("readAMRBoxes called on fileHandle, but handle not found");
	
	n.handle->setGroupToLevel(level);
	Vector<Box> boxes;
	int err = read(*(n.handle), boxes);
	if(err != 0)
	  {
		MayDay::Error("Error reading boxes from filehandle");
	  }
	HDF5HeaderData att;
	att.readFromFile(*(n.handle));
	IntVect ghost = att.m_intvect["outputGhost"];
	int ii = 0;
	for(int i=0; i<boxes.size(); i++,ii+=CH_SPACEDIM)
	  {
		const Box& b = boxes[i];
		IntVect lo = b.smallEnd();
		IntVect hi = b.bigEnd();
		D_TERM(loValid[ii]= lo[0];,loValid[ii+1]=lo[1];,loValid[ii+2]=lo[2]);
		D_TERM(hiValid[ii]= hi[0];,hiValid[ii+1]=hi[1];,hiValid[ii+2]=hi[2]);
		lo-=ghost;
		hi+=ghost;
		D_TERM(loRange[ii]= lo[0];,loRange[ii+1]=lo[1];,loRange[ii+2]=lo[2]);
		D_TERM(hiRange[ii]= hi[0];,hiRange[ii+1]=hi[1];,hiRange[ii+2]=hi[2]);
	  }
	n.handle->setGroup("/");
  }






  void AMR_ReadRefFactors(int hdf5, int *refs) 
  {
	HandleNode& n = getHandle(hdf5);
	if(n.ID==-1) MayDay::Error("readAMRRefFactors called on fileHandle, but handle not found");
	
	HDF5HeaderData att;
	att.readFromFile(*(n.handle));
	int numLevels = att.m_int["num_levels"];
	
	for(int i=0; i<numLevels; i++)
	  {
		att.clear();
		n.handle->setGroupToLevel(i);
		att.readFromFile(*(n.handle));
		refs[i] = att.m_int["ref_ratio"];
	  }
	refs[numLevels-1] = 1;
	n.handle->setGroup("/");
  }




  void AMR_ReadLevel(int hdf5, int level, int levelDatahandle, int compBegin, int ncomp) 
  {

	HandleNode& n = getHandle(hdf5);
	if(n.ID==-1) MayDay::Error("readAMRLevel called on fileHandle, but handle not found");

	n.handle->setGroupToLevel(level);

	DisjointBoxLayout layout;
	LevelData<FArrayBox> data;
	int components;
	buildAliasData(levelDatahandle, layout, data, components);
	assert(ncomp>0);
	if(ncomp > components){
	  MayDay::Error("Requested a read of more components that available space");
	}
	Interval comps(compBegin, compBegin+ncomp-1);
	int err = read(*(n.handle), data, "data", 
				   layout, comps, false);
	if(err != 0){
	  MayDay::Error("Error reading floating point data in readAMRLevel");
	}  
	n.handle->setGroup("/");
  }




  void AMR_ReadPatch(int hdf5, int level, int globalpatch, int comp,
		    void *d_ptr) 
  {
	HandleNode& n = getHandle(hdf5);
	if(n.ID==-1) MayDay::Error("readAMRPatch called on fileHandle, but handle not found");
	
	n.handle->setGroupToLevel(level);

	FArrayBox a_fab;
	
	Interval a_comps(comp, comp);
	int err = readFArrayBox(*(n.handle),
							a_fab,
							level,
							globalpatch,
							a_comps);
	if(err!= 0){
	  MayDay::Error("Error reading patch data");
	}
	int num = a_fab.box().numPts()*sizeof(Real);
	memcpy(a_fab.dataPtr(), d_ptr, num);
	n.handle->setGroup("/");	
  }



  void AMR_ReadStringAttribute(int hdf5, char* name, char** value)
  {
	static char* nullPtr = NULL;
	nullPtr = NULL;

	HandleNode& n = getHandle(hdf5);
	if(n.ID==-1) MayDay::Error("readAMRString called on fileHandle, but handle not found");

	HDF5HeaderData att;
	att.readFromFile(*(n.handle));
	
	if(att.m_string.find(name) != att.m_string.end()){
	  std::string val  = att.m_string[name];
	  *value = (char*)malloc(val.size()+1);
	  strcpy(*value, val.c_str());
	}
	else
	  *value = nullPtr;

  }
}
