/**  \mainpage AMR Interoperability Aplication Programming Interfaces
  <b>  Data Aliasing:  @see AMRData.h
</b>
  <br>
<br>

<p>

<ul>
<li>Problem


<ul>
<li>The largest impediment to interoperability of various 
Structured AMR efforts within the DOE (and beyond) are related to data
description.
</li>
</ul>

<p>
</li>
<li>Goal

<ul>
<li>Provide a community-standard interface for describing Structured AMR
data sets and associated data structures.
</li>
<li>Enable different community members to focus efforts on component development
 without suffering from data-conversion losses (performance and memory).
</li>
<li>examples:  Chombo Elliptic Solver Component, AMR-HDF5 File IO Component.
</li>
</ul>
</li>

<li>Approach

<ul>
<li>A API written in ANSI-C for client components to describe their
AMR data sets to server components.
</ul>

</ul>

<BR>

<B>     AMR-HDF5 Interface  @see AMR-HDF5.h
</B>
  <BR>
<BR>

<P>

<UL>
<LI>HDF5
  
<UL>
<LI>portable binary files.
</LI>
<LI>supports additional meta data for specific application domains.
</LI>
<LI>hyper-slab. (out-of-core tools). Ability to read in specific
component of a specific box.
</LI>
<LI>parallel I/O support on most platforms.
  
</LI>
</UL>
</LI>
<LI>readable/(writable) by ChomboVis
</LI>
<LI>File organization
  
<UL>
<LI>component names
</LI>
<LI>grid spacing
</LI>
<LI>centerings
</LI>
<LI>dimensions
</LI>
<LI>data organized as <I>levels of grids</I>
</LI>
<LI>parallel neutral (but contains processor assignments from when the
file was written)
  
</LI>
</UL>
</LI>

</UL>

*/

// multi-include guard
#ifndef AMR_HDF5_h
#define AMR_HDF5_h

#include "AMRData.h"


/* Keep C++ compilers from getting confused */
#if defined (__cplusplus)
extern "C"
{
#endif

 /**
	 \name Client-side functions */
  /*@{*/

  /**
     Opens an AMR-HDF5 file
     @param filename : name of hdf5 file
     @param m : which mode, 0 - Create, 1 - OPEN_RDONLY, 2 - OPEN_RDWR
     @param MPI_Communicator: Integer representation of your MPI communicator.
	 When not running under MPI this value is ignored.  When running under MPI
	 this communicator is used for the parallel file opening 
	 (MPI-IO interface). 
	 you can create an integer representation for the communicator using 
	 the functions <code>MPI_Comm_c2f</code> and <code>MPI_Comm_f2c</code>
     @param hdf5 : an integer supplied by the caller, which is filled in
     with a handle.
  */
  void AMR_OpenHDF5(char *filename, int MPI_Communicator, int m, int *hdf5) ;

  /**
     Close an HDF5 file.  this integer opaque handle is will now cause
	 the code to abort if used in other AMR_ functions.
  */
  void AMR_CloseHDF5( int hdf5 ) ;


  /**
     Write out a set of levels to file. This assumes that you have 
     gone through the functions in AMRData.h, creates a number of
     Levels, have their handles and have filled them with data. Now
     it's time to write them to file

     @param hdf5 : the file handle, returned by openHDF5() above.

     @param numLevels : an integer which tells the callee how many levels
     exist

     @param levelData : an int array, numLevels long,  containing the 
     level handles created by the methods in AMRData.h

     @param ncomp : number of variables per cell, being stored in the
     HDF5 file.

     @param compNames : pointer to an array of pointers to char (array
     length = ncomp). The callee will malloc memory to each of the
     array elements and fill them up with the names of the components
     (e.g. "Pressure", "rho"). The caller deletes this memory.

     @param refineFactor : an array, numLevels long, that the caller
     fills up with the refinement factor. refinement factor is the factor
     by which a level is coarser that the immediately finer level. The
     finest level will have its refineFactor set to -1 - this serves
     as an error check, in case someone misinterprets refineFactor.
     NOTE: For GrACE, refineFactor is a constant; for Chombo it varies.

     @param spacing : a double array, dimension long, into
     which dx, dy and dz for the coarse mesh has been written by the caller.
     dx, dy, dz for finer meshes can be derived using refineFactor.

     @param double time : unsteady simulation's time variable at which this
     data existed.

     @param dt : delta t on the coarse mesh, when the data was dumped.

     @param domain_lo : int array, dimension long, containing the indices
     of the lower bounding box corner for the entire domain (and not just
     the subdomain on this processor).

     @param domain_hi : ditto as above, but for upper bounding box corner.
  */
  void AMR_WriteData(int hdf5, int numLevels, int *levelData, int ncomp,
		    char **compNames, int *refineFactor, double *spacing,
		    double time, double dt, CONST int *domain_lo, CONST int *domain_hi) ;

  /**
	 @param hdf5: file handle.
	 @param name: attribute name.  Cannot use '/' in name
	 @param value: value of attribute. 
  */
  
  void AMR_WriteStringAttribute(int hdf5, char* name, char* value);
  /**
     The following methods are to be used when reading in a file for
     the first time. The caller has no idea how many levels exist and
     neither does he have the LevelDatas created. This it time t = 0.

     In these calls, all the meomy is allocated by the caller and 
     the callee only fills them up. The caller deletes the memory.

     @param hdf5 : file handle, as obtained above.

     @param numLevels : an integer, supplied by caller and filled by
     callee, re the number of levels in the file.

     @parma ncomp : number of variables per grid cell stored in
     the file.

     @param ndim : 1D/2D/3D ?

     @param spacingCoarse : an oversized double array, at least 3 long,
     into which the callee will write in the {dx, dy, dz} of the 
     coarse mesh.

     @param domain_lo_valid : an oversized array, at least 3 long, into
     which the lower bounding box corner indices of the entire domain
     (and not just the piece on this processor) is written in by the callee.
     These indices exclude the ghost/halo cells

     @param domain_hi_valid : same as above, but for the upper bounding
     box corner

     @param time : This data dump corresponds to time = ?

     @param dtCoarse : And the dt on the coarse mesh at thie momemt was ?
     
  */
  //@{
  void  AMR_ReadMetaData(int hdf5, int *numLevels, int *ncomp, 
			int *ndim, double *spacingCoarse, 
			int *domain_lo_valid, int * domain_hi_valid,
			double *time, double* dtCoarse) ;

  
  /**
     Now that the caller know how many vars per cell exist, he 
     can allocate an array of pointers to char  (array length =
     ncomp, as obtained above). The caller then sends them in,
     where the callee mallocs memory to each element of the array
     and writes in the variable names. The caller frees the memory.
     @param hdf5 : file handle
     @param names : pointer to the array of pointers that the caller 
     allocates.
  */
  void AMR_ReadCompName(int hdf5, char **names) ;

  /**
     The caller now knows the number of levels but needs to know the
     number of patches on each level. This method is called in a 
     loop over numLevels, obtained above.

     @param hdf5 : file handle
     @param level : give the caller info regarding this level
     @param ntotalBoxes : an integer, sent in by the caller, into which 
     the callee fills in the total number of patches, on level "level" 
     across all processors.
  */
  void AMR_ReadNumBoxes(int hdf5, int level, int *ntotalBoxes) ;

  /**

  Caller allocates space to store arrays of data for the region definitions
  of all the patches and then makes a call to read them.  Every processor that
  makes this call will receive a full list of box definitions.  the int arrays
  are assumed to be allocated to hold ntotalBoxes of integers.
  */

  void AMR_ReadBoxes(int hdf5, int level, 
					 int *loValid, int* hiValid, 
					 int* loRange, int* hiRange) ;
  
  /**
     Now that the caller knows the total number of level, he allocates
     an int array, numLevels long into which the refinement factor for
     each level will be written.

     @param hdf5 : file handle

     @param refs : an int array, numLevels long

     NOTE : refinement factor for the level denotes how coarse it is
     w.r.t to the immediately finer level. The finest level will have its
     ref set to 1.
  */
  void AMR_ReadRefFactors(int hdf5, int *refs) ;
 

  /**
     At this point the caller has enough info to create Levels as per 
     the methods in AMRData.h. Once these empty levels have been created,
     they can be filled up from the file. This method is called in a loop,
     looping over all the levels

     @param hdf5 : file handle
     @param level : The caller needs data on this level
     @param levelDatahandle : used by the callee to refer back to his
     data structures
	 @param compBegin: at which component in the file reading should begin
	 @param ncomp: number of contiguous components to be read, 
	 starting at compBegin.
	 it is assumed that the user is passing in a levelDatahandle 
	 that has sufficient
	 storage allocated for this amount of data.
  */
  void AMR_ReadLevel(int hdf5, int level, int levelDatahandle, int compBegin, int ncomp) ;

  /** This method provides the ability to read data in a controled manner.
	  This is useful for navigating a large data file on a computer that
	  cannot read the entire data set.
      @param hdf5 : file pointer
      @param level :  which level
      @param patch : patch of interest
      @param comp  : which variable is the caller interested in.
      @param d_ptr : start putting in data here.  User is responsible for ensuring the
	  memory address pointed to by d_ptr has sufficient space to hold the read data.
  */
  void AMR_ReadPatch(int hdf5, int level, int patch, int comp,
		    void *d_ptr) ;

  /* 
	 @param hdf5: file handle
	 @param name: attribute name
	 @param value: value of attribute.  pointer to null pointer if not found.
  */
  void AMR_ReadStringAttribute(int hdf5, char* name, char** value);

  /*@}*/

  /**
	 \name Fortran binding for AMR-HDF5 */
  /*@{*/

  /*============= Fortran bindings for AMR-HDF5 functions===========*/  

  /**
    amropenhdf5_  takes is invoked in Fortran as :
    <PRE>
	INTEGER communicator, m, filehandle
	CHARACTER(LEN=12):: filename = 'output.hdf5'
	filename(12) = CHAR(0)
	CALL amropenhdf5(filename, communicator, m, filehandle)
    </PRE>
	This works for some fortran compilers.  Fortran compilers automatically
	add the 'length' argument for you. I'm not an expert in 
	passing strings between fortran and C. the c implmentation is
	expecting a null terminated string.  If somone wants to help
	debug the fortran binding, that would be nice.

	This applies for the component name writing routines as well
  */
  void amropenhdf5_(char *filename, long int length, 
					int* MPI_Communicator, int* m, int *hdf5) ;

  void amrclosehdf5_( int* hdf5 ) ;

  /** 
	  Fortran wrapper for AMR_WriteData, except it does not pass in an array
	  of strings for component names.  passing strings from Fortan and C is
	  tricky.  Other functions handle the work of IO for fortran component names
	  @see AMR_WriteData
	  @see amrwritecompname_
	  @see amrreadcompname_
  */
  void amrwritedata_(int* hdf5, int* numLevels, int *levelData, int* ncomp,
		    int *refineFactor, double *spacing,
		    double* time, double* dt, int *domain_lo, int *domain_hi) ;

  /**
	 @see amropenhdf5_
  */
  void amrwritecompname_(int* hdf5, int* compIndex, char* compname, long int length);
 
  void amrreadmetadata_(int* hdf5, int *numLevels, int *ncomp, 
			int *ndim, double *spacingCoarse, 
			int *domain_lo_valid, int * domain_hi_valid,
			double *time, double* dtCoarse) ;

 
  /**
	 @see amropenhdf5_
	 @see amrwritecompname_
  */
  void amrreadcompname_(int* hdf5, int* compIndex, char *names, long int length) ;

  void amrreadnumboxes_(int* hdf5, int* level, int *ntotalBoxes) ;

  void amrreadboxes_(int* hdf5, int* level, 
					 int *loValid, int* hiValid, 
					 int* loRange, int* hiRange) ;
  

  void amrreadreffactors_(int* hdf5, int *refs) ;

  void amrreadlevel_(int* hdf5, int* level, int* levelDatahandle, int* compBegin, int* ncomp) ;


  void amrreadpatch_(int* hdf5, int* level, int* patch, int* comp,
		    void *d_ptr) ;
  /*@}*/
 

#if defined(__cplusplus)
}
#endif

#endif /*AMR_HDF5_h*/
     


