
#include "AMRData.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

typedef struct Patch{
  struct Patch* next;
  int valid[6];
  int range[6];
  void* dataPtr;
} Patch ;

typedef struct Level {
  int ID;
  struct Level* next;
  struct Level* prev;
  struct Patch* patches;
  struct Patch* current;
  int currentIndex;
  int components;
  int dimension;
  int centering;
  int type;
  int ordering;
  int count;
  int closed;
  int MPIcomm;
} Level ;

Level* AMR_levels = NULL;
int nextID = 1002;

void defaultErrorHandler(char* message)
{
  fwrite(message, strlen(message), 1, stderr);
  abort();
}

void (*err)(char*) = &defaultErrorHandler;

void AMR_setErrorHandler(void (*e)(char*))
{
  if(e == NULL) 
	err = &defaultErrorHandler;
  else
	err = e;
}

void (*AMR_getErrorHandler())(char*)
{
  return err;
}

void amrcreate_(int* MPI_communicator,
                int* components,
                int* dimension,
                int* centering,
                int* type,
                int* ordering,
                int* levelhandle)
{
  AMR_Create(*MPI_communicator,
			 *components,
			 *dimension,
			 *centering,
			 *type,
			 *ordering,
			 levelhandle);
}

void AMR_Create(int MPI_communicator,
                int components,
                int dimension,
                int centering,
                int type,
                int ordering,
                int* handle)
{
  /*  for thread-safe code, I would put a lock on the AMR_levels pointer */
  Level* newLevel = NULL;

  if(AMR_levels == NULL){
	AMR_levels = (Level*)malloc(sizeof(Level));
	newLevel = AMR_levels;
	newLevel->next = NULL;
	newLevel->prev = NULL;
  } 
  else {
	newLevel = (Level*)malloc(sizeof(Level));
	newLevel->next = AMR_levels->next;
	AMR_levels->next = newLevel;
	newLevel->prev = AMR_levels;
  }

  newLevel->ID = nextID;
  *handle = nextID;
  nextID++;
  newLevel->patches = NULL;
  newLevel->current = NULL;
  newLevel->currentIndex = -1;
  newLevel->components = components;
  newLevel->type = type;
  newLevel->dimension = dimension;
  newLevel->centering = centering;
  newLevel->ordering  = ordering;
  newLevel->count     = 0;
  newLevel->closed    = 0;
  newLevel->MPIcomm   = MPI_communicator;
}

Level* getLevel(int handle)
{
  Level*  t = AMR_levels;

  if(AMR_levels == NULL){
	err("AMRData: handle used before any AMR data created");

  }

  while(t != NULL){
	if(t->ID == handle) return t;
	t = t->next;
  }
  err("AMRData: handle not found");
  return t;
}

void amrclose_(int* levelhandle)
{
  AMR_Close(*levelhandle);
}


void amrdelete_(int* levelhandle)
{
  AMR_Delete(*levelhandle);
}

void AMR_Close(int handle)
{
  Level* level = getLevel(handle);
  level->closed = 1;
}

/** AMR_add* functions can no longer be called on this AMR object */
void AMR_Delete(int handle)
{
  Level* level = getLevel(handle);
  Patch* p;
  Patch* pp;
  if(level->prev == NULL){
	if(level->next != NULL){
	  AMR_levels = level->next;
	  AMR_levels->prev = level->prev;
	}
	else{
	  AMR_levels = NULL; /* all empty now */
	}
  }
  else {
	level->prev->next = level->next;
  }
  p = level->patches;
  while(p != NULL)
	{
	  pp = p->next;
	  free(p);
	  p = pp;
	}
  free(level);
}

/* for a threadsafe version of this interface, it would be sufficient to
  lock on 'level' I think, after the 'getHandle' call, and then track
  exit points for the lock release */
#define ADDPATCH(TYPE, TYPENAME, TYPEID)                          \
void AMR_AddPatch ## TYPENAME (   int handle,                     \
                        int* loRange, int* hiRange,               \
                        int* loValid, int* hiValid,               \
                        TYPE * dataPtr)                           \
{ \
  Level* level = getLevel(handle);                                \
  Patch* p     = (Patch*)malloc(sizeof(Patch));                   \
  int dim = level->dimension, i;                                  \
                                                                  \
  if(level->type != TYPEID){                                      \
	err("AMRData: calling addPatch* with wrong function-type");   \
  }                                                               \
  if(level->closed != 0){ err("AMRData: calling addPatch* with closed level");}\
  p->dataPtr = dataPtr;                                           \
  for(i=0; i<dim; ++i)                                            \
	{                                                             \
	  p->range[i]   = loRange[i];                                 \
	  p->range[i+3] = hiRange[i];                                 \
	  p->valid[i]   = loValid[i];                                 \
	  p->valid[i+3] = hiValid[i];                                 \
	}                                                             \
  p->next = level->patches;                                       \
  level->patches = p;                                             \
  level->count++;                                                 \
}                                                                 \
void amraddpatch ## TYPE ## _( int* handle,                       \
                        int* loRange, int* hiRange,               \
                        int* loValid, int* hiValid,               \
                        TYPE * dataPtr)                           \
{                                                                 \
  AMR_AddPatch ## TYPENAME (*handle, loRange, hiRange, loValid, hiValid, dataPtr); \
}

ADDPATCH(int, Int, 0)  

ADDPATCH(float, Float, 2)

ADDPATCH(double, Double, 1)
	 
ADDPATCH(long, Long, 3)

void amrgetinfo_(int* levelhandle,
                 int*  MPI_communicator,
                 int*  nComponents,
                 int*  dimension,
                 int*  centering,
                 int*  type,
                 int*  ordering,
                 int*  localcount)
{
  AMR_GetInfo(*levelhandle,
			  MPI_communicator,
			  nComponents,
			  dimension,
			  centering,
			  type,
			  ordering,
			  localcount);
}

void AMR_GetInfo(int handle,
				 int* MPI_c,
                 int* components,
                 int* dimension,
                 int* centering,
                 int* type,
                 int* ordering,
				 int* count)
{
  Level* level = getLevel(handle);
  if(level->closed == 0)
	{ err("AMRData: calling GetPatchInfo with unclosed level");}
  *components = level->components;
  *dimension  = level->dimension;
  *centering  = level->centering;
  *type       = level->type;
  *ordering   = level->ordering;
  *count      = level->count;
  *MPI_c      = level->MPIcomm;
}
void amrgetpatchinfo_(   int* levelhandle, int* patch,
                         int*  loRange, int*  hiRange,
                         int*  loValid, int*  hiValid)
{
   AMR_GetPatchInfo(*levelhandle, *patch,
					loRange, hiRange,
					loValid, hiValid);
}

void AMR_GetPatchInfo(   int levelhandle, int patch,
                         int* loRange, int* hiRange,
                         int* loValid, int* hiValid)
{
  int i;  int* range; int* valid;
  Level* level = getLevel(levelhandle);
  if(level->closed == 0)
	{ err("AMRData: calling GetPatchInfo with unclosed level");}
  if(level->count == 0) err("count == 0");
  if(patch >= level->count) err("patch >= level->count");
  if(patch != level->currentIndex)
	{
	  level->current = level->patches;
	  for(i=0; i<(level->count - patch-1); i++){
		level->current = level->current->next;
		if(level->current == NULL) err("ran out of Patches?");
	  }
	  level->currentIndex = patch;
	}
  range = level->current->range;
  valid = level->current->valid;
  loRange[0] = range[0];
  loRange[1] = range[1];
  loRange[2] = range[2];
  hiRange[0] = range[3];
  hiRange[1] = range[4];
  hiRange[2] = range[5];

  loValid[0] = valid[0];
  loValid[1] = valid[1];
  loValid[2] = valid[2];
  hiValid[0] = valid[3];
  hiValid[1] = valid[4];
  hiValid[2] = valid[5];
}

#define GETPATCH(TYPE, TYPENAME, TYPEID)                          \
void AMR_GetPatch ## TYPENAME ( int handle, int patch, TYPE ** dataPtr) \
{ \
  int i;                                                         \
  Level* level = getLevel(handle);                                \
  if(level->type != TYPEID){                                      \
	err("AMRData: calling GetPatch* with wrong data-type");   \
  }                                                               \
  if(level->count == 0) err("count == 0");                       \
  if(level->closed == 0){ err("AMRData: calling GetPatch* with unclosed level");}\
  if(patch >= level->count) err("patch >= level->count");        \
  if(patch != level->currentIndex)                               \
	{                                                             \
	  level->current = level->patches;                            \
	  for(i=0; i<(level->count - patch-1); i++){                    \
		level->current = level->current->next;                    \
		if(level->current == NULL) err("ran out of Patches?");    \
	  }                                                           \
	  level->currentIndex = patch;                                \
	}                                                             \
  *dataPtr = (TYPE*)(level->current->dataPtr);                     \
}                                                                 \
void amrgetpatch ## TYPE ## _ (int* handle, int* patch, TYPE * data) { \
 AMR_GetPatch ## TYPENAME (*handle, *patch, &data);  }                \

GETPATCH(int, Int, 0)  

GETPATCH(float, Float, 2)

GETPATCH(double, Double, 1)
	 
GETPATCH(long, Long, 3)


