
/*

Test program for AMR-HDF5.cpp, the HDF5 file I/O server interface.

*/


#include "AMR-HDF5.h"
#include "hdf5.h"

#include <iostream>
#include <map>
#include <cstdio>

#include <cassert>
#include <cstring>
#include <string>
#include <vector>
#include <list>
#include <set>
#include <cmath>
#include <typeinfo>
#include <fstream>
#include <algorithm>
#include <cstdlib>
#include <iomanip>
#include <csignal>

#ifdef MPI
#include "mpi.h"
#endif

#define HDF5

#ifndef CH_SPACEDIM
#define CH_SPACEDIM 3 // available options are 1, 2, or 3
#endif

#ifndef CH_USE_FLOAT
#define CH_USE_DOUBLE // options are CH_USE_DOULBE, or CH_USE_FLOAT  for representing 
                      //  floating point values. Currently the code has to be compiled
                      //  for one or the other, as a side effect of using Chombo for the
                      //  I/O
#endif

#include "CFile.h"


void setCircleTags(IntVectSet& ivs, int circleR, int thickness = 1,
              const IntVect& center= IntVect::TheZeroVector());
void
buildDisjointBoxLayout(DisjointBoxLayout& plan, 
                       const IntVectSet& tags, const Box& domain);


struct values
{

  static void setVal3(const Box& box, int comps, FArrayBox& fab)
    {
      int center = (box.smallEnd()[1] + box.bigEnd()[1])/2;
      for(int c=0; c<comps; ++c)
                {
                  for(BoxIterator bit(box); bit.ok(); ++bit)
                        {
                          fab(bit(), c) = c + center + bit()[0];
                        }
                }
    }


};

int main(int argc, char** argv)
{
  int comm = 0;
#ifdef MPI
  MPI_Init(&argc, &argv);
  comm = MPI_Comm_c2f(Chombo_MPI::comm);
#endif

  Box domain(IntVect::TheZeroVector(), 20*IntVect::TheUnitVector());

  DisjointBoxLayout tplan, plan0, plan1;
  {
    IntVectSet tags;

    IntVect center = 10*IntVect::TheUnitVector();

    setCircleTags(tags, 6, 2, center); //circle/sphere

    buildDisjointBoxLayout(plan0, tags, domain);
	

    tags.makeEmpty();

	center*=2;

	domain.refine(2);

    setCircleTags(tags, 12, 2, center);

    //buildDisjointBoxLayout(plan1, tags, domain);
	refine(plan1, plan0, 2);

	domain.coarsen(2);
  }

  LevelData<FArrayBox> level0(plan0, 3);
  DataIterator dit0 = level0.dataIterator();

  LevelData<FArrayBox> level1(plan1, 3);
  DataIterator dit1 = level1.dataIterator();

  level0.setVal(values::setVal3);
	  
  level1.setVal(values::setVal3);
	

#ifdef CH_USE_DOUBLE
#define addP AMR_AddPatchDouble
  int dtype = 1;
#else
#define addP AMR_AddPatchFloat
  int dtype = 2;
#endif

  int level0_handle;
  AMR_Create(comm,//int MPI_communicator,
             3,//   int components,
             CH_SPACEDIM,//   int dimension,
             0,//   int centering,
             dtype,//   int type,
             0,//   int ordering,
             &level0_handle);

  int level1_handle;
  AMR_Create(comm,//int MPI_communicator,
             3,//   int components,
             CH_SPACEDIM,//   int dimension,
             0,//   int centering,
             dtype,//   int type,
             0,//   int ordering,
             &level1_handle);

  for(dit0.begin(); dit0.ok(); ++dit0)
	{
	  const Box& b = level0.box(dit0());
	  
	  addP 
		(level0_handle, 
		 b.loVect(),
		 b.hiVect(),
		 b.loVect(),
		 b.hiVect(),
		 level0[dit0()].dataPtr());
	}

 for(dit1.begin(); dit1.ok(); ++dit1)
	{
	  const Box& b = level1.box(dit1());
	  
	  addP 
		(level1_handle, 
		 b.loVect(),
		 b.hiVect(),
		 b.loVect(),
		 b.hiVect(),
		 level1[dit1()].dataPtr());
	}

 AMR_Close(level0_handle);
 AMR_Close(level1_handle);

 int levels[2] = {level0_handle, level1_handle};
 char*  compNames[3] = {"density", "pressure", "gamma"};
 int refineFactor[2] = {2, -1};
 double spacing[CH_SPACEDIM] = {D_DECL(0.250, 0.250, 0.250)};

 int file;

 AMR_OpenHDF5("test.hdf5", comm, 0, &file);
 AMR_WriteStringAttribute(file, "date", "04-10-2003");

 AMR_WriteData(file, 2, levels, 3,
	    	   compNames, refineFactor, spacing,
			   0.0023, 0.000034, domain.loVect(), domain.hiVect());

 AMR_WriteStringAttribute(file, "prgname", argv[0]);

 AMR_CloseHDF5(file);

 AMR_Delete(level0_handle);
 AMR_Delete(level1_handle);

 /*=================================================*/


 AMR_OpenHDF5("test.hdf5", comm, 1, &file);

 int numlevels, ncomp, ndim;
 double spacingCoarse[CH_SPACEDIM], time, dtCoarse;
 int lo[3], hi[3];

 AMR_ReadMetaData(file, &numlevels, &ncomp, 
				  &ndim, spacingCoarse, 
				  lo, hi,
				  &time, &dtCoarse);
 Box rdomain;
 {
   IntVect l(lo), h(hi);
   rdomain.define(l, h);
 }

 if(domain != rdomain || numlevels!= 2 || ncomp != 3 || dtCoarse != 0.000034)
   MayDay::Error("Error reading meta data back in");

 char*  rcompNames[3];
 int    refs[2];
 AMR_ReadCompName(file, rcompNames) ;
 for(int i=0; i<ncomp; i++)
   {
	 if(strcmp(rcompNames[i], compNames[i]) != 0)
	   MayDay::Error("Error reading component name");
   }

 AMR_ReadRefFactors(file, refs) ;
 if(refs[0] != 2 || refs[1] != 1)
   MayDay::Error("Error reading in refinement factors");


 int numboxes[2];
 AMR_ReadNumBoxes(file, 0, numboxes);
 AMR_ReadNumBoxes(file, 1, numboxes+1);

 int c;
 if(numboxes[0] >= numboxes[1])
   c = numboxes[0]*CH_SPACEDIM;
 else
   c = numboxes[1]*CH_SPACEDIM;

 int* loRange = new int [4*c];
 int* hiRange = loRange + c;
 int* loValid = hiRange + c;
 int* hiValid = loValid + c;

 AMR_ReadBoxes(file, 0, 
			   loValid, hiValid, 
			   loRange, hiRange);

 Vector<Box> boxes_0(numboxes[0]);
 for(int i=0; i<numboxes[0]; ++i)
   {
	 Box& b = boxes_0[i];
	 IntVect lo(loValid);
	 IntVect hi(hiValid);
	 loValid+=CH_SPACEDIM;
	 hiValid+=CH_SPACEDIM;
	 b = Box(lo, hi);
   }


 Vector<int> procIDs;
 LoadBalance(procIDs, boxes_0);
 DisjointBoxLayout plan0_in(boxes_0, procIDs);
 plan0_in.close();
 LevelData<FArrayBox> level0_in(plan0_in, 3);
 DataIterator dit0_in = level0_in.dataIterator();
 AMR_Create(comm,//int MPI_communicator,
			3,//   int components,
			CH_SPACEDIM,//   int dimension,
			0,//   int centering,
			dtype,//   int type,
			0,//   int ordering,
			&level0_handle);
 for(dit0_in.begin(); dit0_in.ok(); ++dit0_in)
   {
	 const Box& b = level0_in.box(dit0_in());
	 
	 addP 
	   (level0_handle, 
		b.loVect(),
		b.hiVect(),
		b.loVect(),
		b.hiVect(),
		level0_in[dit0_in()].dataPtr());
   }
 AMR_Close(level0_handle);
 

 AMR_ReadLevel(file, 0, level0_handle, 0, ncomp) ;

 // in parallel this has to wait for us to settle on ordering issues.
#ifndef MPI
 dit0.begin(); dit0_in.begin();
 for(; dit0.ok(); ++dit0, ++dit0_in)
   {
	 FArrayBox dif(level0.box(dit0()), ncomp);
	 dif.copy(level0[dit0()]);
	 dif.minus(level0_in[dit0_in()]);
	 Real norm = dif.norm(0);
	 if(norm != 0){
	   MayDay::Error("Error reading data back in");
	 }
   }
#endif

 /*  next level */
 hiRange = loRange + c;
 loValid = hiRange + c;
 hiValid = loValid + c;

 AMR_ReadBoxes(file, 1, 
			   loValid, hiValid, 
			   loRange, hiRange);

 Vector<Box> boxes_1(numboxes[1]);
 for(int i=0; i<numboxes[1]; ++i)
   {
	 Box& b = boxes_1[i];
	 IntVect lo(loValid);
	 IntVect hi(hiValid);
	 loValid+=CH_SPACEDIM;
	 hiValid+=CH_SPACEDIM;
	 b = Box(lo, hi);
   }


 LoadBalance(procIDs, boxes_1);
 DisjointBoxLayout plan1_in(boxes_1, procIDs);
 plan1_in.close();
 LevelData<FArrayBox> level1_in(plan1_in, 3);
 DataIterator dit1_in = level1_in.dataIterator();
 AMR_Create(comm,//int MPI_communicator,
			3,//   int components,
			CH_SPACEDIM,//   int dimension,
			0,//   int centering,
			dtype,//   int type,
			0,//   int ordering,
			&level1_handle);
 for(dit1_in.begin(); dit1_in.ok(); ++dit1_in)
   {
	 const Box& b = level1_in.box(dit1_in());
	 
	 addP 
	   (level1_handle, 
		b.loVect(),
		b.hiVect(),
		b.loVect(),
		b.hiVect(),
		level1_in[dit1_in()].dataPtr());
   }
 AMR_Close(level1_handle);
 
 AMR_ReadLevel(file, 1, level1_handle, 0, ncomp) ;

#ifndef MPI
 dit1.begin(); dit1_in.begin();
 for(; dit1.ok(); ++dit1, ++dit1_in)
   {
	 FArrayBox dif(level1.box(dit1()), ncomp);
	 dif.copy(level1[dit1()]);
	 dif.minus(level1_in[dit1_in()]);
	 Real norm = dif.norm(0);
	 if(norm != 0){
	   MayDay::Error("Error reading data back in");
	 }
   }
#endif
 char* att[2];

 AMR_ReadStringAttribute(file, "prgname", att);
 AMR_ReadStringAttribute(file, "date", att+1);
 
 if(strcmp(att[0], argv[0]) !=0 || strcmp(att[1], "04-10-2003"))
   MayDay::Error("Error reading back string attributes");

 free(att[0]); free(att[1]);
 delete [] loRange;

 AMR_CloseHDF5(file);

 AMR_Delete(level0_handle);
 AMR_Delete(level1_handle);

#ifdef MPI
  MPI_Finalize();
#endif
  
}




void
setCircleTags(IntVectSet& ivs, int circleR, int thickness,
              const IntVect& center)
{
  assert(circleR > 0);
  assert(thickness < circleR && thickness > 0);
  for(int t = 0; t<thickness; ++t){
    for(int x=-circleR-t; x<=circleR+t; ++x)
    {
      // not perfect, but I'm not fussy here
      int Y = (int)sqrt((Real)abs(circleR*circleR - x*x)); 
#if CH_SPACEDIM == 2
      // circle
      ivs |= center+IntVect(x,Y);
      ivs |= center+IntVect(x,-Y);
#elif CH_SPACEDIM == 3
      //sphere
      for(int y=-Y; y<=Y; ++y)
      {
        int Z = (int)sqrt((Real)abs(circleR*circleR - x*x - y*y));
        ivs |= center+IntVect(x,y,Z);
        ivs |= center+IntVect(x,y,-Z);
      }
#else
#error compile failed: implemented only for 2D and 3D
#endif
    }
  }
}


void
buildDisjointBoxLayout(DisjointBoxLayout& plan, 
                       const IntVectSet& tags, const Box& domain)
{
  Vector<Box> vectBox;
  Vector<int> assignments;

  IntVectSet domainivs(domain);
  int maxsize = 128;
  Vector<int> fakeNRef(2,2);
  Real fillRatio = 0.5;
  int bufferSize = 0;
  int blockFactor = 1;
  BRMeshRefine mr(domain, fakeNRef, fillRatio, blockFactor, 
                  bufferSize, maxsize);

  mr.makeBoxes(vectBox, tags, domainivs, domain, maxsize);

  vectBox.sort();
  int stat = LoadBalance(assignments,  vectBox);
  if( stat != 0 ) {
    exit(stat) ;
  }

  plan.define(vectBox, assignments);

  plan.close();
};
