#ifndef CH_SPACE_H
#define CH_SPACE_H

#ifdef CH_LANG_CC
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

//
// $Id: CFile.cpp,v 1.6 2003/04/21 22:08:00 bvs Exp $
//
#endif

#if ! defined(CH_SPACEDIM)
#error CH_SPACEDIM must be defined
#endif

#if (CH_SPACEDIM != 1 && CH_SPACEDIM !=2 && CH_SPACEDIM != 3)
#error CH_SPACEDIM must be either 1, 2, or 3
#endif

#if !defined(CH_LANG_FORT) || defined(CH_LANG_CC)

/*@ManDoc:
  The SpaceDim C++ integer constant specifies the dimension of the
  Chombo library to be built.  It is initialized to the value of
  the macro CH\_SPACEDIM.  Hence it will have one of the values 1, 2, or 3.

  C++ code that depends on the dimensionality of the problem that
  you're trying to solve should be cast in terms of this constant,
  instead of the CH\_SPACEDIM macro, if possible, as constants are
  visible to debuggers while macros are not.
*/
   const int SpaceDim = CH_SPACEDIM;
#endif

#if CH_SPACEDIM==1

/*@ManDoc:
  The D\_EXPR(a,b,c) macro expands to a comma expression that will
  evaluate SpaceDim of its arguments, counting from the left; i.e.
  CH\_SPACEDIM==1 implies only `a' is evaluated, CH\_SPACEDIM==2
  implies both `a' and `b', and CH\_SPACEDIM==3 implies all three
  arguments are evaluated.  This macro can be used to enhance
  the portability of code that works with structures whose size
  depends on CH\_SPACEDIM.
           
  For example:

     D\_EXPR(vect[0] *= s, vect[1] *= s, vect[2] *= s);

  More explicitly, this macro evaluates to one of

     ((void)((a),0))

     ((void)((a),(b),0))

     ((void)((a),(b),(c),0))

  corresponding to CH\_SPACEDIM values of 1, 2, or 3.
*/
#  define D_EXPR(a,b,c) ((void)((a),0))

/*@ManDoc:
  The D\_DECL(a,b,c) macro expands to a comma-separated list of
  1, 2, or all 3 of the arguments of the call, depending on the
  value of CH\_SPACEDIM.  This can be used to write portable
  function calls that depend on 1, 2, or 3 arguments, corresponding
  to the value of CH\_SPACEDIM.

  For example:

     return IntVect(D\_DECL(p[0] + s, p[1] + s, p[2] + s));

  More explicitly, this macro evaluates to one of

     a

     a,b

     a,b,c

  corresponding to CH\_SPACEDIM values of 1, 2, or 3.
*/
#  define D_DECL(a,b,c) a

/*@ManDoc:
  The D\_TERM(a,b,c) macro expands to a whitespace-separated list of
  1, 2, or all 3 of the arguments of the call, depending on the value
  of CH\_SPACEDIM.  This can be used to write program logic
  that depend on 1, 2, or 3 arguments, corresponding to
  the value of CH\_SPACEDIM.

  For example:

     return D\_TERM(len[0], *len[1], *len[2]);

  More explicitly, this macro evaluates to one of

     a

     a b

     a b c

  corresponding to CH\_SPACEDIM values of 1, 2, or 3.
*/
#  define D_TERM(a,b,c) a
#elif CH_SPACEDIM==2
#  define D_EXPR(a,b,c) ((void)((a),(b),0))
#  define D_DECL(a,b,c) a,b
#  define D_TERM(a,b,c) a b
#elif CH_SPACEDIM==3
#  define D_EXPR(a,b,c) ((void)((a),(b),(c),0))
#  define D_DECL(a,b,c) a,b,c
#  define D_TERM(a,b,c) a b c
#endif

#endif /*CH_SPACE_H*/
#ifndef CH_REAL_H
#define CH_REAL_H


#ifdef CH_LANG_CC
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#endif


#ifdef CH_USE_FLOAT /* precision */

#undef CH_USE_DOUBLE
#undef CH_USE_FLOAT
#define CH_USE_FLOAT 1

#else   /* precision */     

#  undef  CH_USE_FLOAT
#  undef  CH_USE_DOUBLE
#  define CH_USE_DOUBLE 1

#endif  /* precision */

#if !defined(CH_LANG_FORT) /* language */

#ifdef CH_USE_FLOAT /* precision */

     typedef float Real;
#define  MPI_CH_REAL MPI_FLOAT
#define  H5T_NATIVE_REAL H5T_NATIVE_FLOAT

#else   /* precision */     

     typedef double Real;
#    define MPI_CH_REAL MPI_DOUBLE
#    define H5T_NATIVE_REAL H5T_NATIVE_DOUBLE

#endif  /* precision */

#else /* language */

#ifdef CH_USE_FLOAT /* precision */
#define REAL_T REAL
#else /* precision */

#if defined(CH_CRAY)
#define REAL_T REAL
#else /*cray */
#define REAL_T REAL*8
#endif /*cray*/

#endif /* precision */
#endif /* language */

#if defined(CH_LANG_FORT)
#define Real_T REAL_T
#define Real_t REAL_T
#define real_t REAL_T
#endif

#if !defined(CH_USE_FLOAT) && !defined(CH_USE_DOUBLE)
#error Exactly one of CH_USE_FLOAT or CH_USE_DOUBLE must be defined
#endif

#endif /*CH_REAL_H*/
#ifndef CH_CONSTANTS_H
#define CH_CONSTANTS_H

#ifdef CH_LANG_CC
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
/* CH_COPYRIGHT_NOTICE */

/*
** $Id: CFile.cpp,v 1.6 2003/04/21 22:08:00 bvs Exp $
*/
#endif

#if defined(CH_LANG_FORT) || defined(CH_CRAY)

/* use single-precision constants on the T3E regardless of whether
 * USE_FLOAT or USE_DOUBLE is defined
*/
#if defined(CH_USE_FLOAT)

#define zero     (0.0e0)
#define one      (1.0e0)
#define two      (2.0e0)
#define three    (3.0e0)
#define four     (4.0e0)
#define five     (5.0e0)
#define six      (6.0e0)
#define seven    (7.0e0)
#define eight    (8.0e0)
#define nine     (9.0e0)
#define ten     (10.0e0)
#define twelve  (12.0e0)
#define fifteen (15.0e0)
#define sixteen (16.0e0)
#define twenty  (20.0e0)
#define tenth   (0.100e0)
#define eighth  (0.125e0)
#define sixth   (1.000e0 /  6.000e0)
#define fourth  (0.250e0)
#define third   (1.000e0 /  3.000e0)
#define half    (0.500e0)
#define two3rd  (2.000e0 /  3.000e0)
#define Pi      (3.14159265358979323846e0)

#define ZERO     (0.0E0)
#define ONE      (1.0E0)
#define TWO      (2.0E0)
#define THREE    (3.0E0)
#define FOUR     (4.0E0)
#define FIVE     (5.0E0)
#define SIX      (6.0E0)
#define SEVEN    (7.0E0)
#define EIGHT    (8.0E0)
#define NINE     (9.0E0)
#define TEN     (10.0E0)
#define TWELVE  (12.0E0)
#define FIFTEEN (15.0E0)
#define SIXTEEN (16.0E0)
#define TWENTY  (20.0E0)
#define TENTH   (0.100E0)
#define EIGHTH  (0.125E0)
#define SIXTH   (1.000E0 / 6.000E0)
#define FOURTH  (0.250E0)
#define THIRD   (1.000E0 / 3.000E0)
#define HALF    (0.500E0)
#define TWO3RD  (2.000E0 / 3.000E0)
#define PI      (3.14159265358979323846E0)

#elif defined(CH_USE_DOUBLE)

#define zero     (0.0d0)
#define one      (1.0d0)
#define two      (2.0d0)
#define three    (3.0d0)
#define four     (4.0d0)
#define five     (5.0d0)
#define six      (6.0d0)
#define seven    (7.0d0)
#define eight    (8.0d0)
#define nine     (9.0d0)
#define ten     (10.0d0)
#define twelve  (12.0d0)
#define fifteen (15.0d0)
#define sixteen (16.0d0)
#define twenty  (20.0d0)
#define tenth   (0.100d0)
#define eighth  (0.125d0)
#define sixth   (1.000d0 / 6.000d0)
#define fourth  (0.250d0)
#define third   (1.000d0 / 3.000d0)
#define half    (0.500d0)
#define two3rd  (2.000d0 / 3.000d0)
#define Pi      (3.14159265358979323846264338327950288d0)

#define ZERO    (0.0D0)
#define ONE     (1.0D0)
#define TWO     (2.0D0)
#define THREE   (3.0D0)
#define FOUR    (4.0D0)
#define FIVE    (5.0D0)
#define SIX     (6.0D0)
#define SEVEN   (7.0D0)
#define EIGHT   (8.0D0)
#define NINE    (9.0D0)
#define TEN     (10.0D0)
#define TWELVE  (12.0D0)
#define FIFTEEN (15.0D0)
#define SIXTEEN (16.0D0)
#define TWENTY  (20.0D0)
#define TENTH   (0.100D0)
#define EIGHTH  (0.125D0)
#define SIXTH   (1.000D0 / 6.000D0)
#define FOURTH  (0.250D0)
#define THIRD   (1.000D0 / 3.000D0)
#define HALF    (0.500D0)
#define TWO3RD  (2.000D0 / 3.000D0)
#define PI      (3.14159265358979323846264338327950288D0)

#else

#error One of CH_USE_{FLOAT,DOUBLE} must be defined

#endif

#endif /*CH_LANG_FORT*/

#endif /*CH_CONSTANTS_H*/
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

// 
//  The simplest of all possible vector classes
//  dtg
//  ANAG 10-28-97
// 

#ifndef _SIMP_VECTOR_H_
#define _SIMP_VECTOR_H_


#ifdef ENABLE_MEMORY_TRACKING
//these objects are used for memory tracking system
typedef std::pair<long int*, long int*> ppair;
typedef std::map<std::string, ppair > VectorList;

typedef std::list<int> lst; //not currently used, bvs.
typedef std::map<std::string, lst*> incr; // not currently used, bvs

extern VectorList* vectorList_;
extern incr*       vectorIncr_;
#endif

using std::vector;
using std::ostream;

/// one dimensional dynamic array
/**
   Vector is a resizable one-dimensional array with constant-time
   random access and range checking.  The template type T must have a
   default constructor, a copy constructor, and an assignment
   operator.

   Vector<T> is publically derived from std::vector<T>.  Read the
   friendly STL manual.  
*/
template <class T>

class Vector
{
public:
  /// 
  /** 
      Default constructor. 
      Creates a Vector of zero length with null data.
  */
  inline
  Vector()
  {
  }

  /// 
  /** 
      Destructor. 
  */
  inline
  virtual ~Vector()
  {
#ifdef ENABLE_MEMORY_TRACKING  
    if(size()>0) decrement(size());
#endif
  }

  ///
  /**
     Copy constructor.
  */
  inline
  Vector(const Vector<T>& invec):v(invec.v)
  {
#ifdef ENABLE_MEMORY_TRACKING  
    increment(invec.size());
#endif
  }

  ///
  /**  conversion constructor
  */
  inline
  Vector(const std::vector<T>& invec):v(invec)
  {
#ifdef ENABLE_MEMORY_TRACKING  
    increment(invec.size());
#endif
  }

  ///
  /**
  */
  inline
  Vector<T>& operator=(const std::vector<T>& invec)
  {
#ifdef ENABLE_MEMORY_TRACKING  
    decrement(size());
#endif
    v=invec;
#ifdef ENABLE_MEMORY_TRACKING  
    increment(size());
#endif
    return *this;
  }

  ///
  /**
  */
  inline
  Vector<T>& operator=(const Vector<T>& invec)
  {
#ifdef ENABLE_MEMORY_TRACKING  
    decrement(size());
#endif
    v=invec.v;
#ifdef ENABLE_MEMORY_TRACKING  
    increment(size());
#endif
    return *this;
  }

  /// assign a scalar to every element of the vector
  /** [NOTE: cant use operator=() for this because it would be 
   *         ambiguous with the (int) constructor for number T types.]
  */
  inline
  Vector<T>& assign(const T& inval)
  {
    v.assign( size(),inval );
    return *this;
  }

  ///
  /** 
      Constructs a Vector with given number of elements.\\

      {\bf Arguments:} \\
      size (not modified):  number of elements of Vector to construct.\\
      {\bf This:} \\
      -------The object is modified----------

  */

  inline
  Vector(/*long*/ unsigned int isize)
  {
    resize(isize);
  }

 
  
  ///
  /**
  */
  inline
  void clear()
  {
#ifdef ENABLE_MEMORY_TRACKING  
    decrement(size());
#endif
    v.clear();
  }

  ///
  /**  size function. returns current size of Vector
  */
  inline
  size_t size() const
  {
    return v.size();
  }

  /// 
  /** 
      Constructs a Vector with given number of elements and constant value.\\

      {\bf Arguments:} \\
      size (not modified):  number of elements of Vector to construct.\\
      value (not modified):  value to set every element to.\\
      {\bf This:} \\
      -------The object is modified----------

  */
  inline
  Vector(/*long*/ unsigned int isize, const T& value) : v(isize, value)
  {
#ifdef ENABLE_MEMORY_TRACKING  
    increment(isize);
#endif
  }

  /// 
  /** 
      Returns a modifiable lvalue reference to the value of the given
      element in this Vector.  It is an error if n < 0 or n >= this->size(). \\

      {\bf Arguments: }  \\
      n (not modified) index of desired element.\\
      {\bf Returns:} \\
      modifiable reference to value in Vector at index n.\\
      {\bf This:} \\
      ----- 
      This object is modified if the returned reference is assigned a new value
      ----- 
  */
  inline
  T& operator[] (/*long*/ unsigned int n) 
  { 
    assert(n < size());
    //vector<T>* svec = static_cast<vector<T>* >(this);
    //return  svec->operator[](n);
    //    return  (static_cast<vector<T>* >(this))->operator[](n);
    return v[n];
  } 

  /// 
  /** 
      Returns a constant reference to the given element in this
      Vector.\\

      {\bf Arguments: }  \\
      n (not modified) index of desired element.\\
      {\bf Returns:} \\
      constant reference to value in Vector at index n.\\
      {\bf This:} \\
      This object is not modified.
  */
  inline
  const T& operator[] (/*long*/ unsigned int n) const
  { 
    assert(n < size());
    //const vector<T>* svec = static_cast<const vector<T>* >(this);
    //return  svec->operator[](n);
    //   return  (static_cast<const vector<T>* >(this))->operator[](n);
    return v[n];
  } 

  ///
  /**
  */
  inline 
  void push_back(const T& in)
  {
#ifdef ENABLE_MEMORY_TRACKING  
    increment(1);
#endif
    //  (static_cast<vector<T>* >(this))->push_back(in);
    v.push_back(in);
  }

  ///
  /**
     Modifies this Vector by appending the elements of the argument
     Vector.  The new Vector will have a size of this->size() +
     invec.size() (where this Vector is considered before the append is
     performed).  The first element of invec will have index
     this->size(), the second element will have index this->size()+1,
     etc.\\

     {\bf Arguments:} \\
     invec (not modified):  Vector whose elements to append to this Vector.\\
     {\bf This:} \\
     -------The object is modified----------
  */
  inline
  void  append(const Vector<T>& invec)
  {
#ifdef ENABLE_MEMORY_TRACKING  
    increment(invec.size());
#endif
    for (int ivec = 0; ivec <  invec.size(); ivec++)
      {
        // (static_cast<vector<T>* >(this))->push_back(invec[ivec]);
        v.push_back(invec[ivec]);
      }
  }

  ///
  /**
  */
  inline
  void resize(/*long*/ unsigned int isize)
  {
    //    vector<T>* svec = static_cast<vector<T>* >(this);
#ifdef ENABLE_MEMORY_TRACKING  
    if(isize > size()) increment(isize-size());
    else               decrement(size()-isize);
#endif
    unsigned int l= size();
    //    svec->resize(isize);
    v.resize(isize);
    for(; l<isize; ++l)
      // svec->operator[](l) = T();
      v[l] = T();
  }

  inline 
  void reserve(/*long*/ size_t isize)
  {
	v.reserve(isize);
  }

  inline 
  size_t capacity() const
  {
	return v.capacity();
  }
  ///
  /**
  */
  inline
  void resize(/*long*/ unsigned int isize, const T& value)
  {
    //   vector<T>* svec = static_cast<vector<T>* >(this);
#ifdef ENABLE_MEMORY_TRACKING  
    if(isize > size()) increment(isize-size());
    else               decrement(size()-isize);
#endif
    //    svec->resize(isize, value);
    v.resize(isize, value);
  }

  ///
  /**
  */
  inline void sort()
  {
    std::sort(v.begin(), v.end());
  }

#ifdef ENABLE_MEMORY_TRACKING  
  // static template memory tracking data
  static long int bytes;
  static long int peak;
  //  static lst*  incrementList;
#endif

private:

#ifdef ENABLE_MEMORY_TRACKING  
  inline
  void increment(unsigned int i)
  {
    // warning: pointless comparison of unsigned integer with zero
    // assert(i>=0);  // commented out (ndk)
    static unsigned int sizzle = initFunc();
    i*=sizzle;
    bytes += i;
    //  incrementList->insert(incrementList->begin(), i);
    if(bytes > peak) peak = bytes;
  }

  inline
  void decrement(unsigned int i)
  {
    if(i==0) return;
    //assert(bytes>=0);
    i*=sizeof(T);
    bytes-=i;
    //  incrementList->insert(incrementList->begin(), -i);
    //  assert(bytes>=0);
  }
#endif

  unsigned int initFunc();
  std::vector<T> v;
};

#ifdef ENABLE_MEMORY_TRACKING  
template <class T>
long int Vector<T>::bytes=0;

template <class T>
long int Vector<T>::peak=0;

//template <class T>
//lst*  Vector<T>::incrementList=NULL;
#endif

//class ostream;

template <class T>
ostream& operator<<(ostream& os, const Vector<T>& vec)
{
  for(int i=0; i<vec.size(); i++) os<<vec[i]<<" ";
  return os;
}

#ifdef ENABLE_MEMORY_TRACKING  

template <class T>
unsigned int  Vector<T>::initFunc()
{
  if(vectorList_ == NULL)
  {
    vectorList_ = new VectorList;
    vectorIncr_ = new incr;
  }
  //  if(incrementList == NULL) incrementList = new lst;
  ppair tmp(&bytes, &peak);
  vectorList_->operator[](typeid(T).name()) = tmp;
  //  vectorIncr_->operator[](typeid(T).name()) = incrementList;
  //  bytes = 0;
  //  peak = 0;
  return sizeof(T);
}

#endif
#endif

/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _LOHISIDE_H_
#define _LOHISIDE_H_

/// encapsulation of high or low side 
/**
   Encapsulation of high or low side.
*/
class Side
{
public:
  ///
  /**
     Lo is the low side \\
     Hi is the high side \\
  */
  enum LoHiSide {Invalid=-1, Lo=0, Hi, NUMSIDES};
  Side() {};
  ~Side() {};
  ///
  /**
     Returns the other side.  I.e.  swap Hi for Lo and vice versa.
  */
  friend 
  Side::LoHiSide 
  flip(Side::LoHiSide side);

  ///
  static 
  Side::LoHiSide 
  flip(Side::LoHiSide side);

  ///
  /**
     Returns -1 for Lo side, +1 for Hi Side.
  */
  friend
  int 
  sign(Side::LoHiSide side);


};



/// Iterator for low and high side
/**
   SideIterator iterates over the values Side::Lo and Side::Hi.  
   There are no adjustable parameters whatsoever.
*/
class SideIterator
{
public:

  ///{\bf Constructors, destructors}

  ///
  /** 
      Default, and only, constructor.
  */
  SideIterator();

  ///
  /**
     Default destructor.
  */
  ~SideIterator() {};

  ///{\bf Modification functions}

  ///
  /**
     Set the iterator to the first side. \\

     {\bf Arguments:} none. \\
     {\bf Returns:} none. \\
     {\bf This:} \\
     This object is modified.
  */
  void begin();

  ///
  /**
     Set the iterator to the first side. \\

     {\bf Arguments:} none. \\
     {\bf Returns:} none. \\
     {\bf This:} \\
     This object is modified.
  */
  void reset()
  {begin();}

  void next();

  ///
  /**
     Advances the iterator to the next face. \\

     {\bf Arguments:} none. \\
     {\bf Returns:} none. \\
     {\bf This:} \\
     This object is modified.
  */
  void operator++();

  ///{\bf Access functions}
   
  ///
  /**
     Returns the value of the side for the current iteration. \\

     {\bf Arguments:} none. \\
     {\bf Returns:} \\
     Returns the value of the side for the current iteration. \\
     {\bf This:} \\
     This object is not modified.
  */
  Side::LoHiSide operator() () const;

  ///
  /**
     Checks to see if there is a next side. \\

     {\bf Arguments:} none. \\
     {\bf Returns:} \\
     true if there is a next side. \\
     false otherwise. \\
     {\bf This:} \\
     This object is not modified.
  */
  bool ok() const;

private:
  int current;
};

#endif
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//


Side::LoHiSide
flip(Side::LoHiSide side) 
{
    assert ( (side == Side::Lo) || (side == Side::Hi) );
    return ( (side == Side::Lo)
             ? Side::Hi
             : Side::Lo );
}

Side::LoHiSide
Side::flip(Side::LoHiSide side) 
{
    assert ( (side == Side::Lo) || (side == Side::Hi) );
    return ( (side == Side::Lo)
             ? Side::Hi
             : Side::Lo );
}


int 
sign(Side::LoHiSide side)
{
    assert ( (side == Side::Lo) || (side == Side::Hi) );
    return ( (side == Side::Lo) ? -1 : 1 );
}



SideIterator::SideIterator() : current(-1)
{    
  reset();
}


void
SideIterator::begin()
{
    current = 0;
}


bool 
SideIterator::ok() const
{
    return ( (current > -1) && (current < Side::NUMSIDES) );
}


void
SideIterator::operator++()
{
    ++current;
}

void
SideIterator::next()
{
    ++current;
}


Side::LoHiSide
SideIterator::operator() () const
{
    switch (current)
    {
    case 0:
        return Side::Lo;
    case 1:
        return Side::Hi;
    default:
        return Side::Invalid;
    }
}
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _PARSTREAM_H_
#define _PARSTREAM_H_


std::ostream& pout();

#endif // _PARSTREAM_H_
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef CH_MayDay_H
#define CH_MayDay_H

//
// $Id: CFile.cpp,v 1.6 2003/04/21 22:08:00 bvs Exp $
//

//
//@Man:
//@Memo: Error-reporting Functions and Version Number
/*@Doc:

  This class consists entirely of static error-reporting functions that
  are used throughout the MayDay implementation together with a version
  string.  In the best of all possible worlds we'd use a namespace, but
  we aren't there, yet :-)  Note that all the output functions add a
  terminating exclamation mark, so there is no need to add any punction
  to sentences that get output explicitly.
*/

class MayDay
{
protected:
  //
  // A null string used as a default argument by the following functions.
  //
  static const char* nullString;
public:
  ///
/**
  Print out message to cerr and exit via MayDay::Abort().
*/
  static void Error (const char * msg = nullString);

  ///
/**
  Print out message to cerr and exit via abort().
*/
  static void Abort (const char * msg = nullString);

  /// 
/**
   Issues warning to stderr, but does not halt program execution.
*/
  static void Warning(const char* msg = nullString);

  /* The MayDay version string.
    
    Has the form:
    
    MayDay version 2.0 built Jun 25 1996 at 14:52:36
    */
  static const char * const version;

};


#endif /*CH_MayDay_H*/
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

using std::cerr;
using namespace std;


//
// The definition of our NULL string used as default argument.
//
const char* MayDay::nullString = "";

#define bl_str(s)  # s
#define bl_xstr(s) bl_str(s)

//
// The definition of our version string.
//
const char * const MayDay::version =

"MayDay version "
bl_xstr(CH_VERSION_MAJOR)
"."
bl_xstr(CH_VERSION_MINOR)
" built "
__DATE__
" at "
__TIME__;

#undef bl_str
#undef bl_xstr


//
// This is used by MayDay::Error(), MayDay::Abort(), and MayDay::Assert()
// to ensure that when writing the message to stderr, that no additional
// heap-based memory is allocated.
//

static
void
write_to_stderr_without_buffering (const char* str)
{
    //
    // Flush all buffers.
    //
    fflush(NULL);

    if (str)
    {
        //
        // Add some `!'s and a newline to the string.
        //
        const char * const end = " !!!\n";
        fwrite(str, strlen(str), 1, stderr);
        fwrite(end, strlen(end), 1, stderr);
    }
}

void
MayDay::Error (const char* msg)
{
    write_to_stderr_without_buffering(msg);
    abort();
}

void
MayDay::Abort (const char* msg)
{
    write_to_stderr_without_buffering(msg);
    abort();
}

void
MayDay::Warning (const char* msg)
{
    write_to_stderr_without_buffering(msg);
}


/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef CH_MISC_H
#define CH_MISC_H

//
// $Id: CFile.cpp,v 1.6 2003/04/21 22:08:00 bvs Exp $
//


/*@ManDoc:
  Returns a copy of the minimum of the two values.  Will work with
  any type T that supplies a T::operator<(const T, const T), or an
  equivalent construct, as well as a copy constructor.   In
  particular, it will work with any integral or floating-point
  intrinsic type.
*/

template <class T>
inline
T
Min (const T& a,
     const T& b)
{
    return a < b ? a : b;
}

/*@ManDoc:
  Returns a copy of the maximum of the two values.  Will work with
  any type T that supplies a T::operator>(const T, const T), or an
  equivalent construct, as well as a copy constructor.   In
  particular, it will work with any integral or floating-point
  intrinsic type.
*/

template <class T>
inline
T
Max (const T& a,
     const T& b)
{
    return a > b ? a : b;
}

/*@ManDoc:
  Returns a copy of the absolute value of the value.  Will work with
  any type T that can be compared against zero with
  T::operator>(const T, const T), or an equivalent construct, as
  well as a copy constructor, and an T::operator-() or equivalent.
  In particular, it will work with any integral or floating-point
  intrinsic type.
*/

template <class T>
inline
T
Abs (const T& a)
{
    return a > 0 ? a : -a;
}

/*@ManDoc:
  Swaps the two values.  Type T must have a copy constructor and
  an assignment operator.
*/

template <class T>
inline
void
Swap (T& a, T& b)
{
    T tmp = a;
    a = b;
    b = tmp;
}

#endif /*CH_MISC_H*/
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef CH_ARENA
#define CH_ARENA

//
// $Id: CFile.cpp,v 1.6 2003/04/21 22:08:00 bvs Exp $
//


//
//@Man:
//@Memo: A Virtual Base Class for Dynamic Memory Management
/*@Doc:

  This is a virtual base class for objects that manage their own dynamic
  memory allocation.  Since it is a virtual base class, you have to derive
  something from it to use it.
*/

class Arena;
typedef std::list<Arena*> ArenaList;

class Arena
{
public:
  ///
  /** base class constructor.  registers Arena object with
      memory tracking system */
  Arena();

  ///
  /** base class destructor.  unregisters Arena object with
      memory tracking system */
  virtual ~Arena();

    /*@ManDoc: Allocate a dynamic memory arena of size sz.
               A pointer to this memory should be returned.
    */
    virtual void* alloc (size_t sz) = 0;

    /*@ManDoc: A pure virtual function for deleting the arena pointed
               to by pt.
    */
    virtual void free (void* pt) = 0;

    /*@ManDoc: Given a minimum required arena size of sz bytes, this returns
               the next largest arena size that will hold an integral number
               of objects of the largest of the types void*, long,
               double and function pointer.
    */
    static size_t align (size_t sz);
    typedef void (*FP) ();

#ifdef ENABLE_MEMORY_TRACKING  
  // data members used by memory tracking system.
  long int bytes;
  long int peak;
  static ArenaList* arenaList_;
  char name_[120];
#endif
protected:
    //
    // Types used by align().
    //


    union Word
    {
        void*  p;
        double d;
        long   l;
        FP     f;
    };
};

//
// Inlines.
//

inline
size_t
Arena::align (size_t s)
{
    size_t x = s + sizeof(Word) - 1;
    x -= x%sizeof(Word);
    return x;
}

//
//@Man:
//@Memo: A Concrete Class for Dynamic Memory Management
/*@Doc:

  This is the simplest dynamic memory management class derived from Arena.

  Makes calls to ::operator new() and ::operator delete().
*/

class BArena
    :
    public Arena
{
public:

  BArena(const char* name = "unnamed");
    /*@ManDoc: Allocates a dynamic memory arena of size sz.  Returns a
               pointer to this memory.
    */
    virtual void* alloc (size_t sz);
    //
    //@ManDoc: Deletes the arena pointed to by pt.
    //
    virtual void free (void* pt);
};


using std::set;


//
//@Man:
//@Memo: A Concrete Class for Dynamic Memory Management
/*@Doc:

  This is a coalescing memory manager.  It allocates (possibly) large
  chunks of heap space and apportions it out as requested.  It merges
  together neighboring chunks on each free().
*/

class CArena
    :
    public Arena
{
public:

    /*@ManDoc: Construct a coalescing memory manager.  `hunk\_size' is the
               minimum size of hunks of memory to allocate from the heap.
               If hunk\_size == 0 we use DefaultHunkSize as specified below.
    */
    CArena (size_t hunk_size = 0);
    //
    //@ManDoc: The destructor.
    //
    virtual ~CArena ();
    //
    //@ManDoc: Allocate some memory.
    //
    virtual void* alloc (size_t nbytes);

    /*@ManDoc: Free up allocated memory.  Merge neighboring free memory chunks
               into largest possible chunk.
    */
    virtual void free (void* ap);

    /*@ManDoc: Mirror the C calloc() function.  Returns zero'd memory of
               size nmemb*size.  This is here so that we can implement
               malloc(3) and family.  User's shouldn't use this function.
    */
    void* calloc (size_t nmemb, size_t size);

    /*@ManDoc: Mirror the C realloc() function.  This is here so that we can
               implement malloc(3) and family.  User's shouldn't use this
               function.
    */
    void* realloc (void* ptr, size_t size);
    //
    //@ManDoc: The default memory hunk size to grab from the heap.
    //
    enum { DefaultHunkSize = 1024*1024 };

protected:
    //
    // The nodes in our free list and block list.
    //
    class Node
    {
    public:
        //
        // The default constructor.
        //
        Node ()
            :
            m_block(0), m_size(0) {}
        //
        // Another constructor.
        //
        Node (void* block, size_t size)
            :
            m_block(block), m_size(size) {}
        //
        // The copy constructor.
        //
        Node (const Node& rhs)
            :
            m_block(rhs.m_block), m_size(rhs.m_size) {}
        //
        // The copy assignment constructor.
        //
        Node& operator= (const Node& rhs)
        {
            m_block = rhs.m_block;
            m_size  = rhs.m_size;
            return *this;
        }
        //
        // The "less-than" operator.
        //
        bool operator< (const Node& rhs) const
        {
            return m_block < rhs.m_block;
        }
        //
        // The equality operator. 
        //
        bool operator== (const Node& rhs) const
        {
            return m_block == rhs.m_block;
        }
        //
        // The block address.
        //
        void* block () const { return m_block; }
        //
        // Set block address.
        //
        void block (void* blk) { m_block = blk; }
        //
        // The size of the memory block.
        //
        size_t size () const { return m_size; }
        //
        // Set size.
        //
        void size (size_t sz) { m_size = sz; }

    private:
        //
        // The block of memory we reference.
        //
        void* m_block;
        //
        // The size of the block we represent.
        //
        size_t m_size;
    };
    //
    // The type of our freelist and blocklist.
    // We use a set sorted from lo to hi memory addresses.
    //
    typedef set < Node > NL;
    //
    // The list of blocks allocated via ::operator new().
    //
    std::vector<void*> m_alloc;
    //
    // The free list of allocated but not currently used blocks.
    // Maintained in lo to hi memory sorted order.
    //
    NL m_freelist;
    //
    // The list of busy blocks.
    // A block is either on the freelist or on the blocklist, but not on both.
    //
    NL m_busylist;
    //
    // The minimal size of hunks to request via ::operator new().
    //
    size_t m_hunk;

private:
    //
    // Disallowed.
    //
    CArena (const CArena& rhs);
    CArena& operator= (const CArena& rhs);
};

//
// The Arena used by BaseFab code.
//
extern Arena* The_FAB_Arena;


#endif /*CH_ARENA*/
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//


using std::pair;
// on the usage side, so that we don't have to keep track
// of memory size and pointer lists.  memtrack is too much
// overkill for the Arena, as it can be used cleverly and well
// without any real penalty.

#ifdef ENABLE_MEMORY_TRACKING  
ArenaList* Arena::arenaList_ = NULL;
#endif

//
// Only really use the coalescing FAB arena if CH_COALESCE_FABS.
//
//#ifdef CH_COALESCE_FABS
//static CArena The_Static_FAB_CArena;
//Arena* The_FAB_Arena = &The_Static_FAB_CArena;
//#else
//static BArena The_Static_FAB_BArena;
//Arena* The_FAB_Arena = &The_Static_FAB_BArena;
//#endif


Arena::Arena()
{
#ifdef ENABLE_MEMORY_TRACKING  
  if(arenaList_ == NULL) arenaList_ = new ArenaList;
  arenaList_->insert(arenaList_->begin(), this);
  bytes = 0;
  peak = 0;
#endif
}

Arena::~Arena()
{
#ifdef ENABLE_MEMORY_TRACKING  
  arenaList_->remove(this);
#endif
}

BArena::BArena(const char* a_name)
{
#ifdef ENABLE_MEMORY_TRACKING  
  strcpy(name_, a_name);
#endif
}

void*
BArena::alloc (size_t _sz)
{
    void* ret =  malloc(_sz);
    if(ret==NULL) MayDay::Error("Out of memory in BArena::alloc (BaseFab) ");
    return ret;
}

void
BArena::free (void* pt)
{
    ::free(pt);
}
/*
** This software is copyright (C) by the Lawrence Berkeley National
** Laboratory.  Permission is granted to reproduce this software for
** non-commercial purposes provided that this notice is left intact.
**  
** It is acknowledged that the U.S. Government has rights to this
** software under Contract DE-AC03-765F00098 between the U.S.  Department
** of Energy and the University of California.
**  
** This software is provided as a professional and academic contribution
** for joint exchange. Thus it is experimental, is provided ``as is'',
** with no warranties of any kind whatsoever, no support, no promise of
** updates, or printed documentation. By using this software, you
** acknowledge that the Lawrence Berkeley National Laboratory and Regents
** of the University of California shall have no liability with respect
** to the infringement of other copyrights by any part of this software.
**  
** For further information about this software, contact:
** 
*/

//
// $Id: CFile.cpp,v 1.6 2003/04/21 22:08:00 bvs Exp $
//

CArena::CArena (size_t hunk_size)
{
    //
    // Force alignment of hunksize.
    //
    m_hunk = Arena::align(hunk_size == 0 ? DefaultHunkSize : hunk_size);

    assert(m_hunk >= hunk_size);
    assert(m_hunk%sizeof(Arena::Word) == 0);
}

CArena::~CArena ()
{
    for (unsigned int i = 0; i < m_alloc.size(); i++)
    {
        ::operator delete(m_alloc[i]);
    }
}

void*
CArena::alloc (size_t nbytes)
{
    nbytes = Arena::align(nbytes == 0 ? 1 : nbytes);
    //
    // Find node in freelist at lowest memory address that'll satisfy request.
    //
    NL::iterator free_it = m_freelist.begin();

    for ( ; free_it != m_freelist.end(); ++free_it)
    {
        if ((*free_it).size() >= nbytes)
            break;
    }

    void* vp = 0;

    if (free_it == m_freelist.end())
    {
        vp = ::operator new(nbytes < m_hunk ? m_hunk : nbytes);

        m_alloc.push_back(vp);

        if (nbytes < m_hunk)
        {
            //
            // Add leftover chunk to free list.
            //
            // Insert with a hint -- should be largest block in the set.
            //
            void* block = static_cast<char*>(vp) + nbytes;

            m_freelist.insert(m_freelist.end(), Node(block, m_hunk-nbytes));
        }
    }
    else
    {
        assert((*free_it).size() >= nbytes);

        assert(m_busylist.find(*free_it) == m_busylist.end());

        vp = (*free_it).block();

        if ((*free_it).size() > nbytes)
        {
            //
            // Insert remainder of free block back into freelist.
            //
            // Insert with a hint -- right after the current block being split.
            //
            Node freeblock = *free_it;

            freeblock.size(freeblock.size() - nbytes);

            freeblock.block(static_cast<char*>(vp) + nbytes);

            m_freelist.insert(free_it, freeblock);
        }

        m_freelist.erase(free_it);
    }

    m_busylist.insert(Node(vp, nbytes));

    assert(!(vp == 0));

    return vp;
}

void
CArena::free (void* vp)
{
    if (vp == 0)
        //
        // Allow calls with NULL as allowed by C++ delete.
        //
        return;
    //
    // `vp' had better be in the busy list.
    //
    NL::iterator busy_it = m_busylist.find(Node(vp,0));

    assert(!(busy_it == m_busylist.end()));

    assert(m_freelist.find(*busy_it) == m_freelist.end());

    void* freeblock = static_cast<char*>((*busy_it).block());
    //
    // Put free'd block on free list and save iterator to insert()ed position.
    //
    pair<NL::iterator,bool> pair_it = m_freelist.insert(*busy_it);

    assert(pair_it.second == true);

    NL::iterator free_it = pair_it.first;

    assert(free_it != m_freelist.end() && (*free_it).block() == freeblock);
    //
    // And remove from busy list.
    //
    m_busylist.erase(busy_it);
    //
    // Coalesce freeblock(s) on lo and hi side of this block.
    //
    if (!(free_it == m_freelist.begin()))
    {
        NL::iterator lo_it = free_it;
        --lo_it;
        void* addr = static_cast<char*>((*lo_it).block()) + (*lo_it).size();

        if (addr == (*free_it).block())
        {
            //
            // This cast is needed as iterators to set return const values;
            // i.e. we can't legally change an element of a set.
            // In this case I want to change the size() of a block
            // in the freelist.  Since size() is not used in the ordering
            // relations in the set, this won't effect the order;
            // i.e. it won't muck up the ordering of elements in the set.
            // I don't want to have to remove the element from the set and
            // then reinsert it with a different size() as it'll just go
            // back into the same place in the set.
            //
            Node* node = const_cast<Node*>(&(*lo_it));
            assert(!(node == 0));
            node->size((*lo_it).size() + (*free_it).size());
            m_freelist.erase(free_it);
            free_it = lo_it;
        }
    }

    NL::iterator hi_it = free_it;

    void* addr = static_cast<char*>((*free_it).block()) + (*free_it).size();

    if (++hi_it != m_freelist.end() && addr == (*hi_it).block())
    {
        //
        // Ditto the above comment.
        //
        Node* node = const_cast<Node*>(&(*free_it));
        assert(!(node == 0));
        node->size((*free_it).size() + (*hi_it).size());
        m_freelist.erase(hi_it);
    }
}

void*
CArena::calloc (size_t nmemb,
                size_t size)
{
    assert(!(size == 0));
    assert(!(nmemb == 0));

    void* vp = CArena::alloc(nmemb*size);

    memset(vp, 0, nmemb*size);

    return vp;
}

void*
CArena::realloc (void*  ptr,
                 size_t size)
{
    if (ptr == 0)
    {
        assert(!(size == 0));

        return CArena::alloc(size);
    }
    else
    {
        if (size == 0)
        {
            CArena::free(ptr);
        }
        else
        {
            //
            // It had better be in the busy list.
            //
            NL::iterator busy_it = m_busylist.find(Node(ptr,0));

            assert(!(busy_it == m_busylist.end()));

            assert(m_freelist.find(*busy_it) == m_freelist.end());

            if (size > (*busy_it).size())
            {
                //
                // If the new size is larger than old allocate new chunk.
                //
                void* vp = CArena::alloc(size);

                memcpy(vp, ptr, (*busy_it).size());

                CArena::free(ptr);

                ptr = vp;
            }
        }

        return ptr;
    }
}

/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

// RefCountedPtr.h
// ==================


/*
        A reference counting handle class.

        This is to be used as a pointer to class T.  
        This will feel and smell just like a built-in pointer except:
        1.  There is no need to call delete on the pointer.
        2.  The default copy constructor and assignment implement ref-counting.
        3.  The user may call isNonUnique to determine if this pointer is
            the only pointer to the data.  This can be used to hide the
            ref-counting behavior of a class.
        4.  Checks for dereference of a null pointer.
*/

#ifndef REFCOUNTEDPTR_H
#define REFCOUNTEDPTR_H


//
template <class T>
class RefCountedPtr {
public:
  //
  inline RefCountedPtr(T* ptr = 0);
  //
  inline RefCountedPtr(const RefCountedPtr<T>& other);
  //
  inline ~RefCountedPtr();
  //
  inline const RefCountedPtr<T>& operator =(const RefCountedPtr<T>& rhs);
  //
  inline T* operator ->();
  //
  inline bool isNull() const;
  //
  inline const T* operator ->() const;
  //
  inline T& operator *();
  //
  inline const T& operator *() const;
  //
  // auto conversion to regular const pointer where required.
  inline operator const T* () const;
  //
  inline bool isNonUnique() const;

  inline int refCount() const;

protected:
  T* ptr_;
  int* refCount_;
private:
};

template <class T> inline
RefCountedPtr<T>::RefCountedPtr(T* ptr)
  : ptr_(ptr),
    refCount_(0)
{
  if (ptr_)
    {
      refCount_ = new int;
      if (refCount_ == 0)
        MayDay::Error("RefCountedPtr::RefCountedPtr(T* ptr) out of memory");
      *refCount_ = 1;
    }
}

template <class T> inline
RefCountedPtr<T>::RefCountedPtr(const RefCountedPtr<T>& other)
  : ptr_(other.ptr_),
    refCount_(other.refCount_)
{
  if (refCount_ != 0)
    ++(*refCount_);
}

template <class T> inline
RefCountedPtr<T>::~RefCountedPtr() {
  if (refCount_ != 0 && --(*refCount_) == 0)
    {
      delete ptr_;
      ptr_ = 0;
      delete refCount_;
      refCount_ = 0;
    }
}

template <class T> inline
const RefCountedPtr<T>& RefCountedPtr<T>::operator =(const RefCountedPtr<T>& rhs) {
  if (ptr_ != rhs.ptr_)
    {
      if (refCount_ != 0 && --(*refCount_) == 0)
        {
          delete ptr_;
          delete refCount_;
        }
      ptr_ = rhs.ptr_;
      refCount_ = rhs.refCount_;
      if (refCount_ != 0)
        ++(*refCount_);
    }
  return *this;
}

template <class T> inline
bool RefCountedPtr<T>::isNull() const {
  return (ptr_ == 0);
}

template <class T> inline
T* RefCountedPtr<T>::operator ->() {
  if (ptr_ == 0)
    MayDay::Error("RefCountedPtr<T>::operator ->() on null pointer");
  return ptr_;
}

template <class T> inline
const T* RefCountedPtr<T>::operator ->() const {
  if (ptr_ == 0)
    MayDay::Error("RefCountedPtr<T>::operator ->() on null pointer");
  return ptr_;
}

template <class T> inline
T& RefCountedPtr<T>::operator *() {
  if (ptr_ == 0)
    MayDay::Error("RefCountedPtr<T>::operator *() on null pointer");
  return *ptr_;
}

template <class T> inline
const T& RefCountedPtr<T>::operator *() const {
  if (ptr_ == 0)
    MayDay::Error("RefCountedPtr<T>::operator *() on null pointer");
  return *ptr_;
}

template <class T> inline
RefCountedPtr<T>::operator const T* () const {
  return ptr_;
}

template <class T> inline
bool RefCountedPtr<T>::isNonUnique() const {
  return refCount_ == 0 ? false : *refCount_ != 1;
}

template <class T> inline
int RefCountedPtr<T>::refCount() const {
  return refCount_ == 0 ? 0 : *refCount_;
}

#endif
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef CH_INTVECT_H
#define CH_INTVECT_H

//
// $Id: CFile.cpp,v 1.6 2003/04/21 22:08:00 bvs Exp $
//

#ifndef WRAPPER


class HDF5Handle;


#endif

//
//@Man:
//@Memo: An Integer Vector in SpaceDim-Dimensional Space
/*@Doc:

  The class IntVect is an implementation of an integer vector in a
  SpaceDim-dimensional space.  It represents a point in a discrete space.
  IntVect values are accessed using the operator[] function, as for a normal
  C++ array.  In addition, the basic arithmetic operators have been overloaded
  to implement scaling and translation operations.
*/

class IntVect
{
public:

/// {\bf Constructors and Accessors}

///
/**
  Construct an IntVect whose components are uninitialized.

*/
  IntVect (){;}

///
/**
  Destructor

*/
  ~IntVect() {;}

///
  /**
     Construct an IntVect given the specific values for its
     coordinates.  D\_DECL is a macro that sets the constructor to
     take CH\_SPACEDIM arguments.  

*/
  IntVect (D_DECL(int i, int j, int k));

///
/**
  Construct an IntVect setting the coordinates to the corresponding
  values in the integer array a.  

*/
  explicit IntVect (const int* a);

  ///
  /**
     only for sending to fortran
   */
  const int*  dataPtr() const;

  ///
  /**
     only for sending to fortran
   */
  int*  dataPtr();

  ///
/**
  The copy constructor.

*/
  IntVect (const IntVect& rhs);

  IntVect copy() const {return *this;}
  ///
/**
   The assignment operator.

*/
  IntVect& operator= (const IntVect& rhs);

  ///
/**
   Returns a modifiable lvalue reference to the i'th coordinate of the
   IntVect.

*/
  inline
  int& operator[] (int i);

  ///
/**
   Returns the i'th coordinate of the IntVect.

*/
  inline
  int operator[] (int i) const;

  ///
/**
   Set i'th coordinate of IntVect to val.

*/
  void setVal (int i,
			   int val);
  
  ///
/**
   Returns a const pointer to an array of coordinates of the IntVect.
   Useful for arguments to FORTRAN calls.

*/
  const int* getVect () const;

/// {\bf Comparison Operators}

///
/**
  Returns true if this IntVect is equivalent to argument IntVect.  All
  comparisons between analogous components must be satisfied.  

*/
  bool operator== (const IntVect& p) const;

///
  /**
     Returns true if this IntVect is different from argument IntVect.
     All comparisons between analogous components must be satisfied.  

  */
  bool operator!= (const IntVect& p) const;

///
  /**
     Returns true if this IntVect is less than argument IntVect.  All
     comparisons between analogous components must be satisfied.  Note
     that, since the comparison is componentwise, it is possible for
     an IntVect to be neither greater than, less than, nor equal to
     another.  

  */
  bool operator< (const IntVect& p) const;

///
  /**
     Returns true if this IntVect is less than or equal to argument
     IntVect.  All comparisons between analogous components must be
     satisfied.  Note that, since the comparison is componentwise, it
     is possible for an IntVect to be neither greater than or equal
     to, less than or equal to, nor equal to another.  

  */
  bool operator<= (const IntVect& p) const;

///
  /**
     Returns true if this IntVect is greater than argument IntVect.
     All comparisons between analogous components must be satisfied.
     Note that, since the comparison is componentwise, it is possible
     for an IntVect to be neither greater than, less than, nor equal
     to another.  

  */
  bool operator> (const IntVect& p) const;

///
  /**
     Returns true if this IntVect is greater than or equal to argument
     IntVect.  All comparisons between analogous components must be
     satisfied.  Note that, since the comparison is componentwise, it
     is possible for an IntVect to be neither greater than or equal
     to, less than or equal to, nor equal to another.  

  */

  bool operator>= (const IntVect& p) const;

///
  /**
     Returns true if this IntVect is lexically less than the argument.
     An IntVect MUST BE either lexically less than, lexically greater
     than, or equal to another IntVect.\\

    iv1 is lexically less than iv2 if: \\

    in 2-D: \\
    (iv1[0] < iv2[0]) || ((iv1[0] == iv2[0]) && (iv1[1] < iv2[1])); \\

    in 3-D: \\
  (iv1[0] < iv2[0]) || (iv1[0]==iv2[0] && ((iv1[1] < iv2[1] || ((iv1[1] == iv2[1]) && (iv1[2] < iv2[2]))))); \\ 

  */
  bool lexLT (const IntVect& s) const;

///
  /**
     Returns true if this IntVect is lexically greater than the
     argument.  An IntVect MUST BE either lexically less than,
     lexically greater than, or equal to another IntVect.\\

    iv1 is lexically less than iv2 if: \\

    in 2-D: \\
    (iv1[0] > iv2[0]) || ((iv1[0] == iv2[0]) && (iv1[1] > iv2[1])); \\

    int 3-D: \\
  (iv1[0] > iv2[0]) || (iv1[0]==iv2[0] && ((iv1[1] > iv2[1] || ((iv1[1] == iv2[1]) && (iv1[2] > iv2[2]))))); \\ 

  */
  bool lexGT (const IntVect& s) const;

/// {\bf Arithmetic Operators}

  ///
/**
   Unary plus -- for completeness.

*/
  IntVect operator+ () const;

  ///
/**
   Unary Minus -- negates all components of this IntVect.

*/
  IntVect operator- () const;

/// addition

  ///
/**
  Modifies this IntVect by addition of a scalar to each component.

*/
  IntVect& operator+= (int s);

  ///
/**
   Modifies this IntVect by component-wise addition with argument.

*/
  IntVect& operator+= (const IntVect& p);

  ///
/**
   Returns component-wise sum of this IntVect and argument.

*/
  IntVect operator+ (const IntVect& p) const;

  ///
  /**
     Return an IntVect that is this IntVect with a scalar added to
     each component.

  */
  IntVect operator+ (int s) const;

  ///
/**
   Returns an IntVect that is an IntVect p with a scalar s added to
   each component.

*/
  friend inline IntVect operator+ (int            s,
                                   const IntVect& p);

/// subtraction

  ///
  /**
     Modifies this IntVect by subtraction of a scalar from each
     component.

  */
  IntVect& operator-= (int s);

  ///
/**
   Modifies this IntVect by component-wise subtraction with argument.

*/
  IntVect& operator-= (const IntVect& p);

///
/**
   Returns component-wise difference of this IntVect and argument.

*/
  IntVect operator- (const IntVect& p) const;

  ///
/**
   Returns an IntVect that is this IntVect with a scalar s subtracted
   from each component.

*/
  IntVect operator- (int s) const;

  ///
/**
   Returns -p + s.

*/
  friend inline IntVect operator- (int            s,
                                   const IntVect& p);

/// multiplication

  ///
/**
   Modifies this IntVect by multiplication of a scalar to each
   component.  

*/
  IntVect& operator*= (int s);

  ///
  /**
     Modifies this IntVect by component-wise multiplication with argument.

  */
  IntVect& operator*= (const IntVect& p);

  ///
  /**
     Returns component-wise product of this IntVect and argument.

  */
  IntVect operator* (const IntVect& p) const;

  ///
/**
  Returns an IntVect that is this IntVect with each component
  multiplied by a scalar.

*/
  IntVect operator* (int s) const;

  ///
/**
   Returns and IntVect that is an IntVect p with each component
   multiplied by a scalar s.

*/
  friend inline IntVect operator* (int            s,
                                   const IntVect& p);

/// division

  ///
/**
   Modifies this IntVect by division by a scalar of each component.

*/
  IntVect& operator/= (int s);

  ///
/**
  Modifies this IntVect by component-wise division with IntVect
  argument.

*/
  IntVect& operator/= (const IntVect& p);

///
/**
  Returns component-wise division of this IntVect by IntVect argument.

*/
  IntVect operator/ (const IntVect& p) const;

  ///
/**
  Returns component-wise division of this IntVect by scalar s.

*/
  IntVect operator/ (int s) const;

/// other

  ///
/**
   Modifies this IntVect by taking component-wise min with IntVect
   argument.

*/
  IntVect& min (const IntVect& p);

///
/**
  Returns the IntVect that is the component-wise minimum of two
  argument IntVects.  

*/
  friend inline IntVect min (const IntVect& p1,
                             const IntVect& p2);

  ///
/**
   Modifies this IntVect by taking component-wise max with //IntVect
   argument.

*/
  IntVect& max (const IntVect& p);

///
  /**
     Returns the IntVect that is the component-wise minimum of two
     argument IntVects.  

  */
  friend inline IntVect max (const IntVect& p1,
                             const IntVect& p2);

  ///
/**
  Modifies this IntVect by multiplying each component by a scalar.

*/
  IntVect& scale (int s);

///
  /**
     Returns an IntVect obtained by multiplying each of the components
     of the given IntVect by a scalar.  

  */
  friend inline IntVect scale (const IntVect& p,
                               int            s);

///
  /**
     Modifies IntVect by reflecting it in the plane defined by the
     index ref\_ix and with normal in the direction of idir.
     Directions are zero-based.  

  */
  IntVect& reflect (int ref_ix,
                    int idir);

///
  /**
     Returns an IntVect that is the reflection of the given IntVect in
     the plane which passes through ref\_ix and normal to the
     coordinate direction idir.  

  */
  friend inline IntVect reflect(const IntVect& a,
                                int            ref_ix,
                                int            idir);

///
/**
   Modifies this IntVect by adding s to component in given coordinate
   direction.

*/
  IntVect& shift (int coord,
                  int s);

  ///
/**
   Modifies this IntVect by component-wise addition with IntVect
   argument.

*/
  IntVect& shift (const IntVect& iv);

  ///
/**
   Modifies this IntVect by adding a scalar s to each component.

*/
  IntVect& diagShift (int s);

///
/**
   Returns IntVect obtained by adding a scalar to each of the
   components of the given IntVect.  

*/
  friend inline IntVect diagShift (const IntVect& p,
                                   int            s);

  ///
/**
   Modify IntVect by component-wise integer projection.

*/
  IntVect& coarsen (const IntVect& p);

  ///
/**
   Modify IntVect by component-wise integer projection.

*/
  IntVect& coarsen (int p);

  ///
/**
   Returns an IntVect that is the component-wise integer projection of
   p by s.  

*/
  friend inline IntVect coarsen (const IntVect& p,
                                 int            s);

///
  /**
     Returns an IntVect which is the component-wise integer projection
     of IntVect p1 by IntVect p2.  

  */
  friend inline IntVect coarsen (const IntVect& p1,
                                 const IntVect& p2);


/// {\bf I/O Functions}

  ///
/**
   Print an IntVect to the ostream.

*/
  void printOn (std::ostream& os) const;

  ///
/**
   Print an IntVect to the pout().

*/
  void p() const;

  ///
/**
   Print an IntVect to the ostream a bit more verbosely.

*/
  void dumpOn (std::ostream& os) const;

  ///
/**
   Print the IntVect to given output stream in ASCII.

*/
  friend std::ostream& operator<< (std::ostream&       os,
                              const IntVect& iv);

  ///
/**
   Read next IntVect from given input stream.

*/
  friend std::istream& operator>> (std::istream& os,
                              IntVect& iv);

/// {\bf IntVect Constants}

///
/**
   Returns a basis vector in the given coordinate direction; eg.
   IntVect3d BASISV(1) == (0,1,0).  Note that the coordinate
   directions are zero based.  

*/
  friend inline IntVect BASISV (int dir);

///
/**
   This is an IntVect all of whose components are equal to zero.

*/
  static const IntVect Zero;

///
/**
   This is an IntVect all of whose components are equal to one.

*/
  static const IntVect Unit;

#ifndef WRAPPER
// TheZeroVector is out of favor.  use IntVect::Zero.
  static const IntVect TheZeroVector () {return Zero;}

// TheUnitVector is out of favor.  use IntVect::Unit.
  static const IntVect TheUnitVector () {return Unit;}
#endif

protected:
  //
  // Box is a friend of ours.
  //
  friend class Box;

  friend class HDF5Handle;
  friend class VolIndex;
  friend class FaceIndex;

  int vect[SpaceDim];

  static const size_t IntVectSize;
};

IntVect UnitVector();
IntVect ZeroVector();

#ifndef WRAPPER
//
// Inlines.
//

// try uninitialized IntVect null construction for now.....

// inline
// IntVect::IntVect ()
// {
//     D_EXPR(vect[0] = 0, vect[1] = 0, vect[2] = 0);
// }



inline
IntVect::IntVect (D_DECL(int i, int j, int k))
{
  D_EXPR(vect[0] = i, vect[1] = j, vect[2] = k);
}

inline
IntVect::IntVect (const int *a)
{
  D_EXPR(vect[0] = a[0], vect[1] = a[1], vect[2] = a[2]);
}

inline
IntVect::IntVect (const IntVect &iv)
{
  //D_EXPR(vect[0]=iv.vect[0], vect[1]=iv.vect[1], vect[2]=iv.vect[2]);
  memcpy(vect, iv.vect, IntVectSize);
}

inline
IntVect&
IntVect::operator= (const IntVect &iv)
{
  D_EXPR(vect[0]=iv.vect[0], vect[1]=iv.vect[1], vect[2]=iv.vect[2]);
  return *this;
}

inline
int&
IntVect::operator[] (int i)
{
  assert(i>=0 && i < SpaceDim);
  return vect[i];
}

inline
int
IntVect::operator[] (int i) const
{
  assert(i>=0 && i < SpaceDim);
  return vect[i];
}

inline
void
IntVect::setVal (int i,
                 int val)
{
  assert(i >=0 && i < SpaceDim);
  vect[i] = val;
  //  return *this;
}

inline
const int* 
IntVect::dataPtr() const
{
  return vect;
}

inline
int* 
IntVect::dataPtr()
{
  return vect;
}

inline
const int*
IntVect::getVect () const
{
  return vect;
}

inline
bool
IntVect::operator== (const IntVect& p) const
{
  return D_TERM(vect[0] == p[0], && vect[1] == p[1], && vect[2] == p[2]);
}

inline
bool
IntVect::operator!= (const IntVect& p) const
{
  return D_TERM(vect[0] != p[0], || vect[1] != p[1], || vect[2] != p[2]);
}

inline
bool
IntVect::operator< (const IntVect& p) const
{
  return D_TERM(vect[0] < p[0], && vect[1] < p[1], && vect[2] < p[2]);
}

inline
bool
IntVect::operator<= (const IntVect& p) const
{
  return D_TERM(vect[0] <= p[0], && vect[1] <= p[1], && vect[2] <= p[2]);
}

inline
bool
IntVect::operator> (const IntVect& p) const
{
  return D_TERM(vect[0] > p[0], && vect[1] > p[1], && vect[2] > p[2]);
}

inline
bool
IntVect::operator>= (const IntVect& p) const
{
  return D_TERM(vect[0] >= p[0], && vect[1] >= p[1], && vect[2] >= p[2]);
}

inline
bool
IntVect::lexLT (const IntVect &s) const
{
#define LLT0 (vect[0] < s[0])
#define LLT1 ((vect[0] == s[0]) && (vect[1] < s[1]))
#define LLT2 ((vect[1] == s[1]) && (vect[2] < s[2]))
#if   CH_SPACEDIM == 1
  return LLT0;
#elif CH_SPACEDIM == 2
  return LLT0 || LLT1;
#elif CH_SPACEDIM == 3
  return LLT0 || (vect[0]==s[0] && ((vect[1] < s[1] || LLT2)));
#endif
#undef LLT0
#undef LLT1
#undef LLT2
}

inline
bool
IntVect::lexGT (const IntVect& s) const
{
#define LGT0 (vect[0] > s[0])
#define LGT1 ((vect[0] == s[0]) && (vect[1] > s[1]))
#define LGT2 ((vect[1] == s[1]) && (vect[2] > s[2]))
#if   CH_SPACEDIM == 1
  return LGT0;
#elif CH_SPACEDIM == 2
  return LGT0 || LGT1;
#elif CH_SPACEDIM == 3
  return LGT0 || (vect[0] == s[0] && ((vect[1] > s[1] || LGT2)));
#endif
#undef LGT0
#undef LGT1
#undef LGT2
}

inline
IntVect
IntVect::operator+ () const
{
  return IntVect(*this);
}

inline
IntVect
IntVect::operator- () const
{
  return IntVect(D_DECL(-vect[0], -vect[1], -vect[2] ));
}

inline
IntVect&
IntVect::operator+= (int s)
{
  D_EXPR(vect[0] += s, vect[1] += s, vect[2] += s);
  return *this;
}

inline
IntVect&
IntVect::operator+= (const IntVect& p)
{
  D_EXPR(vect[0] += p[0], vect[1] += p[1], vect[2] += p[2]);
  return *this;
}

inline
IntVect&
IntVect::operator*= (int s)
{
  D_EXPR(vect[0] *= s, vect[1] *= s, vect[2] *= s);
  return *this;
}

inline
IntVect&
IntVect::operator*= (const IntVect &p)
{
  D_EXPR(vect[0] *= p[0], vect[1] *= p[1], vect[2] *= p[2]);
  return *this;
}

inline
IntVect&
IntVect::operator/= (int s)
{
  D_EXPR(vect[0] /= s, vect[1] /= s, vect[2] /= s);
  return *this;
}

inline
IntVect&
IntVect::operator/= (const IntVect& p)
{
  D_EXPR(vect[0] /= p[0], vect[1] /= p[1], vect[2] /= p[2]);
  return *this;
}

inline
IntVect&
IntVect::operator-= (int s)
{
  D_EXPR(vect[0] -= s, vect[1] -= s, vect[2] -= s);
  return *this;
}

inline
IntVect&
IntVect::operator-= (const IntVect& p)
{
  D_EXPR(vect[0] -= p[0], vect[1] -= p[1], vect[2] -= p[2]);
  return *this;
}

inline
IntVect
IntVect::operator+ (const IntVect& p) const
{
  return IntVect(D_DECL(vect[0] + p[0], vect[1] + p[1], vect[2] + p[2]));
}

inline
IntVect
IntVect::operator+ (int s) const
{
  return IntVect(D_DECL(vect[0] + s, vect[1] + s, vect[2] + s));
}

inline
IntVect
IntVect::operator- (const IntVect& p) const
{
  return IntVect(D_DECL(vect[0] - p[0], vect[1] - p[1], vect[2] - p[2]));
}

inline
IntVect
IntVect::operator- (int s) const
{
  return IntVect(D_DECL(vect[0] - s, vect[1] - s, vect[2] - s));
}

inline
IntVect
IntVect::operator* (const IntVect& p) const
{
  return IntVect(D_DECL(vect[0] * p[0], vect[1] * p[1], vect[2] * p[2]));
}

inline
IntVect
IntVect::operator* (int s) const
{
  return IntVect(D_DECL(vect[0] * s, vect[1] * s, vect[2] * s));
}

inline
IntVect
IntVect::operator/ (const IntVect& p) const
{
  return IntVect(D_DECL(vect[0] / p[0], vect[1] / p[1], vect[2] / p[2]));
}

inline
IntVect
IntVect::operator/ (int s) const
{
  return IntVect(D_DECL(vect[0] / s, vect[1] / s, vect[2] / s));
}

inline
IntVect&
IntVect::min (const IntVect& p)
{
  D_EXPR(vect[0] = Min(vect[0], p.vect[0]),
         vect[1] = Min(vect[1], p.vect[1]),
         vect[2] = Min(vect[2], p.vect[2]));
  return *this;
}

inline
IntVect&
IntVect::max (const IntVect& p)
{
  D_EXPR(vect[0] = Max(vect[0], p.vect[0]),
         vect[1] = Max(vect[1], p.vect[1]),
         vect[2] = Max(vect[2], p.vect[2]));
  return *this;
}

inline
IntVect&
IntVect::scale (int s)
{
  D_EXPR(vect[0] *= s, vect[1] *= s, vect[2] *= s);
  return *this;
}

inline
IntVect&
IntVect::reflect (int ref_ix,
                  int idir)
{
  assert(idir >= 0 && idir < SpaceDim);
  vect[idir] = -vect[idir] + 2*ref_ix;
  return *this;
}

inline
IntVect&
IntVect::shift (int coord,
                int s)
{
  assert(coord >= 0 && coord < SpaceDim);
  vect[coord] += s;
  return *this;
}

inline
IntVect&
IntVect::shift (const IntVect& iv)
{
  *this += iv;
  return *this;
}

inline
IntVect&
IntVect::diagShift (int s)
{
  D_EXPR(vect[0] += s, vect[1] += s, vect[2] += s);
  return *this;
}

inline
IntVect
operator+ (int            s,
           const IntVect& p)
{
  return IntVect(D_DECL(p[0] + s, p[1] + s, p[2] + s));
}

inline
IntVect
operator- (int            s,
           const IntVect& p)
{
  return IntVect(D_DECL(s - p[0], s - p[1], s - p[2]));
}

inline
IntVect
operator* (int            s,
           const IntVect& p)
{
  return IntVect(D_DECL(s * p[0], s * p[1], s * p[2]));
}

inline
IntVect
scale (const IntVect& p,
       int            s)
{
  return IntVect(D_DECL(s * p[0], s * p[1], s * p[2]));
}

inline
IntVect
diagShift (const IntVect &p, int s)
{
  return IntVect(D_DECL(p[0] + s, p[1] + s, p[2] + s));
}

inline
IntVect
min (const IntVect& p1,
     const IntVect& p2)
{
  IntVect p(p1);
  return p.min(p2);
}

inline
IntVect
max (const IntVect& p1,
     const IntVect& p2)
{
  IntVect p(p1);
  return p.max(p2);
}

inline
IntVect
BASISV (int dir)
{
  assert(dir >= 0 && dir < SpaceDim);
  IntVect tmp = IntVect::TheZeroVector() ;
  tmp.vect[dir] = 1;
  return tmp;
}

inline
IntVect
reflect (const IntVect& a,
         int            ref_ix,
         int            idir)
{
  assert(idir >= 0 && idir < SpaceDim);
  IntVect b(a);
  b.vect[idir] = -b.vect[idir] + 2*ref_ix;
  return b;
}

inline
IntVect
coarsen (const IntVect& p,
         int            s)
{
  assert(s > 0);
  return IntVect(
                 D_DECL((p.vect[0]<0) ? -abs(p.vect[0]+1)/s-1 : p.vect[0]/s ,
                        (p.vect[1]<0) ? -abs(p.vect[1]+1)/s-1 : p.vect[1]/s ,
                        (p.vect[2]<0) ? -abs(p.vect[2]+1)/s-1 : p.vect[2]/s ));
}

inline
IntVect
coarsen (const IntVect& p1,
         const IntVect& p2)
{
  assert(p2 > IntVect::TheZeroVector());
  return IntVect(
                 D_DECL(
                        (p1.vect[0]<0)?-abs(p1.vect[0]+1)/p2.vect[0]-1:p1.vect[0]/p2.vect[0],
                        (p1.vect[1]<0)?-abs(p1.vect[1]+1)/p2.vect[1]-1:p1.vect[1]/p2.vect[1],
                        (p1.vect[2]<0)?-abs(p1.vect[2]+1)/p2.vect[2]-1:p1.vect[2]/p2.vect[2])
                 );
}

inline
IntVect&
IntVect::coarsen (int s)
{
  assert(s > 0);
  for (int i = 0; i < SpaceDim; ++i)
    vect[i] = ((vect[i]<0) ? -abs(vect[i]+1)/s-1 : vect[i]/s);
  return *this;
}

inline
IntVect&
IntVect::coarsen (const IntVect& p)
{
  assert(p > IntVect::TheZeroVector());
  for (int i = 0; i <SpaceDim; ++i)
    {
      const int s = p.vect[i];
      vect[i] = ((vect[i]<0) ? -abs(vect[i]+1)/s-1 : vect[i]/s);
    }
  return *this;
}

#endif /* WRAPPER */

#endif /*CH_INTVECT_H*/

/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//


using std::ostream;
using std::istream;
using std::ws;


const size_t  IntVect::IntVectSize = SpaceDim*sizeof(int);
const IntVect IntVect::Unit(D_DECL(1,1,1));
const IntVect IntVect::Zero(D_DECL(0,0,0));

//
// Returns IntVect which is the componentwise integer projection
// of IntVect p1 by IntVect p2.
//

ostream&
operator<< (ostream&       os,
            const IntVect& p)
{
    os << D_TERM( '(' << p[0] , <<
                  ',' << p[1] , <<
                  ',' << p[2])  << ')';
    if (os.fail())
        MayDay::Error("operator<<(ostream&,IntVect&) failed");
    return os;
}

//
// Copied from <Utility.H>
//
#define CH_IGNORE_MAX 100000

istream&
operator>> (istream& is,
            IntVect& p)
{
    is >> ws;
    char c;
    is >> c;
    is.putback(c);
    if (c == '(')
    {
        D_EXPR(is.ignore(CH_IGNORE_MAX, '(') >> p[0],
               is.ignore(CH_IGNORE_MAX, ',') >> p[1],
               is.ignore(CH_IGNORE_MAX, ',') >> p[2]);
        is.ignore(CH_IGNORE_MAX, ')');
    }
    else if (c == '<')
    {
        D_EXPR(is.ignore(CH_IGNORE_MAX, '<') >> p[0],
               is.ignore(CH_IGNORE_MAX, ',') >> p[1],
               is.ignore(CH_IGNORE_MAX, ',') >> p[2]);
        is.ignore(CH_IGNORE_MAX, '>');
    }
    else
        MayDay::Error("operator>>(istream&,IntVect&): expected \'(\' or \'<\'");

    if (is.fail())
        MayDay::Error("operator>>(istream&,IntVect&) failed");

    return is;
}

void
IntVect::printOn (ostream& os) const
{
    os << "IntVect: " << *this << '\n';
}

void
IntVect::p() const
{
    pout() << *this << '\n';
}

void
IntVect::dumpOn (ostream& os) const
{
    os << "IntVect(" << MayDay::version << ")= " << *this << '\n';
}
IntVect UnitVector()
{
  return IntVect::TheUnitVector();
}

IntVect ZeroVector()
{
  return IntVect::TheZeroVector();
}
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef CH_REALVECT_H
#define CH_REALVECT_H

//
// $Id: CFile.cpp,v 1.6 2003/04/21 22:08:00 bvs Exp $
//



//
//@Man:
//@Memo: An Real Vector in SpaceDim-Dimensional Space
/*@Doc:

  The class RealVect is an implementation of an Real vector in a
  SpaceDim-dimensional space.  
  RealVect values are accessed using the operator[] function, as for a normal
  C++ array.  In addition, the basic arithmetic operators have been overloaded
  to implement scaling and translation operations.
*/

class RealVect
{
public:

  /// {\bf Constructors and Accessors}

  ///
  /**
     Construct an RealVect whose components are zero.

  */
  RealVect ();

  ///
  /**
     Construct an RealVect given the specific values for its
     coordinates.  D\_DECL is a macro that sets the constructor to
     take CH\_SPACEDIM arguments.  

  */
  RealVect (D_DECL(Real i, Real j, Real k));

  ///
  /**
     The copy constructor.

  */
  RealVect (const RealVect& rhs);

  ///
  /**
     Construct a RealVect from an IntVect by coercing each component
     from int to Real.
  */
  RealVect (const IntVect & iv){
    for(int d=0 ; d<SpaceDim ; ++d ){ vect[d] = (Real)iv[d] ; }
  }
  ///
  /**
     The assignment operator.

  */
  RealVect& operator= (const RealVect& rhs);

  ///
  /**
     Returns a modifiable lvalue reference to the i'th coordinate of the
     RealVect.

  */
  Real& operator[] (int i);

  ///
  /**
     Returns the i'th coordinate of the RealVect.

  */
  const Real& operator[] (int i) const;


  /// {\bf Comparison Operators}

  ///
  /**
     Returns true if this RealVect is equivalent to argument RealVect.  All
     comparisons between analogous components must be satisfied.  

  */
  bool operator== (const RealVect& p) const;

  ///
  /**
     Returns true if this RealVect is different from argument RealVect.
     All comparisons between analogous components must be satisfied.  

  */
  bool operator!= (const RealVect& p) const;

  ///
  /**
     Modifies this RealVect by addition of a scalar to each component.

  */
  RealVect& operator+= (Real s);

  ///
  /**
     Modifies this RealVect by component-wise addition with argument.

  */
  RealVect& operator+= (const RealVect& p);

  ///
  /**
     Modifies this RealVect by subtraction of a scalar from each
     component.

  */
  RealVect& operator-= (Real s);

  ///
  /**
     Modifies this RealVect by component-wise subtraction with argument.

  */
  RealVect& operator-= (const RealVect& p);

  ///
  /**
     Returns an RealVect that is this RealVect with a scalar s subtracted
     from each component.

  */
  RealVect operator- (Real s) const;
 
  ///
  /**
     Modifies this RealVect by multiplication of a scalar to each
     component.  

  */
  RealVect& operator*= (Real s);

  ///
  /**
     Modifies this RealVect by component-wise multiplication with argument.

  */
  RealVect& operator*= (const RealVect& p);

  ///
  /**
     Returns component-wise product of this RealVect and argument.

  */
  RealVect operator* (const RealVect& p) const;

  ///
  /**
     Returns an RealVect that is this RealVect with each component
     multiplied by a scalar.

  */
  RealVect operator* (Real s) const;

  ///
  /**
     Modifies this RealVect by division by a scalar of each component.

  */
  RealVect& operator/= (Real s);

  ///
  /**
     Modifies this RealVect by component-wise division with RealVect
     argument.

  */
  RealVect& operator/= (const RealVect& p);

  ///
  /**
     Returns component-wise division of this RealVect by RealVect argument.

  */
  RealVect operator/ (const RealVect& p) const;

  ///
  /**
     Returns component-wise division of this RealVect by scalar s.

  */
  RealVect operator/ (Real s) const;

  ///
  /**
     only for sending stuff to fortran
   */
  const Real* dataPtr() const;

  ///
  /**
     only for sending stuff to fortran
   */
  Real* dataPtr() ;

  ///
  /**
     Returns a basis vector in the given coordinate direction; eg.
     RealVect3d BASISV(1) == (0,1,0).  Note that the coordinate
     directions are zero based.  

  */
  friend RealVect BASISREALV(int dir);

  ///
  /**
     This is an RealVect all of whose components are equal to zero.
     
  */
  static const RealVect Zero;

  ///
  /**
     This is an RealVect all of whose components are equal to one.

  */
  static const RealVect Unit;

  ///
  /**
     Returns an RealVect that is an RealVect p with a scalar s added to
     each component.

  */
  friend RealVect operator+ (Real            s,
                             const RealVect& p);

  ///
  /**
     Returns -p + s.

  */
  friend RealVect operator- (Real            s,
                             const RealVect& p);

  ///
  /**
     Returns and RealVect that is an RealVect p with each component
     multiplied by a scalar s.

  */
  friend RealVect operator* (Real            s,
                             const RealVect& p);
  ///
  /**
     Returns and RealVect that is an RealVect p with each component
     divided by a scalar s.

  */
  friend RealVect operator/ (Real            s,
                             const RealVect& p);

  ///
  /**
     Returns an RealVect that is an RealVect p with a scalar s added to
     each component.

  */
  friend RealVect operator+ (const RealVect& s,
                             const RealVect& p);

  ///
  /**
     Returns -p + s.

  */
  friend RealVect operator- (const RealVect& s,
                             const RealVect& p);

  ///
  /**
     Returns and RealVect that is an RealVect p with each component
     multiplied by a scalar s.

  */
  friend RealVect operator* (const RealVect& s,
                             const RealVect& p);
  ///
  /**
     Returns and RealVect that is an RealVect p with each component
     divided by a scalar s.

  */
  friend RealVect operator/ (const RealVect& s,
                             const RealVect& p);

  ///
  /**
     Print to the given output stream in ASCII.
  */
  friend std::ostream& operator<< (std::ostream& ostr,
                                   const RealVect& p);

  friend class HDF5Handle;

  static size_t io_offset;
private:

  Real vect[SpaceDim];

};

/*

template < >
int linearSize(const RealVect& vindex);
  
//VolIndex specialization of linearIn
template < >
void linearIn(RealVect& a_outputT, const void* const inBuf);
  
//VolIndex specialization of linearOut
template < >
void linearOut(void* const a_outBuf, const RealVect& a_inputT); 

*/

#endif /*CH_RealVect_H*/

/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

using std::ostream;
using std::istream;
using std::ws;

size_t RealVect::io_offset = HOFFSET(RealVect, vect[0]);

const RealVect RealVect::Unit(D_DECL(1.0,1.0,1.0));
const RealVect RealVect::Zero(D_DECL(0.0,0.0,0.0));


const Real* 
RealVect::dataPtr() const
{
  return vect;
}

Real* 
RealVect::dataPtr()
{
  return vect;
}

RealVect::RealVect (D_DECL(Real i, Real j, Real k))
{
  D_EXPR(vect[0] = i, vect[1] = j, vect[2] = k);
}

RealVect::RealVect (const RealVect &iv)
{
  D_EXPR(vect[0]=iv.vect[0], vect[1]=iv.vect[1], vect[2]=iv.vect[2]);
}

RealVect::RealVect ()
{
  D_EXPR(vect[0]=0.0, vect[1]=0.0, vect[2] = 0.0);
}

RealVect&
RealVect::operator= (const RealVect &iv)
{
  D_EXPR(vect[0]=iv.vect[0], vect[1]=iv.vect[1], vect[2]=iv.vect[2]);
  return *this;
}

Real&
RealVect::operator[] (int i)
{
  assert(i>=0 && i < SpaceDim);
  return vect[i];
}

const Real&
RealVect::operator[] (int i) const
{
  assert(i>=0 && i < SpaceDim);
  return vect[i];
}

bool
RealVect::operator== (const RealVect& p) const
{
  return D_TERM(vect[0] == p[0], && vect[1] == p[1], && vect[2] == p[2]);
}

bool
RealVect::operator!= (const RealVect& p) const
{
  return D_TERM(vect[0] != p[0], || vect[1] != p[1], || vect[2] != p[2]);
}


RealVect&
RealVect::operator+= (Real s)
{
  D_EXPR(vect[0] += s, vect[1] += s, vect[2] += s);
  return *this;
}

RealVect&
RealVect::operator+= (const RealVect& p)
{
  D_EXPR(vect[0] += p[0], vect[1] += p[1], vect[2] += p[2]);
  return *this;
}

RealVect&
RealVect::operator*= (Real s)
{
  D_EXPR(vect[0] *= s, vect[1] *= s, vect[2] *= s);
  return *this;
}

RealVect&
RealVect::operator*= (const RealVect &p)
{
  D_EXPR(vect[0] *= p[0], vect[1] *= p[1], vect[2] *= p[2]);
  return *this;
}

RealVect
RealVect::operator* (const RealVect& p) const
{
  RealVect v(D_DECL(vect[0]*p[0], vect[1]*p[1], vect[2]*p[2]));
  return v;
}

RealVect
RealVect::operator* (Real s) const
{
  RealVect v(D_DECL(vect[0]*s, vect[1]*s, vect[2]*s));
  return v;
}

RealVect&
RealVect::operator/= (Real s)
{
  D_EXPR(vect[0] /= s, vect[1] /= s, vect[2] /= s);
  return *this;
}

RealVect&
RealVect::operator/= (const RealVect& p)
{
  D_EXPR(vect[0] /= p[0], vect[1] /= p[1], vect[2] /= p[2]);
  return *this;
}

RealVect
RealVect::operator/ (const RealVect & p) const
{
  RealVect result( D_DECL( vect[0] / p[0], vect[1] / p[1], vect[2] / p[2] ) );
  return result ;
}

RealVect
RealVect::operator/ (Real s) const
{
  RealVect result( D_DECL( vect[0] / s, vect[1] / s, vect[2] / s) );
  return result ;
}

RealVect&
RealVect::operator-= (const RealVect& p)
{
  D_EXPR(vect[0] -= p[0], vect[1] -= p[1], vect[2] -= p[2]);
  return *this;
}

RealVect
BASISREALV (int dir)
{
  assert(dir >= 0 && dir < SpaceDim);
  RealVect tmp = RealVect::Zero ;
  tmp.vect[dir] = 1;
  return tmp;
}


RealVect
operator/ (Real            s,
           const RealVect& p)
{
  return RealVect(D_DECL(p[0] / s, p[1] /s, p[2] / s));
}
RealVect
operator+ (Real            s,
           const RealVect& p)
{
  return RealVect(D_DECL(p[0] + s, p[1] + s, p[2] + s));
}

RealVect
operator- (Real            s,
           const RealVect& p)
{
  return RealVect(D_DECL(s - p[0], s - p[1], s - p[2]));
}

RealVect
operator* (Real            s,
           const RealVect& p)
{
  return RealVect(D_DECL(s * p[0], s * p[1], s * p[2]));
}





RealVect
operator/ (const RealVect& s,
           const RealVect& p)
{
  return RealVect(D_DECL(p[0] / s[0], p[1] /s[1], p[2] / s[2]));
}

RealVect
operator+ (const RealVect& s,
           const RealVect& p)
{
  return RealVect(D_DECL(p[0] + s[0], p[1] +s[1], p[2] + s[2]));
}

RealVect
operator- (const RealVect& s,
           const RealVect& p)
{
  return RealVect(D_DECL(s[0] - p[0], s[1] - p[1], s[2] - p[2]));
}

RealVect
operator* (const RealVect& s,
           const RealVect& p)
{
  return RealVect(D_DECL(p[0] * s[0], p[1] *s[1], p[2] * s[2]));
}

std::ostream&
operator<< (std::ostream& ostr, const RealVect& p)
{
  ostr << "(" << D_TERM( p[0] ,<< "," << p[1], << "," << p[2] ) << ")" ;
  if (ostr.fail()) MayDay::Error("operator<<(ostream&,RealVect&) failed");
  return ostr;
}
/*  
template < >
int linearSize(const RealVect& vindex)
{
  return sizeof(RealVect);
}

//RealVect specialization of linearIn
template < >
void linearIn(RealVect& a_outputT, const void* const inBuf)
{
  unsigned char* bob = (unsigned char*)inBuf;
  unsigned char* to = (unsigned char*)&a_outputT;
  memcpy(to + RealVect::io_offset, bob, SpaceDim*sizeof(Real));
}

//RealVect specialization of linearOut
template < >
void linearOut(void* const a_outBuf, const RealVect& a_inputT)
{
  unsigned char* bob = (unsigned char*)a_outBuf;
  const unsigned char* from = (const unsigned char*)&a_inputT;
  memcpy(bob, from + RealVect::io_offset, SpaceDim*sizeof(Real));
}

*/
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

// Interval.H
// ============

//
// Little struct for passing component ranges in code.
//
// 

#ifndef INTERVAL_H
#define INTERVAL_H

/// Structure for passing component ranges in code.
/**
   A class to encapsulate component ranges
*/
struct Interval
{
  Interval():
        m_begin(0), m_end(-1) {}
  ///
  /**
      initialize with first and last component numbers
  */
  Interval(int a_firstComp, int a_lastComp):
    m_begin(a_firstComp), m_end(a_lastComp){}
  void define(int a_firstComp, int a_lastComp)
  {m_begin= a_firstComp; m_end=a_lastComp;}

  ///
  /**
     return first component number
  */
  int begin() const;

  ///
  /**
     return last component number
  */
  int end() const;

  ///
  /**
     return last - first + 1
   */
  int size() const {return m_end-m_begin+1;}


private:
  int m_begin, m_end;

};

#ifndef WRAPPER
inline int Interval::begin() const {return m_begin;}
inline int Interval::end()   const {return m_end;}
#endif

#endif // INTERVAL_H
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef CH_BOX_H
#define CH_BOX_H

#ifdef NODEV
#undef NODEV
#endif


#ifndef WRAPPER





//
//@Man:
//@Memo: Cell-Based or Node-Based Indices
/*@Doc:

  The class IndexType defines an index as being cell based or node (face)
  based in each of the CH\_SPACEDIM directions.  This class defines an
  enumerated type CellIndex to be either CELL or NODE; i.e. each of the
  CH\_SPACEDIM dimensions must be either CELL or NODE.
*/

class IndexType
{
public:
  ///
  /**
     The cell index type: one of CELL or NODE.

  */
  enum CellIndex { CELL = 0, NODE = 1 };

  ///
  /**
     The default constructor.

  */
  IndexType ();

  ///
  /** 
      The copy constructor.

  */
  IndexType (const IndexType& rhs);

  ///
  /**
     Constructs an IndexType identical to an IntVect.

  */
  explicit IndexType (const IntVect& iv);

  ///
  /**
     The assignment operator.

  */
  IndexType& operator= (const IndexType& rhs);

  ///
  /** 
      Constructs an IndexType given an explicit CellIndex for each
      direction.  D\_DECL is a macro that sets the constructor to take
      BL\_SPACEDIM arguments.  

  */
  IndexType (D_DECL(CellIndex i, CellIndex j, CellIndex k));

  ///
  /**
     Sets this IndexType to be NODE based in direction dir.

  */
  void set (int dir);

  ///
  /**
     Sets this IndexType to be CELL based in direction dir.

  */
  void unset (int dir);

  ///
  /**
     True if this IndexType is NODE based in direction dir.

  */
  bool test (int dir) const;

  ///
  /**
     Set this IndexType to be NODE based in all directions.

  */
  void setall ();

  ///
  /**
     Set this IndexType to be CELL based in all directions.

  */
  void clear ();

  ///
  /**
     True if this IndexType is NODE based in any direction.

  */
  bool any () const;

  ///
  /**
     True if this IndexType is valid.

  */
  bool ok () const;

  ///
  /**
     Change this IndexType from CELL to NODE or NODE to CELL
     in direction dir.

  */
  void flip (int i);

  ///
  /**
     True if IndexTypes are identical.

  */
  bool operator== (const IndexType& t) const;

  ///
  /**
     True if IndexTypes are not identical.

  */
  bool operator!= (const IndexType& t) const;

  ///
  /**
     True if this IndexType is CELL based in all directions.

  */
  bool cellCentered () const;

  ///
  /**
     True if this IndexType is NODE based in all directions.

  */
  bool nodeCentered () const;

  ///
  /**
     Set this IndexType to CellIndex type t in direction dir.

  */
  void setType (int       dir,
                CellIndex t);

  ///
  /**
     Returns  the CellIndex of this IndexType in direction dir.

  */
  CellIndex ixType (int dir) const;

  ///
  /**
     Return an integer representing this IndexType in direction dir.

  */
  int operator[] (int dir) const;

  ///
  /**
     Convert this IndexType to an IntVect via component-wise
     conversion of enum CellIndex to int.

  */
  IntVect ixType () const;

  ///
  /** 
      This static member function returns an IndexType object all of
      whose components are of value IndexType::CELL.  It is provided as
      a convenience to our users when defining a Box all of whose faces
      should be of type IndexType::CELL.

  */
  static IndexType TheCellType ();

  ///
  /**
     This static member function returns an IndexType object all of
     whose components are of value IndexType::NODE.  It is provided as
     a convenience to our users when defining a Box all of whose faces
     should be of type IndexType::NODE.

  */
  static IndexType TheNodeType ();

  ///
  /**
     Write an IndexType to an ostream in ASCII.

  */
  friend std::ostream& operator<< (std::ostream&         os,
                                   const IndexType& itype);

  ///
  /**
     Read an IndexType from an istream.

  */
  friend std::istream& operator>> (std::istream&   is,
                                   IndexType& itype);
private:
  //
  // Returns 1<<k.
  //
  static int mask (int k);
  //
  // An integer holding the CellIndex in bits 0 - BL\_SPACEDIM-1.
  //
  unsigned int itype;
};

//
// Inlines.
//


inline
int
IndexType::mask (int k)
{
  return 1<<k;
}

inline
IndexType::IndexType ()
  : itype(0)
{}

inline
IndexType::IndexType (const IndexType& bt)
  : itype(bt.itype)
{}

inline
IndexType& IndexType::operator= (const IndexType& bt)
{
  itype = bt.itype;
  return *this;
}

inline
IndexType::IndexType (const IntVect& iv)
{
  itype = D_TERM((iv[0]?1:0), | ((iv[1]?1:0)<<1), | ((iv[2]?1:0)<<2));
}

inline
IndexType::IndexType (D_DECL(CellIndex i, CellIndex j, CellIndex k))
{
  itype = D_TERM(i, | (j<<1), | (k<<2));
}

inline
void
IndexType::set (int dir)
{
  itype |= mask(dir);
}

inline
void
IndexType::unset (int dir)
{
  itype &= ~mask(dir);
}

inline
bool
IndexType::test (int dir) const
{
  return (itype & mask(dir)) != 0;
}

inline
void
IndexType::setall ()
{
  itype = (1 << SpaceDim) - 1;
}

inline
void
IndexType::clear ()
{
  itype = 0;
}

inline
bool
IndexType::any () const
{
  return itype != 0;
}

inline
bool
IndexType::ok () const
{
  return itype < (1 << SpaceDim);
}

inline
void
IndexType::flip (int i)
{
  itype ^= mask(i);
}

inline
bool
IndexType::operator== (const IndexType& t) const
{
  return t.itype == itype;
}

inline
bool
IndexType::operator!= (const IndexType& t) const
{
  return t.itype != itype;
}

inline
bool
IndexType::cellCentered () const
{
  return itype == 0;
}

inline
bool
IndexType::nodeCentered () const
{
  return itype == (1<<SpaceDim)-1;
}

inline
void
IndexType::setType (int       dir,
                    CellIndex t)
{
  t == CELL ? unset(dir) : set(dir);
}

inline
IndexType::CellIndex
IndexType::ixType (int dir) const
{
  return (CellIndex) ((itype & (1<<dir)) >> dir);
}

inline
int
IndexType::operator[] (int dir) const
{
  return test(dir);
}

inline
IntVect
IndexType::ixType () const
{
  return IntVect(D_DECL(itype&1, (itype>>1)&1, (itype>>2)&1));
}

#endif /* WRAPPER */

//
//@Man:
//@Memo: A Rectangular Domain on an Integer Lattice
/*@Doc: 

  A Box is an abstraction for defining discrete rectangular regions of
  SpaceDim-dimensioned indexing space.  Boxes have an IndexType, which
  defines IndexType::CELL or IndexType::NODE based points for each
  direction and a low and high IntVect which defines the lower and
  upper corners of the Box.  Boxes can exist in positive and negative
  indexing space.

  Box is a dimension dependent class, so SpaceDim must be 
  defined as either 1, 2, or 3 when compiling.  
*/

class Box
{
public:

  /// {\bf Constructors}

  ///
  /**
     The default constructor.  The constructed Box is empty. 

  */
  Box ();

  ///
  /**
     The destructor.

  */
  ~Box () {;}


  ///
  /**
     Construct cell-centered type Box.  It is an error if small is
     greater than big.

  */
  Box (const IntVect& small,
       const IntVect& big);

  ///
  /**
     Construct cell-centered type Box.  It is an error if small is
     greater than big.

  */
  void define(const IntVect& small, const IntVect& big);

  ///
  /**
     Construct Box with specified lengths.  It is an error
     if the lengths are negative.

  */
  Box (const IntVect& small,
       const int*     vec_len);

  ///
  /** 
      Construct Box with given type.  small and big are expected to be
      consistent with given type.  It is an error if small is greater
      than big.  It is an error if typ is not convertible to an
      IndexType.  

  */
  Box (const IntVect& small,
       const IntVect& big,
       const IntVect& typ);

  ///
  /** 
      Construct Box with given type.  small and big are expected to be
      consistent with given type.  It is an error if small is greater
      than big.

  */
  Box (const IntVect&   small,
       const IntVect&   big,
       const IndexType& t);

  ///
  /**
     The copy constructor.

  */
  Box (const Box& b);

  void define(const Box& b);

  Box copy() const {return *this;}

  /// {\bf Accessors}

  ///
  /** 
      Returns the lower corner of this Box.

  */
  const IntVect& smallEnd () const;

  ///
  /**
     return smallend if side is low, 
     bigend if side is high.
  */
  IntVect sideEnd(Side::LoHiSide a_side) const;


  ///
  /**
     Returns the coordinate of the low end of this Box in the given
     direction.  Directions are zero-based.  It is an error if not 0
     <= dir < SpaceDim.

  */
  int smallEnd (int dir) const;

  ///
  /**
     Returns the upper corner of this Box.

  */
  const IntVect& bigEnd () const;

  ///
  /**
     Returns the coordinate of the high end of this Box in the given
     direction.  Directions are zero-based.  It is an error if not 0
     <= dir < SpaceDim.

  */
  int bigEnd (int dir) const;

  ///
  /**
     Returns a constant pointer to the array of low end coordinates.
     Useful for calls to Fortran.  It should not be used in any other
     circumstances.  

  */
  const int* loVect () const;

  ///
  /**
     Returns a constant pointer the array of high end coordinates.
     Useful for calls to Fortran.  It should not be used in any other
     circumstances.  

  */
  const int* hiVect () const;

  ///
  /**
     Returns a constant pointer to the array of coordinates in the
     Box.  Useful for calls to Fortran, but otherwise too dangerous
     for use.  

  */
  const int* getVect () const;

  ///
  /**
     Returns offset of point from smallend; i.e.  index(smallend) ->
     0, bigend would return volume()-1.  Is used in accessing
     FArrayBox.  

  */
  long index (const IntVect& v) const;

  /// centering

  ///
  /**
     Return the indexing type of this Box.

  */
  IndexType ixType () const;

  ///
  /**
     Return the indexing type of this Box.

  */
  IntVect type () const;

  ///
  /**
     Return the indexing type of this Box in the specified direction.
     Directions are zero-based.  It is an error if not 0 <= dir <
     SpaceDim.

  */
  IndexType::CellIndex type (int dir) const;

  /// size functions

  ///
  /**
     Return an IntVect containing the size of this Box in each
     coordinate direction.

  */
  const IntVect& size () const;

  ///
  /**
     Return the size of this Box in the specified coordinate
     direction.  Directions are zero-based.  It is an error
     if not 0 <= dir < SpaceDim.

  */
  int size (int dir) const;

  ///
  /**
     Is the number calculated by numPts() representable in a long?

  */
  bool numPtsOK () const;

  ///
  /**
     Return the volume, in indexing space, of region enclosed by this
     Box.  This is identical to volume().  Abort()s if the number
     cannot be represented in a long.\\

  */
  long numPts () const;

  ///
  /**
     Is the number calculated by volume() representable in a long?

  */
  bool volumeOK () const;

  ///
  /**
     Return the volume, in indexing space, of region enclosed by this
     Box.  This is identical to numPts().  Abort()s if the number
     cannot be represented in a long.\\

  */
  long volume () const;

  ///
  /**
     Returns length of the longest side of this Box.  The argument dir
     is modified to give direction with longest side:
     0...SpaceDim-1. Ignores type.  

  */
  int longside (int& dir) const;

  ///
  /**
     Returns length of the longest side of this Box.  Ignores type.

  */
  int longside () const;

  ///
  /**
     Returns length of shortest side of this Box.  The argument dir is
     modified to give direction with shortest side: 0...SpaceDim-1.
     Ignores type.  

  */
  int shortside (int& dir) const;

  ///
  /**
     Returns length of the shortest side of this Box.  Ignores type.

  */
  int shortside () const;


  /// {\bf Comparison Functions}

  ///
  /**
     Returns true if this Box is empty.

  */
  bool isEmpty () const;

  ///
  /**
     Returns true if argument is contained within this Box.  An empty
     Box does not contain and is not contained by any Box, including
     itself.

  */
  bool contains (const IntVect& p) const;

  ///
  /**
     Returns true if argument is contained within this Box.  It is an
     error if the Boxes have different types.  An empty Box does not
     contain any IntVect.  

  */
  bool contains (const Box& b) const;

  /// 
  /**
     Returns true if this Box and the argument have non-null
     intersections.  It is an error if the Boxes have different types.
     An empty Box does not intersect any Box, including itself.  

  */
  bool intersects (const Box& b) const;

  /// 
  /**
     Returns true if this Box and the argument have non-null
     intersections.  It is an error if the Boxes have different types.
     This routine does not perform the check to see if *this or b are
     empty boxes.  It is the callers responsibility to ensure that
     this never happens.  If you are unsure, the use the .intersects(..) routine.

  */
  bool intersectsNotEmpty (const Box& b) const;

  /// 
  /**
     Returns true if this Box and the argument are the same size, ie
     translates of each other.  It is an error if they have different
     types.  
  */
  bool sameSize (const Box& b) const;

  ///
  /**
     Returns true if this Box and the argument have same type.

  */
  bool sameType (const Box &b) const;

  ///
  /**
     Returns true if this Box and the argument are identical,
     including type.

  */
  bool operator== (const Box& b) const;

  bool eq(const Box& b) const;
  ///
  /**
     Returns true if this Box and the argument differ, including type.

  */
  bool operator!= (const Box& b) const;

  bool neq(const Box& b) const;
  ///
  /**
     Returns true if this Box is cell-centered in all indexing
     directions.

  */
  bool cellCentered () const;

  // following operators added Sept. 14, 1999.  bvs
  ///
  /**
     Returns true if this Box is lexigraphically less than rhs box.
     All comparison is based on lower box corner.  In the name of
     coding efficiency, we do not handle the case where two boxes have
     the same lower left corner. In DEBUG mode, this is checked for
     with an assert.

  */
  bool operator < (const Box& rhs) const;

  bool lt(const Box& rhs) const;


  /// {\bf Modification Functions}

  ///
  /**
     The assignment operator.

  */
  Box& operator= (const Box& b);

  ///
  /**
     Redefine the lower corner of this Box.  It is an error
     if the specified corner is greater than the exisiting upper
     corner of this Box.

  */
  Box& setSmall (const IntVect& sm);

  ///
  /**
     Redefines the lower end of this Box in the specified coordinate
     direction.  It is an error if the specified value is greater than
     the exisiting upper end of this Box.  Directions are zero-based.
     It is an error if not 0 <= dir < SpaceDim.

  */
  Box& setSmall (int dir,
                 int sm_index);

  ///
  /**
     Redefines the upper corner of this Box.  It is an error if the
     specified corner is less than the exisiting lower corner of this
     Box.

  */
  Box& setBig (const IntVect& bg);

  ///
  /**
     Redefines the upper end of this Box in the specified coordinate
     direction.  It is an error if the specified value is less than the
     exisiting lower end of this Box.  Directions are zero-based.  It
     is an error if not 0 <= dir < SpaceDim.

  */
  Box& setBig (int dir,
               int bg_index);

  ///
  /**
     Set the entire range in a given direction, starting at sm\_index
     with length n\_cells.  It is an error if n\_cells <= 0.  

  */
  Box& setRange (int dir,
                 int sm_index,
                 int n_cells = 1);

  /// centering type conversion functions

  ///
  /**
     Modifies this Box by converting from the current type into the
     argument type.  This may change the Box coordinates:\\ 

     type CELL -> NODE : increase coordinate by one on high end.\\

     type NODE -> CELL : reduce coordinate by one on high end.\\ 

     other type mappings make no change.

  */
  Box& convert (IndexType typ);

  ///
  /**
     Modifies this Box by converting from the current type into the
     argument type.  This may change the Box coordinates:\\ type CELL
     -> NODE : increase coordinate by one on high end.\\ type NODE ->
     CELL : reduce coordinate by one on high end.\\ other type
     mappings make no change.  It is an error if typ is not
     convertible to an IndexType.  

  */
  Box& convert (const IntVect& typ);

  ///
  /**
     Modifies this Box by converting from the current type into the
     argument type.  This may change the Box coordinates:\\

     type CELL -> NODE : increase coordinate by one on high end.\\

     type NODE -> CELL : reduce coordinate by one on high end.\\

     Other type mappings make no change.  Directions are zero-based.
     It is an error if not 0 <= dir < SpaceDim.  

  */
  Box& convert (int                  dir,
                IndexType::CellIndex typ);

  ///
  /**
     Modifies this Box by converting it to NODE type in all
     directions.  This increases all components of the upper corner by
     one.  The Empty Box is not modified by this function except for
     changing of type.

  */
  Box& surroundingNodes ();

  ///
  /**
     Modifies this Box by converting it to NODE type in given
     direction.  This increases the component of the upper corner in
     the given direction by one.  Directions are zero-based.  It is an
     error if not 0 <= dir < SpaceDim.  The Empty Box is not modified
     by this function except for changing of type.

  */
  Box& surroundingNodes (int dir);

  Box& surroundingNodes_int(int dir);
  ///
  /**
     Returns a Box with NODE based coordinates in direction dir that
     encloses the argument Box.  NOTE: equivalent to
     b.convert(dir,NODE).  NOTE: error if b.type(dir) == NODE.
     Directions are zero-based.  It is an error if not 0 <= dir <
     SpaceDim.  The Empty Box is not modified by this function except
     for changing of type.  

  */
  friend  Box surroundingNodes (const Box& b,
                                int        dir);

  
  ///
  /**
     Returns a Box with NODE based coordinates in all directions that
     encloses argument Box.  The upper corner of the return Box will
     be one more in all components than the upper corner of the
     argument Box.  The Empty Box is not modified by this function
     except for changing of type.  

  */
  friend Box surroundingNodes (const Box& b);

  ///
  /**
     Modifies this Box by converting it to CELL type in all directions.
     This decreases all components of the upper corner by one.  The
     Empty Box is not modified by this function except for changing of
     type.

  */
  Box& enclosedCells ();

  ///
  /**
     Modifies this Box by converting it to CELL type in given
     direction.  This decreases the component of the upper corner in
     the given direction by one.  Directions are zero-based.  It is an
     error if not 0 <= dir < SpaceDim.  The Empty Box is not modified
     by this function except for changing of type.

  */
  Box& enclosedCells (int dir);

  Box& enclosedCells_int (int dir);

  ///
  /**
     Returns a Box with CELL based coordinates in direction dir that
     is enclosed by argument Box.  NOTE: equivalent to
     b.convert(dir,CELL).  NOTE: error if b.type(dir) == CELL.
     Directions are zero-based.  It is an error if not 0 <= dir <
     SpaceDim.  The Empty Box is not modified by this function except
     for changing of type.  

  */
  friend Box enclosedCells (const Box& b,
                            int        dir);

  ///
  /**
     Returns a Box with CELL based coordinates in all directions that
     is enclosed by argument Box.  The upper corner of the return Box
     will be one less in all components than the upper corner of the
     argument Box. The Empty Box is not modified by this function
     except for changing of type.  

  */
  friend Box enclosedCells (const Box& b);

  /// shift functions

  ///
  /**
     Modifies this Box by shifting it nzones indexing positions in
     coordinate direction dir.  Directions are zero-based.  It is an
     error if not 0 <= dir < SpaceDim.  The Empty Box is not modified
     by this function.  

  */
  Box& shift (int dir,
              int nzones);

  ///
  /**
     Modifies this Box by shifting.  Equivalent to
     b.shift(0,iv[0]).shift(1,iv[1]) ....  The Empty Box is not
     modified by this function.

  */
  Box& shift (const IntVect& iv);
  
  Box& shift_intvect (const IntVect& iv);

  ///
  /**
     This member modifies this Box by shifting by "half" indices,
     thereby converting the Box from type CELL to NODE or vice-versa.
     b.shiftHalf(0,1) shifts b to the right by 1/2 cells.
     b.shiftHalf(1,-3) shifts b in the -j direction by 3/2 cells.
     NOTE: If num\_halfs is EVEN the shift is num\_halfs/2 full zones
     and hence will not change the type.  This is: b.shifthalf(4) ==
     b.shift(2).  Directions are zero-based.  It is an error if not 0
     <= dir < SpaceDim.  The Empty Box is not modified by this
     function except for changing of type.  

  */
  Box& shiftHalf (int dir,
                  int num_halfs);

  ///
  /**
     Modifies this Box by shifting by half indices.  Equivalent to
     b.shiftHalf(0,iv[0]).shiftHalf(1,iv[1]) ...  The Empty Box is not
     modified by this function except for changing of type.

  */
  Box& shiftHalf (const IntVect& iv);
  Box& shiftHalf_intvect (const IntVect& iv);

  ///
  /**
     Modifies this Box by shifting it by given IntVect.  The Empty Box
     is not modified by this function.

  */
  Box& operator+= (const IntVect& v);

  ///
  /**
     Returns a Box that is this Box shifted by the given IntVect.  The
     Empty Box is not modified by this function.

  */
  Box  operator+  (const IntVect& v) const;

  ///
  /**
     Modifies this Box by shifting it by given IntVect.  The Empty Box
     is not modified by this function.

  */
  Box& operator-= (const IntVect& v);

  ///
  /**
     Returns a Box that is this Box shifted by the given IntVect.  The
     Empty Box is not modified by this function.

  */
  Box  operator-  (const IntVect& v) const;

  /// neighbor box functions

  ///
  friend Box bdryBox(const Box& b,
                     int        dir,
                     Side::LoHiSide a_sd,
                     int        len);


  ///
  /**
     Returns the edge-centered Box (in direction dir) defining the low
     side of the argument Box.  The output Box will have the given
     length in the given direction.  Directions are zero-based.  It is
     an error if not 0 <= dir < SpaceDim.  The neighbor of an Empty Box
     is an Empty Box of the appropriate type.  

  */
  friend Box bdryLo (const Box& b,
                     int        dir,
                     int        len);

  /// 
  /**
     Returns the edge-centered Box (in direction dir) defining the
     high side of the argument Box.  The return Box will have the
     given length in the given direction.  Directions are zero-based.
     It is an error if not 0 <= dir < SpaceDim.  The neighbor of an
     Empty Box is an Empty Box of the appropriate type.  

  */
  friend Box bdryHi (const Box& b,
                     int        dir,
                     int        len);

  ///
  /**
     Returns the cell centered Box of the given length adjacent to the
     argument Box on the low end along the given coordinate direction.
     The return Box is identical to the argument Box in the other
     directions.  The return Box and the argument Box have an empty
     intersection.  \\

     NOTE: len >= 1. \\

     NOTE: Box retval = adjCellLo(b,dir,len) is equivalent to the
     following set of operations: \\ 

     Box retval(b); \\

     retval.convert(dir,Box::CELL); \\

     retval.setrange(dir,retval.smallEnd(dir)-len,len); \\

     Directions are zero-based.  It is an error if not 0 <= dir <
     SpaceDim.  The neighbor of an Empty Box is an Empty Box of the
     appropriate type.  

  */
  friend Box adjCellLo (const Box& b,
                        int        dir,
                        int        len);

  ///
  /**
     Returns the cell centered Box of the given length adjacent to the
     argument Box on the high end along the given coordinate
     direction.  The return Box is identical to the argument Box in
     the other directions.  The return Box and the argument Box have
     an empty intersection.  \\

     NOTE: len >= 1. \\

     NOTE: Box retval = adjCellHi(b,dir,len) is equivalent to the
     following set of operations: \\ 

     Box retval(b); \\

     retval.convert(dir,Box::CELL); \\

     retval.setrange(dir,retval.bigEnd(dir)+1,len);\\

     Directions are zero-based.  It is an error if not 0 <= dir <
     SpaceDim.  The neighbor of an Empty Box is an Empty Box of the
     appropriate type.  

  */
  friend Box adjCellHi (const Box& b,
                        int        dir,
                        int        len);


  /// intersection functions 

  ///
  /**
     Returns the Box that is intersection of this Box and the argument
     Box.  Intersection is commutative.  The Boxes MUST be of same
     type.  The intersection of the Empty Box and any Box is the Empty
     Box.  

  */
  Box operator& (const Box&) const;

  ///
  /**
     Modifies this Box by intersection with the argument Box.  The
     Boxes MUST be of the same type.  The intersection of the Empty
     Box and any Box is the Empty Box.  

  */
  Box& operator&= (const Box&);


  ///
  friend Box adjCellBox (const Box& b,
                         int        dir,
                         Side::LoHiSide a_side,
                         int        len);
  /// 
  /**
     Modifies this Box to that of the minimum Box containing both this
     Box and the argument Box.  Both Boxes must have identical type.  

  */
  Box& minBox (const Box& b);

  ///
  /**
     Returns a Box that is the minimum Box containing both the
     argument Boxes.  Both Boxes must have identical type.  

  */
  friend Box minBox (const Box& b1,
                     const Box& b2);

  /// grow functions

  ///
  /**
     Modifies this Box by growing it in all directions by given
     amount.  The Empty Box is not modified by this function.  NOTE:
     n\_cell negative shrinks the Box by that number of cells.  If
     shrinking produces an empty Box, the Box is transformed into the
     canonical Empty Box.  

  */
  Box& grow (int i);

  ///
  /**
     Returns a Box that is the argument Box grown in all directions by
     given amount.  The Empty Box is not modified by this function.
     NOTE: n\_cell negative shrinks the Box by that number of cells.
     If shrinking produces an empty Box, the Box is transformed into
     the canonical Empty Box.  

  */
  friend Box grow (const Box& b,
                   int        i);

  ///
  /**
     Modifies this Box by growing it in each direction by specified
     amount.  The Empty Box is not modified by this function.  NOTE:
     components of iv may be negative, which would shrink this Box.  If
     shrinking produces an empty Box, the Box is transformed into the
     canonical Empty Box.

  */
  Box& grow (const IntVect& v);

  ///
  /**
     Returns a Box that is the argument Box grown in each
     direction by specified amount.  The Empty Box is not modified by
     this function.  NOTE: components of iv may be negative, which
     would return a shrunken Box.  If shrinking produces an empty
     Box, the Box is transformed into the canonical Empty Box.

  */
  friend  Box grow (const Box&     b,
                    const IntVect& v);

  ///
  /**
     Modifies this Box by growing it on the low and high end by n\_cell
     cells in direction idir.  The Empty Box is not modified by this
     function.  NOTE: n\_cell negative shrinks this Box by that number
     of cells.  If shrinking produces an empty Box, the Box is
     transformed into the canonical Empty Box.  Directions are
     zero-based.  It is an error if not 0 <= dir < SpaceDim.  

  */
  Box& grow (int idir,
             int n_cell);

  ///
  /**
     Modifies this Box by growing it on the low end by n\_cell cells
     in direction idir.  The Empty Box is not modified by this
     function.  NOTE: n\_cell negative shrinks this Box by that number
     of cells.  If shrinking produces an empty Box, the Box is
     transformed into the canonical Empty Box.  Directions are
     zero-based.  It is an error if not 0 <= dir < SpaceDim.

  */
  Box& growLo (int idir,
               int n_cell=1);

  ///
  /**
     growLo if a_sd == Side::Lo \\
     growHi if a_sd= = Side::Hi 
   */
  Box& growDir (int a_idir,
                const Side::LoHiSide& a_sd,
                int a_cell);

  ///
  /**
     Modifies this Box by growing it on the high end by n\_cell cells
     in direction idir.  The Empty Box is not modified by this
     function.  NOTE: n\_cell negative shrinks the Box by that number
     of cells.  If shrinking produces an empty Box, the Box is
     transformed into the canonical Empty Box.  Directions are
     zero-based.  It is an error if not 0 <= dir < SpaceDim.  

  */
  Box& growHi (int idir,
               int n_cell=1);

  /// refinement

  ///
  /**
     Modifies this Box by refining it by given (positive) refinement
     ratio.  The Empty Box is not modified by this function.  \\

     NOTE: if type(dir) = CELL centered: lo <- lo*ratio and hi <-
     (hi+1)*ratio - 1. \\

     NOTE: if type(dir) = NODE centered: lo <- lo*ratio and hi <-
     hi*ratio.  

  */
  Box& refine (int refinement_ratio);

  ///
  /**
     Returns a Box that is the argument Box refined by given
     (positive) refinement ratio.  The Empty Box is not modified by
     this function.  \\

     NOTE: if type(dir) = CELL centered: lo <- lo*ratio and hi <-
     (hi+1)*ratio - 1.\\

     NOTE: if type(dir) = NODE centered: lo <- lo*ratio and hi <-
     hi*ratio.  

  */
  friend Box refine (const Box& b,
                     int        refinement_ratio);

  ///
  /**
     Modifies this Box by refining it by given (positive) refinement
     ratio.  The Empty Box is not modified by this function.  \\

     NOTE: if type(dir) = CELL centered: lo <- lo*ratio and hi <-
     (hi+1)*ratio - 1.\\

     NOTE: if type(dir) = NODE centered: lo <- lo*ratio and hi <-
     hi*ratio.

  */
  Box& refine (const IntVect& refinement_ratio);

  ///
  /**
     Returns a Box that is the argument Box refined by given
     (positive) refinement ratio.  The Empty Box is not modified by
     this function.  \\

     NOTE: if type(dir) = CELL centered: lo <- lo*ratio and hi <-
     (hi+1)*ratio - 1.\\

     NOTE: if type(dir) = NODE centered: lo <- lo*ratio and hi <-
     hi*ratio.

  */
  friend Box refine (const Box&     b,
                     const IntVect& refinement_ratio);

  /// coarsening

  ///
  /**
     Modifies this Box by coarsening it by given (positive) refinement
     ratio.  The Empty Box is not modified by this function.  \\

     NOTE: if type(dir) = CELL centered: lo <- lo/ratio and hi <-
     hi/ratio.\\

     NOTE: if type(dir) = NODE centered: lo <- lo/ratio and hi <-
     hi/ratio + ((hi%ratio)==0 ? 0 : 1).\\

     That is, refinement of coarsened Box must contain the original
     Box.

  */

  Box& coarsen (int refinement_ratio);

  ///
  /**
     Returns a Box that is the argument Box coarsened by given
     (positive) refinement ratio.  The Empty Box is not modified by
     this function.  \\

     NOTE: if type(dir) = CELL centered: lo <- lo/ratio and hi <-
     hi/ratio.\\

     NOTE: if type(dir) = NODE centered: lo <- lo/ratio and hi <-
     hi/ratio + ((hi%ratio)==0 ? 0 : 1).\\

     That is, refinement of coarsened Box must contain the original
     Box.

  */
  friend Box coarsen (const Box& b,
                      int        refinement_ratio);

  ///
  /**
     Modifies this Box by coarsening by given (positive) refinement
     ratio.  The Empty Box is not modified by this function.  \\

     NOTE: if type(dir) = CELL centered: lo <- lo/ratio and hi <-
     hi/ratio.\\

     NOTE: if type(dir) = NODE centered: lo <- lo/ratio and hi <-
     hi/ratio + ((hi%ratio)==0 ? 0 : 1).\\

     That is, refinement of coarsened Box must contain the original
     Box.

  */
  Box& coarsen (const IntVect& refinement_ratio);

  ///
  /**
     Returns a Box that is the argument Box coarsened by given
     (positive) refinement ratio.  The Empty Box is not modified by
     this function.  \\

     NOTE: if type(dir) = CELL centered: lo <- lo/ratio and hi <-
     hi/ratio.\\

     NOTE: if type(dir) = NODE centered: lo <- lo/ratio and hi <-
     hi/ratio + ((hi%ratio)==0 ? 0 : 1).\\

     That is, refinement of coarsened Box must contain the original
     Box.

  */
  friend Box coarsen (const Box&     b,
                      const IntVect& refinement_ratio);

  // next(...) is out of favor.  use BoxIterator.
  /*
    Step through the rectangle.  It is a runtime error to give
    a point not inside rectangle.  Iteration may not be efficient.
  */
  void next (IntVect &) const;

  /*
    Scan argument IntVect over object second arg is
    increment vector.  Runtime error if IntVect is not
    contained in object Box.  Iteration may not be efficient.
  */
  void next (IntVect&   p,
             const int* shv) const;


  /// misc

  ///
  /**
     Chops the Box at the chop\_pnt in the dir direction returns one
     Box, modifies the object Box.  The union of the two is the
     original Box.  The modified Box is the low end, the returned Box
     is the high end.  If type(dir) = CELL, the Boxes are disjoint with
     if chop\_pnt is the low end of the orig Box.  If type(dir) = NODE,
     common.  It is also an error if the chop\_pnt is an end node of
     the Box.  Directions are zero-based.  It is an error if not 0 <=
     dir < SpaceDim.  

  */
  Box chop (int dir,
            int chop_pnt);

  ///{\bf I/O Functions}

  ///
  /**
     Write an ASCII representation to the ostream.

  */
  friend std::ostream& operator<< (std::ostream&   os,
                                   const Box& bx);

  ///
  /**
     Read from istream.

  */
  friend std::istream& operator>> (std::istream& os,
                                   Box&     bx);


  ///
  /**
     print to pout()
  */
  void p() const;

  ///
  /**
     Gives more detail than printOn.  Useful for exiting due to an
     error.  

  */
  void dumpOn (std::ostream& strm) const;

  /// {\bf Box Constants}

  ///
  /**
     This is a canonical empty Box of cell-centered type.

  */
  //static const Box Empty;


  // TheUnitBox is out of favor.
  /*
    This static member function returns a constant reference to 
    an object of type Box representing the unit box in
    BL\_SPACEDIM-dimensional space.
  */

  //static const Box& TheUnitBox ();

  //
  // Sets the 'len' element of the Box.  Aborts on integer overflow.
  //
  void computeBoxLen ();
  void computeBoxLenNotEmpty();

protected:
  friend class HDF5Handle;
  //
  // A helper function for numPtsOK() and numPts().
  //
  bool numPtsOK (long& N) const;
  //
  // A helper function for volumeOK() and volume().
  //
  bool volumeOK (long& N) const;

  IntVect   smallend;
  IntVect   bigend;
  IntVect   len;
  IndexType btype;

};

// global function prototypes
Box surroundingNodes (const Box& b,
                      int        dir);
Box surroundingNodes (const Box& b);
Box enclosedCells (const Box& b,
                   int        dir);
Box enclosedCells (const Box& b);
Box bdryBox(const Box& b,
            int        dir,
            Side::LoHiSide a_sd,
            int        len=1);
Box bdryLo (const Box& b,
            int        dir,
            int        len=1);
Box bdryHi (const Box& b,
            int        dir,
            int        len=1);
Box adjCellLo (const Box& b,
               int        dir,
               int        len=1);
Box adjCellHi (const Box& b,
               int        dir,
               int        len=1);
Box minBox (const Box& b1,
            const Box& b2);
//
// Inlines. 
//

#ifndef WRAPPER

inline
Box::Box (const Box& b)
  : smallend(b.smallend),
    bigend(b.bigend),
    btype(b.btype)
{
  D_EXPR(len[0] = b.len[0],
         len[1] = b.len[1],
         len[2] = b.len[2]);
}

inline
Box&
Box::operator= (const Box& b)
{
  smallend = b.smallend;
  bigend = b.bigend;
  btype = b.btype;
  D_EXPR(len[0] = b.len[0],
         len[1] = b.len[1],
         len[2] = b.len[2]);
  return *this;
}

inline
IntVect
Box::sideEnd(Side::LoHiSide a_side) const
{
  IntVect retval;
  if(a_side == Side::Lo)
    retval = smallEnd();
  else
    retval = bigEnd();
  return retval;
}

inline
const IntVect&
Box::smallEnd () const
{
  return smallend;
}

inline
int
Box::smallEnd (int dir) const
{
  return smallend[dir];
}

inline
const IntVect&
Box::bigEnd () const
{
  return bigend;
}

inline
int
Box::bigEnd (int dir) const
{
  return bigend[dir];
}

inline
IndexType
Box::ixType () const
{
  return btype;
}

inline
IntVect
Box::type () const
{
  return btype.ixType();
}

inline
IndexType::CellIndex
Box::type (int dir) const
{
  return btype.ixType(dir);
}

inline
const IntVect&
Box::size () const
{
  return len;
}

inline
int
Box::size (int dir) const
{
  return len[dir];
}

inline
const int*
Box::loVect () const
{
  return smallend.getVect();
}

inline
const int*
Box::hiVect () const
{
  return bigend.getVect();
}

inline
const int*
Box::getVect () const
{
  return smallend.getVect();
}


inline
bool
Box::numPtsOK () const
{
  long ignore;
  return numPtsOK(ignore);
}

inline
bool
Box::isEmpty () const
{
  //    return numPts() == 0;
  return (!(bigend >= smallend));
}

inline
bool
Box::contains (const IntVect& p) const
{

  return ( !isEmpty() && (p >= smallend && p <= bigend) );
}

inline
bool
Box::sameType (const Box &b) const
{
  return btype == b.btype;
}

inline
bool
Box::contains (const Box& b) const
{
  assert(sameType(b));
  return ( !isEmpty() && !b.isEmpty() && 
           (b.smallend >= smallend && b.bigend <= bigend) );
}

inline
bool
Box::sameSize (const Box& b) const
{
  assert(sameType(b));
  return D_TERM(len[0] == b.len[0],
                && len[1]==b.len[1],
                && len[2]==b.len[2]);
}

inline
bool
Box::operator== (const Box& b) const
{
  return smallend == b.smallend && bigend == b.bigend && b.btype == btype;
}

inline
bool
Box::operator!= (const Box& b) const
{
  return !operator==(b);
}

inline
bool
Box::cellCentered () const
{
  return !btype.any();
}

inline
bool
Box::volumeOK () const
{
  return numPtsOK();
}

inline
long
Box::index (const IntVect& v) const
{
  long result = v.vect[0]-smallend.vect[0];
#if   CH_SPACEDIM==2
  result += len[0]*(v.vect[1]-smallend.vect[1]);
#elif CH_SPACEDIM==3
  result += len[0]*(v.vect[1]-smallend.vect[1]
                    +(v.vect[2]-smallend.vect[2])*len[1]);
#endif
  return result;
}

inline
void
Box::computeBoxLen ()
{
  if (isEmpty())
    {
      len = IntVect::Zero;
    }
  else
    {
      D_EXPR(len[0] = bigend[0]-smallend[0] + 1,
             len[1] = bigend[1]-smallend[1] + 1,
             len[2] = bigend[2]-smallend[2] + 1);
    }
}

inline
void
Box::computeBoxLenNotEmpty()
{
  D_EXPR(len[0] = bigend[0]-smallend[0] + 1,
         len[1] = bigend[1]-smallend[1] + 1,
         len[2] = bigend[2]-smallend[2] + 1);
}

inline
Box&
Box::setSmall (const IntVect& sm)
{
  assert (sm <= bigend);

  smallend = sm;
  computeBoxLen();
  return *this;
}

inline
Box&
Box::setSmall (int dir,
               int sm_index)
{
  assert (sm_index <= bigend[dir]);

  smallend.setVal(dir,sm_index);
  computeBoxLen();
  return *this;
}

inline
Box&
Box::setBig (const IntVect& bg)
{
  assert (bg >= smallend);

  bigend = bg;
  computeBoxLen();
  return *this;
}

inline
Box&
Box::setBig (int dir,
             int bg_index)
{
  assert (bg_index >= smallend[dir]);

  bigend.setVal(dir,bg_index);
  computeBoxLen();
  return *this;
}

inline
Box&
Box::setRange (int dir,
               int sm_index,
               int n_cells)
{
  assert (n_cells > 0);

  smallend.setVal(dir,sm_index);
  bigend.setVal(dir,sm_index+n_cells-1);
  computeBoxLen();
  return *this;
}

inline
Box&
Box::shift (int dir,
            int nzones)
{
  if (!isEmpty())
    {
      smallend.shift(dir,nzones);
      bigend.shift(dir,nzones);
    }
  return *this;
}

inline
Box&
Box::shift (const IntVect& iv)
{
  if (!isEmpty())
    {
      smallend.shift(iv);
      bigend.shift(iv);
    }
  return *this;
}

inline
Box&
Box::convert (const IntVect& typ)
{
  assert(typ >= IntVect::TheZeroVector() && typ <= IntVect::TheUnitVector());
  if (!isEmpty())
    {
      IntVect shft(typ - btype.ixType());
      bigend += shft;
    }
  btype = IndexType(typ);
  computeBoxLen();
  return *this;
}

inline
Box&
Box::surroundingNodes (int dir)
{
  if (!(btype[dir]))
    {
      if (!isEmpty())
        {
          bigend.shift(dir,1);
        }
      //
      // Set dir'th bit to 1 = IndexType::NODE.
      //
      btype.set(dir);
      computeBoxLen();
    }
  return *this;
}

inline
Box&
Box::enclosedCells (int dir)
{
  if (btype[dir])
    {
      if (!isEmpty())
        {
          bigend.shift(dir,-1);
        }
      //
      // Set dir'th bit to 0 = IndexType::CELL.
      //
      btype.unset(dir);
      computeBoxLen();
    }
  return *this;
}

inline
Box
surroundingNodes (const Box& b,
                  int        dir)
{
  Box bx(b);
  return bx.surroundingNodes(dir);
}

inline
Box
surroundingNodes (const Box& b)
{
  Box bx(b);
  return bx.surroundingNodes();
}

inline
Box
enclosedCells (const Box& b,
               int        dir)
{
  Box bx(b);
  return bx.enclosedCells(dir);
}

inline
Box
enclosedCells (const Box& b)
{
  Box bx(b);
  return bx.enclosedCells();
}

inline
Box&
Box::operator+= (const IntVect& v)
{
  if (!isEmpty())
    {
      smallend += v;
      bigend += v;
    }
  return *this;
}

inline
Box
Box::operator+  (const IntVect& v) const
{
  if (isEmpty())
    {
      return(Box().convert(btype));
    }
  else
    {
      IntVect small(smallend);
      small += v;
      IntVect big(bigend);
      big += v;
      return Box(small,big,btype);
    }
}

inline
Box&
Box::operator-= (const IntVect& v)
{
  if (!isEmpty())
    {
      smallend -= v;
      bigend -= v;
    }
  return *this;
}

inline
bool 
Box::operator < (const Box& rhs) const
{
  return(!isEmpty() && (rhs.isEmpty() || smallend.lexLT(rhs.smallend)));
}

inline
Box
Box::operator-  (const IntVect& v) const
{
  if (isEmpty())
    {
      return(Box().convert(btype));
    }
  else
    {
      IntVect small = smallend;
      small -= v;
      IntVect big = bigend;
      big -= v;
      return Box(small,big,btype);
    }
}

inline
Box&
Box::grow (int i)
{
  if (!isEmpty())
    {
      smallend.diagShift(-i);
      bigend.diagShift(i);
      if (!(bigend >= smallend)) *this = Box().convert(btype);
      computeBoxLen();
    }
  return *this;
}

inline
Box
grow (const Box& b,
      int        i)
{
  if (b.isEmpty())
    {
      return (Box().convert(b.btype));
    }
  else
    {
      IntVect small = diagShift(b.smallend,-i);
      IntVect big   = diagShift(b.bigend,i);
      if (!(big >= small))
        {
          return(Box().convert(b.btype));
        }
      else
        {
          return Box(small,big,b.btype);
        }
    }
}

inline
Box&
Box::grow (const IntVect& v)
{
  if (!isEmpty())
    {
      smallend -= v;
      bigend   += v;
      if (!(bigend >= smallend)) *this = Box().convert(btype);
      computeBoxLen();
    }
  return *this;
}

inline
Box
grow (const Box&     b,
      const IntVect& v)
{
  if (b.isEmpty())
    {
      return(Box().convert(b.btype));
    }
  else
    {
      IntVect small = b.smallend - v;
      IntVect big   = b.bigend   + v;
      if (!(big >= small))
        {
          return(Box().convert(b.btype));
        }
      else
        {
          return Box(small,big,b.btype);
        }
    }
}

inline
Box&
Box::grow (int idir,
           int n_cell)
{
  if (!isEmpty())
    {
      smallend.shift(idir, -n_cell);
      bigend.shift(idir, n_cell);
      if (!(bigend >= smallend)) *this = Box().convert(btype);
      computeBoxLen();
    }
  return *this;
}

inline
Box&
Box::growLo (int idir,
             int n_cell)
{
  if (!isEmpty())
    {
      smallend.shift(idir, -n_cell);
      if (!(bigend >= smallend)) *this = Box().convert(btype);
      computeBoxLen();
    }
  return *this;
}

inline
Box&
Box::growDir (int idir,
              const Side::LoHiSide& a_sd,
              int n_cell)
{
  if(a_sd == Side::Lo)
    {
      growLo(idir, n_cell);
    }
  else
    {
      growHi(idir, n_cell);
    }
  return *this;
}

inline
Box&
Box::growHi (int idir,
             int n_cell)
{
  if (!isEmpty())
    {
      bigend.shift(idir,n_cell);
      if (!(bigend >= smallend)) *this = Box().convert(btype);
      computeBoxLen();
    }
  return *this;
}


inline
Box
Box::operator& (const Box& rhs) const
{
  Box lhs(*this);
  return lhs &= rhs;
}

#endif /* WRAPPER */

#endif /*CH_BOX_H*/
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

//
// $Id: CFile.cpp,v 1.6 2003/04/21 22:08:00 bvs Exp $
//



using std::cout;
using std::cerr;
using std::endl;
using std::ws;
using std::ostream;
using std::istream;

// const Box Box::Empty = Box(IntVect::TheUnitVector(),
//                            IntVect::TheZeroVector());

IndexType
IndexType::TheCellType ()
{
    static const IndexType Cell(D_DECL(IndexType::CELL,
                                       IndexType::CELL,
                                       IndexType::CELL));
    return Cell;
}

IndexType
IndexType::TheNodeType ()
{
    static const IndexType Node(D_DECL(IndexType::NODE,
                                       IndexType::NODE,
                                       IndexType::NODE));
    return Node;
}

ostream&
operator<< (ostream&         os,
            const IndexType& it)
{
    os << '('
       << D_TERM( (it.test(0)?'N':'C'),
                  << ',' << (it.test(1)?'N':'C'),
                  << ',' << (it.test(2)?'N':'C')) << ')' ;

    if (os.fail())
        MayDay::Error("operator<<(ostream&,IndexType&) failed");

    return os;
}

//
// Copied from <Utility.H>
//
#define CH_IGNORE_MAX 100000

istream&
operator>> (istream&   is,
            IndexType& it)
{
    char D_DECL(t0,t1,t2);

    D_EXPR( is.ignore(CH_IGNORE_MAX, '(') >> t0,
            is.ignore(CH_IGNORE_MAX, ',') >> t1,
            is.ignore(CH_IGNORE_MAX, ',') >> t2);
    is.ignore(CH_IGNORE_MAX, ')');
    D_TERM(
        assert(t0 == 'C' || t0 == 'N'); t0=='N'?it.set(0):it.unset(0); ,
        assert(t1 == 'C' || t1 == 'N'); t1=='N'?it.set(1):it.unset(1); ,
        assert(t2 == 'C' || t2 == 'N'); t2=='N'?it.set(2):it.unset(2));

    if (is.fail())
        MayDay::Error("operator>>(ostream&,IndexType&) failed");

    return is;
}



// const Box&
// Box::TheUnitBox ()
// {
//     static const Box Unit(IntVect::TheZeroVector(), IntVect::TheUnitVector());
//     return Unit;
// }

//
// Administrative functions.
//
Box::Box ()
    : smallend(IntVect::TheUnitVector()),
      bigend(IntVect::TheZeroVector()),
      len(IntVect::Zero),
      btype()
{
}

Box::Box (const IntVect& small,
          const int*     vec_len)
    : smallend(small),
      bigend(D_DECL(small[0]+vec_len[0]-1,
                    small[1]+vec_len[1]-1,
                    small[2]+vec_len[2]-1))
{
  assert(D_TERM(vec_len[0] >= 0, && vec_len[1] >= 0, && vec_len[2] >= 0));

  D_EXPR(len[0] = vec_len[0], len[1] = vec_len[1], len[2] = vec_len[2]);
}

Box::Box (const IntVect&   small,
          const IntVect&   big,
          const IndexType& t)
    : smallend(small),
      bigend(big),
      btype(t)
{
  assert (small <= big);

  computeBoxLen();
}

Box::Box (const IntVect& small,
          const IntVect& big)
    : smallend(small),
      bigend(big)
{
  assert (small <= big);

  computeBoxLen();
}

void Box::define (const IntVect& small,
				  const IntVect& big)
{
  smallend = small;
  bigend=big;
  assert (small <= big);
  computeBoxLen();
}

Box::Box (const IntVect& small,
          const IntVect& big,
          const IntVect& typ)
    : smallend(small),
      bigend(big),
      btype(typ)
{
  assert (small <= big);
  assert(typ >= IntVect::TheZeroVector() && typ <= IntVect::TheUnitVector());

  computeBoxLen();
}

bool
Box::numPtsOK (long& N) const
{
    N = len[0];

    for (int i = 1; i < CH_SPACEDIM; i++)
    {
        if (len[i] == 0)
        {
            N = 0;
            return true;
        }
        else if (N <= LONG_MAX/len[i])
        {
            N *= len[i];
        }
        else
        {
            //
            // The return value `N' will be undefined.
            //
            return false;
        }
    }

    return true;
}

long
Box::numPts () const
{
  if (isEmpty()) return 0;

  long result;
  if (!numPtsOK(result))
    {
      cerr << "Bad Box::numPts:  box = " << *this << endl;
      MayDay::Error("Arithmetic overflow in Box::numPts()");
    }
  return result;
}

bool
Box::volumeOK (long& N) const
{
  return(numPtsOK(N));
}

long
Box::volume () const
{
  return(numPts());
}

Box&
Box::shiftHalf (int dir,
                int nzones)
{
  if (!isEmpty())
  {
    int nbit = (nzones<0 ? -nzones : nzones)%2;
    int nshift = nzones/2;
    unsigned int bit_dir = btype[dir];
    //
    // Toggle btyp bit if nzones is odd.
    //
    if (nbit)
        btype.flip(dir);
    if (nzones < 0)
        nshift -= (bit_dir ? nbit : 0);
    else
        nshift += (bit_dir ? 0 : nbit);
    smallend.shift(dir,nshift);
    bigend.shift(dir,nshift);
  }
  return *this;
}

//
// Boolean functions.
//

bool
Box::intersects (const Box& b) const
{
  if(b.isEmpty() || isEmpty()) return false;
  assert(sameType(b));
  IntVect low(smallend);
  IntVect hi(bigend);
  low.max(b.smallend);
  hi.min(b.bigend);
  return low <= hi;
}

// do boxes intersect? 
bool Box::intersectsNotEmpty(const Box& b) const
{
  IntVect low(smallend);
  IntVect hi(bigend);
  low.max(b.smallend);
  hi.min(b.bigend);
  return low <= hi;
}

//
// Intersection functions.
//

Box&
Box::operator&= (const Box& b)
{

  assert(sameType(b));
  if (isEmpty() || b.isEmpty())
  {
    *this = Box().convert(b.btype);
  }
  else
  {
    smallend.max(b.smallend);
    bigend.min(b.bigend);
    if (!(bigend >= smallend)) *this = Box().convert(b.btype);
    else computeBoxLen(); // I'd like to stop doing this....bvs
  }

  return *this;
}

Box&
Box::surroundingNodes ()
{
  if (!isEmpty())
  {
    for (int i = 0; i < CH_SPACEDIM; ++i)
        if ((btype[i] == 0))
            bigend.shift(i,1);
  }
    btype.setall();
    computeBoxLen();
    return *this;
}

Box&
Box::enclosedCells ()
{
  if (!isEmpty())
  {
    for (int i = 0 ; i < CH_SPACEDIM; ++i)
        if (btype[i])
            bigend.shift(i,-1);
  }
    btype.clear();
    computeBoxLen();
    return *this;
}

//
// Next:step through the rectangle with unit increment.
//

void
Box::next (IntVect& p) const
{
    assert(contains(p));

    p.shift(0,1);
#if CH_SPACEDIM==2
    if (!(p <= bigend))
    {
        p.setVal(0,smallend[0]);
        p.shift(1,1);
    }
#elif CH_SPACEDIM==3
    if (!(p <= bigend))
    {
        p.setVal(0,smallend[0]);
        p.shift(1,1);
        if (!(p <= bigend))
        {
            p.setVal(1,smallend[1]);
            p.shift(2,1);
        }
    }
#endif
}

//
// Scan point over region of object Box with a vector incrment
// point incrments by 0 direction portion of vector.  When end of
// Box is reached, an increment is made with the 1 direction portion
// of the vector, and the 0 direction scan is resumed.
// effectively, we are scanning a Box, whose length vector is the argument
// vector over the object Box.
// when scan is over, the argument point is over edge of object Box
// this is the signal that we can go no further.
//

void
Box::next (IntVect&   p,
           const int* shv) const
{
    assert(contains(p));

#if   CH_SPACEDIM==1
    p.shift(0,shv[0]);
#elif CH_SPACEDIM==2
    p.shift(0,shv[0]);
    if (!(p <= bigend))
    {
        //
        // Reset 1 coord is on edge, and 2 coord is incremented.
        //
        p.setVal(0,smallend[0]);
        p.shift(1,shv[1]);
    }
#elif CH_SPACEDIM==3
    p.shift(0,shv[0]);
    if (!(p <= bigend))
    {
        //
        // Reset 1 coord is on edge, and 2 coord is incremented.
        //
        p.setVal(0,smallend[0]);
        p.shift(1,shv[1]);
        if(!(p <= bigend))
        {
            p.setVal(1,smallend[1]);
            p.shift(2,shv[2]);
        }
    }
#endif
}

Box&
Box::refine (int refinement_ratio)
{
  assert(refinement_ratio > 0);
  if (!isEmpty())
  {
    IntVect shft(IntVect::TheUnitVector());
    shft -= btype.ixType();
    smallend.scale(refinement_ratio);
    bigend += shft; // Bigend does more than just multiply.
    bigend.scale(refinement_ratio);
    bigend -= shft;
    computeBoxLen();
  }
    return *this;
}

Box&
Box::refine (const IntVect& refinement_ratio)
{
  if (!isEmpty())
  {
    IntVect shft(IntVect::TheUnitVector());
    shft -= btype.ixType();
    smallend *= refinement_ratio;
    bigend += shft;
    bigend   *= refinement_ratio;
    bigend -= shft;
    computeBoxLen();
  }
    return *this;
}

//
// Define a macro which will compute an object's length vector from
// the smallend and bigend.  Do several versions according to dimension
// requires you to be in a member functions.

int
Box::longside () const
{
    int maxlen = len[0];
    for (int i = 1; i < CH_SPACEDIM; i++)
        if (len[i] > maxlen)
            maxlen = len[i];
    return maxlen;
}

int
Box::longside (int& dir) const
{
    int maxlen = len[0];
    dir = 0;
    for (int i = 1; i < CH_SPACEDIM; i++)
    {
        if (len[i] > maxlen)
        {
            maxlen = len[i];
            dir = i;
        }
    }
    return maxlen;
}

int
Box::shortside () const
{
    int minlen = len[0];
    for (int i = 1; i < CH_SPACEDIM; i++)
        if(len[i] < minlen)
            minlen = len[i];
    return minlen;
}

int
Box::shortside (int& dir) const
{
    int minlen = len[0];
    dir = 0;
    for (int i = 1; i < CH_SPACEDIM; i++)
    {
        if(len[i] < minlen)
        {
            minlen = len[i];
            dir = i;
        }
    }
    return minlen;
}

//
// Modified Box is low end, returned Box is high end.
//

Box
Box::chop (int dir,
           int chop_pnt)
{
  assert(!isEmpty());
  //
  // Define new high end Box including chop_pnt.
  //
  IntVect sm(smallend);
  IntVect bg(bigend);
  sm.setVal(dir,chop_pnt);
  if (btype[dir])
  {
    //
    // NODE centered Box.
    //
    assert(chop_pnt > smallend[dir] && chop_pnt < bigend[dir]);
    //
    // Shrink original Box to just contain chop_pnt.
    //
    bigend.setVal(dir,chop_pnt);
  }
  else
  {
    //
    // CELL centered Box.
    //
    assert(chop_pnt > smallend[dir] && chop_pnt <= bigend[dir]);
    //
    // Shrink origional Box to one below chop_pnt.
    //
    bigend.setVal(dir,chop_pnt-1);
  }
  computeBoxLen();
  return Box(sm,bg,btype);
}

//
// Shift by half increments.
//

Box&
Box::shiftHalf (const IntVect& nz)
{
  if (!isEmpty())
  {
   for (int dir = 0; dir < CH_SPACEDIM; dir++)
   {
       int nzones = nz[dir];
       int nbit = (nzones<0 ? -nzones : nzones)%2;
       int nshift = nzones/2;
       unsigned int bit_dir = btype[dir];
       //
       // Toggle btype bit if nzones is odd.
       //
       if (nbit)
           btype.flip(dir);
       if (nzones < 0)
           nshift -= (bit_dir ? nbit : 0);
       else
           nshift += (bit_dir ? 0 : nbit);
       smallend.shift(dir,nshift);
       bigend.shift(dir,nshift);
   }
  }
   return *this;
}

Box& Box::shift_intvect (const IntVect& iv)
{
  return shift(iv);
}

Box& Box::enclosedCells_int (int dir)
{
  return enclosedCells(dir);
}

Box& Box::surroundingNodes_int(int dir)
{
  return surroundingNodes(dir);
}

bool Box::lt(const Box& rhs) const
{
  return *this < rhs;
}

bool Box::eq(const Box& b) const
{
  return *this == b;
}
bool Box::neq(const Box& b) const
{
  return *this != b;
}

Box& Box::shiftHalf_intvect (const IntVect& iv)
{
  return shiftHalf(iv);
}


Box&
Box::convert (int                  dir,
              IndexType::CellIndex typ)
{
  if (!isEmpty())
  {
   unsigned int bitval = btype[dir];
   int off = typ - bitval;
   bigend.shift(dir,off);
   if (off != 0)
      computeBoxLen();
  }
   //
   // Set dir'th bit to typ.
   //
   btype.setType(dir,typ);
   return *this;
}

Box&
Box::convert (IndexType t)
{
   for (int dir = 0; dir < CH_SPACEDIM; dir++)
   {
      unsigned int typ = t[dir];
      if (!isEmpty())
      {
        unsigned int bitval = btype[dir];
        int off = typ - bitval;
        bigend.shift(dir,off);
      }
      btype.setType(dir, (IndexType::CellIndex) typ);
   }
   computeBoxLen();
   return *this;
}

//
// Refinement functions.
//

Box
refine (const Box& b,
        int        refinement_ratio)
{
  assert(refinement_ratio > 0);
  if (b.isEmpty()) return(b);

    IntVect small(b.smallend);
    small.scale(refinement_ratio);
    IntVect big(b.bigend);
    IntVect shft(IntVect::TheUnitVector());
    shft -= b.btype.ixType();
    big += shft;        // Large end is not just multiply.
    big.scale(refinement_ratio);
    big -= shft;
    return Box(small,big,b.btype);
}

Box
refine (const Box&     b,
        const IntVect& refinement_ratio)
{
  if (b.isEmpty()) return(b);

    IntVect small(b.smallend);
    small *= refinement_ratio;
    IntVect big(b.bigend);
    IntVect shft(IntVect::TheUnitVector());
    shft -= b.btype.ixType();
    big += shft;
    big *= refinement_ratio;
    big -= shft;
    return Box(small,big,b.btype);
}

//
// Coarsening.
//

Box&
Box::coarsen (const IntVect& refinement_ratio)
{
  if (!isEmpty())
  {
    smallend.coarsen(refinement_ratio);

    if (btype.any())
    {
        IntVect off(IntVect::TheZeroVector());
        for (int dir = 0; dir < CH_SPACEDIM; dir++)
        {
            if (btype[dir])
            {
                int b = bigend[dir];
                int r = refinement_ratio[dir];
                if (b%r)
                    off.setVal(dir,1);
            }
        }
        bigend.coarsen(refinement_ratio);
        bigend += off;
    }
    else
        bigend.coarsen(refinement_ratio);

    computeBoxLen();
  }
    return *this;
}

Box
coarsen (const Box& b,
         int  refinement_ratio)
{
  if (b.isEmpty()) return(b);

    IntVect small(b.smallend);
    small.coarsen(refinement_ratio);
    IntVect big(b.bigend);

    if (b.btype.any())
    {
        IntVect off(IntVect::TheZeroVector());
        for (int dir = 0; dir < CH_SPACEDIM; dir++)
        {
            if (b.btype[dir])
                if (big[dir]%refinement_ratio)
                    off.setVal(dir,1);
        }
        big.coarsen(refinement_ratio);
        big += off;
    }
    else
        big.coarsen(refinement_ratio);

    return Box(small,big,b.btype);
}

Box&
Box::coarsen (int refinement_ratio)
{
  if (!isEmpty())
  {
    smallend.coarsen(refinement_ratio);
    if (btype.any())
    {
        IntVect off(IntVect::TheZeroVector());
        for (int dir = 0; dir < CH_SPACEDIM; dir++)
        {
            if (btype[dir])
                if (bigend[dir]%refinement_ratio)
                    off.setVal(dir,1);
        }
        bigend.coarsen(refinement_ratio);
        bigend += off;
    }
    else
        bigend.coarsen(refinement_ratio);

    computeBoxLen();
  }
    return *this;
}

Box
coarsen (const Box&     b,
         const IntVect& refinement_ratio)
{
  if (b.isEmpty()) return(b);

    IntVect small(b.smallend);
    small.coarsen(refinement_ratio);
    IntVect big(b.bigend);

    if (b.btype.any())
    {
        IntVect off(IntVect::TheZeroVector());
        for (int dir = 0; dir < CH_SPACEDIM; dir++)
        {
            if (b.btype[dir])
                if (big[dir]%refinement_ratio[dir])
                    off.setVal(dir,1);
        }
        big.coarsen(refinement_ratio);
        big += off;
    }
    else
        big.coarsen(refinement_ratio);

    return Box(small,big,b.btype);
}

//
// I/O functions.
//

ostream&
operator<< (ostream&   os,
            const Box& b)
{
    os << '('
       << b.smallend << ' '
       << b.bigend   << ' '
       << b.btype.ixType()
       << ')';

    if (os.fail())
        MayDay::Error("operator<<(ostream&,Box&) failed");

    return os;
}

void Box::p() const
{
  pout() << *this << "\n";
}

//
// Moved out of Utility.H
//
#define CH_IGNORE_MAX 100000

istream&
operator>> (istream& is,
            Box&     b)
{
    is >> ws;
    char c;
    is >> c;
    is.putback(c);
    if (c == '(')
    {
        is.ignore(CH_IGNORE_MAX, '(');
        is >> b.smallend ;
        is >> b.bigend ;
        IntVect v;
        is >> v;
        b.btype = IndexType(v);
        assert(b.btype.ok());
        is.ignore(CH_IGNORE_MAX,')');
        b.computeBoxLen();
    }
    else if (c == '<')
    {
        is >> b.smallend;
        is >> b.bigend;
        IntVect v;
        is >> v;
        b.btype = IndexType(v);
        assert(b.btype.ok());
        b.computeBoxLen();
    }
    else
        MayDay::Error("operator>>(istream&,Box&): expected \'<\'");

    if (is.fail())
        MayDay::Error("operator>>(istream&,Box&) failed");

    return is;
}

void
Box::dumpOn (ostream& strm) const
{
    strm << "Box ("
         << MayDay::version
         << ")"
         << smallend
         << " to "
         << bigend
         << " type ["
         << btype.ixType()
         << "]"
         << '\n';

    if (strm.fail())
        MayDay::Error("Box::dumpOn(ostream&) failed");
}

//
// Give smallest Box containing both Boxes.
//

Box&
Box::minBox (const Box &b)
{
  assert(sameType(b));
  if (isEmpty())
  {
    *this = b;
  }
  else if (!b.isEmpty())
  {
    smallend.min(b.smallend);
    bigend.max(b.bigend);
    computeBoxLen();
  }
  return *this;
}

Box
minBox (const Box& b,
        const Box& o)
{
    assert(o.sameType(b));
    if (b.isEmpty())
    {
      return(o);
    }
    else if (o.isEmpty())
    {
      return(b);
    }
    else
    {
      IntVect small = b.smallend;
      IntVect big = b.bigend;
      small.min(o.smallend);
      big.max(o.bigend);
      return Box(small,big,b.btype);
    }
}

//
// Translation functions acting on relative vectors.
//

Box
bdryBox (const Box& b,
         int        dir,
         Side::LoHiSide a_side,
         int        len)
{
  Box retval;
  if(a_side == Side::Lo)
    retval = bdryLo(b, dir, len);
  else
    retval = bdryHi(b, dir, len);

  return retval;
}


Box
bdryLo (const Box& b,
        int        dir,
        int        len)
{
  //
  // set dir'th bit to 1 = IndexType::NODE.
  //
  IndexType typ(b.btype);
  typ.set(dir);
  if (b.isEmpty())
  {
    return(Box().convert(typ));
  }
  else
  {
    IntVect low(b.smallend);
    IntVect hi(b.bigend);
    int sm = low[dir];
    hi.setVal(dir,sm+len-1);
    return Box(low,hi,typ);
  }
}

Box
bdryHi (const Box& b,
        int        dir,
        int        len)
{
  //
  // Set dir'th bit to 1 = IndexType::NODE.
  //
  IndexType typ(b.btype);
  typ.set(dir);
  if (b.isEmpty())
  {
    return(Box().convert(typ));
  }
  else
  {
    IntVect low(b.smallend);
    IntVect hi(b.bigend);
    unsigned int bitval = b.btype.ixType(dir);
    int bg              = hi[dir]  + 1 - bitval%2;
    low.setVal(dir,bg);
    hi.setVal(dir,bg+len-1);
    return Box(low,hi,typ);
  }
}


Box
adjCellBox (const Box& b,
            int        dir,
            Side::LoHiSide a_side,
            int        len)
{
  Box retval;
  if(a_side == Side::Lo)
    retval = adjCellLo(b, dir, len);
  else
    retval = adjCellHi(b, dir, len);

  return retval;
}

Box
adjCellLo (const Box& b,
           int        dir,
           int        len)
{
  assert(len > 0);
  //
  // Set dir'th bit to 0 = IndexType::CELL.
  //
  IndexType typ(b.btype);
  typ.unset(dir);
  if (b.isEmpty())
  {
    return(Box().convert(typ));
  }
  else
  {
    IntVect low(b.smallend);
    IntVect hi(b.bigend);
    int sm = low[dir];
    low.setVal(dir,sm - len);
    hi.setVal(dir,sm - 1);
    return Box(low,hi,typ);
  }
}

Box
adjCellHi (const Box& b,
           int        dir,
           int        len)
{
  assert(len > 0);
  //
  // Set dir'th bit to 0 = IndexType::CELL.
  //
  IndexType typ(b.btype);
  typ.unset(dir);
  if (b.isEmpty())
  {
    return(Box().convert(typ));
  }
  else
  {
    IntVect low(b.smallend);
    IntVect hi(b.bigend);
    unsigned int bitval = b.btype.ixType(dir);
    int bg              = hi[dir]  + 1 - bitval%2;
    low.setVal(dir,bg);
    hi.setVal(dir,bg + len - 1);
    return Box(low,hi,typ);
  }
}


/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef CH_PROBDOMAIN_H
#define CH_PROBDOMAIN_H

#ifndef WRAPPER

#endif


/// clase to manage box-shifting used to enforce periodic BC's
/** The ShiftIterator class contains a list of shift vectors necessary
    to enforce periodic boundary conditions (enforced by shifting the 
    Box of each Fab, then copying any overlapping valid cells with
    ghost cells of the shifted Fab.  Depending on the number of periodic 
    directions, the list of shift vectors over which to iterate differs.
*/
class ShiftIterator
{
public:
  /// basic constructor
  ShiftIterator();

  /// 
  ShiftIterator(const bool* a_isPeriodic);

  /// copy constructor
  ShiftIterator(const ShiftIterator& a_shiftIt);

  /// 
  ~ShiftIterator();

  ///
  ShiftIterator& operator=(const ShiftIterator& a_src);

  /// recompute shift vectors based on periodic directions
  void computeShifts(const bool* a_isPeriodic);

  /// returns the current shift unit vector
  inline IntVect operator()() const;
  IntVect i() const {return this->operator()();}

  /// increment to the next shift unit vector
  inline void operator++();
  void incr(){++(*this);}

  ///
  inline bool ok() const;

  /// reset to first shift unit vector
  inline void reset();
  
  /// 
  inline void begin();

  /// 
  /** skip the iterator to the end.  ok() will return false after this
      method is called.
  */
  void end();

private:

  unsigned int m_index;

  Vector<IntVect> m_shift_vectors;

};



//
//@Man:
//@Memo: A class to facilitate interaction with physical boundary conditions
/*@Doc: 

  ProblemDomain is a class which facilitates the application of physical 
  boundary conditions, both periodic and non-periodic.  This class contains 
  much of the functionality of the Box class, since logically the 
  computational domain is generally a Box.  
  
  Intersection with a ProblemDomain object will result in only removing 
  regions which are outside the physical domain in non-periodic directions.
  Regions outside the logical computational domain in periodic directions will
  be treated as ghost cells which can be filled with an exchange() function 
  or through suitable interpolation from a coarser domain.

  Since ProblemDomain will contain a Box, it is a dimension dependent class, 
  so SpaceDim must be defined as either 1, 2, or 3 when compiling.  

  Note that this implementation of ProblemDomain is inherently 
  cell-centered.


*/
class ProblemDomain
{
public:

  /// {\bf Constructors}

  ///
  /**
     The default constructor.  The constructed domain box is empty. 

  */
    ProblemDomain ();

  ///
  /**
     Construct ProblemDomain with a_domBox as computational domain
     This constructor defaults to non-periodic domain
  */
  ProblemDomain(const Box& a_domBox);

  ///
  /**
     Construct ProblemDomain with a_domBox as computational domain.
     a_isPeriodic is a SpaceDim array of bools, true is the physical
     boundary condition is periodic in the coordinate direction
  */
  ProblemDomain(const Box& a_domBox, const bool* a_isPeriodic);

  ///
  /**
     Construct a ProblemDomain.  It is an error if small is
     greater than big.

  */
    ProblemDomain (const IntVect& small,
                   const IntVect& big);


  ///
  /**
     Construct a ProblemDomain.  It is an error if small is
     greater than big.
     a_isPeriodic is a SpaceDim array of bools, true is the physical
     boundary condition is periodic in the coordinate direction

  */
    ProblemDomain (const IntVect& small,
                   const IntVect& big,
                   const bool* a_isPeriodic);


  ///
  /**
     Construct ProblemDomain with specified lengths.  It is an error
    if the lengths are negative.

  */
    ProblemDomain (const IntVect& small,
         const int*     vec_len);


  ///
  /**
     Construct ProblemDomain with specified lengths.  It is an error
    if the lengths are negative.
     a_isPeriodic is a SpaceDim array of bools, true is the physical
     boundary condition is periodic in the coordinate direction

  */
    ProblemDomain (const IntVect& small,
                   const int*     vec_len,
                   const bool* a_isPeriodic);


  ///
  /**
     The copy constructor.

  */
    ProblemDomain (const ProblemDomain& a_src);

  /// {\bf Accessors}

  ///
  /** 
      Returns the logical computational domain
  */
  const Box& domainBox() const;
  

  ///
  /**
     returns true if BC is periodic in direction a_dir
  */
  bool isPeriodic(int a_dir) const;

  ///
  /**
     returns true is BC is periodic in _any_ direction
  */
  bool isPeriodic() const;
  
  
  /// 
  /** 
      returns the shiftIterator for this ProblemDomain
  */
  ShiftIterator shiftIterator() const;


  ///
  /**
     Returns true if this ProblemDomain is empty or undefined.     
  */
  bool isEmpty () const;
  
  ///
  /**
     Returns true if argument is contained within this ProblemDomain.  An empty
     ProblemDomain does not contain and is not contained by any ProblemDomain, 
     including itself.    
  */
  bool contains (const IntVect& p) const;
  ///
  /**
	 Returns the periodic image of this IntVect inside of the ProblemDomain.
	 Return true if the domain contains this IntVect, returning the image
	 in 'p', otherwise returns false
  */
  bool image (IntVect& p) const;
  
  ///
  /**
     Returns true if argument is contained within this ProblemDomain.  It 
     is an error if the Box is not cell-centered.  An empty
     ProblemDomain does not contain any Box.
  */
    bool contains (const Box& b) const;
  bool contains_box(const Box& b) const {return contains(b);}

  /// 
  /**
     Returns true if this ProblemDomain and the argument have non-null
     intersections.  It is an error if a_box is not cell-centered.
     An empty ProblemDomain does not intersect any Box.
     This will do nothing in periodic directions (since a periodic domain 
     is infinite in the periodic direction.  If periodic in all dimensions, 
     this will always return true.
  */
  bool intersects (const Box& a_box) const;

  /// 
  /**
     Returns true if this ProblemDomain and the argument have non-null
     intersections.  It is an error if a_box is not cell-centered.
     This routine does not perform the check to see if *this or b are
     empty boxes.  It is the callers responsibility to ensure that
     this never happens.  If you are unsure, the use the .intersects(..) 
     routine.  In periodic directions, will always return true.

*/
    bool intersectsNotEmpty (const Box& a_box) const;


  /// 
  /** returns true if box1 and box2 and any of their periodic 
      images intersect (useful for checking disjointness)
  */ 
  bool intersects(const Box& box1, const Box& box2) const;

  ///
  /** modifies a_box to be the intersection of a_box and the 
      problemDomain
  */
  friend void operator &=(Box& a_box, const ProblemDomain& a_probomain);

  ///
  /**  returns a box which is the interesection of a_box and the 
       problemDomain
  */
  friend Box operator & (const Box& a_box, const ProblemDomain& a_probdomain);

  /// {\bf Modification Functions}

  ///
  /**
    The assignment operator.

  */
  ProblemDomain& operator= (const ProblemDomain& b);

  ///
  /** 
      sets whether BC is periodic in direction a_dir (true == periodic)
  */
  void setPeriodic(int a_dir, bool a_isPeriodic);
  
  ///
  /** grows (or shrinks) the domain box by i in all directions
   */
  inline ProblemDomain& grow (int i);

  ///
  /** returns a ProblemDomain with a domainBox grown by the given amount
   */
  friend inline ProblemDomain grow(const ProblemDomain& pd, 
                                   int i);

  ///
  /**  
       modifies this ProblemDomain by growing the domainBox in each
       direction by the specified amount
  */
  inline ProblemDomain& grow(const IntVect& v);

  ///
  /**
     Returns a ProblemDomain that is the argument ProblemDomain
     with a DomainBox grown by the given amount.
  */
  friend inline ProblemDomain grow (const ProblemDomain& pd,
                                     const IntVect& v);

  ///
  /** Modifies this ProblemDomain by growing it on the low and high end
      by n_cell cells in direction idir
  */
  inline ProblemDomain& grow(int idir, int n_cell);

  ///
  /** Modifies this ProblemDomain by growing it on the low end by n_cell
      cells in direction idir.
  */
  inline ProblemDomain& growLo(int idir, int n_cell=1);

  ///
  /** Modifies this ProblemDomain by growing it on the high end by n_Cell
      cells in direction idir
  */
  inline ProblemDomain& growHi(int idir, int n_cell=1);
     


  ///
  /**
     Returns the edge-centered Box (in direction dir) defining the low
     side of the argument ProblemDomain.  The output Box will have the given
     length in the given direction.  Directions are zero-based.  It is
     an error if not 0 <= dir < SpaceDim.  The neighbor of an Empty ProblemDomain
     is an Empty Box of the appropriate type.  If dir is a periodic
     direction, will return an empty box.
     
  */
  friend Box bdryLo (const ProblemDomain& a_pd,
                     int        a_dir,
                     int        a_len=1);
  
  /// 
  /**
     Returns the edge-centered Box (in direction dir) defining the
     high side of the argument ProblemDomain.  The return Box will have the
     given length in the given direction.  Directions are zero-based.
     It is an error if not 0 <= dir < SpaceDim.  The neighbor of an
     Empty ProblemDomain is an Empty Box of the appropriate type.  
     If dir is a periodic direction, will return an empty box.
     
  */
  friend Box bdryHi (const ProblemDomain& a_pd,
                     int        a_dir,
                     int        a_len=1);
  
  ///
  /**
     Returns the cell centered Box of the given length adjacent to the
     argument ProblemDomain on the low end along the given coordinate direction.
     The return Box is identical to the argument ProblemDomain in the other
     directions.  The return ProblemDomain and the argument ProblemDomain have an 
     empty intersection.\\

     NOTE: len >= 1. \\

     NOTE: Box retval = adjCellLo(b,dir,len) is equivalent to the
     following set of operations: \\ 

     Box retval(b); \\

     retval.convert(dir,ProblemDomain::CELL); \\

     retval.setrange(dir,retval.smallEnd(dir)-len,len); \\

     Directions are zero-based.  It is an error if not 0 <= dir <
     SpaceDim.  The neighbor of an Empty ProblemDomain is an Empty Box of the
     appropriate type. If dir is a periodic direction, will return
     an empty Box as well.

  */
  friend Box adjCellLo (const ProblemDomain& a_pd,
                        int        a_dir,
                        int        a_len=1);
  
  ///
  /**
     Returns the cell centered Box of the given length adjacent to the
     argument ProblemDomain on the high end along the given coordinate
     direction.  The return Box is identical to the argument ProblemDomain in
     the other directions.  The return Box and the argument ProblemDomain have
     an empty intersection.  \\
     
     NOTE: len >= 1. \\
     
     NOTE: Box retval = adjCellHi(b,dir,len) is equivalent to the
     following set of operations: \\ 
     
     Box retval(b); \\
     
     retval.convert(dir,ProblemDomain::CELL); \\
     
     retval.setrange(dir,retval.bigEnd(dir)+1,len);\\
     
     Directions are zero-based.  It is an error if not 0 <= dir <
     SpaceDim.  The neighbor of an Empty ProblemDomain is an Empty Box of the
     appropriate type.    If dir is a periodic direction, will return
     an empty Box.
     
  */
  friend Box adjCellHi (const ProblemDomain& a_pd,
                        int        a_dir,
                        int        a_len=1);
  
  
  /// intersection functions 
  
  ///
  /**
     Returns the Box that is intersection of this ProblemDomain and the argument
     ProblemDomain.  The Box MUST be cell-centered.
     The intersection of the Empty ProblemDomain and any Box is the Empty
     Box. This operator does nothing in periodic directions (since
     a periodic domain is an infinite domain).
     
  */
  Box operator& (const Box& a_b) const;
  



  /// refinement

  ///
  /**
     Modifies this ProblemDomain by refining it by given (positive) refinement
     ratio.  The Empty ProblemDomain is not modified by this function.  \\

  */
    ProblemDomain& refine (int a_refinement_ratio);

  ///
  /**
     Returns a ProblemDomain that is the argument ProblemDomain refined by given
     (positive) refinement ratio.  The Empty ProblemDomain is not modified by
     this function.  \\

  */
  friend ProblemDomain refine (const ProblemDomain& a_probdomain,
                               int   a_refinement_ratio);
  
  ///
  /**
     Modifies this ProblemDomain by refining it by given (positive) refinement
     ratio.  The Empty ProblemDomain is not modified by this function.  \\
  */
  ProblemDomain& refine (const IntVect& a_refinement_ratio);
  
  ///
  /**
     Returns a ProblemDomain that is the argument ProblemDomain refined 
     by given (positive) refinement ratio.  The Empty ProblemDomain is 
     not modified by this function.  \\
     
  */
  friend ProblemDomain refine (const ProblemDomain&     a_probdomain,
                            const IntVect& a_refinement_ratio);

  /// coarsening
  
  ///
  /**
     Modifies this ProblemDomain by coarsening it by given (positive) 
     refinement ratio.  The Empty ProblemDomain is not modified by 
     this function.  \\
  */
  
  ProblemDomain& coarsen (int a_refinement_ratio);
  
  ///
  /**
     Returns a ProblemDomain that is the argument ProblemDomain coarsened by given
     (positive) refinement ratio.  The Empty ProblemDomain is not modified by
     this function.  \\
     
  */
  friend ProblemDomain coarsen (const ProblemDomain& a_probdomain,
                             int        a_refinement_ratio);
  
  ///
  /**
     Modifies this ProblemDomain by coarsening by given (positive) refinement
     ratio.  The Empty ProblemDomain is not modified by this function.  \\
  */
  ProblemDomain& coarsen (const IntVect& refinement_ratio);
  
  ///
  /**
     Returns a ProblemDomain that is the argument ProblemDomain coarsened 
     by given (positive) refinement ratio.  The Empty ProblemDomain is 
     not modified by this function.  \\
     
  */
  friend ProblemDomain coarsen (const ProblemDomain&  a_probdomain,
                             const IntVect& a_refinement_ratio);
  


  ///{\bf I/O Functions}
  
  ///
  /**
     Write an ASCII representation to the ostream.
     
  */
  friend std::ostream& operator<< (std::ostream&   os,
                                   const ProblemDomain& bx);

  ///
  /**
     Read from istream.
     
  */
  friend std::istream& operator>> (std::istream& is,
                                   ProblemDomain&     bx);
  
  ///
  /**
     Gives more detail than printOn.  Useful for exiting due to an
     error.  
     
  */
  void dumpOn (std::ostream& strm) const;
  
  
protected:
  friend class HDF5Handle;

  bool m_isPeriodic[SpaceDim];

  Box m_domainBox;

  ShiftIterator m_shiftIt;



};

//
// Inlines.
//
#ifndef WRAPPER

inline
ShiftIterator::ShiftIterator()
  : m_index(100), m_shift_vectors()
{
}

inline 
ShiftIterator::ShiftIterator(const ShiftIterator& a_src)
{
  m_index = a_src.m_index;
  m_shift_vectors = a_src.m_shift_vectors;
}

inline
ShiftIterator&
ShiftIterator::operator=(const ShiftIterator& a_src)
{
  m_index = a_src.m_index;
  m_shift_vectors = a_src.m_shift_vectors;
  return *this;
}

inline
IntVect
ShiftIterator::operator()() const
{
  assert(ok());
  return m_shift_vectors[m_index];
}

inline
void
ShiftIterator::operator++() 
{
  m_index++;
}

inline 
bool
ShiftIterator::ok() const
{
  return (m_index < m_shift_vectors.size());
}

inline 
void
ShiftIterator::reset()
{
  m_index = 0;
}

inline
void
ShiftIterator::begin()
{
  m_index = 0;
}

inline 
void
ShiftIterator::end()
{
  m_index = m_shift_vectors.size();
}

inline
ProblemDomain::ProblemDomain ()
{
  // default is a non-periodic domain
  for (int dir=0; dir<SpaceDim; dir++) 
    {
      m_isPeriodic[dir] = false;
    }
  
}

inline
ProblemDomain::ProblemDomain (const ProblemDomain& b)
  : m_domainBox(b.m_domainBox), m_shiftIt(b.m_shiftIt)
{  
  for (int dir=0; dir<SpaceDim; dir++) 
    {
      m_isPeriodic[dir] = b.m_isPeriodic[dir];
    }
}


inline
ProblemDomain&
ProblemDomain::operator= (const ProblemDomain& b)
{  
  m_domainBox = b.m_domainBox;
  for (int dir=0; dir<SpaceDim; dir++) 
    {
      m_isPeriodic[dir] = b.m_isPeriodic[dir];
    }
  m_shiftIt = b.m_shiftIt;
  return *this;
}

inline 
const Box& 
ProblemDomain::domainBox() const
{
  return m_domainBox;
}

inline
bool
ProblemDomain::isPeriodic(int a_dir) const
{
  return m_isPeriodic[a_dir];
}

inline
bool
ProblemDomain::isPeriodic() const
{
  return D_TERM(m_isPeriodic[0],
                || m_isPeriodic[1],
                || m_isPeriodic[2]);
}

inline
ProblemDomain& 
ProblemDomain::grow(int i)
{
  m_domainBox.grow(i);
  return *this;
}

inline 
ProblemDomain 
grow(const ProblemDomain& pd, int i)
{
  ProblemDomain newPd(pd);
  newPd.grow(i);
  return newPd;
}

inline
ProblemDomain&
ProblemDomain::grow(const IntVect& v)
{
  m_domainBox.grow(v);
  return *this;
}

inline
ProblemDomain
grow(const ProblemDomain& pd, const IntVect& v)
{
  ProblemDomain newPd(pd);
  newPd.grow(v);
  return newPd;
}

inline
ProblemDomain&
ProblemDomain::grow(int idir, int n_cell)
{
  m_domainBox.grow(idir, n_cell);
  return *this;
}

inline 
ProblemDomain&
ProblemDomain::growLo(int idir, int n_cell)
{
  m_domainBox.growLo(idir, n_cell);
  return *this;
}

inline 
ProblemDomain&
ProblemDomain::growHi(int idir, int n_cell)
{
  m_domainBox.growHi(idir, n_cell);
  return *this;
}

inline 
ShiftIterator
ProblemDomain::shiftIterator() const
{
  return m_shiftIt;
}

inline
bool
ProblemDomain::isEmpty () const
{
  return (m_domainBox.isEmpty());
}

inline
bool
ProblemDomain::contains (const IntVect& p) const
{

  // boy is this ugly!
  return ( !isEmpty() && 
           (D_TERM((m_isPeriodic[0] 
                    || (p[0]>m_domainBox.smallEnd(0) 
                        && p[0]<m_domainBox.bigEnd(0))),
                   && (m_isPeriodic[1] 
                       || (p[1]>m_domainBox.smallEnd(1) 
                           && p[1]<m_domainBox.bigEnd(1))),
                   && (m_isPeriodic[2]
                       || (p[2]>m_domainBox.smallEnd(2)
                           && p[2]<m_domainBox.bigEnd(2))))));
}
                     
inline
bool
ProblemDomain::image(IntVect& p) const
{
  if(m_domainBox.contains(p)) return true;
  if(!contains(p)) return false;

  if(m_isPeriodic[0]){
	if(p[0]<m_domainBox.smallEnd(0))    p[0]+= m_domainBox.size(0);
	else if(p[0]>m_domainBox.bigEnd(0)) p[0]-= m_domainBox.size(0);
  }
  if(m_isPeriodic[1]){
	if(p[1]<m_domainBox.smallEnd(1))    p[1]+= m_domainBox.size(1);
	else if(p[1]>m_domainBox.bigEnd(1)) p[1]-= m_domainBox.size(1);
  }
  
#if (CH_SPACEDIM == 3)
  if(m_isPeriodic[2]){
	if(p[2]<m_domainBox.smallEnd(2))    p[2]+= m_domainBox.size(2);
	else if(p[2]>m_domainBox.bigEnd(2)) p[2]-= m_domainBox.size(2);
  }
#endif
  return true;
}


inline
bool
ProblemDomain::contains (const Box& b) const
{
  // boy is this ugly!
  return ( !isEmpty() && 
           (D_TERM((m_isPeriodic[0] 
                    || (b.smallEnd(0)>=m_domainBox.smallEnd(0) 
                        && b.bigEnd(0)<=m_domainBox.bigEnd(0))),
                   && (m_isPeriodic[1] 
                       || (b.smallEnd(1)>=m_domainBox.smallEnd(1) 
                           && b.bigEnd(1)<=m_domainBox.bigEnd(1))),
                   && (m_isPeriodic[2]
                       || (b.smallEnd(2)>=m_domainBox.smallEnd(2)
                           && b.bigEnd(2)<=m_domainBox.bigEnd(2))))));
  
}

#endif /*WRAPPER*/


#endif /*CH_PROBLEM_DOMAIN_H*/
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//


using std::istream;
using std::ws;
ShiftIterator::ShiftIterator(const bool* a_isPeriodic)
{
  m_index = 0;
  computeShifts(a_isPeriodic);
}

ShiftIterator::~ShiftIterator()
{
  m_shift_vectors.clear();
}

void
ShiftIterator::computeShifts(const bool* a_isPeriodic)
{
  m_shift_vectors.resize(0);

  // handle no directions are periodic
  if (D_TERM(!a_isPeriodic[0], 
             && !a_isPeriodic[1],
             && !a_isPeriodic[2])) 
    {
      // return with no periodic directions defined
    }
  // now handle if all directions are periodic
  else if (D_TERM(a_isPeriodic[0],
                  && a_isPeriodic[1],
                  && a_isPeriodic[2])) 
    {
      // loop over all possible shifts in this case (including corners)
      D_TERM(
             for (int xVect=-1; xVect<2; xVect++),
             for (int yVect=-1; yVect<2; yVect++),
             for (int zVect=-1; zVect<2; zVect++) )
        {
          IntVect shiftVect(D_DECL(xVect,yVect,zVect));
          // no point in adding the zero vector
          if (shiftVect != IntVect::Zero) 
            {
              m_shift_vectors.push_back(shiftVect);
            }
        } // end loop over possible shift vectors
    }
  // case where some but not all directions are periodic
  // this could probably also swallow the completely periodic case
  else
    {
      // i think this should work 
      // this is designed to filter out non-periodic directions
      D_TERM(int xMult = (a_isPeriodic[0]?1:0);,
             int yMult = (a_isPeriodic[1]?1:0);,
             int zMult = (a_isPeriodic[2]?1:0););
      
      // in non-periodic dirctions, shift vector component should
      // always be 0
      D_TERM(
             for (int xVect=-1*xMult; xVect<1+xMult; xVect++), 
             for (int yVect=-1*yMult; yVect<1+yMult; yVect++), 
             for (int zVect=-1*zMult; zVect<1+zMult; zVect++)  )
        {
          IntVect shiftVect(D_DECL(xVect, yVect, zVect));
          // no point in adding the zero vector
          if (shiftVect != IntVect::Zero) 
            {
              m_shift_vectors.push_back(shiftVect);
                   }
        } // end loop over possible shift vectors
      
      
    } // end case where some directions are periodic

}


ProblemDomain::ProblemDomain(const Box& a_domBox)
  : m_domainBox(a_domBox)
{
  // default is a non-periodic domain
  for (int dir=0; dir<SpaceDim; dir++) {
    m_isPeriodic[dir] = false;
  }
  m_shiftIt.computeShifts(m_isPeriodic);
} 


ProblemDomain::ProblemDomain(const Box& a_domBox,
                             const bool* a_isPeriodic)
  : m_domainBox(a_domBox)
{
  for (int dir=0; dir<SpaceDim; dir++) {
    m_isPeriodic[dir] = a_isPeriodic[dir];
  }
  m_shiftIt.computeShifts(m_isPeriodic);
} 

ProblemDomain::ProblemDomain(const IntVect& small, const IntVect& big)
  : m_domainBox(small, big)
{
  // default is a non-periodic domain
  for (int dir=0; dir<SpaceDim; dir++) {
    m_isPeriodic[dir] = false;
  } 
  m_shiftIt.computeShifts(m_isPeriodic);
}

ProblemDomain::ProblemDomain(const IntVect& small, const IntVect& big,
                             const bool* a_isPeriodic)
  : m_domainBox(small, big)
{
  // default is a non-periodic domain
  for (int dir=0; dir<SpaceDim; dir++) {
    m_isPeriodic[dir] = a_isPeriodic[dir];
  } 
  m_shiftIt.computeShifts(m_isPeriodic);
}

ProblemDomain::ProblemDomain(const IntVect& small, const int* vec_len)
  : m_domainBox(small, vec_len)
{
  // default is a non-periodic domain
  for (int dir=0; dir<SpaceDim; dir++) {
    m_isPeriodic[dir] = false;
  } 
  m_shiftIt.computeShifts(m_isPeriodic);
}


ProblemDomain::ProblemDomain(const IntVect& small, const int* vec_len,
                             const bool* a_isPeriodic)
  : m_domainBox(small, vec_len)
{
  // default is a non-periodic domain
  for (int dir=0; dir<SpaceDim; dir++) {
    m_isPeriodic[dir] = a_isPeriodic[dir];
  } 
  m_shiftIt.computeShifts(m_isPeriodic);
}




// returns true if box intersects problem domain
bool
ProblemDomain::intersects(const Box& a_box) const
{
  if (isEmpty() || a_box.isEmpty()) return false;
  // this is kinda ugly, but it should work
  return (D_TERM((m_isPeriodic[0] 
                  || (a_box.smallEnd(0)>m_domainBox.smallEnd(0) 
                      || a_box.bigEnd(0)<m_domainBox.bigEnd(0))),
                 || (m_isPeriodic[1] 
                     || (a_box.smallEnd(1)>m_domainBox.smallEnd(1) 
                         || a_box.bigEnd(1)<m_domainBox.bigEnd(1))),
                 || (m_isPeriodic[2]
                     || (a_box.smallEnd(2)>m_domainBox.smallEnd(2)
                         || a_box.bigEnd(2)<m_domainBox.bigEnd(2)))));
}


bool
ProblemDomain::intersectsNotEmpty(const Box& a_box) const
{
  // this is kinda ugly, but it should work
  return (D_TERM((m_isPeriodic[0] 
                  || (a_box.smallEnd(0)>m_domainBox.smallEnd(0) 
                      || a_box.bigEnd(0)<m_domainBox.bigEnd(0))),
                 || (m_isPeriodic[1] 
                     || (a_box.smallEnd(1)>m_domainBox.smallEnd(1) 
                         || a_box.bigEnd(1)<m_domainBox.bigEnd(1))),
                 || (m_isPeriodic[2]
                     || (a_box.smallEnd(2)>m_domainBox.smallEnd(2)
                         || a_box.bigEnd(2)<m_domainBox.bigEnd(2)))));

}


bool
ProblemDomain::intersects(const Box& box1, const Box& box2) const
{

  bool boxesIntersect = false;
  // first do the obvious and easy check
  boxesIntersect = box1.intersects(box2);
  
  // now do the harder check whether periodic images intersect
  if (!boxesIntersect && isPeriodic())
    {
      if (m_domainBox.contains(box1) && m_domainBox.contains(box2)) 
        { 
          // no need to do any periodic checking
        } 
      else
        {
          // loop over periodic shift directions, shift box1, and 
          // check for intersections
          IntVect shiftMult(m_domainBox.size());
          ShiftIterator shiftIt = m_shiftIt;
          for (shiftIt.begin(); shiftIt.ok(); ++shiftIt) 
            {
              Box shiftBox(box1);
              IntVect shiftVect = shiftMult*shiftIt();
              shiftBox.shift(shiftVect);
              if (shiftBox.intersects(box2)) boxesIntersect = true;
            }
        }
    } // end if we need to do periodic check
  return boxesIntersect;
}
              
          


void
ProblemDomain::setPeriodic(int a_dir, bool a_isPeriodic) 
{
  assert (a_dir < SpaceDim);
  m_isPeriodic[a_dir] = a_isPeriodic;
  m_shiftIt.computeShifts(m_isPeriodic);
}






Box
bdryLo(const ProblemDomain& a_pd, int a_dir, int a_len)
{
  Box bndryBox;
  if (!a_pd.isPeriodic(a_dir))
    {
      bndryBox = bdryLo(a_pd.domainBox(),a_dir,a_len);
    }
  return bndryBox;
}

Box
bdryHi(const ProblemDomain& a_pd, int a_dir, int a_len)
{
  Box bndryBox;
  if (!a_pd.isPeriodic(a_dir))
    {
      bndryBox = bdryHi(a_pd.domainBox(),a_dir,a_len);
    }
  return bndryBox;
}

Box
adjCellLo(const ProblemDomain& a_pd, int a_dir, int a_len)
{
  Box adjBox;
  if (!a_pd.isPeriodic(a_dir)) 
    {
      adjBox = adjCellLo(a_pd.domainBox(),a_dir, a_len);
    }
  return adjBox;

}



Box
adjCellHi(const ProblemDomain& a_pd, int a_dir, int a_len)
{
  Box adjBox;
  if (!a_pd.isPeriodic(a_dir)) 
    {
      adjBox = adjCellHi(a_pd.domainBox(),a_dir, a_len);
    }
  return adjBox;
}



Box
ProblemDomain::operator&(const Box& b) const
{

  Box intersectBox;
  // check for empty -- if either box is empty, return empty intersectBox
  if (!b.isEmpty() && !isEmpty()) 
    {         
      // do this by creating an intersection box which will
      // take the periodic case into account
      intersectBox = m_domainBox;
      if (isPeriodic()) 
        {
          for (int dir=0; dir<SpaceDim; dir++) 
            {
              if (m_isPeriodic[dir])
                {
                  // the idea behind this is that intersect does nothing in
                  // periodic directions
                  intersectBox.setRange(dir, b.smallEnd(dir), b.size(dir));
                }
            }
        }

      intersectBox &= b;            
    } // end if neither box is empty

  
  return intersectBox;
}



ProblemDomain& 
ProblemDomain::refine(int a_refinement_ratio)
{
  m_domainBox.refine(a_refinement_ratio);
  return *this;
}

ProblemDomain
refine(const ProblemDomain& a_probdomain, int a_refinement_ratio)
{
  
  if (a_probdomain.isEmpty()) return a_probdomain;
  Box newdomain = a_probdomain.domainBox();
  newdomain.refine(a_refinement_ratio);
  
  return ProblemDomain(newdomain, a_probdomain.m_isPeriodic);
}

ProblemDomain& 
ProblemDomain::refine(const IntVect& a_refinement_ratio)
{
  m_domainBox.refine(a_refinement_ratio);
  return *this;
}


ProblemDomain
refine(const ProblemDomain& a_probdomain, const IntVect& a_refinement_ratio)
{
  Box newdomain(a_probdomain.domainBox());
  newdomain.refine(a_refinement_ratio);
  
  return ProblemDomain(newdomain, a_probdomain.m_isPeriodic);
}


ProblemDomain& 
ProblemDomain::coarsen(int a_refinement_ratio)
{
  m_domainBox.coarsen(a_refinement_ratio);
  return *this;
}

ProblemDomain
coarsen(const ProblemDomain& a_probdomain, int a_refinement_ratio)
{
  Box newdomain(a_probdomain.domainBox());
  
  newdomain.coarsen(a_refinement_ratio);
  return ProblemDomain(newdomain, a_probdomain.m_isPeriodic);
}


ProblemDomain&
ProblemDomain::coarsen(const IntVect& a_refinement_ratio)
{
  m_domainBox.refine(a_refinement_ratio);
  return *this;
}

ProblemDomain
coarsen(const ProblemDomain& a_probdomain, 
        const IntVect& a_refinement_ratio)
{
  
  Box newdomain(a_probdomain.domainBox());
  newdomain.coarsen(a_refinement_ratio);

  return ProblemDomain(newdomain, a_probdomain.m_isPeriodic);
}


ostream& operator<< (ostream& os, const ProblemDomain& a_probdomain)
{
  os << a_probdomain.domainBox()
     << ", " << D_TERM( a_probdomain.m_isPeriodic[0],
                        << " " << a_probdomain.m_isPeriodic[1],
                        << " " << a_probdomain.m_isPeriodic[2]);

  if (os.fail())
    MayDay::Error("operator<<(ostream&,ProblemDomain&) failed");

  return os;
}

//
// Copied from <Utility.H>
//
#define CH_IGNORE_MAX 100000

istream&
operator>> (istream& is, ProblemDomain& a_probdomain)
{

  is >> a_probdomain.m_domainBox;
  is >> ws;
  char c;
  is >> c;
  is.putback(c);
  if (c == ',')
    {
      is.ignore(CH_IGNORE_MAX, ',');
      for (int dir=0; dir<SpaceDim; dir++) 
        {
          is >> a_probdomain.m_isPeriodic[dir];
        }
    }

  else
    MayDay::Error("operator>>(istream&,ProblemDomain&): expected \',\'");
  
  if (is.fail())
    MayDay::Error("operator>>(istream&,ProblemDomain&) failed");
  
    return is;
}

  
void 
ProblemDomain::dumpOn(ostream& strm) const
{

  strm << "domainBox: ";
  m_domainBox.dumpOn(strm);
  strm << " isPeriodic = " 
       << D_TERM(m_isPeriodic[0], 
                 << " " << m_isPeriodic[1], 
                 << " " << m_isPeriodic[2])
       << '\n';
  
  if (strm.fail())
    MayDay::Error("ProblemDomain::dumpOn(ostream&) failed");

}

void operator &= (Box& a_box, const ProblemDomain& a_probdomain)
{
  a_box = a_probdomain & a_box;
}

Box operator & (const Box& a_box, const ProblemDomain& a_probdomain)
{
  Box returnBox = a_probdomain & a_box;
  return returnBox;
}
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// DTGraves

#ifndef _BOX_ITERATOR_H_
#define _BOX_ITERATOR_H_



///iterates through the IntVects of a Box
/**
   BoxIterator iterates through the IntVects of a box.  The actual
   sqeuence of IntVects is implementation-specific.  
   Typical usage:\\

\begin{verbatim}
   Box b; 
   ...
   BoxIterator bit (b); 
   for (bit.begin(); bit.ok(); ++bit)  
   {  
     IntVect iv = bit(); 
     (do operations involving iv) 
   }
\end{verbatim}
*/
class BoxIterator
{
public:

  ///{\bf Constructors, destructors, defines}

  ///
  /**
     Default constructor.  This constructs an invalid iterator.
     The user must call define before using.
  */    
  BoxIterator ();

  ///
  /**
     Constructs a BoxIterator and associates it with a Box.\\
     {\bf Arguments:}\\
     bx (not modified) the Box to iterate over.\\
     {\bf Returns:} none.\\
     {\bf This:}\\
     This object is modified.
  */
  BoxIterator(const Box& bx);

  void setBox(const Box& bx);

  ///
  /**
     Associates a Box with this BoxIterator.\\
     {\bf Arguments:}\\
     bx (not modified) the Box to iterate over. \\
     {\bf Returns:} none.\\
     {\bf This:}\\
     This object is modified.
  */
  void define (const Box& bx);

  ///
  /**
     Copy constructor.\\
     {\bf Arguments:}\\
     tx (not modified) the BoxIterator to copy.\\
     {\bf Returns:} none.\\
     {\bf This:}\\
     This object is modified.
  */
  BoxIterator (const BoxIterator& tx);
    
  ///
  /**
     Destructor.
  */
  ~BoxIterator ();

  ///{\bf Modification functions}

  ///
  /**
     Sets this BoxIterator to the first IntVect in its Box.  The
     definition of the "first" IntVect is
     implementation-dependent.\\
     {\bf Arguments:} none.\\
     {\bf Returns:} none.\\
     {\bf This:}\\
     This object is modified.
  */
  void begin();

  ///
  /**
     Sets this BoxIterator to the first IntVect in its Box.  The
     definition of the "first" IntVect is
     implementation-dependent.\\
     {\bf Arguments:} none.\\
     {\bf Returns:} none.\\
     {\bf This:}\\
     This object is modified.
  */
  void reset();

  ///
  /**
     Modifies this BoxIterator to set it to the next location in its
     Box.  The definition of the "next location" of a Box is
     implementation-dependent.\\ 
     {\bf Arguments:} none.  \\
     {\bf Returns:} none. \\
     {\bf This:} \\This object is modified.  

  */
  void operator++();

  void next();



  ///{\bf Access functions}
   
  ///
  /**
     Returns the value of the InVect for the current location of this BoxIterator.\\
     {\bf Arguments:} none.\\
     {\bf Returns:} \\
     the value of the intvect for the current location.\\
     {\bf This:}\\
     This object is not modified.
  */
  const IntVect& operator() () const;

  ///
  /**
     Returns true if this BoxIterator's location is within its Box.\\
     {\bf Arguments:} none.\\
     {\bf Returns:} \\
     true if this BoxIterator's current location is in its Box.\\
     {\bf This:}\\
     This object is not modified.
  */
  bool ok();

protected:

  IntVect m_current;
  IntVect m_boxLo;
  IntVect m_boxHi;
};

#endif
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// DTGraves


/*******/
/*******/
BoxIterator::
BoxIterator ()
{
  m_current = IntVect::Unit;
  m_boxLo   = IntVect::Unit;
  m_boxHi   = IntVect::Zero;
}

/*******/
/*******/
BoxIterator::
BoxIterator(const Box& bx)
{
  define(bx);
}

/*******/
/*******/
void 
BoxIterator::
setBox(const Box& bx)
{
  define(bx);
}


/*******/
/*******/
void 
BoxIterator::
define (const Box& a_bx)
{
  //this algorithm breaks for 
  //screwy empty boxes
  if(a_bx.smallEnd() <= a_bx.bigEnd())
    {
      m_current = a_bx.smallEnd();
      m_boxLo   = a_bx.smallEnd();
      m_boxHi   = a_bx.bigEnd();
    }
  else
    {
      m_current = IntVect::Unit;
      m_boxLo   = IntVect::Unit;
      m_boxHi   = IntVect::Zero;
    }
}

/*******/
/*******/
BoxIterator::
BoxIterator (const BoxIterator& a_iterIn)
{
  m_current = a_iterIn.m_current;
  m_boxLo =   a_iterIn.m_boxLo;
  m_boxHi =   a_iterIn.m_boxHi;
}

/*******/
/*******/
BoxIterator::
~BoxIterator ()
{
}

/*******/
/*******/
void 
BoxIterator::
begin()
{
  if(m_boxLo <= m_boxHi)
    {
      m_current = m_boxLo;
    }
}
/*******/
/*******/

void 
BoxIterator::
reset()
{
  begin();
}
/*******/
/*******/

void 
BoxIterator::
operator++()
{
  next();
}
/*******/
/*******/
void 
BoxIterator::
next()
{
  m_current[0]++;
#if CH_SPACEDIM >= 2
  if(m_current[0] > m_boxHi[0])
    {
      m_current[0] = m_boxLo[0];
      m_current[1]++;
#if CH_SPACEDIM == 3 
      if(m_current[1] > m_boxHi[1])
        {
          m_current[1] = m_boxLo[1];
          m_current[2]++;
        }
#endif
    }
#endif
}
/*******/
/*******/
const IntVect& 
BoxIterator::
operator() () const
{
  assert(m_current <= m_boxHi);
  assert(m_current >= m_boxLo);
  return m_current;
}
/*******/
/*******/
bool 
BoxIterator::
ok()
{
  return (m_current  <= m_boxHi);
}
/*******/
/*******/
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

//CH_COPYRIGHT_NOTICE

#ifndef CH_BASEFAB_H
#define CH_BASEFAB_H

//
// $Id: CFile.cpp,v 1.6 2003/04/21 22:08:00 bvs Exp $
//



//
//@Man:
//@Memo: A Fortran Array-like Container
/*@Doc:

  BaseFab emulates the Fortran array concept.  
  Useful operations can be performed upon
  BaseFab's in C++, and they provide a convenient interface to
  Fortran when it is necessary to retreat into that language.
        
  `BaseFab' is a template class.  Through use of the
  template, a `BaseFab' may be based upon any class.  So far at least,
  most applications have been based upon simple types like `integer's,
  `real's, or `doubleprecision's.  Most applications do not use BaseFab's 
  directly, but utilize specialized classes derived from BaseFab.
        
  It will be easier to use a `BaseFab' if you understand the following
  concepts.  `BaseFab' objects depend on the dimensionality of space
  (indirectly through the DOMAIN `Box' member).  It is
  typical to define the macro `CH_SPACEDIM' to be 1, 2, or 3 to indicate
  the dimension of space.  See the discussion of class `Box' for more
  information.  A `BaseFab' contains a `Box' DOMAIN, which indicates the
  integer indexing space over which the array is defined.  A `BaseFab'
  also has NVAR components.  By components, we mean that for each
  point in the rectangular indexing space, there are NVAR values
  associated with that point.  A Fortran array corresponding to a
  `BaseFab' would have (CH_SPACEDIM+1) dimensions.
        
  By design, the array layout in a `BaseFab' mirrors that of a
  Fortran array.  The first index (x direction for example) varies
  most rapidly, the next index (y direction), if any, varies next
  fastest. The component index varies last, after all the spatial
  indices.
        
  It is sometimes convenient to be able to treat a sub-array within an
  existing `BaseFab' as a `BaseFab' in its own right.  This is often
  referred to as 'aliasing' the `BaseFab'.  Note that when aliasing is 
  used, the BaseFab's domain will not, in general, be the same as the 
  parent BaseFab's domain, nor will the number of components.
  BaseFab is a dimension dependent class, so CH_SPACEDIM must be 
  defined as either 1, 2, or 3 when compiling.

  This is NOT a polymorphic class.

  It does NOT provide a copy constructor or assignment operator.

  T MUST have a default constructor and an assignment operator.
*/

template <class T>
class BaseFab
{
public:

  /// {\bf constructors, destructor and defines}

  /*@ManDoc: Constructs an invalid `BaseFab'.  The domain is invalid, the
    number of components is zero, and no actual array memory is
    allocated.  An invalid `BaseFab' must be resize()d
    (see `BaseFab::resize') before use.
  */
  BaseFab ();

  ///
  /**
     Constructs a BaseFab with desired domain and number of components.
*/
  BaseFab (const Box& bx,
           int        n,
		   T*         alias=NULL);

  ///
  /**
     The destructor deletes the array memory.
*/
  virtual ~BaseFab ();

  /*@ManDoc: This function resizes a `BaseFab' so it covers the `Box' B
               with N components.  The default action
               is that under resize()ing, the memory allocated for the
               `BaseFab' only grows and never shrinks.  This function is
               particularly useful when a `BaseFab' is used as a temporary
               space which must be a different size whenever it is used.
               Resize()ing a temp will often be faster than re-allocating a
               `BaseFab' because memory allocation can often be avoided.
    */
  void resize (const Box& b,
               int        N = 1,
			   T*         alias=NULL);

  /*@ManDoc: Make BaseFab with desired domain and number of components.
               existing data is lost.  Data is in uninialized state.
    */
  virtual void define(const Box& box, int comps, 
					  T* alias = NULL) { resize(box, comps, alias);}

  /*@ManDoc: The function returns the `BaseFab' to the invalid state.  (See
               comments for constructors above.)  The memory is freed.
    */
  void clear ();


  /// {\bf accessors}

///
  /**
  Returns the number of components.
  */
  int nComp () const;

  ///
  /**
     Returns the domain (box) where the array is defined.
  */
  const Box& box () const;

  /*@ManDoc: Returns a pointer to an array of SpaceDim integers
               giving the length of the domain in each direction.
    */
  const int* size () const;

  /*@ManDoc: Returns the lower corner of the domain.  See
               class `Box' for analogue.
    */
  const IntVect& smallEnd () const;
 
  /*@ManDoc:  Returns the upper corner of the domain.  See
                class `Box' for analogue.
    */
  const IntVect& bigEnd () const;

  ///
  /**
     Returns an Interval for the entire range on components.
*/
  Interval interval() const
  { return Interval(0, nvar-1); }

  /*@ManDoc: Returns a modifiable lvalue reference to the Nth component value
               defined at position p in the domain.  This operator may be
               inefficient if the C++ compiler is unable to optimize the
               C++ code.
    */
  T& operator() (const IntVect& p,
                 int            N);

  T& operator() (const IntVect& p);

  /*@ManDoc: Returns a conatant reference to the Nth component value
    defined at position p in the domain.  This operator may be
    inefficient if the C++ compiler is unable to optimize the
    C++ code.
    */
  const T& operator() (const IntVect& p,
                       int            N) const;

  const T& operator() (const IntVect& p) const;

  /*@ManDoc: This function puts numcomp component values, starting at
               component N, from position pos in the domain into array data,
               that must be allocated by the user.
    */
  void getVal (T*             data,
               const IntVect& pos,
               int            N,
               int            numcomp) const;

  /*@ManDoc: This function puts all component values, starting at
               component 0, from position pos in the domain into array data,
               that must be allocated by the user.
    */
  void getVal (T*             data,
               const IntVect& pos) const;

  ///{\bf Fortran interface functions}

  /*@ManDoc: Returns the lower corner of the domain.  Instead of
               returning them in the form of IntVects, as in smallEnd
               and bigEnd, it returns the values as a pointer to an
               array of constant integers.  This is useful when
               interfacing to Fortran subroutines.  It should not be
               used in any other context!!!

    */
  const int* loVect () const;
 
  /*@ManDoc: Returns the upper corner of the domain.  Instead of
               returning them in the form of IntVects, as in smallEnd
               and bigEnd, it returns the values as a pointer to an
               array of constant integers.  This is useful when
               interfacing to Fortran subroutines.  It should not be
               used in any other context!!!

    */
  const int* hiVect () const;
 
  /*@ManDoc: Returns a pointer to an integer that contains the 
               number of components in the BaseFab. This is useful when
               interfacing to Fortran subroutines.  It should not be
               used in any other context!!!

    */
  const int* nCompPtr () const;
 
  /*@ManDoc: Returns a pointer to an object of type T that is the
               value of the Nth component associated with the cell at
               the low end of the domain.  This is commonly used to
               get a pointer to data in the array which is then handed
               off to a Fortran subroutine. It should not be used in
               any other context!!!  Remember that data is stored in
               Fortran array order, with the component index coming
               last.  In other words, `dataPtr' returns a pointer to
               all the Nth components.  

*/
  T* dataPtr (int N = 0);

  /*@ManDoc: Returns a constant pointer to an object of type T that is
               the value of the Nth component associated with the cell
               at the low end of the domain.  This is commonly used to
               get a pointer to data in the array which is then handed
               off to a Fortran subroutine.  It should not be used in
               any other context!!! Remember that data is stored in
               Fortran array order, with the component index coming
               last.  In other words, `dataPtr' returns a pointer to
               all the Nth components.  

*/
  const T* dataPtr (int N = 0) const;

  ///{\bf comparison functions}

  /*@ManDoc: Returns true if the domain of fab is totally contained within
               the domain of this `BaseFab'.
    */
  bool contains (const BaseFab<T>& fab) const;
 
  /*@ManDoc: Returns true if bx is totally contained
               within the domain of this `BaseFab'.
    */
  bool contains (const Box& bx) const;

  ///{\bf data modification functions}

  /*@ManDoc: The setVal functions set subregions in the `BaseFab' to a
               constant value.  This most general form specifies the sub-box,
               the starting component number, and the number of components
               to be set.
    */
  void setVal (T          x,
               const Box& bx,
               int        nstart,
               int        ncomp);

  ///
  /**
     Modifies this BaseFab so that all values of component N in the
     specified Box bx are set to the given value x.  

*/
  void setVal (T          x,
               const Box& bx,
               int        N);

  ///
  /**
     Modifies this BaseFab so that all values of component N are set to
     the given value x.

*/
  void setVal (T   x,
               int N);

  ///
  /**
     Modifies this BaseFab so that all values of all components are set to
     the given value x.

*/
  void setVal (T x);


  /*@ManDoc: Modifies this BaseFab by copying the contents of the
               argument BaseFab into it.  This, the most general form
               of copy, specifies the contents of any sub-box srcbox
               in `BaseFab' src may be copied into a (possibly
               different) destbox in the destination `BaseFab'.  Note
               that although the srcbox and the destbox may be
               disjoint, they must be the same size and shape.  If the
               sizes differ, the copy is undefined and a runtime error
               results.  This copy function is the only one of the
               copy functions to allow a copy between differing
               boxes. The user also specifies how many components are
               copied, starting at component srccomp in src and stored
               starting at component destcomp.  The results are
               UNDEFINED if the src and dest BaseFabs are the same and
               the srcbox and destbox overlap.

*/
  BaseFab<T>& copy (const BaseFab<T>& src,
                    const Box&        srcbox,
                    int               srccomp,
                    const Box&        destbox,
                    int               destcomp,
                    int               numcomp);

  /*@ManDoc: Modifies this BaseFab by coping the contents of the
             argument BaseFab into it.  A copy within the intersecting
             region of the domains of the two BaseFabs is performed.
             The user specifies how many components are copied,
             starting at component srccomp in src and stored starting
             at component destcomp.

*/
  BaseFab<T>& copy (const BaseFab<T>& src,
                    int               srccomp,
                    int               destcomp,
                    int               numcomp = 1);

  /*@ManDoc: Modifies this BaseFab by coping the contents of the
             argument BaseFab into it.  A copy within the intersecting
             region of the domains of the two BaseFabs and the
             specified Box destbox is performed.  All components are
             copied.

*/
  BaseFab<T>& copy (const BaseFab<T>& src,
                    const Box&        destbox);

  /*@ManDoc: Modifies this BaseFab by coping the contents of the
             argument BaseFab into it.  A copy within the intersecting
             region of the domains of the two BaseFabs is performed.
             All components are copied.

*/
  BaseFab<T>& copy (const BaseFab<T>& src);

  ///
  /**
     Copy from a subsection of one box into another.  Assumes
     the boxes are both in the same index space, and that box R is completely
     contained in both the source and destination boxes.
    */
  void copy(const Box& RegionFrom, 
            const Interval& Cdest, 
            const Box& RegionTo,
            const BaseFab<T>& src, 
            const Interval& Csrc);


  ///    {\bf domain modification functions}

///
  /**
     Modifies the domain of this BaseFab by shifting.  Equivalent to
     fab.shift(0,iv[0]).shift(1,iv[1]) ....  There is no effect upon
     the array memory.

*/
  BaseFab<T>& shift (const IntVect& v);

  ///
  /**
     Modifies the domain of this BaseFab by shifting it n_cells
     indexing positions in coordinate direction idir.  Directions are
     zero-based.  It is an error if not 0 <= idir < SpaceDim.There is
     no effect upon the array memory.

*/
  BaseFab<T>& shift (int idir,
                     int n_cell);

  ///
  /**
     Modifies the domain of this BaseFab by shifting by "half" indices,
     thereby converting the Box from type CELL to NODE or vice-versa.
     fab.shiftHalf(0,1) shifts the domain to the right by 1/2 cells.
     fab.shiftHalf(1,-3) shifts the domain in the -j direction by 3/2 cells.
     NOTE: If num\_halfs is EVEN the shift is num\_halfs/2 full zones
     and hence will not change the type.  This is: fab.shifthalf(4) ==
     fab.shift(2).  Directions are zero-based.  It is an error if not 0
     <= dir < SpaceDim.   There is no
     effect upon the array memory. 

*/
  BaseFab<T>& shiftHalf (int dir,
                         int num_halfs);

  ///
  /**
     Modifies the domain of this BaseFab by shifting by half indices.
     Equivalent to fab.shiftHalf(0,iv[0]).shiftHalf(1,iv[1]) ...
     There is no effect upon the array memory.

  */
  BaseFab<T>& shiftHalf (const IntVect& num_halfs);

  ///{\bf linearization functions}

  ///
  /**
     Returns the size, in number of bytes, of a flat linear
     representation of the data in this object in the area defined
     by the input Box R and the component Interval comps.  The size
     */
  virtual int size(const Box& b, const Interval& comps) const;

  ///
  /**
     Write a linear representation of the internal data.  Assumes
     that sufficient memory for the buffer has already been allocated by the
     caller.
    */
  virtual void linearOut(void* buf, const Box& R, const Interval& comps) const;

  ///
  virtual void linearIn(void* buf, const Box& R, const Interval& comps);

  ///
  static int preAllocatable(){ return 0;} // static preAllocatable

protected:
  //
  // Allocates memory for the `BaseFab<T>'.
  //
  void define ();
  //
  // Deallocates memory for the `BaseFab<T>'.
  //
  void undefine ();
  //
  // template class static data member.  not sure if this will
  // work with all compilers.  It has been in the draft standard
  // for some time
  //
  static Arena* s_Arena;
  //
  //
  static std::string name();
  //
  // The function called by BaseFab copy operations.
  //
  virtual void performCopy (const BaseFab<T>& src,
                            const Box&        srcbox,
                            int               srccomp,
                            const Box&        destbox,
                            int               destcomp,
                            int               numcomp);

  //
  // This function is called by the `BaseFab' setVal operations.
  //
  void performSetVal (T          x,
                      const Box& bx,
                      int        nstart,
                      int        numcomp);


private:
  //
  // These functions are made private to prevent use of the default 
  // functions provided by the C++ compiler.
  //

  BaseFab<T>& operator= (const BaseFab<T>&);
  BaseFab (const BaseFab<T>&);

protected:

  Box  domain;   // My index space.
  int  nvar;     // Number components.
  long numpts;   // Cached number of points in FAB.
  long truesize; // nvar*numpts that was allocated on heap (only if alias==false).
  T*   dptr;     // The data pointer.
  bool aliased;  // The BaseFab is not allocated memory, but is an alias. bvs
};


#endif /*CH_BASEFAB_H*/

/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _CH_LAME_MACROS_H
#define _CH_LAME_MACROS_H

#if (CH_SPACEDIM == 1)

/*@ManDoc:
  The macro ForAllThisPencil(T,b,ns,nc) is intended to facilitate efficient
  looping over the contents of BaseFabs and objects derived from BaseFab.
  Special attention has been paid to make it work efficiently on vector
  supercomputers.

  This macro acts upon the BaseFab *this.  Instead of iterating over the
  entire Box b, it iterates over components starting at component ns and
  ending at component ns+nc-1, in all directions except the first coordinate
  direction.  The user must iterate over the first coordinate direction within
  the ForAllThisPencil loop.  The macro creates two internal reference
  variables; thisR that references the first element in the pencil, and
  thisLen that gives the length of the pencil.  The first argument of the
  macro is a type: the type contained in the BaseFab that is being iterated
  over.

  We can rewrite the code illustrated in `ForAllThisBNN' in this form:

    template <class T>
    void BaseFab<T>::performSetVal(const T val, const Box bx, int ns, int nc)
    {
       assert(domain.contains(bx));
       ForAllThisPencil(T,bx,ns,nc)
       {
          T* dog = \&thisR;
          for (int i = 0; i < thisLen; i++)
             dog[i] = val;
       } EndForPencil
    }

  Looping macro mnemonics:

    This stands for the current object
    C for a const
    X stands for a BaseFab
    B for a Box
    N for an int
*/
#define ForAllThisPencil(T,b,ns,nc)                                     \
{                                                                       \
    assert(contains(b));                                                \
    assert((ns) >= 0 && (ns) + (nc) <= nComp());                        \
    const int *_th_plo = loVect();                                      \
    const int *_th_plen = size();                                       \
    const int *_b_lo = (b).loVect();                                    \
    const int *_b_len = (b).size();                                     \
    T* _th_p = dptr;                                                    \
    for(int _n = (ns); _n < (ns)+(nc); ++_n) {                          \
        const int nR = _n;                                              \
        T *_th_pp = _th_p                                               \
            + ((_b_lo[0] - _th_plo[0])                                  \
               + _n * _th_plen[0]);                                     \
        T &thisR = * _th_pp;                                            \
        const int thisLen = _b_len[0];                                  \

/*@ManDoc:
 Same as ForAllThisPencil, except that it operates on the
 the argument BaseFab passed in. added by bvs,  05/27/99
*/
#define ForAllXBPencil(T,x,b,ns,nc)                                     \
{                                                                       \
    assert(x.contains(b));                                              \
    assert((ns) >= 0 && (ns) + (nc) <= x.nComp());                      \
    const int *_th_plo = x.loVect();                                    \
    const int *_th_plen = x.size();                                     \
    const int *_b_lo = (b).loVect();                                    \
    const int *_b_len = (b).size();                                     \
    T* _th_p = x.dataPtr();                                             \
    for(int nR = (ns); nR < (ns)+(nc); ++nR) {                          \
        T * xR = _th_p                                                  \
            + ((_b_lo[0] - _th_plo[0])                                  \
               + nR * _th_plen[0]);                                     \
    const int thisLen = _b_len[0];  

/*@ManDoc:
  The macro ForAllThisCPencil(T,b,ns,nc) is intended to facilitate efficient
  looping over the contents of BaseFabs and objects derived from BaseFab.
  Special attention has been paid to make it work efficiently on vector
  supercomputers.

  This is the constant version of ForAllThisPencil; i.e. it works when the
  underlying BaseFab is constant.

  Looping macro mnemonics:

    This stands for the current object
    C for a const
    X stands for a BaseFab
    B for a Box
    N for an int
*/
#define ForAllThisCPencil(T,b,ns,nc)                                    \
{                                                                       \
    assert(contains(b));                                                \
    assert((ns) >= 0 && (ns) + (nc) <= nComp());                        \
    const int *_th_plo = loVect();                                      \
    const int *_th_plen = size();                                       \
    const int *_b_lo = (b).loVect();                                    \
    const int *_b_len = (b).size();                                     \
    const T* _th_p = dptr;                                              \
    for(int _n = (ns); _n < (ns)+(nc); ++_n) {                          \
        int nR = _n; nR += 0;                                           \
        const T *_th_pp = _th_p                                         \
            + ((_b_lo[0] - _th_plo[0])                                  \
               + _n * _th_plen[0]);                                     \
        const T &thisR = * _th_pp;                                      \
        const int thisLen = _b_len[0];

/*@ManDoc:
  The macro ForAllXBNN(T,x,b,ns,nc) is intended to facilitate efficient
  looping over the contents of BaseFabs and objects derived from BaseFab.
  Special attention has been paid to make it work efficiently on vector
  supercomputers.

  This macro acts upon the BaseFab x where the loop runs over the points in
  the Box b and over components starting at ns and ending at ns+nc-1.  The
  first argument of the macro is a type: the type contained in the BaseFab
  that is being iterated over. The reference variable is xR, where x
  is literally replaced by the macros second argument.  Thus an expression
  ForAllXBNN(int,dog,...) would have a reference variable dogR of type int.

  Looping macro mnemonics:

    This stands for the current object
    C for a const
    X stands for a BaseFab
    B for a Box
    N for an int
*/
#define ForAllXBNN(T,x,b,ns,nc)                                         \
{                                                                       \
    assert(x.contains(b));                                              \
    assert((ns) >= 0 && (ns) + (nc) <= (x).nComp());                    \
    const int *_x_plo = (x).loVect();                                   \
    const int *_x_plen = (x).size();                                    \
    const int *_b_lo = (b).loVect();                                    \
    const int *_b_len = (b).size();                                     \
    T* _x_p = (x) .dataPtr();                                           \
    for(int _n = (ns); _n < (ns)+(nc); ++_n) {                          \
        const int nR = _n;                                              \
        T *_x_pp = _x_p                                                 \
            + ((_b_lo[0] - _x_plo[0])                                   \
               + _n * _x_plen[0]);                                      \
        for(int _i = 0; _i < _b_len[0]; ++_i, ++_x_pp) {                \
            const int iR = _i + _b_lo[0];                               \
            T &x##R = * _x_pp;

/*@ManDoc:
  The macro ForAllXCBNN(T,x,b,ns,nc) is intended to facilitate efficient
  looping over the contents of BaseFabs and objects derived from BaseFab.
  Special attention has been paid to make it work efficiently on vector
  supercomputers.

  This is the constant version of ForAllXBNN; i.e. it works when the
  underlying BaseFab is constant.

  Looping macro mnemonics:

    This stands for the current object
    C for a const
    X stands for a BaseFab
    B for a Box
    N for an int
*/
#define ForAllXCBNN(T,x,b,ns,nc)                                        \
{                                                                       \
    assert(x.contains(b));                                              \
    assert((ns) >= 0 && (ns) + (nc) <= (x).nComp());                    \
    const int *_x_plo = (x).loVect();                                   \
    const int *_x_plen = (x).size();                                    \
    const int *_b_lo = (b).loVect();                                    \
    const int *_b_len = (b).size().getVect();                           \
    const T* _x_p = (x).dataPtr();                                      \
    for(int _n = (ns); _n < (ns)+(nc); ++_n) {                          \
        const int nR = _n;                                              \
        const T *_x_pp = _x_p                                           \
            + ((_b_lo[0] - _x_plo[0])                                   \
               + _n * _x_plen[0]);                                      \
        for(int _i = 0; _i < _b_len[0]; ++_i) {                         \
            const int iR = _i + _b_lo[0];                               \
            const T & x##R = _x_pp[_i];

/*@ManDoc:
  The ForAllThisBNN(T,b,ns,nc) macro is intended to facilitate efficient
  looping over the contents of BaseFabs and objects derived from BaseFab.
  Special attention has been paid to make it work efficiently on vector
  supercomputers.

  This macro performs the loop over the current object (*this) where the loop
  runs over the points in the Box b and over components starting at ns and
  ending at ns+nc-1.  The first argument of the macro is a type: the type
  contained in the BaseFab that is being iterated over.  The reference
  variable is thisR.

  For example:

    template<class T>
    void
    BaseFab<T>::performSetVal (const T val, const Box bx, int ns, int num)
    {
      assert(domain.contains(bx));
      ForAllThisBNN(T,bx,ns,num)
      {
        thisR = val;
      } EndFor
    }

  Looping macro mnemonics:

    This stands for the current object
    C for a const
    X stands for a BaseFab
    B for a Box
    N for an int  
*/
#define ForAllThisBNN(T,b,ns,nc)                                        \
{                                                                       \
    assert(contains(b));                                                \
    assert((ns) >= 0 && (ns) + (nc) <= nComp());                        \
    const int *_th_plo = loVect();                                      \
    const int *_th_plen = size();                                       \
    const int *_b_lo = (b).loVect();                                    \
    const int *_b_len = (b).size().getVect();                           \
    T* _th_p = dptr;                                                    \
    for(int _n = (ns); _n < (ns)+(nc); ++_n) {                          \
        int nR = _n; nR += 0;                                           \
        T *_th_pp = _th_p                                               \
            + ((_b_lo[0] - _th_plo[0])                                  \
               + _n * _th_plen[0]);                                     \
        for(int _i = 0; _i < _b_len[0]; ++_i, ++_th_pp) {               \
            int iR = _i + _b_lo[0]; iR += 0;                            \
            T &thisR = * _th_pp;

/*@ManDoc:
  The macro ForAllThisCBNN(T,b,ns,nc) is intended to facilitate efficient
  looping over the contents of BaseFabs and objects derived from BaseFab.
  Special attention has been paid to make it work efficiently on vector
  supercomputers.

  This is the constant version of ForAllThisBNN; i.e. it works when the
  underlying BaseFab is constant.

  Looping macro mnemonics:

    This stands for the current object
    C for a const
    X stands for a BaseFab
    B for a Box
    N for an int
*/
#define ForAllThisCBNN(T,b,ns,nc)                                       \
{                                                                       \
    assert(contains(b));                                                \
    assert((ns) >= 0 && (ns) + (nc) <= nComp());                        \
    const int *_th_plo = loVect();                                      \
    const int *_th_plen = size();                                       \
    const int *_b_lo = (b).loVect();                                    \
    const int *_b_len = (b).size().getVect();                           \
    const T* _th_p = dptr;                                              \
    for(int _n = (ns); _n < (ns)+(nc); ++_n) {                          \
        const int nR = _n;                                              \
        const T *_th_pp = _th_p                                         \
            + ((_b_lo[0] - _th_plo[0])                                  \
               + _n * _th_plen[0]);                                     \
        for(int _i = 0; _i < _b_len[0]; ++_i) {                         \
            const int iR = _i + _b_lo[0];                               \
            const T &thisR = _th_pp[_i];

/*@ManDoc:
  The macro ForAllThisBNNXC(T,b,ns,nc,x,nss) is intended to facilitate
  efficient looping over the contents of BaseFabs and objects derived from
  BaseFab.  Special attention has been paid to make it work efficiently on
  vector supercomputers.

  This macro acts upon the BaseFab *this and in addition is able to utiliize
  values in the const BaseFab x.  The loop runs over the points in the Box b
  and over components starting at ns and ending at ns+nc-1.  The reference
  variables are thisR and xR, respectively.  As usual the x in xR is replaced
  by the macro's fifth argument.  The sixth argument nss is the number of the
  argument in x that corresponds to the ns argument in *this.

  Looping macro mnemonics:

    This stands for the current object
    C for a const
    X stands for a BaseFab
    B for a Box
    N for an int
*/
#define ForAllThisBNNXC(T,b,ns,nc,x,nss)                                \
{                                                                       \
    assert((ns) >= 0 && (ns) + (nc) <= nComp());                        \
    assert((nss) >= 0 && (nss) + (nc) <= (x).nComp());                  \
    Box _subbox_((x).box());                                            \
    _subbox_ &= box();                                                  \
    _subbox_ &= b;                                                      \
    if(!_subbox_.isEmpty()) {                                           \
        const int *_th_plo = loVect();                                  \
        const int *_th_plen = size();                                   \
        const int *_x_plo = (x).loVect();                               \
        const int *_x_plen = (x).size();                                \
        const int *_subbox_lo = _subbox_.loVect();                      \
        const int *_subbox_len = _subbox_.size().getVect();             \
        T* _th_p = dataPtr(ns);                                         \
        const T* _x_p  = (x).dataPtr(nss);                              \
        for(int _n = 0; _n < (nc); ++_n) {                              \
            int nR = _n; nR += 0;                                       \
            T *_th_pp = _th_p                                           \
                + ((_subbox_lo[0] - _th_plo[0])                         \
                   + _n * _th_plen[0]);                                 \
            const T *_x_pp = _x_p                                       \
                + ((_subbox_lo[0] - _x_plo[0])                          \
                   + _n * _x_plen[0]);                                  \
            for(int _i = 0; _i < _subbox_len[0]; ++_i, ++_th_pp) {      \
                int iR = _i + _subbox_lo[0]; iR += 0;                   \
                T &thisR = * _th_pp; const T & x##R = _x_pp[_i];

#define ForAllThisCBNNXC(T,b,ns,nc,x,nss)                                \
{                                                                       \
    assert((ns) >= 0 && (ns) + (nc) <= nComp());                        \
    assert((nss) >= 0 && (nss) + (nc) <= (x).nComp());                  \
    Box _subbox_((x).box());                                            \
    _subbox_ &= box();                                                  \
    _subbox_ &= b;                                                      \
    if(!_subbox_.isEmpty()) {                                           \
        const int *_th_plo = loVect();                                  \
        const int *_th_plen = size();                                   \
        const int *_x_plo = (x).loVect();                               \
        const int *_x_plen = (x).size();                                \
        const int *_subbox_lo = _subbox_.loVect();                      \
        const int *_subbox_len = _subbox_.size().getVect();             \
        const T* _th_p = dataPtr(ns);                                         \
        const T* _x_p  = (x).dataPtr(nss);                              \
        for(int _n = 0; _n < (nc); ++_n) {                              \
            int nR = _n; nR += 0;                                       \
            const T *_th_pp = _th_p                                           \
                + ((_subbox_lo[0] - _th_plo[0])                         \
                   + _n * _th_plen[0]);                                 \
            const T *_x_pp = _x_p                                       \
                + ((_subbox_lo[0] - _x_plo[0])                          \
                   + _n * _x_plen[0]);                                  \
            for(int _i = 0; _i < _subbox_len[0]; ++_i, ++_th_pp) {      \
                int iR = _i + _subbox_lo[0]; iR += 0;                   \
                const T &thisR = * _th_pp; const T & x##R = _x_pp[_i];
/*@ManDoc:
  The macro ForAllThisBNNXCBN(T,b,ns,nc,x,bx,nss) is intended to facilitate
  efficient looping over the contents of BaseFabs and objects derived from
  BaseFab.  Special attention has been paid to make it work efficiently on
  vector supercomputers.

  This macro acts upon the BaseFab *this and in addition is able to utiliize
  values in the const BaseFab x.  The loop runs over the points in the
  Box b with components starting at ns and ending at ns+nc-1.  The reference
  variables are thisR and xR, respectively.  As usual the x in xR is replaced
  by the macro's fifth argument.  The sixth argument nss is the number of the
  argument in x that corresponds to the ns argument in *this.  Box bx must
  be the same size as this->box() intersected with b.

  Looping macro mnemonics:

    This stands for the current object
    C for a const
    X stands for a BaseFab
    B for a Box
    N for an int
*/
#define ForAllThisBNNXCBN(T,b,ns,nc,x,bx,nss)                           \
{                                                                       \
    assert((ns) >= 0 && (ns) + (nc) <= nComp());                        \
    assert((nss) >= 0 && (nss) + (nc) <= (x).nComp());                  \
    assert(bx.sameSize((b)));                                             \
    if(!((b).isEmpty())) {                                                  \
        const int *_th_plo = loVect();                                  \
        const int *_th_plen = size();                                   \
        const int *_x_plo = (x).loVect();                               \
        const int *_x_plen = (x).size();                                \
        const int *_subbox_lo = (b).loVect();                             \
        const int *_subbox_len = (b).size().getVect();                    \
        const int *_bx_lo = (bx).loVect();                              \
        T* _th_p = dataPtr(ns);                                         \
        const T* _x_p  = (x).dataPtr(nss);                              \
        for(int _n = 0; _n < (nc); ++_n) {                              \
            int nR = _n + ns; nR += 0;                                  \
            int n##x##R = _n + nss; n##x##R += 0;                       \
            T *_th_pp = _th_p                                           \
                + ((_subbox_lo[0] - _th_plo[0])                         \
                   + _n * _th_plen[0]);                                 \
            const T *_x_pp = _x_p                                       \
                + ((_bx_lo[0] - _x_plo[0])                              \
                   + _n * _x_plen[0]);                                  \
            for(int _i = 0; _i < _subbox_len[0]; ++_i, ++_th_pp) {      \
                int iR = _i + _subbox_lo[0]; iR += 0;                   \
                int i##x##R = _i + _bx_lo[0]; i##x##R += 0;             \
                T &thisR = * _th_pp; const T & x##R = _x_pp[_i];

/*@ManDoc:
  The macro ForAllThisBNNXCBNYCBN(T,b,ns,nc,x,bx,nsx,y,by,nsy) is intended to
  facilitate efficient looping over the contents of BaseFabs and objects
  derived from BaseFab.  Special attention has been paid to make it work
  efficiently on vector supercomputers.

  This macro acts upon the BaseFab *this and in addition is able to utiliize
  values in the const BaseFab x and const BaseFab y.  The loop runs over the
  points in the intersection of Box b with components starting at ns and
  ending at ns+nc-1.  The reference variables are thisR, xR, and yR
  respectively. As usual the x in xR is replaced by the macro's fifth argument
  and likewise for the y in yR.  The seventh argument nsx is the number of the
  argument in x that corresponds to the ns argument in *this, and the eighth
  argument nsy is the number of the argument in y that corresponds to the ns
  argument in *this.  Boxes bx and by must be the same size as this->box()
  intersected with b.

  Looping macro mnemonics:

    This stands for the current object
    C for a const
    X stands for a BaseFab
    B for a Box
    N for an int
*/
#define ForAllThisBNNXCBNYCBN(T,b,ns,nc,x,bx,nsx,y,by,nsy)              \
{                                                                       \
    assert((ns) >= 0 && (ns) + (nc) <= nComp());                        \
    assert((nsx) >= 0 && (nsx) + (nc) <= (x).nComp());                  \
    assert((nsy) >= 0 && (nsy) + (nc) <= (y).nComp());                  \
    Box _subbox_ = box();                                               \
    _subbox_ &= b;                                                      \
    assert((bx).sameSize(_subbox_));                                    \
    assert((by).sameSize(_subbox_));                                    \
    if(!_subbox_.isEmpty()) {                                           \
        const int *_th_plo = loVect();                                  \
        const int *_th_plen = size();                                   \
        const int *_x_plo = (x).loVect();                               \
        const int *_x_plen = (x).size();                                \
        const int *_y_plo = (y).loVect();                               \
        const int *_y_plen = (y).size();                                \
        const int *_subbox_lo = _subbox_.loVect();                      \
        const int *_subbox_len = _subbox_.size().getVect();             \
        const int *_bx_lo = (bx).loVect();                              \
        const int *_by_lo = (by).loVect();                              \
        T* _th_p = dataPtr(ns);                                         \
        const T* _x_p  = (x).dataPtr(nsx);                              \
        const T* _y_p  = (y).dataPtr(nsy);                              \
        for(int _n = 0; _n < (nc); ++_n) {                              \
            int nR      = _n + ns;  nR      += 0;                       \
            int n##x##R = _n + nsx; n##x##R += 0;                       \
            int n##y##R = _n + nsy; n##y##R += 0;                       \
            T *_th_pp = _th_p                                           \
                + ((_subbox_lo[0] - _th_plo[0])                         \
                   + _n * _th_plen[0]);                                 \
            const T *_x_pp = _x_p                                       \
                + ((_bx_lo[0] - _x_plo[0])                              \
                   + _n * _x_plen[0]);                                  \
            const T *_y_pp = _y_p                                       \
                + ((_by_lo[0] - _y_plo[0])                              \
                   + _n * _y_plen[0]);                                  \
            for(int _i = 0; _i < _subbox_len[0]; ++_i, ++_th_pp) {      \
                int iR = _i + _subbox_lo[0];  iR += 0;                  \
                int i##x##R = _i + _bx_lo[0]; i##x##R += 0;             \
                int i##y##R = _i + _by_lo[0]; i##y##R += 0;             \
                T &thisR = * _th_pp;                                    \
                const T & x##R = _x_pp[_i];                             \
                const T & y##R = _y_pp[_i];

#define ForAllRevXBNYCBNNN(T,x,bx,nsx,y,by,nsy,nc,ri)                   \
{                                                                       \
    assert((nsx) >= 0 && (nsx) + (nc) <= (x).nComp());                  \
    assert((nsy) >= 0 && (nsy) + (nc) <= (y).nComp());                  \
    assert((x).contains(bx));                                           \
    assert((y).contains(by));                                           \
    assert((bx).sameSize(by));                                          \
    const int *_x_plo = (x).loVect();                                   \
    const int *_x_plen = (x).size();                                    \
    const int *_y_plo = (y).loVect();                                   \
    const int *_y_plen = (y).size();                                    \
    const int *_len = (bx).size().getVect();                            \
    const int *_bx_lo = (bx).loVect();                                  \
    const int *_by_lo = (by).loVect();                                  \
    T* _x_p  = (x).dataPtr(nsx);                                        \
    const T* _y_p  = (y).dataPtr(nsy);                                  \
    for(int _n = 0; _n < (nc); ++_n) {                                  \
        int n##x##R = _n + nsx; n##x##R += 0;                           \
        int n##y##R = _n + nsy; n##y##R += 0;                           \
        int _ix = 0;                                                    \
        T *_x_pp = _x_p                                                 \
            + ((_bx_lo[0] - _x_plo[0]) + _len[0] - 1                    \
                + _n * _x_plen[0]);                                     \
        const T *_y_pp = _y_p                                           \
            + ((_by_lo[0] - _y_plo[0])                                  \
                + _n * _y_plen[0]);                                     \
        for(int _i = 0; _i < _len[0]; ++_i, --_ix) {                    \
            T & x##R = _x_pp[_ix];                                      \
            const T & y##R = _y_pp[_i];

/*@ManDoc:
  The macro EndForTX must be used to end all ForAllThisBNNXC,
  ForAllThisBNNXCBN and ForAllThisBNNXCBNYCBN looping constructs.
*/
#define EndForTX }}}}

/*@ManDoc:
  The macro EndFor must be used to end all ForAllXBNN, ForAllXCBNN,
  ForAllThisBNN, and ForAllThisCBNN looping constructs.
*/
#define EndFor }}}

/*@ManDoc:
  The macro EndForPencil must be used to end ForAll*Pencil looping constructs.
*/
#define EndForPencil }}

#elif (CH_SPACEDIM == 2)

#define ForAllThisCPencil(T,b,ns,nc)                                    \
{                                                                       \
    assert(contains(b));                                                \
    assert((ns) >= 0 && (ns) + (nc) <= nComp());                        \
    const int *_th_plo = loVect();                                      \
    const int *_th_plen = size();                                       \
    const int *_b_lo = (b).loVect();                                    \
    const int *_b_len = (b).size().getVect();                           \
    const T* _th_p = dptr;                                              \
    for(int _n = (ns); _n < (ns)+(nc); ++_n) {                          \
        int nR = _n; nR += 0;                                           \
        for(int _j = 0; _j < _b_len[1]; ++_j) {                         \
            const int jR = _j + _b_lo[1];                               \
            const T *_th_pp = _th_p                                     \
                + ((_b_lo[0] - _th_plo[0])                              \
                   + _th_plen[0]*(                                      \
                       (jR - _th_plo[1])                                \
                       + _n * _th_plen[1]));                            \
            const T &thisR = * _th_pp;                                  \
            const int thisLen = _b_len[0];

#define ForAllThisPencil(T,b,ns,nc)                                     \
{                                                                       \
    assert(contains(b));                                                \
    assert((ns) >= 0 && (ns) + (nc) <= nComp());                        \
    const int *_th_plo = loVect();                                      \
    const int *_th_plen = size();                                       \
    const int *_b_lo = (b).loVect();                                    \
    const int *_b_len = (b).size().getVect();                           \
    T* _th_p = dptr;                                                    \
    for(int _n = (ns); _n < (ns)+(nc); ++_n) {                          \
        const int nR = _n;                                              \
        for(int _j = 0; _j < _b_len[1]; ++_j) {                         \
            const int jR = _j + _b_lo[1];                               \
            T *_th_pp = _th_p                                           \
                + ((_b_lo[0] - _th_plo[0])                              \
                   + _th_plen[0]*(                                      \
                       (jR - _th_plo[1])                                \
                       + _n * _th_plen[1]));                            \
            T &thisR = * _th_pp;                                        \
            const int thisLen = _b_len[0];                              \

#define ForAllXBPencil(T,x,b,ns,nc)                                     \
{                                                                       \
    assert((x).contains(b));                                            \
    assert((ns) >= 0 && (ns) + (nc) <= (x).nComp());                    \
    const int *_th_plo = (x).loVect();                                  \
    const int *_th_plen =(x).size();                                    \
    const int *_b_lo = (b).loVect();                                    \
    const int *_b_len = (b).size().getVect();                           \
    T* _th_p = (x) .dataPtr();                                          \
    for(int nR = (ns); nR < (ns)+(nc); ++nR) {                          \
        for(int _j = 0; _j < _b_len[1]; ++_j) {                         \
            const int jR = _j + _b_lo[1];                               \
            T *xR = _th_p                                               \
                + ((_b_lo[0] - _th_plo[0])                              \
                   + _th_plen[0]*(                                      \
                       (jR - _th_plo[1])                                \
                       + nR * _th_plen[1]));                            \
            const int thisLen = _b_len[0];                              \

#define ForAllXBNN(T,x,b,ns,nc)                                         \
{                                                                       \
    assert((x).contains(b));                                            \
    assert((ns) >= 0 && (ns) + (nc) <= (x).nComp());                    \
    const int *_x_plo = (x).loVect();                                   \
    const int *_x_plen = (x).size();                                    \
    const int *_b_lo = (b).loVect();                                    \
    const int *_b_len = (b).size().getVect();                           \
    T* _x_p = (x) .dataPtr();                                           \
    for(int _n = (ns); _n < (ns)+(nc); ++_n) {                          \
        const int nR = _n;                                              \
        for(int _j = 0; _j < _b_len[1]; ++_j) {                         \
            const int jR = _j + _b_lo[1];                               \
            T *_x_pp = _x_p                                             \
                + ((_b_lo[0] - _x_plo[0])                               \
                       + _x_plen[0]*(                                   \
                           (jR - _x_plo[1])                             \
                           + _n * _x_plen[1]));                         \
            for(int _i = 0; _i < _b_len[0]; ++_i, ++_x_pp) {            \
                const int iR = _i + _b_lo[0];                           \
                T &x##R = * _x_pp;

#define ForAllXCBNN(T,x,b,ns,nc)                                        \
{                                                                       \
    assert(x.contains(b));                                              \
    assert((ns) >= 0 && (ns) + (nc) <= (x).nComp());                    \
    const int *_x_plo = (x).loVect();                                   \
    const int *_x_plen = (x).size();                                    \
    const int *_b_lo = (b).loVect();                                    \
    const int *_b_len = (b).size().getVect();                           \
    const T* _x_p = (x).dataPtr();                                      \
    for(int _n = (ns); _n < (ns)+(nc); ++_n) {                          \
        const int nR = _n;                                              \
        for(int _j = 0; _j < _b_len[1]; ++_j) {                         \
            const int jR = _j + _b_lo[1];                               \
            const T *_x_pp = _x_p                                       \
                + ((_b_lo[0] - _x_plo[0])                               \
                       + _x_plen[0]*(                                   \
                           (jR  - _x_plo[1])                            \
                           + _n * _x_plen[1]));                         \
            for(int _i = 0; _i < _b_len[0]; ++_i) {                     \
                const int iR = _i + _b_lo[0];                           \
                const T & x##R = _x_pp[_i];


#define ForAllThisBNN(T,b,ns,nc)                                        \
{                                                                       \
    assert(contains(b));                                                \
    assert((ns) >= 0 && (ns) + (nc) <= nComp());                        \
    const int *_th_plo = loVect();                                      \
    const int *_th_plen = size();                                       \
    const int *_b_lo = (b).loVect();                                    \
    const int *_b_len = (b).size().getVect();                           \
    T* _th_p = dptr;                                                    \
    for(int _n = (ns); _n < (ns)+(nc); ++_n) {                          \
        int nR = _n; nR += 0;                                           \
        for(int _j = 0; _j < _b_len[1]; ++_j) {                         \
            const int jR = _j + _b_lo[1];                               \
            T *_th_pp = _th_p                                           \
                + ((_b_lo[0] - _th_plo[0])                              \
                   + _th_plen[0]*(                                      \
                       (jR - _th_plo[1])                                \
                       + _n * _th_plen[1]));                            \
            for(int _i = 0; _i < _b_len[0]; ++_i, ++_th_pp) {           \
                int iR = _i + _b_lo[0]; iR += 0;                        \
                T &thisR = * _th_pp;

#define ForAllThisCBNN(T,b,ns,nc)                                       \
{                                                                       \
    assert(contains(b));                                                \
    assert((ns) >= 0 && (ns) + (nc) <= nComp());                        \
    const int *_th_plo = loVect();                                      \
    const int *_th_plen = size();                                       \
    const int *_b_lo = (b).loVect();                                    \
    const int *_b_len = (b).size().getVect();                           \
    const T* _th_p = dptr;                                              \
    for(int _n = (ns); _n < (ns)+(nc); ++_n) {                          \
        const int nR = _n;                                              \
        for(int _j = 0; _j < _b_len[1]; ++_j) {                         \
            const int jR = _j + _b_lo[1];                               \
            const T *_th_pp = _th_p                                     \
                + ((_b_lo[0] - _th_plo[0])                              \
                   + _th_plen[0]*(                                      \
                       (_j + _b_lo[1] - _th_plo[1])                     \
                       + _n * _th_plen[1]));                            \
            for(int _i = 0; _i < _b_len[0]; ++_i) {                     \
                const int iR = _i + _b_lo[0];                           \
                const T &thisR = _th_pp[_i];

#define ForAllThisBNNXC(T,b,ns,nc,x,nss)                                \
{                                                                       \
    assert((ns) >= 0 && (ns) + (nc) <= nComp());                        \
    assert((nss) >= 0 && (nss) + (nc) <= (x).nComp());                  \
    Box _subbox_((x).box());                                            \
    _subbox_ &= box();                                                  \
    _subbox_ &= b;                                                      \
    if(!_subbox_.isEmpty()) {                                           \
        const int *_th_plo = loVect();                                  \
        const int *_th_plen = size();                                   \
        const int *_x_plo = (x).loVect();                               \
        const int *_x_plen = (x).size();                                \
        const int *_subbox_lo = _subbox_.loVect();                      \
        const int *_subbox_len = _subbox_.size().getVect();             \
        T* _th_p = dataPtr(ns);                                         \
        const T* _x_p  = (x).dataPtr(nss);                              \
        for(int _n = 0; _n < (nc); ++_n) {                              \
            int nR = _n; nR += 0;                                       \
            for(int _j = 0; _j < _subbox_len[1]; ++_j) {                \
                const int jR = _j + _subbox_lo[1];                      \
                T *_th_pp = _th_p                                       \
                    + ((_subbox_lo[0] - _th_plo[0])                     \
                       + _th_plen[0]*(                                  \
                           (jR - _th_plo[1])                            \
                           + _n * _th_plen[1]));                        \
                const T *_x_pp = _x_p                                   \
                    + ((_subbox_lo[0] - _x_plo[0])                      \
                       + _x_plen[0]*(                                   \
                           (jR - _x_plo[1])                             \
                           + _n * _x_plen[1]));                         \
                for(int _i = 0; _i < _subbox_len[0]; ++_i, ++_th_pp) {  \
                    int iR = _i + _subbox_lo[0]; iR += 0;               \
                    T &thisR = * _th_pp; const T & x##R = _x_pp[_i];

#define ForAllThisCBNNXC(T,b,ns,nc,x,nss)                                \
{                                                                       \
    assert((ns) >= 0 && (ns) + (nc) <= nComp());                        \
    assert((nss) >= 0 && (nss) + (nc) <= (x).nComp());                  \
    Box _subbox_((x).box());                                            \
    _subbox_ &= box();                                                  \
    _subbox_ &= b;                                                      \
    if(!_subbox_.isEmpty()) {                                           \
        const int *_th_plo = loVect();                                  \
        const int *_th_plen = size();                                   \
        const int *_x_plo = (x).loVect();                               \
        const int *_x_plen = (x).size();                                \
        const int *_subbox_lo = _subbox_.loVect();                      \
        const int *_subbox_len = _subbox_.size().getVect();             \
        const T* _th_p = dataPtr(ns);                                         \
        const T* _x_p  = (x).dataPtr(nss);                              \
        for(int _n = 0; _n < (nc); ++_n) {                              \
            int nR = _n; nR += 0;                                       \
            for(int _j = 0; _j < _subbox_len[1]; ++_j) {                \
                const int jR = _j + _subbox_lo[1];                      \
                const T *_th_pp = _th_p                                       \
                    + ((_subbox_lo[0] - _th_plo[0])                     \
                       + _th_plen[0]*(                                  \
                           (jR - _th_plo[1])                            \
                           + _n * _th_plen[1]));                        \
                const T *_x_pp = _x_p                                   \
                    + ((_subbox_lo[0] - _x_plo[0])                      \
                       + _x_plen[0]*(                                   \
                           (jR - _x_plo[1])                             \
                           + _n * _x_plen[1]));                         \
                for(int _i = 0; _i < _subbox_len[0]; ++_i, ++_th_pp) {  \
                    int iR = _i + _subbox_lo[0]; iR += 0;               \
                    const T &thisR = * _th_pp; const T & x##R = _x_pp[_i];

#define ForAllThisBNNXCBN(T,b,ns,nc,x,bx,nss)                           \
{                                                                       \
    assert((ns) >= 0 && (ns) + (nc) <= nComp());                        \
    assert((nss) >= 0 && (nss) + (nc) <= (x).nComp());                  \
    assert(bx.sameSize((b)));                                             \
    if(!((b).isEmpty())) {                                                  \
        const int *_th_plo = loVect();                                  \
        const int *_th_plen = size();                                   \
        const int *_x_plo = (x).loVect();                               \
        const int *_x_plen = (x).size();                                \
        const int *_subbox_lo = (b).loVect();                             \
        const int *_subbox_len = (b).size().getVect();                    \
        const int *_bx_lo = (bx).loVect();                              \
        T* _th_p = dataPtr(ns);                                         \
        int nR = ns; int n##x##R = nss;                                 \
        const T* _x_p  = (x).dataPtr(nss);                              \
        for(int _n = 0; _n < (nc); ++_n, ++nR, ++n##x##R ) {             \
            for(int _j = 0; _j < _subbox_len[1]; ++_j) {                \
                const int jR = _j + _subbox_lo[1];                      \
                const int j##x##R = _j + _bx_lo[1];                     \
                T *_th_pp = _th_p                                       \
                    + ((_subbox_lo[0] - _th_plo[0])                     \
                       + _th_plen[0]*(                                  \
                           (jR - _th_plo[1])                            \
                           + _n * _th_plen[1]));                        \
                const T *_x_pp = _x_p                                   \
                    + ((_bx_lo[0] - _x_plo[0])                          \
                       + _x_plen[0]*(                                   \
                           (j##x##R - _x_plo[1])                        \
                           + _n * _x_plen[1]));                         \
                int iR = _subbox_lo[0]; int i##x##R = _bx_lo[0];        \
                for(int _i = 0; _i < _subbox_len[0]; ++_i, ++_th_pp, ++_x_pp, ++iR, ++i##x##R) {  \
                    T &thisR = * _th_pp; const T & x##R = * _x_pp;

#define ForAllThisBNNXCBNYCBN(T,b,ns,nc,x,bx,nsx,y,by,nsy)              \
{                                                                       \
    assert((ns) >= 0 && (ns) + (nc) <= nComp());                        \
    assert((nsx) >= 0 && (nsx) + (nc) <= (x).nComp());                  \
    assert((nsy) >= 0 && (nsy) + (nc) <= (y).nComp());                  \
    Box _subbox_ = box();                                               \
    _subbox_ &= b;                                                      \
    assert((bx).sameSize(_subbox_));                                    \
    assert((by).sameSize(_subbox_));                                    \
    if(!_subbox_.isEmpty()) {                                           \
        const int *_th_plo = loVect();                                  \
        const int *_th_plen = size();                                   \
        const int *_x_plo = (x).loVect();                               \
        const int *_x_plen = (x).size();                                \
        const int *_y_plo = (y).loVect();                               \
        const int *_y_plen = (y).size();                                \
        const int *_subbox_lo = _subbox_.loVect();                      \
        const int *_subbox_len = _subbox_.size().getVect();             \
        const int *_bx_lo = (bx).loVect();                              \
        const int *_by_lo = (by).loVect();                              \
        T* _th_p = dataPtr(ns);                                         \
        const T* _x_p  = (x).dataPtr(nsx);                              \
        const T* _y_p  = (y).dataPtr(nsy);                              \
        for(int _n = 0; _n < (nc); ++_n) {                              \
            int nR = _n + ns; nR += 0;                                  \
            int n##x##R = _n + nsx; n##x##R += 0;                       \
            int n##y##R = _n + nsy; n##y##R += 0;                       \
            for(int _j = 0; _j < _subbox_len[1]; ++_j) {                \
                const int jR = _j + _subbox_lo[1];                      \
                const int j##x##R = _j + _bx_lo[1];                     \
                const int j##y##R = _j + _by_lo[1];                     \
                T *_th_pp = _th_p                                       \
                    + ((_subbox_lo[0] - _th_plo[0])                     \
                       + _th_plen[0]*(                                  \
                           (jR - _th_plo[1])                            \
                           + _n * _th_plen[1]));                        \
                const T *_x_pp = _x_p                                   \
                    + ((_bx_lo[0] - _x_plo[0])                          \
                       + _x_plen[0]*(                                   \
                           (j##x##R - _x_plo[1])                        \
                           + _n * _x_plen[1]));                         \
                const T *_y_pp = _y_p                                   \
                    + ((_by_lo[0] - _y_plo[0])                          \
                       + _y_plen[0]*(                                   \
                           (j##y##R - _y_plo[1])                        \
                           + _n * _y_plen[1]));                         \
                for(int _i = 0; _i < _subbox_len[0]; ++_i, ++_th_pp) {  \
                    int iR = _i + _subbox_lo[0];  iR += 0;              \
                    int i##x##R = _i + _bx_lo[0]; i##x##R += 0;         \
                    int i##y##R = _i + _by_lo[0]; i##y##R += 0;         \
                    T &thisR = * _th_pp;                                \
                    const T & x##R = _x_pp[_i];                         \
                    const T & y##R = _y_pp[_i];

#define ForAllRevXBNYCBNNN(T,x,bx,nsx,y,by,nsy,nc,ir)                   \
{                                                                       \
    assert((nsx) >= 0 && (nsx) + (nc) <= (x).nComp());                  \
    assert((nsy) >= 0 && (nsy) + (nc) <= (y).nComp());                  \
    assert((ir) >= 0 && (ir) < SpaceDim);                               \
    assert((x).contains(bx));                                           \
    assert((y).contains(by));                                           \
    assert((bx).sameSize(by));                                          \
    const int *_x_plo = (x).loVect();                                   \
    const int *_x_plen = (x).size();                                    \
    const int *_y_plo = (y).loVect();                                   \
    const int *_y_plen = (y).size();                                    \
    const int *_bx_lo = (bx).loVect();                                  \
    const int *_by_lo = (by).loVect();                                  \
    const int *_len = (bx).size().getVect();                            \
    T* _x_p  = (x).dataPtr(nsx);                                        \
    const T* _y_p  = (y).dataPtr(nsy);                                  \
    for(int _n = 0; _n < (nc); ++_n) {                                  \
        int n##x##R = _n + nsx; n##x##R += 0;                           \
        int n##y##R = _n + nsy; n##y##R += 0;                           \
        for(int _j = 0; _j < _len[1]; ++_j) {                           \
            const int j##x##R = _j + _bx_lo[1];                 \
            const int jrev##x##R = _len[1]-1-_j + _bx_lo[1];            \
            const int j##y##R = _j + _by_lo[1];                 \
            T *_x_pp;                                                   \
            int _ix = 0;                                                \
            int _istrd;                                                 \
            if (ir == 0) {                                              \
                _x_pp = _x_p                                            \
                    + ((_bx_lo[0] - _x_plo[0]) + _len[0] - 1            \
                       + _x_plen[0]*(                                   \
                           (j##x##R - _x_plo[1])                        \
                           + _n * _x_plen[1]));                         \
                _istrd = -1;                                            \
            } else {                                                    \
                _x_pp = _x_p                                            \
                    + ((_bx_lo[0] - _x_plo[0])                          \
                       + _x_plen[0]*(                                   \
                           (jrev##x##R - _x_plo[1])                     \
                           + _n * _x_plen[1]));                         \
                _istrd = 1;                                             \
            }                                                           \
            const T *_y_pp = _y_p                                       \
                    + ((_by_lo[0] - _y_plo[0])                          \
                       + _y_plen[0]*(                                   \
                           (j##y##R - _y_plo[1])                        \
                           + _n * _y_plen[1]));                         \
            int _x_rev = _len[0]-1; _x_rev += 0;                        \
            for(int _i = 0; _i < _len[0]; ++_i, _ix+=_istrd) {          \
                T & x##R = _x_pp[_ix];                                  \
                const T & y##R = _y_pp[_i];


#define EndFor }}}}
#define EndForTX }}}}}
#define EndForPencil }}}

#elif (CH_SPACEDIM == 3)

#define ForAllThisCPencil(T,b,ns,nc)                                    \
{                                                                       \
    assert(contains(b));                                                \
    assert((ns) >= 0 && (ns) + (nc) <= nComp());                        \
    const int *_th_plo = loVect();                                      \
    const int *_th_plen = size();                                       \
    const int *_b_lo = (b).loVect();                                    \
    const int *_b_len = (b).size().getVect();                           \
    const T* _th_p = dptr;                                              \
    for(int _n = (ns); _n < (ns)+(nc); ++_n) {                          \
        int nR = _n; nR += 0;                                           \
        for(int _k = 0; _k < _b_len[2]; ++_k) {                         \
            const int kR = _k + _b_lo[2];                               \
            for(int _j = 0; _j < _b_len[1]; ++_j) {                     \
                const int jR = _j + _b_lo[1];                           \
                const T *_th_pp = _th_p                                 \
                    + ((_b_lo[0] - _th_plo[0])                          \
                       + _th_plen[0]*(                                  \
                           (jR - _th_plo[1])                            \
                           + _th_plen[1]*(                              \
                               (kR - _th_plo[2])                        \
                               + _n * _th_plen[2])));                   \
                const T &thisR = * _th_pp;                              \
                const int thisLen = _b_len[0];

#define ForAllThisPencil(T,b,ns,nc)                                     \
{                                                                       \
    assert(contains(b));                                                \
    assert((ns) >= 0 && (ns) + (nc) <= nComp());                        \
    const int *_th_plo = loVect();                                      \
    const int *_th_plen = size();                                       \
    const int *_b_lo = (b).loVect();                                    \
    const int *_b_len = (b).size().getVect();                           \
    T* _th_p = dptr;                                                    \
    for(int _n = (ns); _n < (ns)+(nc); ++_n) {                          \
        const int nR = _n;                                              \
        for(int _k = 0; _k < _b_len[2]; ++_k) {                         \
            const int kR = _k + _b_lo[2];                               \
            for(int _j = 0; _j < _b_len[1]; ++_j) {                     \
                const int jR = _j + _b_lo[1];                           \
                T *_th_pp = _th_p                                       \
                    + ((_b_lo[0] - _th_plo[0])                          \
                       + _th_plen[0]*(                                  \
                           (jR - _th_plo[1])                            \
                           + _th_plen[1]*(                              \
                               (kR - _th_plo[2])                        \
                               + _n * _th_plen[2])));                   \
                T &thisR = * _th_pp;                                    \
                const int thisLen = _b_len[0];                          \

#define ForAllXBPencil(T,x,b,ns,nc)                                     \
{                                                                       \
    assert((x).contains(b));                                            \
    assert((ns) >= 0 && (ns) + (nc) <= (x).nComp());                    \
    const int *_th_plo = (x).loVect();                                  \
    const int *_th_plen = (x).size();                                   \
    const int *_b_lo = (b).loVect();                                    \
    const int *_b_len = (b).size().getVect();                           \
    T* _th_p = (x) .dataPtr();                                          \
    for(int nR = (ns); nR < (ns)+(nc); ++nR) {                          \
        for(int _k = 0; _k < _b_len[2]; ++_k) {                         \
            const int kR = _k + _b_lo[2];                               \
            for(int _j = 0; _j < _b_len[1]; ++_j) {                     \
                const int jR = _j + _b_lo[1];                           \
                T *xR = _th_p                                           \
                    + ((_b_lo[0] - _th_plo[0])                          \
                       + _th_plen[0]*(                                  \
                           (jR - _th_plo[1])                            \
                           + _th_plen[1]*(                              \
                               (kR - _th_plo[2])                        \
                               + nR * _th_plen[2])));                   \
                const int thisLen = _b_len[0];                         


#define ForAllXBNN(T,x,b,ns,nc)                                         \
{                                                                       \
    assert(x.contains(b));                                              \
    assert((ns) >= 0 && (ns) + (nc) <= (x).nComp());                    \
    const int *_x_plo = (x).loVect();                                   \
    const int *_x_plen = (x).size();                                    \
    const int *_b_lo = (b).loVect();                                    \
    const int *_b_len = (b).size().getVect();                           \
    T* _x_p = (x) .dataPtr();                                           \
    for(int _n = (ns); _n < (ns)+(nc); ++_n) {                          \
        const int nR = _n;                                              \
        for(int _k = 0; _k < _b_len[2]; ++_k) {                         \
            const int kR = _k + _b_lo[2];                               \
            for(int _j = 0; _j < _b_len[1]; ++_j) {                     \
                const int jR = _j + _b_lo[1];                           \
                T *_x_pp = _x_p                                         \
                    + ((_b_lo[0] - _x_plo[0])                           \
                       + _x_plen[0]*(                                   \
                           (jR - _x_plo[1])                             \
                           + _x_plen[1]*(                               \
                               (kR - _x_plo[2])                         \
                               + _n * _x_plen[2])));                    \
                for(int _i = 0; _i < _b_len[0]; ++_i, ++_x_pp) {        \
                    const int iR = _i + _b_lo[0];                       \
                    T &x##R = * _x_pp;

#define ForAllXCBNN(T,x,b,ns,nc)                                        \
{                                                                       \
    assert(x.contains(b));                                              \
    assert((ns) >= 0 && (ns) + (nc) <= (x).nComp());                    \
    const int *_x_plo = (x).loVect();                                   \
    const int *_x_plen = (x).size();                                    \
    const int *_b_lo = (b).loVect();                                    \
    const int *_b_len = (b).size().getVect();                           \
    const T* _x_p = (x).dataPtr();                                      \
    for(int _n = (ns); _n < (ns)+(nc); ++_n) {                          \
        const int nR = _n;                                              \
        for(int _k = 0; _k < _b_len[2]; ++_k) {                         \
            const int kR = _k + _b_lo[2];                               \
            for(int _j = 0; _j < _b_len[1]; ++_j) {                     \
                const int jR = _j + _b_lo[1];                           \
                const T *_x_pp = _x_p                                   \
                    + ((_b_lo[0] - _x_plo[0])                           \
                       + _x_plen[0]*(                                   \
                           (jR  - _x_plo[1])                            \
                           + _x_plen[1]*(                               \
                               (kR - _x_plo[2])                         \
                               + _n * _x_plen[2])));                    \
                for(int _i = 0; _i < _b_len[0]; ++_i) {                 \
                    const int iR = _i + _b_lo[0];                       \
                    const T & x##R = _x_pp[_i];


#define ForAllThisBNN(T,b,ns,nc)                                        \
{                                                                       \
    assert(contains(b));                                                \
    assert((ns) >= 0 && (ns) + (nc) <= nComp());                        \
    const int *_th_plo = loVect();                                      \
    const int *_th_plen = size();                                       \
    const int *_b_lo = (b).loVect();                                    \
    const int *_b_len = (b).size().getVect();                           \
    T* _th_p = dptr;                                                    \
    for(int _n = (ns); _n < (ns)+(nc); ++_n) {                          \
        int nR = _n; nR += 0;                                           \
        for(int _k = 0; _k < _b_len[2]; ++_k) {                         \
            const int kR = _k + _b_lo[2];                               \
            for(int _j = 0; _j < _b_len[1]; ++_j) {                     \
                const int jR = _j + _b_lo[1];                           \
                T *_th_pp = _th_p                                       \
                    + ((_b_lo[0] - _th_plo[0])                          \
                       + _th_plen[0]*(                                  \
                           (jR - _th_plo[1])                            \
                           + _th_plen[1]*(                              \
                               (kR - _th_plo[2])                        \
                               + _n * _th_plen[2])));                   \
                for(int _i = 0; _i < _b_len[0]; ++_i, ++_th_pp) {       \
                    int iR = _i + _b_lo[0]; iR += 0;                    \
                    T &thisR = * _th_pp;

#define ForAllThisCBNN(T,b,ns,nc)                                       \
{                                                                       \
    assert(contains(b));                                                \
    assert((ns) >= 0 && (ns) + (nc) <= nComp());                        \
    const int *_th_plo = loVect();                                      \
    const int *_th_plen = size();                                       \
    const int *_b_lo = (b).loVect();                                    \
    const int *_b_len = (b).size().getVect();                           \
    const T* _th_p = dptr;                                              \
    for(int _n = (ns); _n < (ns)+(nc); ++_n) {                          \
        const int nR = _n;                                              \
        for(int _k = 0; _k < _b_len[2]; ++_k) {                         \
            const int kR = _k + _b_lo[2];                               \
            for(int _j = 0; _j < _b_len[1]; ++_j) {                     \
                const int jR = _j + _b_lo[1];                           \
                const T *_th_pp = _th_p                                 \
                    + ((_b_lo[0] - _th_plo[0])                          \
                       + _th_plen[0]*(                                  \
                           (jR - _th_plo[1])                            \
                           + _th_plen[1]*(                              \
                               (kR - _th_plo[2])                        \
                               + _n * _th_plen[2])));                   \
                for(int _i = 0; _i < _b_len[0]; ++_i) {                 \
                    const int iR = _i + _b_lo[0];                       \
                    const T &thisR = _th_pp[_i];

#define ForAllThisBNNXC(T,b,ns,nc,x,nss)                                \
{                                                                       \
    assert((ns) >= 0 && (ns) + (nc) <= nComp());                        \
    assert((nss) >= 0 && (nss) + (nc) <= (x).nComp());                  \
    Box _subbox_((x).box());                                            \
    _subbox_ &= box();                                                  \
    _subbox_ &= b;                                                      \
    if(!_subbox_.isEmpty()) {                                           \
        const int *_th_plo = loVect();                                  \
        const int *_th_plen = size();                                   \
        const int *_x_plo = (x).loVect();                               \
        const int *_x_plen = (x).size();                                \
        const int *_subbox_lo = _subbox_.loVect();                      \
        const int *_subbox_len = _subbox_.size().getVect();             \
        T* _th_p = dataPtr(ns);                                         \
        const T* _x_p  = (x).dataPtr(nss);                              \
        for(int _n = 0; _n < (nc); ++_n) {                              \
            int nR = _n; nR += 0;                                       \
            for(int _k = 0; _k < _subbox_len[2]; ++_k) {                \
                const int kR = _k + _subbox_lo[2];                      \
                for(int _j = 0; _j < _subbox_len[1]; ++_j) {            \
                    const int jR = _j + _subbox_lo[1];                  \
                    T *_th_pp = _th_p                                   \
                        + ((_subbox_lo[0] - _th_plo[0])                 \
                           + _th_plen[0]*(                              \
                               (jR - _th_plo[1])                        \
                               + _th_plen[1]*(                          \
                                   (kR - _th_plo[2])                    \
                                   + _n * _th_plen[2])));               \
                    const T *_x_pp = _x_p                               \
                        + ((_subbox_lo[0] - _x_plo[0])                  \
                           + _x_plen[0]*(                               \
                               (jR - _x_plo[1])                         \
                               + _x_plen[1]*(                           \
                                   (kR - _x_plo[2])                     \
                                   + _n * _x_plen[2])));                \
                    for(int _i = 0; _i < _subbox_len[0]; ++_i, ++_th_pp) { \
                        int iR = _i + _subbox_lo[0]; iR += 0;           \
                        T &thisR = * _th_pp; const T & x##R = _x_pp[_i];

#define ForAllThisCBNNXC(T,b,ns,nc,x,nss)                                \
{                                                                       \
    assert((ns) >= 0 && (ns) + (nc) <= nComp());                        \
    assert((nss) >= 0 && (nss) + (nc) <= (x).nComp());                  \
    Box _subbox_((x).box());                                            \
    _subbox_ &= box();                                                  \
    _subbox_ &= b;                                                      \
    if(!_subbox_.isEmpty()) {                                           \
        const int *_th_plo = loVect();                                  \
        const int *_th_plen = size();                                   \
        const int *_x_plo = (x).loVect();                               \
        const int *_x_plen = (x).size();                                \
        const int *_subbox_lo = _subbox_.loVect();                      \
        const int *_subbox_len = _subbox_.size().getVect();             \
        const T* _th_p = dataPtr(ns);                                         \
        const T* _x_p  = (x).dataPtr(nss);                              \
        for(int _n = 0; _n < (nc); ++_n) {                              \
            int nR = _n; nR += 0;                                       \
            for(int _k = 0; _k < _subbox_len[2]; ++_k) {                \
                const int kR = _k + _subbox_lo[2];                      \
                for(int _j = 0; _j < _subbox_len[1]; ++_j) {            \
                    const int jR = _j + _subbox_lo[1];                  \
                    const T *_th_pp = _th_p                                   \
                        + ((_subbox_lo[0] - _th_plo[0])                 \
                           + _th_plen[0]*(                              \
                               (jR - _th_plo[1])                        \
                               + _th_plen[1]*(                          \
                                   (kR - _th_plo[2])                    \
                                   + _n * _th_plen[2])));               \
                    const T *_x_pp = _x_p                               \
                        + ((_subbox_lo[0] - _x_plo[0])                  \
                           + _x_plen[0]*(                               \
                               (jR - _x_plo[1])                         \
                               + _x_plen[1]*(                           \
                                   (kR - _x_plo[2])                     \
                                   + _n * _x_plen[2])));                \
                    for(int _i = 0; _i < _subbox_len[0]; ++_i, ++_th_pp) { \
                        int iR = _i + _subbox_lo[0]; iR += 0;           \
                        const T &thisR = * _th_pp; const T & x##R = _x_pp[_i];

#define ForAllThisBNNXCBN(T,b,ns,nc,x,bx,nss)                           \
{                                                                       \
    assert((ns) >= 0 && (ns) + (nc) <= nComp());                        \
    assert((nss) >= 0 && (nss) + (nc) <= (x).nComp());                  \
    assert((bx).sameSize((b)));                                           \
    if(!((b).isEmpty())) {                                                  \
        const int *_th_plo = loVect();                                  \
        const int *_th_plen = size();                                   \
        const int *_x_plo = (x).loVect();                               \
        const int *_x_plen = (x).size();                                \
        const int *_subbox_lo = (b).loVect();                             \
        const int *_subbox_len = (b).size().getVect();                    \
        const int *_bx_lo = (bx).loVect();                              \
        T* _th_p = dataPtr(ns);                                         \
        const T* _x_p  = (x).dataPtr(nss);                              \
        for(int _n = 0; _n < (nc); ++_n) {                              \
            int nR = _n + ns; nR += 0;                                  \
            int n##x##R = _n + nss; n##x##R += 0;                       \
            for(int _k = 0; _k < _subbox_len[2]; ++_k) {                \
                const int kR = _k + _subbox_lo[2];                      \
                const int k##x##R = _k + _bx_lo[2];                     \
                for(int _j = 0; _j < _subbox_len[1]; ++_j) {            \
                    const int jR = _j + _subbox_lo[1];                  \
                    const int j##x##R = _j + _bx_lo[1];                 \
                    T *_th_pp = _th_p                                   \
                        + ((_subbox_lo[0] - _th_plo[0])                 \
                           + _th_plen[0]*(                              \
                               (jR - _th_plo[1])                        \
                               + _th_plen[1]*(                          \
                                   (kR - _th_plo[2])                    \
                                   + _n * _th_plen[2])));               \
                    const T *_x_pp = _x_p                               \
                        + ((_bx_lo[0] - _x_plo[0])                      \
                           + _x_plen[0]*(                               \
                               (j##x##R - _x_plo[1])                    \
                               + _x_plen[1]*(                           \
                                   (k##x##R - _x_plo[2])                \
                                   + _n * _x_plen[2])));                \
                    for(int _i = 0; _i < _subbox_len[0]; ++_i, ++_th_pp) { \
                        int iR = _i + _subbox_lo[0]; iR += 0;           \
                        int i##x##R = _i + _bx_lo[0]; i##x##R += 0;     \
                        T &thisR = * _th_pp; const T & x##R = _x_pp[_i];

#define ForAllThisBNNXCBNYCBN(T,b,ns,nc,x,bx,nsx,y,by,nsy)              \
{                                                                       \
    assert((ns) >= 0 && (ns) + (nc) <= nComp());                        \
    assert((nsx) >= 0 && (nsx) + (nc) <= (x).nComp());                  \
    assert((nsy) >= 0 && (nsy) + (nc) <= (y).nComp());                  \
    Box _subbox_(box());                                                \
    _subbox_ &= b;                                                      \
    assert((bx).sameSize(_subbox_));                                    \
    assert((by).sameSize(_subbox_));                                    \
    if(!_subbox_.isEmpty()) {                                           \
        const int *_th_plo = loVect();                                  \
        const int *_th_plen = size();                                   \
        const int *_x_plo = (x).loVect();                               \
        const int *_x_plen = (x).size();                                \
        const int *_y_plo = (y).loVect();                               \
        const int *_y_plen = (y).size();                                \
        const int *_subbox_lo = _subbox_.loVect();                      \
        const int *_subbox_len = _subbox_.size().getVect();             \
        const int *_bx_lo = (bx).loVect();                              \
        const int *_by_lo = (by).loVect();                              \
        T* _th_p = dataPtr(ns);                                         \
        const T* _x_p  = (x).dataPtr(nsx);                              \
        const T* _y_p  = (y).dataPtr(nsy);                              \
        for(int _n = 0; _n < (nc); ++_n) {                              \
            int nR = _n + ns; nR += 0;                                  \
            int n##x##R = _n + nsx; n##x##R += 0;                       \
            int n##y##R = _n + nsy; n##y##R += 0;                       \
            for(int _k = 0; _k < _subbox_len[2]; ++_k) {                \
                const int kR = _k + _subbox_lo[2];                      \
                const int k##x##R = _k + _bx_lo[2];                     \
                const int k##y##R = _k + _by_lo[2];                     \
                for(int _j = 0; _j < _subbox_len[1]; ++_j) {            \
                    const int jR = _j + _subbox_lo[1];                  \
                    const int j##x##R = _j + _bx_lo[1];                 \
                    const int j##y##R = _j + _by_lo[1];                 \
                    T *_th_pp = _th_p                                   \
                        + ((_subbox_lo[0] - _th_plo[0])                 \
                           + _th_plen[0]*(                              \
                               (jR - _th_plo[1])                        \
                               + _th_plen[1]*(                          \
                                   (kR - _th_plo[2])                    \
                                   + _n * _th_plen[2])));               \
                    const T *_x_pp = _x_p                               \
                        + ((_bx_lo[0] - _x_plo[0])                      \
                           + _x_plen[0]*(                               \
                               (j##x##R - _x_plo[1])                    \
                               + _x_plen[1]*(                           \
                                   (k##x##R - _x_plo[2])                \
                                   + _n * _x_plen[2])));                \
                    const T *_y_pp = _y_p                               \
                        + ((_by_lo[0] - _y_plo[0])                      \
                           + _y_plen[0]*(                               \
                               (j##y##R - _y_plo[1])                    \
                               + _y_plen[1]*(                           \
                                   (k##y##R - _y_plo[2])                \
                                   + _n * _y_plen[2])));                \
                    for(int _i = 0; _i < _subbox_len[0]; ++_i, ++_th_pp) { \
                        int iR = _i + _subbox_lo[0];  iR += 0;          \
                        int i##x##R = _i + _bx_lo[0]; i##x##R += 0;     \
                        int i##y##R = _i + _by_lo[0]; i##y##R += 0;     \
                        T &thisR = * _th_pp;                            \
                        const T & x##R = _x_pp[_i];                     \
                        const T & y##R = _y_pp[_i];

#define ForAllRevXBNYCBNNN(T,x,bx,nsx,y,by,nsy,nc,ir)                   \
{                                                                       \
    assert((ir) >= 0 && (ir) < SpaceDim);                               \
    assert((nsx) >= 0 && (nsx) + (nc) <= (x).nComp());                  \
    assert((nsy) >= 0 && (nsy) + (nc) <= (y).nComp());                  \
    assert((x).contains(bx));                                           \
    assert((y).contains(by));                                           \
    assert((bx).sameSize(by));                                          \
    const int *_x_plo = (x).loVect();                                   \
    const int *_x_plen = (x).size();                                    \
    const int *_y_plo = (y).loVect();                                   \
    const int *_y_plen = (y).size();                                    \
    const int *_bx_lo = (bx).loVect();                                  \
    const int *_by_lo = (by).loVect();                                  \
    const int *_len = (bx).size().getVect();                            \
    T* _x_p  = (x).dataPtr(nsx);                                        \
    const T* _y_p  = (y).dataPtr(nsy);                                  \
    for(int _n = 0; _n < (nc); ++_n) {                                  \
        int n##x##R = _n + nsx; n##x##R += 0;                           \
        int n##y##R = _n + nsy; n##y##R += 0;                           \
        for(int _k = 0; _k < _len[2]; ++_k) {                           \
            const int k##x##R = _k + _bx_lo[2];                         \
            const int krev##x##R = _len[2]-1-_k + _bx_lo[2];            \
            const int k##y##R = _k + _by_lo[2];                         \
            for(int _j = 0; _j < _len[1]; ++_j) {                       \
                const int j##x##R = _j + _bx_lo[1];                     \
                const int jrev##x##R = _len[1]-1-_j + _bx_lo[1];        \
                const int j##y##R = _j + _by_lo[1];                     \
                T *_x_pp;                                               \
                int _ix = 0;                                            \
                int _istrd = 1;                                         \
                if (ir == 0) {                                          \
                    _x_pp = _x_p                                        \
                        + ((_bx_lo[0] - _x_plo[0]) + _len[0]-1          \
                           + _x_plen[0]*(                               \
                               (j##x##R - _x_plo[1])                    \
                               + _x_plen[1]*(                           \
                                   (k##x##R - _x_plo[2])                \
                                   + _n * _x_plen[2])));                \
                    _istrd = -1;                                        \
                } else if (ir == 1) {                                   \
                    _x_pp = _x_p                                        \
                        + ((_bx_lo[0] - _x_plo[0])                      \
                           + _x_plen[0]*(                               \
                               (jrev##x##R - _x_plo[1])                 \
                               + _x_plen[1]*(                           \
                                   (k##x##R - _x_plo[2])                \
                                   + _n * _x_plen[2])));                \
                } else {                                                \
                    _x_pp = _x_p                                        \
                        + ((_bx_lo[0] - _x_plo[0])                      \
                           + _x_plen[0]*(                               \
                               (j##x##R - _x_plo[1])                    \
                               + _x_plen[1]*(                           \
                                   (krev##x##R - _x_plo[2])             \
                                   + _n * _x_plen[2])));                \
                }                                                       \
                const T *_y_pp = _y_p                                   \
                    + ((_by_lo[0] - _y_plo[0])                          \
                       + _y_plen[0]*(                                   \
                           (j##y##R - _y_plo[1])                        \
                           + _y_plen[1]*(                               \
                               (k##y##R - _y_plo[2])                    \
                               + _n * _y_plen[2])));                    \
                for(int _i = 0; _i < _len[0]; ++_i, _ix += _istrd) {    \
                    T & x##R = _x_pp[_ix];                              \
                    const T & y##R = _y_pp[_i];

#define EndFor }}}}}
#define EndForTX }}}}}}
#define EndForPencil }}}}

#endif
/*@ManDoc:
  The Macro ForAllXPencil is a shortened version of ForAllXBPencil
  where the box is defaulted to the entire domain the the
  components run over all components of x
  */
#define ForAllXPencil(T,x)  ForAllXBPencil(T,x,((x).box()),0,((x).nComp()))

/*@ManDoc:
  The macro ForAllX(T,x) is a shortened form of `ForAllXBNN' where the Box
  defaults to the domain of x and the components run over all the components
  of x.
*/
#define ForAllX(T,x)            ForAllXBNN(T,x,((x).box()),0,((x).nComp()))

/*@ManDoc:
  The macro ForAllXC(T,x) is the constant form of ForAllX(T,x).
*/
#define ForAllXC(T,x)           ForAllXCBNN(T,x,((x).box()),0,((x).nComp()))

/*@ManDoc:
  The macro ForAllXB(T,x,b) is a shortened form of `ForAllXBNN' 
  where the components run over all the components of x.
*/
#define ForAllXB(T,x,b)         ForAllXBNN(T,x,(b),0,(x).nComp())

/*@ManDoc:
  The macro ForAllXBC(T,x,b) is the constant form of ForAllXB(T,x,b).
*/
#define ForAllXBC(T,x,b)        ForAllXCBNN(T,x,(b),0,(x).nComp())

/*@ManDoc:
  The macro ForAllThis(T) is a shortened form of `ForAllThisBNN' where the Box
  defaults to the domain of x and the components run over all the components
  of x.
*/
#define ForAllThis(T)           ForAllThisBNN(T,domain,0,nComp())

/*@ManDoc:
  The macro ForAllThisC(T) is the constant form of ForAllThis(T).
*/
#define ForAllThisC(T)          ForAllThisCBNN(T,domain,0,nComp())

/*@ManDoc:
  The macro ForAllThisB(T,b) is a shortened form of `ForAllThisBNN'
  where the components run over all the components of x.
*/
#define ForAllThisB(T,b)        ForAllThisBNN(T,(b),0,nComp())

/*@ManDoc:
  The macro ForAllThisCB(T,b) is the constant form of ForAllThisB(T,b).
*/
#define ForAllThisCB(T,b)       ForAllThisCBNN(T,(b),0,nComp())

/*@ManDoc:
  The macro ForAllThisNN(T,ns,nc) is a shortened form of `ForAllThisBNN'
  where the Box defaults to the domain of *this.
*/
#define ForAllThisNN(T,ns,nc)     ForAllThisBNN(T,domain,ns,nc)

/*@ManDoc:
  The macro ForAllThisXC(T,x) is a shortened form of `ForAllThisBNNXC'
  where the Box defaults to the domain of *this and the components run over
  all the components of *this.
*/
#define ForAllThisXC(T,x)       ForAllThisBNNXC(T,domain,0,nComp(),x,0)

#endif /*CH_BASEFAB_H*/

/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef CH_BASEFABIMPLEM_H
#define CH_BASEFABIMPLEM_H

//
// Implementation.=====================================
//

template <class T>
Arena* BaseFab<T>::s_Arena = NULL;

template <class T>
inline
int
BaseFab<T>::nComp () const
{
  return nvar;
}

template <class T>
inline
const Box&
BaseFab<T>::box () const
{
  return domain;
}

template <class T>
inline
const int*
BaseFab<T>::size () const
{
  return domain.size().getVect();
}

template <class T>
inline
const IntVect&
BaseFab<T>::smallEnd () const
{
  return domain.smallEnd();
}

template <class T>
inline
const IntVect&
BaseFab<T>::bigEnd () const
{
  return domain.bigEnd();
}

template <class T>
inline
const int*
BaseFab<T>::loVect () const
{
  return domain.loVect();
}

template <class T>
inline
const int*
BaseFab<T>::hiVect () const
{
  return domain.hiVect();
}

template <class T>
inline
bool
BaseFab<T>::contains (const BaseFab<T>& fab) const
{
  return box().contains(fab.box()) && nvar <= fab.nvar;
}

template <class T>
inline
bool
BaseFab<T>::contains (const Box& bx) const
{
  return box().contains(bx);
}

template <class T>
inline
T*
BaseFab<T>::dataPtr (int n)
{
  assert(!(dptr == 0));
  assert((n >= 0) && (n < nvar));
  return &dptr[n*numpts];
}

template <class T>
inline
const int*
BaseFab<T>::nCompPtr () const
{
  assert(!(dptr == 0));
  return &nvar;
}

template <class T>
inline
const T*
BaseFab<T>::dataPtr (int n) const
{
  assert(!(dptr == 0));
  assert((n >= 0) && (n < nvar));
  return &dptr[n*numpts];
}

template <class T>
inline
T&
BaseFab<T>::operator() (const IntVect& p,
                        int            n)
{
  assert(n >= 0);
  assert(n < nvar);
  assert(!(dptr == 0));
  assert(domain.contains(p));

  return dptr[domain.index(p)+n*numpts];
}

template <class T>
inline
T&
BaseFab<T>::operator() (const IntVect& p)
{
  assert(!(dptr == 0));
  assert(domain.contains(p));

  return dptr[domain.index(p)];
}

template <class T>
inline
const T&
BaseFab<T>::operator() (const IntVect& p,
                        int            n) const
{
  assert(n >= 0);
  assert(n < nvar);
  assert(!(dptr == 0));
  assert(domain.contains(p));

  return dptr[domain.index(p)+n*numpts];
}

template <class T>
inline
const T&
BaseFab<T>::operator() (const IntVect& p) const
{
  assert(!(dptr == 0));
  assert(domain.contains(p));

  return dptr[domain.index(p)];
}

template <class T>
inline
void
BaseFab<T>::getVal  (T*             data,
                     const IntVect& pos,
                     int            n,
                     int            numcomp) const
{
  const int loc      = domain.index(pos);
  const long size    = domain.numPts();

  assert(!(dptr == 0));
  assert(n >= 0 && n + numcomp <= nvar);

  for (int k = 0; k < numcomp; k++)
    data[k] = dptr[loc+(n+k)*size];
}

template <class T>
inline
void
BaseFab<T>::getVal (T*             data,
                    const IntVect& pos) const
{
  getVal(data,pos,0,nvar);
}

template <class T>
inline
BaseFab<T>&
BaseFab<T>::shift (const IntVect& v)
{
  domain += v;
  return *this;
}

template <class T>
inline
BaseFab<T>&
BaseFab<T>::shift (int idir,
                   int n_cell)
{
  domain.shift(idir,n_cell);
  return *this;
}

template <class T>
inline
BaseFab<T> &
BaseFab<T>::shiftHalf (const IntVect& v)
{
  domain.shiftHalf(v);
  return *this;
}

template <class T>
inline
BaseFab<T> &
BaseFab<T>::shiftHalf (int idir,
                       int n_cell)
{
  domain.shiftHalf(idir,n_cell);
  return *this;
}

template <class T>
inline
void
BaseFab<T>::setVal (T val)
{
  performSetVal(val,box(), 0, nvar);
}

template <class T>
inline
void
BaseFab<T>::setVal (T          x,
                    const Box& bx,
                    int        n)
{
  performSetVal(x,bx,n,1);
}

template <class T>
inline
void
BaseFab<T>::setVal (T   x,
                    int n)
{
  performSetVal(x,domain,n,1);
}

template <class T>
inline
void
BaseFab<T>::setVal (T          x,
                    const Box& b,
                    int        ns,
                    int        num)
{
  performSetVal(x,b,ns,num);
}

template <class T>
inline
BaseFab<T>&
BaseFab<T>::copy (const BaseFab<T>& src,
                  const Box&        srcbox,
                  int               srccomp,
                  const Box&        destbox,
                  int               destcomp,
                  int               numcomp)
{
  assert(srcbox.sameSize(destbox));
  assert(src.box().contains(srcbox));
  assert(domain.contains(destbox));
  assert(srccomp >= 0 && srccomp+numcomp <= src.nComp());
  assert(destcomp >= 0 && destcomp+numcomp <= nvar);
  performCopy(src,srcbox,srccomp,destbox,destcomp,numcomp);
  return *this;
}

template <class T>
inline
BaseFab<T>&
BaseFab<T>::copy (const BaseFab<T>& src)
{
  assert(nvar <= src.nvar);
  assert(domain.sameType(src.domain));
  Box overlap(domain);
  overlap &= src.domain;
  if (!overlap.isEmpty())
    performCopy(src,overlap,0,overlap,0,nvar);
  return *this;
}

template <class T>
inline
BaseFab<T>&
BaseFab<T>::copy (const BaseFab<T>& src,
                  const Box&        destbox)
{
  assert(nvar <= src.nvar);
  assert(domain.contains(destbox));
  Box overlap(destbox);
  overlap &= src.domain;
  if (!overlap.isEmpty())
    performCopy(src,overlap,0,overlap,0,nvar);
  return *this;
}

template <class T>
inline
BaseFab<T>&
BaseFab<T>::copy (const BaseFab<T>& src,
                  int               srccomp,
                  int               destcomp,
                  int               numcomp)
{
  assert(srccomp >= 0 && srccomp + numcomp <= src.nvar);
  assert(destcomp >= 0 && destcomp + numcomp <= nvar);
  Box overlap(domain);
  overlap &= src.domain;
  if (!overlap.isEmpty())
    performCopy(src,overlap,srccomp,overlap,destcomp,numcomp);
  return *this;
}

template <class T>
inline
void
BaseFab<T>::define ()
{
  assert(nvar > 0);
  assert(dptr == 0);
  assert(numpts > 0);
  assert(!aliased);
  //    assert(!(The_FAB_Arena == 0));// not a sufficient test !!!
#ifdef ENABLE_MEMORY_TRACKING  
  if(s_Arena == NULL) s_Arena = new BArena(name().c_str());
#else
  if(s_Arena == NULL) s_Arena = new BArena("");
#endif
  if(s_Arena == NULL)
    MayDay::Error("malloc in basefab failed");
  truesize = nvar*numpts;
  dptr     = static_cast<T*>(s_Arena->alloc(truesize*sizeof(T)));

#ifdef ENABLE_MEMORY_TRACKING  
  s_Arena->bytes+=truesize*sizeof(T)+sizeof(BaseFab<T>);
  if(s_Arena->bytes > s_Arena->peak) 
    s_Arena->peak = s_Arena->bytes;
#endif

  //
  // Now call T::T() on the raw memo'ry so we have valid Ts.
  //
  T* ptr = dptr;

  for (int i = 0; i < truesize; i++, ptr++)
    new (ptr) T;
}

template <class T>
inline
void
BaseFab<T>::undefine ()
{
  //    assert(!(The_FAB_Arena == 0));
  //
  // Call T::~T() on the to-be-destroyed memory.
  //
  if(aliased)
	{
	  dptr = 0;
	  return;
	}
  if(dptr == 0) return;
  T* ptr = dptr;

  for (int i = 0; i < truesize; i++, ptr++)
    ptr->~T();

  s_Arena->free(dptr);

#ifdef ENABLE_MEMORY_TRACKING  
  s_Arena->bytes-= truesize*sizeof(T)+sizeof(BaseFab<T>);
#endif

  dptr = 0;
}

template <class T>
inline
BaseFab<T>::BaseFab ()
  :
  domain(Box()),
  nvar(0),
  numpts(0),
  truesize(0),
  dptr(0),
  aliased(false)
{}

template <class T>
inline
BaseFab<T>::BaseFab (const Box& bx,
                     int        n,
					 T*         alias)
  :
  domain(bx),
  nvar(n),
  numpts(bx.numPts()),
  dptr(0),
  aliased(false)
{
  if(alias != NULL)
	{
	  dptr = alias;
	  aliased = true;
	}
  else
	define();
}


template <class T>
void
BaseFab<T>::resize (const Box& b,
                    int        n,
					T*         alias)
{
  //     nvar   = n;
  //     domain = b;
  //     numpts = domain.numPts();

  //     if (dptr == 0)
  //     {
  //         define();
  //     }
  //     else if (nvar*numpts > truesize)
  //     {
  //         undefine();

//         define();
//     }
  undefine();
  nvar = n;
  domain = b;
  numpts = domain.numPts();
  if(alias != NULL)
	{
	  dptr = alias;
	  aliased = true;
	}
  else
	{
	  aliased = false;
	  define();
	}
}

template <class T>
inline
BaseFab<T>::~BaseFab ()
{
  undefine();
}

template <class T>
inline
void
BaseFab<T>::clear ()
{
  undefine();

  domain = Box();
  nvar   = 0;
  numpts = 0;
}

template <class T>
inline
void
BaseFab<T>::performCopy (const BaseFab<T>& src,
                         const Box&        srcbox,
                         int               srccomp,
                         const Box&        destbox,
                         int               destcomp,
                         int               numcomp)
{
  assert(src.box().contains(srcbox));
  assert(box().contains(destbox));
  assert(destbox.sameSize(srcbox));
  assert(srccomp >= 0 && srccomp+numcomp <= src.nComp());
  assert(destcomp >= 0 && destcomp+numcomp <= nComp());

  ForAllThisBNNXCBN(T, destbox, destcomp, numcomp, src, srcbox, srccomp)
    {
      thisR = srcR;
    } EndForTX
  ;
}

template <class T>
inline
void
BaseFab<T>::performSetVal (T         val,
                           const Box& bx,
                           int        ns,
                           int        num)
{
  assert(domain.contains(bx));
  assert(ns >= 0 && ns + num <= nvar);

  if (bx == domain)
    {
      T* data = &dptr[ns*numpts];
      for (long i = 0, N = num*numpts; i < N; i++)
        {
          *data++ = val;
        }
    }
  else
    {
      ForAllThisBNN(T,bx,ns,num)
        {
          thisR = val;
        } EndFor
            }
}


template <class T>
inline
void
BaseFab<T>::copy(const Box& RegionFrom, 
                 const Interval& Cdest, 
                 const Box& RegionTo, 
                 const BaseFab<T>& src, 
                 const Interval& Csrc)
{
  if((this == &src) && (RegionFrom == RegionTo)) return;
  assert(Cdest.size() == Csrc.size()); 
  copy(src, RegionFrom, Csrc.begin(), RegionTo, 
       Cdest.begin(), Cdest.size());
}

template <class T>
inline
void
BaseFab<T>::linearOut(void* buf, const Box& R, const Interval& comps) const
{
  T* buffer = (T*)buf;
  ForAllThisBNN(T,R,comps.begin(), comps.size())
    {
      *buffer = thisR;
      ++buffer;
    } EndFor;
}

template <class T>
inline
void
BaseFab<T>::linearIn(void* buf, const Box& R, const Interval& comps)
{
  T* buffer = (T*)buf;
  ForAllThisBNN(T,R,comps.begin(), comps.size())
    {
      thisR = *buffer;
      ++buffer;
    } EndFor;

}

template <class T>
inline
int
BaseFab<T>::size(const Box& box, const Interval& comps) const
{
  return box.numPts()*sizeof(T)*comps.size();
}

template <class T>
inline
std::string
BaseFab<T>::name()
{
  std::string rtn = (typeid(T)).name();
  return rtn;
}

#endif /*CH_BASEFAB_H*/

/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef CH_FARRAYBOX_H
#define CH_FARRAYBOX_H

//
// $Id: CFile.cpp,v 1.6 2003/04/21 22:08:00 bvs Exp $
//

#ifndef WRAPPER

#endif




//@Man:
//@Memo: A Fortran Array of Reals
/*@Doc: 

  Fortran Array Boxes (generally called FABs) are objects constructed
  to interface with arrays in Fortran.  Useful operations can be performed
  upon FABs in C++, and they provide a convenient interface to
  Fortran when it is necessary to retreat into that language for
  doing arithmetic operations when performance matters.

  FArrayBox is derived from BaseFab<Real>.
  FArrayBox adds additional useful capabilities which make sense
  for Real types, such as I/O and L**p norms.

  The C pre-processor macro `CH\_SPACEDIM' must be defined to use
  this class.  The internal precision of FArrayBox objects is
  set by defining either `CH\_USE\_FLOAT' or `CH\_USE\_DOUBLE' 

  This class does NOT provide a copy constructor or assignment operator.  */


class FArrayBox 
: public BaseFab<Real>
{


public:

/// {\bf constructors and destructor}

  ///
  /**
     Constructs an invalid FArrayBox with no memory.
  */
  FArrayBox ();

  /*@ManDoc: Constructs an initial FArrayBox with the data space
    allocated but not inititialized. ncomp is the number of components
    (variables) at each data point in the Box.  

*/
  FArrayBox (const Box& b, int ncomp, Real* alias=NULL);
  void define(const Box& b, int ncomp, Real* alias = NULL) {resize(b, ncomp, alias);}

///
/**
   Constructs FArrayBox by reading it from istream.
*/
  explicit FArrayBox (std::istream& is);

  ///
/**
  The (virtual) destructor.
*/
  virtual ~FArrayBox ();

///{\bf norm functions}

  /*@ManDoc: Returns the Lp-norm of this FAB using components
    (comp : comp+numcomp-1).  p < 0  -> ERROR.
    p = 0  -> infinity norm (max norm).
    p = 1  -> sum of ABS(FAB)
    p > 1  -> Lp-norm
  */
  virtual Real norm (int p=2,
                     int comp=0,
                     int numcomp=1) const;

  /*@ManDoc: Returns the Lp-norm of this FAB using components (comp :
    comp+numcomp-1) and within the subbox.  p < 0 -> ERROR.  p = 0 ->
    infinity norm (max norm).  p = 1 -> sum of ABS(FAB) p > 1 ->
    Lp-norm 
*/
  virtual Real norm (const Box& subbox,
                     int        p=2,
                     int        comp=0,
                     int        numcomp=1) const;

  /**  
       Returns sum of pow(fab[i,c],p): i in subbox, comp<=c<comp+numcomp
       p>=2 only 

*/
  virtual Real sumPow(const Box& subbox,
                     int        p=2,
                     int        comp=0,
                     int        numcomp=1) const;

  /// Return the dot product of this FArrayBox with another
  /**
     Return the dot product of this FArrayBox and "fab2" over their common
     box and all components.
  */
  Real dotProduct(const FArrayBox& fab2) const;

  ///{\bf min and max functions}

  ///
  /**
     Returns the minimum value of given component of this FArrayBox.
  */
  Real min (int comp = 0) const;

  ///
/**
   Returns the minimum value of given component of this FArrayBox in
   given subbox.  

*/
  Real min (const Box& subbox,
            int        comp = 0) const;

  ///
  /**
     Returns the maximum value of given component of this FArrayBox.
  */
  Real max (int comp = 0) const;

  ///
/**
   Returns the maximum value of given component of this FArrayBox in
   given subbox.  

*/
  Real max (const Box& subbox,
            int        comp = 0) const;

  ///
/**
  Finds location of minimum value in given component of this
  FArrayBox.  */
  IntVect minIndex (int comp = 0) const;

  /*@ManDoc: Returns location of minimum value in given component  of this FArrayBox in
    given subbox.
  */
  IntVect minIndex (const Box& subbox,
                    int        comp = 0) const;
  ///
  /**
     Returns location of maximum value in given component  of this FArrayBox.
  */
  IntVect maxIndex (int comp = 0) const;

  /*@ManDoc: Returns location of maximum value in given component of
    this FArrayBox in given subbox.  

*/
  IntVect maxIndex (const Box& subbox,
                    int        comp = 0) const;

///{\bf mask functions}

///
  /**
     Computes mask array with value of 1 in cells where this FArrayBox
     has value less than val, 0 otherwise.  mask is resized by this
     function.  The number of cells marked with 1 returned.  

*/
  int maskLT (BaseFab<int>& mask,
              Real             val,
              int           comp = 0) const;

///
  /**
     Computes mask array with value of 1 in cells where this FArrayBox
     has value less than or equal to val, 0 otherwise.  mask is
     resized by this function.  The number of cells marked with 1
     returned.  

*/
  int maskLE (BaseFab<int>& mask,
              Real             val,
              int           comp = 0) const;
///
  /**
     Computes mask array with value of 1 in cells where this FArrayBox
     has value equal to val, 0 otherwise.  mask is resized by this
     function.  The number of cells marked with 1 returned.  

*/
  int maskEQ (BaseFab<int>& mask,
              Real             val,
              int           comp = 0) const;
///
  /**
     Computes mask array with value of 1 in cells where this FArrayBox
     has value greater than val, 0 otherwise.  mask is resized by this
     function.  The number of cells marked with 1 returned.  

*/
  int maskGT (BaseFab<int>& mask,
              Real             val,
              int           comp = 0) const;

///
  /**
     Computes mask array with value of 1 in cells where this FArrayBox
     has value greater than or equal to val, 0 otherwise.  mask is
     resized by this function.  The number of cells marked with 1
     returned.

*/
  int maskGE (BaseFab<int>& mask,
              Real             val,
              int           comp = 0) const;


  /// {\bf Arithmetic functions}

  ///
  /**
     Modifies this FArrayBox by replacing each value with its absolute value.

  */
  void abs ();

  ///
  /**
     Modifies this FArrayBox by replacing eahc value with its absolute value,
     for components (comp: comp+numcomp-1).

  */
  void abs (int comp,
            int numcomp=1);

  ///
  /**
     Modifies this FArrayBox by replacing eahc value with its absolute value,
     for components (comp: comp+numcomp-1) and within the subbox.

  */
  void abs (const Box& subbox,
            int        comp = 0,
            int        numcomp=1);

///
/**
   Returns sum of given component of FArrayBox.
*/
  Real sum (int comp,
            int numcomp = 1) const;

  /*@ManDoc: Returns sum of component of this FArrayBox in given
    subbox.  

*/
  Real sum (const Box& subbox,
            int        comp,
            int        numcomp = 1) const;

  ///
/**
   Modifies this FArrayBox by replacing each value x with v/x.  

*/
  FArrayBox& invert (Real v);
  ///
/**
   Modifies this FArrayBox by replacing each value x with v/x.  For
   given range of components and within given subbox.  

*/
  FArrayBox& invert (Real          v,
                     const Box& subbox,
                     int        comp=0,
                     int        numcomp=1);
  ///
/**
   Modifies this FArrayBox by replacing each value x with v/x.  For
   given range of components.

*/
  FArrayBox& invert (Real   v,
                     int comp,
                     int numcomp=1);



  ///
/**
   Modifies this FArrayBox by replacing each value with its additive
   inverse.  For given range of components and within given subbox.  

*/
  FArrayBox& negate (const Box& subbox,
                     int        comp=0,
                     int        numcomp=1);
  ///
/**
   Modifies this FArrayBox by replacing each value with its additive
   inverse.  For given range of components.

*/
  FArrayBox& negate (int comp,
                     int numcomp=1);
  ///
/**
   Modifies this FArrayBox by replacing each value with its additive
   inverse.  

*/
  FArrayBox& negate ();

///
/**
   Modifies this FArrayBox by adding the scalar Real r to all values.  For
   given range of components and within given subbox. 

*/
  FArrayBox& plus (Real          r,
                   const Box& b,
                   int        comp=0,
                   int        numcomp=1);

///
/**
   Modifies this FArrayBox by adding the scalar Real r to all values.  For
   given range of components. 

*/
  FArrayBox& plus (Real   r,
                   int comp,
                   int numcomp=1);


///
/**
   Modifies this FArrayBox by adding the scalar Real r to all values.

*/
  FArrayBox& operator+= (Real r);

///
/**
   Modifies this FArrayBox by pointwise addition of the values of the
   argument FArrayBox.  You might come to grief if the domains of the
   FArrayBoxes don't match.

*/
   FArrayBox& operator+= (const FArrayBox& f);

///
/**
   Modifies this FArrayBox by pointwise addition of the values of the
   argument FArrayBox.  You might come to grief if the domains of the
   FArrayBoxes don't match.  The same as += operator.  

*/
  FArrayBox& plus (const FArrayBox& src);

///
/**
   Modifies this FArrayBox by adding the scalar Real r to all values.

*/
  FArrayBox& plus (Real r);
  FArrayBox& plus_real (Real r) {return this->plus(r);}
///
  /**
     Modifies this FArrayBox by pointwise scaled addition of the
     argument FArrayBox (a[i] <- a[i] + scale * src[i]).  Uses domain
     of the intersection of these two FArrayBoxes.

 */
  FArrayBox& plus (const FArrayBox& src, const Real& scale);
///
  /**
     Modifies this FArrayBox by pointwise scaled addition of the
     argument FArrayBox (a[i] <- a[i] + scale * src[i]).  Uses domain
     of the intersection of these two FArrayBoxes.
 */
  FArrayBox& plus (const FArrayBox& src, const Real& scale, 
                   int                srccomp,
                   int                destcomp,
                   int                numcomp=1);

  /*@ManDoc: Modifies this FArrayBox by pointwise addition of values
    in the argument FArrayBox.  Adds src's components
    (srccomp:srccomp+numcomp-1) to this FArrayBox's components
    (destcomp:destcomp+numcomp-1) where the domains of the two
    FArrayBoxes intersect.  

*/
  FArrayBox& plus (const FArrayBox& src,
                   int                srccomp,
                   int                destcomp,
                   int                numcomp=1);

  /*@ManDoc: Modifies this FArrayBox by pointwise addition of values
    in the argument FArrayBox.  Adds src's components
    (srccomp:srccomp+numcomp-1) to this FArrayBox's components
    (destcomp:destcomp+numcomp-1) where the domain of this FArrayBox
    intersects the subbox.  NOTE: subbox must be contained in this
    FAB.  

*/
  FArrayBox& plus (const FArrayBox& src,
                   const Box&         subbox,
                   int                srccomp,
                   int                destcomp,
                   int                numcomp=1);

  /*@ManDoc: Modifies this FArrayBox by pointwise addition of values
               in the argument FArrayBox.  Adds src's components
               (srccomp:srccomp+numcomp-1) in the Box srcbox to this
               FArrayBox's components (destcomp:destcomp+numcomp-1) in
               the Box destbox.  Corresponding locations within the
               two FArrayBoxes are indexed relative to srcbox and
               destbox, and will in general not be the same.  The
               srcbox and destbox must be same size.  The results are
               UNDEFINED if the src and dest FArrayBoxes are the same
               and the srcbox and destbox overlap.

  */
  FArrayBox& plus (const FArrayBox& src,
                   const Box&         srcbox,
                   const Box&         destbox,
                   int                srccomp,
                   int                destcomp,
                   int                numcomp=1);

///
/**
   Modifies this FArrayBox by subtracting the scalar Real r to all values.
    Note: use plus(-r) for more general operations.
  */
  FArrayBox& operator-= (Real r);

///
/**
   Modifies this FArrayBox by pointwise subtraction of the values of the
   argument FArrayBox.  You might come to grief if the domains of the
   FArrayBoxes don't match.

*/
  FArrayBox& operator-= (const FArrayBox& f);

///
/**
   Modifies this FArrayBox by pointwise subtraction of the values of the
   argument FArrayBox.  You might come to grief if the domains of the
   FArrayBoxes don't match.  The same as -= operator.

  */
  FArrayBox& minus (const FArrayBox& src);

  /*@ManDoc: Modifies this FArrayBox by pointwise subtraction of values
    in the argument FArrayBox.  Subtracts src's components
    (srccomp:srccomp+numcomp-1) from this FArrayBox's components
    (destcomp:destcomp+numcomp-1) where the domains of the two
    FArrayBoxes intersect.  

*/
  FArrayBox& minus (const FArrayBox& src,
                    int                srccomp,
                    int                destcomp,
                    int                numcomp=1);

  /*@ManDoc: Modifies this FArrayBox by pointwise subtraction of values
    in the argument FArrayBox.  Subtracts src's components
    (srccomp:srccomp+numcomp-1) from this FArrayBox's components
    (destcomp:destcomp+numcomp-1) where the domain of this FArrayBox
    intersects the subbox.  NOTE: subbox must be contained in this
    FAB.  

*/
  FArrayBox& minus (const FArrayBox& src,
                    const Box&         subbox,
                    int                srccomp,
                    int                destcomp,
                    int                numcomp=1);

  /*@ManDoc: Modifies this FArrayBox by pointwise subtraction of values
               in the argument FArrayBox.  Subtracts src's components
               (srccomp:srccomp+numcomp-1) in the Box srcbox from this
               FArrayBox's components (destcomp:destcomp+numcomp-1) in
               the Box destbox.  Corresponding locations within the
               two FArrayBoxes are indexed relative to srcbox and
               destbox, and will in general not be the same.  The
               srcbox and destbox must be same size.  The results are
               UNDEFINED if the src and dest FArrayBoxes are the same
               and the srcbox and destbox overlap.

  */
  FArrayBox& minus (const FArrayBox& src,
                    const Box&         srcbox,
                    const Box&         destbox,
                    int                srccomp,
                    int                destcomp,
                    int                numcomp=1);

///
/**
   Modifies this FArrayBox by multiplying all values by the scalar Real r.
  */
  FArrayBox& operator*= (Real r);

///
/**
   Modifies this FArrayBox by multiplying all values by the scalar Real r.
  */
  FArrayBox& mult (Real r);

///
/**
   Modifies this FArrayBox by multiplying all values by the scalar
   Real r.  For given range of components.

  */
  FArrayBox& mult (Real   r,
                   int comp,
                   int numcomp=1);

///
/**
   Modifies this FArrayBox by multiplying all values by the scalar
   Real r.  For given range of components and within given subbox.

  */
  FArrayBox& mult (Real          r,
                   const Box& b,
                   int        comp=0,
                   int        numcomp=1);

///
/**
   Modifies this FArrayBox by pointwise multiplication of the values by the
   argument FArrayBox.  You might come to grief if the domains of the
   FArrayBoxes don't match.

*/
  FArrayBox& operator*= (const FArrayBox& f);

///
/**
   Modifies this FArrayBox by pointwise multiplication by the values in the
   argument FArrayBox.  You might come to grief if the domains of the
   FArrayBoxes don't match.  The same as the *= operator.

*/
  FArrayBox& mult (const FArrayBox& src);

  /*@ManDoc: Modifies this FArrayBox by pointwise multiplication by values
    in the argument FArrayBox.  Multiplies src's components
    (srccomp:srccomp+numcomp-1) by this FArrayBox's components
    (destcomp:destcomp+numcomp-1) where the domains of the two
    FArrayBoxes intersect.  

*/
  FArrayBox& mult (const FArrayBox& src,
                   int                srccomp,
                   int                destcomp,
                   int                numcomp=1);

  /*@ManDoc: Modifies this FArrayBox by pointwise multiplication by values
    in the argument FArrayBox.  Multiplies src's components
    (srccomp:srccomp+numcomp-1) by this FArrayBox's components
    (destcomp:destcomp+numcomp-1) where the domain of this FArrayBox
    intersects the subbox.  NOTE: subbox must be contained in this
    FAB.  

*/
  FArrayBox& mult (const FArrayBox& src,
                   const Box&         subbox,
                   int                srccomp,
                   int                destcomp,
                   int                numcomp=1);

  /*@ManDoc: Modifies this FArrayBox by pointwise multiplication by values
               in the argument FArrayBox.  Multiplies src's components
               (srccomp:srccomp+numcomp-1) in the Box srcbox by this
               FArrayBox's components (destcomp:destcomp+numcomp-1) in
               the Box destbox.  Corresponding locations within the
               two FArrayBoxes are indexed relative to srcbox and
               destbox, and will in general not be the same.  The
               srcbox and destbox must be same size.  The results are
               UNDEFINED if the src and dest FArrayBoxes are the same
               and the srcbox and destbox overlap.

  */
  FArrayBox& mult (const FArrayBox& src,
                   const Box&         srcbox,
                   const Box&         destbox,
                   int                srccomp,
                   int                destcomp,
                   int                numcomp=1);

///
/**
   Modifies this FArrayBox by dividing all values by the scalar Real r.
  */
  FArrayBox& operator/= (Real r);

///
/**
   Modifies this FArrayBox by dividing all values by the scalar Real r.
  */
  FArrayBox& divide (Real r);

///
/**
   Modifies this FArrayBox by dividing all values by the scalar Real r.
   For given range of components.
*/
  FArrayBox& divide (Real   r,
                     int comp,
                     int numcomp=1);

///
/**
   Modifies this FArrayBox by dividing all values by the scalar Real
   r.  For given range of components and within given subbox.  

*/
  FArrayBox& divide (Real          r,
                     const Box& b,
                     int        comp=0,
                     int        numcomp=1);

///
/**
   Modifies this FArrayBox by pointwise division of the values by the
   argument FArrayBox.  You might come to grief if the domains of the
   FArrayBoxes don't match.

*/
  FArrayBox& operator/= (const FArrayBox& src);

///
/**
   Modifies this FArrayBox by pointwise division by the values in the
   argument FArrayBox.  You might come to grief if the domains of the
   FArrayBoxes don't match.  The same as the /= operator.

*/
  FArrayBox& divide (const FArrayBox& src);

  /*@ManDoc: Modifies this FArrayBox by pointwise division by values
    in the argument FArrayBox.  Divides this FArrayBox's components
    (destcomp:destcomp+numcomp-1) by src's components
    (srccomp:srccomp+numcomp-1) where the domains of the two
    FArrayBoxes intersect.  

*/
  FArrayBox& divide (const FArrayBox& src,
                     int                srccomp,
                     int                destcomp,
                     int                numcomp=1);

  /*@ManDoc: Modifies this FArrayBox by pointwise division by values
    in the argument FArrayBox.  Divides this FArrayBox's components
    (destcomp:destcomp+numcomp-1) by src's components
    (srccomp:srccomp+numcomp-1) where the domain of this FArrayBox
    intersects the subbox.  NOTE: subbox must be contained in this
    FAB.  

*/
  FArrayBox& divide (const FArrayBox& src,
                     const Box&         subbox,
                     int                srccomp,
                     int                destcomp,
                     int                numcomp=1);

  /*@ManDoc: Modifies this FArrayBox by pointwise division by values
               in the argument FArrayBox.  Divides this FArrayBox's
               components (destcomp:destcomp+numcomp-1) in the Box
               destbox by src's components (srccomp:srccomp+numcomp-1)
               in the Box srcbox.  Corresponding locations within the
               two FArrayBoxes are indexed relative to srcbox and
               destbox, and will in general not be the same.  The
               srcbox and destbox must be same size.  The results are
               UNDEFINED if the src and dest FArrayBoxes are the same
               and the srcbox and destbox overlap.

  */
  FArrayBox& divide (const FArrayBox& src,
                     const Box&         srcbox,
                     const Box&         destbox,
                     int                srccomp,
                     int                destcomp,
                     int                numcomp=1);

  Real get(const IntVect& iv, int comp) const { return this->operator()(iv, comp);}
  void set(const IntVect& iv, int comp, Real val) {  this->operator()(iv, comp) = val;}

protected:

  virtual void performCopy (const BaseFab<Real>& src,
                            const Box&        srcbox,
                            int               srccomp,
                            const Box&        destbox,
                            int               destcomp,
                            int               numcomp);

private:
  //
  // These are disallowed.
  //
  FArrayBox (const FArrayBox&);
  FArrayBox& operator= (const FArrayBox&);
};


#endif /*CH_FARRAYBOX_H*/
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

//
// $Id: CFile.cpp,v 1.6 2003/04/21 22:08:00 bvs Exp $
//

using std::cin;
using std::cout;
using std::cerr;
using std::setw;
using std::setprecision;
using std::ios;




FArrayBox::FArrayBox ()
  :
  BaseFab<Real>()
{
}

FArrayBox::FArrayBox (const Box& b,
                      int        n,
					  Real* alias)
  :
  BaseFab<Real>(b,n,alias)
{
  if(alias != NULL) return;
#ifndef NDEBUG
  setVal(1.23456789e10);
#endif
}

FArrayBox::~FArrayBox () {}


Real
FArrayBox::norm (const Box& subbox,
                 int        p,
                 int        comp,
                 int        numcomp) const
{
  assert(comp >= 0 && comp+numcomp <= nComp());
  assert(p >= 0);

  Real* tmp = 0;
  int tmplen = 0;
  Real nrm = 0;
  if (p == 0)
    {
      //here  begins a normed fab function piece
      ForAllThisCPencil(Real,subbox,comp,numcomp)
        {
          const Real* row = &thisR;
          if (tmp == 0)
            {
              tmp = new Real[thisLen];
              tmplen = thisLen;
              for (int i = 0; i < thisLen; i++)
                tmp[i] = Abs(Real(row[i]));
            }
          else
            {
              for (int i = 0; i < thisLen; i++)
                tmp[i] = Max(tmp[i],Real(Abs(row[i])));
            }
        } EndForPencil
            nrm = tmp[0];
      for (int i = 1; i < tmplen; i++)
        nrm = Max(nrm, tmp[i]);

      //here it ends
    }
  else if (p == 1)
    {
    //here begins a normed fab function piece
        ForAllThisCPencil(Real,subbox,comp,numcomp)
        {
            const Real* row = &thisR;
            if (tmp == 0)
            {
                tmp = new Real[thisLen];
                tmplen = thisLen;
                for (int i = 0; i < thisLen; i++)
                    tmp[i] = Abs(Real(row[i]));
            }
            else
            {
                for (int i = 0; i < thisLen; i++)
                    tmp[i] += Abs(Real(row[i]));
            }
        } EndForPencil
        nrm = tmp[0];
        for (int i = 1; i < tmplen; i++)
            nrm += tmp[i];
    //here it ends
    }
  else if (p == 2)
    {
      nrm = sqrt(sumPow(subbox, 2, comp, numcomp));
    }
  else //so standard norms weren't good enough for you?
    {
      Real invpwr = 1.0/p;
      nrm = pow(sumPow(subbox, p, comp, numcomp),invpwr);
    }
  delete [] tmp;
  return nrm;
}

// utility function used in norms and such
// only works for p>=2

Real FArrayBox::sumPow(const Box& subbox,
                       int        p,
                       int        comp,
                       int        numcomp) const
{
  assert(p>=2);
  Real sum = 0;
  Real* tmp = NULL;
  int tmplen = 0;
  if (p == 2)
    {
      ForAllThisCPencil(Real,subbox,comp,numcomp)
        {
          const Real* row = &thisR;
          if (tmp == 0)
            {
              tmp = new Real[thisLen];
              tmplen = thisLen;
              for (int i = 0; i < thisLen; i++)
                tmp[i] = row[i]*row[i];
            }
          else
            {
              for (int i = 0; i < thisLen; i++)
                tmp[i] += row[i]*row[i];
            }
        } EndForPencil
            sum = tmp[0];
        for (int i = 1; i < tmplen; i++)
          sum += tmp[i];
    }
  else //so standard norms weren't good enough for you?
    {
      Real pwr = Real(p);
      ForAllThisCPencil(Real,subbox,comp,numcomp)
        {
          const Real* row = &thisR;
          if (tmp == 0)
            {
              tmp = new Real[thisLen];
              tmplen = thisLen;
              for (int i = 0; i < thisLen; i++)
                tmp[i] = pow(row[i],pwr);
            }
          else
            {
              for (int i = 0; i < thisLen; i++)
                tmp[i] += pow(row[i],pwr);
            }
        } EndForPencil
            sum = tmp[0];
        for (int i = 1; i < tmplen; i++)
          sum += tmp[i];
 
    }
  delete [] tmp;
  return sum;
}

// Take the dot product of "this" and "fab2" over their common box and all
// components.
Real FArrayBox::dotProduct(const FArrayBox& fab2) const
{
  Real dot = 0.0;
  const FArrayBox& fab1 = *this;
  Box commonBox = fab1.box() & fab2.box();
  int startcomp = 0;
  int endcomp = fab1.nComp()-1;
  int numcomp = endcomp+1;

  assert(fab1.nComp() == fab2.nComp());

  if (!commonBox.isEmpty())
  {
	/*
    FORT_DOTPRODUCT(CHF_REAL(dot),
                    CHF_CONST_FRA(fab1),
                    CHF_CONST_FRA(fab2),
                    CHF_BOX(commonBox),
                    CHF_CONST_INT(startcomp),
                    CHF_CONST_INT(endcomp));
	*/
	ForAllThisCBNNXC(Real, commonBox, startcomp, numcomp, fab2, startcomp)
	  {
		dot += thisR * fab2R;
	  } EndForTX
		  ;
  }

  return dot;
}

//here is where the orderedfab functions start

Real
FArrayBox::min (int comp) const
{
  Real *_min_row = 0;
  int _X_len = 0;
  ForAllThisCPencil(Real,domain,comp,1)
    {
      const Real* _row = &thisR;
      if (_min_row == 0)
        {
          _min_row = new Real[thisLen];
          _X_len = thisLen;
          for (int i = 0; i < thisLen; i++)
            _min_row[i] = _row[i];
        }
      else
        {
          for (int i = 0; i < thisLen; i++)
            _min_row[i] = Min(_row[i],_min_row[i]);
        }
    } EndForPencil;

  Real _min = _min_row[0];
  for (int i = 1; i < _X_len; i++)
    _min = Min(_min,_min_row[i]);

  delete [] _min_row;

  return _min;
}


Real
FArrayBox::min (const Box& subbox,
                int        comp) const
{
  Real *_min_row = 0;
  int _X_len = 0;
  ForAllThisCPencil(Real,subbox,comp,1)
    {
      const Real* _row = &thisR;
      if (_min_row == 0)
        {
          _min_row = new Real[thisLen];
          _X_len = thisLen;
          for (int i = 0; i < thisLen; i++)
            _min_row[i] = _row[i];
        }
      else
        {
          for (int i = 0; i < thisLen; i++)
            _min_row[i] = Min(_row[i],_min_row[i]);
        }
    } EndForPencil;

  Real _min = _min_row[0];
  for (int i = 1; i < _X_len; i++)
    _min = Min(_min,_min_row[i]);

  delete [] _min_row;

  return _min;
}


Real
FArrayBox::max (int comp) const
{
  Real *_max_row = 0;
  int _X_len = 0;
  ForAllThisCPencil(Real,domain,comp,1)
    {
      const Real* _row = &thisR;
      if (_max_row== 0)
        {
          _max_row = new Real[thisLen];
          _X_len = thisLen;
          for (int i = 0; i < thisLen; i++)
            _max_row[i] = _row[i];
        }
      else
        {
          for (int i = 0; i < thisLen; i++)
            _max_row[i] = Max(_row[i],_max_row[i]);
        }
    } EndForPencil;

  Real _max = _max_row[0];
  for (int i = 1; i < _X_len; i++)
    _max = Max(_max,_max_row[i]);

  delete [] _max_row;

  return _max;
}


Real
FArrayBox::max (const Box& subbox,
                int        comp) const
{
  Real *_max_row = 0;
  int _X_len = 0;
  ForAllThisCPencil(Real,subbox,comp,1)
    {
      const Real* _row = &thisR;
      if (_max_row == 0)
        {
          _max_row = new Real[thisLen];
          _X_len = thisLen;
          for (int i = 0; i < thisLen; i++)
            _max_row[i] = _row[i];
        }
      else
        {
          for (int i = 0; i < thisLen; i++)
            _max_row[i] = Max(_row[i],_max_row[i]);
        }
    } EndForPencil;

  Real _max = _max_row[0];
  for (int i = 1; i < _X_len; i++)
    _max = Max(_max,_max_row[i]);

  delete [] _max_row;

  return _max;
}


IntVect
FArrayBox::minIndex (int comp) const
{
  IntVect _min_loc(domain.smallEnd());
  Real _min_val = (*this).operator()(_min_loc,comp);
  ForAllThisCBNN(Real,domain,comp,1)
    {
      int nr = nR;
      int nr2 = nr;
      nr = nr2;
      if (thisR < _min_val)
        {
          _min_val = thisR;
          D_EXPR(_min_loc[0] = iR,
                 _min_loc[1] = jR,
                 _min_loc[2] = kR);
        }
    } EndFor;

  return _min_loc;
}


IntVect
FArrayBox::minIndex (const Box& subbox,
                     int        comp) const
{
  IntVect _min_loc(subbox.smallEnd());
  Real _min_val = (*this).operator()(_min_loc,comp);
  ForAllThisCBNN(Real,subbox,comp,1)
    {
      int nr = nR;
      int nr2 = nr;
      nr = nr2;
      if (thisR < _min_val)
        {
          _min_val = thisR;
          D_EXPR(_min_loc[0] = iR,
                 _min_loc[1] = jR,
                 _min_loc[2] = kR);
        }
    } EndFor;

  return _min_loc;
}


IntVect
FArrayBox::maxIndex (int comp) const
{
  IntVect _max_loc(domain.smallEnd());
  Real _max_val = (*this).operator()(_max_loc,comp);
  ForAllThisCBNN(Real,domain,comp,1)
    {
      int nr = nR;
      int nr2 = nr;
      nr = nr2;
      if (thisR > _max_val)
        {
          _max_val = thisR;
          D_EXPR(_max_loc[0] = iR,
                 _max_loc[1] = jR,
                 _max_loc[2] = kR);
        }
    } EndFor;

  return _max_loc;
}


IntVect
FArrayBox::maxIndex (const Box& subbox,
                     int        comp) const
{
  IntVect _max_loc(subbox.smallEnd());
  Real _max_val = (*this).operator()(_max_loc,comp);
  ForAllThisCBNN(Real,subbox,comp,1)
    {
      int nr = nR;
      int nr2 = nr;
      nr = nr2;
      if (thisR > _max_val)
        {
          _max_val = thisR;
          D_EXPR(_max_loc[0] = iR,
                 _max_loc[1] = jR,
                 _max_loc[2] = kR);
        }
    } EndFor;

  return _max_loc;
}



int
FArrayBox::maskLT (BaseFab<int>& mask,
                   Real             val,
                   int           comp) const
{
  mask.resize(domain,1);
  mask.setVal(0);

  int *mptr = mask.dataPtr();
  int cnt = 0;

  ForAllThisCBNN(Real,domain,comp,1)
    {
      int nr = nR;
      int nr2 = nr;
      nr = nr2;
      int ir = iR;
      int jr = jR;
      ir = jr;
      jr = ir;
      int ix = D_TERM(_i, +_j*_b_len[0], +_k*_b_len[0]*_b_len[1]);
      if (thisR < val)
        {
          mptr[ix] = 1;
          cnt++;
        }
    } EndFor;

  return cnt;
}


int
FArrayBox::maskLE (BaseFab<int>& mask,
                   Real             val,
                   int           comp) const
{
  mask.resize(domain,1);
  mask.setVal(0);

  int *mptr = mask.dataPtr();
  int cnt = 0;

  ForAllThisCBNN(Real,domain,comp,1)
    {
      int nr = nR;
      int nr2 = nr;
      nr = nr2;
      int ir = iR;
      int jr = jR;
      ir = jr;
      jr = ir;
      int ix = D_TERM(_i, +_j*_b_len[0], +_k*_b_len[0]*_b_len[1]);
      if (thisR <= val)
        {
          mptr[ix] = 1;
          cnt++;
        }
    } EndFor;

  return cnt;
}


int
FArrayBox::maskEQ (BaseFab<int>& mask,
                   Real             val,
                   int           comp) const
{
  mask.resize(domain,1);
  mask.setVal(0);

  int *mptr = mask.dataPtr();
  int cnt = 0;

  ForAllThisCBNN(Real,domain,comp,1)
    {
      int nr = nR;
      int nr2 = nr;
      nr = nr2;
      int ir = iR;
      int jr = jR;
      ir = jr;
      jr = ir;
      int ix = D_TERM(_i, +_j*_b_len[0], +_k*_b_len[0]*_b_len[1]);
      if (thisR == val)
        {
          mptr[ix] = 1;
          cnt++;
        }
    } EndFor;

  return cnt;
}


int
FArrayBox::maskGT (BaseFab<int>& mask,
                   Real             val,
                   int           comp) const
{
  mask.resize(domain,1);
  mask.setVal(0);

  int *mptr = mask.dataPtr();
  int cnt = 0;

  ForAllThisCBNN(Real,domain,comp,1)
    {
      int nr = nR;
      int nr2 = nr;
      nr = nr2;
      int ir = iR;
      int jr = jR;
      ir = jr;
      jr = ir;
      int ix = D_TERM(_i, +_j*_b_len[0], +_k*_b_len[0]*_b_len[1]);
      if (thisR > val)
        {
          mptr[ix] = 1;
          cnt++;
        }
    } EndFor;

  return cnt;
}


int
FArrayBox::maskGE(BaseFab<int>& mask,
                  Real             val,
                  int           comp) const
{
  mask.resize(domain,1);
  mask.setVal(0);

  int *mptr = mask.dataPtr();
  int cnt = 0;

  ForAllThisCBNN(Real,domain,comp,1)
    {
      int nr = nR;
      int nr2 = nr;
      nr = nr2;
      int ir = iR;
      int jr = jR;
      ir = jr;
      jr = ir;
      int ix = D_TERM(_i, +_j*_b_len[0], +_k*_b_len[0]*_b_len[1]);
      if (thisR >= val)
        {
          mptr[ix] = 1;
          cnt++;
        }
    } EndFor;

  return cnt;
}
//here is where the orderedfab functions end


//here is where the normedfab functions begin


void
FArrayBox::abs ()
{
  ForAllThis(Real)
    {
      thisR = Abs(thisR);
    } EndFor
        }


void
FArrayBox::abs (int comp,
                int numcomp)
{
  ForAllThisNN(Real,comp,numcomp)
    {
      thisR = Abs(thisR);
    } EndFor
        }


void
FArrayBox::abs (const Box& subbox,
                int        comp,
                int        numcomp)
{
  ForAllThisBNN(Real,subbox,comp,numcomp)
    {
      thisR = Abs(thisR);
    } EndFor
        }



Real
FArrayBox::norm (int p,
                 int comp,
                 int numcomp) const
{
  return norm(domain,p,comp,numcomp);
}


//here is where the normedfab functions end

//here is where the arithfab functions begin

FArrayBox&
FArrayBox::plus (Real r)
{
  return operator+=(r);
}


FArrayBox&
FArrayBox::plus (const FArrayBox& x)
{
  return operator+=(x);
}


FArrayBox&
FArrayBox::operator-= (Real r)
{
  return operator+=(-r);
}


FArrayBox&
FArrayBox::minus (const FArrayBox& x)
{
  return operator-=(x);
}



FArrayBox&
FArrayBox::mult (Real r)
{
  return operator*=(r);
}


FArrayBox&
FArrayBox::mult (const FArrayBox& x)
{
  return operator*=(x);
}


FArrayBox&
FArrayBox::divide (Real r)
{
  return operator/=(r);
}


FArrayBox&
FArrayBox::divide (const FArrayBox& x)
{
  return operator/=(x);
}




Real
FArrayBox::sum (int comp,
                int numcomp) const
{
  Real *_sum_row = 0;
  int _sum_len = 0;
  ForAllThisCPencil(Real,domain,comp,numcomp)
    {
      const Real* _row = &thisR;
      if (_sum_row == 0)
        {
          _sum_row = new Real[thisLen];
          _sum_len = thisLen;
          for (int i = 0; i < thisLen; i++)
            _sum_row[i] = _row[i];
        }
      else
        {
          for (int i = 0; i < thisLen; i++)
            _sum_row[i] += _row[i];
        }
    } EndForPencil;

  Real _sum = _sum_row[0];
  for (int i = 1; i < _sum_len; i++)
    _sum += _sum_row[i];
  delete [] _sum_row;

  return _sum;
}


Real
FArrayBox::sum (const Box& subbox,
                int        comp,
                int        numcomp) const
{
  Real *_sum_row = 0;
  int _sum_len = 0;
  ForAllThisCPencil(Real,subbox,comp,numcomp)
    {
      const Real* _row = &thisR;
      if (_sum_row == 0)
        {
          _sum_row = new Real[thisLen];
          _sum_len = thisLen;
          for (int i = 0; i < thisLen; i++)
            _sum_row[i] = _row[i];
        }
      else
        {
          for (int i = 0; i < thisLen; i++)
            {
              _sum_row[i] += _row[i];
            }
        }
    } EndForPencil;

  Real _sum = _sum_row[0];
  for (int i = 1; i < _sum_len; i++)
    _sum += _sum_row[i];
  delete [] _sum_row;

  return _sum;
}


FArrayBox&
FArrayBox::negate ()
{
  ForAllThis(Real)
    {
      thisR = - thisR;
    } EndFor
        return *this;
}


FArrayBox&
FArrayBox::negate (int comp,
                   int numcomp)
{
  ForAllThisNN(Real,comp,numcomp)
    {
      thisR = - thisR;
    } EndFor
        return *this;
}


FArrayBox&
FArrayBox::negate (const Box& b,
                   int        comp,
                   int        numcomp)
{
  ForAllThisBNN(Real,b,comp,numcomp)
    {
      thisR = - thisR;
    } EndFor
        return *this;
}


FArrayBox&
FArrayBox::invert (Real r)
{
  ForAllThis(Real)
    {
      thisR = r/thisR;
    } EndFor
        return *this;
}


FArrayBox&
FArrayBox::invert (Real   r,
                   int comp,
                   int numcomp)
{
  ForAllThisNN(Real,comp,numcomp)
    {
      thisR = r/thisR;
    } EndFor
        return *this;
}


FArrayBox&
FArrayBox::invert (Real          r,
                   const Box& b,
                   int        comp,
                   int        numcomp)
{
  ForAllThisBNN(Real,b,comp,numcomp)
    {
      thisR = r/thisR;
    } EndFor
        return *this;
}


FArrayBox&
FArrayBox::operator += (Real r)
{
  ForAllThis(Real)
    {
      thisR += r;
    } EndFor
        return *this;
}


FArrayBox&
FArrayBox::plus (Real   r,
                 int comp,
                 int numcomp)
{
  ForAllThisNN(Real,comp,numcomp)
    {
      thisR += r;
    } EndFor
        return *this;
}


FArrayBox&
FArrayBox::plus (Real          r,
                 const Box& b,
                 int        comp,
                 int        numcomp)
{
  ForAllThisBNN(Real,b,comp,numcomp)
    {
      thisR += r;
    } EndFor
        return *this;
}
 
// added bvs Tue May 18, PDT 1999

FArrayBox&
FArrayBox::plus(const FArrayBox& src, const Real& scale)
{
  ForAllThisBNNXC(Real, src.box()&domain, 0, nvar, src, 0)
    {
      thisR += srcR*scale;
    } EndForTX
        return *this;
}

FArrayBox&
FArrayBox::plus(const FArrayBox& src, const Real& scale,
                int                srccomp,
                int                destcomp,
                int                numcomp)
{
  ForAllThisBNNXC(Real, src.box()&domain, destcomp, numcomp, src, srccomp)
    {
      thisR += srcR*scale;
    } EndForTX
        return *this;
}


FArrayBox&
FArrayBox::operator += (const FArrayBox& x)
{
  ForAllThisXC(Real,x)
    {
      thisR += xR;
    } EndForTX
        return *this;
}

FArrayBox&
FArrayBox::plus (const FArrayBox& src,
                 int                srccomp,
                 int                destcomp,
                 int                numcomp)
{
  ForAllThisBNNXC(Real,domain,destcomp,numcomp,src,srccomp)
    {
      thisR += srcR;
    } EndForTX
        return *this;
}


FArrayBox&
FArrayBox::plus (const FArrayBox& src,
                 const Box&         subbox,
                 int                srccomp,
                 int                destcomp,
                 int                numcomp)
{
  ForAllThisBNNXC(Real,subbox,destcomp,numcomp,src,srccomp)
    {
      thisR += srcR;
    } EndForTX
        return *this;
}


FArrayBox&
FArrayBox::plus (const FArrayBox& src,
                 const Box&         srcbox,
                 const Box&         destbox,
                 int                srccomp,
                 int                destcomp,
                 int                numcomp)
{
  ForAllThisBNNXCBN(Real,destbox,destcomp,numcomp,src,srcbox,srccomp)
    {
      thisR += srcR;
    } EndForTX
        return *this;
}


FArrayBox&
FArrayBox::operator -= (const FArrayBox& x)
{
  ForAllThisXC(Real,x)
    {
      thisR -= xR;
    } EndForTX
        return *this;
}


FArrayBox&
FArrayBox::minus (const FArrayBox& src,
                  int                srccomp,
                  int                destcomp,
                  int                numcomp)
{
  ForAllThisBNNXC(Real,domain,destcomp,numcomp,src,srccomp)
    {
      thisR -= srcR;
    } EndForTX
        return *this;
}


FArrayBox&
FArrayBox::minus (const FArrayBox& src,
                  const Box&         subbox,
                  int                srccomp,
                  int                destcomp,
                  int                numcomp)
{
  ForAllThisBNNXC(Real,subbox,destcomp,numcomp,src,srccomp)
    {
      thisR -= srcR;
    } EndForTX
        return *this;
}


FArrayBox&
FArrayBox::minus (const FArrayBox& src,
                  const Box&         srcbox,
                  const Box&         destbox,
                  int                srccomp,
                  int                destcomp,
                  int                numcomp)
{
  ForAllThisBNNXCBN(Real,destbox,destcomp,numcomp,src,srcbox,srccomp)
    {
      thisR -= srcR;
    } EndForTX
        return *this;
}


FArrayBox&
FArrayBox::operator *= (Real r)
{
  ForAllThis(Real)
    {
      thisR *= r;
    } EndFor
        return *this;
}


FArrayBox&
FArrayBox::mult (Real   r,
                 int comp,
                 int numcomp)
{
  ForAllThisNN(Real,comp,numcomp)
    {
      thisR *= r;
    } EndFor
        return *this;
}


FArrayBox&
FArrayBox::mult (Real          r,
                 const Box& b,
                 int        comp,
                 int        numcomp)
{
  ForAllThisBNN(Real,b,comp,numcomp)
    {
      thisR *= r;
    } EndFor
        return *this;
}


FArrayBox&
FArrayBox::operator *= (const FArrayBox &x)
{
  ForAllThisXC(Real,x)
    {
      thisR *= xR;
    } EndForTX
        return *this;
}


FArrayBox&
FArrayBox::mult (const FArrayBox& src,
                 int                srccomp,
                 int                destcomp,
                 int                numcomp)
{
  ForAllThisBNNXC(Real,domain,destcomp,numcomp,src,srccomp)
    {
      thisR *= srcR;
    } EndForTX
        return *this;
}


FArrayBox&
FArrayBox::mult (const FArrayBox& src,
                 const Box&         subbox,
                 int                srccomp,
                 int                destcomp,
                 int                numcomp)
{
  ForAllThisBNNXC(Real,subbox,destcomp,numcomp,src,srccomp)
    {
      thisR *= srcR;
    } EndForTX
        return *this;
}


FArrayBox&
FArrayBox::mult (const FArrayBox& src,
                 const Box&         srcbox,
                 const Box&         destbox,
                 int                srccomp,
                 int                destcomp,
                 int                numcomp)
{
  ForAllThisBNNXCBN(Real,destbox,destcomp,numcomp,src,srcbox,srccomp)
    {
      thisR *= srcR;
    } EndForTX
        return *this;
}


FArrayBox&
FArrayBox::operator /= (Real r)
{
  ForAllThis(Real)
    {
      thisR /= r;
    } EndFor
        return *this;
}


FArrayBox&
FArrayBox::divide (Real   r,
                   int comp,
                   int numcomp)
{
  ForAllThisNN(Real,comp,numcomp)
    {
      thisR /= r;
    } EndFor
        return *this;
}


FArrayBox&
FArrayBox::divide (Real          r,
                   const Box& b,
                   int        comp,
                   int        numcomp)
{
  ForAllThisBNN(Real,b,comp,numcomp)
    {
      thisR /= r;
    } EndFor
        return *this;
}


FArrayBox&
FArrayBox::operator /= (const FArrayBox &x)
{
  ForAllThisXC(Real,x)
    {
      thisR /= xR;
    } EndForTX
        return *this;
}


FArrayBox&
FArrayBox::divide (const FArrayBox& src,
                   int                srccomp,
                   int                destcomp,
                   int                numcomp)
{
  ForAllThisBNNXC(Real,domain,destcomp,numcomp,src,srccomp)
    {
      thisR /= srcR;
    } EndForTX
        return *this;
}


FArrayBox&
FArrayBox::divide (const FArrayBox& src,
                   const Box&         subbox,
                   int                srccomp,
                   int                destcomp,
                   int                numcomp)
{
  ForAllThisBNNXC(Real,subbox,destcomp,numcomp,src,srccomp)
    {
      thisR /= srcR;
    } EndForTX
        return *this;
}


FArrayBox&
FArrayBox::divide (const FArrayBox& src,
                   const Box&         srcbox,
                   const Box&         destbox,
                   int                srccomp,
                   int                destcomp,
                   int                numcomp)
{
  ForAllThisBNNXCBN(Real,destbox,destcomp,numcomp,src,srcbox,srccomp)
    {
      thisR /= srcR;
    } EndForTX
        return *this;
}

void FArrayBox::performCopy (const BaseFab<Real>& src,
                             const Box&        srcbox,
                             int               srccomp,
                             const Box&        destbox,
                             int               destcomp,
                             int               numcomp)
{
  BaseFab<Real>::performCopy(src, srcbox, srccomp, destbox, destcomp, numcomp);
/*  FArrayBox& dest = *this;
  FORT_PERFORMCOPY(CHF_FRA(dest),
                   CHF_CONST_FRA(src),
                   CHF_BOX(destbox),
                   CHF_BOX(srcbox),
                   CHF_CONST_INT(srccomp),
                   CHF_CONST_INT(destcomp),
                   CHF_CONST_INT(numcomp));   
*/
}
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef POOL_H
#define POOL_H

// Pool is a class to optimize memory allocation.  It is specialized to
// allocate fixed size chunks of memory specified by ptrSize in the
// constructor.  Its operation is analogous to malloc, not new.  It
// does not initialize the memory in any way.  The constructor can
// optionally specify an initial pool size, and memory alignment.  The
// pool size will grow as needed by calling ::new.  The pool size never
// shrinks.  Memory will be reclaimed at ~Pool().  The units of poolSize
// are number-of-ptrSize-chunks.  The units of alignment are bytes.

  ///
class Pool;
typedef std::list<Pool*> PoolList;

class Pool {
public:
  ///
  Pool(int ptrSize, const char* name = "unnamed", 
       int poolSize = 100, int alignment = sizeof(int));
  ///
  ~Pool();
  ///
  void* getPtr();
  ///
  void returnPtr(void* ptr);
  ///
  long memUsage() const;


  // undocumented function, call this at own risk.  You must
  // be absolutely positive that you have returned all the
  // ptr's you've asked for, or you can have major seg faults.
  void clear();

  // not for public consumption. used in memory tracking code.
  static PoolList* poolList_;
  char name_[64];

protected:
private:
  ///
  Vector<char*> pool_;
  ///
  int ptrSize_;
  ///
  int poolSize_;
  ///
  int alignment_;
  ///
  void* next_; 
  ///
  void* getMoreMemory();
  ///
  Pool(const Pool& rhs);  /// Not implemented
  const Pool& operator=(const Pool& rhs);  // Not implemented

};


#endif
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//


PoolList* Pool::poolList_ = NULL;
bool PoolListInit = false;

// Constructor sets Pool parameters, but doesn't actually allocate any memory.
Pool::Pool(int ptrSize,  const char* name, int poolSize, int alignment)
        : pool_(0),              // An array of pointers to pools
        ptrSize_(ptrSize),       // Chunk size (bytes) in each pool
        poolSize_(poolSize),     // Number of chunks in each pool
        alignment_(alignment),   // allocated memory to be aligned on this boundary.
        next_(0)                 // Points to next available chunk or is null.
{
  if (poolSize_ < 1)
    MayDay::Error("Pool::Pool non positive poolSize.");
  if ((unsigned int)ptrSize_ < sizeof(void*))     // Minimum chunk size is size of pointer.
    ptrSize_ = sizeof(void*);
  if ((unsigned int)alignment_ < sizeof(int))   // alignment_ checked out only for 4 and 8.
    alignment_ = sizeof(int);
  while (ptrSize_ % alignment_ != 0)  // chunk size must be multiple of alignment_
    ptrSize_++;

  if(!PoolListInit)
  {
    poolList_ = new PoolList;
    PoolListInit = true;
  }
  poolList_->insert(poolList_->begin(), this);
  strcpy(name_, name);
}

Pool::~Pool()
{
  for (int i = 0; i < pool_.size(); i++)
    delete [] pool_[i];
  pool_.resize(0);
  poolList_->remove(this);
}

void Pool::clear()
{
  for (int i = 0; i < pool_.size(); i++)
    delete [] pool_[i];
  next_ = 0;
  pool_.resize(0);
}

void* Pool::getMoreMemory() {
  // Create another pool
  char* po;
  if (alignment_ == sizeof(int))
    po = (char*)new int [ptrSize_ * poolSize_ / sizeof(int)];
  else if (alignment_ == sizeof(double))
    po = (char*)new double [ptrSize_ * poolSize_ / sizeof(double)];
  else
    MayDay::Error("Pool::Pool : alignments other than int or double not supported");

  pool_.push_back(po);
  // Thread pool:
  // The first few bytes in each chunk is interpreted as a pointer
  // which points to the next free chunk.  With multiple pools, and random
  // news and deletes, the pools will become threaded together.  This is ok.
  char* p = po;
  for (int i = 0; i < poolSize_-1; i++)
  {
    *(char**)p = p + ptrSize_;  // Chunk at i points to chunk at i+1.
    p += ptrSize_;
  }
  *(char**)p = 0;  // Chunk at end of list points to null
  return po;       // return first chunk.
}

/// getPtr and returnPtr must be as fast as possible!

void* Pool::getPtr()
{
  if (next_ == 0) next_ = getMoreMemory(); 
  void* result = next_;  // result points to first free chunk in list
  next_ = *(void**)next_;  // point next_ to next free chunk in list.
  return result;
}

///
void Pool::returnPtr(void* ptr)
{
  *(void**)ptr = next_;  // Stamp *ptr with the first free chunk.
  next_ = ptr;           // Make ptr the first free chunk.
}

long Pool::memUsage() const
{
  return ptrSize_ * poolSize_ * pool_.size();
}
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//


// arch dependant setting for what a WORD should be
// WORD is the smallest efficiently accessible memory
// object.  WORD_BITS is the size, in bits, of this object
//
//  Later I can make this object use a Pool, and we can
// avoid all the memory system call overhead for creation and
// use placement new.

#define WORD unsigned int
const int  WORDSIZE = 8*sizeof(unsigned int);

// WORDSIZE is no the same as sizeof(WORD).  This is the number of bits, not chars

using std::ostream;


class BitSetIterator;

///
/* stores a contiguous memory chunk and represents true with a 1 bit
   and false with a 0 bit. 

   example: 35 bits, set to true, WORDSIZE=8:

   m_index = 5 (need at least 5 WORDS to hold that many bits)
   m_size  = 35

   +-------+-------+-------+-------+-------+
   1111111111111111111111111111111111111111

   now, set bit 35 to 0

   index = 35/WORDSIZE = 4       , hence, we are in the fourth WORD.
   mask  = 35 - 4*WORDSIZE =  3    hence, we are in the third bit of the fourth word

   now, we can use the masks:
static WORD trueMasks[WORDSIZE]; //10000000, 01000000, 00100000, ....

   word[index] = word[index] & ~trueMasks[mask]  (~trueMasks[3] = 11101111)

   now we have:

   +-------+-------+-------+-------+-------+
   1111111111111111111111111111111111101111
*/
class BitSet
{
public:
  ///
  BitSet();

  ///
  BitSet(int bits, bool init);

  ///
  void define(int bits, bool init);

  ///
  BitSet(const BitSet& rhs);

  ///
  BitSet& operator=(const BitSet& rhs);

  ///
  ~BitSet();

  // somewhat slow random access. Fast iterating done
  // with BitSetIterator
  bool operator[](int i) const;

  /* 
      logic operations 
  */

  ///
  void setTrue(int i); // set bit to 1

  ///
  void setFalse(int i); // set bit to 0

  ///
  /**
     returns 'true' if the entire bitset is zero 
  */
  bool isEmpty() const;

  ///
  /**
     returns 'true' if entire bitset is 1 
  */
  bool isFull() const;

  ///
  int size() const;
  static int initialize();

  // not for public, used by memory tracker.
  static long int bytes;
  static long int peak;

private:
  friend class BitSetIterator;

  WORD* m_bits;
  int   m_size;
  int   m_length;  //length of char array, not bit length

  static WORD trueMasks[WORDSIZE]; //10000000, 01000000, 00100000, ....
};

// somewhat slow random access. Fast iterating done
// with BitSetIterator
inline bool BitSet::operator[](int i) const
{
  assert(i>=0);
  assert(i<m_size);
  int index = i/WORDSIZE;
  assert(index < m_length);
  int remainder = i-WORDSIZE*index;
  WORD tmp = m_bits[index] & trueMasks[remainder];
  return tmp > 0;
}

inline int BitSet::size() const
{
  return m_size;
}

///
/* Iterate over bits in a BitSet.  return true if bit is 1

   example: 35 bits in bitset, WORDSIZE=8:

   currently at the 22nd bit: (bit # =21)

   |-m_index = 2---|
   +-------+-------+-------+-------+-------+
   1111111110001111111100011111111100011111
                   ^    ^   
       m_remainder |----| = 6      ^  ^
                                   |--| m_partialBits = 3

  returns: false for operator()
.
*/   
class BitSetIterator
{
public:
  ///
  BitSetIterator();

  ///
  BitSetIterator(const BitSet& bitset);

  // copy and assign should be fine

  ///
  bool operator()() const;

  ///
  bool ok() const;

  ///
  void operator++();

  ///
  void begin();

  ///
  void end();

private:
  int m_index;
  int m_remainder;
  int m_length;

  int m_partialBits;
  const BitSet* m_bitset;
  static BitSet emptyBitSet;
};

inline
BitSetIterator::BitSetIterator()
  : m_index(0), m_remainder(0), m_length(0), m_partialBits(0), m_bitset(&emptyBitSet)
{;}

inline
BitSetIterator::BitSetIterator(const BitSet& a_bitset)
  :m_index(0), m_remainder(0), m_length(a_bitset.m_length - 1),
   m_partialBits(a_bitset.m_size - WORDSIZE*(a_bitset.m_length - 1)),
   m_bitset(&a_bitset)
{
  if(m_partialBits == WORDSIZE)
    { 
      m_partialBits = 0;
      m_length++;
    }
}

inline
bool BitSetIterator::operator()() const
{
  return (m_bitset->m_bits[m_index] & BitSet::trueMasks[m_remainder] ) > 0;
}

inline
bool BitSetIterator::ok() const
{
  if(m_index < m_length) return true;
  if(m_remainder < m_partialBits) return true;
  return false;
}

inline
void BitSetIterator::operator++()
{
  ++m_remainder;
  if(m_remainder == WORDSIZE)
    {
      m_remainder = 0;
      ++m_index;
    }
}
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//


BitSet BitSetIterator::emptyBitSet = BitSet();

WORD BitSet::trueMasks[WORDSIZE];

int BitSet::initialize()
{
  trueMasks[WORDSIZE-1] = 1;
  for(int i=WORDSIZE-2; i>=0 ; --i)
    {
      trueMasks[i] = trueMasks[i+1] << 1;
    }
  return 0;
}

BitSet::BitSet()
  :
  m_bits(NULL),
  m_size(0),
  m_length(0)
{}

long int BitSet::bytes = BitSet::initialize();
long int BitSet::peak = 0;

BitSet::BitSet(int bits, bool init)
  :
  m_bits(NULL),
  m_size(bits),
  m_length(0)
{
  //  static int initialized = initialize();
  assert(bits >= 0);
  if(bits==0) return;

  m_length = bits/WORDSIZE;
  if(m_length*WORDSIZE != bits) m_length +=1;

  m_bits = (WORD*)malloc(m_length*sizeof(WORD));
  bytes+= m_length*sizeof(WORD);
  if(bytes > peak) peak = bytes;

  if(m_bits == NULL)
    {
      MayDay::Error("Memory Error in BitSet::BitSet(int bits, bool init)");
    }
  if(init)
    {
      memset(m_bits, ~0, m_length*sizeof(WORD));
    }
  else
    {
      memset(m_bits, 0, m_length*sizeof(WORD));
    }
}

void BitSet::define(int bits, bool init)
{
  if(m_bits != NULL){
    free(m_bits);
    bytes-=m_length*sizeof(WORD);
  }
  m_size = bits;
  assert(bits >= 0);
  m_length = bits/WORDSIZE;
  if(bits == 0) { m_bits = NULL; return;}
  if(m_length*WORDSIZE != bits) m_length +=1;
  m_bits = (WORD*)malloc(m_length*sizeof(WORD));
  bytes+= m_length*sizeof(WORD);
  if(bytes > peak) peak = bytes;

  if(m_bits == NULL)
    {
      MayDay::Error("Memory Error in BitSet::BitSet(int bits, bool init)");
    }
  if(init)
    {
      memset(m_bits, ~0, m_length*sizeof(WORD));
    }
  else
    {
      memset(m_bits, 0, m_length*sizeof(WORD));
    }
}

BitSet::BitSet(const BitSet& rhs)
  :
  m_bits(NULL),
  m_size(rhs.m_size),
  m_length(rhs.m_length)
{
  if(m_length == 0) return;
  m_bits = (WORD*)malloc(m_length*sizeof(WORD));
  bytes+= m_length*sizeof(WORD);
  if(bytes > peak) peak = bytes;
  if(m_bits == NULL)
    {
      MayDay::Error("Memory Error in BitSet::BitSet(const BitSet& rhs)");
    }

  memcpy(m_bits, rhs.m_bits, m_length*sizeof(WORD));
}

bool BitSet::isEmpty() const
{
  if(m_length == 0) return true;
  for(int i=0; i<m_length-1; ++i)
    {
      if(m_bits[i] > 0) return false;
    }
  for(int i=(m_length-1)*WORDSIZE; i<m_size; ++i)
    {
      if (this->operator[](i)) return false;
    }
  return true;
}

bool BitSet::isFull() const
{
  WORD g=0;
  g=~g;
  for(int i=0; i<m_length-1; ++i)
    {
      if(m_bits[i] != g) return false;
    }
  for(int i=(m_length-1)*WORDSIZE; i<m_size; ++i)
    {
      if (!this->operator[](i)) return false;
    }
  return true;
}

BitSet& BitSet::operator=(const BitSet& rhs)
{
  if(m_bits != NULL) 
    {
      free(m_bits);
      bytes -= m_length*sizeof(WORD);
      m_bits = NULL;
    }
  m_length=rhs.m_length;
  m_size =rhs.m_size;

  if(rhs.m_bits != NULL)
    {
      m_bits = (WORD*)malloc(m_length*sizeof(WORD));
      bytes+= m_length*sizeof(WORD);
      if(bytes > peak) peak = bytes;
    }
  else
    {
      m_bits = NULL;
      return *this;
    }
  if(m_bits == NULL)
    {
      MayDay::Error("Memory Error in BitSet::operator=(const BitSet& rhs)");
    }

  memcpy(m_bits, rhs.m_bits, m_length*sizeof(WORD));
  return *this;
}

BitSet::~BitSet()
{
  if(m_bits != NULL) 
    {
      free(m_bits);
      bytes-=m_length*sizeof(WORD);
    }
}

void BitSet::setFalse(int i)
{
  // WORD* tm = trueMasks;
  assert(i>=0);
  assert(i<m_size);
  int index = i/WORDSIZE;
  assert(index < m_length);
  int remainder = i-WORDSIZE*index;
  WORD* word = m_bits+index;

  *word = *word & ~(trueMasks[remainder]);
}

void BitSet::setTrue(int i)
{
  assert(i>=0);
  assert(i<m_size);
  int index = i/WORDSIZE;
  assert(index < m_length);
  int remainder = i-WORDSIZE*index;
  WORD* word = m_bits+index;

  *word = *word | trueMasks[remainder];
}


void BitSetIterator::end()
{
  m_index = m_length;
  m_remainder = m_partialBits;
}
void BitSetIterator::begin()
{
  m_index = 0;
  m_remainder = 0;
}
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//


class ProblemDomain;
class DenseIntVectSetIterator;
///
/**  Performance-oriented class that reproduces most of the
  functionality of IntVectSet.  The one function that is
  not supported is union (|= ).  This object stores a dense
  representation of an IntVectSet over the box of definition*/
class DenseIntVectSet
{
public:
  ///
  DenseIntVectSet(){;}

  ///
  /** you can either have the domain begin all true, or all false */
  DenseIntVectSet(const Box& a_domain, bool init = true);
  // copy, and operator= should be fine

  /// 
  /**
     make this set the complement of itself and the input
  */
  DenseIntVectSet& operator-=(const Box& box);

  /// 
  /**
     make this set the complement of itself and the input
  */
  DenseIntVectSet& operator-=(const IntVect& intvect);

  /// 
  /**
     make this set the complement of itself and the input
  */
  DenseIntVectSet& operator-=(const DenseIntVectSet& ivs);

  /// 
  /**
     make this set the intersection of itself and the input
  */
  DenseIntVectSet& operator&=(const Box& box);
 
  /// 
  /**
     make this set the intersection of itself and the input
  */
  DenseIntVectSet& operator&=(const ProblemDomain& domain);

  /// 
  /**
     make this set the union of itself and the input.  intvect MUST
         be within the domain of of the DenseIntVectSet or this will throw
         an error
  */
  DenseIntVectSet& operator|=(const IntVect& intvect);

  /// 
  /**
     make this set the union of itself and the input.  b MUST
         be within the domain of of the DenseIntVectSet or this will throw
         an error
  */
  DenseIntVectSet& operator|=(const Box& b);

  /// 
  /**
     make this set the union of itself and the input. resulting
         DenseIntVectSet has domain of the minBox holding both oeprator
         and operand
  */
  DenseIntVectSet& operator|=(const DenseIntVectSet& b);

  /// 
  /**
     make this set the intersection of itself and the input
  */
  DenseIntVectSet& operator&=(const DenseIntVectSet& ivs);

  ///
  /**
     Shift every cell in the set by the input.
  */
  void shift(const IntVect& iv);

  //
  /* slower, constant time pointwise access.
     You should not build a BoxIterator, then access this
     method if you intend to iterate over the objects true
     or false terms.  you should build a DenseIntVectSetIterator */
  bool operator[](const IntVect& index) const;

  ///
  /** returns 'true' if the entire set of points specified
      by the box 'b' is a member of the IntvectSet */
  bool contains(const Box& box) const;

  ///
  /**
     coarsen the set by the input
  */
  void coarsen(int iref);

  ///
  /**
     refine the set by the input
  */
  void refine(int iref);

  ///
  /**
     grow the set by the input
  */
  void grow(int igrow);

  ///
  /**
     grow the set by the input
  */
  void grow(int idir, int igrow);

  ///
  /** Chop the DenseIntVectSet  at the chop_pnt in the dir direction \\
      Returns one DenseIntVectSet and  modifies the object DenseIntVectSet. \\
      The union of the two is the original IntVectSet. \\
      The modified DenseIntVectSet is the low end, the returned DenseIntvectSet \\
      is the high end. */
  DenseIntVectSet chop(int dir, int chop_pnt);

  inline const Box& box() const;

  ///
  /** see IntVectSet
   */
  void nestingRegion(int a_radius, const Box& a_domain);

  ///
  /** see IntVectSet
   */
  void nestingRegion(int a_radius, const ProblemDomain& a_domain);

  ///
  /**
     return true if the set has no points
  */
  bool isEmpty() const; //cheaper than numPts by wide margin

  ///
  /**
     return true if the set consists of a set of a full box
  */
  bool isFull() const;  // cheaper than numPts, bu significant margin

  ///
  /**
     return the number of points in the set
  */
  int  numPts() const;

  ///
  /**
         turn DenseIntVectSet into a Vector of Boxes
  */
  Vector<Box> createBoxes() const;
  ///
  /**
     set object to its minimum representation.  changes data, but
     is a logically const operation
  */
  void compact() const ; 

  void recalcMinBox() const;

  const Box& mBox() const {return m_minBox;}

private:

  void grow(const IntVect& iv);
  // version without checking intersect or empty
  DenseIntVectSet& intersect(const DenseIntVectSet& rhs);
  Box m_domain;
  BitSet m_bits;
  friend class DenseIntVectSetIterator;
  Box m_minBox;
};

///
/** DenseIntVectSetIterator iterates over all the 'true'
    members of a DenseIntVectSet set.*/
class DenseIntVectSetIterator
{
public:

  ///
  DenseIntVectSetIterator();

  ///
  DenseIntVectSetIterator(const DenseIntVectSet& ivs);

  ///
  void define(const DenseIntVectSet& ivs);
  //default null, assignment, copy and destructor should work fine.

  ///
  const IntVect& operator()() const ;

  ///
  bool ok() const;

  ///
  void operator++();

  ///
  void begin();

  ///
  void end();

  static DenseIntVectSet    emptyDenseIntVectSet;

private:
  BitSetIterator m_iterator;
  IntVect        m_current;
  const DenseIntVectSet*    m_ivsPtr;
  int isize, ijsize, bigi, bigj;

  void nextIntVect();
  void nextIntVect(int skip);

};

//===================================================================

// inline functions


inline
const Box& DenseIntVectSet::box() const { return m_domain;}



inline
DenseIntVectSet& DenseIntVectSet::operator-=(const IntVect& intvect)
{
  if(m_domain.contains(intvect))
    m_bits.setFalse(m_domain.index(intvect));
  return *this;
}

inline void DenseIntVectSet::shift(const IntVect& iv)
{
  m_domain.shift(iv);
  m_minBox.shift(iv);
}

inline
DenseIntVectSetIterator::DenseIntVectSetIterator()
  : m_ivsPtr(&emptyDenseIntVectSet)
{
  ;
}

inline
void DenseIntVectSetIterator::define(const DenseIntVectSet& ivs)
{
  m_ivsPtr = &ivs;
  begin();
}

inline
DenseIntVectSetIterator::DenseIntVectSetIterator(const DenseIntVectSet& ivs)
  : m_ivsPtr(&ivs)
{
  begin();
}

inline
const IntVect& DenseIntVectSetIterator::operator()() const
{
  return m_current;
}

inline
bool DenseIntVectSetIterator::ok() const
{
  return m_iterator.ok();
}


inline 
void DenseIntVectSetIterator::end()
{
  m_iterator.end();
}

//  inline 
//  void DenseIntVectSetIterator::operator++()
//  {
//    ++m_iterator;
//    nextIntVect();
//    while(!m_iterator() && m_iterator.ok())
//      {
//        ++m_iterator; // next bit please
//        nextIntVect();
//      }
//  }
inline 
void DenseIntVectSetIterator::operator++()
{
  ++m_iterator;
  int i=1;
  while(m_iterator.ok() && !m_iterator())
    {
      ++m_iterator; // next bit please
      ++i;
    }
  nextIntVect(i);
}
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//


DenseIntVectSet::DenseIntVectSet(const Box& a_domain, bool init)
  :
  m_domain(a_domain),
  m_bits(a_domain.numPts(), init),
  m_minBox(a_domain)
{}

bool DenseIntVectSet::operator[](const IntVect& index) const
{
  if(!m_domain.contains(index)) return false;
  return m_bits[m_domain.index(index)];
}

DenseIntVectSet& DenseIntVectSet::operator-=(const Box& b)
{
  if(isEmpty()) return *this;
  if(!m_domain.intersects(b)) return *this;
  BoxIterator it(b & m_domain);
  for(it.begin(); it.ok(); ++it) this->operator-=(it());
  return *this;
}

DenseIntVectSet& DenseIntVectSet::operator|=(const DenseIntVectSet& d)
{
  if(m_domain.contains(d.m_domain))
        {
          BoxIterator bit(d.m_domain);
          int i=0;
          for(bit.begin(); bit.ok(); ++bit, ++i){
                if(d.m_bits[i]) m_bits.setTrue(m_domain.index(bit()));
          }
        }
  else if(d.m_domain.contains(m_domain))
        {
          DenseIntVectSet newSet = d;
          BoxIterator bit(m_domain);
          int i=0;
          for(bit.begin(); bit.ok(); ++bit, ++i){
                if(m_bits[i]) newSet.m_bits.setTrue(newSet.m_domain.index(bit()));
          }
          *this = newSet;
        }
  else 
        {
          Box newDomain = minBox(m_domain, d.m_domain);
          DenseIntVectSet newSet(newDomain, false);
          newSet |= *this;
          newSet |= d;
          *this = newSet;
        }
  return *this;
}

DenseIntVectSet& DenseIntVectSet::operator|=(const Box& b)
{
  if(!m_domain.contains(b)){
        MayDay::Error("Box union with DenseIntVectSet outside m_domain");
  }
  BoxIterator bit(b);
  for(bit.begin(); bit.ok(); ++bit)
        {
          m_bits.setTrue(m_domain.index(bit()));
        }
  return *this;
}

DenseIntVectSet& DenseIntVectSet::operator|=(const IntVect& intvect)
{
  if(!m_domain.contains(intvect)){
        MayDay::Error("union with DenseIntVectSet outside m_domain");
  }
  m_bits.setTrue(m_domain.index(intvect));
  return *this;
}

DenseIntVectSet& DenseIntVectSet::operator&=(const ProblemDomain& domain)
{
  DenseIntVectSet tmp(*this);
  tmp -= domain.domainBox();
  *this &= domain.domainBox();
  DenseIntVectSetIterator it(tmp);
  for(it.begin(); it.ok(); ++it)
	{
	  IntVect iv = it();
	  if(domain.image(iv)) *this |= iv;
	}
  return *this;
}


DenseIntVectSet& DenseIntVectSet::operator&=(const Box& b)
{
  if(b.contains(m_domain) || b == m_domain) return *this;
  if(isEmpty()) return *this;
  if(!m_domain.intersects(b))
    {
      *this = DenseIntVectSet();
      return *this;
    }
  if(m_domain.numPts() < b.numPts()) 
	{
	  BoxIterator bit(m_domain);
	  int i=0;
	  for(bit.begin(); bit.ok();++i, ++bit)
		{
		  if(b.contains(bit())) { }// do nothing;
		  else m_bits.setFalse(i);
		}
	}
  else
	{
	  Box btmp(b);
	  btmp&=m_domain;
	  DenseIntVectSet tmp = DenseIntVectSet(btmp, true);
	  BoxIterator bit(btmp); int i=0;
	  for(bit.begin(); bit.ok(); ++i, ++bit)
		{
		  if(!this->operator[](bit())) tmp.m_bits.setFalse(i);
		}
	  *this = tmp;
	}
  return *this;
	  
}

Vector<Box> DenseIntVectSet::createBoxes() const
{
  Vector<Box> boxes;
  DenseIntVectSetIterator it(*this);
  for(it.begin(); it.ok(); ++it){
        boxes.push_back(Box(it(), it()));
  }
  return boxes;
}

void DenseIntVectSet::recalcMinBox() const
{
  if(isEmpty()){
	(Box&)m_minBox = Box();
	return;
  }
  DenseIntVectSetIterator it(*this);
  it.begin();
  (Box&)m_minBox = Box(it(), it());
  ++it;
  for(; it.ok(); ++it){
	Box m(it(), it());
	((Box&)m_minBox).minBox(m);
  }
}

DenseIntVectSet DenseIntVectSet::chop(int dir, int chop_pnt)
{
  if(m_minBox.smallEnd(dir) >= chop_pnt)
    {
      DenseIntVectSet rtn(*this);
      *this = DenseIntVectSet();
      return rtn;
    }
  if(m_minBox.bigEnd(dir) < chop_pnt)
    {
      return DenseIntVectSet();
    }
  Box chop = m_domain;
  Box chopThis = chop.chop(dir, chop_pnt);
  DenseIntVectSet left(chop);
  DenseIntVectSet right(chopThis);
  left.intersect(*this);
  right.intersect(*this);
  *this = right;
  return left;
  
}


bool DenseIntVectSet::contains(const Box& box) const
{
  if(!m_minBox.contains(box)) return false;
  for(BoxIterator bit(box); bit.ok(); ++bit)
    if(!this->operator[](bit())) return false;
  return true;
}

DenseIntVectSet& DenseIntVectSet::operator&=(const DenseIntVectSet& ivs)
{
  if(isEmpty()) return *this;
  if(ivs.isEmpty())
    {
      *this = DenseIntVectSet();
      return *this;
    }
  if(!m_domain.intersectsNotEmpty(ivs.m_domain))
    {
      *this = DenseIntVectSet();
      return *this;
    }
  return intersect(ivs);
}

DenseIntVectSet& DenseIntVectSet::intersect(const DenseIntVectSet& ivs)
{
  BoxIterator bit(m_domain);
  int i=0;
  for(bit.begin(); bit.ok();++i, ++bit)
    {
      if(m_bits[i])
        {
          if(ivs[bit()]) { } // do nothing
          else
            m_bits.setFalse(i);
        }
    }
  return *this;
}

void DenseIntVectSet::coarsen(int iref)
{
  if(iref == 1) return;
  assert(iref >= 1);
  int refinements = iref/2;
  assert(refinements*2 == iref); // check iref for power of 2

  Box newDomain(m_domain);
  newDomain.coarsen(iref);
  DenseIntVectSet newSet(newDomain, false);
  BoxIterator bit(m_domain);
  int count=0;
  for(bit.begin(); bit.ok(); ++bit, ++count)
    {
      if(m_bits[count])
        {
          IntVect iv(bit());
          iv.coarsen(iref);
          long index = newDomain.index(iv);
          newSet.m_bits.setTrue(index);
        }
    }

  *this = newSet;
}


void DenseIntVectSet::refine(int iref)
{
  if(iref == 1) return;
  if(isEmpty()) return;
  assert(iref >= 1);
  int refinements = iref/2;
  assert(refinements*2 == iref); // check iref for power of 2 
  Box newDomain(m_domain);
  newDomain.refine(iref);
  DenseIntVectSet newSet(newDomain, false);
  IntVect iv;
  BoxIterator bit(newDomain);
  int count=0;
  for(bit.begin(); bit.ok(); ++bit, ++count)
    {
      iv = bit();
      iv.coarsen(iref);
      if(this->operator[](iv))
        {
          newSet.m_bits.setTrue(count);
        }
    }
  *this = newSet;
}

void DenseIntVectSet::nestingRegion(int radius, const Box& domain)
{
  assert(radius >= 0);
  if (radius == 0) return;

  DenseIntVectSet tmp(*this);

  {
    IntVect lo = m_domain.smallEnd();
    IntVect hi = m_domain.bigEnd();
    for(int i=0; i<SpaceDim; ++i)
      {
        if(lo[i] != domain.smallEnd()[i]) lo[i] += radius;
        if(hi[i] != domain.bigEnd()[i])   hi[i] -= radius;
      }
    Box shrink(lo, hi);
    *this &= shrink;
  }


  Box clobberBox(IntVect::Zero, IntVect::Zero);
  IntVect center(IntVect::Zero);
  clobberBox.grow(radius);
  BoxIterator bit(m_domain);
  int i=0;
  for(bit.begin(); bit.ok();++i, ++bit)
    {
      if(!(tmp.m_bits[i])) // was there a zero at this bit ?
        {
          if(domain.contains(bit()))
            {
              clobberBox.shift(bit()-center);
              *this -= clobberBox;
              center=bit();
            }
        }
    }
}

void DenseIntVectSet::nestingRegion(int radius, const ProblemDomain& a_domain)
{
  assert(radius >= 0);
  if (radius == 0) return;

  DenseIntVectSet tmp;
  Box region = a_domain.domainBox();
  if(!a_domain.isPeriodic()){
        tmp = *this;
  }
  else {
        D_TERM(if(a_domain.isPeriodic(0)) region.grow(0, radius);,
                   if(a_domain.isPeriodic(1)) region.grow(1, radius);,
                   if(a_domain.isPeriodic(2)) region.grow(2, radius);)
        tmp = DenseIntVectSet(region, false);
        tmp |= *this;
        for(int i=0; i<SpaceDim; ++i)
          {
                if(a_domain.isPeriodic(i))
                  {
                        int size = a_domain.domainBox().size(i);
                        Box hiRegion = adjCellHi(a_domain.domainBox(), i, radius);
                        for(BoxIterator bit(hiRegion); bit.ok(); ++bit){
                          IntVect image = bit();
                          image[i] -= size;
                          if(this->operator[](image)){ 
                                tmp |= bit();
                          }
                          image[i] += (size - radius);
                          if(this->operator[](image)){
                                image[i] -= size;
                                tmp |= image;
                          }
                        }
                  }
          }
  }

  {
    IntVect lo = m_domain.smallEnd();
    IntVect hi = m_domain.bigEnd();
    for(int i=0; i<SpaceDim; ++i)
      {
                
                if(lo[i] != region.smallEnd()[i]) lo[i] += radius;
                if(hi[i] != region.bigEnd()[i])   hi[i] -= radius;
                
      }
    Box shrink(lo, hi);
    *this &= shrink;
  }


  Box clobberBox(IntVect::Zero, IntVect::Zero);
  IntVect center(IntVect::Zero);
  clobberBox.grow(radius);
  BoxIterator bit(region);
  int i=0;
  for(bit.begin(); bit.ok();++i, ++bit)
    {
      if(!(tmp.m_bits[i])) // was there a zero at this bit ?
        {
                  clobberBox.shift(bit()-center);
                  *this -= clobberBox;
                  center=bit();        
        }
    }
}

void DenseIntVectSet::grow(int igrow)
{
  if(igrow >= 1)
    {
      IntVect range(IntVect::Unit);
      range*=igrow;
      grow(range);
    }
  else if(igrow < 0)
    {
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          grow(idir, igrow);
        }
    }
  else
    {
      //nothing to do if igrow == 0
      return;
    }

}

void DenseIntVectSet::grow(int idir, int igrow)
{
  assert(idir >= 0);
  assert(idir < SpaceDim);
  if(igrow >= 1)
    {
      IntVect range(IntVect::Zero);
      range[idir] = igrow;
      grow(range);
    }
  else if(igrow < 0)
    {
      IntVect shiftvec= igrow*BASISV(idir);
      DenseIntVectSet ivsShiftPlus = *this;
      DenseIntVectSet ivsShiftMinu = *this;
      ivsShiftPlus.shift(shiftvec);
      ivsShiftMinu.shift(-shiftvec);
      *this &= ivsShiftPlus;
      *this &= ivsShiftMinu;
    }
  else
    {
      //nothing to do if igrow == 0
      return;
    }

}

void DenseIntVectSet::grow(const IntVect& range)
{
  Box newDomain(m_domain);
  newDomain.grow(range);
  DenseIntVectSet newSet(newDomain, false);

  int index = 0;

#if CH_SPACEDIM == 2

  for(int j = 0; j<m_domain.size(1); ++j)
    {
      for(int i=0; i<m_domain.size(0); ++i, ++index)
        {
          if(m_bits[index])
            {
              for(int jj=0; jj<2*range[1] +1 ; ++jj)
                {
                  int indexNew = i + newDomain.size(0)*(j+jj);
                  for(int ii = 0; ii<2*range[0] + 1; ++ii, ++indexNew)
                    newSet.m_bits.setTrue(indexNew);
                }
            }
        }
    }
#endif

#if CH_SPACEDIM == 3
  for(int k=0; k<m_domain.size(2); ++k)
    for(int j = 0; j<m_domain.size(1); ++j)
      for(int i = 0; i<m_domain.size(0); ++i, ++index)
        {
          if(m_bits[index])
            {
              for(int kk=0; kk<2*range[2]+1; ++kk)
                for(int jj=0; jj<2*range[1] +1 ; ++jj)
                  {
                    int indexNew = i+
                      newDomain.size(0)*(j+jj) +
                      newDomain.size(1)*newDomain.size(0)*(k+kk);
                    for(int ii = 0; ii<2*range[0] + 1; ++ii, ++indexNew)
                      newSet.m_bits.setTrue(indexNew);
                  }
            }
        }

#endif

  *this = newSet;
}

DenseIntVectSet& DenseIntVectSet::operator-=(const DenseIntVectSet& ivs)
{
  BoxIterator bit(m_domain);
  int i=0;
  for(bit.begin(); bit.ok();++i, ++bit)
    {
      if(m_bits[i])
        {
          if(ivs[bit()])
            m_bits.setFalse(i);
        }
    }
  return *this;
}

bool DenseIntVectSet::isEmpty() const
{
  return m_bits.isEmpty();
}
bool DenseIntVectSet::isFull() const
{
  return m_bits.isFull();
}

int DenseIntVectSet::numPts() const
{
  int n = m_domain.numPts();
  int c=0;
  for(int i=0; i<n; ++i)
    if(m_bits[i]) ++c;
  return c;
}

void DenseIntVectSet::compact() const
{
  DenseIntVectSet* nthis = (DenseIntVectSet*)this;
  if(isEmpty()){
    *nthis = DenseIntVectSetIterator::emptyDenseIntVectSet;
    return;
  } else if (isFull()){
    return;
  }
  DenseIntVectSetIterator it(*this);
  Box nDomain(it(), it());
  IntVect& lo = (IntVect&)(nDomain.smallEnd());
  IntVect& hi = (IntVect&)(nDomain.bigEnd());
  for(++it; it.ok(); ++it)
    {
      lo.min(it());
      hi.max(it());
    }
  nDomain.computeBoxLenNotEmpty();
  DenseIntVectSet nset(nDomain);
  nset.intersect(*this);
  *nthis = nset;
}

DenseIntVectSet  DenseIntVectSetIterator::emptyDenseIntVectSet;


void DenseIntVectSetIterator::nextIntVect()
{
  const IntVect& big = m_ivsPtr->m_domain.bigEnd();
  const IntVect& small = m_ivsPtr->m_domain.smallEnd();
  ++(m_current[0]);
#if CH_SPACEDIM >= 2
  if(m_current[0] > big[0])
    {
      m_current[0] = small[0];
      ++m_current[1];
    }
#endif
#if CH_SPACEDIM == 3
  if(m_current[1] > big[1])
    {
      m_current[1] = small[1];
      ++m_current[2];
    }
#endif
}

void DenseIntVectSetIterator::nextIntVect(int skip)
{
  if(skip == 1) {nextIntVect(); return;}

#if CH_SPACEDIM == 2

  int j_jump = skip / isize;
  m_current[1] += j_jump;
  m_current[0] += skip - j_jump*isize;
  if(m_current[0] > bigi)
    {
      ++(m_current[1]);
      m_current[0] -= isize;
    }
#endif
#if CH_SPACEDIM == 3

  int jump = skip / ijsize;
  m_current[2] += jump;
  skip -=   jump*ijsize;
  jump  =   skip / isize;
  m_current[1] += jump;
  skip -=   jump*isize;
  m_current[0] += skip;

  if(m_current[0] > bigi)
    {
      ++(m_current[1]);
      m_current[0] -= isize;
    }
  if(m_current[1] > bigj)
    {
      ++(m_current[2]);
      m_current[1] -= ijsize/isize;
    }

  // hack for now to make sure code actually works....
  //for(;skip>0; --skip) nextIntVect();

#endif

}

void DenseIntVectSetIterator::begin()
{
  if(m_ivsPtr->m_bits.isEmpty())
    {
      m_iterator.end();
      return;
    }
  isize =  m_ivsPtr->m_domain.size(0);
  bigi  =   m_ivsPtr->m_domain.bigEnd()[0];
#if CH_SPACEDIM > 2
  ijsize = isize * m_ivsPtr->m_domain.size(1);
  bigj  =   m_ivsPtr->m_domain.bigEnd()[1];
#endif
  m_iterator = BitSetIterator(m_ivsPtr->m_bits);
  m_current = m_ivsPtr->m_domain.smallEnd();
  int i=0;
  while(m_iterator.ok() && !m_iterator())
    {
      ++i;
      ++m_iterator;
    }
  nextIntVect(i);
}

/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef CH_TUPLE_H
#define CH_TUPLE_H

//
// $Id: CFile.cpp,v 1.6 2003/04/21 22:08:00 bvs Exp $
//

using namespace std;


//
//@Man:
//@Memo: Ordered Tuples for Types T
/*@Doc:

  This class represents ordered tuples of some user-specified concrete
  type T for N > 0. The type T must have a default constructor.  If the
  non-default constructor, copy constructor, or copy assignment operator
  are used, T must also have a copy constructor.
*/

template <class T, size_t N>
class Tuple
{
public:
 
    /*@ManDoc: The default constructor.  For user-defined types T, the
               default constructor for T will be run on each of the N
               objects in the Tuple.  For builtin (intrinsic) types,
               the values in the Tuple will be garbage.
    */
    Tuple ();

    /*@ManDoc: Constructs a Tuple, initializing the elements in the Tuple
               with the corresponding elements in the vector v.  This assumes
               that v contains at least N elements of type T -- an assumption
               that is NOT checked.  For user-defined types, T must have a
               well-defined and accessible copy constructor.
    */
  //explicit Tuple (const T* v);
    //
    // The copy constructor.
    //
    Tuple (const Tuple& rhs);
    //
    // The copy assignment operator.
    //
    Tuple& operator= (const Tuple& rhs);
 
    /*@ManDoc: Returns a modifiable lvalue reference to the i'th
               element in the Tuple, counting from zero.  Performs
               range checking when the library is compiled in debug
               mode.  */
    T& operator[] (int i);
 
    /*@ManDoc: Returns a constant reference to the i'th element in the Tuple,
               counting from zero.  Performs range checking when the library
               is compiled in debug mode.
    */
    const T& operator[] (int i) const;
 
    /*@ManDoc: Returns the address of the underlying vector of T
               representation.  This should ONLY be used when interfacing
               to Fortran as it breaks the encapsulation of the class.
    */
    operator const T* () const;

protected:
    //
    // The underlying vector of T representing the Tuple.
    //
    T vect[N];
};

//
// Inlines.
//

template <class T, size_t N>
inline
Tuple<T,N>::Tuple()
{}

template <class T, size_t N>
inline
T&
Tuple<T,N>::operator[] (int i)
{
    assert(0 <= i && i < (int)N);
    return vect[i];
}

template <class T, size_t N>
inline
const T&
Tuple<T,N>::operator[] (int i) const
{
    assert(0 <= i && i < (int)N);
    return vect[i];
}

template <class T, size_t N>
inline
Tuple<T,N>::operator const T* () const
{
    return &vect[0];
}


//template <class T, size_t N>
//Tuple<T,N>::Tuple (const T* v)
//{
//    assert(v != 0);
//    for (size_t i = 0; i < N; ++i)
//        vect[i] = v[i];
//}

template <class T, size_t N>
Tuple<T,N>::Tuple (const Tuple<T,N>& rhs)
{
    for (size_t i = 0; i < N; ++i)
        vect[i] = rhs.vect[i];
}

template <class T, size_t N>
Tuple<T,N>&
Tuple<T,N>::operator= (const Tuple<T,N>& rhs)
{
    for (size_t i = 0; i < N; ++i)
        vect[i] = rhs.vect[i];
    return *this;
}

#endif /*CH_TUPLE_H*/
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef TREEINTVECTSET_H
#define TREEINTVECTSET_H


class IntVectSet;
class ProblemDomain;

// xlC doesn't like the static const int inside this class.
// so istead of a nice we have a macro: TIVS_NODESIZE
// Currently, this only affects two files:  this one and the .cpp
#if (CH_SPACEDIM==2)
#define TIVS_NODESIZE 4
#elif (CH_SPACEDIM==3)
#define TIVS_NODESIZE 8
#else
//error -- only tested for dimensions 2 and 3 (ndk)
// just cause a compile time bug -- not sure what the best way is.
error_only_tested_for_2D_and_3D
#endif

///
/**
   Set calculus class based on trees.
 */
class TreeIntVectSet
{
public:
  ///
  inline TreeIntVectSet();

  ///
  TreeIntVectSet(const Box&);

  ///
  TreeIntVectSet(const TreeIntVectSet& a_sivs);

  ///
  ~TreeIntVectSet();

  ///
  void define(const Box&);

  ///
  void define(const TreeIntVectSet& a_sivs);

  ///
  TreeIntVectSet& operator=(const TreeIntVectSet& a_sivs);

  ///or
  TreeIntVectSet& operator|=(const TreeIntVectSet& a_sivs);

  ///
  TreeIntVectSet& operator|=(const IntVect& a_iv);

  ///
  TreeIntVectSet& operator|=(const Box& a_box);

  ///and
  TreeIntVectSet& operator&=(const TreeIntVectSet& s_sivs);

  ///
  TreeIntVectSet& operator&=(const Box& a_box);

  ///
  TreeIntVectSet& operator&=(const ProblemDomain& a_domain);

  ///not
  TreeIntVectSet& operator-=(const TreeIntVectSet& a_sivs);

  ///
  TreeIntVectSet& operator-=(const IntVect& a_iv);

  ///
  TreeIntVectSet& operator-=(const Box& a_box);

  ///
  /**
     Returns Vector<Box> representation of this IntVectSet.
  */
  Vector<Box> createBoxes() const;


  void createBoxes(Vector<Box>& boxes, int& size) const;

  ///
  bool contains(const IntVect& iv) const;

  ///
  bool contains(const Box& box) const;

  ///
  /**
     somewhat expensive, but shouldn't be ;-)
  */
  TreeIntVectSet chop(int idir, int chop_pnt); 

  ///
  /**
     expensive
  */
  void grow(int igrow);           

  ///
  /**
     expensive
  */
  void grow(int idir, int igrow);

  ///
  /**
     fast if iref is power of 2
  */
  void refine(int iref = 2); 

  ///
  /**
     //fast if iref is power of 2
  */
  void coarsen(int iref = 2); 

  ///
  /**
     very very fast
  */
  void shift(const IntVect& iv); 

  ///
  void clear();

  ///
  void nestingRegion(int a_radius, const Box& a_domain);

  ///
  void nestingRegion(int a_radius, const ProblemDomain& a_domain);

  ///
  inline const Box& minBox() const;

  ///
  bool isEmpty() const;

  ///
  int numPts() const;

  friend void dumpTree(const TreeIntVectSet* set);

  ///
  void compact() const;  // logically const operation;

  ///
  void recalcMinBox() const;

/*
  int linearSize() const;

  void linearIn(const void* const inBuf);

  void linearOut(void* const a_outBuf) const;
*/

private:

  friend class TreeIntVectSetIterator;

  struct TreeNode {
    TreeNode* nodes;
    TreeNode():nodes(0){;}
  };

  TreeNode m_tree;
  Box m_minBox, m_spanBox;
  int m_depth;


  //===============
  static Pool treeNodePool;

  static void   quadrantBox(const Box& inputBox, int quadrant, Box& outputQuadrant);
  static void clearTree(TreeNode& tree);
  static void expandNode(TreeNode& node);
  void growTree();


  static int  oppositeQuadrant(int index);
  static bool nextIntVect(const Box& box, IntVect& iv);
  static void nextNode(int& currentDepth);
  static void cloneNode(const TreeNode& src, TreeNode& dest);
  // data structures that are needed again and again in the routines
  // that only change their size during grow() operations.  Keep them
  // around so that we don't have to create and destroy them on every function
  // call
  static Vector<int> index, bufferOffset;
  static Vector<TreeNode*> parents;
  static Vector<Box> boxes;

  // xlC isn't working me on this one.
  // so istead of a nice const static integer, we have a MACRO TIVS_NODESIZE
  //const static int  nodeSize = D_TERM(2,*2,*2);
  static TreeNode full; 
  friend struct Flag;
  friend class IntVectSet;
};


class TreeIntVectSetIterator
{
public:
  TreeIntVectSetIterator();
  TreeIntVectSetIterator(const TreeIntVectSet& ivs);
  void define(const TreeIntVectSet& ivs);
  const IntVect& operator()() const ;
  bool ok() const;
  void operator++();
  void begin();
  void end();
  void clear();
private:
  const TreeIntVectSet* m_ivs;
  Vector<const TreeIntVectSet::TreeNode*> nodes;
  Vector<Box>   boxes;
  Vector<int>   index;
  int           m_depth;
  IntVect       m_current;

  void findNextNode();
  void findNext(); // parse tree, starting at [depth, index], set m_current
  //when you find an entry, might run to end of iterator.
};

inline
TreeIntVectSetIterator::TreeIntVectSetIterator():m_ivs(0), m_depth(-1) {;}

inline 
TreeIntVectSetIterator::TreeIntVectSetIterator(const TreeIntVectSet& ivs)
{
  define(ivs);
}

inline
void TreeIntVectSetIterator::clear()
{
  m_depth = -1;
  m_ivs = 0;
}

// the printf below. i just didn't like the stdio.h, so changed it to
// cstdio -- but then realized it prolly was meant to be commented
// out. (ndk)

inline
void TreeIntVectSetIterator::define(const TreeIntVectSet& ivs)
{
  m_ivs = &ivs;
  int max = ivs.index.size();
  //  if(max==10) printf("max=10 \n");
  if(boxes.size() < max)
    {
      boxes.resize(max);
      index.resize(max);
      nodes.resize(max);
    }
  begin();
}

inline
bool TreeIntVectSetIterator::ok() const
{
  return m_depth >= 0;
}

inline
void TreeIntVectSetIterator::end()
{
  m_depth = -1;
}  

inline
TreeIntVectSet& TreeIntVectSet::operator|=(const IntVect& iv)
{
  return *this|=Box(iv, iv);
}

inline
TreeIntVectSet& TreeIntVectSet::operator-=(const IntVect& iv)
{
  return *this-=Box(iv, iv);
}

inline
void TreeIntVectSetIterator::operator++()
{
  findNext();
}

inline 
const IntVect& TreeIntVectSetIterator::operator()() const
{
  return m_current;
}

//=======================================================

inline
TreeIntVectSet::~TreeIntVectSet()
{
  clearTree(m_tree);
}
inline
TreeIntVectSet::TreeIntVectSet(const TreeIntVectSet& a_tivs)
{
  define(a_tivs);
}

inline
TreeIntVectSet::TreeIntVectSet(const Box& a_box)
{
  define(a_box);
}

inline
TreeIntVectSet::TreeIntVectSet()
{
  m_tree.nodes = 0;
  m_depth=1;

}

inline
const Box&
TreeIntVectSet::minBox() const
{
  return m_minBox;
}

inline void TreeIntVectSet::nextNode(int& depth)
{
  index[0] = 0;
  index[depth]++;
  while(index[depth] == TIVS_NODESIZE)
    {
      index[depth] = 0;
      depth--;
      index[depth]++;
    }
}

#endif //  TREEINTVECTSET_H
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

//
// These are the inevitable functions that people
// can't avoid using when writing a SPMD program.
// It is a minimal set from which more elaborate
// functionality can be generated.  As always, a
// user is free to utilize the entire MPI programming
// on their own platform.  The functions are 
// assured to work on all machines supported.
//

#ifndef SPMD_H
#define SPMD_H

class IntVectSet;


/// local process ID
/** 
    Returns the ID of the locally running process in the range 0 <=
    procID() < numProc().  This has no relation to the operating system
    pid.  There is always a procID() == 0.  */
int procID();

/// number of parallel processes
/** 
    Returns the number of parallel processes running.
    Always returns at least 1. */
unsigned int numProc();


template <class T>
int linearSize(const T& inputT); 

template <class T>
void linearIn(T& a_outputT, const void* const inBuf); 

template <class T>
void linearOut(void* const a_outBuf, const T& inputT); 

#ifdef MPI
// is this being used anywhere?  xlC was complaining about it... (ndk)
//  ahh.. it's used when MPI=TRUE

///gather from multiple processes
/**
  Gathers a_input of every process onto Vector<T> a_outVec on process
  number a_dest.  It is an error if a_dest < 0 or a_dest >= numProc().
  a_outVec is numProc() long with the value of a_input on every
  process in its elements.\\ 

T must have have the following functions:
  \\

  int linearSize<T>(const T& inputT); \\
  void linearIn<T>(T& a_outputT, const void* const inBuf); \\
  void linearOut<T>(void* a_outBuf, const T& inputT); \\ 

*/
template <class T>
void gather(Vector<T>& a_outVec, const T& a_input, int a_dest);

// this has to be here so that linearSize<IntVectSet>, linearIn<IntVectSet>, 
//  and linearOut<IntVectSet> are defined.
template < >
void gather(Vector<IntVectSet>& a_outVec, 
            const IntVectSet& a_input, 
            int a_dest);
#endif

///broadcast to every process
/**
  Broadcasts a_inAndOut to every process from process number a_src.
  It is an error if a_src < 0 or a_src >= numProc().  \\

  T must have have the following functions: \\

  int linearSize<T>(const T& inputT); \\
  void linearIn<T>(T& a_outputT, const void* const inBuf); \\
  void linearOut<T>(void* a_outBuf, const T& inputT); \\ */
template <class T>
void broadcast(T& a_inAndOut,  int a_src);

//integer specialization of linearSize
template < >
int linearSize(const int& a_input); 

//integer specialization of linearIn
template < >
void linearIn(int& a_outputT, const void* const inBuf); 

//integer specialization of linearOut
template < >
void linearOut(void* const a_outBuf, const int& a_inputT); 

//Real specialization of linearSize
template < >
int linearSize(const Real& a_input); 

//Real specialization of linearIn
template < >
void linearIn(Real& a_outputT, const void* const a_inBuf); 

//Real specialization of linearOut
template < >
void linearOut(void* const a_outBuf, const Real& a_inputT); 

//Box specialization of linearSize
template < >
int linearSize(const Box& a_input); 

//Box specialization of linearIn
template < >
void linearIn(Box& a_outputT, const void* const a_inBuf); 

//Box specialization of linearOut
template < >
void linearOut(void* const a_outBuf, const Box& a_inputT); 

//Vector<int>  specialization
template < > 
int linearSize(const Vector<int>& a_input); 
template < > 
void linearIn(Vector<int>& a_outputT, const void* const inBuf); 
template < > 
void linearOut(void* const a_outBuf, const Vector<int>& a_inputT); 

//Vector<Real>  specialization
template < > 
int linearSize(const Vector<Real>& a_input); 
template < > 
void linearIn(Vector<Real>& a_outputT, const void* const inBuf); 
template < > 
void linearOut(void* const a_outBuf, const Vector<Real>& a_inputT); 

//Vector<Box>  specialization
template < > 
int linearSize(const Vector<Box>& a_input); 
template < > 
void linearIn(Vector<Box>& a_outputT, const void* const inBuf); 
template < > 
void linearOut(void* const a_outBuf, const Vector<Box>& a_inputT); 

//Vector<Vector<Box> > specialization
template < > 
int linearSize(const Vector<Vector<Box> >& a_input); 
template < > 
void linearIn(Vector<Vector<Box> >& a_outputT, const void* const inBuf); 
template < > 
void linearOut(void* const a_outBuf, const Vector<Vector<Box> >& a_inputT); 

// RealVect spcializations of linearization
template < >
int linearSize(const RealVect& vindex);

//VolIndex specialization of linearIn
template < >
void linearIn(RealVect& a_outputT, const void* const inBuf);

//VolIndex specialization of linearOut
template < >
void linearOut(void* const a_outBuf, const RealVect& a_inputT);


//Vector<Vector<int> > specialization
template < > 
int linearSize(const Vector<Vector<int> >& a_input); 
template < > 
void linearIn(Vector<Vector<int> >& a_outputT, const void* const inBuf); 
template < > 
void linearOut(void* const a_outBuf, const Vector<Vector<int> >& a_inputT); 


//Vector<T> specialization of linearSize
template <class T>
int linearListSize(const Vector<T>& a_input); 

//Vector<T> specialization of linearIn
template <class T>
void linearListIn(Vector<T>& a_outputT, const void* const a_inBuf); 

//Vector<T> specialization of linearOut
template <class T>
void linearListOut(void* const a_outBuf, const Vector<T>& a_inputT); 

class SerialTask
{
public:
    enum task { compute=0 };
};

// return id of unique processor for special serial tasks
int
uniqueProc(const SerialTask::task& a_task);

#ifdef MPI
struct Chombo_MPI{ static MPI_Comm comm;};
void setChomboMPIErrorHandler();

#else
// this can be changed for debugging parallel code in serial
extern int num_procs ;
#endif


#endif // SPMD.H

/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _SPMDI_H_
#define _SPMDI_H_


//default implementations for linearization routines.

template <class T>
int linearSize(const T& inputT)
{
  return inputT.linearSize();
}

template <class T>
void linearIn(T& a_outputT, const void* const inBuf)
{
  a_outputT.linearIn(inBuf);
}

template <class T>
void linearOut(void* const a_outBuf, const T& inputT)
{
  inputT.linearOut(a_outBuf);
}

#ifdef MPI

/*****************************/
//gather a_input into a a_outVec
/*****************************/
template <class T>
inline void 
gather(Vector<T>& a_outVec, const T& a_input, int a_dest)
{
  assert (a_dest >= 0);
  assert(a_dest <  numProc());
  //now THIS size lives on THIS processor
  int isize = linearSize(a_input);
    
  //make stuff for linearout
  void* loclBuf = malloc(isize);
  if(loclBuf == NULL)
    MayDay::Error("out of memory in gather 1");
    
  //put linearized T into its proper buffer
  linearOut(loclBuf, a_input);


  int nProcess = numProc(); 
  int sendCount = 1;
  int recdCount = 1;
    
  //need to gather isizes onto processor a_dest
  int* vectSize = NULL;
  int* vectDisp = NULL;
  void* sendBuf = static_cast<void*>(&isize);
  //allocate received buffer
  if(procID() == a_dest)
    {
      vectSize = new int[nProcess];
      vectDisp = new int[nProcess];
    }

  int result1 = MPI_Gather(sendBuf, sendCount, MPI_INT, 
                           vectSize,recdCount, MPI_INT,
                           a_dest,  Chombo_MPI::comm);

  if(result1 != MPI_SUCCESS)
    MayDay::Error("Gather<T> failed in MPI_Gather 1");

  //make memory for gather, linearin
  void* recdBuf = NULL;
  if(procID() == a_dest)
    {
      int itotsize=0;
      for(int iproc = 0; iproc < nProcess; iproc++)
        {
          vectDisp[iproc] = itotsize;
          itotsize += vectSize[iproc];
        }
      recdBuf = malloc(itotsize);
      if(recdBuf == NULL)
        MayDay::Error("out of memory in gather 2");
    }

  //gather data 
  int result2 = MPI_Gatherv(loclBuf, isize, MPI_BYTE, 
                            recdBuf, vectSize, vectDisp, MPI_BYTE,
                            a_dest, Chombo_MPI::comm);
  if(result2 != MPI_SUCCESS)
    MayDay::Error("Gather<T> failed in MPI_Gather 2");


  if(procID() == a_dest)
    {
      //calculate offset into array for current processor
      int ioffset = 0;
      a_outVec.resize(nProcess);
      //need to cast to char* to do pointer arithmetic
      char* arithPtr = (char*)recdBuf;
      for(int iproc = 0; iproc < nProcess; iproc++)
        {
          ioffset = vectDisp[iproc];
          char* thisProcBuf = arithPtr + ioffset;
          linearIn(a_outVec[iproc], thisProcBuf);
        }

      //delete memory for dest-specific arrays
      delete[] vectSize;
      delete[] vectDisp;
      free(recdBuf);
    }

  //delete memory for local buffer
  free(loclBuf);
}


/*****************************/
//broadcast T everywhere
/*****************************/
template <class T>
inline void 
broadcast(T& a_inAndOut,  int a_src)
{
  assert (a_src >= 0);
  assert(a_src <  numProc());
  int isize;  
  if(procID() == a_src)
    isize = linearSize(a_inAndOut);

  MPI_Bcast(&isize, 1, MPI_INT, a_src, Chombo_MPI::comm);

  void* broadBuf = malloc(isize);

  if(broadBuf == NULL)
    MayDay::Error("out of memory in broadcast");

  //take inAndOut from src and put it into broadBuf
  if(procID() == a_src)
      linearOut(broadBuf, a_inAndOut);

  //broadcast broadBuf to all procs
  MPI_Bcast(broadBuf, isize, MPI_BYTE, a_src, Chombo_MPI::comm);
    
  //take broadBuf and put back into inAndOut if not src
  if(procID() != a_src)
      linearIn(a_inAndOut, broadBuf);

  //delete memory for buffer
  free(broadBuf);
}

#else
/*****************************/
//non-mpi version
/*****************************/
template <class T>
inline void 
gather(Vector<T>& a_outVec, const T& a_input, int a_dest)
{
  a_outVec.resize(1);
  a_outVec[0] = a_input;
}
/*****************************/
//non-mpi version
/*****************************/
template <class T>
inline void 
broadcast(T& a_inAndOut,  int a_src)
{
  //nothing to do.  in and out are the same with one proc
}

#endif //the mpi thing

//*************************************
//These should work independent of MPI
//*************************************

//Vector<T> specialization of linearIn
template <class T>
void 
linearListIn(Vector<T>& a_outputT, const void* const a_inBuf)
{
  //first entry is the size of the vector
  const int* const intBuf = (int*)a_inBuf;
  int vecsize = intBuf[0];
  Vector<int> vecOffset(vecsize);
  //next vecsize entries are offsets of data into buffer
  for(int ivec = 0; ivec < vecsize; ivec++)
    {
      vecOffset[ivec] = intBuf[ivec+1];
    }
  //next vecsize entries are the actual data
  //yes I could do this in one loop but that would 
  // either 
  // a) make it less symmetric with linearOut 
  // and/or
  // b) make both of them far less readable
  a_outputT.resize(vecsize);
  const char* const charbuf = (char*)a_inBuf;
  for(int ivec = 0; ivec < vecsize; ivec++)
    {
      const char* const dataLoc = charbuf + vecOffset[ivec];
      linearIn(a_outputT[ivec], dataLoc);
    }
}

//Vector<T> specialization of linearOut
template <class T>
void 
linearListOut(void* const a_outBuf, const Vector<T>& a_input)
{
  //first entry is the size of the vector
  int* const intBuf = (int*)a_outBuf;
  intBuf[0] = a_input.size();
  int vecsize = intBuf[0];
  Vector<int> vecOffset(vecsize);
  //next vecsize entries are offsets of data into buffer
  //next vecsize entries are the actual data
  int ioffset = (vecsize+1)*sizeof(int);
  for(int ivec = 0; ivec < vecsize; ivec++)
    {
      intBuf[ivec+1] = ioffset; 
      vecOffset[ivec] = ioffset; 
      ioffset += linearSize(a_input[ivec]);
    }
  //yes I could do this in one loop but that would 
  // either 
  // a) make it less symmetric with linearIn
  // and/or
  // b) make both of them far less readable
  char* const charBuf = (char*)a_outBuf;
  for(int ivec = 0; ivec < vecsize; ivec++)
    {
      char* const dataLoc = charBuf + vecOffset[ivec];
      linearOut(dataLoc, a_input[ivec]);
    }
}

//Vector<T> specialization of linearSize
template <class T>
int 
linearListSize(const Vector<T>& a_input)
{
  //first entry is the size of the vector (int)
  //next vecsize entries are offsets of data into buffer (int)
  //next vecsize entries are the actual data
  int itotsize = (a_input.size() + 1)*sizeof(int);
  for(int ivec = 0; ivec < a_input.size(); ivec++)
    {
      itotsize += linearSize(a_input[ivec]);
    }
  return itotsize;
}

#endif
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

using std::ostream;
using std::ofstream;



ostream& pout ()
{
#ifdef MPI
  int flag;
  MPI_Initialized( &flag );
  if(flag)
  {
    static char buffer[137];
    static int temp = sprintf (buffer, "%d", procID ());
    static std::string par_name = std::string ("pout.") + buffer;
    static ofstream s_pout (par_name.c_str());

    return (s_pout);
  }
  else
  {
    return std::cout;  // MPI hasn't been started yet....
  }
#else
  return (std::cout);
#endif
}
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//


using std::ostream;
using std::cout;
using std::endl;


// The implementation of the tree here is a little
// weird for some.  It is a strictly 2-ary tree. 
// the terminal nodes of the tree represent a 2-ary
// box of information, not just a point.  it uses
// the bit pattern in the pointer data holder to 
// store the true/false information for the points in the
// 2-ary box. This requires 8 bits for the 3D case, a bit
// for each point in the box.  hence the appearance of
// the magic '256' in the code (2^8).  TreeNode pointers
// that have a value less than '256' are not pointing at
// memory, but holding the leaf information.

//
//  Further details of how non-recursive TreeNode design works: 
//
//  (for a 2D tree) 
//
//    (m_tree)    
//           + -- 0  
//
//             (a)+ - 0
//                    1
//                1   +
//                    1  <------you are here
//              
//                + - + - 0
//                    0   1
//                    0   1
//                    0   0
//
//  for the node indicated, the 'index' vector  would contain
//
//  index=[ 0      1    3  ...............]            
//parents=[&m_tree &a   ..................]
//
// or directly refered to as m_tree.nodes[1].nodes[3]
//
//  the tree indicates a covering of an index space in either
//  1 or 0.   1 is stored in the tree by pointing at the static
//  'full' data member, 0 is stored as a 0.
//
//  every 'nodes' member of a TreeNode object can be either
//
//   0, &full, or a pointer .
//
//   The interpretation of the tree depends on what m_spanBox is.
//     nodes[i] indicates whether the i'th quadrant of the parent
//     Box is full, empty, or needs to be parsed deeper.
//
//


// NOTE:  Had to replace the static const integer nodeSize with
// an ugly macro named TIVS_NODESIZE to attain compilation with xlC.
// Currently, this only affects this file and the .H
// Certainly I'd love to fix it better than this...  (ndk)

// handy macro for tree parsing operations....

#define NEXT(index_local, depth)                 \
      index_local[0] = 0;                        \
      index_local[depth]++;                      \
          while(index_local[depth] == TIVS_NODESIZE)  \
            {                                    \
              index_local[depth] = 0;            \
              depth--;                           \
              index_local[depth]++;              \
            }                


void TreeIntVectSet::define(const Box& a_box)
{

  clearTree(m_tree);
  m_minBox = a_box;

  if(a_box.isEmpty())
    {
      m_spanBox = Box();
      return;
    }

  //    int size = 0;
  //    for(int i=0; i<SpaceDim; ++i) size = Max(size, a_box.size(i));
  //    int clippedSize = (size/2)*2;
  //    if(clippedSize < size) size = clippedSize+2;

  //    int minBoxSize = 2;
  //    m_depth = 1;
  //    while(minBoxSize < size) 
  //      {
  //        minBoxSize*=2;
  //        m_depth++;
  //      }

  //    int m = index.size();
  //    m = Max(m_depth+1, m);
  //    if(m > index.size())
  //      {
  //        index.resize(m);
  //        parents.resize(m);
  //        boxes.resize(m);
  //      }

  //    m_spanBox = Box(a_box.smallEnd(), 
  //                    a_box.smallEnd()+(minBoxSize-1)*IntVect::TheUnitVector());


  *this |= a_box;

}

void TreeIntVectSet::clear()
{
  clearTree(m_tree);
  m_spanBox = Box();
  m_minBox  = Box();
  m_depth = 1;
}


TreeIntVectSet& TreeIntVectSet::operator=(const TreeIntVectSet& rhs)
{
  define(rhs);
  return *this;
}

void TreeIntVectSet::define(const TreeIntVectSet& a_tivs)
{
  if(this == &a_tivs) return;

  clearTree(m_tree);

  m_spanBox = a_tivs.m_spanBox;
  m_minBox  = a_tivs.m_minBox;
  m_depth  = a_tivs.m_depth;

  if(a_tivs.m_tree.nodes == 0 || a_tivs.m_tree.nodes == &full)
    {
      m_tree = a_tivs.m_tree;
      return;
    }

  expandNode(m_tree);

  // perform direct Tree-to-Tree copy operation.
  // this is a little more involved than just
  // adding the boxes from one to the other, but
  // it should be much more efficient and faster

  cloneNode(a_tivs.m_tree, m_tree);

}

void TreeIntVectSet::cloneNode(const TreeNode& src, TreeNode& dest)
{
  static Vector<const TreeNode*> otherParents;
  if(parents.size() > otherParents.size())
    otherParents.resize(parents.size());

  if(src.nodes ==0 || src.nodes == &full)
    {
      dest=src;
      return;
    }
  if(dest.nodes == 0 || dest.nodes == &full)
    {
      expandNode(dest);
    }

  otherParents[0] = &(src);
  parents[0]  = &(dest);
  index[1] = 0;
  int depth = 1;
  while(depth!=0) 
    {
      const TreeNode* otherParent  = otherParents[depth-1];
      TreeNode* thisParent   = (TreeNode*)parents[depth-1];
      int ind = index[depth];
      const TreeNode& otherCurrent = otherParent->nodes[ind];
      TreeNode& thisCurrent  = thisParent->nodes[ind];

      if(otherCurrent.nodes == 0 || otherCurrent.nodes == &full) // terminal
        {
          thisCurrent = otherCurrent;
          nextNode(depth);
        }
      else  // not terminal node, add nodes and traverse deeper
        {
          expandNode(thisCurrent);
          otherParents[depth] = &(otherCurrent);
          parents[depth]  = &(thisCurrent);
          depth++;
          index[depth] = 0;
        }
    }
}

void TreeIntVectSet::clearTree(TreeNode& tree)
{
  // quick check to eliminate the work of tiny tree deletion
  if(tree.nodes == 0 || tree.nodes == &full) {tree.nodes = 0; return; }

  static Vector<TreeNode*> parents_local;
  static Vector<int>       index_local;
  parents_local.resize(TreeIntVectSet::parents.size());
  index_local.resize(TreeIntVectSet::index.size());
  // Line below added by petermc, 11 Dec 2001, to avoid crash
  if (index_local.size() == 0) {tree.nodes = 0; return; }
  index_local[0] = 0;

  int depth = 1;

  parents_local[0] = &(tree);
  index_local[1] = 0;
  while(depth != 0)
    {
      TreeNode* parent =  parents_local[depth-1];
      int ind = index_local[depth];
      //Vector<int>& indexRef = index;
      //Vector<TreeNode*>& parentRef = parents;
      TreeNode& current = parent->nodes[ind];
      if(current.nodes == 0 || current.nodes == &full)
        {
          index_local[depth]++;
          while(index_local[depth] == TIVS_NODESIZE)
            {
              index_local[depth] = 0;
              //delete[] parents[depth-1]->nodes;
              treeNodePool.returnPtr(parents_local[depth-1]->nodes);
              depth--;
              index_local[depth]++;
            }
        }
      else
        {
          parents_local[depth] = &(current);
          depth++;
          index_local[depth] = 0;
        }
    }
  tree.nodes = 0;

}

void TreeIntVectSet::refine(int iref)
{
  if(iref == 1) return;
  assert(iref >= 1);
  int refinements = 1;
  int r = 2;
  while(r < iref)
     {
       refinements++;
       r*=2;
     }

  assert(r == iref); // check iref for power of 2
  m_spanBox.refine(iref);
  m_minBox.refine(iref);
  m_depth+=refinements;
  int m=index.size();
  m = Max(m_depth+1, m);
  if(m > index.size())
    {
      index.resize(m);
      parents.resize(m);
      boxes.resize(m);
          bufferOffset.resize(m);
    }
}

void TreeIntVectSet::shift(const IntVect& iv)
{
  static Vector<Box> boxs;
  int size;

  createBoxes(boxs, size);
  clear();
  for(int i=0; i < size; ++i)
    {
      boxs[i].shift(iv);
    }
  for(int i=0; i<size; ++i)
    {
      *this |= boxs[i];
    } 
}

void TreeIntVectSet::grow(int igrow)
{
  if(igrow == 0 ) return;
  if(igrow < 0) MayDay::Error("TreeIntVectSet::grow(int) called with negative value");

  static Vector<Box> boxs;
  int size;

  createBoxes(boxs, size);
  clear();
  for(int i=0; i < size; ++i)
    {
      boxs[i].grow(igrow);
    }
  for(int i=0; i<size; ++i)
    {
      *this |= boxs[i];
    } 
}

void TreeIntVectSet::grow(int idir, int igrow)
{
  if(igrow == 0 ) return;
  if(igrow < 0) MayDay::Error("TreeIntVectSet::grow(int) called with negative value");
  static Vector<Box> boxs;
  int size;

  createBoxes(boxs, size);
  clear();
  for(int i=0; i<size; ++i)
    {
      boxs[i].grow(idir, igrow);
    }
  for(int i=0; i<size; ++i)
    {
      *this |= boxs[i];
    } 
}

/*
TreeIntVectSet TreeIntVectSet::chop(int dir, int chop_pnt)
{
  if(m_minBox.smallEnd(dir) >= chop_pnt)
    {
      TreeIntVectSet rtn(*this);
      clear();
      return rtn;
    }
  if(m_minBox.bigEnd(dir) < chop_pnt)
    {
      return TreeIntVectSet();
    }


  static Vector<Box> boxs;
  int size;
  createBoxes(boxs, size);

  clear();
  TreeIntVectSet rtn;
  for(int i=0; i<size; ++i)
    {
      Box& box = boxs[i];
      if(box.smallEnd(dir) >= chop_pnt)
        rtn |= box;
      else if(box.bigEnd(dir) < chop_pnt)
        *this |= box;
      else
        {
          Box hi = box.chop(dir, chop_pnt);
          rtn |= hi;
          *this |= box;
        }
    }
  return rtn;
}

*/

TreeIntVectSet TreeIntVectSet::chop(int dir, int chop_pnt)
{
  if(m_minBox.smallEnd(dir) >= chop_pnt)
    {
      TreeIntVectSet rtn(*this);
      clear();
      return rtn;
    }
  if(m_minBox.bigEnd(dir) < chop_pnt)
    {
      return TreeIntVectSet();
    }

  Box min  = m_minBox;
  Box chop = m_spanBox;

  TreeIntVectSet rtn(*this);

  Box chopThis = chop.chop(dir, chop_pnt);
  rtn   -= chop;
  *this -= chopThis;

//  rtn.recalcMinBox();
//  recalcMinBox();

  m_minBox = min;
  rtn.m_minBox = m_minBox.chop(dir, chop_pnt);

  return rtn;

}

/*
  void TreeIntVectSet::coarsen(int icoarse)
  {
    static  Vector<Box> boxs;
   int size;

    createBoxes(boxs, size);
    clear();
    for(int i=0; i<size; ++i)
      {
        Box& box = boxs[i];
        box.coarsen(icoarse);
        this->operator|=(box);
      }
    compact();
  }
*/


void TreeIntVectSet::coarsen(int icoarse)
{ 
  if(icoarse == 1) return; 
  assert(icoarse >= 1); 
  int coarsenings = 1;
  int c = 2;
  while(c < icoarse){
	c*=2;
	coarsenings++;
  }
  assert(c == icoarse); // check icoarse for power of 2 

  compact(); 
  if(m_tree.nodes == 0) return; 
  if(m_tree.nodes == &full) { m_spanBox.coarsen(icoarse); return;} 
  if(coarsenings > m_depth)  
    {  
      clearTree(m_tree); 
      m_tree.nodes = &full; 
      m_spanBox.coarsen(icoarse);
      m_minBox.coarsen(icoarse);
      return; 
    } 

  // first, pure tree manipulation, turn all unit sized terminal nodes 
  // into gathered nodes.
  //      ie.
  // if depth == m_depth-1  && TreeNode.node != 0 or full  
  //     TreeNode.nodes = [full,0,0,full]  -> TreeNode.nodes = full  

  parents[0] = &m_tree; 

  index[1] = 0; 
  index[0] = 0;
  int depth = 1; 
  while(depth != 0) 
    { 
      TreeNode& current = parents[depth-1]->nodes[index[depth]]; 
      if(depth == m_depth-coarsenings-1) 
        { 
          if(current.nodes != 0)
            {
              clearTree(current); 
              current.nodes = &full;
            } 
        } 
      else if(!(current.nodes == 0 || current.nodes == &full)) 
        { 
          parents[depth] = &current; 
          depth++; 
          index[depth] = -1; 
        } 
      nextNode(depth); 
    } 

  m_spanBox.coarsen(icoarse); 
  m_minBox.coarsen(icoarse);
  m_depth-=coarsenings; 

}

void TreeIntVectSet::expandNode(TreeNode& a_node)
{
  // Vector<int>&  indexRef = index;
  TreeNode* tmp = a_node.nodes;
  assert(tmp == 0 || tmp == &full);
  //a_node.nodes = new TreeNode[TIVS_NODESIZE];
  a_node.nodes = (TreeNode*)(treeNodePool.getPtr());
  //printf("alloc e: %p\n",a_node.nodes);
  for(int i=0; i<TIVS_NODESIZE; ++i)
    {  
      a_node.nodes[i].nodes = tmp;
    }

}


TreeIntVectSet& TreeIntVectSet::operator|=(const Box& a_box)
{
  if(a_box.isEmpty()) return *this;

  if(m_tree.nodes == 0 || m_tree.nodes == &full)
    {
      expandNode(m_tree);
    }
  m_minBox.minBox(a_box);

  // one half of the real smarts of this class, the other half being
  // the growTree function

  while(!m_spanBox.contains(a_box)) growTree();


  //Vector<int>& indexRef = index; //used for debugging
  //Vector<Box>& boxRef   = boxes; //  ..............
  index[1] = 0;
  parents[0] = &(m_tree);
  boxes[0] = m_spanBox;
  int depth = 1;

  while(depth != 0)
    {
      TreeNode* parent = parents[depth-1];
      int ind = index[depth];
      TreeNode& current = parent->nodes[ind];
      Box& parentBox  = boxes[depth-1];
      Box& currentBox = boxes[depth];
      quadrantBox(parentBox, index[depth], currentBox);
      // clear up the two most common cases quickly

      // case 0 a_box does not touch this quadrant, or this whole
      //quadrant is already marked 'full'...no change
      if(current.nodes == &full || !a_box.intersectsNotEmpty(currentBox))
        {
          nextNode(depth);
        }
      // case 1 currentBox is completely covered by a_box.
      else if(currentBox.smallEnd() >= a_box.smallEnd() &&
              currentBox.bigEnd()   <= a_box.bigEnd())
        {
          clearTree(current);
          current.nodes = &full;
          nextNode(depth);
        }

      // OK, now things are more tricky, partial intersection of
      // boxes.  Need to parse deeper
      else
        {
          if(current.nodes == 0 || current.nodes == &full)
            expandNode(current);
          parents[depth] = &(current);
          depth++;
          index[depth]= 0;                
        }
    } 

  return *this;

}


TreeIntVectSet& TreeIntVectSet::operator|=(const TreeIntVectSet& set)
{
  if(set.m_tree.nodes==0) return *this;
  if(set.m_tree.nodes == &full)
    {
      this->operator|=(set.m_spanBox);
      return *this;
    }
  if(m_tree.nodes == 0)
    {
      *this = set;
      return *this;
    }
  if(m_tree.nodes == &full)
    {
      Box b = m_spanBox;
      *this = set;
      *this |= b;
      return *this;
    }

  const TreeNode* yourRoot;
  TreeIntVectSet tmp;

  if(m_depth == set.m_depth)
    {
      yourRoot = &(set.m_tree);
    }
  else if(m_depth < set.m_depth)
    {
      while(m_depth < set.m_depth) growTree();
      yourRoot = &(set.m_tree);
    }
  else
    {
      tmp = set;
      while(tmp.m_depth < m_depth) tmp.growTree();
      yourRoot = &(tmp.m_tree);
    }

  static Vector<const TreeIntVectSet::TreeNode*> parents_other, parents_this;
  static Vector<int> indexSet;
  if(parents.size() > parents_other.size())
    {
      parents_other.resize(parents.size());
      parents_this.resize(parents.size());
      indexSet.resize(parents.size());
    }

  parents_other[0] = yourRoot;
  parents_this[0] = &m_tree;
  indexSet[1] = 0;
  int depth = 1;
  while(depth!=0) 
    {
      const TreeNode* otherParent  = parents_other[depth-1];
      TreeNode* thisParent   = (TreeNode*)parents_this[depth-1];
      const TreeNode& otherCurrent = otherParent->nodes[indexSet[depth]];
      TreeNode& thisCurrent  = thisParent->nodes[indexSet[depth]];

      if(otherCurrent.nodes == 0 || thisCurrent.nodes == &full) 
        {
          NEXT(indexSet, depth);
        }
      else if(otherCurrent.nodes == &full)
        {
          clearTree(thisCurrent);
          thisCurrent = otherCurrent;
          // leave the rest of this node alone
          NEXT(indexSet, depth);
        }
      else if(thisCurrent.nodes == 0) 
        {
          cloneNode(otherCurrent, thisCurrent);
          NEXT(indexSet, depth);
        }
      else  // not terminal node for either one, traverse deeper
        {
          parents_other[depth] = &(otherCurrent);
          parents_this[depth]  = &(thisCurrent);
          depth++;
          indexSet[depth] = 0;
        }
    } 
  m_minBox.minBox(set.m_minBox);

  return *this;
}


//old version that didn't rely on common centering techniques.....

// TreeIntVectSet& TreeIntVectSet::operator|=(const TreeIntVectSet& set)
// {

//   if(set.m_tree.nodes==0) return *this;
//   else if(set.m_tree.nodes == &full)
//     {
//       this->operator|=(set.m_spanBox);
//       return *this;
//     }

//   //m_minBox.minBox(set.m_minBox);

//   int depth = set.m_depth+1;
//   static Vector<int> index_local;
//   static Vector<const TreeIntVectSet::TreeNode*> parents_local;
//   static Vector<Box> boxes_local;
//   if(depth > index_local.size())
//     {
//       index_local.resize(depth);
//       parents_local.resize(depth);
//       boxes_local.resize(depth);
//     }
//   index_local[0] = 0;

//   depth = 1;
//   parents_local[0] = &(set.m_tree);
//   boxes_local[0] = set.m_spanBox;
//   index_local[1] = 0;
//   while(depth != 0)
//     {
//       const TreeNode* parent =  parents_local[depth-1];
//       int ind = index_local[depth];
//       const TreeNode& current = parent->nodes[ind];
//       quadrantBox(boxes_local[depth-1], ind, boxes_local[depth]);
//       if(current.nodes == 0 || current.nodes == &full)
//              {
//                if(current.nodes == 0){;}  // do nothing
//                else if(current.nodes == &TreeIntVectSet::full)
//                      {
//                        this->operator|=(boxes_local[depth]);
//                      }
//                index_local[depth]++;
//                while(index_local[depth] == TIVS_NODESIZE)
//                      {
//                        index_local[depth] = 0;
//                        depth--;            
//                        index_local[depth]++;
//                      }
//              }
//       else{
//              parents_local[depth] = &(current);
//              depth++;
//              index_local[depth] = 0;
//       }
//     }
//   return *this;
// } 




TreeIntVectSet& TreeIntVectSet::operator-=(const TreeIntVectSet& set)
{

  if(set.m_tree.nodes==0) return *this;
  else if(set.m_tree.nodes == &full)
    {
      this->operator-=(set.m_spanBox);
      return *this;
    }
  if(m_tree.nodes == 0) return *this;

  int depth = set.m_depth+1;
  static Vector<int> index_local;
  static Vector<const TreeIntVectSet::TreeNode*> parents_local;
  static Vector<Box> boxes_local;
  if(depth > index_local.size())
    {
      index_local.resize(depth);
      parents_local.resize(depth);
      boxes_local.resize(depth);
    }
  index_local[0] = 0;
  depth = 1;
  parents_local[0] = &(set.m_tree);
  boxes_local[0] = set.m_spanBox;
  index_local[1] = 0;
  while(depth != 0)
    {
      const TreeNode* parent =  parents_local[depth-1];
      int ind = index_local[depth];
      const TreeNode& current = parent->nodes[ind];
      quadrantBox(boxes_local[depth-1], ind, boxes_local[depth]);
      if(boxes_local[depth].intersectsNotEmpty(m_minBox))
        {
          if(current.nodes == 0 || current.nodes == &full)
            {
              if(current.nodes == 0){;}  // do nothing
              else if(current.nodes == &TreeIntVectSet::full)
                {
                  this->operator-=(boxes_local[depth]);
                }
              index_local[depth]++;
              while(index_local[depth] == TIVS_NODESIZE)
                {
                  index_local[depth] = 0;
                  depth--;
                  index_local[depth]++;
                }
            }
          else{
            parents_local[depth] = &(current);
            depth++;
            index_local[depth] = 0;
          }
        }
      else
        {
          index_local[depth]++;
          while(index_local[depth] == TIVS_NODESIZE)
            {
              index_local[depth] = 0;
              depth--;
              index_local[depth]++;
            }
        }
    }
  return *this;
} 


TreeIntVectSet& TreeIntVectSet::operator&=(const TreeIntVectSet& set)
{
  if(set.m_tree.nodes==0 || m_tree.nodes == 0 || !m_minBox.intersects(set.m_minBox))
    {
      clear();  
      return *this;
    }

  if(set.m_tree.nodes==&full) { return *this&=set.m_spanBox;}
  else if(m_tree.nodes == &full)
    {
      expandNode(m_tree);
    }

  const TreeNode* yourRoot;
  TreeIntVectSet tmp;

  if(m_depth == set.m_depth)
    {
      yourRoot = &(set.m_tree);
    }
  else if(m_depth < set.m_depth)
    {
      while(m_depth < set.m_depth) growTree();
      yourRoot = &(set.m_tree);
    }
  else
    {
      tmp = set;
      while(tmp.m_depth < m_depth) tmp.growTree();
      yourRoot = &(tmp.m_tree);
    }

  static Vector<const TreeIntVectSet::TreeNode*> parents_other, parents_this;
  static Vector<int> indexSet;
  if(parents.size() > parents_other.size())
    {
      parents_other.resize(parents.size());
      parents_this.resize(parents.size());
      indexSet.resize(parents.size());
    }

  parents_other[0] = yourRoot;
  parents_this[0] = &m_tree;
  indexSet[1] = 0;
  int depth = 1;
  while(depth!=0) 
    {
      const TreeNode* otherParent  = parents_other[depth-1];
      TreeNode* thisParent   = (TreeNode*)parents_this[depth-1];
      const TreeNode& otherCurrent = otherParent->nodes[indexSet[depth]];
      TreeNode& thisCurrent  = thisParent->nodes[indexSet[depth]];

      if(otherCurrent.nodes == 0 || thisCurrent.nodes == 0) 
        {
          clearTree(thisCurrent);
          NEXT(indexSet, depth);
        }
      else if(otherCurrent.nodes == &full)
        {
          // leave the rest of this node alone
          NEXT(indexSet, depth);
        }
      else if(thisCurrent.nodes == &full) // swap nodes
        {
          clearTree(thisCurrent);
          cloneNode(otherCurrent, thisCurrent);
          NEXT(indexSet, depth);
        }
      else  // not terminal node for either one, traverse deeper
        {
          parents_other[depth] = &(otherCurrent);
          parents_this[depth]  = &(thisCurrent);
          depth++;
          indexSet[depth] = 0;
        }
    } 
  m_minBox &= set.m_minBox;

  return *this;
}

/*
TreeIntVectSet& TreeIntVectSet::operator&=(const TreeIntVectSet& set)
{

  if(set.m_tree.nodes==0 || m_tree.nodes == 0 || !m_minBox.intersects(set.m_minBox))
    {
      clear();  
      return *this;
    }
  if(set.m_tree.nodes==&full) { return *this&=set.m_spanBox;}
  else if(m_tree.nodes == &full)
    {
      expandNode(m_tree);
    }

  // step one, trim by be minbox;
  *this &= set.m_minBox;

  // now, subtract the complement of the
  int depth = set.m_depth+1;
  static Vector<int> index_local;
  static Vector<const TreeIntVectSet::TreeNode*> parents_local;
  static Vector<Box> boxes_local;
  if(depth > index_local.size())
    {
      index_local.resize(depth);
      parents_local.resize(depth);
      boxes_local.resize(depth);
    }
  index_local[0] = 0;

  depth = 1;
  parents_local[0] = &(set.m_tree);
  boxes_local[0] = set.m_spanBox;
  index_local[1] = 0;
  while(depth != 0)
    {
      const TreeNode* parent =  parents_local[depth-1];
      int ind = index_local[depth];
      const TreeNode& current = parent->nodes[ind];
      quadrantBox(boxes_local[depth-1], ind, boxes_local[depth]);
      if(current.nodes == 0 || current.nodes == &full)
        {
          if(current.nodes == 0)
            {
              *this -= boxes_local[depth];
            } 
          else if(current.nodes == &TreeIntVectSet::full)
            {
              //do nothing.
            }
          index_local[depth]++;
          while(index_local[depth] == TIVS_NODESIZE)
            {
              index_local[depth] = 0;
              depth--;        
              index_local[depth]++;
            }
        }
      else
        {
          parents_local[depth] = &(current);
          depth++;
          index_local[depth] = 0;
        }
    }
  return *this;
} 

*/


// first pass to help matters a little, next pass can exploit
// the common centering aspect of the tmp and *this tree.  bvs.
void TreeIntVectSet::nestingRegion(int radius, const Box& domain)
{
  assert(radius>=0);
  if (radius == 0) return;

  TreeIntVectSet tmp(*this);
  {
    IntVect lo = m_minBox.smallEnd();
    IntVect hi = m_minBox.bigEnd();
    for(int i=0; i<SpaceDim; ++i)
      {
        if(lo[i] != domain.smallEnd()[i]) lo[i] += radius;
        if(hi[i] != domain.bigEnd()[i])   hi[i] -= radius;
      }
    Box shrink(lo, hi);
    *this &= shrink;
  }

  IntVect lower(-radius * IntVect::Unit), upper(IntVect::Unit * radius);

  int depth  = m_depth+1;
  static Vector<int> index_tmp;
  static Vector<const TreeIntVectSet::TreeNode*> parents_tmp;
  static Vector<Box> boxes_tmp;
  if(depth > index_tmp.size())
    {
      index_tmp  .resize(depth+1);
      parents_tmp.resize(depth+1);
      boxes_tmp  .resize(depth+1);
    }
  index_tmp[0] = 0;

  depth = 1;
  parents_tmp[0] = &(tmp.m_tree);
  boxes_tmp[0] = m_spanBox;
  index_tmp[1] = 0;
  Box clobberBox;
  IntVect& lo = (IntVect&)(clobberBox.smallEnd());
  IntVect& hi = (IntVect&)(clobberBox.bigEnd());
  while(depth != 0)
    {
      const TreeNode* parent =  parents_tmp[depth-1];
      int ind = index_tmp[depth];
      Box&  currentBox = boxes_tmp[depth];
      const TreeNode& current = parent->nodes[ind];
      quadrantBox(boxes_tmp[depth-1], ind, currentBox);
      if(current.nodes == 0 || current.nodes == &full)
        {
          if(current.nodes == 0)
            {
              lo = currentBox.smallEnd();
              hi = currentBox.bigEnd();
              lo.max(domain.smallEnd());
              hi.min(domain.bigEnd());
              if(hi>=lo)
                {
                  hi+=upper;
                  lo+=lower;
                  clobberBox.computeBoxLenNotEmpty();
                  *this -= clobberBox;
                } 
            }
          else if(current.nodes == &TreeIntVectSet::full)
            {
              //do nothing.
            }
          index_tmp[depth]++;
          while(index_tmp[depth] == TIVS_NODESIZE)
            {
              index_tmp[depth] = 0;
              depth--;
              index_tmp[depth]++;
            }
        }
      else
        {
          parents_tmp[depth] = &(current);
          depth++;
          index_tmp[depth] = 0;
        }
    }
}

void TreeIntVectSet::nestingRegion(int radius, const ProblemDomain& domain)
{
  assert(radius>=0);
  if (radius == 0) return;

  const Box& domainBox = domain.domainBox();

  Box domainImaged = domainBox;

  TreeIntVectSet tmp(*this);
  {
    IntVect lo = m_minBox.smallEnd();
    IntVect hi = m_minBox.bigEnd();
    for(int i=0; i<SpaceDim; ++i)
      {
                if(lo[i] != domainBox.smallEnd()[i]) lo[i] += radius;
                if(hi[i] != domainBox.bigEnd()[i])   hi[i] -= radius;
                if(domain.isPeriodic(i)){
                  domainImaged.grow(i, radius);
                }
      }
    Box shrink(lo, hi);
    *this &= shrink;
  }

  // secret of periodic handling.  create phantom IntVects in the
  // tmp TreeIntVectSet for the periodic images up to radius distance
  if(domain.isPeriodic())
    {
      Vector<Box> boxRep = tmp.createBoxes();
      ShiftIterator shiftIt = domain.shiftIterator();
      Box intersectBox(domainBox);
      for (int dir=0; dir<SpaceDim; dir++) 
        {
          if (domain.isPeriodic(dir))
            intersectBox.grow(dir,-radius);
        }
      IntVect shiftMult(domainBox.size());
      for (int b=0; b<boxRep.size(); ++b) 
        {
          Box image = boxRep[b];
          if (!intersectBox.contains(image)) 
            {
              for (shiftIt.begin(); shiftIt.ok(); ++shiftIt) 
                {
                  IntVect shiftVect(shiftMult*shiftIt());
                  image.shift(shiftVect);
                  if (image.intersects(domainImaged))
                    {
                      tmp |= (image & domainImaged);
                    }
                  image.shift(-shiftVect);
                } // end loop over periodic shifts
            } // end if box needs to be checked for periodic intersections
        } // end loop over boxes

      // this is designed to catch case where this level
      // abuts the physical domain, but there are no periodically
      // shifted boxes to subtract (dfm 11/13/01)
      Box tmpMinBox = tmp.m_minBox;
      IntVect loVect = tmpMinBox.smallEnd();
      IntVect hiVect = tmpMinBox.bigEnd();

      for (int dir=0; dir<SpaceDim; dir++) 
      {
        if (domain.isPeriodic(dir))
          {
            if (tmpMinBox.smallEnd()[dir] == domainBox.smallEnd(dir)) 
              loVect[dir] += radius;
            if (tmpMinBox.bigEnd()[dir] == domainBox.bigEnd(dir)) 
              hiVect[dir] -= radius;
          } // end if periodic in this direction
      } // end loop over directions

      Box testBox(loVect,hiVect);
      // if testBox != tmpMinBox then at least one box in this IntVect
      // abuts a periodic boundary, but there are no periodic images
      // which abut it.  So, this edge must be shrunk as well
      if (testBox != tmpMinBox) 
        *this &= testBox;

    } // end if is periodic
                                

  IntVect lower(-radius * IntVect::Unit), upper(IntVect::Unit * radius);

  int depth  = m_depth+1;
  Vector<int> index_tmp;
  Vector<const TreeIntVectSet::TreeNode*> parents_tmp;
  Vector<Box> boxes_tmp;
  if(depth > index_tmp.size())
    {
      index_tmp  .resize(depth+1);
      parents_tmp.resize(depth+1);
      boxes_tmp  .resize(depth+1);
    }
  index_tmp[0] = 0;

  depth = 1;
  parents_tmp[0] = &(tmp.m_tree);
  boxes_tmp[0] = tmp.m_spanBox;
  index_tmp[0] = 0;
  index_tmp[1] = 0;
  Box clobberBox;
  IntVect& lo = (IntVect&)(clobberBox.smallEnd());
  IntVect& hi = (IntVect&)(clobberBox.bigEnd());
  while(depth != 0)
    {
      const TreeNode* parent =  parents_tmp[depth-1];
      int ind = index_tmp[depth];
      Box&  currentBox = boxes_tmp[depth];
      const TreeNode& current = parent->nodes[ind];
      quadrantBox(boxes_tmp[depth-1], ind, currentBox);
      if(current.nodes == 0 || current.nodes == &full)
        {
          if(current.nodes == 0)
            {
              lo = currentBox.smallEnd();
              hi = currentBox.bigEnd();
                          lo.max(domainImaged.smallEnd());
                          hi.min(domainImaged.bigEnd());
              if(hi>=lo)
                {
                  hi+=upper;
                  lo+=lower;
                  clobberBox.computeBoxLenNotEmpty();
                  *this -= clobberBox;
                } 
            }
          else if(current.nodes == &TreeIntVectSet::full)
            {
              //do nothing.
            }
          index_tmp[depth]++;
          while(index_tmp[depth] == TIVS_NODESIZE)
            {
              index_tmp[depth] = 0;
              depth--;
              index_tmp[depth]++;
            }
        }
      else
        {
          parents_tmp[depth] = &(current);
          depth++;
          index_tmp[depth] = 0;
        }
    }
}

bool TreeIntVectSet::contains(const IntVect& iv) const
{
  if(m_tree.nodes == 0 || !m_minBox.contains(iv)) return false;
  if(m_tree.nodes == &full) return true;

  //Vector<int>& indexRef = index; //used for debugging
  //Vector<Box>& boxRef   = boxes; //  ..............

  int depth = 1;
  parents[0] = (TreeNode*)&(m_tree);
  boxes[0] = m_spanBox;
  index[1] = 0;
  while(depth != 0)
    {
      const TreeNode* parent =  parents[depth-1];
      int ind = index[depth];
      const TreeNode& current = parent->nodes[ind];
      quadrantBox(boxes[depth-1], ind, boxes[depth]);
      if(!boxes[depth].contains(iv))
        {
          index[depth]++;
          assert(index[depth] != TIVS_NODESIZE);
        }
      else
        {
          if(current.nodes == 0) return false;
          if(current.nodes == &full) return true;
          parents[depth] = (TreeNode*)&(current);
          depth++;
          index[depth] = 0;
        }
    }
  return false;
} 

bool TreeIntVectSet::isEmpty() const
{
  if(m_tree.nodes == 0) return true;
  if(m_tree.nodes == &full) return false;

  int depth = 1;
  parents[0] = (TreeNode*)&m_tree;
  index[1] = 0;
  while(depth != 0)
    {
      const TreeNode* parent =  parents[depth-1];
      int ind = index[depth];
      const TreeNode& current = parent->nodes[ind];
      if(current.nodes == 0 || current.nodes == &full)
        {
          if(current.nodes == 0){;}  // do nothing
          else if(current.nodes == &full)
            {
                          return false;
            }
          nextNode(depth);
        }
      else
        {
          parents[depth] = (TreeNode*)&(current);
          depth++;
          index[depth] = 0;
        }
    }
  ((TreeIntVectSet*)this)->clear();
  return true;
}

TreeIntVectSet& TreeIntVectSet::operator&=(const ProblemDomain& domain)
{
  TreeIntVectSet tmp(*this);
  tmp -= domain.domainBox();
  *this &= domain.domainBox();
  TreeIntVectSetIterator it(tmp);
  for(it.begin(); it.ok(); ++it)
	{
	  IntVect iv = it();
	  if(domain.image(iv)) *this |= iv;
	}
  return *this;
}

TreeIntVectSet& TreeIntVectSet::operator&=(const Box& a_box)
{
  // one half of the real smarts of this class, the other half being
  // the growTree function

  if(m_tree.nodes == 0) return *this;
  if(a_box.contains(m_minBox)) return *this;
  if(m_tree.nodes == &full) expandNode(m_tree);

  //Vector<int>& indexRef = index; //used for debugging
  //Vector<Box>& boxRef   = boxes; //  ..............
  index[0] = 0;
  index[1] = 0;
  parents[0] = &(m_tree);
  boxes[0] = m_spanBox;
  int depth = 1;
  while(depth != 0)
    {
      TreeNode* parent = parents[depth-1];
      int ind = index[depth];
      TreeNode& current = parent->nodes[ind];
      Box& parentBox  = boxes[depth-1];
      Box& currentBox = boxes[depth];
      quadrantBox(parentBox, index[depth], currentBox);
      if(current.nodes == 0){
        nextNode(depth);
      }
      else if(current.nodes == &full)
        {
          if(a_box.contains(currentBox))
            {
              nextNode(depth);
            }
          else if(!a_box.intersectsNotEmpty(currentBox))
            {
              clearTree(current);
              current.nodes = 0;
              nextNode(depth);
            }
          else
            {
              expandNode(current);
              parents[depth] = &(current);
              depth++;
              index[depth]= 0;
            }
        }
      // case 1 currentBox is completely covered by a_box.
      else
        {
          parents[depth] = &(current);
          depth++;
          index[depth]= 0;
        }
    } 

  return *this; 
}


TreeIntVectSet& TreeIntVectSet::operator-=(const Box& a_box)
{


  if(m_tree.nodes == 0 || (a_box & m_minBox).isEmpty()) return *this;
  if(a_box.contains(m_minBox))
    {
      clear();
      return *this;
    }
  if(m_tree.nodes == &full)
    {
      expandNode(m_tree);
    }

  //Vector<int>& indexRef = index; //used for debugging
  //Vector<Box>& boxRef   = boxes; //  ..............
  index[0] = 0;
  index[1] = 0;
  parents[0] = &(m_tree);
  boxes[0] = m_spanBox;
  int depth = 1;
  while(depth != 0)
    {
      TreeNode* parent = parents[depth-1];
      int ind = index[depth];
      TreeNode& current = parent->nodes[ind];
      Box& parentBox  = boxes[depth-1];
      Box& currentBox = boxes[depth];
      quadrantBox(parentBox, index[depth], currentBox);
      if(a_box.contains(currentBox))
        {
          clearTree(current);
          current.nodes = 0;
          nextNode(depth);
        }
      else if(current.nodes == 0)
        {
          nextNode(depth);
        }
      else if(current.nodes == &full)
        {
          if((depth == m_depth-1) || !a_box.intersectsNotEmpty(currentBox))
            {
              nextNode(depth);
            }
          else
            {
              expandNode(current);
              parents[depth] = &(current);
              depth++;
              index[depth]= 0;
            }   
        }
      else
        {
          parents[depth] = &(current);
          depth++;
          index[depth]= 0;                
        }
    } 
  compact();
  return *this; 
}

void TreeIntVectSet::growTree()
{
  // do the grow operation

  // two cases. top node is terminal, or it has TIVS_NODESIZE leaves to parse out

  if(m_tree.nodes == 0 || m_tree.nodes == &full)
    {
      expandNode(m_tree);
    }
  else
    {
      for(int i=0; i<TIVS_NODESIZE; ++i)
        {
          TreeNode* tmp = m_tree.nodes[i].nodes;
          if(tmp != 0)
            {
              int ind = oppositeQuadrant(i);
              //m_tree.nodes[i].nodes = new TreeNode[TIVS_NODESIZE];
              TreeNode* newPtr = (TreeNode*)(treeNodePool.getPtr());
              m_tree.nodes[i].nodes = newPtr;
              for(int j=0; j< TIVS_NODESIZE; ++j) newPtr[j].nodes = 0;
              //printf("alloc: %p\n", m_tree.nodes[i].nodes );
              //printf("tmp: %p\n", tmp);
              newPtr[ind].nodes = tmp;
            }
        }

    }

  if(m_spanBox.isEmpty()) 
    m_spanBox = Box(-1*IntVect::Unit, IntVect::Zero);
  else
    m_spanBox.grow(m_spanBox.size(0)/2);


  m_depth++;
  int m = m_depth+1;
  if(m > index.size())
    {
      index.resize(m);
      parents.resize(m);
      boxes.resize(m);
          bufferOffset.resize(m);
    }
}


// Basically, a BoxIterator function without the crap of
// a BoxIterator.
bool TreeIntVectSet::nextIntVect(const Box& box, IntVect& iv)
{

  iv[0]++;
  if(iv[0] > box.bigEnd(0))
    {
#if (CH_SPACEDIM >= 2)
      iv[0] = box.smallEnd(0);
      iv[1]++;
#else
      return false;
#endif
    }
  else
    {
      return true;
    }

#if (CH_SPACEDIM >= 2)
  if(iv[1] > box.bigEnd(1))
    {
#if (CH_SPACEDIM == 3)
      iv[1] = box.smallEnd(1);
      iv[2]++;
#else
      return false;
#endif 
    }
  else
    {
      return true;
    }
#endif

#if (CH_SPACEDIM == 3)
  if(iv[2] > box.bigEnd(2))
    {
      return false;
    }
  return true;
#endif

  //assert(iv<=box.bigEnd());
  //assert(iv>=box.smallEnd());

  // won't ever get here...
  //return true;
}

void TreeIntVectSet::quadrantBox(const Box& inputBox, 
                                 int quadrant, 
                                 Box& outputQuadrant)
{
  int halfSize = inputBox.size(0)/2;
  int q   = TIVS_NODESIZE/2;
  //IntVect low, hi;
  int* low = (int*)(outputQuadrant.loVect());
  int* hi  = (int*)(outputQuadrant.hiVect());
  if(quadrant < q)
    {
      low[0] = inputBox.smallEnd(0);
      hi[0]   = inputBox.smallEnd(0)+halfSize-1;
    }
  else
    {
      low[0] = inputBox.smallEnd(0)+halfSize;
      hi[0]   = inputBox.bigEnd(0);
      quadrant-=q;
    }

#if (CH_SPACEDIM >= 2)
  q/=2;
  if(quadrant < q)
    {
      low[1] = inputBox.smallEnd(1);
      hi[1]   = inputBox.smallEnd(1)+halfSize-1;
    }
  else
    {
      low[1] = inputBox.smallEnd(1)+halfSize;
      hi[1]   = inputBox.bigEnd(1);
      quadrant-=q;
    }
#endif

#if (CH_SPACEDIM >= 3)
  q/=2;
  if(quadrant < q)
    {
      low[2] = inputBox.smallEnd(2);
      hi[2]   = inputBox.smallEnd(2)+halfSize-1;
    }
  else
    {
      low[2] = inputBox.smallEnd(2)+halfSize;
      hi[2]   = inputBox.bigEnd(2);
    }
#endif
  assert(!outputQuadrant.isEmpty());
  outputQuadrant.computeBoxLenNotEmpty();
}

inline
void tab(ostream& os, int tab)
{
  for(; tab>=0 ;tab--) os << " ";
}

int TreeIntVectSet::numPts() const
{
  if(isEmpty()) return 0;
  if(m_tree.nodes == &full) return m_spanBox.numPts();

  int rtn= 0;
  int depth = 1;
  parents[0] = (TreeNode*)&m_tree;
  boxes[0] = m_spanBox;
  index[1] = 0;
  while(depth != 0)
    {
      const TreeNode* parent =  parents[depth-1];
      int ind = index[depth];
      const TreeNode& current = parent->nodes[ind];
      quadrantBox(boxes[depth-1], ind, boxes[depth]);
      if(current.nodes == 0 || current.nodes == &full)
        {
          if(current.nodes == 0){;}  // do nothing
          else if(current.nodes == &full)
            {
              rtn += boxes[depth].numPts();
            }
          nextNode(depth);
        }
      else
        {
          parents[depth] = (TreeNode*)&(current);
          depth++;
          index[depth] = 0;
        }
    }
  return rtn;
}  
Vector<Box> TreeIntVectSet::createBoxes() const
{
  static Vector<Box> buffer(200), rtn;
  int size;
  createBoxes(buffer, size);
  rtn.resize(size, Box());
  for(int i=0; i<size; ++i) rtn[i] = buffer[i];
  return rtn;
}

void TreeIntVectSet::createBoxes(Vector<Box>& rtn, int& size) const
{
  size = 0;
  if(m_tree.nodes == 0) return;
  if(m_tree.nodes == &full)
    {
      if(rtn.size() < 1) rtn.resize(1, m_minBox);
      else               rtn[0] = m_minBox;
      size = 1;
      return;
    }
  int depth = 1;
  parents[0] = (TreeNode*)&m_tree;
  boxes[0] = m_spanBox;
  index[1] = 0;
  while(depth != 0)
    {
      const TreeNode* parent =  parents[depth-1];
      int ind = index[depth];
      const TreeNode& current = parent->nodes[ind];
      quadrantBox(boxes[depth-1], ind, boxes[depth]);
      if(current.nodes == 0 || current.nodes == &full)
        {
          if(current.nodes == 0){;}  // do nothing
          else if(current.nodes == &full)
            {
              if(rtn.size() > size) 
                rtn[size] = boxes[depth];
              else
                rtn.push_back(boxes[depth]);
              ++size;
            }
          nextNode(depth);
        }
      else
        {
          parents[depth] = (TreeNode*)&(current);
          depth++;
          index[depth] = 0;
        }
    }
} 

bool TreeIntVectSet::contains(const Box& box) const
{
  if(m_tree.nodes == 0) return false;
  if(!m_minBox.contains(box)) return false;
  if(m_tree.nodes == &full)
    { 
      return true;
    }
  int depth = 1;
  parents[0] = (TreeNode*)&m_tree;
  boxes[0] = m_spanBox;
  index[1] = 0;
  while(depth != 0)
    {
      const TreeNode* parent =  parents[depth-1];
      int ind = index[depth];
      const TreeNode& current = parent->nodes[ind];
      quadrantBox(boxes[depth-1], ind, boxes[depth]);
      if(!box.intersects(boxes[depth]))
        nextNode(depth);
      else if(current.nodes == 0 || current.nodes == &full)
        {
          if(current.nodes == 0)
            {
              // OK, boxes[depth] is NOT in IntVectSet, but it does
              // intersect 'box', hence we do not 'contain' this box.
              return false;
            }
          else if(current.nodes == &full)
            {
              //do nothing
            }
          nextNode(depth);
        }
      else{
        parents[depth] = (TreeNode*)&(current);
        depth++;
        index[depth] = 0;
      }
    }
  return true;
} 
void dumpTree(const TreeIntVectSet* set)
{
  if(set == NULL) return;
  if(set->m_tree.nodes == 0) return;
  if(set->m_tree.nodes == &TreeIntVectSet::full) 
    {
      cout << set->m_spanBox.smallEnd()
           <<"..."<<set->m_spanBox.bigEnd()<<"\n";
      return;
    }
  int depth = set->m_depth+1;
  static Vector<int> index;
  static Vector<const TreeIntVectSet::TreeNode*> parents;
  static Vector<Box> boxes;
  if(depth > index.size())
    {
      index.resize(depth);
      parents.resize(depth);
      boxes.resize(depth);
    }
  depth = 1;
  parents[0] = &(set->m_tree);
  boxes[0] = set->m_spanBox;
  index[0] = 0;
  index[1] = 0;
  while(depth != 0)
    {
      const TreeIntVectSet::TreeNode* parent =  parents[depth-1];
      int ind = index[depth];
      const TreeIntVectSet::TreeNode& current = parent->nodes[ind];
      TreeIntVectSet::quadrantBox(boxes[depth-1], ind, boxes[depth]);
      if(current.nodes == 0 || current.nodes == &TreeIntVectSet::full)
        {
          if(current.nodes == 0)
            {
              cout<<depth;
              tab(cout, depth);
              cout <<"0\n";
            }
          else if(current.nodes == &TreeIntVectSet::full)
            {
              cout<<depth;
              tab(cout, depth);
              cout << boxes[depth].smallEnd()
                   <<"..."<<boxes[depth].bigEnd()<<"\n";
            }

          index[depth]++;
          while(index[depth] == TIVS_NODESIZE)
            {
              index[depth] = 0;
              depth--;
              index[depth]++;
            }
        }
      else
        {
          cout<<depth;
          tab(cout, depth);
          cout <<"+\n";
          parents[depth] = &(current);
          depth++;
          index[depth] = 0;
        }
    }
} 

void TreeIntVectSet::recalcMinBox() const
{
  Box min = Box();
  if(m_tree.nodes == 0)
    {
      return;
    }
  if(m_tree.nodes == &full)
    { 
      min = m_spanBox;
      return;
    }
  int depth = 1;
  parents[0] = (TreeNode*)&m_tree;
  boxes[0] = m_spanBox;
  index[1] = 0;
  while(depth != 0)
    {
      const TreeNode* parent =  parents[depth-1];
      int ind = index[depth];
      const TreeNode& current = parent->nodes[ind];
      quadrantBox(boxes[depth-1], ind, boxes[depth]);
      if(current.nodes == 0 || current.nodes == &full)
        {
          if(current.nodes == 0)
            {
              // do nothing, this region is not part of IntVectSet
            }
          else if(current.nodes == &full)
            {
              min.minBox(boxes[depth]);
            }
          nextNode(depth);
        }
      else{
        parents[depth] = (TreeNode*)&(current);
        depth++;
        index[depth] = 0;
      }
    }
  (Box&)m_minBox = min;
} 


int TreeIntVectSet::oppositeQuadrant(int ind)
{
  assert(ind < TIVS_NODESIZE);
  assert(ind >= 0);
  return TIVS_NODESIZE+~ind;
}




void TreeIntVectSet::compact() const
{
  if(m_tree.nodes == 0 || m_tree.nodes == &full)
    {
      (Box&)m_minBox=m_spanBox;
      return;
    }
  int depth = 1;
  static Vector<Tuple<TreeNode*, 8> > flags;
  flags.resize(index.size());
  parents[0] = (TreeNode*)&(m_tree);
  index[1] = 0;

  while(depth != 0)
    {
      TreeNode* parent =  parents[depth-1];
      int ind = index[depth];
      //Vector<int>& indexRef = index;  //debug aids
      //Vector<TreeNode*>& parentRef = parents; //debug aids
      TreeNode& current = parent->nodes[ind];
      flags[depth][ind] = current.nodes;
      if(current.nodes == 0 || current.nodes == &full)
        {
          index[depth]++;
          while(index[depth] == TIVS_NODESIZE)
            {
              int i=0;
              while(i< TIVS_NODESIZE-1 && flags[depth][i] == flags[depth][i+1]) i++;
              if(i == TIVS_NODESIZE-1)
                {
                  TreeNode* parent = parents[depth-1];
                  clearTree(*parent);
                  parent->nodes = flags[depth][0];
                  flags[depth-1][index[depth-1]] = flags[depth][0] ;
                }
              index[depth] = 0;
              depth--;
              index[depth]++;
            }
        }
      else
        {
          parents[depth] = &(current);
          depth++;
          index[depth] = 0;
        }
    }
  // OK, we have the non-redundant tree, now can optimize a little.
  // Trees are origin centered, so can't re-center, but we can
  // see if reverse of growing makes sense

}


/*
int TreeIntVectSet::linearSize() const
{
  compact();

  int size = 2* linearSize(m_minBox) + linearSize(m_depth);

  if(isEmpty() || m_tree.nodes == &full){
        return size;
  }

  size += TIVS_NODESIZE;

  int depth = 1;
  parents[0] = (TreeNode*)&m_tree;
  index[1] = 0;
  while(depth != 0)
    {
      const TreeNode* parent =  parents[depth-1];
      int ind = index[depth];
      const TreeNode& current = parent->nodes[ind];

      if(current.nodes == 0 || current.nodes == &full)
        {
                  // do nothing
        }
      else{
                size += TIVS_NODESIZE;
        parents[depth] = (TreeNode*)&(current);
        depth++;
        index[depth] = 0;
      }
    }
  return size;
  
}

void TreeIntVectSet::linearOut(void* const a_outBuf) const
{
  unsigned char* buffer = (unsigned char*)a_outBuf;
  Box emptyBox;

  if(isEmpty() || m_tree.nodes == &full){
        linearOut(buffer, emptyBox); // empty box acts as special token
  } else {
        linearOut(buffer, m_minBox);
  }

  buffer += linearSize(m_minBox);

  if(isEmpty()){
        linearOut(buffer, emptyBox);
        return;
  } else {
        linearOut(buffer, m_spanBox);
        buffer += ::linearSize(emptyBox);
        linearOut(buffer, m_depth);
        buffer+= ::linearSize(m_depth);
  }

  int* buf = (int*)buffer;
  
  int depth = 1;
  parents[0] = (TreeNode*)&m_tree;

  index[1] = 0;
  bufferOffset[1] = 0;
  int nextFree = TIVS_NODESIZE;
  while(depth != 0)
    {
      const TreeNode* parent =  parents[depth-1];
      int ind = index[depth];
      const TreeNode& current = parent->nodes[ind];
      if(current.nodes == 0 || current.nodes == &full)
        {
          if(current.nodes == 0)
            {
                          buf[bufferOffset[depth]] = 0;
            }
          else if(current.nodes == &full)
            {
                          buf[bufferOffset[depth]] = 1;
            }
                  bufferOffset[depth]++;
          nextNode(depth);
        }
      else{
                buf[bufferOffset[depth]] = nextFree;
                bufferOffset[depth+1] = nextFree;
                nextFree += TIVS_NODESIZE;
        parents[depth] = (TreeNode*)&(current);
        depth++;
        index[depth] = 0;
      }
    }
  
}

void TreeIntVectSet::linearIn(const void* const inBuf)
{
  clear();
  unsigned char* buffer = (unsigned char*)inBuf;
  ::linearIn(m_minBox, buffer);
  buffer+=::linearSize(m_minBox);
  ::linearIn(m_spanBox, buffer);
  buffer+=::linearSize(m_spanBox);
  ::linearIn(m_depth, buffer);
  buffer+=::linearSize(m_depth);
  
  if(m_minBox.isEmpty()){
        if(m_spanBox.isEmpty()){
          clear();
        } else {
          m_tree.nodes = &full;
        }
        return;
  }

  int* buf = (int*)buffer;

  expandNode(m_tree);

  int depth = 1;
  parents[0] = (TreeNode*)&m_tree;

  index[1] = 0;
  if(bufferOffset.size() < index.size()) bufferOffset.resize(index.size());
  bufferOffset[1] = 0;
  while(depth != 0)
    {
      const TreeNode* parent =  parents[depth-1];
      int ind = index[depth];
      TreeNode& current = parent->nodes[ind];
          int bi = buf[bufferOffset[depth]];
      if(bi == 0 || bi == 1)
        {
          if(bi == 0)
            {
                          current.nodes = 0;
            }
          else if(bi == 1)
            {
                          current.nodes = &full;
            }
                  bufferOffset[depth]++;
          nextNode(depth);
        }
      else{
                bufferOffset[depth+1] = bi;
                expandNode(current);
        parents[depth] = (TreeNode*)&(current);
        depth++;
        index[depth] = 0;
      }
    }
}

*/
  //======================================================================


void TreeIntVectSetIterator::begin()
{
  if(m_ivs->m_tree.nodes == 0)
        {
          m_depth = -1;
          return;
        }
  boxes[0] = m_ivs->m_spanBox;
  TreeIntVectSet::quadrantBox(boxes[0], 0, boxes[1]);
  nodes[0] = (TreeIntVectSet::TreeNode*)&(m_ivs->m_tree); 
  m_depth = 1;
  index[1] = -1;
  if( m_ivs->m_tree.nodes == &TreeIntVectSet::full)
        {
          m_current = boxes[0].smallEnd();
          m_depth = 0;
          return;
        }
  else{
        findNextNode();
  }
  assert((m_depth == -1) || (m_ivs->m_minBox.contains(m_current)));
}

void TreeIntVectSetIterator::findNext()
{
  if(TreeIntVectSet::nextIntVect(boxes[m_depth], m_current))
        return;
  else
        {
          findNextNode();
        }
}

void TreeIntVectSetIterator::findNextNode()
{    
  static TreeIntVectSet::TreeNode* full = &TreeIntVectSet::full;
  index[m_depth]++;
  while(index[m_depth] == TIVS_NODESIZE)
        {
          m_depth--;
          index[m_depth]++;
        }
  while(m_depth > 0)
        {
          const TreeIntVectSet::TreeNode* parent =  nodes[m_depth-1];
          assert(index[m_depth] >= 0);
          const TreeIntVectSet::TreeNode& current = parent->nodes[index[m_depth]];
          TreeIntVectSet::quadrantBox(boxes[m_depth-1], index[m_depth], boxes[m_depth]);
          if(current.nodes == full)
                {
                  m_current = boxes[m_depth].smallEnd();
                  return;
                }
          else if(current.nodes == 0)
                {
                  index[m_depth]++;
                  while(index[m_depth] == TIVS_NODESIZE)
                        {
                          m_depth--;
                          index[m_depth]++;
                        }
                }
          else{
                nodes[m_depth] = &(current);
                m_depth++;
                index[m_depth] = 0;
          }
        }
  m_depth = -1;
}

Vector<Box> TreeIntVectSet::boxes(2);
TreeIntVectSet::TreeNode TreeIntVectSet::full;
Vector<int> TreeIntVectSet::index(2);
Vector<int> TreeIntVectSet::bufferOffset(2);
Vector<TreeIntVectSet::TreeNode*> TreeIntVectSet::parents(2);
Pool TreeIntVectSet::treeNodePool(sizeof(TreeNode[TIVS_NODESIZE]), "TreeIntVectSet Pool");

/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _INTVECTSET_H_
#define _INTVECTSET_H_


#ifndef WRAPPER

#endif

/// An irregular domain on an integer lattice
/**
   IntVectSet represents an irregular region in an integer lattice
   SpaceDim-dimensional index space as an arbitrary collection of
   IntVects.  A full set calculus is defined.  Any IntVect or
   cell-centered Box can be fully represented as an IntVectSet.  There
   is an iterator that provides access to the contents of an
   IntVectSet.  IntVectSets are implicitly cell-centered.
   Intersection, union, and complement operations are defined for
   pairs of IntVectSets (and, by extension, an IntVectSet and an
   IntVect or an IntVectSet and a Box).  The minimum Box of an
   IntVectSet is defined as the smallest Box that contains every
   IntVect in the IntVectSet.

   The IntVects in an IntVectSet do not have a canonical ordering.
*/
class IntVectSet
{
public:
  friend
  class IVSIterator;
  /// {\bf Constructors, Destructor, and defines}

  ~IntVectSet();
  /// 
  /** default constructor: 
      defines an empty IntVectSet. */
  IntVectSet();
  void
  define();

  /// 
  /** copy constructor:
      defines this to be a copy of ige_in. */
  IntVectSet(const IntVectSet& ige_in);
  void
  define(const IntVectSet& ige_in);
  
  IntVectSet copy() const {return *this;}

  void define_intvectset(const IntVectSet& ige_in){ define(ige_in);}
  ///
  /** conversion constructor. */
  explicit
  IntVectSet(const DenseIntVectSet& a_dense);
  void
  define (const DenseIntVectSet& a_dense);

  ///
  /** conversion constructor. */
  explicit
  IntVectSet(const TreeIntVectSet& a_tree);
  void
  define (const TreeIntVectSet& a_tree);

  /// 
  /** IntVect constructor:
      defines this to be an IntVectSet with just one IntVect. */
  explicit
  IntVectSet(const IntVect& iv_in);
  void define(const IntVect& iv_in);
  void define_intvect(const IntVect& iv_in){define(iv_in);}
  /// 
  /** Box constructor:
      defines this to be an IntVectSet with all the cells in a Box. */
  explicit
  IntVectSet(const Box& b);
  void
  define(const Box& b);
  void define_box(const Box& b){define(b);}


  /// {\bf union operators}

  ///
  /** 
      Returns the union of two IntVectSets. \\
      {\bf Arguments:} \\
      ivs1, ivs2 (not modified): \\
      {\bf Returns:} \\
      The union of the two IntVectSets ivs1 and ivs2.\\
  */
  friend
  IntVectSet 
  operator|(const IntVectSet& ivs1, const IntVectSet& ivs2);

  ///
  /** 
      Returns the union of an IntVectSet and an IntVect. \\
      {\bf Arguments:} \\
      ivs (not modified) \\
      iv (not modified): \\
      {\bf Returns:} \\
      The union of the IntVectSet ivs and the IntVect iv.\\
  */
  friend
  IntVectSet 
  operator|(const IntVectSet& ivs, const IntVect& iv);

  ///
  /** 
      Returns the union of an IntVectSet and an IntVect. \\
      {\bf Arguments:} \\
      ivs (not modified) \\
      iv (not modified): \\
      {\bf Returns:} \\
      The union of the IntVectSet ivs and the IntVect iv.\\
  */
  friend
  IntVectSet 
  operator|(const IntVect& iv, const IntVectSet& ivs);

  ///
  /** 
      Returns the union of an IntVectSet and a Box. \\
      {\bf Arguments:} \\
      ivs (not midified): \\
      b (not modified): \\
      {\bf Returns:} \\
      The union of the IntVectSet ivs and the Box b.\\
  */
  friend
  IntVectSet 
  operator|(const IntVectSet& ivs, const Box& b);

  ///
  /** 
      Returns the union of an IntVectSet and a Box. \\
      {\bf Arguments:} \\
      ivs (not midified): \\
      b (not modified): \\
      {\bf Returns:} \\
      The union of the IntVectSet ivs and the Box b.\\
  */
  friend
  IntVectSet 
  operator|(const Box& b, const IntVectSet& ivs);


  ///
  /** 
      Sets this IntVectSet to its union with another IntVectSet. \\
      {\bf Arguments:} \\
      ivs (not modified): \\
      {\bf Returns:} \\
      The union of the two IntVectSets.\\
      {\bf This:} \\
      ---The object is modified---
  */
  IntVectSet&
  operator|=(const IntVectSet& ivs);

  void or_intvectset(const IntVectSet& ivs) { *this |= ivs;}

  ///
  /** 
      Sets this IntVectSet to its union with an IntVect. \\
      {\bf Arguments:} \\
      iv (not modified): \\
      {\bf Returns:} \\
      The union of this IntVectSet and the IntVect.\\
      {\bf This:} \\
      ---The object is modified---
  */
  IntVectSet&
  operator|=(const IntVect& iv);
  void or_intvect(const IntVect& ivs) { *this |= ivs;}


  ///
  /** 
      Sets this IntVectSet to its union with a Box. \\
      {\bf Arguments:} \\
      box (not modified): \\
      {\bf Returns:} \\
      The union of this IntVect and the Box.\\
      {\bf This:} \\
      ---The object is modified---
  */
  IntVectSet&
  operator|=(const Box& b);
  void or_box(const Box& b) { *this |= b;}



  /// {\bf complement operators}

  /// 
  /** 
      Return the complement of the argument IntVectSet within this IntVectSet.
   */
  IntVectSet
  operator-(const IntVectSet& ivs) const;

  /// 
  /** 
      Return the complement of the argument Box within this IntVectSet.
   */
  IntVectSet
  operator-(const Box& b) const;

  /// 
  IntVectSet 
  /** 
      Return the complement of the argument IntVect within this IntVectSet.
   */
  operator-(const IntVect& iv) const;

  /// 
  /**
     Modifies this IntVectSet to be the complement of the argument
     IntVectSet within this IntVectSet.  \\
      {\bf This:} \\
      ---This object is modified.---
  */
  IntVectSet&
  operator-=(const IntVectSet& ivs);
  void minus(const IntVectSet& ivs){*this -= ivs;}

  /// 
  /**
     Modifies this IntVectSet to be the complement of the argument
     Box within this IntVectSet.  \\
      {\bf This:} \\
      ---This object is modified.---
  */
  IntVectSet&
  operator-=(const Box& b);
  void minus_box(const Box& b){ *this -=b;}


  /// 
  /**
     Modifies this IntVectSet to be the complement of the argument
     IntVect within this IntVectSet.\\
      {\bf This:} \\
      ---This object is modified.---  
  */
  IntVectSet&
  operator-=(const IntVect& iv);
  void minus_intvect(const IntVect& iv){ *this -=iv;}

  /// {\bf Intersection operators}

  /// 
  /** 
      Returns the intersection of two IntVectSets.   The result may be empty.\\
      {\bf Arguments:} \\
      ivs1, ivs2 (not modified): \\
      {\bf Returns:} \\
      The intersection of the two IntVectSets ivs1 and ivs2.\\
  */
  friend 
  IntVectSet
  operator&(const IntVectSet& ivs1, const IntVectSet& ivs2);

  /// 
  /** 
      Returns the intersection of an IntVectSet and a Box. \\
      {\bf Arguments:} \\
      ivs (not modified): \\
      {\bf Returns:} \\
      The intersection of the IntVectSet and the Box.\\
  */
  friend
  IntVectSet 
  operator&(const IntVectSet& ivs, const Box& b);

  /// 
  /** 
      Returns the intersection of an IntVectSet and a Box. \\
      {\bf Arguments:} \\
      ivs (not modified): \\
      {\bf Returns:} \\
      The intersection of the IntVectSet and the Box.\\
  */
  friend
  IntVectSet 
  operator&(const Box& b, const IntVectSet& ivs);

  /// 
  /** 
      Sets this IntVectSet to its intersection with another IntVectSet. \\
      {\bf Arguments:} \\
      ivs (not modified): \\
      {\bf Returns:} \\
      The intersection of the two IntVectSets.\\
      {\bf This:} \\
      ---This object is modified.---
  */
  IntVectSet&
  operator&=(const IntVectSet& ivs);
  void and_intvectset(const IntVectSet& ivs){*this &= ivs;}



  /// 
  /** 
      Modifies this IntVectSet to be its intersection with a Box. \\
      {\bf Arguments:} \\
      box (not modified): \\
      {\bf Returns:} \\
      The intersection of the this IntVectSet and the Box.\\
      {\bf This:} \\
      ---This object is modified.---
  */
  IntVectSet&
  operator&=(const Box& b);
  void and_box(const Box& b){*this &= b;}

  /// 
  /** 
      Modifies this IntVectSet to be its intersection with a ProblemDomain. \\
      {\bf Arguments:} \\
      domain (not modified): \\
      {\bf Returns:} \\
      The intersection of the this IntVectSet and the ProblemDomain.\\
      {\bf This:} \\
      ---This object is modified.---
  */
  IntVectSet&
  operator&=(const ProblemDomain& domain);
  void and_domain(const ProblemDomain& d){*this &= d;}


  /// {\bf modification functions}

  ///
  /** 
      Modifies this IntVectSet by growing it by the specified number
      of cells in all directions, including diagonal.  It is an error
      if igrow < 0. \\
      {\bf Arguments:} \\
      igrow (not modified): number of cells to grow.\\
      {\bf Returns:} \\
      This IntVectSet grown by igrow in all directions.\\ 
      {\bf This:} \\
      ---This object is modified.---
*/
  void 
  grow(int igrow);
  ///
  /** 
      Returns the argument IntVectSet grown it by the specified number
      of cells in all directions, including diagonal.  \\
      {\bf Arguments:} \\
      ivs (not modified): \\
      igrow (not midified): number of cells to grow.\\
      {\bf Returns:} \\
      The argument IntVectSet grown by igrow in all directions.\\
*/
  friend
  IntVectSet
  grow(const IntVectSet& ivs, int igrow);



  ///
  /** 
      Modifies this IntVectSet by growing it by the specified number
      of cells in the specified coordinate direction.  Directions are
      zero-based.\\
      {\bf Arguments:} \\
      idir (not modified): zero-based coordinate direction to grow in.\\
      igrow (not modified): number of cells to grow.\\
      {\bf Returns:} \\
      This IntVectSet grown by igrow the direction idir.\\ 
      {\bf This:} \\
      ---This object is modified.---
*/
  IntVectSet&
  grow(int idir, int igrow);
  void grow_dir(int idir, int igrow){grow(idir, igrow);}

  /// 
  /**
     Modifies this IntVectSet by refining it by a specified factor
     iref.  It is an error if iref <= 0.  Definition of refinement:
     For each IntVect iv in the original IntVectSet, the refined
     IntVectSet will contain the Box refine(Box(iv,iv),iref).\\
      {\bf This:} \\
      ---This object is modified.---

   */
  IntVectSet&
  refine(int iref = 2);

  /// 
  /**
     Returns the IntVectSet that is the argument IntVectSet refined by
     a specified factor iref.  It is an error if iref <= 0.
     Definition of refinement: For each IntVect iv in the original
     IntVectSet, the refined IntVectSet will contain the Box
     refine(Box(iv,iv),iref).

   */
  friend
  IntVectSet
  refine(const IntVectSet& ivs, int iref = 2);


  /// 
  /**
     Modifies this IntVectSet by coarsening it by a specified factor
     iref.  It is an error if iref <= 0.  Definition of coarsening:
     For each IntVect iv in the original IntVectSet, the refined
     IntVectSet will contain the IntVect coarsen(iv,iref).\\
      {\bf This:} \\
      ---This object is modified.---

   */
  IntVectSet&
  coarsen(int iref = 2);

  /// 
  /**
     Returns the IntVectSet that is the argument IntVectSet coarsened
     by a specified factor iref.  It is an error if iref <= 0.
     Definition of coarsening: For each IntVect iv in the original
     IntVectSet, the refined IntVectSet will contain the IntVect
     coarsen(iv,iref).

   */
  friend
  IntVectSet
  coarsen(const IntVectSet& ivs, int iref = 2);

  ///
  /** 
      Increment all the IntVects in this IntVectSet by the given iv.\\
      --this IS changed--
   */
  void
  shift(const IntVect& iv);

  ///
  /**
         remove all IntVects that are not at least 'radius' from the the
         edge of the IntVectSet in any direction.  'radius' must be positive
         IntVects that border the box 'domain' are spared from this trimming
  */
  void
  nestingRegion(int radius, const Box& domain);

  ///
  /**
         remove all IntVects that are not at least 'radius' from the the
         edge of the IntVectSet in any direction.  'radius' must be positive
         IntVects that border non-periodic boundaries are spared from this trimming.
                 Radius extends across periodic images.
  */
  void
  nestingRegion(int radius, const ProblemDomain& domain);
  void  nestingRegion_prob(int radius, const ProblemDomain& domain)
  {nestingRegion(radius, domain);}

  ///
  /** sets this IntVectSet to be empty */
  void
  makeEmpty();

  ///
  /** Chop the IntVectSet  at the chop_pnt in the dir direction \\
      Returns one IntVectSet and  modifies the object IntVectSet. \\
      The union of the two is the original IntVectSet. \\
      The modified IntVectSet is the low end, the returned IntvectSet \\
      is the high end. */
  IntVectSet
  chop(int dir, int chop_pnt);


  /// {\bf accessor and inquiry functions}

  ///
  /** Returns number of IntVects in this IntVectSet. */
  int
  numPts() const;

  ///
  /** Returns minimum enclosing box of this IntVectSet. */
  const Box&
  minBox() const;

  /// 
  /** Forces recalculation of actual minimum box for this IntVectSet */
  void recalcMinBox() const;

  ///
  /** Returns true if no cells are in this IntVectSet. */
  bool
  isEmpty() const;

  ///
  /** Returns true if this IntVectSet is currently being represented in a dense fashion */
  bool
  isDense() const;


  ///
  /** Does this IntVectSet contain the argument IntVect? */
  bool
  contains(const IntVect& iv) const;

  ///
  /** Does this IntVectSet contain the argument IntVectSet? */
  bool
  contains(const IntVectSet& iv) const;
  bool contains_intvectset(const IntVectSet& iv) const {return contains(iv);}

  ///
  /** is every IntVect in this 'box' contained in the IntvectSet.
      note: this is not performed in a crude point-wise fashion
      and is faster than a user doing it pointwise themselves */
  bool
  contains(const Box& box) const;
  bool contains_box(const Box& b) const {return contains(b);}

  ///
  /**
     Returns Vector<Box> representation of this IntVectSet.
  */
  Vector<Box>
  boxes() const;

  ///
  /**
     attempts to optimize storage for IntVectSet.  This will also
     optimize iteration through the IntVectSet later.  Best to call
     it when you are no longer modifying the IntVectSet
  */
  void compact() const;

  ///
  /**
     Write Vector<Box> representation to an output stream.
  */
  void
  printBoxes(std::ostream& a_ostrm) const;

  void p() const { printBoxes(pout());}

  friend
  std::ostream& operator<<(std::ostream& os, const IntVectSet& ivs);


  void convert() const; // turn dense rep into Tree rep.  very costly. 
                        // it is 'logically' const, but does modify data structures;

  // not for public consumption.  used in memory tracking.
  static long int count;
  static long int peakcount;
private:

  bool m_isdense;
  TreeIntVectSet m_ivs;
  DenseIntVectSet m_dense;

  // not a user function.  called by memory tracking system on
  // exit to clean up static allocation pools used for the optimization
  // of these routines.
  friend void dumpmemoryatexit();
  static void clearStaticMemory(); // after this functon is called, you cannot
                              // do any other operations with IntVectSet.
};





///Iterator for an IntVectSet
/**
   IVSIterator iterates over every point (IntVect)
   in an IntVectSet.  It has exactly the same
   syntax and sematic as BoxIterator.  Typical usage:\\

\begin{verbatim}
   IntVectSet ivs;
   ...
   IVSIterator ivsit (ivs); 
   for (ivsit.begin(); ivsit.ok(); ++ivsit)  
   {  
     IntVect iv = ivsit(); 
     (do operations involving iv) 
   } 
\end{verbatim}
*/
class IVSIterator
{
public:
  ///
  IVSIterator():m_isdense(true) {;}

  ///
  IVSIterator(const IntVectSet& ivs);

  ~IVSIterator(){;}

  ///
  void define(const IntVectSet& ivs);

  ///
  const IntVect& operator()() const ;

  const IntVect& iv() const { return this->operator()();}

  ///
  bool ok() const;

  ///
  void operator++();

  void incr() { ++(*this);}

  ///
  void begin();

  ///
  void reset();

  ///
  void end();
private:
  bool m_isdense;
  DenseIntVectSetIterator m_dense;
  TreeIntVectSetIterator  m_tree;
};



#ifndef WRAPPER



inline const IntVect& IVSIterator::operator()() const
{
  if(m_isdense) return m_dense();
  return m_tree();
}

inline bool IVSIterator::ok() const
{
  if(m_isdense) return m_dense.ok();
  return m_tree.ok();
}

inline void  IVSIterator::operator++()
{
  if(m_isdense) ++m_dense;
  else          ++m_tree;
}
inline void IVSIterator::reset()
{
  begin();
}

inline void IVSIterator::begin()
{
  if(m_isdense) m_dense.begin();
  else          m_tree.begin();
}

inline void IVSIterator::end()
{
  if(m_isdense) m_dense.end();
  else          m_tree.end();
}


inline IntVectSet::IntVectSet(): m_isdense(true) 
{
  count++;
  if(count > peakcount) peakcount = count;
}

inline void IntVectSet::define(const IntVectSet& ige_in)
{
  *this = ige_in;
}

inline IntVectSet::IntVectSet(const IntVectSet& ige_in)
{
  count++;
  if(count > peakcount) peakcount = count;

  *this = ige_in;
}

inline   bool
IntVectSet::isDense() const
{
  return m_isdense;
}

#endif /* WRAPPER */

#endif
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//


long int IntVectSet::count = 0;
long int IntVectSet::peakcount = 0;

IntVectSet::~IntVectSet(){count--;}

void IntVectSet::define()
{
  m_ivs.clear();
  m_dense = DenseIntVectSet();
  m_isdense = true;
}

void IntVectSet::define(const DenseIntVectSet& a_dense)
{
  m_ivs.clear();
  m_dense = a_dense;
  m_isdense = true;
}

void IntVectSet::define(const TreeIntVectSet& a_tree)
{
  m_ivs = a_tree;
  m_dense = DenseIntVectSet();
  m_isdense = false;
}

IntVectSet::IntVectSet(const DenseIntVectSet& a_dense)
{
  count++;
  define(a_dense);
}

IntVectSet::IntVectSet(const TreeIntVectSet& a_tree)
{
  count++;
  define(a_tree);
}

IntVectSet::IntVectSet(const IntVect& iv_in)
{
  count++;
  if(count > peakcount) peakcount = count;
  define(iv_in);
}

void IntVectSet::define(const IntVect& iv_in)
{
  define(Box(iv_in, iv_in));
}

IntVectSet::IntVectSet(const Box& b)
{
  count++;
  if(count > peakcount) peakcount = count;
  define(b);
}

void IntVectSet::define(const Box& b)
{
  if(b.numPts() < 10000)
    {
      m_ivs.clear();
      m_isdense = true;
      m_dense = DenseIntVectSet(b);
    }
  else
    {
      m_isdense = false;
      m_ivs.define(b);
    }
}

IntVectSet operator|(const IntVectSet& ivs1, const IntVectSet& ivs2)
{
  IntVectSet rtn = ivs1;
  rtn |= ivs2;
  return rtn;
}

IntVectSet operator|(const IntVectSet& ivs, const IntVect& iv)
{
  IntVectSet rtn = ivs;
  rtn |= iv;
  return rtn;
}

IntVectSet operator|(const IntVect& iv, const IntVectSet& ivs)
{
  IntVectSet rtn = ivs;
  rtn |= iv;
  return rtn;
}

IntVectSet operator|(const IntVectSet& ivs, const Box& b)
{
  IntVectSet rtn = ivs;
  rtn |= b;
  return rtn;
}

IntVectSet operator|(const Box& b, const IntVectSet& ivs)
{
  IntVectSet rtn = ivs;
  rtn |= b;
  return rtn;
}

IntVectSet& IntVectSet::operator|=(const IntVect& iv)
{

  if(m_isdense){
        if(!m_dense.box().contains(iv)){
          convert();
          m_ivs |= iv;
        } else {
          m_dense |= iv;
        }
  } else {
        m_ivs |= iv;
  }
  return *this;
}

IntVectSet& IntVectSet::operator|=(const Box& b)
{
  if(m_isdense){
        if(!m_dense.box().contains(b)){
          convert();
          m_ivs |= b;
        } else {
          m_dense |= b;
        }
  } else {
        m_ivs |= b;
  }
  return *this;

}


IntVectSet& IntVectSet::operator|=(const IntVectSet& ivs)
{
  if(m_isdense) convert();
  if(ivs.m_isdense) ivs.convert();
  m_ivs |= ivs.m_ivs;
  return *this;
}

IntVectSet IntVectSet::operator-(const IntVectSet& ivs) const
{
  IntVectSet rtn = *this;
  rtn -= ivs;
  return rtn;
}

IntVectSet IntVectSet::operator-(const IntVect& iv) const
{
  IntVectSet rtn = *this;
  rtn -= iv;
  return rtn;
}

IntVectSet IntVectSet::operator-(const Box& b) const
{
  IntVectSet rtn = *this;
  rtn -= b;
  return rtn;
}

IntVectSet& IntVectSet::operator-=(const IntVectSet& ivs)
{
  if(!(minBox().intersects(ivs.minBox()))) return *this;
  if(m_isdense)
    {
      if(ivs.m_isdense) m_dense-=ivs.m_dense;
      else
        {
          for(TreeIntVectSetIterator it(ivs.m_ivs); it.ok(); ++it) m_dense -= it();
        }
    }
  else
    {
      if(ivs.m_isdense)
        {
          for(DenseIntVectSetIterator it(ivs.m_dense);it.ok(); ++it) m_ivs -= it();
        }
      else
        m_ivs -= ivs.m_ivs;
    }

  return *this;
}

IntVectSet& IntVectSet::operator-=(const IntVect& iv)
{
  if(m_isdense) m_dense -= iv;
  else          m_ivs   -= iv;
  return *this;
}

IntVectSet& IntVectSet::operator-=(const Box& b)
{
  if(m_isdense) m_dense -= b;
  else          m_ivs   -= b;
  return *this;
}


IntVectSet operator&(const IntVectSet& ivs1, const IntVectSet& ivs2)
{
  IntVectSet rtn = ivs1;
  rtn &= ivs2;
  return rtn;
}



IntVectSet operator&(const IntVectSet& ivs, const Box& b)
{
  IntVectSet rtn = ivs;
  rtn &= b;
  return rtn;
}

IntVectSet operator&(const Box& b, const IntVectSet& ivs)
{
  IntVectSet rtn = ivs;
  rtn &= b;
  return rtn;
}

IntVectSet& IntVectSet::operator&=(const Box& b)
{
  if(m_isdense) m_dense &= b;
  else          m_ivs   &= b;
  return *this;
}

IntVectSet& IntVectSet::operator&=(const ProblemDomain& d)
{
  if(m_isdense) m_dense &= d;
  else          m_ivs   &= d;
  return *this;
}

IntVectSet& IntVectSet::operator&=(const IntVectSet& ivs)
{
  if(!(minBox().intersects(ivs.minBox())))
    {
      m_ivs.clear();
      m_isdense = true;
      m_dense = DenseIntVectSet();
      return *this;
    }
  if(m_isdense)
    {
      if(ivs.m_isdense) m_dense&=ivs.m_dense;
      else
        {
          convert();
          m_ivs &= ivs.m_ivs;
        }
    }
  else
    {
      if(ivs.m_isdense) ivs.convert();
        m_ivs &= ivs.m_ivs;
    }

  return *this;
}

IntVectSet grow(const IntVectSet& ivs, int igrow)
{
  IntVectSet rtn(ivs);
  rtn.grow(igrow);
  return rtn;
}

void IntVectSet::grow(int igrow)
{
  if(m_isdense) m_dense.grow(igrow);
  else          m_ivs.grow(igrow);
  //  return *this;
}

void IntVectSet::nestingRegion(int radius, const Box& domain)
{
  if(m_isdense) m_dense.nestingRegion(radius, domain);
  else          m_ivs.nestingRegion(radius, domain);
}

void IntVectSet::nestingRegion(int radius, const ProblemDomain& domain)
{
  if(m_isdense) m_dense.nestingRegion(radius, domain);
  else          m_ivs.nestingRegion(radius, domain);
}

IntVectSet& IntVectSet::grow(int idir, int igrow)
{
  assert(idir >= 0);
  assert(idir < SpaceDim);
  if(m_isdense) m_dense.grow(idir, igrow);
  else          m_ivs.grow(idir, igrow);
  return *this;
}

IntVectSet refine(const IntVectSet& ivs, int iref)
{
  IntVectSet rtn(ivs);
  rtn.refine(iref);
  return rtn;
}

IntVectSet& IntVectSet::refine(int iref)
{
  if(m_isdense) m_dense.refine(iref);
  else          m_ivs.refine(iref);
  return *this;
}

IntVectSet coarsen(const IntVectSet& ivs, int iref)
{
  IntVectSet rtn(ivs);
  rtn.coarsen(iref);
  return rtn;
}

IntVectSet& IntVectSet::coarsen(int iref)
{
  if(m_isdense) m_dense.coarsen(iref);
  else          m_ivs.coarsen(iref);
  return *this;
}

void IntVectSet::shift(const IntVect& iv)
{
  if(m_isdense) m_dense.shift(iv);
  else          m_ivs.shift(iv);
}

void IntVectSet::makeEmpty()
{
  if(m_isdense) m_dense = DenseIntVectSet();
  else          m_ivs.clear();
}

void IntVectSet::compact() const
{
  if(m_isdense) m_dense.compact();
  else          m_ivs.compact();
}

IntVectSet IntVectSet::chop(int dir, int chop_pnt)
{
  if(m_isdense) {
        DenseIntVectSet r = m_dense.chop(dir, chop_pnt);
        IntVectSet rtn(r);
        return rtn;
  }
  else {
        TreeIntVectSet t = m_ivs.chop(dir, chop_pnt);
        IntVectSet rtn(t);
        return rtn;
  }

}

const Box& IntVectSet::minBox() const
{
  if(m_isdense) return m_dense.mBox();
  m_ivs.recalcMinBox();
  return m_ivs.minBox();
}

void IntVectSet::recalcMinBox() const
{
   if(m_isdense)
	 m_dense.recalcMinBox();
   else
	 m_ivs.recalcMinBox();
}

bool IntVectSet::isEmpty() const
{
  if(m_isdense) return m_dense.isEmpty();
  return m_ivs.isEmpty();
}

int IntVectSet::numPts() const
{
  if(m_isdense) return m_dense.numPts();
  return m_ivs.numPts();
}

bool IntVectSet::contains(const IntVect& iv) const
{
  if(m_isdense) return m_dense[iv];
  return m_ivs.contains(iv);
}

bool IntVectSet::contains(const IntVectSet& ivs) const
{
  if(&ivs == this) return true;
  for(IVSIterator ivsit(ivs); ivsit.ok(); ++ivsit)
    {
      if(!contains(ivsit()))
        return false;
    }
  //return true for empty input
  return true;
}

bool IntVectSet::contains(const Box& box) const
{
  if(m_isdense) return m_dense.contains(box);
  return m_ivs.contains(box);
}

Vector<Box> IntVectSet::boxes() const
{
  if(m_isdense) return m_dense.createBoxes();
  return m_ivs.createBoxes();
}

void IntVectSet::printBoxes(std::ostream& a_ostrm) const
{
  Vector<Box> b = boxes();
  for(int i=0; i<b.size(); ++i) a_ostrm<<b[i]<<std::endl;
}

std::ostream& operator<<(std::ostream& os, const IntVectSet& ivs)
{
  ivs.printBoxes(os);
  return os;
}

void IntVectSet::convert() const
{
  if(!m_isdense) return; //already converted
  if(m_dense.isEmpty()) { } // do nothing
  else if(m_dense.isFull())
    {
      ((TreeIntVectSet&)m_ivs).define(m_dense.box());
    }
  else
    {
      DenseIntVectSetIterator it(m_dense);
      for(it.begin(); it.ok(); ++it)
        {
          ((TreeIntVectSet&)m_ivs) |= it();
        }
      m_ivs.compact();
    }
  (bool&)m_isdense = false;
}

// if you are in this function, you better really
// have returned all the TreeNodes to the pool before
// you call this function!!

void IntVectSet::clearStaticMemory()
{
  TreeIntVectSet::treeNodePool.clear();
  TreeIntVectSet::index.clear();
  TreeIntVectSet::parents.clear();
  TreeIntVectSet::boxes.clear();
  TreeIntVectSet::bufferOffset.clear();
}




//====================================================================

IVSIterator::IVSIterator(const IntVectSet& ivs)
{
  define(ivs);
}

void IVSIterator::define(const IntVectSet& ivs)
{
  if(ivs.m_isdense)
    {
      m_isdense = true;
      m_dense.define(ivs.m_dense);
    }
  else
    {
      m_isdense = false;
      m_tree.define(ivs.m_ivs);
    }
}
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

using std::cout;
using std::endl;

#ifndef MPI

int procID(){ return 0;}

// reset this to fool serial code into thinking its parallel
int num_procs = 1 ;

unsigned int numProc(){ return num_procs;}

#else // MPI version

int procID()
{
  int ret;
  MPI_Comm_rank(Chombo_MPI::comm, &ret);
  return ret;
}

unsigned int numProc()
{
  int ret;
  MPI_Comm_size(Chombo_MPI::comm, &ret);
  return ret;
}

// hopefully static copy of opaque handles
MPI_Comm Chombo_MPI::comm = MPI_COMM_WORLD;


#endif // MPI


/// these should work independent of MPI's existence
template < >
void linearIn<Real>(Real& a_outputT, const void* const a_inBuf)
{
  //this fandango is to avoid unaligned accesses
  char realBuf[sizeof(Real)];
  memcpy(realBuf, a_inBuf, sizeof(Real));
  Real* buffer = (Real*)realBuf;
  a_outputT = *buffer;
}

template < >
void linearOut<Real>(void* const a_outBuf, const Real& a_inputT)
{
  //this fandango is to avoid unaligned accesses
  char realBuf[sizeof(Real)];
  Real* realPtr = (Real*)realBuf;
  *realPtr = a_inputT;
  memcpy(a_outBuf, realBuf, sizeof(Real));
}

template < >
int linearSize<Real>(const Real& inputReal)
{
  return sizeof(Real);
}

template < >
int linearSize(const RealVect& vindex)
{
  return sizeof(RealVect);
}

//RealVect specialization of linearIn
template < >
void linearIn(RealVect& a_outputT, const void* const inBuf)
{
  unsigned char* bob = (unsigned char*)inBuf;
  unsigned char* to = (unsigned char*)&a_outputT;
  memcpy(to + RealVect::io_offset, bob, SpaceDim*sizeof(Real));
}

//RealVect specialization of linearOut
template < >
void linearOut(void* const a_outBuf, const RealVect& a_inputT)
{
  unsigned char* bob = (unsigned char*)a_outBuf;
  const unsigned char* from = (const unsigned char*)&a_inputT;
  memcpy(bob, from + RealVect::io_offset, SpaceDim*sizeof(Real));
}

template < >
void linearIn<int>(int& a_outputT, const void* const a_inBuf)
{
  int* buffer = (int*)a_inBuf;
  a_outputT = *buffer;
}

template < >
void linearOut<int>(void* const a_outBuf, const int& a_inputT)
{
  int* buffer = (int*)a_outBuf;
  *buffer = a_inputT;   
}

template < >
int linearSize<int>(const int& a_input)
{
  return sizeof(int);
}


template < >
void linearIn<Box>(Box& a_outputT, const void* const a_inBuf)
{
  int* intBuf = (int*)a_inBuf;
  IntVect lo, hi;
  //output is lo[0],lo[1],...
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      lo[idir] = intBuf[idir];
      hi[idir] = intBuf[idir+SpaceDim];
    }
  a_outputT = Box(lo,hi);
}

template < >
void linearOut<Box>(void* const a_outBuf, const Box& a_inputT)
{
  int* intBuf = (int*)a_outBuf;
  const IntVect& lo = a_inputT.smallEnd();
  const IntVect& hi = a_inputT.bigEnd();
  //output is lo[0],lo[1],...
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      intBuf[idir] = lo[idir];
      intBuf[idir+SpaceDim] = hi[idir];
    }
}

template < >
int linearSize<Box>(const Box& a_input)
{
  //box is stored as 2*spaceDim integers
  return(2*SpaceDim*sizeof(int));
}

//Vector<int>  specialization
template < > int linearSize(const Vector<int>& a_input)
{
  return linearListSize(a_input);
}
template < > void linearIn(Vector<int>& a_outputT, const void* const inBuf)
{
  linearListIn(a_outputT, inBuf);
}
template < > void linearOut(void* const a_outBuf, const Vector<int>& a_inputT)
{
  linearListOut(a_outBuf, a_inputT);
}

//Vector<Real>  specialization
template < > int linearSize(const Vector<Real>& a_input)
{
  return linearListSize(a_input);
}
template < > void linearIn(Vector<Real>& a_outputT, const void* const inBuf)
{
  linearListIn(a_outputT, inBuf);
}
template < > void linearOut(void* const a_outBuf, const Vector<Real>& a_inputT)
{
  linearListOut(a_outBuf, a_inputT);
}

//Vector<Box>  specialization
template < > int linearSize(const Vector<Box>& a_input)
{
  return linearListSize(a_input);
}
template < > void linearIn(Vector<Box>& a_outputT, const void* const inBuf)
{
  linearListIn(a_outputT, inBuf);
}
template < > void linearOut(void* const a_outBuf, const Vector<Box>& a_inputT)
{
  linearListOut(a_outBuf, a_inputT);
}

//Vector<Vector<Box> >  specialization
template < > int linearSize(const Vector<Vector<Box> >& a_input)
{
  return linearListSize(a_input);
}
template < > void linearIn(Vector<Vector<Box> >& a_outputT, const void* const inBuf)
{
  linearListIn(a_outputT, inBuf);
}
template < > void linearOut(void* const a_outBuf, const Vector<Vector<Box> >& a_inputT)
{
  linearListOut(a_outBuf, a_inputT);
}

//Vector<Vector<int> >  specialization
template < > int linearSize(const Vector<Vector<int> >& a_input)
{
  return linearListSize(a_input);
}
template < > void linearIn(Vector<Vector<int> >& a_outputT, const void* const inBuf)
{
  linearListIn(a_outputT, inBuf);
}
template < > void linearOut(void* const a_outBuf, const Vector<Vector<int> >& a_inputT)
{
  linearListOut(a_outBuf, a_inputT);
}
template < >
void gather<IntVectSet>(Vector<IntVectSet>& a_outVec, 
            const IntVectSet& a_input, 
            int a_dest)
{
  Vector<Box> boxes = a_input.boxes();
  Vector<Vector<Box> > all_boxes;

  gather (all_boxes, boxes, a_dest);

  const int num_vecs = all_boxes.size();
  a_outVec.resize(num_vecs);

  for (int i = 0; i < num_vecs; ++i)
  {
    IntVectSet& ivs = a_outVec[i];
    const Vector<Box>& vb = all_boxes[i];
    for (int ibox = 0; ibox < vb.size(); ++ibox)
    {
      ivs |= vb[ibox];
    }
  }
}


// return id of unique processor for special serial tasks
int
uniqueProc(const SerialTask::task& a_task)
{
#ifdef NDEBUG
    switch (a_task)
    {
    case SerialTask::compute:
    default:
        return (0);
        //break;  // unreachable break can generate compiler warning
    }
#else
// in mpi, the debugger attaches to process 0
    return (0);
#endif
}

/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef DATAINDEX_H
#define DATAINDEX_H

#ifndef WRAPPER
class BoxLayout;


class DataIterator;
class LayoutIterator;

#endif /* WRAPPER */

///An index for LayoutIterator
class LayoutIndex
{
public:

  ///
  LayoutIndex();

  ~LayoutIndex() {;}
  // default copy and assign shoule be fine.

  ///
  bool operator == (const LayoutIndex& rhs) const;
  bool eq(const LayoutIndex& rhs) const { return *this == rhs;}

  ///
  bool operator != (const LayoutIndex& rhs) const
  { return !(*this == rhs);}

  /// returns 'true' if this DataIndex has been null constructed.
  bool isNull() const;

  // not user function. breaks design encapsulation. method not
  // intended for users. The alternative is to make this class a friend
  // of BoxLayoutData<T>, which implicitly makes this class templated, which
  // makes no bloody sense.  The behaviour of this method is undefined for
  // users.  bvs
  int intCode() const;

private:
  friend class DataIterator;
  friend class LayoutIterator;
  friend class BoxLayout;
  friend class DataIndex;

  int m_index;
  const int* m_layoutIntPtr;
  LayoutIndex(int a_index, const int* layoutID)
    : m_index(a_index), m_layoutIntPtr(layoutID){}
};


class DataIndex : public LayoutIndex
{
public:
  explicit DataIndex(const LayoutIndex& promotion)
    :LayoutIndex(promotion){}
  DataIndex()
    : LayoutIndex(){;}
  ~DataIndex(){;}
private:
  friend class LayoutIterator;
  friend class DataIterator;
  friend class BoxLayout;
  friend class Copier;
  DataIndex(int a_index, const int* layoutID)
    :LayoutIndex(a_index, layoutID) {}
};

#ifndef WRAPPER
inline LayoutIndex::LayoutIndex()
  : m_index(0), m_layoutIntPtr(NULL){}

inline bool LayoutIndex::operator==(const LayoutIndex& rhs) const
{
  return (m_index == rhs.m_index);
}

inline int LayoutIndex::intCode() const
{
  return m_index;
}

inline bool LayoutIndex::isNull() const
{
  return m_layoutIntPtr == NULL;
}
#endif

#endif // DATAITERATOR_H
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef BOXLAYOUT_H
#define BOXLAYOUT_H

#ifndef WRAPPER
class DataIterator;
class LayoutIterator;
#endif /*WRAPPER*/

// This header file defines the BoxLayout oject.

///A not-necessarily disjoint collective of boxes
/**
  A BoxLayout is a collection of Box objects that are assigned
  to a process number.  Each box is associated with only one
  process.  processes are numbered from 0 to n-1 (for an n process
  parallel job)

  BoxLayout can be either open or closed.  null construction or deepcopy
  creates an open BoxLayout.  Boxes may be added to an open BoxLayout. 
  non-const operations may be performed on the boxes existing in an
  open BoxLayout.  closed BoxLayouts are sorted.  Many uses of BoxLayouts
  require a closed BoxLayout.  A closed BoxLayout cannot be modified.

  Ref-counting
  ------------
  BoxLayout is an explicitly ref-counted object.

  Assignment and copy are compiler generated.  They increment the refcount
  on the contained data members.  They perform shallow, ref-counted
  operations.

  Refcounting is a process whereby multiple instaniations make
  use of a single implementation of that object and keep a tally
  of how many instantiations are sharing.  Thus

  \begin{verbatim}
  BoxLayout b1(boxes, procIDs);

            b1  ----> refcount = 1
                ---->  m_boxes
                ---->  m_processors

  BoxLayout b2(b1)

            b1  ----> refcount = 2  <---- b2
                ---->  m_boxes      <----
                ---->  m_processors <----

  BoxLayout b3;
  b3 = b2;

            b1  ----> refcount = 3  <---- b2
                ---->  m_boxes      <----
                ---->  m_processors <----
                        ^^^
                        |||
                         b3
  \end{verbatim}
*/

struct Entry
{
  Entry(): m_procID(procID()){;}
  Entry(const Box& a_box)
	:box(a_box), m_procID(procID()){}
  Entry(const Box& a_box, const unsigned int a_index)
	:box(a_box), index(a_index), m_procID(procID()){}
  bool operator < (const Entry& rhs) const
  { return box < rhs.box;}
  
  Box box;
  unsigned int index;
  unsigned int m_procID;// not used in serial code.
};

class BoxLayout
{
public:
  ///
  BoxLayout();

  ///
  /** Construct from a Vector of Boxes and  a Vector of
      processor assignments.  On exit, the BoxLayout will be closed.
  */
  BoxLayout(const Vector<Box>& a_boxes,
            const Vector<int>& a_procIDs);

  ///
  /**
    Ref-counted destruction.  Once the last reference to the 
    implementation of this class is destroyed, the data members
    are cleaned up
  */
  virtual
  ~BoxLayout();

  /// refcounted assignment
  /**
   */
  BoxLayout& operator=(const BoxLayout& a_rhs);

  ///
  /** Define this BoxLayout from a Vector of Boxes and a 
      Vector of processor assignments.  Any pre-existing layout will
      be lost (ref-count dropped by one).  The processor assignment Vector
      must be the same length
      as the Box Vector.  On exit, the BoxLayout will be closed.  
  */
  virtual void
  define(const Vector<Box>& a_boxes,
         const Vector<int>& a_procIDs);


  ///
  /** const accessor operator.  see also {\tt get(const DataIndex&)}
   */
  const Box&
  operator[](const LayoutIndex& it) const;

  ///
  Box&
  operator[](const LayoutIndex& it);

  ///
  /**  Refcounted pointer check.  True if these two objects
    share the same implementation.
    */
  inline bool
  operator==(const BoxLayout& rhs) const ;
  bool eq(const BoxLayout& rhs) const { return *this == rhs;}


  // not a user function
  bool check(const LayoutIndex& index) const
    { return index.m_layoutIntPtr == m_layout;}


  ///
  /** add a box to this BoxLayout. Takes the processor
    assignment. The input 'box' is copied.  BoxLayout must be 'open'
    for this operation, otherwise a runtime error will occur.
      */
  DataIndex
  addBox(const Box& box, int procID);

  ///
  /** Mark this BoxLayout as complete and unchangeable.
   */
  virtual void
  close();

  ///
  /** Actual deep copy operation.  New object created with copied
      data.  This object disassociates itself with original implementation
      safely.  This object now is considered 'open' and can be non-const
      modified.  There is no assurance that the order in which this BoxLayout
      is indexed corresponds to the indexing of a_source.

       BoxLayout b1(boxes, procIDs);
\begin{verbatim}
            b1  ----> refcount = 1
                ---->  m_boxes
                ---->  m_processors

       BoxLayout b2(b1)

            b1  ----> refcount = 2  <---- b2
                ---->  m_boxes      <----
                ---->  m_processors <----

       BoxLayout b3;
       b3.deepCopy(b2);
            b1  ----> refcount = 2  <---- b2  b3 ----> refcount = 1 
                ---->  m_boxes      <----        ---->  m_boxes
                ---->  m_processors <----        ---->  m_processors
\end{verbatim}
  */ 
  virtual void
  deepCopy(const BoxLayout& a_source);

  ///
  /** As a consquence of the C++ compiler being free to choose which
      version of operator[] when the object is technically non-const, we very
      often get 'BoxLayout closed' errors.  This is a non-overloaded get method.
  */
  Box get(const LayoutIndex& it) const;


  /** return true if close() has been called.  closed BoxLayout is always sorted.
   */
  bool
  isClosed() const;


  ///
  /** return true if sort() has been called
   */
  bool
  isSorted() const {return isClosed();}

  ///
  /* parallel iterator.  Returns DataIndex objects that that only correspond to boxes
     on the local processor */
  DataIterator
  dataIterator() const;

  ///
  /** If BoxLayout is closed, then LayoutIterator will return the Box's in sorted
    order.  Processes through ALL the boxes in a BoxLayout */
  LayoutIterator
  layoutIterator() const;

  // invokes cout<<*this;  pretty-print dump of BoxLayout
#ifndef WRAPPER
  void
  print() const;
#endif
  void p() const { print();}

  ///
  /** Returns the processor to which this box has been assigned.
      Not a user function, at least, not a new user function.  It can
      be used safely at anytime, closed or open.  A person needing this
      level of knowledge of the processor assignment should have non-trivial
      needs, like writing your own load balancer or such.  Most user level parallel
      coding needs can be addressed using the DataIterator class.
  */
  unsigned int
  procID(const LayoutIndex& a_index) const ;

  ///
  /* returns the number of boxes assigned to a given procID
   */
  int numBoxes(const int procID) const;

  ///
  /** Assign a Box in the BoxLayout to a processor.
      Requires the BoxLayout to be open.
  */
  void
  setProcID(const LayoutIndex& a_index, unsigned int a_procID);

  ///
  inline unsigned int
  size() const ;


  //const or non-const operation ?.....I can think of usages either way...bvs
  ///
  /**
    Sort the boxes.  
    */
  void
  sort();

  ///
  /**  'output' must be open, 'input' must be closed, 'refinement' must be
    a positive non-zero integer.  'output' and 'input' do not share an
    implementation.  'output' is first deepcopy'ed from 'input', 
    then coarsen(refinement) is called on each box of 'output'.  'output'
    returns from this function closed.  LayoutIterators from 'input' and 'output'
    can perform interchangeably. */
  friend void coarsen(BoxLayout& output, const BoxLayout& input, int refinement);

  ///
  /** see 'coarsen(BoxLayout& output, const BoxLayout& input, int refinement) '
    and substitute the word "refine" for "coarsen" */
  friend void refine(BoxLayout& output, const BoxLayout& input, int refinement);


  // not a user function.  used in IO routine.
  unsigned int index(const LayoutIndex& index) const;

  //not user functions, used in Chombo interface to AMRData.h interoperability tools
  void aliasAddBox(const Box& box);
  void aliasClose();

protected:

  void setIndexVector();
  friend class LayoutIterator;
  friend class DataIterator;

  RefCountedPtr<Vector<Entry> >        m_boxes;
  RefCountedPtr<Vector<unsigned int> > m_index;
  RefCountedPtr<int>                   m_layout;
  RefCountedPtr<bool>                  m_closed;
  RefCountedPtr<DataIterator>          m_dataIterator;

private:

};

void coarsen_bl(BoxLayout& output, const BoxLayout& input, int refinement);

void refine_bl(BoxLayout& output, const BoxLayout& input, int refinement);

#ifndef WRAPPER

inline
void coarsen_bl(BoxLayout& output, const BoxLayout& input, int refinement)
{ coarsen(output, input, refinement);}

inline
void refine_bl(BoxLayout& output, const BoxLayout& input, int refinement)
{ refine(output, input, refinement);}

//========================================================


// inlined functions
// =================


// constructors
// ============



/*
inline
BoxLayout::BoxLayout(int a_boxes)
  :m_boxes(new std::vector<Entry>(a_boxes)),
   m_index(new std::vector<unsigned int>(a_boxes)),
   m_layout(new int),
   m_closed(new bool(false))
{
  std::vector<Entry>& boxes = *m_boxes;
  std::vector<unsigned int>& index = *m_index;
  for(unsigned int i=0; i<boxes.size(); ++i)
    {
      boxes[i].index = i;
      index[i] = i;
    }
}
*/

// operators
// =========

inline const Box&
BoxLayout::operator[](const LayoutIndex& index) const
{
  assert(check(index));// make sure this LayoutIndex came from my own iterator
  return m_boxes->operator[](m_index->operator[](index.m_index)).box;
}

inline Box&
BoxLayout::operator[](const LayoutIndex& index)
{
  if(*m_closed)
    {
      MayDay::Error("attempt to modify closed BoxLayout");
    }
  assert(check(index)); // make sure this LayoutIndex came from my own iterator
  return m_boxes->operator[](m_index->operator[](index.m_index)).box;
}

inline bool
BoxLayout::operator==(const BoxLayout& rhs) const
{
  return m_boxes == rhs.m_boxes;
}


// member functions
// ================


inline  Box
BoxLayout::get(const LayoutIndex& index) const
{
  assert(check(index)); // make sure this LayoutIndex came from my own iterator
  return m_boxes->operator[](m_index->operator[](index.m_index)).box;
}

inline unsigned int
BoxLayout::index(const LayoutIndex& a_index) const
{
  return m_index->operator[](a_index.m_index);
}

inline bool
BoxLayout::isClosed() const
{
  return *m_closed;
}

inline unsigned int
BoxLayout::procID(const LayoutIndex& a_index) const
{
  assert(check(a_index));
  return m_boxes->operator[](m_index->operator[](a_index.m_index)).m_procID;
}

inline void
BoxLayout::setProcID(const LayoutIndex& a_index, unsigned int a_procID)
{
  assert(check(a_index));
  m_boxes->operator[](m_index->operator[](a_index.m_index)).m_procID = a_procID;
}

inline unsigned int
BoxLayout::size() const 
{
  return m_boxes->size();
}


// global functions
// ================

std::ostream& operator<<(std::ostream& os, const BoxLayout& a_layout);

#endif /*WRAPPER*/


#endif // DATAPLAN_H

/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
#ifndef LAYOUTITERATOR_H
#define LAYOUTITERATOR_H

#ifndef WRAPPER


#endif

///An Iterator based on a BoxLayout object. 
/**
   An Iterator based on a BoxLayout object.  It does not
   support a dereferencing operation(1), since it is intended
   to work with all of BoxLayouts, DisjointBoxLayouts, BoxLayoutDatas
   LevelData's, and any object that is built on top
   of a BoxLayout object.  LayoutIterator access the data in a BoxLayout-based
   object in a NON-data-parallel manner.

   BoxLayout-based objects act as the Factory for the LayoutIterator. 

 (1) STL-speak.  not critical for comprehension, but can help people
 familiar with STL iterators and expecting similar behaviour.

*/
class LayoutIterator
{
public:
  // default copy and null constructor should be fine
  // not useful to someone using the iterator.  Used by
  // other classes in Chombo.  Could make it private and
  // let BoxLayout have access....
  ~LayoutIterator() {;}
  ///
  const LayoutIndex& operator()() const;
  LayoutIndex i() const { return this ->operator()();}

  ///
  void operator++();
  void incr(){ ++(*this);}

  ///
  bool ok() const;

  ///
  void reset();

  ///
  void begin();

  ///
  /**
     skip the iterator to the end.
  */
  void end();



private:

  friend class BoxLayout;

  LayoutIterator(const BoxLayout& boxlayout, const int* layoutID);

  //int m_begin, m_end;

  const BoxLayout m_layout;

  unsigned int m_index;

  LayoutIndex m_current;

};


#ifndef WRAPPER

inline const LayoutIndex& LayoutIterator::operator()() const
{
  assert(ok());
  return m_current;
}

inline void LayoutIterator::operator++()
{
  ++m_index;
  if(ok())
    m_current.m_index = (*(m_layout.m_boxes))[m_index].index;
}

inline bool LayoutIterator::ok() const
{
  return m_index < m_layout.size();
}


inline void LayoutIterator::begin()
{
  m_index = 0;
  if(ok())
    m_current.m_index = (*(m_layout.m_boxes))[m_index].index;
}
#endif /*WRAPPER*/

#endif 
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef DATAITERATOR_H
#define DATAITERATOR_H

#ifndef WRAPPER
#endif /* WRAPPER */

///An Iterator based on a BoxLayout object. 
/**
   An Iterator based on a BoxLayout object.  It does not
   support a dereferencing operation, since it is intended
   to work with all of BoxLayouts, DisjointBoxLayouts, BoxLayoutDatas
   LevelDatas, and any object that is built on top
   of a BoxLayout object.  DataIterator access the data in a BoxLayout-based
   object in a data-parallel manner. Order is not defined.


*/


class DataIterator
{
public:
  // default copy constructor should be fine

  /// a null constructed DataIterator will return false on ok()
  DataIterator();

  ~DataIterator() {;}
  ///
  inline const DataIndex& operator()() const;
  DataIndex i() const { return this->operator()();}

  ///
  inline void operator++();
  void incr() { ++(*this);}

  ///
  bool ok() const;

  ///
  void reset();

  ///
  void begin();

  ///
  /**
     skip the iterator to the end. ok() will return false after this
     method is called.
  */
  void end();

private:

  friend class BoxLayout;
  friend class DisjointBoxLayout;

  DataIterator(const BoxLayout& boxlayout, const int* layoutID);

  //int m_begin, m_end;

  const BoxLayout m_layout;

  unsigned int m_index;

  DataIndex m_current;

  unsigned int m_procID;

};

#ifndef WRAPPER

inline DataIterator::DataIterator()
  : m_index(0), m_procID(0)
{}


inline const DataIndex& DataIterator::operator()() const
{
  assert(ok());
  return m_current;
}

inline void DataIterator::operator++()
{
  const Entry* box;
  while(++m_index < m_layout.size())
    {
      box = &(*(m_layout.m_boxes))[m_index];
      if(box->m_procID == m_procID)
        {
          m_current.m_index = box->index;
          return;
        }
    }

}

inline bool DataIterator::ok() const
{
  return m_index < m_layout.size();
}



inline void DataIterator::reset()
{
  begin();
}

inline void DataIterator::begin()
{
  m_index = 0; 
  const Entry* box;
  while(m_index < m_layout.size())
    {
      box = &(*(m_layout.m_boxes))[m_index];
      if(box->m_procID == m_procID)
        {
          m_current.m_index = box->index;
          return;
        }
      ++m_index;
    }
}
#endif

#endif 
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//


using std::ostream;

//need at least one non-inlined function, otherwise
//some compilers don't build a class description in the
//object file.

BoxLayout::~BoxLayout()
{}

BoxLayout::BoxLayout()
  :m_boxes(new Vector<Entry>()),
   m_index(new Vector<unsigned int>()),

   m_layout(new int),
   m_closed(new bool(false)),
   m_dataIterator(NULL)
{}

BoxLayout& BoxLayout::operator=(const BoxLayout& a_rhs)
{
  if(this == &a_rhs) return *this;
  m_boxes = a_rhs.m_boxes;
  m_index = a_rhs.m_index;
  m_layout = a_rhs.m_layout;
  m_closed = a_rhs.m_closed;
  m_dataIterator = a_rhs.m_dataIterator;
  return *this;
}


void BoxLayout::sort()
{
  if(!*m_closed)
    {
      //      std::sort(m_boxes->begin(), m_boxes->end());
      m_boxes->sort();
      setIndexVector();
    }
}

void BoxLayout::close()
{
  //  assert(m_index->size() != 0);
  if(!*m_closed){
    sort();
    *m_closed = true;
    m_dataIterator = new DataIterator(*this, m_layout);
  }
}

void BoxLayout::setIndexVector()
{
  Vector<Entry>& boxes = *m_boxes;
  Vector<unsigned int>& index = *m_index;
  index.resize(boxes.size());
  for(unsigned int i=0; i<boxes.size(); ++i)
    {
      index[boxes[i].index] = i;
    }
}

// Constructors and such
// =====================

DataIterator BoxLayout::dataIterator() const
{
  assert(*m_closed);
  //DataIterator rtn(*this, m_layout);
  return *m_dataIterator;
  //return rtn;
}

LayoutIterator BoxLayout::layoutIterator() const
{
  return LayoutIterator(*this, m_layout);
}


BoxLayout::BoxLayout(const Vector<Box>& a_boxes, const Vector<int>& assignments)
  :m_boxes(new Vector<Entry>()),
   m_index(new Vector<unsigned int>()),
   m_layout(new int),
   m_closed(new bool(false))
{
  define(a_boxes, assignments);
}

void
BoxLayout::define(const Vector<Box>& a_boxes, const Vector<int>& a_procIDs)
{
  if(*m_closed)
    {
      MayDay::Error("attempt to define(..) a closed BoxLayout");
    }
  const int num_boxes = a_boxes.size();
  const int num_procs = a_procIDs.size();
  if( num_procs > 0 && num_boxes != num_procs )
    {
      MayDay::Error("BoxLayout::define(): vector of processor assignments is different length from vector of boxes");
    }
  m_boxes->resize(num_boxes);
  for (unsigned int i = 0; i < num_boxes; ++i)
    {
      m_boxes->operator[](i) = a_boxes[i];
      m_boxes->operator[](i).index = i;
      if( num_procs > 0 ) m_boxes->operator[](i).m_procID = a_procIDs[i];
    }
  m_index->resize(1); //signal for close routine.
  close();
}


// Other member functions
// ======================

DataIndex
BoxLayout::addBox(const Box& box, int procID)
{
  if(*m_closed)
    {
      MayDay::Error("attempt to addBox to closed BoxLayout");
    }
  if(m_layout.isNonUnique())
    {
      m_layout = new int;
    }
  Vector<Entry>& boxes = *m_boxes;
  unsigned int i = boxes.size();
  boxes.push_back(box);
  boxes[i].index = i;
  boxes[i].m_procID = procID;
  (*m_index).push_back(i);
  return DataIndex(i, m_layout);
}

void BoxLayout::aliasAddBox(const Box& box)
{
  if(*m_closed)
    {
      MayDay::Error("attempt to addBox to closed BoxLayout");
    }
  if(m_layout.isNonUnique())
    {
      m_layout = new int;
    }
  m_boxes->push_back(box);
  assert(m_index->size() == 0);
}


void
BoxLayout::deepCopy(const BoxLayout& a_source)
{
  m_boxes =  new Vector<Entry>(*(a_source.m_boxes));
  m_index =  new Vector<unsigned int>(*(a_source.m_index));
  m_layout = a_source.m_layout;
  *m_closed = false;
}


/*
void
BoxLayout::resize(unsigned int n, const Box& box) 
{
  if(*m_closed)
    {
      MayDay::Error("attempt to resize closed BoxLayout");
    }
  if(n == m_boxes->size()) return;
  if(m_layout.isNonUnique())
    {
      m_layout = new int;
    }
  unsigned int size = m_boxes->size();
  m_boxes->resize(n, box);
  m_index->resize(n);
  if(n > size)
    {
      for(unsigned int i=size; i<n; ++i)
        m_boxes->operator[](i).index = i;
    }
  setIndexVector();
}

*/

void BoxLayout::aliasClose() 
{

  assert(m_index->size()==0); //check to make sure the other addBox function was never called

  Vector<Vector<Box> > allBoxes;
  Vector<Box> vbox(m_boxes->size());

  for(int i=0; i<m_boxes->size(); i++)
	{
	  vbox[i] = m_boxes->operator[](i).box;
	}

  gather(allBoxes, vbox, 0);
  broadcast(allBoxes, 0);
  vbox.resize(0);
  Vector<int> procID;
  for(int p=0; p<allBoxes.size(); p++)
	{
	  Vector<Box>& b = allBoxes[p];
	  for(int i=0; i<b.size(); i++)
		{
		  procID.push_back(p);
		  vbox.push_back(b[i]);
		}
	}
  
  m_boxes->resize(vbox.size());
  for (unsigned int i = 0; i < vbox.size(); ++i)
    {
      m_boxes->operator[](i) = vbox[i];
      m_boxes->operator[](i).index = i;
      m_boxes->operator[](i).m_procID = procID[i];
    }
  setIndexVector();
  *m_closed = true;
  m_dataIterator = new DataIterator(*this, m_layout);

}

// Global functions
// ================

// For now, we can just have the one coarsen funtion.  If a DisjointBoxLayout
// enters this function, is coarsened, and then doesn't remain disjoint, it
// will be caught here at the call to close().  Debugging should not be

void
coarsen(BoxLayout& a_output, const BoxLayout& a_input, int a_refinement)
{
   if(!a_input.isClosed())
    {
      MayDay::Error("input to coarsen must be called with closed BoxLayout");
    }
  if(a_output.isClosed())
    {
      MayDay::Error("output of coarsen must be called on open BoxLayout");
    }
  a_output.deepCopy(a_input);


  for(LayoutIterator it(a_input.layoutIterator()); it.ok(); ++it)
    {
      a_output[it()].coarsen(a_refinement);
    }
  a_output.close();
}

// we have an easier time with refine, since we know that refinement will
// not change the state of a sort, but, we will play it safe for now
// until this function shows up in the profiler.

void refine(BoxLayout& a_output, const BoxLayout& a_input, int a_refinement)
{
  if(!a_input.isClosed())
    {
      MayDay::Error("input to refine must be called with closed BoxLayout");
    }
  if(a_output.isClosed())
    {
      MayDay::Error("output of refine must be called on open BoxLayout");
    }
  a_output.deepCopy(a_input);

  for(LayoutIterator it(a_input.layoutIterator()); it.ok(); ++it)
    {
      a_output[it()].refine(a_refinement);
    }
  a_output.close();
}

ostream& operator<<(ostream& os, const BoxLayout& a_layout)
{
  int i=0;
  for(LayoutIterator it(a_layout.layoutIterator()); it.ok(); ++it)
    {
      os << a_layout.get(it())<<"["<<a_layout.procID(it())<<"]";
      ++i;
      if(i==4){ os <<"\n"; i=0;}
      else {os <<" # ";}
    }

  os <<"\n";
  return os;
}


void BoxLayout::print() const { std::cout << *this;}

int BoxLayout::numBoxes(const int procID) const
{
  int num = 0;
  for(int i=0; i<m_boxes->size(); ++i)
    {
      if(m_boxes->operator[](i).m_procID == procID) ++num;
    }
  return num;
}


/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef DISJOINTBOXLAYOUT_H
#define DISJOINTBOXLAYOUT_H

#ifndef WRAPPER
#endif /*WRAPPER*/

///A BoxLayout that has a concept of disjointedness
/**
   DisjointBoxLayout is-a BoxLayout that has a concept of disjointedness
   of the boxes.  DisjointBoxLayout is only different from BoxLayout in
   that it has a method isDisjoint().  DisjointBoxLayouts also have
   a ProblemDomain member to ensure disjointness in the case of periodic 
   boundary conditions, among other things.
*/

class DisjointBoxLayout: public BoxLayout
{

public:

  friend class Copier;

  // Constructors, destuctors, defines
  // =================================

  ///
  DisjointBoxLayout();

  ///
  /** Construct from a Vector of Boxes and optionally a Vector of
      processor assignments.  If the processor assignment Vector
      is present, it must be either zero-length or the same length
      as the Box Vector.  On exit, the DisjointBoxLayout will be closed.
  */
  DisjointBoxLayout(const Vector<Box>& a_boxes,
                    const Vector<int>& a_procIDs);

  ///
  /** Construct from a Vector of Boxes and optionally a Vector of
      processor assignments.  If the processor assignment Vector
      is present, it must be either zero-length or the same length
      as the Box Vector.  On exit, the DisjointBoxLayout will be closed.
  */
  DisjointBoxLayout(const Vector<Box>& a_boxes,
                    const Vector<int>& a_procIDs,
                    const ProblemDomain& a_physDomain);

  ///
  virtual
  ~DisjointBoxLayout(){;}

  ///
  /** Define this DisjointBoxLayout from a Vector of Boxes.  This is a
      wrapper for a call to BoxLayout::define, required because
      DisjointBoxLayout::define is overloaded and the compiler will not
      look into the base class for name resolution.
  */
  virtual void
  define(const Vector<Box>& a_boxes,
         const Vector<int>& a_procIDs);
  ///
  /** Shallow define. Only way to promote a BoxLayout.  If BoxLayout
      has been closed, then this method checks isDisjoint and throws an
      error if not disjoint.  If a_layout is disjoint, then this object
      becomes a closed DisjointBoxLayout object
  */
  void
  define(BoxLayout& a_layout);

  ///
  /** Shallow define. Only way to promote a BoxLayout.  If BoxLayout
      has been closed, then this method checks isDisjoint and throws an
      error if not disjoint.  If a_layout is disjoint, then this object
      becomes a closed DisjointBoxLayout object
  */
  void
  define(BoxLayout& a_layout, const ProblemDomain& a_physDomain);




  ///
  /** Define this DisjointBoxLayout from a Vector of Boxes.  This is a
      wrapper for a call to BoxLayout::define, required because
      DisjointBoxLayout::define is overloaded and the compiler will not
      look into the base class for name resolution.
  */
  virtual void
  define(const Vector<Box>& a_boxes,
         const Vector<int>& a_procIDs,
         const ProblemDomain& a_physDomain);
  virtual void
  define_pd(const Vector<Box>& a_boxes,
            const Vector<int>& a_procIDs,
			const ProblemDomain& a_physDomain) {define(a_boxes, a_procIDs,  a_physDomain);}



  // other member functions
  // ======================

  ///
  /** Mark this DisjointBoxLayout as complete and unchangeable.
   */
  virtual void
  close();

  ///
  /** Doesn't check for disjointness, since a_source is already
      guaranteed to be disjoint.  Also copies problem domain 
      info */
  virtual void
  deepCopy(const DisjointBoxLayout& a_source);

  ///
  /** checks that the source BoxLayout isDisjoint, throws an error
      if it is not. Otherwise, same as BoxLayout::deepCopy */
  virtual void
  deepCopy(const BoxLayout& a_source);


  ///
  /** checks that the source BoxLayout isDisjoint, throws an error
      if it is not. Otherwise, same as BoxLayout::deepCopy */
  virtual void
  deepCopy(const BoxLayout& a_source, const ProblemDomain& a_physDomain);

  ///
  /** Returns true if this object contains a disjoint
      union of Boxes. 
       The Disjoint testing algorithm assumes that the
      boxes are CELL-Centered boxes.  We can expand the algorithm
      when someone needs it.  bvs
  */
  bool
  isDisjoint() const;

  ///
  /** Checks to see that problem domains are compatible
      To be compatible, a_domain and the ProblemDomain in 
      this DisjointBoxLayout must have the same periodicity.
      They must be periodic in the same dimensions, and in
      all periodic directions, the period must be the same.
      There are no checks in non-periodic directions, since 
      DisjointBoxLayouts don't care about physical domains
      except in the periodic case.
   */
  bool checkPeriodic(const ProblemDomain& a_domain) const;

  /** performs problem domain check when we don't have direct
      access to a problemDomain -- returns true if this and
      a_dbl have compatible ProblemDomains
  */
  bool checkDomains(const DisjointBoxLayout& a_dbl) const;

  /// 
  /**  'output' must be open, 'input' must be closed, 'refinement' must be
    a positive non-zero integer.  'output' and 'input' do not share an
    implementation.  'output' is first deepcopy'ed from 'input', 
    then coarsen(refinement) is called on each box of 'output'.  'output'
    returns from this function closed.  LayoutIterators from 'input' and 
    'output' can perform interchangeably. */
  friend void coarsen(DisjointBoxLayout& output, 
                      const DisjointBoxLayout& input, 
                      int refinement);

  ///
  /** see 'coarsen(BoxLayout& output, const BoxLayout& input, int refinement) '
    and substitute the word "refine" for "coarsen" */
  friend void refine(DisjointBoxLayout& output, 
                     const DisjointBoxLayout& input, 
                     int refinement);


  ///
  /** returns the set of boxes which you get from calling Box::adjCellLo
      on each box in input. output must be open, is closed upon exit from 
      this function.
  */
  friend void adjCellLo(DisjointBoxLayout& a_output,
                        const DisjointBoxLayout& a_input,
                        int a_dir, int a_len=1);


  ///
  /** returns the set of boxes which result from calling Box::adjCellHi
      on each box in input.  output must be open, will be closed upon
      exiting this function.
  */
  friend void adjCellHi(DisjointBoxLayout& a_output,
                        const DisjointBoxLayout& a_input,
                        int a_dir, int a_len=1);
  

protected:
  const ProblemDomain& physDomain() const;



private:
  ///
  ProblemDomain m_physDomain;

};
void adjCellLo_dbl(DisjointBoxLayout& a_output,
				   const DisjointBoxLayout& a_input,
				   int a_dir, int a_len);
void adjCellHi_dbl(DisjointBoxLayout& a_output,
				   const DisjointBoxLayout& a_input,
				   int a_dir, int a_len);
void coarsen_dbl(DisjointBoxLayout& output, 
				 const DisjointBoxLayout& input, 
				 int refinement);
void refine_dbl(DisjointBoxLayout& output, 
				const DisjointBoxLayout& input, 
				int refinement);

#ifndef WRAPPER
inline
void adjCellLo_dbl(DisjointBoxLayout& a_output,
				   const DisjointBoxLayout& a_input,
				   int a_dir, int a_len) { adjCellLo(a_output,a_input,a_dir,a_len);}
inline
void adjCellHi_dbl(DisjointBoxLayout& a_output,
			   const DisjointBoxLayout& a_input,
			   int a_dir, int a_len){ adjCellHi(a_output,a_input,a_dir,a_len);}
inline
void coarsen_dbl(DisjointBoxLayout& output, 
				 const DisjointBoxLayout& input, 
				 int refinement){ coarsen(output, input, refinement);}
inline
void refine_dbl(DisjointBoxLayout& output, 
			const DisjointBoxLayout& input, 
				int refinement){ refine(output, input, refinement);}
#endif

#endif
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//


DisjointBoxLayout::DisjointBoxLayout()
  :BoxLayout()
{
}


DisjointBoxLayout::DisjointBoxLayout(const Vector<Box>& a_boxes,
                                     const Vector<int>& a_procIDs)
  :BoxLayout(a_boxes,a_procIDs)
{
  assert(isDisjoint());
}


DisjointBoxLayout::DisjointBoxLayout(const Vector<Box>& a_boxes,
                                     const Vector<int>& a_procIDs,
                                     const ProblemDomain& a_physDomain)
  :BoxLayout(a_boxes,a_procIDs), m_physDomain(a_physDomain)
{
  assert(isDisjoint());
}


void
DisjointBoxLayout::define(const Vector<Box>& a_boxes,
                          const Vector<int>& a_procIDs)
{
  BoxLayout::define(a_boxes,a_procIDs);
#if 0
  assert(isDisjoint());
#endif
}


void
DisjointBoxLayout::define(const Vector<Box>& a_boxes,
                          const Vector<int>& a_procIDs,
                          const ProblemDomain& a_physDomain)
{
  m_physDomain = a_physDomain;
  BoxLayout::define(a_boxes,a_procIDs);
#if 0
  assert(isDisjoint());
#endif
}


void
DisjointBoxLayout::define(BoxLayout& a_layout)
{
  ((BoxLayout*)this)->operator=(a_layout);
  if(*m_closed)
    {
      assert(isDisjoint());
    }
}


void
DisjointBoxLayout::define(BoxLayout& a_layout, 
                          const ProblemDomain& a_physDomain)
{
  m_physDomain = a_physDomain;

  ((BoxLayout*)this)->operator=(a_layout);
  if(*m_closed)
    {
      assert(isDisjoint());
    }
}

void
DisjointBoxLayout::close()
{
  if(!*m_closed)  //do nothing if already closed
    {
      sort();
      assert(isDisjoint());
      *m_closed = true;
      m_dataIterator = new DataIterator(*this, m_layout);
    }
}


void
DisjointBoxLayout::deepCopy(const BoxLayout& a_layout)
{
  BoxLayout::deepCopy(a_layout);
  if(!a_layout.isClosed())
    sort();
  assert(isDisjoint());
}


void
DisjointBoxLayout::deepCopy(const DisjointBoxLayout& a_layout)
{
  BoxLayout::deepCopy(a_layout);
  m_physDomain = a_layout.m_physDomain;
  if(!a_layout.isClosed())
    sort();
  assert(isDisjoint());
}

void
DisjointBoxLayout::deepCopy(const BoxLayout& a_layout, 
                            const ProblemDomain& a_physDomain)
{
  m_physDomain = a_physDomain;
  BoxLayout::deepCopy(a_layout);
  if(!a_layout.isClosed())
    sort();
  assert(isDisjoint());
}

bool
DisjointBoxLayout::isDisjoint() const
{
  // this is a holder for boxes in the periodic case -- if a box
  // needs to be checked for its periodic images, then save it
  // and come back to it later (to preserve O(N) algorithm)
  Vector<Box> periodicCheckBoxes;
  for(unsigned int i=0; i<size(); ++i){
    const Box& a = m_boxes->operator[](i).box;
    // in the periodic case, no box may be larger than the domain
    if (m_physDomain.isPeriodic())
      {
        if (m_physDomain.domainBox().size() < a.size())
          return false;
        // if a extends out of basic domain box, save it for
        // periodic checking later
        if (!m_physDomain.domainBox().contains(a))
          periodicCheckBoxes.push_back(a);
      }
    for(unsigned int j=i+1; j<size(); ++j){
      bool disjoint = false;
      const Box& b = m_boxes->operator[](j).box;
      D_EXPR(disjoint = disjoint || a.bigEnd(0) < b.smallEnd(0) || a.smallEnd(0) > b.bigEnd(0),
             disjoint = disjoint || a.bigEnd(1) < b.smallEnd(1) || a.smallEnd(1) > b.bigEnd(1),
             disjoint = disjoint || a.bigEnd(2) < b.smallEnd(2) || a.smallEnd(2) > b.bigEnd(2));
      if(!disjoint)
        {
          return false;
        }
      else if(b.smallEnd(0) > a.bigEnd(0)) 
        {
          //can skip the rest, since we know that the smallEnd of all 
          // the remaining boxes are going to also return true for this test.
          j=size();
        }
    }
  }

  // now come back to boxes we need to check for periodic images
  if (periodicCheckBoxes.size() > 0)
    {
      ShiftIterator shiftIt = m_physDomain.shiftIterator();
      IntVect shiftMult = m_physDomain.domainBox().size();
      for (shiftIt.begin(); shiftIt.ok(); ++shiftIt) 
        {
          IntVect shiftVect(shiftMult*shiftIt());
          for (unsigned int i=0; i<periodicCheckBoxes.size(); i++) 
            {
              Box& a = periodicCheckBoxes[i];
              a.shift(shiftVect);
              // only check for intersections with other boxes if 
              // the shifted box intersects the computational domain.
              if (a.intersects(m_physDomain.domainBox()))
                {
                  // now repeat disjointness test with shifted box
                  for (unsigned int j=0; j<size(); ++j)
                    {
                      bool disjoint = false;
                      const Box& b = m_boxes->operator[](j).box;
                      D_EXPR(disjoint = disjoint || a.bigEnd(0) < b.smallEnd(0) || a.smallEnd(0) > b.bigEnd(0),
                             disjoint = disjoint || a.bigEnd(1) < b.smallEnd(1) || a.smallEnd(1) > b.bigEnd(1),
                             disjoint = disjoint || a.bigEnd(2) < b.smallEnd(2) || a.smallEnd(2) > b.bigEnd(2));
                      if(!disjoint)
                        {
                          return false;
                        }
                      else if(b.smallEnd(0) > a.bigEnd(0)) 
                        {
                          //can skip the rest, since we know that the smallEnd 
                          // of all the remaining boxes are going to also return
                          // true for this test.
                          j=size();
                        }
                    } // end loop over j
                } // end if shifted box intersects computational domain
              // shift box back
              a.shift(-shiftVect);
            } //end loop over i
        } // end loop over shift directions
    } // end if we need to check periodic images
  
  return true;
}


bool
DisjointBoxLayout::checkPeriodic(const ProblemDomain& a_domain) const
{
  bool goodMatch = true;
  if (a_domain.isPeriodic())
    {
      for (int dir=0; dir<SpaceDim; dir++) 
        {
          if (a_domain.isPeriodic(dir)) {
            const Box& thisDomBox = m_physDomain.domainBox();
            const Box& inputDomBox = m_physDomain.domainBox();
            
            // make sure both are periodic in direction dir
            if (m_physDomain.isPeriodic(dir))
              {
                // both domains are periodic in direction dir -- check
                // that period is the same as well
                if ((thisDomBox.smallEnd(dir) != inputDomBox.smallEnd(dir)) 
                    ||(thisDomBox.bigEnd(dir) != inputDomBox.bigEnd(dir)))
                  {
                    // periods don't agree
                    goodMatch = false;
                  } // end if periods don't agree
              } // end if both are periodic in direction dir
            else
              {
                // a_domain is periodic, but m_physDomain is not!
                goodMatch = false;
              }
          } else if (m_physDomain.isPeriodic(dir)) {
            
            // in this case, a_domain is not periodic, but m_physDomain is!
            goodMatch = false;
          }
        } // end loop over directions 
          
    } // end if a_domain is periodic in any direction
  else if (m_physDomain.isPeriodic()) 
    {
      // if a_domain is not periodic, then m_physdomain better not be either
      goodMatch = false;
    } 

  return goodMatch;
}
         

bool
DisjointBoxLayout::checkDomains(const DisjointBoxLayout& a_dbl) const
{
  const ProblemDomain& dblDomain = a_dbl.physDomain();
  return checkPeriodic(dblDomain);
}


// Global functions
// ================

void
coarsen(DisjointBoxLayout& a_output, const DisjointBoxLayout& a_input, 
        int a_refinement)
{
  
  if(!a_input.isClosed())
    {
      MayDay::Error("input to coarsen must be called with closed BoxLayout");
    }
  if(a_output.isClosed())
    {
      MayDay::Error("output of coarsen must be called on open BoxLayout");
    }
  
  // copy first, then coarsen everything
  a_output.deepCopy(a_input);

  // now coarsen the physDomain
  a_output.m_physDomain = coarsen(a_input.m_physDomain, a_refinement);
  
  
  for(LayoutIterator it(a_input.layoutIterator()); it.ok(); ++it)
    {
      a_output[it()].coarsen(a_refinement);
    }
  a_output.close();
}



// we have an easier time with refine, since we know that refinement will
// not change the state of a sort, but, we will play it safe for now
// until this function shows up in the profiler.

void refine(DisjointBoxLayout& a_output, 
            const DisjointBoxLayout& a_input, 
            int a_refinement)
{
  if(!a_input.isClosed())
    {
      MayDay::Error("input to refine must be called with closed BoxLayout");
    }
  if(a_output.isClosed())
    {
      MayDay::Error("output of refine must be called on open BoxLayout");
    }
  

  // first copy, then refine everything 
  a_output.deepCopy(a_input);

  // start by refining the physDomain
  a_output.m_physDomain = refine(a_input.m_physDomain,a_refinement);
  
  for(LayoutIterator it(a_input.layoutIterator()); it.ok(); ++it)
    {
      a_output[it()].refine(a_refinement);
    }
  a_output.close();
}


void
adjCellLo(DisjointBoxLayout& a_output, 
          const DisjointBoxLayout& a_input,
          int a_dir, int a_len)
{
  if(!a_input.isClosed())
    {
      MayDay::Error("input to adjCellLo must be called with closed BoxLayout");
    }
  if(a_output.isClosed())
    {
      MayDay::Error("output of adjCellLo must be called on open BoxLayout");
    }
  
  // copy first, then adjCellLo everything
  a_output.deepCopy(a_input);

  for(LayoutIterator it(a_input.layoutIterator()); it.ok(); ++it)
    {
      a_output[it()] = adjCellLo(a_output[it()],a_dir, a_len);
    }
  // shouldn't need to resort anything
  a_output.close();
}


void
adjCellHi(DisjointBoxLayout& a_output, 
          const DisjointBoxLayout& a_input,
          int a_dir, int a_len)
{
  if(!a_input.isClosed())
    {
      MayDay::Error("input to adjCellHi must be called with closed BoxLayout");
    }
  if(a_output.isClosed())
    {
      MayDay::Error("output of adjCellHi must be called on open BoxLayout");
    }
  
  // copy first, then adjCellHi everything
  a_output.deepCopy(a_input);

  for(LayoutIterator it(a_input.layoutIterator()); it.ok(); ++it)
    {
      a_output[it()] = adjCellHi(a_output[it()],a_dir, a_len);
    }
  // shouldn't need to resort anything
  a_output.close();
}


const
ProblemDomain&
DisjointBoxLayout::physDomain() const
{
  return m_physDomain;
}
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//


LayoutIterator::LayoutIterator(const BoxLayout& plan, 
                               const int* layoutID)
  :m_layout(plan), m_index(0), m_current(0, layoutID)
{
  if(ok())
    m_current.m_index = (*(m_layout.m_boxes))[m_index].index;
}

void LayoutIterator::end()
{
  m_index = m_layout.size();
}

void LayoutIterator::reset()
{
  m_index = 0;
  if(ok())
    m_current.m_index = (*(m_layout.m_boxes))[m_index].index;
}
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//


DataIterator::DataIterator(const BoxLayout& plan, 
                           const int* layoutID)
  :m_layout(plan), m_index(0), m_current(0, layoutID), m_procID(0)
{
  // in many programs a user will call MPI_Finalize, at the end of
  // the main routine.  But some objects, like LayoutData, have not
  // had their destructors called.
  m_procID = procID();
  begin(); // might result in ok()==false if there were no boxes on this processor
}

void DataIterator::end()
{
  m_index = m_layout.size();
}
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef LAYOUTDATA_H
#define LAYOUTDATA_H


//class ArrayViewData;

///  Data that maintains a one-to-one mapping to the boxes in a BoxLayout
/**

  A collection of Box-oriented objects.  The arrangement
  of the T objects is given by the underlying BoxLayout object.
  LayoutData does not attempt to prevent users from manipulating
  the template class T to become out of synch with the
  boxes in the BoxLayout.  Caveat emptor.

  Non-local (off-processor access) to a LayoutData index is an error.
  you can assure that access to the data stored in a LayoutData is
  local to your processor by using the DataIterator object for indexing.
  (In fact, you are pretty much forced to by the 

  Data in a LayoutData CANNOT be communicated to other
  processors using the API presented here.  

  class T must provide the following methods:
  \begin{verbatim}
  {
    T()
  }
  \end{verbatim}

  This class is useful for processor-local data that needs to be indexable
  along with a BoxLayoutData.  Auxillary data objects, etc.

*/
template<class T> class LayoutData
{
public:

  friend class ArrayViewData;

  ///
  LayoutData();

  ///
  /** constructor.  allocates a T object for every box in the BoxLayout dp
    using the T null constructor.  dp must be closed. */
  LayoutData(const BoxLayout& dp);

  ///
  virtual ~LayoutData();

  ///
  /** constructor.  allocates a T object for every box in the BoxLayout dp
    using the T null constructor.  dp must be closed.  previously stored
    T objects are removed */
  virtual void define(const BoxLayout& dp);

  // this approach, of using the iterator itself for indexing is
  // my own brilliance.  This way you don't need all the
  // crappy [Const][Dependent] Iterators.  It also forces all
  // data access to be done through the DataIterator, which will
  // know how to talk to KeLP (or other parallel architectures,
  // like multithreading, etc).

  ///
  DataIterator dataIterator() const;

  ///
  const T& operator[](const DataIndex& index) const;

  ///
  T& operator[](const DataIndex& index);

  ///
  Box box(const DataIndex& index) const;


  // not really a protection mechanism, since a user can always
  // perform a shallow copy of the BoxLayout, and be free to manipulate
  // the data in the object.  If you wish to have an actual copy 
  // of the BoxLayout, then you need to invoke the clone() method.

  ///
  const BoxLayout& boxLayout() const
    {
      return m_boxLayout;
    }


protected:

  BoxLayout m_boxLayout;

  // this used to be std::vector<T>, and I could let vector handle
  // destruction for me, but vector.resize() absolutely demands that
  // I provide a copy constructor for T.  People get uncomfortable when
  // I make them provide me with copy constructors.
  Vector<T*> m_vector;

  // thinking about making this class self-documenting to a greater degree
  // and having the component names also be kept in the class and carried
  // around through various transformations.....
  //  vector<string> m_componentNames;

private:

  // disallow copy constructors and assignment operators
  // to avoid very hard-to-find performance problems 
  LayoutData<T>& operator= (const LayoutData<T>& rhs);
  LayoutData(const LayoutData& rhs);

 // handy usage function, tied to the implementation of this class
  // as a vector<T*>.  Assumes that m_comps and m_boxLayout have already
  // been initialized correctly.  Sets the correct size for the data
  // vector, deletes the extra T objects if m_vector is to shorten, and
  // performs either construction or define on the remaining T objects.

  void allocate();

};


//======================================================================
//======================================================================




//========= inline functions ===========================

// not literally a .H file, but just an experiment in
// having useable, readable headers, while having 
// the dependeny system work properly.  Since LayoutData.H
// now depends on LayoutDataI.H, then changing either
// recompiled.  This way people can just read the interface
// in this file.  Implementation is put in the *I.H file.


#endif // LAYOUTDATA_H
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//


template<class T>
inline const T& LayoutData<T>::operator[](const DataIndex& index) const
{
  assert(m_boxLayout.check(index));
  assert(m_boxLayout.procID(index) == procID()); // using a DataIndex for data on another processor
                                       //if you are reading this error in a debugger, then
                                       // you haven't used addBox(Box, int) correctly, or
                                       // you are using a LayoutIterator to access a data holder.
  return *(m_vector[m_boxLayout.index(index)]);
}

template<class T>
inline T& LayoutData<T>::operator[](const DataIndex& index)
{
  assert(m_boxLayout.check(index));
  assert(m_boxLayout.procID(index) == procID());// using a DataIndex for data on another processor
                                       //if you are reading this error in a debugger, then
                                       // you haven't used addBox(Box, int) correctly, or
                                       // you are using a LayoutIterator to access a data holder.

  return *(m_vector[m_boxLayout.index(index)]);
}

template<class T>
inline Box LayoutData<T>::box(const DataIndex& index) const
{
  return m_boxLayout.get(index);
}

template<class T>
inline DataIterator LayoutData<T>::dataIterator() const
{
  return m_boxLayout.dataIterator();
}

template<class T>
inline LayoutData<T>::LayoutData()
{
  m_boxLayout.close();
}

template<class T>
inline LayoutData<T>::LayoutData(const BoxLayout& dp)
  : m_boxLayout(dp)
{
  assert(dp.isClosed());
  allocate();
}

template<class T>
inline void LayoutData<T>::define(const BoxLayout& dp)
{
  assert(dp.isClosed());
  m_boxLayout = dp;
  allocate();
}

template<class T>
inline LayoutData<T>::~LayoutData()
{
  for(DataIterator it(dataIterator()); it.ok(); ++it)
    {
      delete m_vector[m_boxLayout.index(it())];
    }
}


template<class T>
inline void LayoutData<T>::allocate()
{

  for(unsigned int i=0; i<m_vector.size(); ++i)
    {
      delete m_vector[i];
      m_vector[i] = NULL;
    }

  m_vector.resize(m_boxLayout.size(), NULL);

  for(DataIterator it(dataIterator()); it.ok(); ++it)
    {
      unsigned int index = m_boxLayout.index(it());
      if(m_vector[index] == NULL) 
        {
          m_vector[index] = new T();
          if(m_vector[index] == NULL)
          {
            MayDay::Error("OutOfMemory in boxlayoutdata::setVector");
          }
        }
 
    }
}


/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef BOXLAYOUTDATA_H
#define BOXLAYOUTDATA_H


/// Factory object to data members of a BoxLayoutData container
template <class T> class DataFactory
{
public:
  /// factory function.  creates a new 'T' object
  /** creates a new 'T' object and returns a pointer to it.  Responsiblitly
      for calling operator 'delete' on this pointer is passed to the user. */
  virtual T* create(const Box& box, int ncomps, const DataIndex& a_datInd) const=0;

};


/// Factory object to data members of a BoxLayoutData container
template <class T> class DefaultDataFactory : public DataFactory<T>
{
public:
  /// factory function.  creates a new 'T' object
  /** creates a new 'T' object and returns a pointer to it.  Responsiblitly
      for calling operator 'delete' on this pointer is passed to the user. */
  virtual T* create(const Box& box, int ncomps, const DataIndex& a_datInd) const;

};

class FABAliasDataFactory : public DataFactory<FArrayBox>
{
public:
  virtual ~FABAliasDataFactory(){;}
  FABAliasDataFactory(const LayoutData<Real*>& aliases);
  void define(const LayoutData<Real*>& aliases);
  /// factory function.  creates a new 'T' object using an aliased dataPtr for T
  /** creates a new 'T' object and returns a pointer to it.  Responsiblitly
      for calling operator 'delete' on this pointer is passed to the user. */
  virtual FArrayBox* create(const Box& box, int ncomps, const DataIndex& a_datInd) const;

protected:
  LayoutData<Real*> aliasPtrs;
};


template<class T> class LevelData;


/// Data on a BoxLayout
/**

  A collection of Box-oriented objects.  The arrangement
  of Boxes is given by the underlying BoxLayout object.
  BoxLayoutData attempt to prevent users from maniupulating
  the template class T to become out of synch with the
  boxes in the BoxLayout.  Caveat emptor.

  All access to the data in a BoxLayoutData is forced to be
  data-parallel, by use of the DataIterator.

  class T must provide the following methods:
\begin{itemize}
    \item {\tt T()}
    \item {\tt T(const Box& box, int comps)}
    \item {\tt define(const Box& box, int comps)}
    \item {\tt void copy(const Box& R, const Interval& Cd, const T& source, const Interval Cs)}
       Cs and Cd must have same length, but can be displaced

    \item {\tt static int preAllocatable()} 
     returns 0 if the size(...) 
        function is strictly
        a function of Box and Interval, and does not depend
        on the current state of the T object.
     return 1 if size(..) is symmetric, in that sender and receiver T object
        can size their message buffers, but a static object cannot.
     returns 2 if the object is truly dynamic.  the message size is subject to 
        unique object data.
    \item {\tt int size(const Box& R, const Interval& comps)}
    \item {\tt void linearOut(void* buf, const Box& R, const Interval& comps)}
    \item {\tt void linearIn(void* buf, const Box& R, const Interval& comps)}
  \end{itemize}

  In sequential codes, many of these functions will not be called, but in the
  general parallel case, they all will be.

*/

template<class T> 
class BoxLayoutData : public LayoutData<T>
{
public:
  ///
  BoxLayoutData():m_comps(0) {m_isdefined = false;}

  virtual ~BoxLayoutData(){;}
  ///
  BoxLayoutData(const BoxLayout& boxes, int comps, 
                const DataFactory<T>& factory = DefaultDataFactory<T>());

  ///
  virtual void define(const BoxLayout& boxes, int comps, 
                      const DataFactory<T>& factory = DefaultDataFactory<T>());

  ///
  virtual void define(const BoxLayoutData<T>& da,
                      const DataFactory<T>& factory = DefaultDataFactory<T>());

  ///
  /** Define this BoxLayoutData to be the subset of the
      data in da defined by the Interval comps.
  */
  virtual void define(const BoxLayoutData<T>& da, const Interval& comps,
                      const DataFactory<T>& factory = DefaultDataFactory<T>());

  /// overridden and throws an error.  
  virtual void define(const BoxLayout& boxes);

  ///
  int nComp() const { return m_comps;}

  ///
  Interval interval() const
  {
    Interval outint(0, m_comps-1);
    return(outint);
  }


    ///
  /* User writes a function with the signature:
     \begin{verbatim}
     void setFunction(const Box& box, int comps, T& t){ your code here;}
     \end{verbatim}
     They can then hand this off to LayoutData::setVal.  This class
     then cycles through all the T objects and invokes this function.  Function
     must not be inline. (I'm still trying to figure out a nice way to send
     in non-static member functions).
     */
  void setVal(void (*setFunction)(const Box& box, int comps, T& t));


  ///
  virtual bool isDefined() const;
protected:

  int             m_comps;
  bool            m_isdefined;

  friend class LevelData<T>;
  void setVector(const BoxLayoutData<T>& da, 
                 const Interval& srcComps,
                 const Interval& destComps);
  void allocateGhostVector(const DataFactory<T>& factory, 
                           const IntVect& ghost = IntVect::TheZeroVector());
};

///   
/** not actually L-p norms, since it doesn't take into
  account the dx of the system.  A user can take that
  into account or not.

  For p != 0, returns pth root of sum of pth powers over all
  points in all fabs and all components in the interval:

  ( sum [ |A[i][pt,var]|^p : FArrayBox A[i], point pt in A[i].box(), var in interval ] )^(1/p)

  To turn into an L-p norm, one needs to multiply this by dx^(SpaceDim/p).

  For p == 0, returns global max over all points in all fabs and all
  components in the interval:

  max [ |A[i][pt,var]| : FArrayBox A[i], point pt in A[i].box(), var in interval ]

  Some people don't like that this norm is not normalized based on
  number of points in A.  Normalization is your problem.
  */
Real norm(const BoxLayoutData<FArrayBox>& A, 
          const Interval& interval,
          const int& p = 2);


//======================================================================
template < >
BaseFab<int>* DefaultDataFactory<BaseFab<int> >::create(const Box& box, 
                                                        int ncomps, 
                                                        const DataIndex& a_datInd) const;

template < >
FArrayBox* DefaultDataFactory<FArrayBox>::create(const Box& box, 
                                                 int ncomps,
                                                 const DataIndex& a_datInd) const;

template <class T>
T* DefaultDataFactory<T>::create(const Box& box, 
                                 int ncomps, 
                                 const DataIndex& a_datInd) const
{
  return new T(box, ncomps);
}


template<class T>
inline bool BoxLayoutData<T>::isDefined() const
{
  return m_isdefined;
}

template <class T>
inline void BoxLayoutData<T>::setVector(const BoxLayoutData<T>& da, 
                                        const Interval& srcComps,
                                        const Interval& destComps)
{
  if(&da != this)
    {
      for(DataIterator it(dataIterator()); it.ok(); ++it)
        {
          m_vector[m_boxLayout.index(it())]->copy( box(it()), destComps, 
                                                   box(it()), da[it()], srcComps);
        }
    }
}

template<class T>
inline void BoxLayoutData<T>::define(const BoxLayoutData<T>& da, const Interval& comps,
                                     const DataFactory<T>& factory)
{
  if(this == &da){
    MayDay::Error("BoxLayoutData<T>::define(const LayoutData<T>& da,.....) called with 'this'");
  }
  assert(comps.size()>0);
  assert(comps.end()<=m_comps);
  assert(comps.begin()>=0);
  m_boxLayout = da.boxLayout();

  m_comps = comps.size();

  Interval dest(0, m_comps-1);
  allocateGhostVector(factory);
  setVector(da, comps, dest);

}

template<class T>
inline void BoxLayoutData<T>::define(const BoxLayout& boxes, int comps,
                                     const DataFactory<T>& factory)
{
  assert(boxes.isClosed());
  m_boxLayout = boxes;
  m_comps = comps;
  m_isdefined = true;
  allocateGhostVector(factory);
}

template<class T>
inline void BoxLayoutData<T>::define(const BoxLayout& boxes)
{
  MayDay::Error("BoxLayoutData<T>::define(const BoxLayout& boxes)...needs comps");
}

template<class T>
inline BoxLayoutData<T>::BoxLayoutData(const BoxLayout& boxes, int comps,
                                       const DataFactory<T>& factory)
  :m_comps(comps)
{
  assert(boxes.isClosed());
  m_boxLayout = boxes;
  m_isdefined = true;
  allocateGhostVector(factory);
}

template<class T>
inline void BoxLayoutData<T>::define(const BoxLayoutData<T>& da,
                                     const DataFactory<T>& factory)
{
  if(this != &da){
    m_isdefined = da.m_isdefined;
    m_boxLayout = da.boxLayout();
    m_comps    = da.nComp();
    Interval srcAnddest(0, m_comps-1);
    allocateGhostVector(factory);
    setVector(da, srcAnddest, srcAnddest);
  }
}


template<class T>
inline void BoxLayoutData<T>::allocateGhostVector(const DataFactory<T>& factory, const IntVect& ghost)
{
  for(unsigned int i=0; i<m_vector.size(); ++i)
    {
      delete m_vector[i];
      m_vector[i] = NULL;
    }

  m_vector.resize(m_boxLayout.size(), NULL);

  for(DataIterator it(dataIterator()); it.ok(); ++it)
    {
      unsigned int index = m_boxLayout.index(it());
      Box abox = box(it());
      abox.grow(ghost);
      m_vector[index] = factory.create(abox, m_comps, it());
      if(m_vector[index] == NULL)
        {
          MayDay::Error("OutOfMemory in boxlayoutdata::allocate");
        }
    }
}

template<class T>
inline void BoxLayoutData<T>::setVal(void (*a_setVal)(const Box& box, int comps, T& t))
{
  for(DataIterator it(dataIterator()); it.ok(); ++it)
    {
      unsigned int index = m_boxLayout.index(it());
      a_setVal(box(it()), m_comps, *(m_vector[index]));
    }
}



//======================================================================

#endif // BOXLAYOUTDATA_H
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//


#ifdef MPI
#endif

Real norm(const BoxLayoutData<FArrayBox>& a_layout, 
          const Interval& interval,
          const int& p)
{
  const BoxLayout validRegion = a_layout.boxLayout();
  DataIterator it = validRegion.dataIterator();
  Real norm = 0;
  if(p == 0)  // max norm
    {
      for(;it.ok(); ++it)
        {
          norm = Max(norm, a_layout[it()].norm(validRegion[it()],
                                               0,
                                               interval.begin(),
                                               interval.size()));
        }
#     ifdef MPI
      Real recv;
      int result = MPI_Allreduce(&norm, &recv, 1, MPI_CH_REAL,
                                 MPI_MAX, Chombo_MPI::comm);
      if(result != MPI_SUCCESS){ //bark!!!
        MayDay::Error("sorry, but I had a communcation error on norm");
      }
      norm = recv;
#     endif
    }
  else if(p == 1) // abs sum norm
    {
      for(;it.ok(); ++it)
        {
          norm += a_layout[it()].norm(validRegion[it()],
                                      1,
                                      interval.begin(),
                                      interval.size());
        }
#     ifdef MPI
      Real recv;
      int result = MPI_Allreduce(&norm, &recv, 1, MPI_CH_REAL,
                                 MPI_SUM, Chombo_MPI::comm);
      if(result != MPI_SUCCESS){ //bark!!!
        MayDay::Error("sorry, but I had a communcation error on norm");
      }
      norm = recv;
#     endif
    }
  else
    {
      for(;it.ok(); ++it)
        {
          norm+=a_layout[it()].sumPow(validRegion[it()],
                                      p,
                                      interval.begin(),
                                      interval.size());
        }
#     ifdef MPI
      Real recv;
      int result = MPI_Allreduce(&norm, &recv, 1, MPI_CH_REAL,
                                 MPI_SUM, Chombo_MPI::comm);
      if(result != MPI_SUCCESS){ //bark!!!
        MayDay::Error("sorry, but I had a communcation error on norm");
      }
      norm = recv;
#     endif
      Real invpwr = 1.0/p;
      norm = pow(norm, invpwr);
    }

  return norm;
}

template <>
BaseFab<int>* DefaultDataFactory<BaseFab<int> >::create(const Box& box, 
                                                        int ncomps,
                                                        const DataIndex& a_datInd) const
{
  return new BaseFab<int>(box, ncomps);
}

template <>
FArrayBox* DefaultDataFactory<FArrayBox>::create(const Box& box, int ncomps,
                                          const DataIndex& a_datInd) const
{
  return new FArrayBox(box, ncomps);
}

FABAliasDataFactory::FABAliasDataFactory(const LayoutData<Real*>& aliases)
{
  define(aliases);
}

void FABAliasDataFactory::define(const LayoutData<Real*>& aliases)
{
  aliasPtrs.define(aliases.boxLayout());
  DataIterator dit = aliases.dataIterator();
  for(dit.begin(); dit.ok(); ++dit)
	{
	  aliasPtrs[dit()] = aliases[dit()];
	}
}

FArrayBox* FABAliasDataFactory::create(const Box& box, int ncomps, const DataIndex& a_datInd) const
{
  const Box& b = aliasPtrs.box(a_datInd);
  if(b != box)
	{
	  MayDay::Error("Aliased data holder dimensions do not match across LevelData const.");
	}
  FArrayBox* rtn = new FArrayBox(box, ncomps, aliasPtrs[a_datInd]);
  return rtn;
}
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef COPIER_H
#define COPIER_H


class MotionItem;

class CopyIterator;


/// A strange but true thing to make copying from one boxlayoutdata to another fast
/**
  class to handle the of data from a LevelData to a BoxLayoutData
  defined over the same index space.  The idea behind this object
  is twofold:
  A)  there is a significant amount of information that can be
  computed once and reused from one copy to the next for
  a LevelData and BoxLayoutData that have unchanging BoxLayouts
  underneath them.  In these cases, LevelData::copyTo methods
  has an optional argument.  That way, in cases where the operation
  is only performed a single time, we do not waste our time building
  optimizing data, etc.
  B)  when we interface with KeLP, this class maps quite well
  with the existing KeLP API.

  pains have been taken to not have this class be templated.  Thus
  it does not actually perform the copying itself (which would require
  knowledge of the template class type).  It provides an iterator
  interface to MotionItems that specify the required work.  LevelData
  provides the copyTo methods.
*/

class Copier
{
public:

  ///null constructor, copy constructor and operator= can be compiler defined.
  Copier(){;}
  ///
  Copier(const DisjointBoxLayout& a_level, const BoxLayout& a_dest);

  /// this constructor contains support for periodic BC's
  Copier(const DisjointBoxLayout& a_level, const BoxLayout& a_dest,
         const ProblemDomain& a_domain);

  /// Copier to copy data into the valid and invalid regions of a_dest from a_level
  Copier(const DisjointBoxLayout& a_level, 
         const BoxLayout& a_dest, 
         const IntVect& a_destGhost);

  /// Copier to copy data into the valid and invalid regions of a_dest from a_level with support for periodic BCs
  Copier(const DisjointBoxLayout& a_level, 
         const BoxLayout& a_dest, 
         const ProblemDomain& a_domain,
         const IntVect& a_destGhost);

  ///
  virtual ~Copier();

  ///
  virtual void define(const DisjointBoxLayout& a_level, const BoxLayout& a_dest);


  /// contains support for periodic BCs
  virtual void define(const DisjointBoxLayout& a_level, 
                      const BoxLayout& a_dest,
                      const ProblemDomain& a_domain);

  ///
  virtual void define(const DisjointBoxLayout& a_level, 
                      const BoxLayout& a_dest,
                      const IntVect& a_destGhost);

  /// contains support for periodic BCs
  virtual void define(const DisjointBoxLayout& a_level, 
                      const BoxLayout& a_dest,
                      const ProblemDomain& a_domain,
                      const IntVect& a_destGhost);

  ///
  virtual void clear();

  /// check that this Copier object corresponds to these two BoxLayouts.
  bool check(const DisjointBoxLayout& from, const BoxLayout& to) const;

  int print() const;

  bool bufferAllocated() const;
  void setBufferAllocated(bool arg) const;

protected:

  friend class CopyIterator;

  Vector<MotionItem*> m_localMotionPlan;
  Vector<MotionItem*> m_fromMotionPlan;
  Vector<MotionItem*> m_toMotionPlan;

  friend void dumpmemoryatexit();
  static Pool s_motionItemPool;
  mutable bool buffersAllocated;
private:

  // keep a refcounted reference around for debugging purposes, we can
  // decide afterwards if we want to eliminate it.
  DisjointBoxLayout m_originPlan;
  BoxLayout  m_dest;

};

std::ostream& operator<<(std::ostream& os, const Copier& copier);

//===========================================================================
// end of public interface for Copier.  
//===========================================================================


//  These classes are public because I can't find a nice
//  way to make a class a friend of all the instantiations
//  of a template class.  These classes are not part of
//  the public interface for the Array API.  
//
//  Later, if MotionItem shows up in the profiler, we
//  can start using a pool allocation scheme and placement new

class MotionItem{
public:
  DataIndex fromIndex, toIndex;
  Box fromRegion;
  Box toRegion;
  int procID;
  // this constructor will probably eventually go away
  MotionItem(const DataIndex& a_from, 
             const DataIndex& a_to,
             const Box&       a_region);
  MotionItem(const DataIndex& a_from, 
             const DataIndex& a_to,
             const Box&       a_fromRegion,
             const Box&       a_toRegion);
};

inline MotionItem::MotionItem(const DataIndex& a_from, 
                              const DataIndex& a_to,
                              const Box&       a_region)
  :fromIndex(a_from), toIndex(a_to), fromRegion(a_region), 
  toRegion(a_region), procID(-1)
{;}

inline MotionItem::MotionItem(const DataIndex& a_from, 
                              const DataIndex& a_to,
                              const Box&       a_fromRegion,
                              const Box&       a_toRegion)
  :fromIndex(a_from), toIndex(a_to), fromRegion(a_fromRegion), 
  toRegion(a_toRegion), procID(-1)
{;}



class CopyIterator
{
public:
  enum local_from_to {LOCAL, FROM, TO};

  inline CopyIterator(const Copier& a_copier, local_from_to);

  inline const MotionItem& operator()() const;

  inline void operator++();

  inline bool ok() const;

  inline void reset();
private:
  const Vector<MotionItem*>* m_motionplanPtr;
  unsigned int  m_current;
};

//====== inlined functions====================================

inline CopyIterator::CopyIterator(const Copier& a_copier, local_from_to type)
  :m_current(0)
{
  switch(type){
  case LOCAL:
    m_motionplanPtr = &(a_copier.m_localMotionPlan);
    break;
  case FROM:
    m_motionplanPtr = &(a_copier.m_fromMotionPlan);
    break;
  case TO:
    m_motionplanPtr = &(a_copier.m_toMotionPlan);
    break;
  default:
    MayDay::Error("illegal local_from_to option for CopyIterator");
  }
}

inline const MotionItem& CopyIterator::operator()() const
{
  assert(m_current < m_motionplanPtr->size());
  return *(m_motionplanPtr->operator[](m_current));
}

inline void CopyIterator::operator++() {++m_current;}

inline bool CopyIterator::ok() const
{
  return m_current < m_motionplanPtr->size();
}

inline void CopyIterator::reset()
{
  m_current = 0;
}




#endif 
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//



using std::ostream;

Pool Copier::s_motionItemPool(sizeof(MotionItem), "Copier::MotionItem");

Copier::Copier(const DisjointBoxLayout& a_level, const BoxLayout& a_dest)
{
  const ProblemDomain& domain = a_level.physDomain();
  define(a_level, a_dest, domain, IntVect::TheZeroVector());
}

Copier::Copier(const DisjointBoxLayout& a_level, const BoxLayout& a_dest,
               const ProblemDomain& a_domain)
{
  define(a_level, a_dest, a_domain, IntVect::TheZeroVector());
}


Copier::Copier(const DisjointBoxLayout& a_level,
               const BoxLayout& a_dest,
               const IntVect& a_destGhost)
{
  const ProblemDomain domain = a_level.physDomain();
  define(a_level, a_dest, domain, a_destGhost);
}

Copier::Copier(const DisjointBoxLayout& a_level,
               const BoxLayout& a_dest,
               const ProblemDomain& a_domain,
               const IntVect& a_destGhost)
{
  define(a_level, a_dest, a_domain, a_destGhost);
}


void Copier::clear()
{
  for(unsigned int i=0; i<m_localMotionPlan.size(); ++i)
    {
      s_motionItemPool.returnPtr( m_localMotionPlan[i] );
    }
  for(unsigned int i=0; i<m_fromMotionPlan.size(); ++i)
    {
      s_motionItemPool.returnPtr( m_fromMotionPlan[i] );
    }
  for(unsigned int i=0; i<m_toMotionPlan.size(); ++i)
    {
      s_motionItemPool.returnPtr( m_toMotionPlan[i] );
    }

  m_localMotionPlan.resize(0);
  m_fromMotionPlan.resize(0);
  m_toMotionPlan.resize(0);
}

void Copier::define(const DisjointBoxLayout& a_level,
                    const BoxLayout& a_dest)
{
  const ProblemDomain domain = a_level.physDomain();
  define(a_level, a_dest, domain, IntVect::TheZeroVector());
}

void Copier::define(const DisjointBoxLayout& a_level,
                    const BoxLayout& a_dest,
                    const ProblemDomain& a_domain)
{
  define(a_level, a_dest, a_domain, IntVect::TheZeroVector());
}

void Copier::define(const DisjointBoxLayout& a_level,
                    const BoxLayout& a_dest,
                    const IntVect& a_ghost)
{
  const ProblemDomain domain = a_level.physDomain();
  define (a_level, a_dest, domain, a_ghost);
}

void Copier::define(const DisjointBoxLayout& a_level,
                    const BoxLayout& a_dest,
                    const ProblemDomain& a_domain,
                    const IntVect& a_ghost)
{
  assert(a_level.isClosed());
  assert(a_dest.isClosed());
  assert(a_level.checkPeriodic(a_domain));

  clear();
  buffersAllocated = false;
  //bool self = a_dest == a_level;  
  const DisjointBoxLayout& level= a_level;
  const BoxLayout&         dest = a_dest;

  // set up vector of dataIndexes to keep track of which
  // "to" boxes are not completely contained within the primary
  // domain.  these boxes are then candidates for filling by 
  // periodic images of the "from" data.
  Vector<DataIndex> periodicallyFilledToVect;

  // in order to cull which "from" data may be needed to 
  // fill the "to" data, keep track of the radius around the 
  // primary domain in which all these cells lie. 
  // do this by incrementally growing the domain box and 
  // keeping track of what this radius is.
  // just to make things simpler, start off with a radius of one
  Box grownDomainCheckBox = a_domain.domainBox();
  grownDomainCheckBox.grow(1);
  int periodicCheckRadius = 1;

  // since valid regions of the "from" DBL may also be outside
  // the primary domain, need to keep track of whether any of these
  // need to be checked separately.  
  Vector<DataIndex> periodicFromVect;
  // use same domain trick here as well
  Box grownFromDomainCheckBox = a_domain.domainBox();
  int periodicFromCheckRadius = 1;

  Box domainBox(a_domain.domainBox());
  bool isPeriodic = false;
  if (!domainBox.isEmpty())
    isPeriodic = a_domain.isPeriodic();
  
                    
  unsigned int myprocID = procID();


  // The following 4 for loops are the result of a performance optimization.
  // When increasing the size of the problem, we found that the code was 
  // looping over every destination box for every source box which was N1*N2
  // loop iterations (essentially an N-squared approach).
  // The following code attempts to simply reduce N1 and N2 by first separating 
  // the boxes (or LayoutIndexes to boxes) that reside on the current processor.
  // Then the loop to determine which boxes of the first list intersect with
  // which boxes of the second list can be done in N1' * N2' iterations,
  // where N1' is the reduced N1 and N2' is the reduced N2.
  // We have to break up the assigning of MotionItems into two separate
  // loops and be careful about the local copies.  These 4 loops are
  // significantly faster than the original for loop -- _especially_
  // for large problems.  (ndk)

#ifdef MPI  // don't need to do this in serial
  // make a vector of boxes (or LayoutIndexes to boxes) from destination layout 
  // that are known to reside on this processor.
  vector<DataIndex> vectorDestDI;
  vector<DataIndex> vectorDestOnProcDI;
  for(LayoutIterator to(a_dest.layoutIterator()); to.ok(); ++to) {
    vectorDestDI.push_back(DataIndex(to()));
    if (myprocID == dest.procID(to())) {
      vectorDestOnProcDI.push_back(DataIndex(to()));
    }
  }

  // make a vector of boxes (or LayoutIndexes to boxes) from "level"/src layout 
  // that are known to reside on this processor.
  vector<DataIndex> vectorLevelDI;
  vector<DataIndex> vectorLevelOnProcDI;
  for(LayoutIterator from(a_level.layoutIterator()); from.ok(); ++from) {
    vectorLevelDI.push_back(DataIndex(from()));
    if (myprocID == level.procID(from())) {
      vectorLevelOnProcDI.push_back(DataIndex(from()));
    }
  }
#else
  // in serial, it's not very interesting as it's all of them.
  vector<DataIndex> vectorDestOnProcDI;
  vector<DataIndex> vectorLevelDI;
  for(LayoutIterator to(a_dest.layoutIterator()); to.ok(); ++to) {
    vectorDestOnProcDI.push_back(DataIndex(to()));
  }
  for(LayoutIterator from(a_level.layoutIterator()); from.ok(); ++from) {
    vectorLevelDI.push_back(DataIndex(from()));
  }
#endif


  // loop over all dest/to DI's on my processor
  for(vector<DataIndex>::iterator vdi=vectorDestOnProcDI.begin();
      vdi != vectorDestOnProcDI.end(); ++vdi) {
    
    // at this point, i know myprocID == toProcID
    const DataIndex todi(*vdi);
    Box ghost(dest[todi]);
    ghost.grow(a_ghost);

    // then for each level/from DI, see if they intersect
    for(vector<DataIndex>::iterator vli = vectorLevelDI.begin();
        vli != vectorLevelDI.end(); ++vli) {
      
      const DataIndex fromdi(*vli);
      const unsigned int fromProcID = level.procID(fromdi);
      const Box& fromBox = level[fromdi];
      if(fromBox.bigEnd(0) < ghost.smallEnd(0)) {
        //can skip rest cuz we haven't gotten to something interesting
        continue;
      }

      if (ghost.intersectsNotEmpty(fromBox)) {
        Box box(ghost); // ??
        box&=fromBox;   // ??
        MotionItem* item = new (s_motionItemPool.getPtr())
          MotionItem(fromdi, todi, box);
        if(item == NULL) {
          MayDay::Error("Out of Memory in copier::define");
        }
        if(fromProcID == myprocID) { // local move
          m_localMotionPlan.push_back(item);
        } else {
          item->procID = fromProcID;
          m_toMotionPlan.push_back(item);
        }
      }
      if(fromBox.smallEnd(0) > ghost.bigEnd(0)) { 
        //can break out of loop, since we know that the smallEnd
        // of all the remaining boxes are lexigraphically beyond this ghosted box.
        break;      
      }

    }
  }

  // Don't need to worry about this in serial as we already
  // took care of the local copy motion items just above.  skip this.
#ifdef MPI
  // loop over all dest/to DI's 
  for(vector<DataIndex>::iterator vdi=vectorDestDI.begin();
      vdi != vectorDestDI.end(); ++vdi) {
    
    const DataIndex todi(*vdi);
    Box ghost(dest[todi]);
    ghost.grow(a_ghost);

    const unsigned int toProcID = dest.procID(todi);

    // then for each level/from DI on this processor, see if they intersect
    for(vector<DataIndex>::iterator vli = vectorLevelOnProcDI.begin();
        vli != vectorLevelOnProcDI.end(); ++vli) {

      // at this point, i know myprocID == fromProcID
   
      const DataIndex fromdi(*vli);
      const Box& fromBox = level[fromdi];

      if(fromBox.bigEnd(0) < ghost.smallEnd(0)) {
        //can skip rest cuz we haven't gotten to something interesting
        continue;
      }

      if (ghost.intersectsNotEmpty(fromBox)) {
        Box box(ghost); // ??
        box&=fromBox;   // ??

        if(toProcID == myprocID) { // local move
          // don't push back here!  or you will get two.
          //     we already did it above...
          //m_localMotionPlan.push_back(item);
        } else {
          MotionItem* item = new (s_motionItemPool.getPtr())
            MotionItem(fromdi, todi, box);
          if(item == NULL) {
            MayDay::Error("Out of Memory in copier::define");
          }

          item->procID = toProcID;
          m_fromMotionPlan.push_back(item);
        }
      }
      if(fromBox.smallEnd(0) > ghost.bigEnd(0)) {
        //can break out of loop, since we know that the smallEnd
        // of all the remaining boxes are lexigraphically beyond this ghosted box.
        break;      
      }      
    }
  }
#endif

  // put periodic intersection checking in here for "to" boxes
  if (isPeriodic) {
    for(LayoutIterator to(a_dest.layoutIterator()); to.ok(); ++to) {
      
      Box ghost(dest[to()]);
      ghost.grow(a_ghost);
      //unsigned int toProcID = dest.procID(to());  // unused variable
      
      // only do this if ghost box hangs over domain edge 
      if (!domainBox.contains(ghost)) { 
        // add the dataIndex for this box to the list 
        // of boxes which we need to come back to
        periodicallyFilledToVect.push_back(DataIndex(to()));
        // now check to see if we need to grow the 
        // periodic check radius
        if (!grownDomainCheckBox.contains(ghost)) {
          // grow the domainCheckBox until it contains ghost
          while (!grownDomainCheckBox.contains(ghost)) {
            grownDomainCheckBox.grow(1);
            periodicCheckRadius++;
          }
        } // end if we need to grow radius around domain
        
      } //end if ghost box is not contained in domain
    } // end if periodic
  }

  // Here ends the so-called N-squared optimizations.  the rest is unchanged. (ndk)
  
  // now do periodic checking, if necessary
  if (isPeriodic)
    {

      // the only "from" boxes we will need to check 
      // will be those within periodicCheckRadius of the
      // domain boundary. so, create a box to screen out
      // those which we will need to check.
      Box shrunkDomainBox = a_domain.domainBox();
      shrunkDomainBox.grow(-periodicCheckRadius);
      
      ShiftIterator shiftIt = a_domain.shiftIterator();
      IntVect shiftMult(domainBox.size());
      
      // now loop over "from" boxes
      for (LayoutIterator from(a_level.layoutIterator()); from.ok(); ++from)
        {
          // first check to see whether we need to look at this box
          const Box& fromBox = level[from()];
          
          if (!shrunkDomainBox.contains(fromBox))
            {
              unsigned int fromProcID = level.procID(from());

              // check to see if fromBox is contained in domain, 
              // if not, add it to the list of fromBoxes we need to 
              // go back and check separately to see if it will 
              // fill one of the "to" boxes
              if (!domainBox.contains(fromBox))
                {
                  periodicFromVect.push_back(DataIndex(from()));
                
                  if (!grownFromDomainCheckBox.contains(fromBox))
                    {
                      while (!grownFromDomainCheckBox.contains(fromBox))
                        {
                          grownFromDomainCheckBox.grow(1);
                          periodicFromCheckRadius++;
                        }
                    } // end if we need to grow domain check box
                } // end if fromBox is outside domain
              
                           
              // now loop over those "to" boxes which were not contained
              // in the domain
              for (int toRef=0; toRef<periodicallyFilledToVect.size(); toRef++)
                {
                  DataIndex toIndex = periodicallyFilledToVect[toRef];
                  unsigned int toProcID = dest.procID(toIndex);                  
                  
                  // don't worry about anything that doesn't involve this proc
                  if (toProcID != myprocID && fromProcID != myprocID)
                    {
                      // do nothing
                    } 
                  else
                    {
                      Box ghost(dest[toIndex]);
                      ghost.grow(a_ghost);
                      // now need to loop over shift vectors and look at images
                      for (shiftIt.begin(); shiftIt.ok(); ++shiftIt)
                        {
                          IntVect shiftVect(shiftIt()*shiftMult);
                          ghost.shift(shiftVect);
                          if (ghost.intersectsNotEmpty(fromBox)) // rarely happens
                            {
                              Box intersectBox(ghost);
                              intersectBox &= fromBox;
                              Box toBox(intersectBox);
                              toBox.shift(-shiftVect);
                              MotionItem* item = new (s_motionItemPool.getPtr())
                                MotionItem(DataIndex(from()), DataIndex(toIndex),
                                           intersectBox, toBox);
                              if (item == NULL) 
                                { 
                                  MayDay::Error("Out of Memory in copier::define");
                                }
                              if (toProcID == fromProcID) // local move
                                m_localMotionPlan.push_back(item);
                              else if (fromProcID == myprocID)
                                {
                                  item->procID = toProcID;
                                  m_fromMotionPlan.push_back(item);
                                }
                              else
                                {
                                  item->procID = fromProcID;
                                  m_toMotionPlan.push_back(item);
                                }
                              
                            } // end if shifted box intersects
                          
                          ghost.shift(-shiftVect);
                        } // end loop over shift vectors                  
                    } // end if either from box or to box are on this proc
                } // end loop over destination boxes
            } // end if source box is close to domain boundary
        } // end loop over destination boxes

      // now go back through the "from" boxes which were outside 
      // the domain and see if they intersect any toBoxes
      if (periodicFromVect.size() != 0)
        {
          // the only "to" boxes we will need to check 
          // will be those within periodicCheckRadius of the
          // domain boundary. so, create a box to screen out
          // those which we will need to check.
          shrunkDomainBox = a_domain.domainBox();
          shrunkDomainBox.grow(-periodicFromCheckRadius);
          
          // now loop over the "to" boxes 
          for (LayoutIterator to(a_dest.layoutIterator()); to.ok(); ++to)
            {
              // first check to see whether we need to look at this box
              Box ghost(dest[to()]);
              ghost.grow(a_ghost);
              
              if (!shrunkDomainBox.contains(ghost))
                {
                  unsigned int toProcID = a_dest.procID(to());
                  
                  // now loop over those "from" boxes which are not
                  // contained by the domain
                  for (int fromRef = 0; fromRef<periodicFromVect.size(); fromRef++)
                    {
                      DataIndex fromIndex = periodicFromVect[fromRef];
                      const Box& fromBox = level[fromIndex];
                      unsigned int fromProcID = level.procID(fromIndex);
                      
                      // don't worry about anything which doesn't involve
                      // this proc
                      if (toProcID != myprocID && fromProcID != myprocID)
                        {
                          // do nothing
                        } 
                      else
                        {
                          // now need to loop over shift vectors and look at images
                          for (shiftIt.begin(); shiftIt.ok(); ++shiftIt)
                            {
                              IntVect shiftVect(shiftIt()*shiftMult);
                              ghost.shift(shiftVect);
                              if (ghost.intersectsNotEmpty(fromBox))
                                {
                                  Box intersectBox(ghost);
                                  intersectBox &= fromBox;
                                  Box toBox(intersectBox);
                                  toBox.shift(-shiftVect);
                                  MotionItem* item = new (s_motionItemPool.getPtr())
                                    MotionItem(DataIndex(fromIndex), DataIndex(to()),
                                               intersectBox, toBox);
                                  if (item == NULL) 
                                    { 
                                      MayDay::Error("Out of Memory in copier::define");
                                    }
                                  if (toProcID == fromProcID) // local move
                                    m_localMotionPlan.push_back(item);
                                  else if (fromProcID == myprocID)
                                    {
                                      item->procID = toProcID;
                                      m_fromMotionPlan.push_back(item);
                                    }
                                  else
                                    {
                                      item->procID = fromProcID;
                                      m_toMotionPlan.push_back(item);
                                    }
                                  
                                } // end if shifted box intersects
                              
                              ghost.shift(-shiftVect);
                            } // end loop over shift vectors                  
                        } // end if either from box or to box are on this proc
                    } // end loop over "from" boxes
                } // end if destination box is close to domain boundary
            } // end loop over destination boxes
        } // end if any of the "From" boxes were outside the domain          
      
    } // end if we need to do anything for periodicity
}




int Copier::print() const
{
  std::cout << *this;
  return 0;
}

ostream& operator<<(ostream& os, const Copier& copier)
{
  os << "local("<<procID()<<"): ";
  for(CopyIterator it(copier, CopyIterator::LOCAL); it.ok(); ++it)
    {
      os << it().toRegion;
    }
  os << "\nfrom("<<procID()<<"): ";
  for(CopyIterator it(copier, CopyIterator::FROM); it.ok(); ++it)
    {
      os << it().fromRegion<<"["<<it().procID<<"]";
    }
  os << "\nto("<<procID()<<"): ";
  for(CopyIterator it(copier, CopyIterator::TO); it.ok(); ++it)
    {
      os << it().toRegion<<"["<<it().procID<<"]";
    }
  os<<"\n";
  return os;
}

Copier::~Copier()
{
  clear();
}

void Copier::setBufferAllocated(bool arg) const
{
  buffersAllocated  = arg;
}

bool Copier::bufferAllocated() const
{
  return buffersAllocated;
}
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef LEVELARRAY_H
#define LEVELARRAY_H


///Data over a disjoint union of rectangles

/**
 LevelData is-a BoxLayoutData.  It is built from a DisjointBoxLayout
 instead though, thus making the data in a DisjointBoxLayout uniquely
 defined for all spatial indices defined by the underlying
 DisjointBoxLayout.  It carries within it an inherent concept of
 ghost cells (which the user can choose to be empty).

  Since LevelData is-a BoxLayoutData, all the methods
  required of the template class T for BoxLayoutData are also
  required for LevelData. LevelData does not have any extra requirements.
*/

template<class T> class LevelData : public BoxLayoutData<T>
{
public:

  ///
  LevelData();

  ///
  LevelData(const DisjointBoxLayout& dp, int comps,
            const IntVect& ghost = IntVect::TheZeroVector(),
            const DataFactory<T>& a_factory = DefaultDataFactory<T>());

  virtual ~LevelData();
  ///

  ///
  virtual void define(const DisjointBoxLayout& dp, int comps, 
                      const IntVect& ghost = IntVect::TheZeroVector(),
                      const DataFactory<T>& a_factory = DefaultDataFactory<T>());


  ///
  /** 
    copy definer.  this LevelData thrown away and da's LevelData copied
    */
  virtual void define(const LevelData<T>& da, 
                      const DataFactory<T>& a_factory = DefaultDataFactory<T>());

  ///
  /**
    Copy definer.  'this' LevelData thrown away and da's LevelData copied.
    This LevelData is now defined for the range [0, comps.size()] and filled
    with the data in da from [comps.begin(), comps.end()]
    */
  virtual void define(const LevelData<T>& da, const Interval& comps,
                      const DataFactory<T>& a_factory = DefaultDataFactory<T>());

  ///
  virtual void copyTo(const Interval& srcComps, 
                      BoxLayoutData<T>& dest,
                      const Interval& destComps) const;

  ///
  virtual void copyTo(const Interval& srcComps,
                      BoxLayoutData<T>& dest,
                      const Interval& destComps,
                      const Copier& copier) const;
  ///
  /** same as copyTo that takes a BoxLayoutData, except that it fills the
    ghost cells of 'dest' with data from 'this' also. */
  virtual void copyTo(const Interval& srcComps, 
                      LevelData<T>& dest,
                      const Interval& destComps) const;

  ///
  virtual void copyTo(const Interval& srcComps,
                      LevelData<T>& dest,
                      const Interval& destComps,
                      const Copier& copier) const;
  ///
  virtual void exchange(const Interval& comps);

  ///
  virtual void exchange(const Interval& comps,
                        const Copier& copier);

  ///
  const IntVect& ghostVect() const { return m_ghost;}

  /*
    These functions will invoke error messages when invoked.
    C++ will ensure that constructors are not called for the
    base class by a user, but a 'define' function has no such protection,
    hence the need to prevent such usage.  A runtime error is not
    a perfect solution...(strong construction gets around this  *sigh*).
    classes that derive from LevelData will have to turn it's valid
    defines into runtime errors also and make it's own defines.  Thus
    taking over the job of the compiler.
    */

  ///
  virtual void define(const BoxLayout& dp, int comps,
                      const DataFactory<T>& factory);

  ///
  virtual void define(const BoxLayoutData<T>& da,
                      const DataFactory<T>& factory = DefaultDataFactory<T>());

  ///
  virtual void define(const BoxLayoutData<T>& da, const Interval& comps,
                      const DataFactory<T>& factory = DefaultDataFactory<T>());

  virtual void define(const BoxLayout& deadFunction);

  ///
  const DisjointBoxLayout& getBoxes() const
  {
    return m_disjointBoxLayout;
  }

  ///
  const DisjointBoxLayout& disjointBoxLayout() const
  {
    return m_disjointBoxLayout;
  }

protected:

  void makeItSo(const Interval&     a_srcComps, 
                const LevelData<T>& a_src,
                BoxLayoutData<T>&   a_dest,
                const Interval&     a_destComps,
                const Copier&       a_copier) const;

  DisjointBoxLayout m_disjointBoxLayout;

  IntVect   m_ghost;

  // attempt to speed up the periodic code
  //Copier m_exchangeCopier;

  //========================================================================
  //
  // data structures used by makeItSo when we have some
  // data that needs to be moved (ie. there are entries
  // in the 'FROM' or 'TO' CopyIterators)
  //
  void completePendingSends() const;

  void allocateBuffers(const LevelData<T>& a_src, 
                       const Interval& a_srcComps,
                       const BoxLayoutData<T>& a_dest,
                       const Interval& a_destComps,
                       const Copier&   a_copier) const;

  void writeSendDataFromMeIntoBuffers(const LevelData<T>& a_src, 
                                      const Interval& a_srcComps) const;

  void postSendsFromMe() const ;

  void postReceivesToMe() const ;

  void unpackReceivesToMe(BoxLayoutData<T>& a_dest, 
                          const Interval&   a_destComps) const ;


  mutable void*  m_sendbuffer; // pointer member OK here, 
                               // since LevelData<T> has no copy
  mutable size_t m_sendcapacity;
  mutable void*  m_recbuffer;  // pointer member OK here, 
                               // since LevelData<T> has no copy
  mutable size_t m_reccapacity;

#ifdef MPI

  struct bufEntry
  {
    void* bufPtr; // pointer into contiguous m_buffer
    size_t size;
    const MotionItem* item;
    unsigned int procID;
    bool operator < (const bufEntry& rhs) const
      { 
        if(procID == rhs.procID) 
          {
            const Box& left = item->toRegion;
            const Box& right= rhs.item->toRegion;
            if(left.smallEnd() == right.smallEnd())
              {
                return left.bigEnd().lexLT(right.bigEnd());
              }
            else
              {
                return item->toRegion < rhs.item->toRegion;
              }
          }
        //else
        return procID < rhs.procID;
      }
  };


  mutable std::vector<bufEntry> m_fromMe;
  mutable std::vector<bufEntry> m_toMe;

  mutable MPI_Request  *m_sendRequests, *m_receiveRequests;
  mutable MPI_Status   *m_sendStatus,   *m_receiveStatus;
  mutable int numSends, numReceives;

#endif

};


// ====== inlined function definitions ================

// As with the BoxLayoutData implementation file.  This file
// just gives the interface and the inline implmentations
// are given in LevelDataI.H




#endif
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//


using std::sort;

template<class T> inline
LevelData<T>::LevelData()
  :m_sendbuffer(NULL), m_sendcapacity(0),
   m_recbuffer(NULL), m_reccapacity(0)
{
#ifdef MPI
  numSends = 0;
  numReceives = 0;
#endif
}

template<class T> inline
LevelData<T>::LevelData(const DisjointBoxLayout& dp, int comps, const IntVect& ghost,
                        const DataFactory<T>& a_factory)
  : m_disjointBoxLayout(dp), m_ghost(ghost), m_sendbuffer(NULL),
    m_sendcapacity(0),  m_recbuffer(NULL),m_reccapacity(0)
{ 
#ifdef MPI
  numSends = 0;
  numReceives = 0;
#endif
  m_boxLayout = dp; 
  m_comps = comps;
  m_isdefined = true;

  if(!dp.isClosed())
    {
      MayDay::Error("non-disjoint DisjointBoxLayout: LevelData<T>::LevelData(const DisjointBoxLayout& dp, int comps)");
    }

  Interval interval(0, comps-1);
  allocateGhostVector(a_factory, ghost);
  setVector(*this, interval, interval);
  
 
}



// Since I need to thwart the user from invoking the
// 'define' methods that use a general BoxLayout, I cannot
// get at said functions myself now. Ha!  So, I have to recode
// them here.

template<class T> inline
void LevelData<T>::define(const DisjointBoxLayout& dp, int comps, const IntVect& ghost,
                          const DataFactory<T> & a_factory)
{
  m_isdefined = true;
  if(!dp.isClosed())
    {
      MayDay::Error("non-disjoint DisjointBoxLayout: LevelData<T>::define(const DisjointBoxLayout& dp,....)");
    }
  if(comps<=0)
    {
      MayDay::Error("LevelData::LevelData(const BoxLayout& dp, int comps)  comps<=0");
    }
  m_comps = comps;
  m_boxLayout = dp;

  m_disjointBoxLayout = dp;
  m_ghost = ghost;

  Interval interval(0, comps-1);
  allocateGhostVector(a_factory, ghost);
  setVector(*this, interval, interval);

#ifdef MPI  
  m_fromMe.resize(0);
  m_toMe.resize(0);
#endif

}


template<class T> inline
void LevelData<T>::define(const LevelData<T>& da,  const DataFactory<T> & a_factory)
{
  m_isdefined = true;
  if(this == &da) return;
  m_disjointBoxLayout = da.m_disjointBoxLayout;
  m_boxLayout  = da.m_disjointBoxLayout;
  m_comps     = da.m_comps;
  m_ghost     = da.m_ghost;

  Interval srcAnddest(0, m_comps-1);

  allocateGhostVector(a_factory, m_ghost);
  setVector(da, srcAnddest, srcAnddest);

#ifdef MPI
  m_fromMe.resize(0);
  m_toMe.resize(0);
#endif
}


template<class T> inline
void LevelData<T>::define(const LevelData<T>& da, const Interval& comps,
                          const DataFactory<T>& a_factory)
{
  m_isdefined = true;
  if(this == &da){
    MayDay::Error(" LevelData<T>::define(const LevelData<T>& da, const Interval& comps) called with 'this'");
  }
  assert(comps.size()>0);
  // this line doesn't make any sense!
  //assert(comps.end()<=m_comps);
  assert(comps.begin()>=0);

  m_disjointBoxLayout = da.m_disjointBoxLayout;
  m_boxLayout  = da.m_disjointBoxLayout;

  m_comps = comps.size();

  m_ghost = da.m_ghost;

  Interval dest(0, m_comps-1);

  allocateGhostVector(a_factory, m_ghost);

  setVector(da, comps, dest);

#ifdef MPI
  m_fromMe.resize(0);
  m_toMe.resize(0);
#endif
}

template<class T> inline
void LevelData<T>::copyTo(const Interval& srcComps, 
                          BoxLayoutData<T>& dest,
                          const Interval& destComps) const
{
  if((BoxLayoutData<T>*)this == &dest) return;

  if(boxLayout() == dest.boxLayout())
    {
      // parallel direct copy here, no communication issues
      for(DataIterator it(dataIterator()); it.ok(); ++it)
        {
          dest[it()].copy(box(it()), 
                          destComps,
                          box(it()),
                          this->operator[](it()),
                          srcComps);
        }
      return;
    }

  Copier copier(m_disjointBoxLayout, dest.boxLayout());
  copyTo(srcComps, dest, destComps, copier);
}

template<class T> inline
void LevelData<T>::copyTo(const Interval& srcComps, 
                          LevelData<T>& dest,
                          const Interval& destComps) const
{
  if(this == &dest){
    MayDay::Error("src == dest in copyTo function. Perhaps you want exchange ?");
  }

  if(boxLayout() == dest.boxLayout()  && dest.ghostVect() == IntVect::Zero)
    {
      // parallel direct copy here, no communication issues
      for(DataIterator it(dataIterator()); it.ok(); ++it)
        {
          dest[it()].copy(box(it()), 
                          destComps,
                          box(it()),
                          this->operator[](it()),
                          srcComps);
        }
      return;
    }

  Copier copier(m_disjointBoxLayout, dest.getBoxes(), dest.m_ghost);
  copyTo(srcComps, dest, destComps, copier);
}


template<class T> inline
void LevelData<T>::copyTo(const Interval& srcComps, 
                          BoxLayoutData<T>& dest,
                          const Interval& destComps,
                          const Copier& copier) const
{

  makeItSo(srcComps, *this, dest, destComps, copier);

}

template<class T> inline
void LevelData<T>::copyTo(const Interval& srcComps, 
                          LevelData<T>& dest,
                          const Interval& destComps,
                          const Copier& copier) const
{

  makeItSo(srcComps, *this, dest, destComps, copier);

}

template<class T> inline 
void LevelData<T>::exchange(const Interval& comps)
{
  // later on we can code this part as a direct algorithm
  // by copying and pasting the code from the Copier::define code
  // for now, just do the easy to debug approach.
  Copier copier(m_disjointBoxLayout, m_disjointBoxLayout, m_ghost);
  exchange(comps, copier);

  // if there aren't any ghost cells, there isn't really anything
  // to do here (also, if m_ghost == Zero, m_exchangeCopier
  // wasn't defined!
  //if (m_ghost != IntVect::Zero)
  //makeItSo(comps, *this, *this, comps, m_exchangeCopier);
}

template<class T> inline
void LevelData<T>::exchange(const Interval& comps,
                            const Copier& copier)
{
  makeItSo(comps, *this, *this, comps, copier);

}

template<class T> inline
void LevelData<T>::makeItSo(const Interval&   a_srcComps, 
                            const LevelData<T>& a_src,
                            BoxLayoutData<T>& a_dest,
                            const Interval&   a_destComps,
                            const Copier&     a_copier) const
{
  // The following five functions are nullOps in uniprocessor mode

  // Instead of doing this here, do it an end of makeItSo (ndk)
  //completePendingSends(); // wait for sends from possible previous operation

  // new evil logic to determine under what conditions we can just
  // re-use our messaging pattern and buffers from the last call to
  // makeItSo.  pretty elaborate, I know.  bvs
#ifdef MPI
  static Copier* lastCopier=NULL;
  
    if(T::preAllocatable() == 2 || !a_copier.bufferAllocated() || 
  	 (m_fromMe.size() + m_toMe.size() == 0) ||lastCopier != &a_copier){
	allocateBuffers(a_src,  a_srcComps,
					a_dest, a_destComps,
					a_copier);  //monkey with buffers, set up 'fromMe' and 'toMe' queues
		a_copier.setBufferAllocated(true);
	  }
	lastCopier = (Copier*)(&a_copier);
#endif

  writeSendDataFromMeIntoBuffers(a_src, a_srcComps);


  // If there is nothing to recv/send, don't go into these functions 
  // and allocate memory that will not be freed later.  (ndk)
  // The #ifdef MPI is for the m_toMe and m_fromMe
#ifdef MPI
  if (m_toMe.size() > 0) {
    postReceivesToMe(); // all non-blocking
  }

  if (m_fromMe.size() > 0) {
    postSendsFromMe();  // all non-blocking
  }
#endif

  //  computation that could occur during communication should really
  //  go here somehow.  while all the non-blocking sending and receiving is
  //  going on.  
  //
  //  my thought is to return from this function at this point an object
  //  that encapsulates the argument list above.  
  //  a "ChomboMessaging" object.
  //  The user can keep a reference
  //  to this object and do computations.  When they reach the limit of what
  //  they can compute without this communication completing, they call the
  //  "finalize()" function of their ChomboMessaging object and the rest of this
  //  code below gets executed.
  //  a real question though is: is there really enough computation to do while
  //  messaging is going on to justify the effort, and what machines really have
  //  good asynchronous messaging to make the work worthwhile.
  //  
  //  the other approach is to more finely decompose the overlapping of
  //  messaging and computation by using the ChomboMessaging object in the
  //  DataIterator construction.  The DataIterator returns T objects as they
  //  are completed from messaging.  This preserves almost all of the Chombo
  //  code as is but would be mucho tricky to actually implement and might only
  //  gain little.  This would not be a thing to try unitl Chombo is
  //  heavily instrumented for performance measuring.  in this design, unpackRecievesToMe()
  //  would become a complicated process interwoven with a DataIterator.

  //  postReceivesToMe();

  // perform local copy
  for(CopyIterator it(a_copier, CopyIterator::LOCAL); it.ok(); ++it)
    {
      const MotionItem& item = it();
      a_dest[item.toIndex].copy(item.fromRegion, 
                                a_destComps,
                                item.toRegion,
                                a_src[item.fromIndex],
                                a_srcComps);
    }

  // Uncomment and Move this out of unpackReceivesToMe()  (ndk)
  completePendingSends(); // wait for sends from possible previous operation

  unpackReceivesToMe(a_dest, a_destComps); // nullOp in uniprocessor mode

}

template<class T> inline
void LevelData<T>::define(const BoxLayout& dp, int comps,  const DataFactory<T>& a_factory)
{
  MayDay::Error("LevelData<T>::define called with BoxLayout input");
}

template<class T> inline
void LevelData<T>::define(const BoxLayout& dp)
{
  MayDay::Error("LevelData<T>::define called with BoxLayout input");
}

template<class T> inline
void LevelData<T>::define(const BoxLayoutData<T>& da, const DataFactory<T>& a_factory )
{
  MayDay::Error("LevelData<T>::define called with BoxLayout input");
}

template<class T> inline
void LevelData<T>::define(const BoxLayoutData<T>& da, const Interval& comps,
                          const DataFactory<T>& a_factory)
{
  MayDay::Error("LevelData<T>::define called with BoxLayout input");
}

template<class T> inline
LevelData<T>::~LevelData()
{
  completePendingSends();
  free(m_sendbuffer);
  free(m_recbuffer);
}

#ifndef MPI
// uniprocessor version of all these nullop functions.
template<class T> inline
void LevelData<T>::completePendingSends() const
{;}

template<class T> inline
void LevelData<T>::allocateBuffers(const LevelData<T>& a_src, 
                                   const Interval& a_srcComps,
                                   const BoxLayoutData<T>& a_dest,
                                   const Interval& a_destComps,
                                   const Copier&   a_copier) const
{;}  

template<class T> inline
void LevelData<T>::writeSendDataFromMeIntoBuffers(const LevelData<T>& a_src, 
                                                  const Interval&     a_srcComps) const
{;}

template<class T> inline
void LevelData<T>::postSendsFromMe() const
{;}

template<class T> inline
void LevelData<T>::postReceivesToMe() const
{;}

template<class T> inline
void LevelData<T>::unpackReceivesToMe(BoxLayoutData<T>& a_dest, 
                                      const Interval&   a_destComps) const
{;}

#else

// MPI versions of the above codes.

template<class T> inline
void LevelData<T>::completePendingSends() const
{
  if(numSends > 0){
    int result = MPI_Waitall(numSends, m_sendRequests, m_sendStatus);
    if(result != MPI_SUCCESS)
      {
        //hell if I know what to do about failed messaging here
      }

    delete[] m_sendRequests;
    delete[] m_sendStatus;
  }
  numSends = 0;
}

template<class T> inline
void LevelData<T>::allocateBuffers(const LevelData<T>& a_src, 
                                   const Interval& a_srcComps,
                                   const BoxLayoutData<T>& a_dest,
                                   const Interval& a_destComps,
                                   const Copier&   a_copier) const
{
  m_fromMe.resize(0);
  m_toMe.resize(0);
  size_t sendBufferSize = 0;
  size_t recBufferSize  = 0;
  // two versions of code here.  one for preAllocatable T, one not.

  //  T dummy;
  for(CopyIterator it(a_copier, CopyIterator::FROM); it.ok(); ++it)
    {
      const MotionItem& item = it();
      bufEntry b;
      b.item = &item;
      b.size = a_src[item.fromIndex].size(item.fromRegion, a_srcComps);
      sendBufferSize+=b.size;
      b.procID = item.procID;
      m_fromMe.push_back(b);
    }
  sort(m_fromMe.begin(), m_fromMe.end());
  for(CopyIterator it(a_copier, CopyIterator::TO); it.ok(); ++it)
    {
      const MotionItem& item = it();
      bufEntry b;
      b.item = &item;
      if (!(T::preAllocatable() == 2))
        {
          b.size = a_dest[item.toIndex].size(item.toRegion, a_destComps);
          recBufferSize+=b.size;
        }
      b.procID = item.procID;
      m_toMe.push_back(b);
    }
  sort(m_toMe.begin(), m_toMe.end());

  if(T::preAllocatable() == 2) // dynamic allocatable, need two pass
    {
      // in the non-preallocatable case, I need to message the
      // values for the m_toMe[*].size 
      if(m_fromMe.size() > 0)
        {
          MPI_Request nullrequest;

          int lastProc = -1;
          int messageIndex = 0;
          for(int i=0; i<m_fromMe.size(); ++i)
            {
              bufEntry& b = m_fromMe[i];
              if(b.procID == lastProc) messageIndex++;
              else                     messageIndex = 0;
              lastProc = b.procID;
              MPI_Isend(&(b.size), 1, MPI_INT, b.procID, 
                        messageIndex, Chombo_MPI::comm, &(nullrequest));
              MPI_Request_free(&(nullrequest));  // we don't wait on these sends, assume
              //    this memory really doesn't go anywhere.  Most MPI systems will fast message
              //    a single integer message
            }
        }
      if(m_toMe.size() > 0)
        {
          m_receiveRequests = new MPI_Request[m_toMe.size()];
          m_receiveStatus   = new MPI_Status[m_toMe.size()];
          int lastProc = -1;
          int messageIndex = 0;
          for(int i=0; i<m_toMe.size(); ++i)
            {
              bufEntry& b = m_toMe[i];
              if(b.procID == lastProc) messageIndex++;
              else                     messageIndex = 0;
              lastProc = b.procID;
              MPI_Irecv(&(b.size), 1, MPI_INT, b.procID, 
                        messageIndex, Chombo_MPI::comm, m_receiveRequests+i);
            }

          int result = MPI_Waitall(m_toMe.size(), m_receiveRequests, m_receiveStatus);
          if(result != MPI_SUCCESS)
            {
              MayDay::Error("First pass of two-phase communication failed");
            }
          for(int i=0; i<m_toMe.size(); ++i)  recBufferSize+= m_toMe[i].size;
          delete[] m_receiveRequests;
          delete[] m_receiveStatus;
        }
    }

  // allocate send and receveive buffer space.

  if(sendBufferSize > m_sendcapacity)
    {
      free(m_sendbuffer);
      m_sendbuffer = malloc(sendBufferSize);
      if(m_sendbuffer == NULL)
        {
          MayDay::Error("Out of memory in LevelData::allocatebuffers");
        }
      m_sendcapacity = sendBufferSize;
    }

  if(recBufferSize > m_reccapacity)
    {
      free(m_recbuffer);
      m_recbuffer = malloc(recBufferSize);
      if(m_recbuffer == NULL)
        {
          MayDay::Error("Out of memory in LevelData::allocatebuffers");
        }
      m_reccapacity = recBufferSize;
    }


  /*
  pout()<<"\n";
  for(int i=0; i<m_fromMe.size(); i++) 
    pout()<<m_fromMe[i].item->region<<"{"<<m_fromMe[i].procID<<"}"<<" ";
  pout() <<"::::";
  for(int i=0; i<m_toMe.size(); i++) 
    pout()<<m_toMe[i].item->region<<"{"<<m_toMe[i].procID<<"}"<<" ";
  pout() << endl;
  */

  char* nextFree = (char*)m_sendbuffer;
  if(m_fromMe.size() > 0)
  {
    for(unsigned int i=0; i<m_fromMe.size(); ++i)
      {
        m_fromMe[i].bufPtr = nextFree;
        nextFree += m_fromMe[i].size;
      }
  }

  nextFree = (char*)m_recbuffer;
  if(m_toMe.size() > 0)
  {
    for(unsigned int i=0; i<m_toMe.size(); ++i)
      {
        m_toMe[i].bufPtr = nextFree;
        nextFree += m_toMe[i].size;
      }
  }

  // since fromMe and toMe are sorted based on procID, messages can now be grouped
  // together on a per-processor basis.

}



template<class T> inline
void LevelData<T>::writeSendDataFromMeIntoBuffers(const LevelData<T>& a_src, 
                                                  const Interval&     a_srcComps) const
{


  for(unsigned int i=0; i<m_fromMe.size(); ++i)
    {
      const bufEntry& entry = m_fromMe[i];
      a_src[entry.item->fromIndex].linearOut(entry.bufPtr, entry.item->fromRegion, a_srcComps);
    }

}

template<class T> inline
void LevelData<T>::postSendsFromMe() const
{

  // now we get the magic of message coalescence
  // fromMe has already been sorted in the allocateBuffers() step.

  numSends = m_fromMe.size();
  if(numSends > 1){
  for(unsigned int i=m_fromMe.size()-1; i>0; --i)
    {
      if(m_fromMe[i].procID == m_fromMe[i-1].procID) 
        {
          numSends--;
          m_fromMe[i-1].size+=m_fromMe[i].size;
          m_fromMe[i].size = 0;
        }
    }
  }
  m_sendRequests = new MPI_Request[numSends];
  m_sendStatus = new MPI_Status[numSends];


  unsigned int next=0;
  for(int i=0; i<numSends; ++i)
    {
      const bufEntry& entry = m_fromMe[next];
      // cout<<procID()<< ": sending message of "<<entry.size;
      // cout<<" to proc "<<  entry.procID<<endl;
      MPI_Isend(entry.bufPtr, entry.size, MPI_BYTE, entry.procID, 
                0, Chombo_MPI::comm, m_sendRequests+i);
      ++next;
      while(next < m_fromMe.size() && m_fromMe[next].size == 0) ++next;
    }


}

template<class T> inline
void LevelData<T>::postReceivesToMe() const
{
  numReceives = m_toMe.size();

  if(numReceives > 1){
  for(unsigned int i=m_toMe.size()-1; i>0; --i)
    {
      if(m_toMe[i].procID == m_toMe[i-1].procID) 
        {
          numReceives--;
          m_toMe[i-1].size+=m_toMe[i].size;
          m_toMe[i].size = 0;
        }
    }
  }
  m_receiveRequests = new MPI_Request[numReceives];
  m_receiveStatus = new MPI_Status[numReceives];


  unsigned int next=0;
  for(int i=0; i<numReceives; ++i)
    {
      const bufEntry& entry = m_toMe[next];
      //cout<<procID()<< ": receiving message of "<<entry.size;
      //cout<<" from proc "<<  entry.procID<<endl;
      MPI_Irecv(entry.bufPtr, entry.size, MPI_BYTE, entry.procID, 
                0, Chombo_MPI::comm, m_receiveRequests+i);
      ++next;
      while(next < m_toMe.size() && m_toMe[next].size == 0) ++next;
    }

}


template<class T> inline
void LevelData<T>::unpackReceivesToMe(BoxLayoutData<T>& a_dest, 
                                      const Interval&   a_destComps) const
{

  if(numReceives > 0){
    int result = MPI_Waitall(numReceives, m_receiveRequests, m_receiveStatus);
    if(result != MPI_SUCCESS)
      {
        //hell if I know what to do about failed messaging here
      }

    for(unsigned int i=0; i<m_toMe.size(); ++i)
      {
        const bufEntry& entry = m_toMe[i];
        a_dest[entry.item->toIndex].linearIn(entry.bufPtr, entry.item->toRegion, a_destComps);
      }

    delete[] m_receiveRequests;
    delete[] m_receiveStatus;
  }
  numReceives = 0;
}

#endif








/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL


#ifndef HDF5PORTABLE_H
#define HDF5PORTABLE_H

#ifdef CH_CYGWIN
//       because Cygwin doesn't distinguish case in filenames and Chombo 
//       has its own "HDF5.H". -dbs]
#else
#endif

template <class T>
void dataTypes(Vector<hid_t>& a_types, const T& a_dummySpecializationArg);

template <class T>
void dataSize(const T& item, Vector<int>& a_sizes, 
              const Box& box, const Interval& comps);

template <class T>
void write(const T& item, Vector<void*>& a_allocatedBuffers, 
           const Box& box, const Interval& comps);

template <class T>
void read(T& item, Vector<void*>& a_allocatedBuffers, const Box& box, 
          const Interval& comps);

template <class T>
const char* name(const T& a_dummySpecializationArg);

// These are the default template declarations for the HDF5 I/O
// tempalte functions needed by T objects for binary IO.  The default
// implementations use the linearOut/linearIn/size functions of the
// T tempalte classes to procide non-portable binary IO.
//     A user can add their own template specializations for
// portable binary IO abilities. 
//
//  header file.
// 
//
//
//  http://hdf.ncsa.uiuc.edu/HDF5/
//
// available datatypes in HDF5 
//
//            Example
//                                                  Corresponding C Type
//            H5T_NATIVE_CHAR
//                                                  char
//            H5T_NATIVE_UCHAR
//                                                  unsigned char
//            H5T_NATIVE_SHORT
//                                                  short
//            H5T_NATIVE_USHORT
//                                                  unsigned short
//            H5T_NATIVE_INT
//                                                  int
//            H5T_NATIVE_UINT
//                                                  unsigned
//            H5T_NATIVE_LONG
//                                                  long
//            H5T_NATIVE_ULONG
//                                                  unsigned long
//            H5T_NATIVE_FLOAT
//                                                  float
//            H5T_NATIVE_DOUBLE
//                                                  double
//
// and some special ones defined for Chombo:
//
//            H5T_NATIVE_REAL                       Real
//
//            box_id                                Box
//
//            intvext_id                            IntVect
#endif
/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL

#ifdef HDF5  // if you don't have HDF5, then this file is useless

#ifndef HDF5_H
#define HDF5_H

using std::cout;
using std::endl;

#ifdef MPI
#endif

// hdf5 #defines inline.... duh!
#undef inline
#ifdef AMRNODEELLIPTIC
#endif

using std::map;

class HDF5Handle;

// HDF5.H
// ============


/// user-friendly function to write out data on a AMR level
/**
   all data is IN data.
   a_handle:  handle open and ready.
   a_data:    data, what else do you want to know ?
   a_dx:      the grid spacing at this level
   a_dt:      the timestep size that was last completed
   a_time:    the time of this level (might not be the same as other levels)
   a_domain:  the problem domain, represented at this level of refinement
   a_refRatio:the refinement of a_level+1 wrt a_level. for vis systems it
   would probably help if you use 1 for a_level==max_level.
*/
template <class T>
int writeLevel(HDF5Handle& a_handle, 
               const int& a_level, 
               const LevelData<T>& a_data,
               const Real& a_dx, 
               const Real& a_dt, 
               const Real& a_time, 
               const Box& a_domain,
               const int& a_refRatio,
               const IntVect& outputGhost = IntVect::Zero);

template <class T>
int readLevel(HDF5Handle& a_handle,  
              const int& a_level, 
              LevelData<T>& a_data, 
              Real& a_dx, 
              Real& a_dt, 
              Real& a_time, 
              Box& a_domain,
              int& a_refRatio, 
              const Interval& a_comps = Interval(), 
              const IntVect& ghost = IntVect::Zero, 
              bool setGhost = false);


// More basic HDF5 functions.  These can be used at a persons own
// discretion.  Refer to the User's Guide for a description of how these
// functions interact with the convenience functions writeLevel, etc.

/// writes BoxLayout to HDF5 file.
/** 
    Writes BoxLayout to HDF5 file.  Only one BoxLayout per group is permitted, this operation overwrites
    previous entries.
    This operation assumes boxes are cell-centered.\\
    returns: success:    0\\
    HDF5 error: negative error code.\\
*/
int write(HDF5Handle& a_handle, 
          const BoxLayout& a_layout);

/// writes a BoxLayoutData<T> to an HDF5 file.
/**
   writes a BoxLayoutData<T> to an HDF5 file.\\
   returns: success:    0\\
   HDF5 error: negative error code.\\
*/
template <class T>
int write(HDF5Handle& a_handle, 
          const BoxLayoutData<T>& a_data, 
          const std::string& a_name,
          const IntVect& outputGhost = IntVect::Zero);

/// writes a LevelData<T> to an HDF5 file.
/**
   Writes a LevelData<T> to an HDF5 file.
   the DisjointBoxLayout is not written out with the data, the user is required to
   handle that object seperately. (see the "read" interface below).\\
   returns: success:    0\\
   HDF5 error: negative error code.\\
*/
template <class T>
int write(HDF5Handle& a_handle, 
          const LevelData<T>& a_data, 
          const std::string& a_name,
          const IntVect& outputGhost = IntVect::Zero);

/// reads Vector<Box> from location specified by a_handle.
/**  
     Reads BoxLayout from the group specified by a_handle.  
     Only one BoxLayout per group is permitted, this operation overwrites.
     This operation assumes boxes are cell-centered.\\
     returns: success:    0\\
     HDF5 error: negative error code.\\
*/
int read(HDF5Handle& a_handle, 
         Vector<Box>& boxes);

/// reads the set of Boxes out from the level_* groups of a Chombo HDF5 AMR file
/**
   goes to all groups named level_n for level n = 0 to numLevel and fills the
   vector of boxes.  HDF5 Handle must be set to the root group (the default location when the
   file is just opened.  

   A handy *get everything* version of int read(HDF5Handle& a_handle, Vector<Box>& boxes)
*/
int readBoxes(HDF5Handle& a_handle,
              Vector<Vector<Box> >& boxes);
 

/// FArrayBox-at-a-time read function.  FArrayBox gets redefined in the function.  Reads data field named by a_dataName.
/**  FArrayBox gets redefined in the function.
     it will be sized to the box at the [level,boxNumber] position and have components
     the go from [0,a_components.size()-1]

     some meta-data has to be read again and again internally to do this function, but
     it should make some out-of-core tools possible.
*/

int readFArrayBox(HDF5Handle& a_handle,
                  FArrayBox&  a_fab,
                  int a_level,
                  int a_boxNumber,
                  const Interval& a_components,
                  const std::string& a_dataName = "data" );

/// read BoxLayoutData named a_name from location specified by a_handle.
/** 
    Read BoxLayoutData named a_name from location specified by a_handle.  User must supply the correct BoxLayout for this function if redefineData == true.  \\
    returns: success:      0\\
    bad location: 1\\
    HDF5 error:   negative error code.\\
*/
template <class T>
int read(HDF5Handle& a_handle, 
         BoxLayoutData<T>& a_data, 
         const std::string& a_name, 
         const BoxLayout& a_layout,
         const Interval&  a_comps = Interval(),
         bool redefineData = true);

/// read LevelData named a_name from location specified by a_handle.
/** 
    Read LevelData named a_name from location specified by a_handle.  User must supply the correct BoxLayout for this function if redefineData == true.  \\
    returns: success:      0\\
    bad location: 1\\
    HDF5 error:   negative error code.\\
*/
template <class T>
int read(HDF5Handle& a_handle, 
         LevelData<T>& a_data, 
         const std::string& a_name, 
         const DisjointBoxLayout& a_layout,
         const Interval& a_comps = Interval(),
         bool redefineData = true);

/// Handle to a particular group in an HDF file.  
/** 
    HDF5Handle is a handle to a particular group in an HDF file.  Upon
    construction, group is defined to be the root.  All data is
    written and read assuming the native representations for data on
    the architecture it is running on.  When a file is opened, these
    settings are checked and an error is flagged when things don't
    match up.  It is the USER'S responsibility to close() this object
    when it is no longer needed.

*/
class HDF5Handle
{
public:
  ///
  /**
     Enumeration of opening modes for HDF5 files.  \\

     CREATE: file is created if it didn't exist, or an existing file of
     the same name is clobbered.\\

     OPEN_RDONLY: existing file is opened in read-only mode.  If the
     file doesn't already exist then open fails and isOpen() returns
     false.\\

     OPEN_RDWR: existing file is opened in read-write mode.  If the file
     doesn't already exist then open fails and isOpen() returns false.\\ 

  */
  enum mode {CREATE, OPEN_RDONLY, OPEN_RDWR};

  ///    {\bf constructor}

  ///
  /**
     Default constructor.  User must call open() prior to using
     constructed object.

  */
  HDF5Handle();

  /// 
  /** Opens file and sets the current group to the root "/" group. \\

  if mode == CREATE, then file is created if it didn't exist, or an
  existing file of the same name is clobbered.\\

  if mode == OPEN_*, then existing file is opened, if the file doesn't
  already exist then open fails and isOpen() returns false.\\
  */
  HDF5Handle(const std::string& a_filename, mode a_mode);

  /// {\bf File functions}

  /// 
  /** 
      Opens file and sets the current group of this HDF5Handle to the
      root "/" group.  File that this HDF5Handle previously pointed at is
      NOT closed, that is the users responsibility.\\

      if mode == CREATE, then file is created if it didn't exist, or an
      existing file of the same name is clobbered.\\

      if mode == OPEN_*, then existing file is opened, if the file doesn't
      already exist then open fails and isOpen() returns false.\\

      returns:\\
      0  on success\\
      negative number if file open failed (return code from HDF5)\\
      1  file does not appear to contain datacheck info, probably not a Chombo file\\
      2  on data bit size differences between code and file.\\

      aborts on SpaceDim not matching between code and file\\

  */
  int open(const std::string& a_filename, mode a_mode);

  /// 
  /**
     A NULL or failed constructed HDF5Handle will return false.
  */
  bool isOpen() const;

  ///  
  /**
     Closes the file.  Must be called to close file.  Files are not
     automatically closed.  

  */
  void close();

  /// {\bf Group functions}

  /// 
  /**
     Sets the current group to be "/level_x" where x=a_level.
  */
  void setGroupToLevel(int a_level);

  /// 
  /** 
      Set group to users choice, referenced from file root.
      groupAbsPath will look like a Unix file path:
      "/mySpecialData/group1/" "/" is the root group of the
      file. returns a negative value on failure

  */
  int setGroup(const std::string& groupAbsPath);

  ///
  /**
     Returns name of current group.  groupAbsPath will look like a Unix
     file path: "/mySpecialData/group1/" "/" is the root group of the
     file.  

  */
  const std::string& getGroup() const;

  const hid_t& fileID() const;
  const hid_t& groupID() const;
  static hid_t box_id;
  static hid_t intvect_id;
  static hid_t realvect_id;

private:

  HDF5Handle(const HDF5Handle&);
  HDF5Handle& operator=(const HDF5Handle&);

  hid_t         m_fileID;
  hid_t         m_currentGroupID;
  bool          m_isOpen;
  std::string   m_filename; // keep around for debugging
  std::string   m_group;
  int           m_level;

  //  static hid_t  file_access;
  static bool   initialized;
  static void   initialize();

};

/// data to be added to HDF5 files.
/**
   HDF5HeaderData is a wrapper for some data maps to be added to HDF5
   files.  instead of an overdose of access functions, the maps are
   made public and they can be manipulated by the user at will.  They
   maintain type safety.

   to add a Real data entry, a user can simply program as follows:

   \begin{verbatim}
   Real dx;
   .
   .
   HDF5HeaderData metaData;
   metaData.m_real["dx"] = dx;
   \end{verbatim}

   If "dx" already existed, then it is overwritten, otherwise an entry is
   created and added with the new value;

   To search for entries, the user does the following:
   \begin{verbatim}

   HDF5HeaderData metaData;
   HDF5Handle currentStep(filename);
   currentStep.setGroupToLevel(0);
   metaData.readFromFile(currentStep);
   if(metaData.m_intvect.find("ghost") != metaData.m_intvect.end())
   ghost = metaData.m_intvect["ghost"];
   else
   ghost = defaultGhostIntVect;

   \end{verbatim}

   A user can skip the check for existence if they have reason to "know" the
   data will be there.  It is just good coding practice.

   To erase an entry, you can use:
   \begin{verbatim}
   metaData.m_real.erase("dx");
   \end{verbatim}

*/
class HDF5HeaderData
{
public:

  /// 
  /** 
      Writes this HDF5HeaderData's current attribute list to the
      current group in 'file.'  Returns 0 on success, returns the
      error code from HDF5 on failure.

  */
  int writeToFile(HDF5Handle& file) const;

  /// 
  /** 
      Reads into this HDF5HeaderData's attribute list from file.  Read
      process is add/change, does not remove key-value pairs. Reads
      from current group.  Returns 0 on success, positive number if a
      particular member of group caused an error, negative on general
      error. 

  */
  int readFromFile(HDF5Handle& file);

  ///
  void clear();

  ///
  map<std::string, Real>        m_real;

  ///
  map<std::string, int>         m_int;

  ///
  map<std::string, std::string> m_string;

  ///
  map<std::string, IntVect>     m_intvect;

  ///
  map<std::string, Box>         m_box;

  ///
  map<std::string, RealVect>    m_realvect;

  //users should not need these functions in general

  int writeToLocation(hid_t loc_id) const;
  int readFromLocation(hid_t loc_id);

private:
  static herr_t attributeScan(hid_t loc_id, const char *name, void *opdata);
};

extern "C"{
  herr_t HDF5HeaderDataattributeScan(hid_t loc_id, const char *name, void *opdata);
}

std::ostream& operator<<(std::ostream& os, const HDF5HeaderData& data);


//=============================================================================
//
// end of declarations.
//
//=============================================================================

/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL



// non-user code used in implementation of communication

struct OffsetBuffer
{
  Vector<int> index;
  Vector<Vector<int> > offsets;
  void operator=(const OffsetBuffer& rhs);
};

ostream& operator<<(ostream& os, const OffsetBuffer& ob);

//OffsetBuffer specialization of linearSize
template < >
int linearSize(const OffsetBuffer& a_input); 

//OffsetBuffer specialization of linearIn
template < >
void linearIn(OffsetBuffer& a_outputT, const void* const a_inBuf); 

//OffsetBuffer specialization of linearOut
template < >
void linearOut(void* const a_outBuf, const OffsetBuffer& a_inputT); 

template < > int linearSize(const Vector<OffsetBuffer>& a_input);
template < > void linearIn(Vector<OffsetBuffer>& a_outputT, const void* const inBuf);
template < > void linearOut(void* const a_outBuf, const Vector<OffsetBuffer>& a_inputT);

// First, template specializations for read/write for FArrayBox.

template <>
inline void dataTypes(Vector<hid_t>& a_types, const BaseFab<int>& dummy)
{  
  a_types.resize(1); 
  a_types[0] = H5T_NATIVE_INT;
}

/* since many compilers choke on the proper template specialization syntax
   I am forced to pass in a dummy specialization argument
*/
template <>
inline void dataTypes(Vector<hid_t>& a_types, const FArrayBox& dummy)
{  a_types.resize(1); a_types[0] = H5T_NATIVE_REAL;}

#ifdef AMRNODEELLIPTIC
template <>
inline void dataTypes(Vector<hid_t>& a_types, const NodeFArrayBox& dummy)
{
  a_types.resize(1);
  a_types[0] = H5T_NATIVE_REAL;
}
#endif

template <>
inline void dataSize(const BaseFab<int>& item, Vector<int>& a_sizes, 
                     const Box& box, const Interval& comps)
{
  a_sizes[0] = box.numPts() * comps.size();
}

template <>
inline void dataSize(const FArrayBox& item, Vector<int>& a_sizes, 
                     const Box& box, const Interval& comps)
{
  a_sizes[0] = box.numPts() * comps.size();
}

#ifdef AMRNODEELLIPTIC
template <>
inline void dataSize(const NodeFArrayBox& item, Vector<int>& a_sizes, 
                     const Box& box, const Interval& comps)
{
  Box boxNodes = surroundingNodes(box);
  a_sizes[0] = boxNodes.numPts() * comps.size();
}
#endif

template <>
inline const char* name(const FArrayBox& a_dummySpecializationArg)
{
  // Attempt to get rid of warnings on IBM...
  //static const char* name = "FArrayBox";
  const char* name = "FArrayBox";
  return name;
}

#ifdef AMRNODEELLIPTIC
template <>
inline const char* name(const NodeFArrayBox& a_dummySpecializationArg)
{
  // Attempt to get rid of warnings on IBM...
  //static const char* name = "NodeFArrayBox";
   const char* name = "NodeFArrayBox";
  return name;
}
#endif

template <>
inline const char* name(const BaseFab<int>& a_dummySpecializationArg)
{
  // Attempt to get rid of warnings on IBM...
  //static const char* name = "BaseFab<int>";
  const char* name = "BaseFab<int>";
  return name;
}

//
// now, generic, non-binary portable version of template functions
// for people who just want ot rely on linearIn/linearOut

template <class T>
inline void dataTypes(Vector<hid_t>& a_types, const T& dummy)
{  a_types.resize(1); a_types[0] = H5T_NATIVE_CHAR;}

template <class T>
inline void dataSize(const T& item, Vector<int>& a_sizes, 
                     const Box& box, const Interval& comps)
{
  a_sizes[0] = item.size(box, comps);
}


template <class T>
inline void write(const T& item, Vector<void*>& a_allocatedBuffers, 
                  const Box& box, const Interval& comps)
{
  item.linearOut(a_allocatedBuffers[0], box, comps);
}

template <class T>
inline void read(T& item, Vector<void*>& a_allocatedBuffers, 
                 const Box& box, const Interval& comps)
{
  item.linearIn(a_allocatedBuffers[0], box, comps);
}

template <class T>
inline const char* name(const T& a_dummySpecializationArg)
{
  static const char* name = "unknown";
  return name;
}

template <class T>
void getOffsets(Vector<Vector<long long> >& offsets, const BoxLayoutData<T>& a_data,  
                int types, const Interval& comps, const IntVect& outputGhost)
{


  const BoxLayout& layout = a_data.boxLayout();
  offsets.resize(types, Vector<long long>(layout.size()+1));
  //  offsets.resize(layout.size() + 1, Vector<long long>(types));
  for(int t=0; t<types; t++) offsets[t][0] = 0;
  Vector<int> thisSize(types);
  if(T::preAllocatable()==0)
    { // static preAllocatable
      T dummy;
      unsigned int index = 1;
      for(LayoutIterator it(layout.layoutIterator()); it.ok(); ++it)
        {
          Box region = layout[it()];
          region.grow(outputGhost);
          dataSize(dummy, thisSize, region, comps);
          for(unsigned int i=0; i<thisSize.size(); ++i) 
            {
              //offsets[index][i] = offsets[index-1][i] + thisSize[i];
              offsets[i][index] = offsets[i][index-1] + thisSize[i];
            }
          ++index;
        }
    }
  else
    { // symmetric and dynamic preallocatable need two pass I/O
      OffsetBuffer buff;
      //int index = 0;
      for(DataIterator dit(a_data.dataIterator()); dit.ok(); ++dit)
        {
		  int index = a_data.boxLayout().index(dit());
          //int index = dit().intCode();
          buff.index.push_back(index);
          Box region = layout[dit()];
          region.grow(outputGhost);
          dataSize(a_data[dit()], thisSize, region, comps);
          buff.offsets.push_back(thisSize);
        }
      Vector<OffsetBuffer> gathering(numProc());
      gather(gathering, buff, uniqueProc(SerialTask::compute));
      broadcast(gathering,  uniqueProc(SerialTask::compute));
      //  pout() << gathering<<endl;
      for(int i=0; i<numProc(); ++i)
        {
          OffsetBuffer& offbuf = gathering[i];
          for(int num=0; num<offbuf.index.size(); num++)
            {
              int index = offbuf.index[num];
              for(unsigned int j=0; j<types; ++j) 
                {
                  //offsets[index+1][j] = offbuf.offsets[num][j];
                  offsets[j][index+1] = offbuf.offsets[num][j];
                }
            }
        }
      for(int i=0; i<layout.size(); i++)
        {
          for(unsigned int j=0; j<types; ++j) 
            {
              //offsets[i+1][j] += offsets[i][j];
              offsets[j][i+1] += offsets[j][i];
            }
        }
    }

  // pout() << offsets<<endl;
}


//==================================================================
//
// Now, linear IO routines for a BoxLayoutData of T
//

template <class T>
int write(HDF5Handle& a_handle, const BoxLayoutData<T>& a_data, 
          const std::string& a_name, const IntVect& outputGhost)
{
  int ret = 0;

  Interval comps = a_data.interval(); // later, we can make comps an argument.....
  T dummy; // used for preallocatable methods for dumb compilers.
  Vector<hid_t> types;
  dataTypes(types, dummy);

  Vector<Vector<long long> > offsets;
  Vector<long long> bufferCapacity(types.size(), 1);  // noel (was 0)
  Vector<void*> buffers(types.size(), NULL);

  getOffsets(offsets, a_data, types.size(), comps, outputGhost);

  // create datasets collectively.
  hsize_t flatdims[1];
  char dataname[100];
  Vector<hid_t> dataspace(types.size());
  Vector<hid_t> dataset(types.size());



  herr_t err;
  hsize_t count[1];
  hssize_t offset[1];

  for(unsigned int i=0; i<types.size(); ++i) 
    {
      flatdims[0] = offsets[i][offsets[i].size()-1];
      sprintf(dataname, "%s:datatype=%i",a_name.c_str(), i);
      dataspace[i]      = H5Screate_simple(1, flatdims, NULL);
      assert(dataspace[i] >=0);
      dataset[i]        = H5Dcreate(a_handle.groupID(), dataname,  
                                    types[i],
                                    dataspace[i], H5P_DEFAULT);
      assert(dataset[i] >= 0);
    }

  hid_t offsetspace, offsetData;
  for(unsigned int i=0; i<types.size(); ++i) 
    {
      flatdims[0] =  offsets[i].size();
      sprintf(dataname, "%s:offsets=%i",a_name.c_str(), i);
      offsetspace =  H5Screate_simple(1, flatdims, NULL);
      assert(offsetspace >= 0);
      offsetData  =   H5Dcreate(a_handle.groupID(), dataname,  
                                H5T_NATIVE_LLONG, offsetspace, H5P_DEFAULT);
      assert(offsetData >= 0);
      if(procID() == 0)
        {
          hid_t memdataspace = H5Screate_simple(1, flatdims, NULL);
          assert(memdataspace >= 0);
          err = H5Dwrite(offsetData, H5T_NATIVE_LLONG, memdataspace, offsetspace,
                         H5P_DEFAULT, &(offsets[i][0]));
          assert(err >= 0);
          H5Sclose(memdataspace);
        }
      H5Sclose(offsetspace);
      H5Dclose(offsetData);
    }

  // write BoxLayoutData attributes into Dataset[0]
  HDF5HeaderData info;
  info.m_int["comps"] = comps.size();
  info.m_string["objectType"] = name(dummy);
  std::string group = a_handle.getGroup();
  a_handle.setGroup(group+"/"+a_name+"_attributes");
  info.writeToFile(a_handle);
  a_handle.setGroup(group);

  // collective operations finished, now perform parallel writes
  // to specified hyperslabs.

  Vector<size_t> type_size(types.size());
  for(unsigned int i=0; i<types.size(); ++i) 
    {
      type_size[i] = H5Tget_size(types[i]);
    }

  Vector<int> thisSize(types.size());

  // step 1, create buffer big enough to hold the biggest linearized T
  // that I will have to output.

  //  pout()<<"offsets ";
   for(int i=0; i<offsets[0].size(); i++)
 	{
	  // 	  pout()<<" "<<offsets[0][i];
 	}
   pout()<<"\n";
  for(DataIterator it = a_data.dataIterator(); it.ok(); ++it)
    {
      unsigned int index = a_data.boxLayout().index(it());
      for(unsigned int i=0; i<types.size(); ++i) 
        {
          long long size = (offsets[i][index+1] - offsets[i][index]) 
            * type_size[i];
		  // pout()<<"index:size "<<index<<" "<<size<<"\n";
          assert(size >= 0);
          if(size > bufferCapacity[i]) // grow buffer if necessary.....
            {
              bufferCapacity[i] = size;
            } 
        }
    }
  //assert(bufferCapacity[0] > 1);
  for(unsigned int i=0; i<types.size(); ++i) 
    {
      buffers[i] = malloc(bufferCapacity[i]);
      if(buffers[i] == NULL) {
        pout() << " i=" << i
               << " types.size() = " << types.size()
               << " bufferCapacity[i] = " << (int)bufferCapacity[i]
               << endl;
        MayDay::Error("memory error in buffer allocation write");
      }
    }

  // Step 2.  actually a) write each of my T objects into the
  // buffer, then b) write that buffered data out to the
  // write position in the data file using hdf5 hyperslab functions.
  for(DataIterator it = a_data.dataIterator(); it.ok(); ++it)
    {
      const T& data = a_data[it()];
      unsigned int index = a_data.boxLayout().index(it());
      Box box = a_data.box(it());
      box.grow(outputGhost);
      write(data, buffers, box, comps); //write T to buffer
      for(unsigned int i=0; i<types.size(); ++i) 
        {
          offset[0] = offsets[i][index];
          count[0] = offsets[i][index+1] - offset[0];
          if(count[0] > 0){
			err =  H5Sselect_hyperslab(dataspace[i], H5S_SELECT_SET, 
									   offset, NULL, 
									   count, NULL);
			assert(err >= 0);
			hid_t memdataspace = H5Screate_simple(1, count, NULL);
			assert(memdataspace >= 0);
			err = H5Dwrite(dataset[i], types[i], memdataspace, dataspace[i],
						   H5P_DEFAULT, buffers[i]);
			assert(err >= 0);
			H5Sclose(memdataspace);
			if(err < 0) { ret = err; goto cleanup;}
		  }
        }
    }

  // OK, clean up data structures

 cleanup:
  for(unsigned int i=0; i<types.size(); ++i) 
    {
      free(buffers[i]);
      H5Sclose(dataspace[i]);
      H5Dclose(dataset[i]);
    }
  return ret;

}

template <class T>
int write(HDF5Handle& a_handle, const LevelData<T>& a_data, 
          const std::string& a_name, const IntVect& outputGhost)
{
  HDF5HeaderData info;
  info.m_intvect["ghost"] = a_data.ghostVect();
  IntVect og(outputGhost);
  og.min(a_data.ghostVect());
  info.m_intvect["outputGhost"] = og;
  std::string group = a_handle.getGroup();
  a_handle.setGroup(group+"/"+a_name+"_attributes");
  info.writeToFile(a_handle);
  a_handle.setGroup(group);
  return write(a_handle, (const BoxLayoutData<T>&)a_data, a_name, og);
}

template <class T>
int read(HDF5Handle& a_handle, LevelData<T>& a_data, const std::string& a_name, 
         const DisjointBoxLayout& a_layout, const Interval& a_comps, bool a_redefineData)
{
  if(a_redefineData)
    {
      HDF5HeaderData info;
      std::string group = a_handle.getGroup();
      if(a_handle.setGroup(group+"/"+a_name+"_attributes"))
        {
          std::string message = "error opening "+a_handle.getGroup()+"/"+a_name ;
          MayDay::Warning(message.c_str());
          return 1;
        }
      info.readFromFile(a_handle);
      a_handle.setGroup(group);
      int ncomp =  info.m_int["comps"];
      IntVect ghost = info.m_intvect["ghost"];
      if(a_comps.end() > 0 && ncomp < a_comps.end())
        {
          MayDay::Error("attempt to read component interval that is not available");
        }
      if(a_comps.size() == 0)
        a_data.define(a_layout, ncomp, ghost);
      else
        a_data.define(a_layout, a_comps.size(), ghost);
    }
  return read(a_handle, (BoxLayoutData<T>&)a_data, a_name, a_layout, a_comps, false);

}
template <class T>
int read(HDF5Handle& a_handle, BoxLayoutData<T>& a_data, const std::string& a_name, 
         const BoxLayout& a_layout, const Interval& a_comps, bool a_redefineData)
{
  int ret = 0; // return value;

  herr_t err;

  char dataname[100];
  hsize_t count[1];
  hssize_t offset[1];
  Vector<Vector<long long> > offsets;

  T dummy;
  Vector<hid_t> types;
  dataTypes(types, dummy);
  Vector<hid_t> dataspace(types.size());
  Vector<hid_t> dataset(types.size());
  offsets.resize(types.size(), Vector<long long>(a_layout.size() +1));

  Vector<int> bufferCapacity(types.size(), 500);
  Vector<void*> buffers(types.size(), NULL);

  for(unsigned int i=0; i<types.size(); ++i) 
    {
      sprintf(dataname, "%s:datatype=%i",a_name.c_str(), i);
      dataset[i]        = H5Dopen(a_handle.groupID(), dataname);
      if(dataset[i] < 0) {MayDay::Warning("dataset open failure"); return dataset[i];}
      dataspace[i]      = H5Dget_space(dataset[i]);
      if(dataspace[i] < 0) {MayDay::Warning("dataspace open failure"); return dataspace[i];}
    }

  hid_t offsetspace, offsetData;
  hsize_t flatdims[1];
  for(unsigned int i=0; i<types.size(); ++i) 
    {
      flatdims[0] =  offsets[i].size();
      sprintf(dataname, "%s:offsets=%i",a_name.c_str(), i);
      offsetspace =  H5Screate_simple(1, flatdims, NULL);
      assert(offsetspace >= 0);
      offsetData  =   H5Dopen(a_handle.groupID(), dataname);  
      assert(offsetData >= 0);
      hid_t memdataspace = H5Screate_simple(1, flatdims, NULL);
      assert(memdataspace >= 0);
      err = H5Dread(offsetData, H5T_NATIVE_LLONG, memdataspace, offsetspace,
                    H5P_DEFAULT, &(offsets[i][0]));
      assert(err >=0);
      H5Sclose(memdataspace);
      H5Sclose(offsetspace);
      H5Dclose(offsetData);
    }

  HDF5HeaderData info;
  std::string group = a_handle.getGroup();
  if(a_handle.setGroup(a_handle.getGroup()+"/"+a_name+"_attributes"))
    {
      std::string message = "error opening "+a_handle.getGroup()+"/"+a_name ;
      MayDay::Warning(message.c_str());
      return 1;
    }

  info.readFromFile(a_handle);
  a_handle.setGroup(group);
  int ncomps = info.m_int["comps"];
  IntVect outputGhost(IntVect::Zero); // backwards file compatible mode.
  if(info.m_intvect.find("outputGhost") != info.m_intvect.end())
    {
      outputGhost = info.m_intvect["outputGhost"];
    }
  if(ncomps <= 0){
    MayDay::Warning("ncomps <= 0 in read");
    return ncomps;
  }

  if(a_redefineData){
    if(a_comps.size() != 0){
      a_data.define(a_layout, a_comps.size());
    } else {
      a_data.define(a_layout, ncomps);
    }
  }

  Interval comps(0, ncomps-1);
  //huh?
  //  if(a_comps.size() != 0)  comps = Interval(0, a_comps.size());

  //  getOffsets(offsets, a_data, types.size(), comps);

  Vector<size_t> type_size(types.size());
  for(unsigned int i=0; i<types.size(); ++i) 
    {
      type_size[i] = H5Tget_size(types[i]);
      buffers[i] = malloc(bufferCapacity[i]);
      if(buffers[i] == NULL) {
        pout() << " i= " << i 
               << " types.size()=" << types.size() 
               << " bufferCapacity[i] = " << (int)bufferCapacity[i]
               << endl; 
        MayDay::Error("memory error in buffer allocation read");
      }
    }

  Vector<int> thisSize(types.size());
  for(DataIterator it = a_data.dataIterator(); it.ok(); ++it)
    {
      T& data = a_data[it()];
      unsigned int index = a_data.boxLayout().index(it());
      Box box = a_data.box(it());

      for(unsigned int i=0; i<types.size(); ++i) 
        {
          if(a_comps.size() == 0){
            offset[0] = offsets[i][index];
            count[0] = offsets[i][index+1] - offset[0];
          } else {
            offset[0] = offsets[i][index] + box.numPts()*a_comps.begin();
            count[0]  = a_comps.size() * box.numPts();
          }
          if(count[0] > 0)
			{
			  int size = count[0] * type_size[i];
			  if(size > bufferCapacity[i])
				{
				  free(buffers[i]);
				  bufferCapacity[i] = Max(2*bufferCapacity[i], size);
				  buffers[i] = malloc(bufferCapacity[i]);
				  if(buffers[i] == NULL) {
					MayDay::Error("memory error in buffer allocation read");
				  }
				} 
			  
			  err =  H5Sselect_hyperslab(dataspace[i], H5S_SELECT_SET, 
                                     offset, NULL, 
                                     count, NULL);
			  assert(err >= 0);
			  hid_t memdataspace = H5Screate_simple(1, count, NULL);
			  assert(memdataspace >= 0);
			  err = H5Dread(dataset[i], types[i], memdataspace, dataspace[i],
							H5P_DEFAULT, buffers[i]);
			  assert(err >= 0);
			  H5Sclose(memdataspace);
			  if(err < 0) { ret = err; goto cleanup;}
			}
        }
      box.grow(outputGhost);
      read(data, buffers,  box, comps);
    }

 cleanup:
  for(unsigned int i=0; i<types.size(); ++i) 
    {
      free(buffers[i]);
      H5Sclose(dataspace[i]);
      H5Dclose(dataset[i]);
    }
  return ret;
}


template <class T>
int writeLevel(HDF5Handle& a_handle, 
               const int&  a_level, 
               const LevelData<T>& a_data,
               const Real& a_dx, 
               const Real& a_dt, 
               const Real& a_time, 
               const Box&  a_domain,
               const int&  a_refRatio,
               const IntVect& outputGhost)
{
  int error;
  char levelName[10];
  std::string currentGroup = a_handle.getGroup();
  sprintf(levelName, "/level_%i",a_level);
  error = a_handle.setGroup(currentGroup + levelName);
  if(error != 0) return 1;

  HDF5HeaderData meta;
  meta.m_real["dx"] = a_dx;
  meta.m_real["dt"] = a_dt;
  meta.m_real["time"] = a_time;
  meta.m_box["prob_domain"] = a_domain;
  meta.m_int["ref_ratio"] = a_refRatio;

  error = meta.writeToFile(a_handle);
  if(error != 0) return 2;

  error = write(a_handle, a_data.boxLayout());
  if(error != 0) return 3;

  error = write(a_handle, a_data, "data", outputGhost);
  if(error != 0) return 4;

  a_handle.setGroup(currentGroup);

  return 0;
}


template <class T>
int readLevel(HDF5Handle&   a_handle,  
              const int&    a_level, 
              LevelData<T>& a_data, 
              Real& a_dx, 
              Real& a_dt, 
              Real& a_time, 
              Box&  a_domain,
              int&  a_refRatio,
              const Interval&   a_comps,
              const IntVect& ghost,
              bool  setGhost)
{
  HDF5HeaderData header;
  header.readFromFile(a_handle);
  int nComp = header.m_int["num_components"];

  int error;
  char levelName[10];
  std::string currentGroup = a_handle.getGroup();
  sprintf(levelName, "/level_%i",a_level);
  error = a_handle.setGroup(currentGroup + levelName);
  if(error != 0) return 1;

  HDF5HeaderData meta;
  error = meta.readFromFile(a_handle);
  if(error != 0) return 2;
  a_dx       = meta.m_real["dx"];
  a_dt       = meta.m_real["dt"];
  a_time     = meta.m_real["time"];
  a_domain   = meta.m_box["prob_domain"];
  a_refRatio = meta.m_int["ref_ratio"];

  Vector<Box> boxes;
  error = read(a_handle, boxes);
  Vector<int> procIDs;
  LoadBalance(procIDs, boxes);

  DisjointBoxLayout layout(boxes, procIDs);

  layout.close();
  if(error != 0) return 3;

  if(!setGhost)
    error = read(a_handle, a_data, "data", layout, a_comps);
  else
    {
      if(a_comps.size() !=  0)
        a_data.define(layout, a_comps.size(), ghost);
      else
        a_data.define(layout, nComp, ghost);

      error = read(a_handle, a_data, "data", layout, a_comps, false);

    }
  if(error != 0) return 4;

  a_handle.setGroup(currentGroup);

  return 0;
}



#endif  // HDF5_H

#endif // HDF5
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL

#ifdef HDF5

using std::ostream;
using std::cout;
using std::endl;

// template <class T>
// ostream& operator<<(ostream& os, const Vector<T>& vec)
// {
//   for(int i=0; i<vec.size(); i++) os<<vec[i]<<" ";
//   return os;
// }

void OffsetBuffer::operator=(const OffsetBuffer& rhs)
{
  if(rhs.index.size() == 0)
    {
      index.resize(0);
      offsets.resize(0);
      return;
    }
  index.resize(rhs.index.size());
  int types = rhs.offsets[0].size();
  offsets.resize(rhs.offsets.size(), Vector<int>(types, 0));
  for(int i=0; i<rhs.index.size(); i++)
    {
      index[i] = rhs.index[i];
      for(int j=0; j<types; j++)
        {
          offsets[i][j] = rhs.offsets[i][j];
        }
    }
}

ostream&  operator<<(ostream& os, const OffsetBuffer& ob)
{
  os << ob.index <<"|"<<ob.offsets<<"\n";
  return os;
}

//OffsetBuffer specialization of linearSize
template < >
int linearSize(const OffsetBuffer& a_input)
{
  int r = 0;
  r += 2;   // number of entries from this processor, number of types.
  r += a_input.index.size();// indices for these offset entries
  r += a_input.offsets[0].size()*a_input.offsets.size();  // offsets
  return r * sizeof(int);
}

//OffsetBuffer specialization of linearIn
template < >
void linearIn(OffsetBuffer& a_outputT, const void* const a_inBuf)
{
  // pout() << "in:  ";
  const int* data = (const int*) a_inBuf;
  int num = data[0],  numTypes = data[1];
  a_outputT.index.resize(data[0]);
  a_outputT.offsets.resize(data[0], Vector<int>(data[1]));
  const int* off = data+2+num;
  for(int i=0; i<num; i++)
    {
      a_outputT.index[i] = data[i+2];
      Vector<int>& offsets = a_outputT.offsets[i];
      for(int j=0; j<numTypes; j++)
        {
          offsets[j] = off[j];
		  //  pout() << offsets[j]<<" ";
        }
      off+= numTypes;
    }
  //  pout() << endl;
}

//OffsetBuffer specialization of linearOut
template < >
void linearOut(void* const a_outBuf, const OffsetBuffer& a_inputT)
{
  //  pout() << "out: ";
  int* data = (int*) a_outBuf;
  data[0] =  a_inputT.index.size();
  data[1] =  a_inputT.offsets[0].size();
  data+=2;
  for(int i=0; i< a_inputT.index.size(); ++i, ++data)
    *data =  a_inputT.index[i];

  for(int i=0; i<a_inputT.offsets.size(); ++i)
    {
      const Vector<int>& offset = a_inputT.offsets[i];
      for(int t=0; t < offset.size(); ++t, ++data)
        {
          *data = offset[t];
		  //  pout() << *data <<" ";
        }
    }

  // pout() << endl;
}
//Vector<OffsetBuffer>  specialization
template < > int linearSize(const Vector<OffsetBuffer>& a_input)
{
  return linearListSize(a_input);
}
template < > void linearIn(Vector<OffsetBuffer>& a_outputT, const void* const inBuf)
{
  linearListIn(a_outputT, inBuf);
}
template < > void linearOut(void* const a_outBuf, const Vector<OffsetBuffer>& a_inputT)
{
  linearListOut(a_outBuf, a_inputT);
}

int  write(HDF5Handle& a_handle, const BoxLayout& a_layout)
{
  assert(a_layout.isClosed());
  herr_t ret;
  hssize_t offset[1];
  hsize_t  flatdims[1], count[1];
  count[0] = 1;
  flatdims[0] = a_layout.size();

  hid_t boxdataspace = H5Screate_simple(1, flatdims, NULL);

  hid_t boxdataset   = H5Dcreate(a_handle.groupID(), "boxes",  a_handle.box_id, 
                                 boxdataspace, H5P_DEFAULT);
  if(boxdataset < 0) return boxdataset;
  hid_t memdataspace = H5Screate_simple(1, count, NULL);

  // write boxes in parallel
  for(DataIterator it = a_layout.dataIterator(); it.ok(); ++it)
    {
      offset[0] = a_layout.index(it());
      ret = H5Sselect_hyperslab (boxdataspace, H5S_SELECT_SET, offset, NULL, 
                                 count, NULL);
      if(ret < 0) return ret;
      ret = H5Dwrite(boxdataset, a_handle.box_id, memdataspace, boxdataspace,
                     H5P_DEFAULT, &(a_layout[it()]));
      if(ret < 0) return ret;
    }

  H5Sclose(boxdataspace);
  H5Sclose(memdataspace);
  H5Dclose(boxdataset);

  return 0;
}

int read(HDF5Handle& a_handle, Vector<Box>& boxes)
{
  hid_t  boxdataset = H5Dopen(a_handle.groupID(), "boxes");
  if(boxdataset < 0) return boxdataset;
  hid_t boxdataspace =  H5Dget_space(boxdataset);
  if(boxdataspace < 0) return boxdataspace;
  hsize_t dims[1], maxdims[1];
  H5Sget_simple_extent_dims(boxdataspace, dims, maxdims);

  hid_t memdataspace = H5Screate_simple(1, dims, NULL);

  Box* rawboxes = new Box[dims[0]];
  if(rawboxes == NULL)
    MayDay::Error("out of memory in read(HDF5Handle& a_handle, BoxLayout& a_layout)");

  herr_t error = H5Dread(boxdataset, a_handle.box_id, memdataspace, boxdataspace,
                         H5P_DEFAULT, rawboxes);
  if(error < 0) return error;

  boxes.resize(dims[0]);
  for(int index = 0; index < dims[0]; ++index)
    {
      rawboxes[index].computeBoxLen();
      boxes[index] = rawboxes[index];
    }

  delete[] rawboxes;
  H5Dclose(boxdataset);
  H5Sclose(boxdataspace);
  H5Sclose(memdataspace);
  return 0;
}

int readBoxes(HDF5Handle& a_handle, Vector<Vector<Box> >& boxes)
{
  int error;
  char levelName[10];
  std::string currentGroup = a_handle.getGroup();

  // we want to start in root group.  currentGroup reset on normal exit
  error = a_handle.setGroup("/");
  if(error != 0) return 1;
  std::string workingGroup = a_handle.getGroup();

  HDF5HeaderData header;
  header.readFromFile(a_handle);
  int numLevels = header.m_int["num_levels"];
  boxes.resize(numLevels);

  for(int i = 0; i<numLevels; i++){
        sprintf(levelName, "/level_%i",i);
        error = a_handle.setGroup(workingGroup + levelName);
        if(error != 0) return 1;
        read(a_handle, boxes[i]);
  }
  
  a_handle.setGroup(currentGroup); // put handle back to where it stated.

  return 0;
}

int readFArrayBox(HDF5Handle& a_handle,
                                  FArrayBox&  a_fab,
                                  int a_level,
                                  int a_boxNumber,
                                  const Interval& a_components,
                                  const std::string& a_dataName)
{
  char levelName[10];
  herr_t err;
  int error=0;
  hsize_t count[1];
  hssize_t offset[1];
  hsize_t flatdims[1];
  hid_t dataspace, dataset, memdataspace;

  std::string currentGroup = a_handle.getGroup();

  // we want to start in root group.  currentGroup reset on normal exit
  error = a_handle.setGroup("/");
  if(error != 0) return 1;
  std::string workingGroup = a_handle.getGroup();

  sprintf(levelName, "/level_%i",a_level);
  error = a_handle.setGroup(workingGroup + levelName);
  if(error != 0) return 1;

  // first, fetch the correct box from file.

  Box box;
  flatdims[0]  = 1;
  memdataspace = H5Screate_simple(1, flatdims, NULL);
  count[0]  = 1;
  offset[0] = a_boxNumber;
  dataset   = H5Dopen(a_handle.groupID(),  "boxes");      
  dataspace = H5Dget_space(dataset);
  err =  H5Sselect_hyperslab(dataspace, H5S_SELECT_SET, 
							 offset, NULL, 
							 count, NULL);
  if(err != 0 ) return 2;
  err = H5Dread(dataset, HDF5Handle::box_id, memdataspace, dataspace,
				H5P_DEFAULT, &box);
  if(err != 0 ) return 3;

  H5Dclose(dataset);
  H5Sclose(dataspace);
  
  box.computeBoxLen();
  
  //next, fetch the data offset that we need from the file.
  std::string namebuff1 = a_dataName + ":offsets=0";
  dataset = H5Dopen(a_handle.groupID(), namebuff1.c_str());
  dataspace = H5Dget_space(dataset);
  err =  H5Sselect_hyperslab(dataspace, H5S_SELECT_SET, 
							 offset, NULL, 
							 count, NULL);
  if(err != 0 ) return 2;

  long long offsetLong;
  err = H5Dread(dataset, H5T_NATIVE_LLONG, memdataspace, dataspace,
				H5P_DEFAULT, &offsetLong);
  if(err != 0 ) return 3;
  
  H5Dclose(dataset);
  H5Sclose(dataspace);
  H5Sclose(memdataspace);
  
  //  OK, now lets go get our FArrayBox data:
  int ncomp = a_components.size();
  if(a_fab.box() == box && ncomp == a_fab.nComp()){
	//FArrayBox already defined appropriately for this operation
  } else {
	a_fab.define(box, ncomp);
  }

  offsetLong += box.numPts() * a_components.begin();
  count[0]    = box.numPts() * ncomp;
  flatdims[0] = count[0];
  offset[0]   = offsetLong;
  memdataspace = H5Screate_simple(1, flatdims, NULL);
  std::string namebuff2 = a_dataName + ":datatype=0";
  dataset  = H5Dopen(a_handle.groupID(),  namebuff2.c_str() );
  dataspace = H5Dget_space(dataset);
  err =  H5Sselect_hyperslab(dataspace, H5S_SELECT_SET, 
                             offset, NULL, count, NULL);
  if(err != 0 ) return 2;
  err = H5Dread(dataset, H5T_NATIVE_REAL, memdataspace, dataspace,
				H5P_DEFAULT, a_fab.dataPtr());
  if(err != 0 ) return 3;
  
  H5Dclose(dataset);
  H5Sclose(dataspace);
  H5Sclose(memdataspace);
  
  a_handle.setGroup(currentGroup);
  return 0;

}


bool HDF5Handle::initialized = false;
hid_t HDF5Handle::box_id = 0;
hid_t HDF5Handle::intvect_id = 0;
hid_t HDF5Handle::realvect_id = 0;


void HDF5Handle::initialize()
{


  /*  old composite-of-composite version of box 
  intvect_id = H5Tcreate (H5T_COMPOUND, sizeof(IntVect));
#if CH_SPACEDIM == 2
  H5Tinsert (intvect_id, "intvecti", HOFFSET(IntVect, vect[0]), H5T_NATIVE_INT);
  H5Tinsert (intvect_id, "intvectj", HOFFSET(IntVect, vect[1]), H5T_NATIVE_INT);
#endif
#if CH_SPACEDIM == 3
  H5Tinsert (intvect_id, "intvecti", HOFFSET(IntVect, vect[0]), H5T_NATIVE_INT);
  H5Tinsert (intvect_id, "intvectj", HOFFSET(IntVect, vect[1]), H5T_NATIVE_INT);
  H5Tinsert (intvect_id, "intvectk", HOFFSET(IntVect, vect[2]), H5T_NATIVE_INT);
#endif

  box_id = H5Tcreate (H5T_COMPOUND, sizeof(Box));
  H5Tinsert (box_id, "smallend", HOFFSET(Box, smallend), intvect_id);
  H5Tinsert (box_id, "bigend",   HOFFSET(Box, bigend), intvect_id);
  */

  intvect_id = H5Tcreate (H5T_COMPOUND, sizeof(IntVect));
  box_id = H5Tcreate (H5T_COMPOUND, sizeof(Box));
  realvect_id = H5Tcreate(H5T_COMPOUND, sizeof(RealVect));
#if CH_SPACEDIM == 2
  H5Tinsert (intvect_id, "intvecti", HOFFSET(IntVect, vect[0]), H5T_NATIVE_INT);
  H5Tinsert (intvect_id, "intvectj", HOFFSET(IntVect, vect[1]), H5T_NATIVE_INT);
  H5Tinsert (box_id, "lo_i", HOFFSET(Box, smallend.vect[0]), H5T_NATIVE_INT);
  H5Tinsert (box_id, "lo_j", HOFFSET(Box, smallend.vect[1]), H5T_NATIVE_INT);
  H5Tinsert (box_id, "hi_i", HOFFSET(Box, bigend.vect[0]),   H5T_NATIVE_INT);
  H5Tinsert (box_id, "hi_j", HOFFSET(Box, bigend.vect[1]),   H5T_NATIVE_INT);
  H5Tinsert (realvect_id, "x", HOFFSET(RealVect, vect[0]),   H5T_NATIVE_REAL);
  H5Tinsert (realvect_id, "y", HOFFSET(RealVect, vect[1]),   H5T_NATIVE_REAL);

#endif
#if CH_SPACEDIM == 3
  H5Tinsert (intvect_id, "intvecti", HOFFSET(IntVect, vect[0]), H5T_NATIVE_INT);
  H5Tinsert (intvect_id, "intvectj", HOFFSET(IntVect, vect[1]), H5T_NATIVE_INT);
  H5Tinsert (intvect_id, "intvectk", HOFFSET(IntVect, vect[2]), H5T_NATIVE_INT);
  H5Tinsert (box_id, "lo_i", HOFFSET(Box, smallend.vect[0]), H5T_NATIVE_INT);
  H5Tinsert (box_id, "lo_j", HOFFSET(Box, smallend.vect[1]), H5T_NATIVE_INT);
  H5Tinsert (box_id, "lo_k", HOFFSET(Box, smallend.vect[2]), H5T_NATIVE_INT);
  H5Tinsert (box_id, "hi_i", HOFFSET(Box, bigend.vect[0]),   H5T_NATIVE_INT);
  H5Tinsert (box_id, "hi_j", HOFFSET(Box, bigend.vect[1]),   H5T_NATIVE_INT);
  H5Tinsert (box_id, "hi_k", HOFFSET(Box, bigend.vect[2]),   H5T_NATIVE_INT);
  H5Tinsert (realvect_id, "x", HOFFSET(RealVect, vect[0]),   H5T_NATIVE_REAL);
  H5Tinsert (realvect_id, "y", HOFFSET(RealVect, vect[1]),   H5T_NATIVE_REAL);
  H5Tinsert (realvect_id, "z", HOFFSET(RealVect, vect[2]),   H5T_NATIVE_REAL);
#endif


  initialized = true;
}

HDF5Handle::HDF5Handle(): m_isOpen(false)
{
  if(!initialized) initialize();
}

HDF5Handle::HDF5Handle(const std::string& a_filename, mode a_mode)
  : m_isOpen(false), m_level(-1)
{
  int err = open(a_filename, a_mode);
  if (err < 0 ){
	char buf[70];
	sprintf(buf, "Problem opening file %s ",a_filename.c_str());
	MayDay::Error(buf);
  }
}


int HDF5Handle::open(const std::string& a_filename, mode a_mode)
{
  int ret = 0;
  if(m_isOpen)
    {
      MayDay::Error("Calling 'open' on already open file.  use 'close' on finished files");
    }

  m_filename = a_filename;
  if(!initialized) initialize();
  m_group    = "/";

  hid_t file_access = 0;
#ifdef MPI
  file_access = H5Pcreate (H5P_FILE_ACCESS);

#if H5_VERS_MINOR==2
  H5Pset_mpi(file_access,  Chombo_MPI::comm, MPI_INFO_NULL);
#else
  H5Pset_fapl_mpio(file_access,  Chombo_MPI::comm, MPI_INFO_NULL);
#endif
#else
  file_access = H5P_DEFAULT;
#endif

  switch(a_mode){
  case CREATE:
    m_fileID = H5Fcreate(a_filename.c_str(), H5F_ACC_TRUNC, H5P_DEFAULT, file_access);
    if(m_fileID < 0) return m_fileID;
    break;
  case OPEN_RDONLY:
    m_fileID = H5Fopen(a_filename.c_str(), H5F_ACC_RDONLY, file_access);
    if(m_fileID < 0) return m_fileID;
    break;
  case OPEN_RDWR:
    m_fileID = H5Fopen(a_filename.c_str(), H5F_ACC_RDWR, file_access);
    if(m_fileID < 0) return m_fileID;
    break;
  default:
    MayDay::Error("unrecognized file access mode:  HDF5Handle::open");
  }
  m_currentGroupID = H5Gopen(m_fileID, m_group.c_str());
  if(m_fileID >= 0 && m_currentGroupID >= 0) m_isOpen = true;

  // Write or read dimension checks and accuracy data

  HDF5HeaderData info;
  char buf[100];
  hid_t attr, datatype, group;
  size_t codeprecision, fileprecision;
  switch(a_mode){
  case CREATE:
    group = H5Gcreate(m_fileID, "Chombo_global", 0); 
    info.m_int["SpaceDim"] = SpaceDim;
    info.m_real["testReal"] = 0.0;
    info.writeToLocation(group);
    break;
  default:
    group =  H5Gopen(m_fileID, "Chombo_global");
    if(group < 0)
      {
        MayDay::Warning("This files appears to be missing a 'Chombo_global' section");
      }
    info.readFromLocation(group);
    if(info.m_int["SpaceDim"] == 0)
      {
        sprintf(buf, "file %s appears to lack a SpaceDim definition",a_filename.c_str());
        MayDay::Error(buf);
      }
    if(info.m_int["SpaceDim"] != SpaceDim)
      {
        MayDay::Error("SpaceDim of %s does not match code");
      }
    attr = H5Aopen_name(group, "testReal");
    if(attr < 0) return 1;
    datatype = H5Aget_type(attr);
    fileprecision = H5Tget_precision(datatype);
    codeprecision = H5Tget_precision(H5T_NATIVE_REAL);
    if(fileprecision > codeprecision) ret = 2;
    if(codeprecision > fileprecision)
      {
        sprintf(buf, "code is compiled with Real=%i bits, file %s has Real=%i bits",
                codeprecision, a_filename.c_str(), fileprecision);
        MayDay::Warning(buf);
        ret = 2;
      }
    H5Aclose(attr);
    H5Tclose(datatype);
  }
  H5Gclose(group);


  return ret;
}


bool HDF5Handle::isOpen() const{ return m_isOpen;}

void HDF5Handle::setGroupToLevel(int a_level)
{
  char ch[15];
  sprintf(ch, "level_%i", a_level);
  setGroup(ch);
} 

int HDF5Handle::setGroup(const std::string& group)
{
  if(!m_isOpen)
    {
      MayDay::Error("cannot access group until file is successfully opened");
    }
  if(m_group == group) return 0;

  int ret = 0;

  ret = H5Gclose(m_currentGroupID);
  if(ret < 0)
    { 
      MayDay::Warning(" Error closing old group");
      return ret;
    }

  H5E_auto_t efunc; void* edata; // turn auto error messaging off 
  H5Eget_auto(&efunc, &edata);
  H5Eset_auto(NULL, NULL);
  m_currentGroupID = H5Gopen(m_fileID, group.c_str());
  if(m_currentGroupID < 0)
    {
      H5Eset_auto(efunc, edata); //turn error messaging back on.
      //open failed, go to group creation
      m_currentGroupID = H5Gcreate(m_fileID, group.c_str(), 0);
    }
  if(m_currentGroupID < 0)
    ret = -1;

  H5Eset_auto(efunc, edata); //turn error messaging back on.
  m_group = group;
  return ret;
}
void HDF5Handle::close()
{
  if(m_isOpen){
    if(!(m_currentGroupID < 0)) H5Gclose(m_currentGroupID);
    if(!(m_fileID < 0)) H5Fclose(m_fileID);
    m_isOpen = false;
  }
}



const std::string&  HDF5Handle::getGroup() const
{
  return m_group;
}
const hid_t& HDF5Handle::fileID() const {return m_fileID;}
const hid_t& HDF5Handle::groupID() const {return m_currentGroupID;}



//=====================================================================================
  /// writes the current attribute list to the current group in 'file'
int HDF5HeaderData::writeToFile(HDF5Handle& file) const
{
  return writeToLocation(file.groupID());
}

int HDF5HeaderData::writeToLocation(hid_t loc_id) const
{
  H5E_auto_t efunc; void* edata;  
  H5Eget_auto(&efunc, &edata);

  herr_t  ret;
  char messg[80];
#define INSERT(Ttype, mapName, H5Ttype)                                   \
  for(map<std::string, Ttype>::const_iterator p = mapName.begin();        \
      p!= mapName.end(); ++p)                                             \
    {                                                                     \
      hid_t aid  = H5Screate(H5S_SCALAR);                                 \
      H5Eset_auto(NULL, NULL);                                            \
      hid_t attr = H5Acreate(loc_id, p->first.c_str(), H5Ttype,           \
                             aid, H5P_DEFAULT);                           \
      if(attr < 0)                                                        \
        {                                                                 \
          H5Adelete(loc_id, p->first.c_str());                            \
          attr = H5Acreate(loc_id, p->first.c_str(), H5Ttype,             \
                           aid, H5P_DEFAULT);                             \
          if(attr < 0)                                                    \
            {                                                             \
              sprintf(messg," Problem writing attribute %s",p->first.c_str());  \
              MayDay::Warning(messg);                                     \
            }                                                             \
        }                                                                 \
      H5Eset_auto(efunc, edata);                                          \
      Ttype tmp = p->second;                                              \
      ret = H5Awrite(attr, H5Ttype, &tmp);                                \
      if(ret < 0) return ret;                                             \
      H5Sclose(aid);                                                      \
      H5Aclose(attr);                                                     \
    }                                                                     \

    INSERT(Real, m_real, H5T_NATIVE_REAL);
    INSERT(int, m_int, H5T_NATIVE_INT);
    INSERT(IntVect, m_intvect, HDF5Handle::intvect_id);
    INSERT(Box, m_box, HDF5Handle::box_id);
    INSERT(RealVect, m_realvect, HDF5Handle::realvect_id);

    // string is different, of course

    for(map<std::string, std::string>::const_iterator p = m_string.begin(); 
        p!= m_string.end(); ++p)                                        
    {
      hid_t s_type = H5Tcopy(H5T_C_S1);
      H5Tset_size(s_type, p->second.length()); //extra requirement for strings
      hid_t aid  = H5Screate(H5S_SCALAR);
      H5Eset_auto(NULL, NULL);             
      hid_t attr = H5Acreate(loc_id, p->first.c_str(), s_type,
                             aid, H5P_DEFAULT); 
      if(attr < 0)          
        {                   
          H5Adelete(loc_id, p->first.c_str());     
          attr = H5Acreate(loc_id, p->first.c_str(), s_type, 
                           aid, H5P_DEFAULT);       
          if(attr < 0)                                     
            {                                              
              sprintf(messg," Problem writing attribute %s",p->first.c_str()); 
              MayDay::Warning(messg);
            }                   
        }
      H5Eset_auto(efunc, edata);
      char* tmp = (char*)p->second.c_str();
      ret = H5Awrite(attr, s_type, tmp);           
      if(ret < 0) return ret;                                 
      H5Sclose(aid);                                           
      H5Aclose(attr);
      H5Tclose(s_type);
    }


    return 0;
}


  /// read process is add/change, does not remove key-value pairs. reads from current group.
int HDF5HeaderData::readFromFile(HDF5Handle& file)
{
  return H5Aiterate(file.groupID(), NULL, HDF5HeaderDataattributeScan , this);
}

int HDF5HeaderData::readFromLocation(hid_t loc_id)
{
   return H5Aiterate(loc_id, NULL, HDF5HeaderDataattributeScan , this);
}

extern "C"{
herr_t HDF5HeaderDataattributeScan(hid_t loc_id, const char *name, void *opdata)
{
  herr_t ret = 0;
  HDF5HeaderData& data = *(static_cast<HDF5HeaderData*>(opdata));

  hid_t attr   = H5Aopen_name(loc_id, name);
  hid_t atype  = H5Aget_type(attr);
  hid_t aclass = H5Tget_class(atype);
  char* buf = NULL;  size_t size = 0;

  switch(aclass){
  case H5T_INTEGER :
    int Ivalue;
    ret  = H5Aread(attr, H5T_NATIVE_INT, &Ivalue);
    if(ret < 0) break;
    data.m_int[name] = Ivalue;
    break;
  case H5T_FLOAT:
    Real Rvalue;
    ret = H5Aread(attr, H5T_NATIVE_REAL, &Rvalue);
    if(ret < 0) break;
    data.m_real[name] = Rvalue;
    break;
  case H5T_STRING:
    size = H5Tget_size(atype);
    buf = new char[size+1];
    ret = H5Aread(attr, atype, buf);
    if(ret < 0) break;
    buf[size] = 0; // for some reason HDF5 is not null terminating strings correctly
    data.m_string[name] = std::string(buf);
    break;
  case H5T_COMPOUND:
    if(strcmp(H5Tget_member_name(atype, 0), "lo_i") == 0)
      {
        Box value;
        ret = H5Aread(attr, HDF5Handle::box_id, &value);
        if(ret < 0) break;
        value.computeBoxLen();
        data.m_box[name] = value;
        break;
      }
    else if(strcmp(H5Tget_member_name(atype, 0), "intvecti") == 0)
      {
        IntVect value;
        ret = H5Aread(attr, HDF5Handle::intvect_id, &value);
        if(ret < 0) break;
        data.m_intvect[name] = value;
        break;
      }
    else if(strcmp(H5Tget_member_name(atype, 0), "x") == 0)
      {
        RealVect value;
        ret = H5Aread(attr, HDF5Handle::realvect_id, &value);
        if(ret < 0) break;
        data.m_realvect[name] = value;
        break;
      }
  default:
    // heel if I know what to do about attributes I don't recognize
    MayDay::Warning("HDF5HeaderData::readFromFile encountered unrecognized attribute");
  }
  delete[] buf;
  H5Tclose(atype);
  H5Aclose(attr);
  return ret;
}
          }
void HDF5HeaderData::clear()
{
  m_real.clear();
  m_int.clear();
  m_string.clear();
  m_intvect.clear();
  m_box.clear();
}

ostream& operator<<(ostream& os, const HDF5HeaderData& data)
{

#define PRINT(Ttype, mapName) \
  for(map<std::string, Ttype>::const_iterator p = data.mapName.begin();        \
      p!= data.mapName.end(); ++p)                   \
  os<<p->first<<" :"<<p->second<<"\n";

  PRINT(Real, m_real);
  PRINT(int, m_int);
  PRINT(std::string, m_string);
  PRINT(IntVect, m_intvect);
  PRINT(Box, m_box);
  return os;
}                        
#endif

/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.

#ifndef _MESH_REFINE_H_
#define _MESH_REFINE_H_
// Include files:



// Constants:


/// class which manages grid generation
class MeshRefine
/**
   This class manages grid generation from sets of tagged cells. It is 
   designed to be a pure virtual base class from which another class
   may be derived with a specific grid-generation algorithm (for example,
   the BRMeshRefine class).
   
   There are two ways grids can be defined based on tagged cells.
   one takes a single IntVectSet of tags defined on the BaseLevel
   mesh and uses that set of tags for every level to be refined;
   the other takes a Vector<IntVectSet> of tags defined on all the
   mesh levels to be refined and uses those. \\ \\
   
   
   {\bf Long Description:} \\
   
   Create new meshes based on tagged cells on a range of levels of a mesh
   hierarchy.  Each level of tagged cells is used to generate a new mesh at
   the next finer level.  The finest level in the output mesh will be one
   level higher than the top of the range of levels given as input.  As a 
   special case, use the same tags (appropriately refined) for all levels.
   
   Usage:
   
   Call the regrid functions after computing error estimates and tagging cells.
   To add a new mesh level, set TopLevel to the index of the finest level
   in the existing mesh and define tags on the finest level.  To keep the
   existing number of mesh levels, set TopLevel to one less than the
   index of the finest level and don't define any tags on the finest level.
   If a single IntVectSet of tags is passed (instead of a
   Vector<IntVectSet>) then the same tags (properly refined) will be used
   for all the new meshes up to level TopLevel+1.  In any case, the
   meshes at levels BaseLevel and below are not modified.  The output
   argument newmeshes will be reallocated to the necessary size before
   being used.  When this function returns, the elements of the
   newmeshes vector corresponding to the unchanged levels will be
   filled in with copies of the levels from the old mesh vector.  The
   variable tags is modified in an undefined way, so its contents
   should not be relied upon.  The variable BlockFactor specifies the
   amount by which each box will be coarsenable. Every grid box will
   have an integral multiple of BlockFactor cells in each dimension and
   also lower index values that are integral multiples.  As a side effect,
   the minimum box size will be BlockFactor.
   
   Expensive validations are done only when debugging is enabled
   (i.e. the DEBUG make variable is "TRUE").
   
   
   Usage Notes:
   
   All the input vectors should be defined with max index >= TopLevel.
   They should have values for indices [BaseLevel:TopLevel].
   (except for OldMeshes, which must be defined for all indices).  The
   new mesh vector newmeshes will be redefined up to index
   TopLevel+1.  RefRatios should be defined such that
   RefRatios[L] is the value to use to refine the level L mesh to
   produce the level L+1 mesh.  The tags vector is modified in an
   undefined manner.  The output variable newmeshes may not be
   completely defined if an exception occurs.  
   The BlockFactor can be used to force a minimum box size.
*/
{
public:
  /// default constructor -- leaves object in an unusable state
  MeshRefine();
  
  /// full constructor -- leaves object in usable state
  MeshRefine(/// level 0 domain
               const Box& a_baseDomain,
               /// refinement ratios -- refRatio[0] is btwn levels 0 and 1
               const Vector<int>& a_refRatios,
               /// measure of how efficiently tagged cells will be covered
               const Real a_fillRatio,
               /// amount by which grids are guaranteed to be coarsenable
               const int a_blockFactor,
               /// proper nesting buffer amount
               const int a_bufferSize,
               /// maximum grid length in any direction -- 0 means no limit.
               const int a_maxSize);

  /// full constructor -- leaves object in usable state
  MeshRefine(/// level 0 domain
               const ProblemDomain& a_baseDomain,
               /// refinement ratios -- refRatio[0] is btwn levels 0 and 1
               const Vector<int>& a_refRatios,
               /// measure of how efficiently tagged cells will be covered
               const Real a_fillRatio,
               /// amount by which grids are guaranteed to be coarsenable
               const int a_blockFactor,
               /// proper nesting buffer amount
               const int a_bufferSize,
               /// maximum grid length in any direction -- 0 means no limit.
               const int a_maxSize);
  

  /// destructor
  virtual ~MeshRefine();
  
  /// define function -- size of RefRatios will define maximum number of levels
  void define(/// level 0 domain
              const Box& a_baseDomain,
              /// refinement ratios -- refRatio[0] is btwn levels 0 and 1
              const Vector<int>& a_refRatios,
              /// measure of how efficiently tagged cells will be covered
              const Real a_fillRatio,
              /// amount by which grids are guaranteed to be coarsenable
              const int a_blockFactor,
              /// proper nesting buffer amount
              const int a_bufferSize,
              /// maximum grid length in any direction -- 0 means no limit
              const int a_maxSize);


  /// define function -- size of RefRatios will define maximum number of levels
  void define(/// level 0 domain
              const ProblemDomain& a_baseDomain,
              /// refinement ratios -- refRatio[0] is btwn levels 0 and 1
              const Vector<int>& a_refRatios,
              /// measure of how efficiently tagged cells will be covered
              const Real a_fillRatio,
              /// amount by which grids are guaranteed to be coarsenable
              const int a_blockFactor,
              /// proper nesting buffer amount
              const int a_bufferSize,
              /// maximum grid length in any direction -- 0 means no limit
              const int a_maxSize);
  
  ///create hierarchy of grids from a single level of tags
  /** This function creates a hierarchy of grids from a single level of 
      tags on BaseLevel.  If tags exist, then all levels will have grids.
      Returns the new finest level of grids.
  */
  virtual int regrid(/// new set of grids at every level
                     Vector<Vector<Box> >& a_newmeshes,
                     /// tagged cells on baseLevel
                     const IntVectSet& a_tags,
                     /// index of base mesh level (finest unchanged level)
                     const int a_baseLevel,
                     /// top level to refine (one less than finest possible level)
                     const int a_topLevel,
                     /// existing grids (if no previous grids, set to domains)
                     const Vector<Vector<Box> >& a_oldMeshes) const;
  
  /// create hierarchy of grids from tags at all levels
  /** This function creates a hierarchy of grids from tags at all 
      refinement levels.  It is possible that not all levels will 
      return with grids, since there may not be tags at all levels.
      Returns the new finest level of grids.
  */
  virtual int regrid(/// new set of grids at every level
                     Vector<Vector<Box> >& a_newmeshes,
                     /// tagged cells on each existing level
                     const Vector<IntVectSet>& a_tags,
                     /// index of base mesh level (finest unchanged level)
                     const int a_baseLevel,
                     /// top level to refine (one less than finest possible level)
                     const int a_topLevel,
                     /// existing grids (if no previous grids, set to domains)
                     const Vector<Vector<Box> >& a_oldMeshes) const;
  
  
  /// Access functions
  
  /// returns vector of refinement ratios
  const Vector<int>& refRatios() const;
  
  /// returns fillRatio
  Real fillRatio() const;
  
  /// returns blocking factor
  int blockFactor() const;
  
  /// returns proper nesting buffer size 
  int bufferSize() const;
  
  /// returns maximum box size in any dimension -- 0 means no limit
  int maxSize() const;
  
  /// sets vector of refinement ratios
  void refRatios(const Vector<int>& a_nRefVect);
  
  /// sets fillRatio
  void fillRatio(const Real a_fill_ratio);
  
  /// sets blocking factor
  void blockFactor(const int a_block_factor);
  
  /// sets proper nesting buffer size 
  void bufferSize(const int a_buffer_size);
  
  /// sets maximum box size in any dimension -- 0 means no limit
  void  maxSize(const int a_max_size);
  
  /// has this object been defined properly?
  bool isDefined() const;
  
  
  /// constructs a set of boxes which covers a set of tagged cells 
  /** constructs a set of boxes which covers a set of tagged cells
      by using the algorithm of choice.  Everything should
      be on the same level, and blocking factor is not applied.
      Boxes will be on the same refinement level as the tags.
      This would normally be a protected function, but it can be useful
      to call it on its own, so it has been left public. 
  */
  virtual void 
  makeBoxes(/// output: refined boxes at each new level
            Vector<Box>&      a_mesh,        
            /// input: set of tagged cells to cover
            const IntVectSet& a_tags,        
            /// input: proper nesting domain in which mesh boxes must live
            const IntVectSet& a_pnd,         
            /// input: physical domain
            const ProblemDomain& a_domain,
            ///input: largest number of cells in any dimension for any box
            const int         a_maxSize) const = 0;
  

protected:

  /// computes local blockFactors used internally to enforce the BlockFactor
  /** This function computes values for m_local_blockfactors array, which
      is the amount that tags on a level are coarsened in order to guarantee
      that the grids on the next finer level are coarsenable by the 
      BlockFactor. */
  virtual void
  computeLocalBlockFactors();
  
  
  /*@ManMemo: Computes proper nesting domains.  This should only be 
    called by refine. it assumes that everything has already been
    coarsened by the local blocking factor
  */
  virtual void
  makePNDs(///output: proper nesting domains at each level 
           Vector<IntVectSet>& a_pnds,    
           /// input: index of highest AMR level not to be refined
           const int           a_baseLevel,       
           /// input: index of highest AMR level in output
           const int           a_topLevel,
           /// input: (same as in \func{meshRefine})
           const Vector<ProblemDomain>&  a_domains,        
           /// input: boxes at mesh level \var{BaseLevel} 
           const IntVectSet&   a_baseMesh, 
           ///input: (similar to \func{meshRefine; but with level-dependent coarsening factors})
           const Vector<int>&  a_bufferSize ) const;  
  

  // local data members
  bool m_isDefined;
  
  Vector<ProblemDomain> m_vectDomains;

  Vector<int> m_nRefVect;

  Real m_fillRatio;

  int m_blockFactor;
  
  Vector<int> m_level_blockfactors;
  
  int m_bufferSize;
  
  int m_maxSize;
  
};


#endif
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//


// base Mesh refinement class
// ---------
//  class MeshRefine
//
// Short Description:
// ------------------
//  Create new meshes based on tagged cells on a range of levels of a mesh
//  hierarchy.  Each level of tagged cells is used to generate a new mesh at
//  the next finer level.  The finest level in the output mesh will be one
//  level higher than the top of the range of levels given as input.  As a 
//  special case, use the same tags (appropriately refined) for all levels.
//  MeshRefine is intended as a pure virtual base class, from which should
//  be derived a class which supplied the actual box-generation algorithm
//  (for example, the BRMeshRefine class implements the Berger-Rigoutsos
//  algorithm).
//
// Usage:
// ------
//  Call the regrid functions after computing error estimates and tagging
//  cells.  To add a new mesh level, set \var{TopLevel} to the index of 
//  the finest level in the existing mesh and define tags on the finest
//   level.  To keep the existing number of mesh levels, set \var{TopLevel} 
//  to one less than the index of the finest level and don't define any tags 
//  on the finest level. If a single IntVectSet of tags is passed (instead 
//  of a Vector<IntVectSet>) then the same tags (properly refined) will be 
//  used for all the new meshes up to level \var{TopLevel}+1.  In any case, 
//  the meshes at levels \var{BaseLevel} and below are not modified.  The 
//  output argument \var{newmeshes} will be reallocated to the necessary 
//  size before being used.  When this function returns, the elements of the
//  \var{newmeshes} vector corresponding to the unchanged levels will be
//  filled in with copies of the levels from the old mesh vector.  The
//  variable \var{tags} is modified in an undefined way, so its contents
//  should not be relied upon.  The variable \var{BlockFactor} specifies a
//  minimum grid size.  Every grid box will have an integral
//  multiple of \var{RefRatio}*\var{ceiling(BlockFactor/RefRatio)} cells 
//  in each dimension and also lower index values that are integral multiples.
//
//
// Arguments:
// ----------
//  \var{newmeshes}                type: Vector<Vector<Box>>
//      (output, index[0:\var{TopLevel}+1]) vector containing the new mesh hierarchy
//
//  \var{tags}                     type: Vector<IntVectSet>
//      (input/output, index [0:\var{TopLevel}]) tagged cells to be covered by the
//      new meshes.  For any level L, the new mesh at level L+1 will try to cover
//      the tagged cells at level L.  NOTE: This variable is modified in an
//      undefined manner so its values should not be used on exit
// --or--
//  \var{tags}                     type: IntVectSet
//      (input/output) tagged cells on the \var{BaseLevel} mesh.  These tags
//      will be refined and reused for the other mesh levels.
//
//  \var{BaseLevel}                type: int
//      (input, range[0:\var{TopLevel}]) index of the finest mesh level that
//      will not be modified.
//
//  \var{TopLevel}                 type: int
//      (input, range[\var{BaseLevel}:\var{OldMeshes.size()-1]) index of the
//      finest mesh level that will be refined (i.e. the finest mesh level for
//      which tags are defined); the new mesh vector \var{newmeshes} will have
//      an element at index [\var{TopLevel}+1].
//
//  \var{Domains}                  type: Vector<ProblemDomain>
//      (input, index[0:\var{TopLevel}]) boundaries of the physical domain,
//      one mesh box at each level.  Must be consistent with \var{RefRatios}
//      (i.e. Domains[levelL+1] == Domains[levelL] * RefRatios[levelL])
//
//  \var{OldMeshes}                type: Vector<Vector<Box>>
//      (input, index[0:\var{TopLevel}]) the existing mesh hierarchy vector.
//      The \var{BaseLevel} boxes should be aligned on \var{BufferSize}-sized
//      internal boundaries.
//
//  \var{RefRatios}                type: Vector<int>&
//      (input, index[0:\var{TopLevel}], range[2^i,i>0])
//      the number of cells (in each direction) in level L+1 for each cell in
//      level L.  Must be an integral power of 2.
//
//
//  \var{FillRatio}                type: Real
//      (input, range[0:1]) the minimum acceptable ratio
//      of filled cells in a box in the Berger-Rigoutsos algorithm in
//      \func{makeBoxes} (typical value = 0.75
//
//  \var{BlockFactor}              type: int
//      (input, range[1:) )
//      factor  that determines the minimum size of a box
//      and what index values a box corner may have (i.e. what alignment is 
//      has). Grids are guaranteed to be coarsenable by the BlockFactor.
//
//  \var{BufferSize}               type: int
//      (input, range[1:), default=1)
//      minimum number of level L mesh cells between the boundary of a level L+1
//      mesh and the nearest level L-1 mesh cell.  This is used to guarantee 
//      that stencils can be computed across coarse-fine boundaries.
//
//  \var{MaxBoxSize}               type: int
//      (input, range[2*BufferSize:), default=\infinity)
//      largest number of grid points in any dimension that will be generated
//      in an individual grid box.  Must be at least twice the BufferSize.
//      
// Returns:
// --------
//  The first two arguments (\var{tags}, \var{newmeshes}) are modified.
//
//  Expensive validations are done only when debugging is enabled
//   (i.e. the DEBUG make variable is "TRUE").
//   
//
// Usage Notes:
// ------------
//  All the input vectors should be defined with max index >= \var{TopLevel}.
//  They should have values for indices [\var{BaseLevel}:\var{TopLevel}].
//  (except for \var{OldMeshes}, which must be defined for all indices).  The
//  new mesh vector \var{newmeshes} will be redefined up to index
//  \var{TopLevel}+1.  \var{RefRatios} should be defined such that
//  \var{RefRatios}[L] is the value to use to refine the level L mesh to
//  produce the level L+1 mesh.  The \var{tags} vector is modified in an
//  undefined manner.  The output variable \var{newmeshes} may not be
//  completely defined if an exception occurs.  If you do not know what the
//  optional arguments do, it is probably safe to ignore them.
//  The BlockFactor can be used to force a minimum box size.
//
// Long Description:
// -----------------
//  MeshRefine replaces the meshes at levels \var{BaseLevel}+1 to
//  \var{TopLevel} and creates a new mesh at level \var{TopLevel}+1.  Levels
//  \var{BaseLevel} and below in the new mesh hierarchy are the same as in
//  the old one.  The new meshes are generated so that the tagged cells at
//  for tagged cells that aren`t properly nested in the \var{BaseLevel}
//  mesh).  If TopLevel<0, no new meshes are created and a warning code
//  is returned.
//
// Numerical Algorithm:
// --------------------
//  The tagged cells at level L are used to generate a new (refined) mesh at
//  level L+1.  The \var{BaseLevel} mesh is unchanged.  The new meshes must
//  satisfy the proper nesting requirements, i.e. each new mesh box must be
//  contained in the proper nesting domain (PND) for the next lower level.  The
//  PNDs of all levels are derived from the \var{BaseLevel} mesh (see
//  \func{makePNDs}).  The basic idea is that the proper nesting domain of a
//  level L mesh contains the interior cells of the level L-1 PND, except for
//  faces that are on the physical boundary.  The base level PND is defined as
//  the interiors of the base level mesh boxes, except for boundary faces
//  (again) and box face cells that are surrounded by cells of other boxes.
//
//  In the following graphic, the PND for two mesh boxes with BufferSize=1 is
//  shown.  
//
//        ###################################################
//        #     |   |   |   |   |   |                       #
//        #     |   | o | o | o |   |                       #
//        #     |   |   |   |   |   |                       #
//        #     +---+---+---+---+---%---+---+---+---+---+---#
//        #     |   |   |   |   |   %   |   |   |   |   |   #
//        #     |   | o | o | o |   %   |   |   |   |   |   #
//        #     |   |   |   |   |   %   |   |   |   |   |   #
//        #     +---+---+---+---+---%---+---+---+---+---+---#
//        #     |   |   |   |   |   %   |   |   |   |   |   #
//        #     |   | o | o | o | o % o | o | o | o | o | o #
//        #     |   |   |   |   |   %   |   |   |   |   |   #
//        #     +---+---+---+---+---%---+---+---+---+---+---#
//        #     |   |   |   |   |   %   |   |   |   |   |   #
//        #     |   |   |   |   |   %   | o | o | o | o | o #
//        #     |   |   |   |   |   %   |   |   |   |   |   #
//        #     +---+---+---+---+---%---+---+---+---+---+---#
//        #                         |   |   |   |   |   |   #
//        #                         |   | o | o | o | o | o #
//        #                         |   |   |   |   |   |   #
//        ###################################################
//
//  where:
//   #    are physical boundaries (non periodic)
//   | +  are cell boundaries of the two mesh boxes
//   %    are box boundaries shared between boxes
//   o    are cells in the proper nesting domain
//
//  The tagged cells outside the proper nesting domains are discarded by
//  intersecting the set of tagged cells at each level with the PND at that
//  level.  This ensures that all cells in the refined meshes are properly
//  nested.  The actual mesh generation starts at the top of the existing mesh
//  (\var{TopLevel}) and works down.  At each level a set of mesh boxes that
//  \func{makeBoxes}, which is provided in the derived class (e.g. 
//  BRMeshRefine).  The goal is to minimize the number of boxes 
//  and maximize the ratio of tagged to untagged cells in each box.  
//  The boxes are refined to the next higher level and saved as the new 
//  mesh (in \var{newmeshes}).  The unrefined boxes are coarsened and
//  projected down to the next lower level where a buffer of bufferSize cells
//  is added around all the boxes (the opposite of taking the interior 
//  cells that is done in building the PND) and added to the tagged cells 
//  at the coarser level.  This ensures that the next coarser level will 
//  have meshes that will properly nest the meshes just generated.  This 
//  is repeated until the BaseLevel+1 mesh is replaced.
//
//
// Implementation Method:
// ----------------------
//  There are 5 basic steps to perform.  Two functions (\func{makePNDs} and
//  \func{makeBoxes}) are called to handle the most complex steps and the code
//  in this function handles the rest.
//  Start
//   1) Compute the proper nesting domains (PNDs).  A \var{Vector<IntVectSet>}
//      is created and passed to \func{makePNDs}, along with the grid and
//      boundary data.  Only the part of the Vector that is active
//      [\var{BaseLevel:TopLevel}] is modified.  The other elements of the
//      vector are not touched.
//   2) Removing (clipping) the tag cells outside the PNDs is trivial:
//      intersect the two \var{IntVectSets} (PND and tag cells) at each level.
//      [Note: The PND data is not needed after this step.]
//  Loop for L from \var{TopLevel} downto \var{BaseLevel}
//   3) \func{makeBoxes} is called to find a set of boxes that cover the tag
//      cells at this level.  The output is stored in a temp \var{BoxArray}
//   4) The temp \var{BoxArray} is refined and the result is stored in the
//      output mesh vector at level L+1.
//   5) The temp \var{BoxArray} is coarsened to the next coarser level, then
//      expanded by the buffer cells in each direction (clipping at the 
//      physical boundaries) and then  unioned with the tags at level L-1.  
//      This step does not have to be done for the base level.  The resulting
//      tags are used to generate the mesh at the next coarser level.
//  EndLoop
//
// Implementation Notes:
// ---------------------
//  The output vector \var{newmeshes} is redefined to the necessary size
//  without necessarily reusing the existing memory space.
//
///////////////////////////////////////////////////////////////////////////////


// Include files:


MeshRefine::MeshRefine() : m_isDefined(false)
{
}

MeshRefine::~MeshRefine()
{
}

MeshRefine::MeshRefine(const Box& a_baseDomain,
                           const Vector<int>& a_refRatios,
                           const Real a_fillRatio,
                           const int a_blockFactor,
                           const int a_bufferSize,
                           const int a_maxSize)
{
  ProblemDomain crseDom(a_baseDomain);
  define(crseDom, a_refRatios, a_fillRatio, a_blockFactor,
         a_bufferSize, a_maxSize);
}


MeshRefine::MeshRefine(const ProblemDomain& a_baseDomain,
                           const Vector<int>& a_refRatios,
                           const Real a_fillRatio,
                           const int a_blockFactor,
                           const int a_bufferSize,
                           const int a_maxSize)
{
  define(a_baseDomain, a_refRatios, a_fillRatio, a_blockFactor,
         a_bufferSize, a_maxSize);
}

void
MeshRefine::define(const Box& a_baseDomain,
                     const Vector<int>& a_refRatios,
                     const Real a_fillRatio,
                     const int a_blockFactor,
                     const int a_bufferSize,
                     const int a_maxSize)
{
  ProblemDomain crseDom(a_baseDomain);
  define(crseDom, a_refRatios, a_fillRatio, a_blockFactor,
         a_bufferSize, a_maxSize);
}



void
MeshRefine::define(const ProblemDomain& a_baseDomain,
                     const Vector<int>& a_refRatios,
                     const Real a_fillRatio,
                     const int a_blockFactor,
                     const int a_bufferSize,
                     const int a_maxSize)
{
  int maxLevel = a_refRatios.size();
  m_nRefVect.resize(maxLevel);
  m_vectDomains.resize(maxLevel+1);
  m_level_blockfactors.resize(maxLevel);

  m_vectDomains[0] = a_baseDomain;
  for (int lev=0; lev<maxLevel; lev++) {
    m_nRefVect[lev] = a_refRatios[lev];
    m_vectDomains[lev+1] = refine(m_vectDomains[lev], a_refRatios[lev]);
  }
  
  // do some quick sanity checks
  assert( a_blockFactor >= 1 );                      
  assert( a_bufferSize >= 0 );                         
  assert( (a_maxSize >= 2*a_bufferSize) || (a_maxSize == 0) );
  assert( (a_blockFactor <= a_maxSize) || (a_maxSize == 0) );             
  assert( a_fillRatio > 0.0 || a_fillRatio <= 1.0 );
  
  m_fillRatio = a_fillRatio;
  m_blockFactor = a_blockFactor;
  m_bufferSize = a_bufferSize;
  m_maxSize = a_maxSize;

  computeLocalBlockFactors();

  m_isDefined = true;

}


//
// This regrid function takes a single IntVectSet of tags
//
int
MeshRefine::regrid(Vector<Vector<Box> >&   a_newmeshes,
                   const IntVectSet&             a_tags,
                   const int               a_baseLevel,
                   const int               a_topLevel,
                   const Vector<Vector<Box> >& a_oldMeshes) const
{
    //
    // Handle special cases
    //
    assert (a_topLevel >= 0 );
    assert ( a_baseLevel < (a_topLevel+1) && a_baseLevel >= 0 );

    //
    // Create a vector of tags from the single set of tags that is given
    Vector<IntVectSet> tags_vector( a_topLevel+1 ,a_tags );

    // Refine the levels above the base level
    for( int i = 1 ; i <= a_topLevel ; i++ )
      {
         tags_vector[i] = refine(tags_vector[i-1], m_nRefVect[i-1] ) ;
      }
    
    // run the version of refine that takes a vector of tags
    return regrid(a_newmeshes, tags_vector,
                  a_baseLevel, a_topLevel, a_oldMeshes);
}




//
// This version of refine takes a Vector of tags
//
int
MeshRefine::regrid(Vector<Vector<Box> >&   a_newmeshes,
                   const Vector<IntVectSet>&     a_tags,
                   const int               a_baseLevel,
                   const int               a_topLevel,
                   const Vector<Vector<Box> >& a_OldMeshes) const
{

  //
  // Validate arguments and handle special cases
  //
  assert (isDefined());
  
  assert ( a_topLevel >= 0 );
  assert ( a_baseLevel < (a_topLevel+1) && a_baseLevel >= 0 );
  assert ( a_OldMeshes.size() >= a_topLevel + 1 );
  assert (   m_vectDomains.size() >= a_topLevel + 1 );
  assert( m_nRefVect.size() >= a_topLevel + 1 );
  assert(      a_tags.size() >= a_topLevel + 1 );           

  Vector<IntVectSet> modifiedTags;

  //
  // set the top level to be the finest level which actually has tags
  bool found_top = false;
  int TopLevel = a_topLevel;
  int new_finest_level;
  // need this because .size() returns an unsigned int
  int isize = a_tags.size();
  for (int level = Min(a_topLevel, isize-1);
       !found_top &&(level >= a_baseLevel);
       --level) 
    {
      if (!a_tags[level].isEmpty()) 
        { 
          found_top = true;
          TopLevel = level;
        }
    }

  if (found_top) 
    {

      // copy tags to local storage so we can mess with them.
      modifiedTags.resize(a_tags.size());
      for (int ilev = a_baseLevel; ilev <= TopLevel; ilev++)
        {
          modifiedTags[ilev] = a_tags[ilev];
        }
  

      
      //
      // reinitialize the output array
      //
      a_newmeshes.resize( TopLevel+2 ) ;
      
      //
      // Generate new meshes if requested.
      //
      if( TopLevel+1 > a_baseLevel ) 
        {
          
          Box domaint = m_vectDomains[a_baseLevel].domainBox();
          Box testdom = coarsen(domaint, m_blockFactor);
          testdom.refine(m_blockFactor);
          if(domaint != testdom)
            MayDay::Error("MeshRefine:domain and Blocking Factor incompatible");
          // [NOTE: the following validations are expensive so only do them when
          //        debugging.]
          
#if ! defined( NDEBUG )
          // all existing meshes should be within the boundaries
          for( int i = a_baseLevel ; i <= TopLevel ; i++ ) {
            Box minbox = a_OldMeshes[i][0];
            for (int ibox = 0; ibox < a_OldMeshes[i].size(); ++ibox)
              {
                minbox.minBox(a_OldMeshes[i][ibox]);
              }
            assert(m_vectDomains[i].contains(minbox));
          }
#endif
          // all tagged cells must be within the existing meshes
          // make sure that each level of \var{Domains} is consistent with the 
          // previous level and the refinement ratio.
          for( int i = a_baseLevel ; i < TopLevel ; i++ ) 
            {
              if(m_vectDomains[i+1].domainBox() 
                 != refine(m_vectDomains[i].domainBox(), m_nRefVect[i])) 
                MayDay::Error("MeshRefine:domains and refratios incompatible");
            }
          //
          // coarsen old meshes, tags, problem domains by the appropriate
          //  blocking factor  
          Vector<Vector<Box> > OldMeshes = a_OldMeshes;
          Vector<ProblemDomain> Domains = m_vectDomains;
          for (int level = 0; level < modifiedTags.size (); ++level)
            {
              // first intersect tags with domains
              // (note that in periodic case, this can
              // result in wrapping of tags that are 
              // outside the base domain box
              modifiedTags[level] &= m_vectDomains[level];

              // coarsen blocking factor, rounding upwards
              // the goal of this is to coarsen everything down
              // to a level which is m_blockFactor coarser than
              // the new fine level we're going to generate.  By
              // generating grids at this fake level, we can then
              // refine up to the new level, and the blocking factor
              // will be automatically enforced.
              modifiedTags[level].coarsen (m_level_blockfactors[level]);
            }
          for (int level = 0; level < OldMeshes.size (); ++level)
            {
              if (level < m_level_blockfactors.size())
                {
                  for (int box = 0; box < OldMeshes[level].size (); ++box)
                    {
                      OldMeshes[level][box].coarsen(m_level_blockfactors[level]);
                    }
                  Domains[level].coarsen(m_level_blockfactors[level]);
                }
            }
          //
          // Compute proper nesting domains from a_baseLevel up to TopLevel.
          // [NOTE: this leaves pnds[0:a_baseLevel-1] empty.]
          //dtg
          Vector<IntVectSet> pnds( TopLevel+1 ) ;
          IntVectSet oldmesh_ivs;
          for (int box = 0; box < OldMeshes[a_baseLevel].size(); ++box)
            {
              oldmesh_ivs |= OldMeshes[a_baseLevel][box];
            }
          
          // GHM 7/30/01
          // the buffer size argument to makePND needs to be coarsened by
          // the BlockFactor, since all the other arguments have been so
          // coarsened.  If not, a circumstance may occur where a set of
          // grids defined with a certain set of tags does not "refine" to
          // itself given the exact same set of tags.
          
          Vector<int> blocked_BufferSize(TopLevel+1);
          for( int level=0; level<=TopLevel; level++ ) {
            blocked_BufferSize[level] = 
              (m_bufferSize +m_level_blockfactors[level] - 1 )/
              m_level_blockfactors[level];
          }
          
          // note that this is called _after_ we do all the coarsening to 
          // enforce blocking factor!
          makePNDs(pnds, a_baseLevel, TopLevel, Domains, oldmesh_ivs,
                   blocked_BufferSize);
          
          
          //
          // Clip the tagged cells to the proper nesting domains by
          // intersecting the two sets.
          //
          for( int lvl = a_baseLevel ; lvl <= TopLevel ; lvl++ ) {
            modifiedTags[lvl] &= pnds[lvl] ;
          }
          
          //
          // Generate new meshes.
          //
          
          // At each level, starting at the top, generate boxes that cover the tag
          // cells using makeBoxes(), refine these boxes and save them in the
          // output \var{a_newmeshes}.  Take the unrefined boxes, add a buffer zone 
          // around each box, coarsen, and union with the tag cells on the next 
          // coarser level.  This modifies the tags variable.  To handle 
          // \var{BlockFactor}, coarsen everything before making the new
          // meshes and then refine the resulting mesh boxes.
          Vector<Box> lvlboxes ;  // new boxes on this level
          for( int lvl = TopLevel ; lvl >= a_baseLevel ; lvl-- ) {
            ProblemDomain lvldomain ; // domain of this level
            
            lvldomain = Domains[lvl] ; 
            
            // Union the meshes from the previous level with the tags on this
            // level to guarantee that proper nesting is satisfied.  On the
            // first iteration this is a no-op because \var{lvlboxes} is empty.
            // [NOTE: for every iteration after the first, \var{lvlboxes} will
            //        already be coarsened by \var{BlockFactor} so it will be
            //        at the same refinement level as \var{tags[lvl]}, which 
            //        has also been coarsened]
            // this is simple in the non-periodic case, more complicated 
            // in the periodic case 
            if (lvldomain.isPeriodic() )
              {
                const Box domainBox = lvldomain.domainBox();
                ShiftIterator shiftIt = lvldomain.shiftIterator();
                IntVect shiftMult(domainBox.size());
                for (int i=0; i<lvlboxes.size(); i++)
                  {
                    Box localBox(lvlboxes[i]);
                    // will handle periodic wraparound through shifting
                    // and adding shifted image to tags, which will all
                    // remain within the domainBox
                    localBox &= domainBox;
                    modifiedTags[lvl] |= localBox;
                    // now do shifts to capture periodic images 
                    // necessary to enforce proper nesting
                    // only do this if original box was not 
                    // contained by domainBox
                    if (localBox != lvlboxes[i]) 
                      {
                        for (shiftIt.begin(); shiftIt.ok(); ++shiftIt)
                          {
                            IntVect shiftVect(shiftIt()*shiftMult);
                            Box localShiftedBox(lvlboxes[i]);
                            localShiftedBox.shift(shiftVect);
                            localShiftedBox &= domainBox;
                            if (!localShiftedBox.isEmpty())
                              modifiedTags[lvl] |= localShiftedBox;
                          } // end loop over shift directions
                      } // end whether periodic checking was needed
                  } // end loop over finer-level boxes to enforce nesting 
              } 
            else 
              {
                // non periodic case is simple
                for( int i = 0 ; i < lvlboxes.size() ; i++ )
                  modifiedTags[lvl] |= lvlboxes[i] ;
              }
            
            // make a new mesh at the same level as the tags
            makeBoxes(lvlboxes, modifiedTags[lvl], pnds[lvl], 
                      lvldomain, m_maxSize/m_blockFactor);
            
            
            // refine the new mesh and save it
            //[NOTE: we have to undo the coarsening by \var{BlockFactor} as well
            //       as refine to the next level.]
            //[NOTE: refine() operates in-place so copy first then refine()
            //       because the unrefined mesh will be needed later.]
            a_newmeshes[lvl+1] = lvlboxes ;
            for (int ibox = 0; ibox < a_newmeshes[lvl+1].size(); ++ibox)
              {
                a_newmeshes[lvl+1][ibox].refine(m_level_blockfactors[lvl]*m_nRefVect[lvl]);
              }
            // Make the boxes ready for the next iteration.
            // Don't have to do this for the last iteration.
            if( lvl > a_baseLevel ) 
              {
                
                // coarsen the unrefined new mesh so it matches the tags
                // at that level, then add the buffer cells, clip at the domain
                // boundaries so we can union them in the next iteration
                for( int i = 0 ; i < lvlboxes.size() ; i++ ) 
                  {
                    // since we want to have a BufferSize border
                    // of (lvl-1) cells, this translates into 
                    // growing the coarsened boxes by the BufferSize
                    // coarsened by the same factor we used to get the
                    // tags from level (lvl-1) to this intermediate level
                    lvlboxes[i].grow(Max(1, m_bufferSize/m_level_blockfactors[lvl]));
                    lvlboxes[i] &= Domains[lvl];
                    
                    // (dfm 3/26/2001) since you're working on a set of 
                    // faux levels which are a factor of m_blockFactor
                    // coarser than the levels you are _GENERATING_
                    // (in this case lvl+1), when you coarsen these
                    // boxes down to the next coarser level, you
                    // want to coarsen them by the refinement 
                    // ratio between levels (lvl+1) and (lvl),
                    // instead of the ratio between (lvl) and
                    // (lvl-1).
                    lvlboxes[i].coarsen( m_nRefVect[lvl] ) ;
                  }
                
              }
          }
        }
      
      //
      // Finally, copy the old mesh levels that didn't change.
      //
      for( int i = 0 ; i <= a_baseLevel ; i++ )
        a_newmeshes[i] = a_OldMeshes[i] ;
      
      //
      // Done generating grids
      //
      
      // set new finest level
      new_finest_level = TopLevel+1;
      // this is designed to catch the pathological but possible case
      // where there were tags on the TopLevel, but no grids were generated
      // (possibly if all tags were outside pnds)
      while (a_newmeshes[new_finest_level].size() == 0 
             && new_finest_level > a_baseLevel) 
        {
          new_finest_level -= 1;
        }

    } else {
      // if no tags on any level, just return
      new_finest_level = a_baseLevel;
    }
  
  return new_finest_level;
  
  
}// end of regrid function


void
MeshRefine::computeLocalBlockFactors()
{
  // computes amount that tags on a level l must be coarsened so
  // that grids on next finer level l+1 will be guaranteed to 
  // satisfy the blocking factor.
  // The easiest way to do this is to coarsen everything down
  // to a level which is m_blockFactor coarser than the new 
  // fine level which will be generated.  So, m_level_blockfactor[lvl]
  // is the amount by which the tags, etc at level lvl must
  // be coarsened so that they are m_blockFactor coarser than
  // the (lvl+1) grids which will be generated.  If m_BlockFactor
  // is less than the refinement ratio between levels (lvl) and (lvl+1),
  // then no coarsening needs to be done, so we default to one, in that case.
  for (int lev=0; lev<(m_level_blockfactors.size()); lev++)
    {
      m_level_blockfactors[lev] = (m_blockFactor + m_nRefVect[lev] -1)/m_nRefVect[lev];
    }
}

///////////////////////////////////////////////////////////////////////////////
//
// Function:
// ---------
//  int makePNDs()
//
// Short Description:
// ------------------
//  Compute the proper nesting domains for a subset of the levels in a
//  multilevel adaptive mesh.
//
// Usage:
// ------
//  This is called by \func{regrid}.  It wasn't designed to be called
//  directly so if you do so you're on your own.  The output is a
//  \var{Vector<IntVectSet>}.  This function is called to compute the proper
//  nesting domains needed to do mesh refinement.  The function that calls this
//  one must have the mesh and problem domain data structures already defined 
//  for all the levels and have allocated a vector<IntVectSet> for the proper 
//  nesting domains.  This function computes the proper nesting domains of the
//  levels to be refined (i.e., the levels \var{a_baseLevel} and above) and 
//  modifies the \var{pnds} array.  The output vector is undefined for 
//  indices corresponding to mesh levels below \var{a_baseLevel}
//  (i.e. [0:\var{a_baseLevel}-1]).
//
//  NOTE: in MeshRefine::regrid, this function is called _after_ everthing 
//        has been coarsened to enforce the blocking factor, so this behaves
//        differently than one might expect -- call outside of the regrid fn
//        at your own risk! (DFM 8/24/01)
//
// Arguments:
// ----------
//  \var{pnds}                     type: Vector<IntVectSet>
//      (output, index[\var{a_baseLevel}:\var{TopLevel}]) proper nesting domains
//
//  \var{a_baseLevel}                type: int
//      (input, range[0:\var{TopLevel}]) index of the mesh level to use as the
//      source for the proper nesting domains
//
//  \var{TopLevel}                 type: int
//      (input, range[\var{a_baseLevel}:\var{OldMeshes.size()-1]) index of the
//      finest mesh level to compute a PND for
//
//  \var{BaseMesh}                 type: Vector<Box>
//      (input) the boxes at level \var{a_baseLevel} in the mesh hierarchy.
//
//  \var{Domains}                  type: Vector<Box>
//      [same as for \func{MeshRefine}.]
//
//  \var{BufferSize}              type: int
//      [same as for \func{MeshRefine}.]
//
// Returns:
// --------
//  An int giving an exception number is returned.  The argument
//  \var{pnds} is modified.
//
// References:
// -----------
//  See ??? for a description of Proper Nesting Domains.
//  meshRefine -- calls this function
//
// Numerical Algorithm:
// --------------------
//  The proper nesting domain (PND) of level L+1 is defined as the interior of
//  the PND for level L (the next coarser) except at the boundaries of the
//  problem domain (ie, the level 0 box), in which case the PNDs of levels L
//  and L+1 both contain the boundary.  The PND of the base level is defined as
//  the interior of the union of all the grid boxes in the base level.  Given
//  the PND of level L (denoted by PND[L]) this function computes PND[L+1] by
//  iterating over the coordinate directions and computing the cells to remove
//  from the proper nesting domain due to the boundaries in that direction.
//  For each direction the domain is shifted (first up, then down) and only the
//  part that is shifted _out_ of the PND but remains _inside_ the problem
//  boundary is kept and then shifted back into the PND and subtracted
//  (removed) from it.  So if the shift moves part of the PND outside the
//  problem boundary then no cells will be removed.  If the cells that get
//  shifted out of the PND remain inside the boundary then they will be removed
//  from the PND when they get shifted back.  When all the directions have been
//  handled, the PND is refined to make it correct for level L+1.  The PND of
//  the base level is the union of the mesh boxes in the base level and serves
//  as the starting point.  In set notation the operation is:
//
//            D - ( ( ( (D << d) - D ) * B ) >> d )
//
//  where:
//    d is the current direction (+i,-i,+j,etc)
//    B is the boundary of the problem domain at the current level
//    - is the subtraction (removal) operator
//    * is the intersection operator
//   << is the down-shift operator (shift in negative direction)
//   >> is the up-shift operator (shift in positive direction)
//
//
// Implementation Method:
//  The PNDs are implemented as a vector of \type{IntVectSet}s.  A scratch 
//  IntVectSet is used to store the PND of the current level while it is 
//  being computed.  When the PND for a level is complete, it is refined 
//  and used for the next level.  The base PND is computed as the union of 
//  the mesh boxes in the base level mesh.  The PNDs for each level are 
//  stored in the appropriate elements of the pnds vector as they are 
//  computed.  Levels below a_baseLevel are not modified.  The code loops 
//  through each level starting from Baselevel.  Loop through each coordinate
//  direction applying the operation defined in the "Numerical Algorithm" 
//  section above.  Copy the final result into the PNDs vector.  If an 
//  exception occurs, do not change PNDs on the level that took the
//   exception.
//
// Implementation Notes:
//  This assumes cell-centers are being manipulated, not vertices.  The whole
//  vector of pnds is passed even though only some are accessed
//  because it is more convenient since these variables are likely to exist
//  already.
//
///////////////////////////////////////////////////////////////////////////////

void
MeshRefine::makePNDs(Vector<IntVectSet>&       a_pnds, 
                       const int            a_baseLevel,
                       const int             a_topLevel,
                       const Vector<ProblemDomain>&     a_domains,
                       const IntVectSet&    a_baseMesh, 
                       const Vector<int>&   a_bufferSize) const
{
  //
  // Validate inputs
  //
  assert( a_baseLevel <= a_topLevel && a_baseLevel >= 0 );
  assert(   a_domains.size() >= a_topLevel + 1 );       
  assert( m_nRefVect.size() >= a_topLevel + 1 );       
  // all existing boxes on the base level must be aligned to \var{BuffSize}


  a_pnds[a_baseLevel] = a_baseMesh;
  
  for( int lvl = a_baseLevel ; lvl <= a_topLevel ; lvl++)
    {
      IntVectSet& pnd = a_pnds[lvl];
      
      pnd.nestingRegion(a_bufferSize[lvl], a_domains[lvl]);

      if( (a_topLevel - lvl) > 0)
        {
          a_pnds[lvl+1] = pnd;


          // note that we use the refinement ratio btwn 
          // lvl+1 and lvl+2 in this case.  this is because
          // when this has been called from MeshRefine,
          // each level has already been brought the the
          // blocking-factor-enforcing coarseness of 
          // lvl+1 coarsened by the blocking factor
          // (since that's where the grids produced by
          // the tags on lvl will eventually wind up)
          a_pnds[lvl+1].refine(m_nRefVect[lvl+1]);
        }
      
    }
  
}
//end of makePNDs



// ---------------------------------------------------------------
// Utility and access functions
// --------------------------------------------------------------

  /// returns vector of refinement ratios
const Vector<int>& 
MeshRefine::refRatios() const
{
  return m_nRefVect;
}
  
/// returns fillRatio
Real 
MeshRefine::fillRatio() const
{
  return m_fillRatio;
}
  
/// returns blocking factor
int 
MeshRefine::blockFactor() const
{
  return m_blockFactor;
}
  
/// returns proper nesting buffer size 
int 
MeshRefine::bufferSize() const
{
  return m_bufferSize;
}
  
/// returns maximum box size in any dimension
int 
MeshRefine::maxSize() const
{
  return m_maxSize;
}

/// sets vector of refinement ratios
void 
MeshRefine::refRatios(const Vector<int>& a_nRefVect)
{
  int isize = a_nRefVect.size();
  m_nRefVect.resize(isize);
  for (int i=0; i<isize; i++) 
    {
      m_nRefVect[i] = a_nRefVect[i];
    }
  // will need to recompute local blockfactors in this case
  computeLocalBlockFactors();
}
  
/// sets fillRatio
void 
MeshRefine::fillRatio(const Real a_fill_ratio)
{
  m_fillRatio = a_fill_ratio;
}
  
/// sets blocking factor
void 
MeshRefine::blockFactor(const int a_block_factor)
{
  m_blockFactor = a_block_factor;
  // then need to recompute local blockFactors
  computeLocalBlockFactors();
}

/// sets proper nesting buffer size 
void 
MeshRefine::bufferSize(const int a_buffer_size)
{
  m_bufferSize = a_buffer_size;
}
  
/// sets maximum box size in any dimension
void  
MeshRefine::maxSize(const int a_max_size)
{
  m_maxSize = a_max_size;
}

/// has this object been defined properly?
bool 
MeshRefine::isDefined() const
{
  return m_isDefined;
}


/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef MEMTRACK_H
#define MEMTRACK_H

#ifdef ENABLE_MEMORY_TRACKING

using std::ostream;

// user functions

void ReportUnfreedMemory(ostream& os);
void ReportMemoryUsage(ostream& os);
void UnfreedMemory(); // send output to pout()
void MemoryUsage();   // sned output to pout()

void memTrackingOn();
void memtrackingOff();

// =========== end user functions.===================================


void AddTrack(void* addr,  size_t asize,  const char *fname, 
              int lnum, bool malloc);
void RemoveTrack(void* addr, bool malloc);


// void * operator new(size_t size, char const *file, int line);

// void operator delete(void *p) throw();

// void * operator new[](size_t size, char const *file, int line);

// void operator delete[](void *p) throw();


// these are ordered in this way so that we don't double track malloc
// functions. 
void* mallocp(size_t size, const char* file, int line);
void* reallocp(void* p, size_t size, const char* file, int line);
void* callocp(size_t nelem, size_t elsize, const char* file, int line);
void  freep(void* p);

#define calloc(a,b)   callocp(a,b,__FILE__, __LINE__)
#define realloc(a, b) reallocp(a, b, __FILE__, __LINE__)
#define malloc(a)     mallocp(a, __FILE__, __LINE__)
#define free(a)       freep(a)

#endif

#endif

/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

// track the allocation tracker itself.
//
//  addendum.  ha ha , my tracker outfoxed me anyways.  since
//  our heavy use allocators have their own tracking mechanisms
//  (BaseFab, Pool, Vector) I end up tracking some of the trackers
//  memory anyways.  I'm aware of it all, just have to be careful
//  about things.  bvs

#ifdef ENABLE_MEMORY_TRACKING


extern "C" {
}


long int BYTES_PER_MEG = 1024*1024;

using std::list;
using std::cout;
using std::endl;

void UnfreedMemory();

// function to register for listening to signal SIGUSR1 to
// provide a snapshot of the current memory layout
void dumpmemorymap(int sig)
{
  UnfreedMemory(); // act on signal
#ifndef CH_IRIX64
  signal(sig, dumpmemorymap); // reset self for future requests
#endif
}

void dumpmemoryatexit()
{
  Copier::s_motionItemPool.clear();
  IntVectSet::clearStaticMemory();
  /*  this code was from my mad debugging of the static order
      initialization bug in the memory tracking code that was
      making the optimized code behave differently than the 
      debug code, because static initialization order is not
      preserved during an optimizaed compilation.  bvs

  incr::iterator i;
  lst::iterator  L;
  for(i=vectorIncr_->begin(); i!= vectorIncr_->end(); i++)
    {
      cout<<i->first<<"  ";
      for(L=i->second->begin(); L!= i->second->end(); L++)
        cout << " "<<*L;
      cout<<"\n";
    }
  */ 
  UnfreedMemory();
//  delete  Arena::arenaList_;
//  delete  Pool::poolList_;
//  delete  vectorList_;
//  delete  vectorIncr_;
//  delete  Arena::s_Arena;
}

void dumpmemoryabort(int sig)
{
  UnfreedMemory();
  signal(sig, SIG_DFL);
  raise(sig);
}

int registerMemorySignals()
{
#ifndef MPI
  atexit(dumpmemoryatexit);
#ifndef CH_IRIX64
  signal(SIGABRT, dumpmemoryabort);
  signal(SIGUSR2, dumpmemorymap);
#endif
#endif
  return 1;
}

static int registerAbortHandler = registerMemorySignals();


VectorList* vectorList_ = NULL;
incr*       vectorIncr_ = NULL;

typedef struct{
  void*  address;
  size_t size;
} chunk ;

typedef struct {
  std::list<chunk*>    chunks;
  char            file[64];
  int             line;
  long            total;
  long            peak;
  bool            mlloc;
} ALLOC_INFO;

Pool chunkPool(sizeof(chunk), "memtrack:chunk");
Pool infoPool(sizeof(ALLOC_INFO), "memtrack:locations"); 

typedef std::list<ALLOC_INFO*> AllocList;

AllocList *allocList;
unsigned long grandTotal = 0;
unsigned long grandPeak  = 0;

bool trackingOn = true;

void memTrackingOn()
{
  trackingOn = true;
}

void memtrackingOff()
{
  trackingOn = false;
}


void AddTrack(void* addr, size_t asize, const char *fname, int lnum,
              bool mlloc)
{
  if(!allocList)
  {
    allocList = new(AllocList);
  }
  if(!trackingOn) return;
  if(addr==NULL) return;

  chunk* tmp = (chunk*)chunkPool.getPtr();
  tmp->address=addr;
  tmp->size=asize;

  AllocList::iterator i;
  for(i = allocList->begin(); i != allocList->end(); i++)
    {
      ALLOC_INFO& info = *(*i);
      if(info.line == lnum)
        if(strncmp(fname, info.file, strlen(fname))==0)
          {
            info.chunks.insert(info.chunks.begin(), tmp);
            info.total += asize;
            if(info.total > info.peak) info.peak=info.total;
            grandTotal += asize;
            if(grandTotal > grandPeak) grandPeak = grandTotal;
            return;
          }
    }
  ALLOC_INFO* info;
  info = (ALLOC_INFO*)infoPool.getPtr();
  new (info) ALLOC_INFO;
  info->chunks.insert(info->chunks.begin(), tmp);
  info->total = asize;
  info->peak  = asize;
  strncpy(info->file, fname, 63);
  info->line = lnum;
  info->mlloc = mlloc;
  allocList->insert(allocList->begin(), info);

}

void RemoveTrack(void* addr, bool mlloc)
{
  if(!trackingOn) return;
  if(addr==NULL || !allocList) return;

  AllocList::iterator i;
  list<chunk*>::iterator c;
  for(i = allocList->begin(); i != allocList->end(); i++)
    {
      for(c=(*i)->chunks.begin(); c!= (*i)->chunks.end(); c++)
        {
          if((*c)->address == addr)
            {
              bool m = (*i)->mlloc;
              if(mlloc != m)
                {
                  if(!m)
                    MayDay::Error("memory allocated with operator NEW returned using free");
                  else
                    MayDay::Error("memory allocated with malloc-family returned using DELETE");
                }

              (*i)->chunks.remove(*c);
              (*i)->total -= (*c)->size;
              chunkPool.returnPtr(*c);
              // if((*i)->chunks.empty()){
              //    allocList->remove(*i);
              //    infoPool.returnPtr((*i));
              // }
              return;
            }
        }
    }   
  MayDay::Error("free/DELETE called on un-malloced/NEW'd pointer, or double freed");
}



void ReportUnfreedMemory(ostream& os)
{
  AllocList::iterator  i;
  PoolList::iterator   p;
  ArenaList::iterator  a;
  VectorList::iterator v;

  long int totalSize = 0;
  long int peak = 0;

  if(trackingOn && allocList){
    for(i = allocList->begin(); i != allocList->end(); i++)
    {
      peak += (*i)->peak;
      if( (*i)->total != 0)
      {
        totalSize += (*i)->total;
        //os<< "-----------------------------------------------------------\n";
        os << (*i)->file<<":"<<(*i)->line<<":  "<< (*i)->total<<  "\n";
      }
    }
    if(totalSize != 0)
    {
      os << "Total Unfreed from new,malloc, etc: "<<totalSize<<"  bytes\n";
    }
  } 

  if(vectorList_ != NULL)
  {
    //  os<<"---------------------------------\n";
    //  os<<"Vector-based memory allocation\n";
    for(v = vectorList_->begin(); v!=vectorList_->end(); v++)
      {
        long int* bytes = (v->second.first);

        if(*bytes != 0 && *bytes != 8)
        {
          os<<"Vector "<<v->first<<": "<<*bytes<<" bytes ("
            <<(*bytes)/BYTES_PER_MEG<<" Mb)\n";
          totalSize+=*bytes;
        }
        peak += *(v->second.second);
        // probably wondering about my silly "8" check.  workaround for
        // now since I can't clear chunkPool and infoPool unless this
        // is an atexit call, since they need to stay fresh in case this
        // is a transient run call to memusage.  will think up a better
        // way for this at another time.  Can't kill myself for 8 bytes
        //   bvs.
      }
  }

  //  os<<"---------------------------------\n";
  //  os<<"Pool-based memory allocation\n";
  for(p = Pool::poolList_->begin(); p!=Pool::poolList_->end(); p++)
    {
      peak+= (*p)->memUsage();

      if(*p == &chunkPool || *p == &infoPool)
      {
        // os<<(*p)->name_<<": "<<(*p)->memUsage()<<" bytes ("
        //   <<(*p)->memUsage()/BYTES_PER_MEG<<" Mb) (not added to total)\n";
      }
      else if ((*p)->memUsage() != 0)
      {
        os<<(*p)->name_<<": "<<(*p)->memUsage()<<" bytes ("
          <<(*p)->memUsage()/BYTES_PER_MEG<<" Mb)\n";
        totalSize+= (*p)->memUsage();
      }
    }

  if(Arena::arenaList_ != NULL)
  {
    // os<<"---------------------------------\n";
    // os<<"BaseFab-based memory allocation\n";
  for(a = Arena::arenaList_->begin(); a!=Arena::arenaList_->end(); a++)
    {
      peak+=(*a)->peak;
      if((*a)->bytes != 0)
      {
        os<<"BaseFab "<<(*a)->name_<<": "<<(*a)->bytes<<" bytes ("
          <<(*a)->bytes/BYTES_PER_MEG<<" Mb)\n";
        totalSize+= (*a)->bytes;
      }
    }
  }

  peak+= BitSet::peak + IntVectSet::peakcount*sizeof(IntVectSet);

  if(BitSet::bytes != 0)
  {
    //os<<"---------------------------------\n";
    os<<" BitSet allocation "<< BitSet::bytes<<" bytes ("
      <<BitSet::bytes/BYTES_PER_MEG<<" Mb)\n";
    totalSize += BitSet::bytes;
  }
  if(IntVectSet::count != 0)
  {
    os<<" IntVectSet allocation "<< IntVectSet::count*sizeof(IntVectSet)
      <<" bytes ("
      <<IntVectSet::count*sizeof(IntVectSet)/BYTES_PER_MEG<<" Mb)\n";
    totalSize += IntVectSet::count*sizeof(IntVectSet);
  }
  if(totalSize != 0)
  {
    os<<"---------------------------------\n";  
    os << "Total Unfreed : "<<totalSize<<" bytes ("
       <<totalSize/BYTES_PER_MEG<<" Mb)\n";
  }
  os<<"peak memory usage: "<<peak<<" bytes ("
    <<peak/BYTES_PER_MEG<<" Mb)\n";
  //XXX <dbs>  I dont like the extra blank line
  //XXXos<<std::endl;

}
void UnfreedMemory()
{
  ReportUnfreedMemory(pout());
}

void ReportMemoryUsage(ostream& os)
{
  ReportUnfreedMemory(os);
}

void MemoryUsage()
{
  ReportUnfreedMemory(pout());
}

void overallMemoryUsage(long int& currentTotal, long int& peak)
{
  currentTotal = 0;
  peak = 0;
  AllocList::iterator  i;
  PoolList::iterator   p;
  ArenaList::iterator  a;
  VectorList::iterator v;

  if(trackingOn && allocList)
  {
    for(i = allocList->begin(); i != allocList->end(); i++)
    {
      if( (*i)->total != 0)
      {
        currentTotal += (*i)->total;
        peak += (*i)->peak;
      }
    }
  } 

  if(vectorList_ != NULL)
  {
    for(v = vectorList_->begin(); v!=vectorList_->end(); v++)
      {
        long int* bytes = (v->second.first);
        currentTotal += *bytes;
        peak += *(v->second.second); // little trick might break some day.  bvs
      }
  }

  for(p = Pool::poolList_->begin(); p!=Pool::poolList_->end(); p++)
    {
      if(*p == &chunkPool || *p == &infoPool)
      {
        // don't report on the tracker itself.
      }
      else
      {
        currentTotal += (*p)->memUsage();
        peak+= (*p)->memUsage();
      }
    }

  if(Arena::arenaList_ != NULL){
  for(a = Arena::arenaList_->begin(); a!=Arena::arenaList_->end(); a++)
    {
      currentTotal += (*a)->bytes;
      peak += (*a)->peak;
    }
  }

  currentTotal += BitSet::bytes;
  peak += BitSet::peak;

  currentTotal += IntVectSet::count*sizeof(IntVectSet);
  peak += IntVectSet::peakcount*sizeof(IntVectSet);

}


//extra piece of Brian trickier to make sure that mallocp etc.
// do not become recursive and ridiculous.  now we need access
// to the real memory functions.

#undef malloc
#undef realloc
#undef calloc
#undef free


void* mallocp(size_t size, const char* file, int line)
{
  void *ptr = (void *)malloc(size);
  AddTrack(ptr, size, file, line, true);
  return(ptr);
}

void* reallocp(void* p, size_t size, const char* file, int line)
{
  RemoveTrack(p, true);
  void *ptr = (void *)realloc(p,size);
  AddTrack(ptr, size, file, line, true);
  return(ptr);
}

void *callocp(size_t nelem, size_t elsize, const char* file, int line)
{
  void *ptr = (void *)calloc(nelem, elsize);
  AddTrack(ptr, elsize*nelem, file, line, true);
  return(ptr);
}

void freep(void* p)
{
  RemoveTrack(p, true);
  free(p);
}

// void * operator new(size_t size, char const *file, int line)
// {
//   void *ptr = (void *)malloc(size);
//   AddTrack(ptr, size, file, line, false);
//   return(ptr);
// }

// void operator delete(void *p) throw()
// {
//   RemoveTrack(p, false);
//   free(p);
// }

// void * operator new[](size_t size, char const *file, int line)
// {
//   void *ptr = (void *)malloc(size);
//   AddTrack(ptr, size, file, line, false);
//   return(ptr);
// }

// void operator delete[](void *p) throw()
// {
//   RemoveTrack(p, false);
//   free(p);
// }



#ifdef MPI

void mpierrorfunction(MPI_Comm* comm, int* error, ...)
{
  char string[MPI_MAX_ERROR_STRING];
  int resultlen;
  MPI_Error_string(*error, string, &resultlen);
  cout<< string<<"\n";
  dumpmemoryatexit();
  // it would be nice to have a stack dump at this point also.
  MPI_Abort(*comm, *error);
}

void setChomboMPIErrorHandler()
{
  MPI_Errhandler handler;

  // Some MPI implementations are having trouble with this.
  // In particular, halem at GSFC.
  // For now, i'm just going to not do this for machines
  // running OSF (which currently, is only halem for us) (ndk)
#ifdef CH_OSF1
#else
  MPI_Errhandler_create(mpierrorfunction, &handler);
  MPI_Errhandler_set(Chombo_MPI::comm, handler);
#endif
}
#endif

#else    // on ENABLE_MEMORY_TRACKING
// make some stub functions

void dumpmemoryatexit(){}
void setChomboMPIErrorHandler(){}




#endif  // on ENABLE_MEMORY_TRACKING


/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.

#ifndef _BR_MESH_REFINE_H_
#define _BR_MESH_REFINE_H_
// Include files:



// Constants:

// Minimum acceptable ratio of tagged cells to total cells for 
// the Berger-Rigoutsos algorithm in \func{makeBoxes}.
// Used as default for \var{FillRatio} optional argument.
#ifndef _BR_MIN_BOX_FILL_RATIO_
#define _BR_MIN_BOX_FILL_RATIO_ ( 0.75 )
#endif


/// class which manages Berger-Rigoutsos grid generation
class BRMeshRefine : public MeshRefine
/**
   This class manages grid generation from sets of tagged cells using the 
   Berger-Rigoutsos algorithm in the context of the MeshRefine class
   from which it is derived
   
   There are two ways grids can be defined based on tagged cells.
   one takes a single IntVectSet of tags defined on the BaseLevel
   mesh and uses that set of tags for every level to be refined;
   the other takes a Vector<IntVectSet> of tags defined on all the
   mesh levels to be refined and uses those. \\ \\
   
   
   {\bf Long Description:} \\
   
   Create new meshes based on tagged cells on a range of levels of a mesh
   hierarchy.  Each level of tagged cells is used to generate a new mesh at
   the next finer level.  The finest level in the output mesh will be one
   level higher than the top of the range of levels given as input.  As a 
   special case, use the same tags (appropriately refined) for all levels.
   
   Usage:
   
   Call the regrid functions after computing error estimates and tagging cells.
   To add a new mesh level, set TopLevel to the index of the finest level
   in the existing mesh and define tags on the finest level.  To keep the
   existing number of mesh levels, set TopLevel to one less than the
   index of the finest level and don't define any tags on the finest level.
   If a single IntVectSet of tags is passed (instead of a
   Vector<IntVectSet>) then the same tags (properly refined) will be used
   for all the new meshes up to level TopLevel+1.  In any case, the
   meshes at levels BaseLevel and below are not modified.  The output
   argument newmeshes will be reallocated to the necessary size before
   being used.  When this function returns, the elements of the
   newmeshes vector corresponding to the unchanged levels will be
   filled in with copies of the levels from the old mesh vector.  The
   variable tags is modified in an undefined way, so its contents
   should not be relied upon.  The variable BlockFactor specifies the
   amount by which each box will be coarsenable. Every grid box will
   have an integral multiple of BlockFactor cells in each dimension and
   also lower index values that are integral multiples.  As a side effect,
   the minimum box size will be BlockFactor.
   
   Expensive validations are done only when debugging is enabled
   (i.e. the DEBUG make variable is "TRUE").
   
   
   Usage Notes:
   
   All the input vectors should be defined with max index >= TopLevel.
   They should have values for indices [BaseLevel:TopLevel].
   (except for OldMeshes, which must be defined for all indices).  The
   new mesh vector newmeshes will be redefined up to index
   TopLevel+1.  RefRatios should be defined such that
   RefRatios[L] is the value to use to refine the level L mesh to
   produce the level L+1 mesh.  The tags vector is modified in an
   undefined manner.  The output variable newmeshes may not be
   completely defined if an exception occurs.  
   The BlockFactor can be used to force a minimum box size.
*/
{
public:
  /// default constructor -- leaves object in an unusable state
  BRMeshRefine();
  
  /// full constructor -- leaves object in usable state
  BRMeshRefine(/// level 0 domain
               const Box& a_baseDomain,
               /// refinement ratios -- refRatio[0] is btwn levels 0 and 1
               const Vector<int>& a_refRatios,
               /// measure of how efficiently tagged cells will be covered
               const Real a_fillRatio,
               /// amount by which grids are guaranteed to be coarsenable
               const int a_blockFactor,
               /// proper nesting buffer amount
               const int a_bufferSize,
               /// maximum grid length in any direction -- 0 means no limit.
               const int a_maxSize);

  /// full constructor -- leaves object in usable state
  BRMeshRefine(/// level 0 domain
               const ProblemDomain& a_baseDomain,
               /// refinement ratios -- refRatio[0] is btwn levels 0 and 1
               const Vector<int>& a_refRatios,
               /// measure of how efficiently tagged cells will be covered
               const Real a_fillRatio,
               /// amount by which grids are guaranteed to be coarsenable
               const int a_blockFactor,
               /// proper nesting buffer amount
               const int a_bufferSize,
               /// maximum grid length in any direction -- 0 means no limit.
               const int a_maxSize);
  

  /// destructor
  virtual ~BRMeshRefine();
  
  /// define function -- size of RefRatios will define maximum number of levels
  void define(/// level 0 domain
              const Box& a_baseDomain,
              /// refinement ratios -- refRatio[0] is btwn levels 0 and 1
              const Vector<int>& a_refRatios,
              /// measure of how efficiently tagged cells will be covered
              const Real a_fillRatio,
              /// amount by which grids are guaranteed to be coarsenable
              const int a_blockFactor,
              /// proper nesting buffer amount
              const int a_bufferSize,
              /// maximum grid length in any direction -- 0 means no limit
              const int a_maxSize);


  /// define function -- size of RefRatios will define maximum number of levels
  void define(/// level 0 domain
              const ProblemDomain& a_baseDomain,
              /// refinement ratios -- refRatio[0] is btwn levels 0 and 1
              const Vector<int>& a_refRatios,
              /// measure of how efficiently tagged cells will be covered
              const Real a_fillRatio,
              /// amount by which grids are guaranteed to be coarsenable
              const int a_blockFactor,
              /// proper nesting buffer amount
              const int a_bufferSize,
              /// maximum grid length in any direction -- 0 means no limit
              const int a_maxSize);
  

  /// constructs a set of boxes which covers a set of tagged cells 
  /** constructs a set of boxes which covers a set of tagged cells
      by using the Berger-Rigoutsos algorithm.  Everything should
      be on the same level, and blocking factor is not applied.
      Boxes will be on the same refinement level as the tags.
      This would normally be a protected function, but it can be useful
      to call it on it's own, so it has been left public. 
  */
  void 
  makeBoxes(/// output: refined boxes at each new level
            Vector<Box>&      a_mesh,        
            /// input: set of tagged cells to cover
            const IntVectSet& a_tags,        
            /// input: proper nesting domain in which mesh boxes must live
            const IntVectSet& a_pnd,         
            /// input: physical domain
            const ProblemDomain& a_domain,
            ///input: largest number of cells in any dimension for any box
            const int         a_maxSize) const;
  

protected:
  void 
  makeBoxes(/// output: refined boxes at each new level
            std::list<Box>&      a_mesh,        
            /// input: set of tagged cells to cover
            const IntVectSet& a_tags,        
            /// input: proper nesting domain in which mesh boxes must live
            const IntVectSet& a_pnd,         
            /// input: physical domain
            const ProblemDomain& a_domain,
            ///input: largest number of cells in any dimension for any box
            const int         a_maxSize,
			///input: depth of this recursion in the algorithm
			const int         a_depth
			) const;
  ///
  void
  splitBox(std::list<Box> & a_boxes ,
		   const std::list<Box>::iterator& a_boxindex,
           const int a_dimension ,const int a_maxboxsize ) const;
  
  ///
   Vector<int>
  makeTrace( const IntVectSet& a_Ivs ,int a_dir ) const;
  
  ///
  void
  makeTraces( const IntVectSet& a_Ivs ,Vector<int>* a_traces ) const;
  ///
   int
  findSplit( const Vector<int>& a_trace ) const;
  
  ///
   int
  findMaxInflectionPoint( const Vector<int>& a_trace ,int& a_maxVal ) const;
  
  ///
   void
  splitTags( const IntVectSet& a_tags,
             const int a_split_dir ,const int a_split_indx,
             IntVectSet& a_tags_lo ,IntVectSet& a_tags_hi ) const;
  
  ///
  void 
  breakBoxes(Vector<Box>& a_vboxin,  const int& a_maxSize, 
             const int& a_idir) const;
  
  ///
   int
  maxloc( const int* a_V ,const int a_Size ) const;


};

///
/**
   Splits domain into a vector of disjoint boxes with
   maximum size maxsize.
   blocking factor is default to one.
   If you make minimum size > 1, then domain must
   be coarsenable and refineable by blockfactor
   (refine(coarsen(domain,blockfactor), minsize) == domain)
   or an error is thrown.  This would be defined in 
   MeshRefine.H, except that it needs to use a BRMeshRefine object.
*/
extern void domainSplit(const ProblemDomain& a_domain, Vector<Box>& a_vbox, 
                        int a_maxSize, int a_blockfactor=1);

///
/**
   Splits domain into a vector of disjoint boxes with
   maximum size maxsize.
   blocking factor is default to one.
   If you make minimum size > 1, then domain must
   be coarsenable and refineable by blockfactor
   (refine(coarsen(domain,blockfactor), minsize) == domain)
   or an error is thrown. This would be defined in 
   MeshRefine.H, except that it needs to use a BRMeshRefine object.
*/
extern void domainSplit(const Box& a_domain, Vector<Box>& a_vbox, 
                        int a_maxSize, int a_blockfactor=1);

#endif
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//


// Constants for Berger-Rigoutsos algorithm

#if ! defined(_BR_MIN_INFLECTION_MAG_)
#define _BR_MIN_INFLECTION_MAG_ ( 3 )
#endif


// Berger-Rigoutsos Mesh refinement class
// ---------
//  class BRMeshRefine
//
///////////////////////////////////////////////////////////////////////////////


// Include files:


BRMeshRefine::BRMeshRefine() 
{
}

BRMeshRefine::~BRMeshRefine()
{
}

BRMeshRefine::BRMeshRefine(const Box& a_baseDomain,
                           const Vector<int>& a_refRatios,
                           const Real a_fillRatio,
                           const int a_blockFactor,
                           const int a_bufferSize,
                           const int a_maxSize)
{
  ProblemDomain crseDom(a_baseDomain);
  define(crseDom, a_refRatios, a_fillRatio, a_blockFactor,
         a_bufferSize, a_maxSize);
}


BRMeshRefine::BRMeshRefine(const ProblemDomain& a_baseDomain,
                           const Vector<int>& a_refRatios,
                           const Real a_fillRatio,
                           const int a_blockFactor,
                           const int a_bufferSize,
                           const int a_maxSize)
{
  define(a_baseDomain, a_refRatios, a_fillRatio, a_blockFactor,
         a_bufferSize, a_maxSize);
}

void
BRMeshRefine::define(const Box& a_baseDomain,
                     const Vector<int>& a_refRatios,
                     const Real a_fillRatio,
                     const int a_blockFactor,
                     const int a_bufferSize,
                     const int a_maxSize)
{
  ProblemDomain crseDom(a_baseDomain);
  define(crseDom, a_refRatios, a_fillRatio, a_blockFactor,
         a_bufferSize, a_maxSize);
}



void
BRMeshRefine::define(const ProblemDomain& a_baseDomain,
                     const Vector<int>& a_refRatios,
                     const Real a_fillRatio,
                     const int a_blockFactor,
                     const int a_bufferSize,
                     const int a_maxSize)
{
  // nothing new here
  MeshRefine::define(a_baseDomain, a_refRatios, a_fillRatio,
                     a_blockFactor, a_bufferSize, a_maxSize);

}


///////////////////////////////////////////////////////////////////////////////
//
// Function: makeBoxes
//
// Short Description:
//  Construct a set of boxes that cover a set of tagged cells.
//
// Purpose:
//  Given a set of tagged cells defined on a single level of an AMR grid,
//  construct a BoxArray that covers all these cells that minimizes the
//  number of boxes and maximizes the ratio of tagged cells in each box.
//  This is part of the process of refining a mesh hierarchy based on error
//  estimates.
// 
// Caveat:
//   The mesh that is created must lie within the proper nesting domain (Pnd)
//
// Usage Notes:
//
// Numerical Algorithm:
//  <Berger-Rigoutsos algorithm>
//
// Implementation Method:
//
// Implementation Notes:
//  This gets tricky when the minbox around a set of tags has a non-zero offset.
//  Have to be careful to remember this when looking at box sizes.
//
// References:
//  M. Berger, I. Rigoutsos, "An Algorithm for Point Clustering and Grid Generation", 
//  _IEEE_Transactions_on_Systems,_Man,_and_Cybernetics_, Vol. 21, No.5, pp. 1278-1286, 
//  Sept./Oct. 1991.
//
///////////////////////////////////////////////////////////////////////////////

void
BRMeshRefine::makeBoxes(Vector<Box>&         a_mesh,
                        const IntVectSet&    a_tags,
                        const IntVectSet&    a_pnd,
                        const ProblemDomain& a_domain,
                        const int            a_maxSize
                        ) const
{
  std::list<Box> boxes;
  makeBoxes(boxes, a_tags, a_pnd, a_domain, a_maxSize, 0);
  //boxes.sort();
  a_mesh.resize(boxes.size());
  std::list<Box>::iterator it = boxes.begin();
  for(int i=0; i<a_mesh.size(); ++i, ++it) a_mesh[i]=*it;
}


void
BRMeshRefine::makeBoxes(std::list<Box>&      a_mesh,
                        const IntVectSet&    a_tags,
                        const IntVectSet&    a_pnd,
                        const ProblemDomain& a_domain,
                        const int            a_maxSize,
						const int            a_depth
                        ) const
{

  assert (isDefined());

  Box minbx ;                                  //min box around Tags
  std::list<Box> mesh_hi ;                     //boxes from recursion
  IntVectSet tags_lo, tags_hi ;                //tags for recursion
  
  //
  // Handle special cases of no tags
  //
  if( a_tags.isEmpty() ) {    //return null box
    a_mesh.clear() ;
    return;
  }
  
  //
  // Validate inputs
  //
  
  // The number of tagged cells in a box cannot exceed the number
  // of cells in the box so enforce an upper bound on \var{FillRatio}.
  //[NOTE: 0 or negative values are allowed -- they mean any box
  //       is acceptable.  This will probably be a mistake by the
  //       caller, but there is no obvious lower valid value for
  //       this variable (1e-6 is just as likely a mistake as 0).]
  assert ( m_fillRatio <= 1.0 );
  
  //
  // Handle the case of all tags fitting in one box.
  // This always happens at the bottom of the recursion.
  //
  minbx = a_tags.minBox() ;
  // If minbox has enough tagged cells and is properly nested, want 
  // to add this minbox to the mesh.
  // If not, continue with splitting it and recursing on the pieces.
  if( (a_tags.numPts() >= minbx.numPts() * m_fillRatio) 
      && ( a_pnd.contains(minbx))) {
    // no IntVects in the box are outside the PND so this box can
    // be accepted.  If it is larger than the maximum box size, split
    // it into two boxes
    //[NOTE: one of the boxes may violate the FillRatio requirement, but
    //       we will ignore this.]
//	pout()<< "split Box "<<minbx<<"  maxsize "<<a_maxSize<<std::endl;
    a_mesh.clear(); 
    a_mesh.push_front(minbx) ;
    if (a_maxSize > 0)
      {
	
        for( int i = 0 ; i < SpaceDim ; i++ ) {
          // divide the box along this dimension until the box is small enough,
          // replacing the original box with one half and appending the second half
          // to the end of the vector and recursing on both halves

		  std::list<Box>::iterator it;
		  for(it=a_mesh.begin(); it!=  a_mesh.end(); ++it){
			splitBox( a_mesh,  it, i, a_maxSize ) ;
		  }
        }
      } // end if we are enforcing a max box size
    

  } else {
    // if efficiency criterion not met or box not properly nested...
    //
    // Compute the traces (signatures) of the minbox in each direction, and
    // find a hole in the trace (zero value) and an inflection point (zero
    // Laplacian) in each direction; keep the best of each.
    //[NOTE: both \func{find*} functions return -1 if nothing was found.]
    //[NOTE: \var{infl_val} is modified by \func{findMaxInflectionPoint}.]
    //[NOTE: \var{trace} indexes from 0, so indices into \var{trace} must 
    //       be shifted to be used as indices into the \var{a_tags}.]
    //
    int hole_indx[SpaceDim], best_hole_dim,       //holes in traces
      infl_indx[SpaceDim], best_infl_dim,       //inflection points in traces
      infl_val [SpaceDim] ;                    //magnitudes of infl.points
	Vector<int> traces[SpaceDim];
	makeTraces( a_tags, traces) ;
    for( int idim = 0 ; idim < SpaceDim ; idim++ ) {
	  // Vector<int> trace = makeTrace( a_tags, idim ) ;
      hole_indx[idim] = findSplit( traces[idim] ) ;
      infl_indx[idim] = findMaxInflectionPoint(traces[idim], infl_val[idim] ) ;
    }
    // Take the highest index as the best one because we want to take as large
    // a box as possible  (fewer large boxes are better than many small ones)
    best_hole_dim = maxloc( hole_indx, SpaceDim ) ;
    best_infl_dim = maxloc( infl_indx, SpaceDim ) ;
    
    //
    // Split the Tag set at a hole in one of the traces, if there is one, or an
    // inflection point in the Laplacian of the traces.  Failing that, split
    // at the middle of the longest dimension of the enclosing box.
    //
    if( hole_indx[best_hole_dim] >= 0 ) {
      // split at a hole in the trace, adjusting the trace index for the
      // offset into \var{a_tags} indices
      splitTags( a_tags, best_hole_dim,
                 hole_indx[best_hole_dim] + minbx.smallEnd(best_hole_dim ),
                 tags_lo, tags_hi ) ;
      
    } else if( infl_indx[best_infl_dim] >= 0 ) {
      // split at an inflection point in the trace, adjusting the trace
      // index for the offset into \var{a_tags} indices
      splitTags(a_tags, best_infl_dim,
                infl_indx[best_infl_dim] + minbx.smallEnd( best_infl_dim ), 
                tags_lo, tags_hi ) ;
      
    } else {
      // split on the midpoint of the longest side of \var{minbx}, rounding up,
      // allowing for \var{minbx} to have a non-zero offset
      int longdim ;  //[NOTE: longdim is set by \func{longside}]
      minbx.longside( longdim ) ;  //ignore the return value
      splitTags(a_tags, longdim,
                (minbx.smallEnd(longdim) + minbx.bigEnd(longdim) +1)/2,
                tags_lo, tags_hi ) ;
    }
    
    // Recurse on the two halves of the Tags 
    if(!tags_lo.isEmpty())
      makeBoxes( a_mesh, tags_lo, a_pnd, a_domain, a_maxSize, a_depth+1) ;
    if(! tags_hi.isEmpty() )
      makeBoxes( mesh_hi, tags_hi, a_pnd, a_domain, a_maxSize, a_depth+1) ;
    
    // combine the results into a single mesh
	a_mesh.splice(a_mesh.begin(), mesh_hi);
  } // done if we need to split the box

  // Done
} //end of makeBoxes


///////////////////////////////////////////////////////////////////////////////
//
// Function: splitBox()
//
// Short description:
//  If the box in the given Vector pointed to by the given index is larger
//  than the given maximum size in the given dimension, split it into two
//  boxes and recurse on each.
//
///////////////////////////////////////////////////////////////////////////////

void
BRMeshRefine::splitBox( std::list<Box>&                 a_boxes,
                        const std::list<Box>::iterator& a_box,
                        const int a_dim,
                        const int a_maxSize) const
{
  // See if the box needs to be split in this dimension.  If not, do nothing.
  Box& b = *a_box;
  if( b.size( a_dim ) > a_maxSize ) 
    {
     // pout() <<"splitting box "<<a_boxes[a_boxIndex]<<std::endl;
      // chop() returns the upper half as a new box (which is appended to
      // the vector) and modifies the original box to be the lower half
      int len = 
        ( b.smallEnd( a_dim )+ 
          b.bigEnd(   a_dim )+1 )/2;
      
	  std::list<Box>::iterator c = a_boxes.insert(a_box, b.chop(a_dim, len));
 
      // recurse on the two boxes
 
      splitBox( a_boxes, a_box , a_dim, a_maxSize ) ;
      splitBox( a_boxes, c     , a_dim, a_maxSize ) ;
    }
  
  // Done
}

///////////////////////////////////////////////////////////////////////////////
//
// Function: makeTrace()
//
// Short description:
//  Count the number of cells in the IVS at every index in the specified
//  coordinate direction.  In other words, make a histogram using the
//  \var{Dir}'th coordinate as data.
//
// Note: Vectors index from 0 so the trace indices have to be shifted from
//       the box indices by the value of the lower bound.
//
// References:
//  The trace computed here is the \Sigma value computed in Berger-Rigoutsos.
//
///////////////////////////////////////////////////////////////////////////////

Vector<int>
BRMeshRefine::makeTrace( const IntVectSet& a_Ivs, int a_dir ) const
{
  //
  // Validate inputs
  //
  assert( a_dir >= 0 & a_dir < SpaceDim ) ;

  //
  // Histogram the coordinates of \var{a_Ivs} in direction \var{a_dir} into
  // \var{trace}.  The index range of \var{trace} must start at 0
  // regardless of the bounds of the minbox of \var{a_Ivs}
  //
  Box minbox = a_Ivs.minBox() ;
  int offset = minbox.smallEnd(a_dir) ;
  Vector<int> trace( minbox.size(a_dir), 0 ) ;  //initialize to 0
  IVSIterator i(a_Ivs) ;
  
  for( i.begin() ; i.ok() ; ++i ) {
    trace[i()[a_dir]-offset] += 1 ;
  }
  
  // Done
  return( trace ) ;
}

void BRMeshRefine::makeTraces ( const IntVectSet& a_Ivs, Vector<int>* traces) const
{
  Box minbox = a_Ivs.minBox() ;
  IntVect offset = minbox.smallEnd() ;
  const IntVect& size = minbox.size();
  D_TERM(traces[0].resize(size[0],0);,traces[1].resize(size[1],0);,traces[2].resize(size[2],0););
  IVSIterator i(a_Ivs) ;
  for( i.begin() ; i.ok() ; ++i ) {
	IntVect iv = i();
	iv-=offset;
	D_TERM(traces[0][iv[0]]++,; traces[1][iv[1]]++,;traces[2][iv[2]]++);
  }

}
///////////////////////////////////////////////////////////////////////////////
//
// Function: findSplit()
//
// Short Description:
//  Given a trace (ie. signature) of a Tag set, find a place in the trace
//  with a zero count.  Assumes the first and last elements of the trace
//  are non-zero.
//
// Usage Note:
//  If there are no acceptable zeros in the trace, this returns -1.
//  \var{Vector}s index from 0, so the result may have to be shifted to be 
//  used as an index into the \var{IntVectSect} which produced the trace.
//
///////////////////////////////////////////////////////////////////////////////

 int
BRMeshRefine::findSplit( const Vector<int>& a_trace ) const
{
  // look for a place to split at -- begin at the index after the first nonzero
  for( int i = 1 ; i < a_trace.size() -1 ; i++ ) {
    if( a_trace[i] == 0 ) return( i ) ;
  }
  
  // nothing acceptable
  return( -1 ) ;
}


///////////////////////////////////////////////////////////////////////////////
//
// Function: findMaxInflectionPoint
//
// Short Description:
//  Find the largest inflection point in the given trace (ie. signature)
//  and return the index and set the value of the inflection in the last
//  argument.  The index will be the edge-centered location where a 
//  chop should be made in the BR algorithm.
//
//  If there are no inflection points or if none of them are larger than the
//  cutoff tolerance, -1 is returned.
//
// Implementation Note:
//  The smallest acceptable inflection magnitude is the global constant
//  _MIN_INFLECTION_MAG_
//
///////////////////////////////////////////////////////////////////////////////

 int
BRMeshRefine::findMaxInflectionPoint(const Vector<int>& a_trace, 
                                     int& a_maxVal) const 
{
  // first compute the discrete Laplacian of the trace
  Vector<int> d2Trace( a_trace.size(), 0 ) ;
  for( int i = 1 ; i < a_trace.size()-1 ; i++ ) {
    d2Trace[i] = a_trace[i-1] - 2*a_trace[i] + a_trace[i+1] ;
  }
  
  // find inflection points and save one with the largest magnitude
  int absval, imax ;
  a_maxVal = -1 ;
  for( int i = 2 ; i < a_trace.size()-1 ; i++ ) {
    absval = abs( d2Trace[i-1]  - d2Trace[i] ) ;
    if( d2Trace[i-1] * d2Trace[i] < 0  && absval > a_maxVal ) {
      imax = i ; a_maxVal = absval ;
    }
  }
  
  // Find the largest inflection point, if one exists and
  // has magnitude large enough and return its index.
  // return edge-centered location of chop point.
  if( a_maxVal == -1 )                          return( -1 ) ;
  else if( a_maxVal < _BR_MIN_INFLECTION_MAG_ ) return( -1 ) ;
  else                                        return( imax);
}

///////////////////////////////////////////////////////////////////////////////
//
// Function: splitTags
//
// Purpose:
//  Given a direction and an index in that direction, split the IntVectSet of
//  tags into two sets (lo,hi) on either side of the index.  Tags that match
//  the index exactly go into the hi set.
//
// Implementation Method:
//  For each tag in the set, look at the coordinate in the direction specified,
//  and if it is <= the specified index, copy the tag into the lo set.  Else
//  copy it into the hi set.
//
// Usage Note:
//  It is acceptable for the split index to be outside of the box containing Tags.
//  This forces all the Tags to end up in one of the two output sets.
//
///////////////////////////////////////////////////////////////////////////////

void
BRMeshRefine::splitTags(const IntVectSet& a_tags,
                        const int a_split_dir, const int a_split_indx,
                        IntVectSet& a_tags_lo, IntVectSet& a_tags_hi ) const
{
  // Validate inputs
  //[NOTE: it is ok if the a_split_indx is outside the box containing a_tags.]
  assert( a_split_dir >= 0 && a_split_dir < SpaceDim );
  assert( !a_tags.isEmpty() );
  
  // Copy the whole set into a_tags_lo and extract the cells that aren't
  // below the split index into a_tags_hi, keeping the leftover in a_tags_lo
  // (i.e., the a_split_indx'th cells go into the high set).
  a_tags_lo = a_tags ;
  a_tags_hi = a_tags_lo.chop( a_split_dir, a_split_indx );
  
}

//recursive function to enforce max size of boxes in a given direction
void 
BRMeshRefine::breakBoxes(Vector<Box>& a_vboxin,  
                         const int& a_maxSize, const int& a_idir) const
{
  int nboxes = a_vboxin.size();
  //need to use STL vector for bools.
  using std::vector;
  vector<bool> splitDec(nboxes);
  bool anyBoxesToSplit = false;
  //find out which boxes need to be chopped and in what direction
  for(int ibox = 0; ibox < nboxes; ibox++)
    {
      if( a_vboxin[ibox].size(a_idir ) > a_maxSize ) 
        {
          splitDec[ibox] = true;
          anyBoxesToSplit = true;
        }
      else
        {
          splitDec[ibox] = false;
        }
    }
  //if there are no boxes to split, just return
  //otherwise, split all the boxes that need to be 
  //split ONCE and then call function recursively
  // and set the return vector to the temporary after
  // the recursion
  if(anyBoxesToSplit)
    {
      Vector<Box> vboxtemp;
      for(int ibox = 0; ibox < nboxes; ibox++)
        {
          Box boxone = a_vboxin[ibox];
          if(splitDec[ibox])
            {
              int len = (boxone.smallEnd(a_idir) + 
                         boxone.bigEnd(a_idir))/2;
              Box boxtwo = boxone.chop(a_idir, len);
              vboxtemp.push_back(boxone);
              vboxtemp.push_back(boxtwo);
            }
          else
            {
              vboxtemp.push_back(boxone);
            }
        } // end loop over boxes
      breakBoxes(vboxtemp, a_maxSize, a_idir);
      a_vboxin = vboxtemp;
    }
  return;
}



///////////////////////////////////////////////////////////////////////////////
//
// Function: maxloc
//
// Purpose: find the maximum value in a Vector<int> or an int[] and return its index.
//
// Usage Notes:
//  Returns -1 if the Vector has no entries.
//
///////////////////////////////////////////////////////////////////////////////


int
BRMeshRefine::maxloc( const int* a_V, const int a_Size ) const
{
  int imax = 0 ;
  for( int i=1 ; i<a_Size ; i++ ) if( a_V[i] > a_V[imax] ) imax = i ;
  return( imax ) ;
}


// at the moment, this is a pretty hokey approach
// (first build a BRMeshRefine, then "generate" grids
// hopefully will have a better version soon...
void
domainSplit(const ProblemDomain& a_domain, Vector<Box>& a_vbox, 
            int a_maxSize, int a_blockFactor)
{
  const Box& domBox = a_domain.domainBox();
  domainSplit(domBox, a_vbox, a_maxSize, a_blockFactor);
}




void
domainSplit(const Box& a_domain, Vector<Box>& a_vbox, int a_maxSize, int a_blockFactor)
{
  assert(!a_domain.isEmpty());
  assert(a_maxSize > 0);
  int iref = 1;
  Box coardom = coarsen(a_domain, iref);
  Vector<Box> domains(2);
  domains[0] = coardom;
  domains[1] = a_domain;
  Vector<Box> oldGrid0(1, coardom);
  Vector<Vector<Box> > oldMeshes(1);
  oldMeshes[0] = oldGrid0;
  if(a_maxSize/a_blockFactor/iref < 1)
    MayDay::Error("DomainSplit: maxsize and blocking factor incompatible");
  Vector<Vector<Box> > newMeshes;
  Vector<int> refratio(1,iref);
  IntVectSet tags(coardom);
  // make irreg_boxes using berger-rigoutsos
  Real fillrat = 0.7;
  int buffersize = 1;
  int block_factor= a_blockFactor;
  Box testdom = coarsen(a_domain, iref*block_factor);
  testdom.refine(iref*block_factor);
  if(a_domain != testdom)
    MayDay::Error("DomainSplit:domain and Blocking Factor incompatible");
  BRMeshRefine mrobject(domains[0], refratio, fillrat, 
                        block_factor, buffersize, a_maxSize);

  mrobject.regrid(newMeshes, tags, 0, 0, 
                                  oldMeshes);


  assert(newMeshes.size() == 2);
  a_vbox = newMeshes[1];
}




/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

// Purpose:
//  LoadBalance() is a global function to compute an assignment
//  of boxes to processors for an AMR mesh hierarchy.  The assignment
//  is made so as to balance the computation and communication
//  workload on each processor (i.e. make it as even as possible).
//
//  ********************************************************
//  CAVEAT: this version ignores the communication workload.
//          It balances ONLY the compute workload.
//  ********************************************************
//
// Usage:
//  The meshes in the AMR hierarchy are represented using
//   a Vector of Vectors of Boxes.
//
//  The computational workload is a real number for each box in the
//   hierarchy, also represented as a Vector of Vectors.
//
//  The communication workload is a real number for each connection
//   between pairs of boxes, which specifies the cost of communication
//   between the two boxes if they are on different processors.  No
//   allowance is made for varying the communication cost depending on the
//   distance between processors (ie. all processors are equally far apart)
//   and it assumes that the effective cost of communicating between two
//   boxes on the same processor is zero.  The communication workload is
//   represented as a graph.
//
//  ********************************************************
//  CAVEAT: the communication workload argument is not
//           present in this version of the function.
//  ********************************************************
//
//  The resulting assignment is an integer for each box in the hierarchy,
//   which gives the processor number (starting at zero) on which each box
//   should reside.  Again, represented as a Vector of Vectors.
//
//  The other output argument is the efficiency ratio, which is a measure
//   of how close to perfect the final load balance is.  The \var{effRatio}
//   output variable is defined as the smallest load on any processor divided
//   by the largest load.  A perfect load balance has efficiency == 1.
//
//  ********************************************************
//  CAVEAT: in this version, each level in the AMR hierarchy
//          is balanced independently, so the efficiency ratio
//          that is returned is the worst over all levels.
//  ********************************************************
//
//  It is important to note that it is the sum of the computation cost 
//   and communication cost that is balanced, so the values in the two
//   variables must be in the same units.  It doesn't matter what the
//   units are.
//
//  ********************************************************
//  CAVEAT: the communication workload argument is not
//           present in this version of the function.
//  ********************************************************
// 
// Interface Specification:
//  int LoadBalance() arguments are:
//       Vector<Vector<int>>& procAssignments  //output: processor number 
//                                             //   for each box
//       Real&                effRatio         //output: efficiency ratio
// const Vector<Vector<Box>>& Grids            //input: meshes to balance
// const Vector<Vector<long>>&ComputeLoads     //input: computational cost
//                                             //   of each box
//###not used###
//### const [...something...]&CommunicateLoads //input: communication cost
//                                             //   between each pair of 
//                                             //   neighboring boxes
// const Vector<int>&         RefRatios        //input: refinement ratio
//                                             //   for each level
//
// Returns: integer status_code
//    =0  means succesful completion
//  exceptions: (output variables are not defined)
//  -1011 input vectors (\var{Grids} and \var{ComputeLoads}) are not
//        the same size
//  -1012 one of the vector elements of the input vectors (\var{Grids} 
//        and \var{ComputeLoads}) are not the same size
//  -1013 input vectors (\var{Grids} and \var{RefRatios}) are not
//        the same size
//
// Modification History
//  19Nov99 <dbs> initial design and coding
//


#ifndef _LOADBALANCE_H_
#define _LOADBALANCE_H_


int
LoadBalance( Vector<Vector<int> >& procAssignments  //output: processor number  
//                                                  //   for each box           
            ,Real&                 effRatio         //output: ratio of min load
//                                                  //   to max load
     ,const Vector<Vector<Box> >&  Grids            //input: meshes to balance  
     ,const Vector<Vector<long> >& ComputeLoads     //input: computational cost 
//                                                  //   of each box            
//## ,const [...something...]&     CommunicateLoads //input: communication cost 
//                                                  //   between each pair of   
//                                                  //   neighboring boxes      
     ,const Vector<int>&           RefRatios        //input: refinement ratio   
//                                                  //   for each level        
            ,int nProc = numProc()
           ) ;


int
LoadBalance( Vector<BoxLayout>&    Grids            //in-out: input grids to balance
//                                                  //   and output proc. numbers
            ,Real&                 effRatio         //output: ratio of min load
//                                                  //   to max load
     ,const Vector<Vector<long> >& ComputeLoads     //input: computational cost 
//                                                  //   of each box            
//## ,const [...something...]&     CommunicateLoads //input: communication cost 
//                                                  //   between each pair of   
//                                                  //   neighboring boxes      
     ,const Vector<int>&           RefRatios        //input: refinement ratio   
//                                                  //   for each level         
            ,int nProc = numProc()
           ) ;
///
/** convenience function to load balance a Vector of Boxes based 
 on load=box.numPts() */
int LoadBalance(Vector<int>& procAssignments, const Vector<Box>& boxes);

#endif
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

// Purpose:
//  This file contains a global function to compute an assignment
//  of boxes to processors for an AMR hierarchy.  The assignment
//  is made so as to balance the computation and communication
//  workload on each processor (i.e. make it as even as possible).
//
//  ********************************************************
//  CAVEAT: this version ignores the communication workload.
//          It balances ONLY the compute workload.
//  ********************************************************
//
// Usage:
//  The boxes in the AMR hierarchy are stored as a Vector of
//  Vectors of Boxes.  The first (outer) vector indexes over
//  the levels and the second (inner) vector indexes over the boxes
//  on each level.
//  The computational workload is represented as a real number for
//  each box in the hierarchy, stored as a Vector of Vectors of Reals
//  in the same across-level/in-level ordering as the boxes.
//  The communication workload is represented as a real number for
//  each connection between pairs of boxes, which specifies the cost of 
//  communication between the two boxes if they are on different 
//  processors.  No allowance is made for varying the communication 
//  cost depending on the distance between processors (ie. all processors
//  are equally far apart) and it assumes that the effective cost of
//  communicating between two boxes on the same processor is zero.
//  The resulting assignment is represented as an integer for 
//  each box in the hierarchy, which gives the processor number (starting
//  at zero) on which each box should reside.  Represented as a 
//  Vector of Vectors of ints in the same ordering as the boxes and
//  loads.
//  The other output argument is a measure of how close to perfect the
//  final load balance is.  The \var{effRatio} output is defined
//  as the smallest load divided by the largest load.  A perfect load
//  balance has efficiency == 1.
//  It is important to note that it is the sum of the computation cost 
//  and communication cost that is balanced, so the values in the two
//  variables must be in the same units.  It doesn't matter what the
//  units are.  
//
// Interface Specifications:
//  There are several interfaces to the LoadBalance function that allow
//  for different ways to represent the boxes and loads to be balanced.
//
//  This version is the full-functionality version: it takes a hierarchy
//  of boxes (represented as a Vector<Vector<Box>> and returns processor
//  assignments for all levels of the hierarchy.
//  
//  int LoadBalance(
//    Vector<Vector<int>>  procAssignments  //output: processor number for each
//                                          //   box indexed by [level][box].
//    Real                 effRatio         //output: efficiency ratio
//    Vector<Vector<Box>>  Grids            //input: meshes to balance, indexed
//                                          //   by [level][box]
//    Vector<Vector<long>> ComputeLoads     //input: computational cost of each
//                                          //   box, indexed by [level][box]
//    [...something...]    CommunicateLoads //input: communication cost
//                                          //   between each pair of 
//                                          //   neighboring boxes
//    Vector<int>          RefRatios        //input: refinement ratio
//    )                                     //   for each level
//
//  This version takes the boxes from an existing BoxLayout and modifies
//  it with the processor assignments.
//
//  int LoadBalance(
//    Vector<BoxLayout>    Grids            //in-out: input boxes to balance,
//                                          //   output processor numbers
//    Real                 effRatio         //same as above
//    Vector<Vector<long>> ComputeLoads     //   "
//    [...something...]    CommunicateLoads //   "
//    Vector<int>          RefRatios        //   "
//    )
//
//  This version balances only a single level grid and implicitly assumes
//  the load is proportional to the number of points in each box.
//
//  int LoadBalance(
//    Vector<int>          procAssignments  //output: processor number for each
//                                              box in a single level
//    Vector<Box>          boxes            //input: boxes to balance
//    )
//
//  Return value: integer status_code (see "Exceptions:" below)
//    =0  means succesful completion
//    <0  means a fatal exception occurred; the output is undefined
//    >0  means a non-fatal exception occurredl but the output is defined
//
//
// Method:
//  This subroutine uses the Kernighan-Lin algorithm for
//  solving knapsack problems.  It operates one level at a time.
//  The algorithm is:
//   0. loop over levels
//   1.   sort loads by decreasing value
//   2.   loop over loads
//   3.     assign load to processor with lowest total load
//   4.   loop until done
//   5.     select the processor with the worst load balance
//   6.     loop over loads il in this processor ip
//   7.       loop over loads jl in other processors jp
//   8.         if (swapping il and jl improves load balance more 
//                       than any previous swap) then
//   9.             save jp, il,jl and the improvement_in_balance
//  10.     if any improvement_in_balance, then swap ip,il with jp,jl
//  11.   repeat (4) until no swap improves the load balance
//
//   Complexity is: O( NP * NL/NP * NL * nr ) + O( NP^2)
//     where: NP=#processors, NL=#loads(ie. boxes),
//            nr is the # of repetitions of step 4 (may be O(NP))
//            and the O(NP^2) term is due to Step 11.
//
// Implementation Notes:
//  There are several places in this implementation where the
//  cost could be reduced significantly with more complex coding.
//  In particular, most of the searches are done by brute force,
//  resulting in O(N) or O(N^2) cost where a better implementation
//  using sorted lists woudl cost only O(log N) or O(N log N).
//  This isn't a big deal for a prototype, but for a problem with
//  1000s of boxes this will be significant.
//
// Exceptions:
//  -1011 input vectors (\var{Grids} and \var{ComputeLoads}) are not
//        the same size
//  -1012 one of the vector elements of the input vectors (\var{Grids} 
//        and \var{ComputeLoads}) are not the same size
//  -1013 input vectors (\var{Grids} and \var{RefRatios}) are not
//        the same size
//
// References:
// B. Kernighan and S. Lin, "An effective heuristic procedure for partitioning
// graphs", The Bell System Technial Journal, pp. 291--308, Feb 1970
//
// Modification History
//  19Nov99 <dbs> initial design and coding


using std::cout;



// local prototypes
int
min_element( const Vector<long>& Vect ) ;
void
min_max_elements( int& imin ,int& imax ,const Vector<long>& Vect ) ;

void
min_max_elements( int& imin ,int& imax ,const Vector<long long>& Vect ) ;

// Local class definition (needed to sort loads)
class Load
{
  public:
    Load():load(0) ,grid_index(0){}
    bool operator < (const Load& rhs) const {return load < rhs.load;}

    long load ;      //actual load on this box
    int grid_index ; //link to Grids[]
};


// Code:

///
// This version takes a Vector<BoxLayout> and builds a matching Vector of
// Vector<Box> and uses it to call the full version.
///
int
LoadBalance( Vector<BoxLayout>&    Grids            //in-out: input grids to balance
//                                                  //   and output proc. numbers
            ,Real&                 effRatio         //output: ratio of min load
//                                                  //   to max load
     ,const Vector<Vector<long> >&  ComputeLoads     //input: computational cost 
//                                                  //   of each box; indexed by [Level][Box]
//## ,const [...something...]&     CommunicateLoads //input: communication cost 
//                                                  //   between each pair of   
//                                                  //   neighboring boxes      
     ,const Vector<int>&           RefRatios        //input: refinement ratio   
//                                                  //   for each level 
            ,int nProc
           )
{
  Vector<Vector<Box> > boxes( Grids.size() ) ;

  for( int i = 0 ; i < Grids.size() ; ++i ) {
    LayoutIterator j = Grids[i].layoutIterator() ;
    int p ;
    for( j.reset() ,p = 0 ; j.ok() ; ++j ,++p ) {
      boxes[i][p] = Grids[i][j()] ;
    }
  }

  Vector<Vector<int> > procIDs( Grids.size() ) ;
  int status = LoadBalance( procIDs ,effRatio
                           ,boxes ,ComputeLoads ,RefRatios, nProc ) ;

  if( status < 0 ) return status ;  //LoadBalance() failed

  for( int i = 0 ; i < Grids.size() ; ++i ) {
    LayoutIterator j = Grids[i].layoutIterator() ;
    int p ;
    for( j.reset() ,p = 0 ; j.ok() ; ++j ,++p ) {
      Grids[i].setProcID( j() ,procIDs[i][p] ) ;
    }
  }

  return status ; 
}


///
// This version takes a single BoxLayout (i.e. one level)
// and uses the box volumes to construct the compute loads
// and calls the full version.
///
int LoadBalance(Vector<int>& procs, const Vector<Box>& boxes)
{

//   Vector<Vector<Box> > layouts(1, boxes);
//   Vector<Vector<long> > computeLoads(1, Vector<long>(boxes.size()));
//   Vector<int> refRatios(1,1);
//   Vector<Vector<int> > assignments(1,Vector<int>(boxes.size(), -1));
//   Real effRatio;

//   for(int index = 0; index < boxes.size(); ++index)
//     {
//       computeLoads[0][index] = layouts[0][index].numPts();
//     }

//   int ret = LoadBalance(assignments, effRatio, layouts, computeLoads, refRatios);

//   if(ret == 0)
//     procs = assignments[0];


  // Phase 1  modified knapsack algorithm.  this one doesn't use
  // load sorting followed by round robin.  This one does bin packing
  // first by finding vector divisors, then does regular knapsack
  // optimization.

  int ret = 0;
  procs.resize(0);
  procs.resize(boxes.size(), 0);

  if(numProc() == 1 || procs.size() == 1) return 0;

  // first, compute 'total load'
  double totalLoad = 0; long long tload = 0;
  for(int index = 0; index < boxes.size(); ++index)
	{
	  totalLoad += boxes[index].numPts();
	  tload += boxes[index].numPts();
	  procs[index] = -index;
	}
  
  // figure out roughly single processor load goal
  double goal = totalLoad/numProc();

  // debugging prints
  //pout()<<"--- LoadBalance  --- Number of Boxes: "<<boxes.size() << std::endl;
  //pout() << " Box sizes: " << std::endl;  
  //for(int i=0; i < boxes.size(); i++) pout() <<" "<<boxes[i].numPts();;
  //pout() << std::endl;

  // now, first bin assignments

  // change goal to dynamicgoal (ndk)
  // We desire equal amount of cells per processor, which
  // is the total number of cells divided by the number of processors: ie the goal.
  // But because the cells are grouped into boxes of sometimes very different sizes,
  // and because we assume that the boxes are in a locality-preferable vector,
  // using a constant goal can allow for several consecutive processors to be 
  // assigned a number of cells _less_ than (or _more_ than) the goal.  
  // This can lead to the very last processor receiving an abnormally large
  // (or small) workload.  Changing the algorithm to compute a new dynamic
  // goal after each processor-box assignment eliminates this problem.
  // However, there are still improvements that can be made when the
  // number of boxes to be farmed out is not much larger than the total number
  // of processors doing the work.  (ndk)

  Vector<long long> loads(numProc(), 0);
  int bin = 0;
  double dynamicLoad = totalLoad;
  double dynamicgoal = goal;
  double load = 0;
  long long numP;
  //pout() << " first time thru: dynamicLoad =" << dynamicLoad 
  //       << " dynamicgoal = " << dynamicgoal << std::endl;

  for(int i=0; i < boxes.size(); i++)
	{
	  numP = boxes[i].numPts();
	  if(bin != numProc()-1)  // NOT last rank
		{	  
		  load += numP;
		  if(load >= dynamicgoal){
			if( ( load - dynamicgoal ) > numP/2 )
			  {
				procs[i] = bin+1;
				loads[bin+1] += numP;
				load = numP;
			  }
			else
			  {
				procs[i] = bin;
				loads[bin] += numP;
				load = 0;
			  }
                        dynamicLoad -= (double)loads[bin];
                        dynamicgoal = dynamicLoad/(double)(numProc()-(bin+1));
			bin++;
		  }
		  else
                    {  // last rank
			  loads[bin] += numP;
			  procs[i] = bin;
			}
		}
          else 
            {
              procs[i] = bin;
              loads[bin] += numP;
            }
          
	}

  //pout()<<"tload = "<<(double)tload<< std::endl;
  //pout() << " loads: ";
  //for(int i=0; i<numProc(); i++) pout()<<" "<<(long)loads[i];
  //pout()<<std::endl;

  //Phase 2 of loadbalance, actual knapsack swapping to improve efficiency
  // first , scan for highest load processor
  long long mload = loads[0];
  for(int i=1; i<loads.size(); ++i) mload = Max(mload, loads[i]);
  
  //parallel inefficiency: sum of idle time of processors waiting for max load processor
  // dividedd by total load
  //double eff = 1 - ((double)(mload*numProc() - totalLoad))/totalLoad;
  //pout() << "loadbalance efficiency (before knapsack )= "<<eff
  //     << "  Nboxes=" << boxes.size()
  //     <<std::endl;
	
  

  // Phase 3 of loadbalance:  exhanges of identical sized boxes in order
  // to improve locality.  Need to handle periodicity somehow also, but
  // that will have to wait for loadbalance to be redesigned.
  
  bool sorted = true;
  for(int i=1; i<boxes.size(); i++) 
	{
	  if(boxes[i] < boxes[i-1]) sorted = false;
	}
  Vector<std::list<int> > neighbours(boxes.size());

  for(int i=0; i<boxes.size(); i++)
	{
	  Box b = boxes[i];
	  b.grow(1);
	  std::list<int>& list = neighbours[i];
	  for(int ii=0; ii<boxes.size(); ii++)
		{
		  if(ii == i); //don't bother with self intersect
		  else{
			const Box& bb = boxes[ii];
			
			if(b.intersects(bb))
			  {
				list.push_front(ii);
				
			  }
			if(sorted && bb.smallEnd(0) > b.bigEnd(0)) ii=boxes.size();
		  }
		}
	}
  
  int swapcount = 6;
  int diffBoxes ;
  int sameBoxes ;
  int iter = 0;
  while (iter < 20 && swapcount > 0)
	{
	  swapcount = 0;
	  diffBoxes = 0;
	  sameBoxes = 0;
	  iter++;

	  for(int i=0; i<neighbours.size(); i++)
		{

		  std::list<int>& list = neighbours[i];
		  for(int ii=i+1; ii<neighbours.size(); ii++)
			{
			  int p = procs[i];
			  int pii = procs[ii];
			  if(boxes[i].numPts() != boxes[ii].numPts()) diffBoxes++; // don't swap
			  else if(p == pii) ; //do nothing, no swaps on same proc
			  else{
				sameBoxes++;
				std::list<int>& listii = neighbours[ii];

				std::list<int>::iterator it;
				// count : number of "local" exchanges before swap
				// countafter: number of "local" exchanges after swap
				int count = 0; int countafter = 0;
				for(it = list.begin();it != list.end(); it++)
				  {
					if(procs[*it] == p)
					  count++;
					if(procs[*it] == pii)
					  countafter++;
				  }
				for(it = listii.begin();it != listii.end(); it++)
				  {
					if(procs[*it] == pii)
					  count++;
					if(procs[*it] == p)
					  countafter++;
				  }	
				if(countafter > count){
				  swapcount++;
				  procs[i] = pii;
				  procs[ii] = p;
 				  //pout()<<"count : "<<count<<" countafter :"<<countafter
                                  //	<<std::endl;
				}
			  }
			}
		}
 	  //pout()<<"Boxes: "<<boxes.size()<<" same: "<<sameBoxes
          //	<<" swaps: "<<swapcount<<std::endl;
	}
  
  return ret;
}


///
// This version does the real work.
///
int
LoadBalance(Vector<Vector<int> >& procAssignments  //output: processor number  
//                                                  //   for each box           
            ,Real&                 effRatio         //output: ratio of min load
//                                                  //   to max load
            ,const Vector<Vector<Box> >&  Grids            //input: meshes to balance  
            ,const Vector<Vector<long> >& ComputeLoads     //input: computational cost 
//                                                  //   of each box            
//## ,const [...something...]&     CommunicateLoads //input: communication cost 
//                                                  //   between each pair of   
//                                                  //   neighboring boxes      
            ,const Vector<int>&           RefRatios        //input: refinement ratio   
                                                    //   for each level   
            ,int nProc                            // number of procs to assugn to
           )
{
  // local variables
  Real eff_ratio ; // efficiency ratio on a level
  int status = 0 ; // return code

  // Validate inputs
  if( Grids.size() != ComputeLoads.size() ) { return -1011 ; }
  if( Grids.size() != RefRatios.size() ) { return -1013 ; }
  for( int lvl=0 ; lvl<Grids.size() ; ++lvl )
    {
      if( Grids[lvl].size() != ComputeLoads[lvl].size() ) { return -1012 ; }
    }

  // set the number of elements in the output vector to the number
  // of levels and the number of elements in each element to the 
  // number of boxes on each level and set the value of each 
  // element to zero
  procAssignments.resize( Grids.size() ) ;
  for( int lvl=0 ; lvl<Grids.size() ; ++lvl )
    {
      procAssignments[lvl].resize( Grids[lvl].size(),0 ) ;
    }

  // check for special case of all loads on 1 processor
  if( nProc == 1 )
    {
      for( int lvl=0 ; lvl<Grids.size() ; ++lvl )
        {
          for( int i=0 ; i<Grids[lvl].size() ; ++i )
            {
              procAssignments[lvl][i] = 0 ;
            }
        }
      effRatio = 1.0 ;
      status = 0 ;
    }
  else
    {
      // general case: loads on more than one processor
      effRatio = 1.0 ;

      // balance each level separately
      for( int lvl=0 ; lvl<Grids.size() ; ++lvl )
        {
          // first, build the load structure and sort by compute_load
          Vector<Load> loads( Grids[lvl].size() ) ;
          for( int i=0 ; i<Grids[lvl].size() ; ++i )
            {
              loads[i].load = ComputeLoads[lvl][i] ;
              loads[i].grid_index = i ;
            }
          //          std::sort( loads.begin() ,loads.end() ) ;
          loads.sort();
          // do the initial assignments by sequentially
          // `handing out' the loads from largest to smallest
          Vector<long> total_loads( nProc,0 ) ; //total load per processor
          Vector<Vector<Load> > proc_loads( nProc ) ; //loads per processor
          int iproc_minload = 0 ; //processor with lowest load
          // loads are sorted in increasing order, so work backwards through the vector
          for( int i=loads.size()-1 ; i>=0 ; --i )
            {
              // put the next load on the processor with the lowest total load
              proc_loads[iproc_minload].push_back( loads[i] ) ;
              total_loads[iproc_minload] += loads[i].load ;

              // recompute which processor has the lowest load
              //[NOTE: this would be faster if the loads were sorted]
              iproc_minload = min_element( total_loads ) ;
            }
          // compute average load per processor, truncated to int
          long avg_load = 0 ;
          for( int i=0 ; i<total_loads.size() ; ++i ) avg_load += total_loads[i] ;
          avg_load /= nProc ;

          // optimize the assignments by swapping a load off the
          // processor with the max load onto another processor
          // such that the load balance is improved
          int iter_count = 0, swap_count = 0 ;
          int iproc_maxload ;
          long max_change ; //largest change in load balance
          int ibmax,jbmax,ipmax,jpmax ;  //box and processor indices corresponding to max_change

          while( 1 )
            {
              max_change = 0 ;

              // find the processor that has the largest deviation from perfect load balance
              min_max_elements( iproc_minload ,iproc_maxload ,total_loads ) ;
              if( iproc_minload == iproc_maxload )
                {
                  // load balance is perfect
                  // (this won't happen except in test cases)
                  break ;
                }
              assert( total_loads[iproc_minload] <= avg_load &&
                      avg_load <= total_loads[iproc_maxload] ) ;
              if( avg_load - total_loads[iproc_minload] > total_loads[iproc_maxload] - avg_load )
                ipmax = iproc_minload ;
              else
                ipmax = iproc_maxload ;

              //[NOTE: dont need this here, but it may be useful for debugging.]
              eff_ratio = (Real)total_loads[iproc_minload] / (Real)total_loads[iproc_maxload] ;

              // deviation from perfect load balance for this proc
              long devib = total_loads[ipmax] - avg_load ;

              // search all the other processors for the swap that has the maximum
              // reduction in the total deviation from the perfect load balance
              for( int j=0 ; j<proc_loads.size() ; ++j )
                {
                  if( j != ipmax )
                    {
                      long devjb = total_loads[j] - avg_load ;

                      // loop over all boxes on both processors
                      for( int ibox=0 ; ibox<proc_loads[ipmax].size() ; ++ibox )
                        {
                          for( int jbox=0 ; jbox<proc_loads[j].size() ; ++jbox )
                            {
                              iter_count++ ;
                              // how much bigger is the ibox load than the jbox load?
                              long diff = proc_loads[ipmax][ibox].load
                                        - proc_loads[  j  ][jbox].load ;
                              // change in total deviation from swapping boxes
                              long change = Abs( devib ) + Abs( devjb ) 
                                          - Abs( devib - diff ) - Abs( devjb + diff ) ;
                              // remember this pair of boxes if the change is better
                              //[NOTE: max_change starts at 0, so this is always an improvement]
                              if( change > max_change )
                                {
                                  max_change = change ;
                                  ibmax = ibox ; jbmax = jbox ; jpmax = j ;
                                }
                            }
                        }
                    }
                }
              // if there is a swap that improves load balance, take it; else stop
              if( max_change > 0 )
                {
                  // adjust the total loads on each processor
                  long load_diff = proc_loads[ipmax][ibmax].load
                                 - proc_loads[jpmax][jbmax].load ;
                  assert( load_diff != 0 ) ;
                  total_loads[ipmax] -= load_diff ;
                  total_loads[jpmax] += load_diff ;
                  // swap the loads
                  Load tmp = proc_loads[ipmax][ibmax] ;
                  proc_loads[ipmax][ibmax] = proc_loads[jpmax][jbmax] ;
                  proc_loads[jpmax][jbmax] = tmp ;

                  swap_count++ ;
                }
              else
                {
                  break ;
                }
            }

          // Done with this level.

          // Compute the final efficiency ratio and save it if appropriate.
          min_max_elements( iproc_minload ,iproc_maxload ,total_loads ) ;
          eff_ratio = (Real)total_loads[iproc_minload] / (Real)total_loads[iproc_maxload] ;
          if( eff_ratio < effRatio ) effRatio = eff_ratio ;

          // Assign boxes to processors for this level.
          for( int ip=0 ; ip<proc_loads.size() ; ++ip )
            {
              for( int jb=0 ; jb<proc_loads[ip].size() ; ++jb )
                {
                  procAssignments[lvl][proc_loads[ip][jb].grid_index] = ip ;
                }
            }

#ifndef NDEBUG
//           if( iter_count > 0 )
//             {
//               cout << "    debug: LoadBalance: level " << lvl << " used "
//                    << iter_count << " iterations and "
//                    << swap_count << " swaps to get efficiency ratio "
//                    << eff_ratio << std::endl ;
//             }
#endif
        }

      // Done with all levels.  
      // We could try to permute processors assignments between levels to
      // reduce communication, but it is probably not worth the effort
      // since it probably would have O(N^4) cost (N==#boxes).
    }

  return status;
}


////////////////////////////////////////////////////////////////
//                utility functions                           //
////////////////////////////////////////////////////////////////

//
// Find the index of the small value in a non-empty (long) Vector
//
int
min_element( const Vector<long>& Vect )
{
  assert( Vect.size() > 0 );
  int imin = 0 ;
  for( int i=1 ; i<Vect.size() ; ++i )
    {
      if( Vect[i] < Vect[imin] ) imin = i ;
    }
  return imin;
}


//
// Find the indices of the smallest and largest values in a non-empty (long) Vector
//
void
min_max_elements( int& imin ,int& imax ,const Vector<long>& Vect )
{
  assert( Vect.size() > 0 );
  imin = 0 ; imax = 0 ;
  for( int i=1 ; i<Vect.size() ; ++i )
    {
      if( Vect[i] < Vect[imin] ) imin = i ;
      if( Vect[i] > Vect[imax] ) imax = i ;
    }
  return;
}
void
min_max_elements( int& imin ,int& imax ,const Vector<long long>& Vect )
{
  assert( Vect.size() > 0 );
  imin = 0 ; imax = 0 ;
  for( int i=1 ; i<Vect.size() ; ++i )
    {
      if( Vect[i] < Vect[imin] ) imin = i ;
      if( Vect[i] > Vect[imax] ) imax = i ;
    }
  return;
}

/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef LDF_H
#define LDF_H

#ifndef WRAPPER
#endif


class LDF
{
public:
  LDF()
  {
	m_ptr = new LevelData<FArrayBox>();}

  LDF(LevelData<FArrayBox>* newPtr)
  {
	m_ptr = RefCountedPtr<LevelData<FArrayBox> >(newPtr);}

  void define(const DisjointBoxLayout& dp, int comps,
			  const IntVect& ghost) 
  {
	m_ptr->define(dp, comps, ghost);}

  ~LDF(){;}

  void exchange(const Interval& comps)
  {
	m_ptr->exchange(comps);}

  void copyTo(const Interval& srcComps, 
			  LDF& dest,
			  const Interval& destComps) const
  {
	m_ptr-> copyTo(srcComps, *(dest.m_ptr), destComps);}

  IntVect ghostVect() const {return m_ptr->ghostVect();}

  const DisjointBoxLayout& getBoxes() const {return m_ptr->getBoxes();}

  const DisjointBoxLayout& disjoiuntBoxLayout() const {return m_ptr->getBoxes();}

  Interval interval() const {return m_ptr->interval();}

  DataIterator dataIterator() const {return m_ptr->dataIterator();}

  Box box(const DataIndex& index) const { return m_ptr->box(index);}

  FArrayBox& get(const DataIndex& index) { return m_ptr->operator[](index);}

  const LevelData<FArrayBox>* dataPtr() const {return m_ptr;}

  void operator=(LevelData<FArrayBox>* newPtr){
	m_ptr = RefCountedPtr<LevelData<FArrayBox> >(newPtr);}

private:
  RefCountedPtr<LevelData<FArrayBox> > m_ptr;
};

#endif
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// DTGraves, Fri, Dec 3, 1999

#ifndef _AMRIO_
#define _AMRIO_
#ifndef WRAPPER

using std::string;
using std::istream;
#endif

class AMRData 
{
public:
  AMRData(){ numLevels = 0; dx=0.0; dt=0.0; time=0.0; ghost=ZeroVector();}
  Vector< DisjointBoxLayout > vectGrids;  
  Vector< LDF >  vectData;  
  Vector< string > vectNames; 
  Box domain;  
  Real dx;  
  Real dt;  
  Real time; 
  Vector< int > vectRatio;   
  int numLevels;
  IntVect ghost;
};

///
/**
   Writes hierarchy of levels in HDF5 format.  Only available if the
   preprocessor macro HDF5 is defined at compilation.

   {\bf Arguments:}\\
 filename  :  file to output to.\\
 a_vectGrids : grids at each level.\\
 a_vectData :  data at each level.\\
 a_vectNames:  names of variables.\\
 a_domain :  domain at coarsest level.\\
 a_dx     :  grid spacing at coarsest level.\\
 a_dt     :  time step at coarsest level.\\
 a_time     :  time.\\
 a_vectRatio :  refinement ratio at all levels
 (ith entry is refinement ratio between levels i and i + 1).\\
 a_numLevels :  number of levels to output.\\ 
This is blocking.

*/
void WriteAMRHierarchyHDF5(string filename,  const AMRData& data);
///
/**
   Writes hierarchy of levels in HDF5 format.  Only available if the
   preprocessor macro HDF5 is defined at compilation.

   {\bf Arguments:}\\
 filename  :  file to output to.\\
 a_vectGrids : grids at each level.\\
 a_vectData :  data at each level.\\
 a_vectNames:  names of variables.\\
 a_domain :  domain at coarsest level.\\
 a_dx     :  grid spacing at coarsest level.\\
 a_dt     :  time step at coarsest level.\\
 a_time     :  time.\\
 a_vectRatio :  refinement ratio at all levels
 (ith entry is refinement ratio between levels i and i + 1).\\
 a_numLevels :  number of levels to output.\\ 
This is blocking.

*/
void
WriteAMRHierarchyHDF5(const string& filename,                             
                      const Vector<DisjointBoxLayout>& a_vectGrids, 
                      const Vector<LevelData<FArrayBox>* > & a_vectData, 
                      const Vector<string>& a_vectNames,                 
                      const Box& a_domain,                               
                      const Real& a_dx,                                         
                      const Real& a_dt,                                         
                      const Real& a_time,
                      const Vector<int>& a_vectRatio,                           
                      const int& a_numLevels);


///
/**
   Writes hierarchy of levels in HDF5 format.  Only available if the
   preprocessor macro HDF5 is defined at compilation.

   {\bf Arguments:}\\
handle  :  handle to output to.\\
 a_vectGrids : grids at each level.\\
 a_vectData :  data at each level.\\
 a_vectNames:  names of variables.\\
 a_domain :  domain at coarsest level.\\
 a_dx     :  grid spacing at coarsest level.\\
 a_dt     :  time step at coarsest level.\\
 a_time     :  time.\\
 a_vectRatio :  refinement ratio at all levels
 (ith entry is refinement ratio between levels i and i + 1).\\
 a_numLevels :  number of levels to output.\\ 
This is not blocking.

*/
void
WriteAMRHierarchyHDF5(HDF5Handle& handle,                             
                      const Vector<DisjointBoxLayout>& a_vectGrids, 
                      const Vector<LevelData<FArrayBox>* > & a_vectData, 
                      const Vector<string>& a_vectNames,                 
                      const Box& a_domain,                               
                      const Real& a_dx,                                         
                      const Real& a_dt,                                         
                      const Real& a_time,
                      const Vector<int>& a_vectRatio,                           
                      const int& a_numLevels);


///
/**
   Writes hierarchy of levels in HDF5 format.  Only available if the
   preprocessor macro HDF5 is defined at compilation.

   {\bf Arguments:}\\
 filename  :  file to output to.\\
 a_vectGrids : grids at each level.\\
 a_vectData :  data at each level.\\
 a_domain :  domain at coarsest level.\\
 a_vectRatio :  refinement ratio at all levels
 (ith entry is refinement ratio between levels i and i + 1).\\
 a_numLevels :  number of levels to output.\\ 
This is blocking.
*/
void
WriteAMRHierarchyHDF5(const string& filename,                             
                      const Vector<DisjointBoxLayout>& a_vectGrids, 
                      const Vector<LevelData<FArrayBox>* > & a_vectData, 
                      const Box& a_domain,                               
                      const Vector<int>& a_vectRatio,                           
                      const int& a_numLevels);



///
/**
   Writes hierarchy of levels in HDF5 format.  Only available if the
   preprocessor macro HDF5 is defined at compilation.

   {\bf Arguments:}\\
 handle  :  handle to output to.\\
 a_vectGrids : grids at each level.\\
 a_vectData :  data at each level.\\
 a_domain :  domain at coarsest level.\\
 a_vectRatio :  refinement ratio at all levels
 (ith entry is refinement ratio between levels i and i + 1).\\
 a_numLevels :  number of levels to output.\\ 
This is not blocking.
*/
void
WriteAMRHierarchyHDF5(HDF5Handle& handle, 
                      const Vector<DisjointBoxLayout>& a_vectGrids, 
                      const Vector<LevelData<FArrayBox>* > & a_vectData, 
                      const Box& a_domain,                               
                      const Vector<int>& a_vectRatio,                           
                      const int& a_numLevels);


///
/**
   Reads hierarchy of levels in HDF5 format.  Only available if the
   preprocessor macro HDF5 is defined at compilation.  Returns 0 on success.

   {\bf Arguments:}\\
filename  :  file to input from.\\
a_vectGrids : grids at each level.\\
a_vectData :  data at each level.\\
a_vectNames:  names of variables.\\
a_domain :  domain at coarsest level.\\
a_dx     :  grid spacing at coarsest level.\\
a_dt     :  time step at coarsest level.\\
a_time     :  time.\\
a_vectRatio :  refinement ratio at all levels
(ith entry is refinement ratio between levels i and i + 1).\\
a_numLevels :  number of levels to read.\\

{\bf Returns:} \\
status code with values:\\
0: success \\
-1: bogus number of levels \\
-2: bogus number of components \\
-3: error in readlevel \\
-4: file open failed \\
This is blocking 
*/

int
ReadAMRHierarchyHDF5(string filename,  AMRData& data);

#ifndef WRAPPER
///
/**
   Reads hierarchy of levels in HDF5 format.  Only available if the
   preprocessor macro HDF5 is defined at compilation.  Returns 0 on success.

   {\bf Arguments:}\\
filename  :  file to input from.\\
a_vectGrids : grids at each level.\\
a_vectData :  data at each level.\\
a_vectNames:  names of variables.\\
a_domain :  domain at coarsest level.\\
a_dx     :  grid spacing at coarsest level.\\
a_dt     :  time step at coarsest level.\\
a_time     :  time.\\
a_vectRatio :  refinement ratio at all levels
(ith entry is refinement ratio between levels i and i + 1).\\
a_numLevels :  number of levels to read.\\

{\bf Returns:} \\
status code with values:\\
0: success \\
-1: bogus number of levels \\
-2: bogus number of components \\
-3: error in readlevel \\
-4: file open failed \\
This is blocking
*/
int
ReadAMRHierarchyHDF5(const string& filename,                             
                     Vector<DisjointBoxLayout>& a_vectGrids, 
                     Vector<LevelData<FArrayBox>* > & a_vectData, 
                     Vector<string>& a_vectNames,                 
                     Box& a_domain,                               
                     Real& a_dx,                                         
                     Real& a_dt,                                         
                     Real& a_time,                                         
                     Vector<int>& a_refRatio,                           
                     int& a_numLevels,
                     const IntVect& a_ghostVector = IntVect::TheZeroVector());

///
/**
   Reads hierarchy of levels in HDF5 format.  Only available if the
   preprocessor macro HDF5 is defined at compilation.  Returns 0 on success.

   {\bf Arguments:}\\
handle  :  handle to input from.\\
a_vectGrids : grids at each level.\\
a_vectData :  data at each level.\\
a_vectNames:  names of variables.\\
a_domain :  domain at coarsest level.\\
a_dx     :  grid spacing at coarsest level.\\
a_dt     :  time step at coarsest level.\\
a_time     :  time.\\
a_vectRatio :  refinement ratio at all levels
(ith entry is refinement ratio between levels i and i + 1).\\
a_numLevels :  number of levels to read.\\

{\bf Returns:} \\
status code with values:\\
0: success \\
-1: bogus number of levels \\
-2: bogus number of components \\
-3: error in readlevel \\
-4: file open failed \\
This is not blocking
*/
int
ReadAMRHierarchyHDF5(HDF5Handle& handle,
                     Vector<DisjointBoxLayout>& a_vectGrids, 
                     Vector<LevelData<FArrayBox>* > & a_vectData, 
                     Vector<string>& a_vectNames,                 
                     Box& a_domain,                               
                     Real& a_dx,                                         
                     Real& a_dt,                                         
                     Real& a_time,                                         
                     Vector<int>& a_refRatio,                           
                     int& a_numLevels,
                     const IntVect& a_ghostVector = IntVect::TheZeroVector());

///
/**
   Reads hierarchy of levels in HDF5 format.  Only available if the
   preprocessor macro HDF5 is defined at compilation.  Returns 0 on success.

   {\bf Arguments:}\\
filename  :  file to input from.\\
a_vectGrids : grids at each level.\\
a_vectData :  data at each level.\\
a_domain :  domain at coarsest level.\\
a_vectRatio :  refinement ratio at all levels
(ith entry is refinement ratio between levels i and i + 1).\\
a_numLevels :  number of levels to read.\\

{\bf Returns:} \\
status code with values:\\
0: success \\
-1: bogus number of levels \\
-2: bogus number of components \\
-3: error in readlevel \\
-4: file open failed \\
This is blocking.
*/
int
ReadAMRHierarchyHDF5(const string& filename,                             
                     Vector<DisjointBoxLayout>& a_vectGrids, 
                     Vector<LevelData<FArrayBox>* > & a_vectData, 
                     Box& a_domain,                               
                     Vector<int>& a_refRatio,                           
                     int& a_numLevels,
                     const IntVect& a_ghostVector = IntVect::TheZeroVector());


///
/**
   Reads hierarchy of levels in HDF5 format.  Only available if the
   preprocessor macro HDF5 is defined at compilation.  Returns 0 on success.

   {\bf Arguments:}\\
handle  :  handle to input from.\\
a_vectGrids : grids at each level.\\
a_vectData :  data at each level.\\
a_domain :  domain at coarsest level.\\
a_vectRatio :  refinement ratio at all levels
(ith entry is refinement ratio between levels i and i + 1).\\
a_numLevels :  number of levels to read.\\

{\bf Returns:} \\
status code with values:\\
0: success \\
-1: bogus number of levels \\
-2: bogus number of components \\
-3: error in readlevel \\
-4: file open failed \\
This is not blocking.
*/
int
ReadAMRHierarchyHDF5(HDF5Handle& handle,
                     Vector<DisjointBoxLayout>& a_vectGrids, 
                     Vector<LevelData<FArrayBox>* > & a_vectData, 
                     Box& a_domain,                               
                     Vector<int>& a_refRatio,                           
                     int& a_numLevels,
                     const IntVect& a_ghostVector = IntVect::TheZeroVector());
// ///
/**
   Reads hierarchy of EB levels in ASCII format.   Returns 0 on success.
   Data is read in as FArrayBoxes because the data file does not 
   {\bf Arguments:}\\
   filename  :  file to input from.\\
   a_vectGrids : grids at each level.\\
   a_vectData :  data at each level.\\
   a_vectNames:  names of variables.\\
   a_domain :  domain at coarsest level.\\
   a_vectRatio :  refinement ratio at all levels
   (ith entry is refinement ratio between levels i and i + 1).\\
   a_multiValuedCells :  multi-valued cells at all levels \\
   (ith entry is refinement ratio between levels i and i + 1).\\
   a_coveredCells :  covered cells at all levels \\
   (ith entry is refinement ratio between levels i and i + 1).\\
   a_numLevels :  number of levels to read.\\
   returns 0 if successful.
*/
 int
 ReadEBAMRASCII(const string& filename,                             
                Vector<DisjointBoxLayout>& a_vectGrids, 
                Vector<LevelData<FArrayBox>* > & a_vectData, 
                Vector<string>& a_vectNames,                 
                Box& a_domain,                               
                Vector<IntVectSet>& a_coveredCells,                               
               Vector<IntVectSet>& a_multiValuedCells,                               
               Vector<int>& a_refRatio,                           
               int& a_numLevels,
               const IntVect& a_ghostVector = IntVect::TheZeroVector());

#endif

/// 
/** Writes a plotfile using the same format as WriteAMRHierarchyHDF5, but
    for a FArrayBox.  This is useful for debugging.  *a_dataPtr is written 
    to a file named fab.hdf5
*/
void
writeFAB(const FArrayBox* a_dataPtr);

/// 
/** View *a_dataPtr by writing it to an HDF5 plotfile (to a temporary file)
    and then running ChomboVis with a python script which brings up a data
    browser by default. The file has the same format as WriteAMRHierarchyHDF5,
    but for a single FArrayBox.  This is useful for debugging.
*/
void
viewFAB(const FArrayBox* a_dataPtr);

///
/** Writes a plotfile using the same format as WriteAMRHierarchyHDF5, but
    for a FArrayBox.  This is useful for debugging.  *a_dataPtr is written
    to the file given by a_filename.
*/
void
writeFABname(const FArrayBox* a_dataPtr,
             const char*      a_filename);

///
/** Write a plotfile using the same format as WriteAMRHierarchyHDF5, but
    for a single LevelData<FArrayBox>.  Useful for debugging.  *a_dataPtr is
    written to a file named LDF.hdf5.
*/
void
writeLDF(const LDF* a_dataPtr);

///
/** Write a plotfile using the same format as WriteAMRHierarchyHDF5, but
    for a single LevelData<FArrayBox>.  Useful for debugging.  *a_dataPtr is
    written to a file named LDF.hdf5.
*/
void
writeLevel(const LevelData<FArrayBox>* a_dataPtr);

/// 
/** View *a_dataPtr by writing it to an HDF5 plotfile (to a temporary file)
    and then running ChomboVis with a python script which brings up a data
    browser by default. The file has the same format as WriteAMRHierarchyHDF5,
    but for a single LevelData<FArrayBox>.  This is useful for debugging.
*/
void
viewLevel(const LevelData<FArrayBox>* a_dataPtr);

///
/** Write a plotfile using the same format as WriteAMRHierarchyHDF5, but
    for a single LevelData<FArrayBox>. Useful for debugging.  *a_dataPtr is
    written to the file given by a_filename.
*/
void
writeLDFname(const LDF*    a_dataPtr,
             const string& a_filename);

///
/** Write a plotfile using the same format as WriteAMRHierarchyHDF5, but
    for a single LevelData<FArrayBox>. Useful for debugging.  *a_dataPtr is
    written to the file given by a_filename.
*/
void
writeLevelname(const LevelData<FArrayBox>* a_dataPtr,
               const char*                 a_filename);

///
/** Write a plotfile with levels that don't necessarily begin with 
0 (will translate things so that lBase comes out as level 0, etc)
Levels in Vector arguments should correspond with actual levels.
(and so should be defined for a_levels.begin() through a_levels.end().
   {\bf Arguments:}\\
 filename  :  file to output to.\\
 a_vectGrids : grids at each level.\\
 a_vectData :  data at each level.\\
 a_vectNames:  names of variables.\\
 a_domain :  domain at base level (given by  a_levels.begin()).\\
 a_dx     :  grid spacing at base level.\\
 a_dt     :  time step at base level.\\
 a_time     :  time.\\
 a_vectRatio :  refinement ratio at all levels
 (ith entry is refinement ratio between levels i and i + 1).\\
 a_numLevels :  number of levels to output.\\ 
This is blocking.

*/
void
WritePartialAMRHierarchyHDF5(const string& filename,
                             const Vector<DisjointBoxLayout>& a_vectGrids, 
                             const Vector<LevelData<FArrayBox>* > & a_vectData,
                             const Vector<string>& a_vectNames,
                             const Box& a_baseDomain,
                             const Real& a_baseDx,
                             const Real& a_dt,
                             const Real& a_time,
                             const Vector<int>& a_vectRatio,
                             const Interval& a_levels);





// //write a fab out in ascii very slowly
 extern void 
 writeAFabASCII(std::ostream& os, const FArrayBox& fab);

// //read a  fab in, in ascii, very slowly
 extern void 
 readAFabASCII(istream& is, FArrayBox& fab);

 void writeAFabASCII(std::ostream& os, const FArrayBox& fab);

 void readAFabASCII(istream& is, FArrayBox& fab);

// ///
/**
 Writes hierarchy of levels in ASCII format.

    {\bf Arguments:}\\
 filename  :  file to output to.\\
 a_vectGrids : grids at each level.\\
 a_vectData :  data at each level.\\
 a_vectNames:  names of variables.\\
 a_domain :  domain at coarsest level.\\
 a_dx     :  grid spacing at coarsest level.\\
 a_dt     :  time step at coarsest level.\\
  a_time     :  time.\\
 a_vectRatio :  refinement ratio at all levels
 (ith entry is refinement ratio between levels i and i + 1).\\
 a_numLevels :  number of levels to output.\\

*/
 void
WriteAMRHierarchyASCII(const string& filename,                             
                       const Vector<DisjointBoxLayout>& a_vectGrids, 
                       const Vector<LevelData<FArrayBox>* > & a_vectData, 
                       const Vector<string>& a_vectNames,                 
                       const Box& a_domain,                               
                       const Real& a_dx,
                       const Real& a_dt,
                       const Real& a_time,
                       const Vector<int>& a_vectRatio,
                       const int& a_numLevels);


// ///
/**
Reads hierarchy of levels in ASCII format.  Returns 0 on success.

   {\bf Arguments:}\\
filename  :  file to output to.\\
a_vectGrids : grids at each level.\\
a_vectData :   data at each level.\\
a_vectNames:  names of variables.\\
a_domain :  domain at coarsest level.\\
a_dx     :  grid spacing at coarsest level.\\
a_dt     :  time step at coarsest level.\\
a_time     :  time.\\
a_vectRatio :  refinement ratio at all levels
(ith entry is refinement ratio between levels i and i + 1).\\
a_numLevels :  number of levels to output.\\
a_ghostVector :  number of ghost cells on output .\\

{\bf Returns:}\\
status code with values:\\
0: success\\
-1: failure\\

*/
int
ReadAMRHierarchyASCII(const string& filename,                             
                      Vector<DisjointBoxLayout>& a_vectGrids, 
                      Vector<LevelData<FArrayBox>* > & a_vectData, 
                      Vector<string>& a_vectNames,                 
                      Box& a_domain,                               
                      Real& a_dx,                                         
                      Real& a_dt,                                         
                      Real& a_time,                                         
                      Vector<int>& a_vectRatio,                           
                      int& a_numLevels,
                      const IntVect& ghostVector = IntVect::TheZeroVector());

#endif
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// DTGraves, Fri, Dec 3, 1999


using std::fstream;
using std::string;

using std::sprintf;
//using std::tempnam;

using std::system;



/*
\\ write out hierarchy of amr data in HDF5 format
\\ filename,  == file to output to
\\ a_vectData == data at each level
\\ a_vectNames== names of variables
\\ a_domain == domain at coarsest level
\\ a_dx     == grid spacing at coarsest level
\\ a_dt     == time step at coarsest level
\\ a_time     == time 
\\ a_vectRatio == refinement ratio at all levels 
\\ (ith entry is refinement ratio between levels i and i + 1)
\\ a_numLevels == number of levels to output
*/
void
WriteAMRHierarchyHDF5(const string& filename,                             
                      const Vector<DisjointBoxLayout>& a_vectGrids, 
                      const Vector<LevelData<FArrayBox>* > & a_vectData, 
                      const Vector<string>& a_vectNames,                 
                      const Box& a_domain,                               
                      const Real& a_dx,
                      const Real& a_dt,
                      const Real& a_time,
                      const Vector<int>& a_refRatio,                           
                      const int& a_numLevels)
{
  HDF5Handle handle(filename.c_str(),  HDF5Handle::CREATE);

  WriteAMRHierarchyHDF5(handle, a_vectGrids, a_vectData, a_vectNames,
                        a_domain, a_dx, a_dt, a_time, a_refRatio, a_numLevels);

#ifdef MPI  
  MPI_Barrier(Chombo_MPI::comm);
#endif   

  handle.close();
}
void
WriteAMRHierarchyHDF5(string filename, const AMRData& data)
{
  HDF5Handle handle(filename.c_str(),  HDF5Handle::CREATE);

  Vector<LevelData<FArrayBox>* >  vectData(data.vectData.size(), NULL);
  for(int i=0; i<vectData.size(); ++i)
	{
	  vectData[i] = (LevelData<FArrayBox>*) (data.vectData[i].dataPtr());
	}
  WriteAMRHierarchyHDF5(handle, data.vectGrids, vectData, data.vectNames,
                        data.domain, data.dx, data.dt, data.time, 
						data.vectRatio, data.numLevels);

#ifdef MPI  
  MPI_Barrier(Chombo_MPI::comm);
#endif   

  handle.close();
}



void
WriteAMRHierarchyHDF5(HDF5Handle& handle,                             
                      const Vector<DisjointBoxLayout>& a_vectGrids, 
                      const Vector<LevelData<FArrayBox>* > & a_vectData, 
                      const Vector<string>& a_vectNames,                 
                      const Box& a_domain,                               
                      const Real& a_dx,                                         
                      const Real& a_dt,                                         
                      const Real& a_time,                                         
                      const Vector<int>& a_refRatio,                           
                      const int& a_numLevels)
{
  assert(a_numLevels > 0);
  assert(a_vectData.size()  >= a_numLevels);
  assert(a_refRatio.size() >= a_numLevels-1);

  HDF5HeaderData header;
  int nComp = a_vectNames.size();

  string filedescriptor("VanillaAMRFileType");
  header.m_string ["filetype"]      = filedescriptor;
  header.m_int ["num_levels"]       = a_numLevels;
  header.m_int ["num_components"]    = nComp;

  for(int ivar = 0; ivar < nComp; ivar++)
    {
      char labelChSt[80];
      sprintf(labelChSt, "component_%d", ivar);
      string label(labelChSt);
      header.m_string[label] = a_vectNames[ivar];
    }
  header.writeToFile(handle);

  Box domainLevel = a_domain;
  Real dtLevel = a_dt;
  Real dxLevel = a_dx;
  for(int ilev = 0; ilev < a_numLevels; ilev++)
    {
      int refLevel = 1;
      if(ilev != a_numLevels -1)
          refLevel = a_refRatio[ilev];
      if(ilev != 0)
        {
          domainLevel.refine(a_refRatio[ilev-1]);
          dtLevel /= a_refRatio[ilev-1];
          dxLevel /= a_refRatio[ilev-1];
        }
      assert(a_vectData[ilev] != NULL);
      const LevelData<FArrayBox>& dataLevel = *a_vectData[ilev];
      assert(dataLevel.nComp() == nComp);
      int eek = writeLevel(handle, ilev, dataLevel, 
                           dxLevel, dtLevel,  a_time,
                           domainLevel, refLevel);
      if(eek != 0)
        MayDay::Error("WriteAMRHierarchyHDF5: Error in writeLevel");
    }
}

void
WriteAMRHierarchyHDF5(const string& filename,                             
                      const Vector<DisjointBoxLayout>& a_vectGrids, 
                      const Vector<LevelData<FArrayBox>* > & a_vectData, 
                      const Box& a_domain,                               
                      const Vector<int>& a_refRatio,                           
                      const int& a_numLevels)
{

  HDF5Handle handle(filename.c_str(),  HDF5Handle::CREATE);
  WriteAMRHierarchyHDF5(handle, a_vectGrids, a_vectData, 
                        a_domain, a_refRatio, a_numLevels);

#ifdef MPI
  MPI_Barrier(Chombo_MPI::comm);
#endif  
  handle.close();
}

void
WriteAMRHierarchyHDF5(HDF5Handle& handle,                             
                      const Vector<DisjointBoxLayout>& a_vectGrids, 
                      const Vector<LevelData<FArrayBox>* > & a_vectData, 
                      const Box& a_domain,                               
                      const Vector<int>& a_refRatio,                           
                      const int& a_numLevels)
{
  assert(a_numLevels > 0);
  assert(a_vectData.size()  >= a_numLevels);
  assert(a_refRatio.size() >= a_numLevels-1);
  Real dxin = 1.0;
  Real dtin = 1.0;
  Real time = 1.0;

  HDF5HeaderData header;
  int nComp = a_vectData[0]->nComp();

  string filedescriptor("VanillaAMRFileType");
  header.m_string ["filetype"]      = filedescriptor;
  header.m_int ["num_levels"]       = a_numLevels;
  header.m_int ["num_components"]    = nComp;

  for(int ivar = 0; ivar < nComp; ivar++)
    {
      char labelChSt[80];
      sprintf(labelChSt, "component_%d", ivar);
      string label(labelChSt);
      header.m_string[label] = label;
    }
  header.writeToFile(handle);

  Box domainLevel = a_domain;
  Real dtLevel = dtin;
  Real dxLevel = dxin;
  for(int ilev = 0; ilev < a_numLevels; ilev++)
    {
      int refLevel = 1;
      if(ilev != a_numLevels -1)
          refLevel = a_refRatio[ilev];
      if(ilev != 0)
        {
          domainLevel.refine(a_refRatio[ilev-1]);
          dtLevel /= a_refRatio[ilev-1];
          dxLevel /= a_refRatio[ilev-1];
        }
      assert(a_vectData[ilev] != NULL);
      const LevelData<FArrayBox>& dataLevel = *a_vectData[ilev];
      assert(dataLevel.nComp() == nComp);
      int eek = writeLevel(handle, ilev, dataLevel, 
                           dxLevel, dtLevel,  time,
                           domainLevel, refLevel);
      if(eek != 0)
        MayDay::Error("WriteAMRHierarchyHDF5: Error in writeLevel");
    }
}

//
/*
\\ Read in hierarchy of amr data in HDF5 format
\\ filename,  == file to output to
\\ a_vectData == data at each level
\\ a_vectNames== names of variables
\\ a_domain == domain at coarsest level
\\ a_dx     == grid spacing at coarsest level
\\ a_dt     == time step at coarsest level
\\ a_time     == time 
\\ a_vectRatio == refinement ratio at all levels 
\\ (ith entry is refinement ratio between levels i and i + 1)
\\ a_numLevels == number of levels to output

return values:
0: success
-1: bogus number of levels
-2: bogus number of components
-3: error in readlevel
-4: file open failed
*/
int
ReadAMRHierarchyHDF5(const string& filename,                             
                     Vector<DisjointBoxLayout>& a_vectGrids, 
                     Vector<LevelData<FArrayBox>* > & a_vectData, 
                     Vector<string>& a_vectNames,                 
                     Box& a_domain,                               
                     Real& a_dx,                                         
                     Real& a_dt,                                         
                     Real& a_time,                                         
                     Vector<int>& a_refRatio,                           
                     int& a_numLevels,
                     const IntVect& ghostVector)
{
  HDF5Handle handle;
  int err = handle.open(filename.c_str(),  HDF5Handle::OPEN_RDONLY);
  if( err < 0) return -4;
  int eekflag = ReadAMRHierarchyHDF5(handle, a_vectGrids, a_vectData,
                                     a_vectNames, a_domain, a_dx, a_dt,
                                     a_time, a_refRatio, a_numLevels, 
                                     ghostVector);

#ifdef MPI
  MPI_Barrier(Chombo_MPI::comm);
#endif  
  handle.close();

  return (eekflag);
}

int
ReadAMRHierarchyHDF5(string filename,  AMRData& data)
{
  HDF5Handle handle;
  int err = handle.open(filename.c_str(),  HDF5Handle::OPEN_RDONLY);
  if( err < 0) return -4;
  Vector<LevelData<FArrayBox>* > vectData;
  int eekflag = ReadAMRHierarchyHDF5(handle, data.vectGrids, vectData,
                                     data.vectNames, data.domain, data.dx, data.dt,
                                     data.time, data.vectRatio, data.numLevels, 
                                     data.ghost);
  data.vectData.resize(vectData.size());
  for(int i=0; i<vectData.size(); ++i)
	{
	  data.vectData[i] = vectData[i];
	}

#ifdef MPI
  MPI_Barrier(Chombo_MPI::comm);
#endif  
  handle.close();

  return (eekflag);
}

int
ReadAMRHierarchyHDF5(HDF5Handle& handle,           
                     Vector<DisjointBoxLayout>& a_vectGrids, 
                     Vector<LevelData<FArrayBox>* > & a_vectData, 
                     Vector<string>& a_vectNames,                 
                     Box& a_domain,                               
                     Real& a_dx,                                         
                     Real& a_dt,                                         
                     Real& a_time,                                         
                     Vector<int>& a_refRatio,                           
                     int& a_numLevels,
                     const IntVect& ghostVector)
{

  HDF5HeaderData header;
  header.readFromFile(handle);


  a_numLevels = header.m_int["num_levels"];
  if(a_numLevels <= 0)
  {
    MayDay::Warning("ReadAMRHierarchyHDF5: Bogus number of levels");
    return (-1);
  }
  a_vectData.resize(a_numLevels);
  a_refRatio.resize(a_numLevels);
  a_vectGrids.resize(a_numLevels);

  int nComp = header.m_int["num_components"];
  if(nComp <= 0)
  {
    MayDay::Warning("ReadAMRHierarchyHDF5: Bogus number of Components");
    return (-2);
  }
  a_vectNames.resize(nComp);

  for(int ivar = 0; ivar < nComp; ivar++)
    {
      char labelChSt[80];
      sprintf(labelChSt, "component_%d", ivar);
      string label(labelChSt);
      a_vectNames[ivar] = header.m_string[label];
    }
  for(int ilev = 0; ilev < a_numLevels; ilev++)
    {
      int refLevel = 0;
      Box domainLevel;
      Real dtLevel;
      Real dxLevel;
      a_vectData[ilev] = new LevelData<FArrayBox>();
      int eek = readLevel(handle, ilev, *(a_vectData[ilev]), 
                          dxLevel, dtLevel,  a_time,
                          domainLevel, refLevel, Interval(), ghostVector, true);
      if(eek != 0)
      {
        MayDay::Warning("ReadAMRHierarchyHDF5: readLevel failed");
        return (-3);
      }

      const DisjointBoxLayout& dbl = a_vectData[ilev]->getBoxes();
      a_vectGrids[ilev]= dbl;

      if(ilev == 0)
        {
          a_domain = domainLevel;
          a_dt = dtLevel;
          a_dx = dxLevel;
        }
      a_refRatio[ilev] = refLevel;
    }
  return (0);
}

int
ReadAMRHierarchyHDF5(const string& filename,                             
                     Vector<DisjointBoxLayout>& a_vectGrids, 
                     Vector<LevelData<FArrayBox>* > & a_vectData, 
                     Box& a_domain,                               
                     Vector<int>& a_refRatio,                           
                     int& a_numLevels,
                     const IntVect& ghostVector)
{
  HDF5Handle handle;
  int err = handle.open(filename.c_str(),  HDF5Handle::OPEN_RDONLY);
  if( err < 0) return -4;

  int eekflag = ReadAMRHierarchyHDF5(handle, a_vectGrids, a_vectData,
                                     a_domain, a_refRatio, a_numLevels, 
                                     ghostVector);
#ifdef MPI
  MPI_Barrier(Chombo_MPI::comm);
#endif  
  handle.close();
  return (eekflag);
}

int
ReadAMRHierarchyHDF5(HDF5Handle& handle,     
                     Vector<DisjointBoxLayout>& a_vectGrids, 
                     Vector<LevelData<FArrayBox>* > & a_vectData, 
                     Box& a_domain,                               
                     Vector<int>& a_refRatio,                           
                     int& a_numLevels,
                     const IntVect& ghostVector)
{
  HDF5HeaderData header;
  header.readFromFile(handle);


  a_numLevels = header.m_int["num_levels"];
  if(a_numLevels <= 0)
  {
    MayDay::Warning("ReadAMRHierarchyHDF5: Bogus number of levels");
    return (-1);
  }
  a_vectData.resize(a_numLevels);
  a_refRatio.resize(a_numLevels);
  a_vectGrids.resize(a_numLevels);

  //  int nComp = header.m_int["num_components"];
  for(int ilev = 0; ilev < a_numLevels; ilev++)
    {
      int refLevel = 0;
      Box domainLevel;
      Real dtLevel;
      Real dxLevel;
      Real time;
      a_vectData[ilev] = new LevelData<FArrayBox>();
      int eek = readLevel(handle, ilev, *(a_vectData[ilev]), 
                          dxLevel, dtLevel,  time,
                          domainLevel, refLevel, Interval(), ghostVector, true);
      if(eek != 0)
      {
        MayDay::Warning("ReadAMRHierarchyHDF5: readLevel failed");
        return (-3);
      }

      const DisjointBoxLayout& dbl = a_vectData[ilev]->getBoxes();
      a_vectGrids[ilev]= dbl;

      if(ilev == 0)
        {
          a_domain = domainLevel;
        }
      a_refRatio[ilev] = refLevel;
    }

  return (0);
}

// put simple debugging functions here, at the end of the hdf5 stuff

void
writeFAB(const FArrayBox* a_dataPtr)
{
  if (a_dataPtr == NULL) return;

  char* fname = "fab.hdf5";
  writeFABname(a_dataPtr, fname);
}

void
viewFAB(const FArrayBox* a_dataPtr)
{
  if (a_dataPtr == NULL) return;

  char* fname = tempnam(NULL,NULL);
  writeFABname(a_dataPtr, fname);

  char command[2000];
  sprintf(command,"$CHOMBOVIS_HOME/chombovis user_script=$CHOMBOVIS_HOME/examples/browser.py %s &",fname);
  system(command);
}

void
writeFABname(const FArrayBox* a_dataPtr,
             const char*      a_filename)
{
  if (a_dataPtr == NULL) return;

  const FArrayBox& data = *a_dataPtr;

  HDF5Handle handle(a_filename, HDF5Handle::CREATE);
  HDF5HeaderData header;

  int numlevels= 1;
  int nComp = data.nComp();
  
  string filedescriptor("VanillaAMRFileType");
  header.m_string ["filetype"]      = filedescriptor;
  header.m_int ["num_levels"]       = numlevels;
  header.m_int ["num_components"]    = nComp;

  for(int ivar = 0; ivar < nComp; ivar++)
    {
      char labelChSt[80];
      sprintf(labelChSt, "component_%d", ivar);
      string label(labelChSt);
      header.m_string[label] = label;
    }
  header.writeToFile(handle);

  Box domainLevel = data.box();
  // put bogus numbers here
  Real dtLevel = 1.0;
  Real dxLevel = 1.0;
  Real time = 1.0;
 
  int refLevel = 1;
  
  // build bogus DisjointBoxLayout here
  Vector<Box> boxes(1,domainLevel);
  unsigned int myprocID= procID();
  Vector<int> procAssign(1,myprocID);
  DisjointBoxLayout grids(boxes, procAssign);
  LevelData<FArrayBox> ldf(grids, nComp);
  // now copy fab into ldf 
  DataIterator dit = ldf.dataIterator();
  ldf[dit()].copy(data);

  int eek = writeLevel(handle, 0, ldf, dxLevel, dtLevel, time,
                       domainLevel, refLevel);
  if (eek != 0) 
    MayDay::Error("writeFABname: error in writeLEvel");

  
  handle.close();
}

void
writeLDF(const LDF* a_dataPtr)
{
  if (a_dataPtr == NULL) return;

  char* fname = "LDF.hdf5";
  writeLevelname(a_dataPtr->dataPtr(), fname);
}

void 
writeLevel(const LevelData<FArrayBox>* a_dataPtr)
{
  if (a_dataPtr == NULL) return;

  char* fname = "LDF.hdf5";
  writeLevelname(a_dataPtr, fname);
}

void 
viewLevel(const LevelData<FArrayBox>* a_dataPtr)
{
  if (a_dataPtr == NULL) return;

  char* fname = tempnam(NULL,NULL);
  writeLevelname(a_dataPtr, fname);

  char command[2000];
  sprintf(command,"$CHOMBOVIS_HOME/chombovis user_script=$CHOMBOVIS_HOME/examples/browser.py %s &",fname);
  system(command);
}

void
writeLDFname(const LDF*    a_dataPtr,
             const string& a_filename)
{
  if (a_dataPtr == NULL) return;

  writeLevelname(a_dataPtr->dataPtr(), a_filename.c_str());
}

void 
writeLevelname(const LevelData<FArrayBox>* a_dataPtr,
               const char*                 a_filename)
{
  if (a_dataPtr == NULL) return;

  const LevelData<FArrayBox>& data = *a_dataPtr;

  HDF5Handle handle(a_filename, HDF5Handle::CREATE);
  HDF5HeaderData header;

  int numlevels = 1;
  int nComp = data.nComp();
  
  string filedescriptor("VanillaAMRFileType");
  header.m_string ["filetype"]      = filedescriptor;
  header.m_int ["num_levels"]       = numlevels;
  header.m_int ["num_components"]    = nComp;

  for(int ivar = 0; ivar < nComp; ivar++)
    {
      char labelChSt[80];
      sprintf(labelChSt, "component_%d", ivar);
      string label(labelChSt);
      header.m_string[label] = label;
    }
  header.writeToFile(handle);


  // need to figure out what domain will contain this LevelData
  DataIterator dit = data.dataIterator();
  const DisjointBoxLayout& levelBoxes = data.getBoxes();
  dit.reset();
  Box domain = levelBoxes[dit()];
  for (dit.reset(); dit.ok(); ++dit)
    {
      const Box thisBox = levelBoxes[dit()];
      D_TERM(
             if (thisBox.smallEnd(0)<domain.smallEnd(0)) 
                domain.setSmall(0,thisBox.smallEnd(0)); , 
             if (thisBox.smallEnd(1)<domain.smallEnd(1)) 
                domain.setSmall(1,thisBox.smallEnd(1)); , 
             if (thisBox.smallEnd(2)<domain.smallEnd(2)) 
                domain.setSmall(2, thisBox.smallEnd(2)); );

      D_TERM(
             if (thisBox.bigEnd(0)>domain.bigEnd(0)) 
                domain.setBig(0,thisBox.bigEnd(0)); , 
             if (thisBox.bigEnd(1)>domain.bigEnd(1)) 
                domain.setBig(1,thisBox.bigEnd(1)); , 
             if (thisBox.bigEnd(2)>domain.bigEnd(2)) 
                domain.setBig(2, thisBox.bigEnd(2)); );
             
    } // end loop over levels to determine "domain"

  // put bogus numbers here
  Real dtLevel = 1.0;
  Real dxLevel = 1.0;
  Real time = 1.0;
 
  int refLevel = 1;
  
  IntVect ghostVect = data.ghostVect();
  int eek = writeLevel(handle, 0, data, dxLevel, dtLevel, time,
                       domain, refLevel, ghostVect);

  if (eek != 0) 
    MayDay::Error("writeLDFname: error in writeLEvel");

  
  handle.close();
}

void
WritePartialAMRHierarchyHDF5(const string& filename,
                             const Vector<DisjointBoxLayout>& a_vectGrids, 
                             const Vector<LevelData<FArrayBox>* > & a_vectData,
                             const Vector<string>& a_vectNames,
                             const Box& a_baseDomain,
                             const Real& a_baseDx,
                             const Real& a_dt,
                             const Real& a_time,
                             const Vector<int>& a_vectRatio,
                             const Interval& a_levels)
{
  int numLevels = a_levels.size();
  
  // now make new dataholders which only have numLevels entries, 
  // and which will move the baseLevel to level 0

  Vector<DisjointBoxLayout> newVectGrids(numLevels);
  Vector<LevelData<FArrayBox> * > newVectData(numLevels);
  Vector<int> newVectRatio(numLevels);

  int leveloffset = a_levels.begin();
  for (int srcLevel = a_levels.begin(); srcLevel <= a_levels.end(); srcLevel++)
    {
      int destLevel = srcLevel-leveloffset;
      newVectGrids[destLevel] = a_vectGrids[srcLevel];
      newVectData[destLevel] = a_vectData[srcLevel];
      newVectRatio[destLevel] = a_vectRatio[srcLevel];
    }

  WriteAMRHierarchyHDF5(filename, newVectGrids, newVectData, a_vectNames,
                        a_baseDomain, a_baseDx, a_dt, a_time,
                        newVectRatio, numLevels);
}


  
void writeAFabASCII(ostream& os, const FArrayBox& fab)
{
    os << fab.box() << "  " << fab.nComp() << endl;;
    BoxIterator bit(fab.box());
    for(bit.begin(); bit.ok(); bit.next())
    {
        for(int ivar = 0; ivar < fab.nComp(); ivar++)
        {
            os << fab(bit(), ivar) << endl;
        }
    }

}
void readAFabASCII(istream& is, FArrayBox& fab)
{
    Box fabbox;
    int ncomp;
    is >> fabbox >> ncomp;
    fab.resize(fabbox, ncomp);
    BoxIterator bit(fabbox);
    for(bit.begin(); bit.ok(); bit.next())
    {
        for(int ivar = 0; ivar < fab.nComp(); ivar++)
        {
            Real temp;
            is >> temp;
            fab(bit(), ivar) = temp;
        }
    }

}

//this outputs stuff to the amrascii format very slowly
/*
\\ write out hierarchy of amr data in HDF5 format
\\ filename,  == file to output to
\\ a_vectData == data at each level
\\ a_vectNames== names of variables
\\ a_domain == domain at coarsest level
\\ a_dx     == grid spacing at coarsest level
\\ a_dt     == time step at coarsest level
\\ a_time     == time 
\\ a_refRatio == refinement ratio at all levels 
\\ (ith entry is refinement ratio between levels i and i + 1)
\\ a_numLevel s== number of levels to output
*/
void
WriteAMRHierarchyASCII(const string& filename,                             
                       const Vector<DisjointBoxLayout>& a_vectGrids, 
                       const Vector<LevelData<FArrayBox>* > & a_vectData, 
                       const Vector<string>& a_vectNames,                 
                       const Box& a_domain,                               
                       const Real& a_dx,                                         
                       const Real& a_dt,                                         
                       const Real& a_time,                                         
                       const Vector<int>& a_refRatio,                           
                       const int& a_numLevels)
{
  assert(a_numLevels > 0);
  assert(a_vectData.size()  >= a_numLevels);
  assert(a_refRatio.size() >= a_numLevels-1);


  //sanity checks
  int nvar = a_vectData[0]->nComp();
  for (int ivec = 0; ivec < a_vectData.size(); ivec++)
    assert(nvar == a_vectData[ivec]->nComp());
  assert(a_vectNames.size() == nvar);
  for (int ivec = 0; ivec < a_refRatio.size(); ivec++)
    assert(a_refRatio[ivec] > 0);

  fstream os;
  os.open(filename.c_str(), ios::out);

  os <<  nvar           << endl; // Number of states dumped
  for(int iname = 0; iname < a_vectNames.size(); iname++)
    {
      os << a_vectNames[iname]  << endl; // Name of state
    }

  os << SpaceDim     << endl; // Dimension of data
  os << a_numLevels  << endl; // number of levels dumped

  for(int irat = 0; irat < a_numLevels-1; irat++)
    os << a_refRatio[irat] << " " ;          //refinement ratios
  os << endl;                   
  os << a_domain << endl;                   //coarse level domain
  os << a_dx     << endl;                   //coarse level grid spacing
  os << a_dt     << endl;                   //coarse level time step
  os << a_time   << endl;                   // Simulation time of dump

  for (int ilevel = 0; ilevel < a_numLevels; ilevel++)
    {
      os << ilevel << endl;  //level number
      const LevelData<FArrayBox>& mf_lev = *(a_vectData[ilevel]);
      const DisjointBoxLayout& bxa = a_vectGrids[ilevel];

      // dump actual boxes
      LayoutIterator lit = bxa.layoutIterator();
      os << bxa.size() << endl;              // number of boxes
      for(lit.reset(); lit.ok(); ++lit)
        os << bxa[lit()] << endl;              // For each grid, dump box

      // dump actual data
      DataIterator dit = mf_lev.dataIterator();
      for(dit.reset(); dit.ok(); ++dit)
        {
          const FArrayBox& dat = mf_lev[dit()];
          writeAFabASCII(os, dat);
        }
    }
  os.close();
}

//this reads stuff in  amrascii format very slowly
/*
\\ write out hierarchy of amr data in HDF5 format
\\ filename,  == file to output to
\\ a_vectData == data at each level
\\ a_vectNames== names of variables
\\ a_domain == domain at coarsest level
\\ a_dx     == grid spacing at coarsest level
\\ a_dt     == time step at coarsest level
\\ a_time     == time 
\\ a_refRatio == refinement ratio at all levels 
\\ (ith entry is refinement ratio between levels i and i + 1)
\\ a_numLevels == number of levels to output

return values:
0 = success
-1 = failure
*/
int
ReadAMRHierarchyASCII(const string& filename,                             
                      Vector<DisjointBoxLayout>& a_vectGrids, 
                      Vector<LevelData<FArrayBox>* > & a_vectData, 
                      Vector<string>& a_vectNames,                 
                      Box& a_domain,                               
                      Real& a_dx,                                         
                      Real& a_dt,                                         
                      Real& a_time,                                         
                      Vector<int>& a_refRatio,                           
                      int& a_numLevels,
                      const IntVect& ghostVector)
{
  fstream is;
  if(numProc() > 1)
    {
      MayDay::Warning("ReadAMRHierarchyASCII is written for one processor");
    }
  is.open(filename.c_str(), ios::in);

  int nvar;
  is  >>  nvar;             // number of states
  a_vectNames.resize(nvar);
  for(int iname = 0; iname < a_vectNames.size(); iname++)
    {
      is >> a_vectNames[iname];   // Name of state
    }

  int dimdat;
  is >> dimdat; //dimension of problem
  if(dimdat != SpaceDim)
    {
      MayDay::Warning("ReadAMRHierarchyASCII: dimension of data != dimension of reader");
      return (-1);
    }
  is >> a_numLevels; // number of levels dumped

  a_refRatio.resize(a_numLevels);
  a_vectGrids.resize(a_numLevels);
  a_vectData.resize(a_numLevels);
  for(int irat = 0; irat < a_numLevels-1; irat++)
    is >> a_refRatio[irat];          //refinement ratios

  is >> a_domain;                //coarse level domain
  is >> a_dx;                   //coarse level grid spacing
  is >> a_dt;                   //coarse level time step
  is >> a_time;                 // Simulation time of dump

  for (int ilevel = 0; ilevel < a_numLevels; ilevel++)
    {
      int itemp;
      is  >> itemp;  //level number
      if(itemp != ilevel)
        {
          MayDay::Warning("ReadAMRHierarchyASCII: Input file wrong");
          return (-1);
        }

      DisjointBoxLayout& dbl = a_vectGrids[ilevel];
      int nboxes;
      is >> nboxes;
      for(int ibox = 0; ibox < nboxes; ibox++)
        {
          Box btemp;
          is >> btemp;
          dbl.addBox(btemp, 0);
        }
      dbl.close();
      LevelData<FArrayBox> mf_lev(dbl,nvar,ghostVector);
      DataIterator dit = mf_lev.dataIterator();
      // read actual data
      for(dit.reset(); dit.ok(); ++dit)
        {
          FArrayBox datnoGhost;
          readAFabASCII(is, datnoGhost);
          FArrayBox& dat = mf_lev[dit()];
          dat.copy(datnoGhost);
        }
      a_vectData[ilevel] = new LevelData<FArrayBox>();
      a_vectData[ilevel]->define(mf_lev);
    }
  is.close();

  return (0);
}

// Read in hierarchy of amr data in ASCII format
int
ReadEBAMRASCII(const string& filename,                             
               Vector<DisjointBoxLayout>& a_vectGrids, 
               Vector<LevelData<FArrayBox>* > & a_vectData, 
               Vector<string>& a_vectNames,                 
               Box& a_domain,                               
               Vector<IntVectSet>& a_coveredCells,                               
               Vector<IntVectSet>& a_multiValuedCells,                               
               Vector<int>& a_refRatio,                           
               int& a_numLevels,
               const IntVect& a_ghostVector)
{
  fstream is;
  if(numProc() > 1)
    {
      MayDay::Warning("ReadAMRHierarchyASCII is written for one processor");
    }
  is.open(filename.c_str(), ios::in);

  int nvar;
  is  >>  nvar;             // number of states
  a_vectNames.resize(nvar);
  for(int iname = 0; iname < a_vectNames.size(); iname++)
    {
      is >> a_vectNames[iname];   // Name of state
    }

  int dimdat;
  is >> dimdat; //dimension of problem
  if(dimdat != SpaceDim)
    {
      MayDay::Warning("ReadAMRHierarchyASCII: dimension of data != dimension of reader");
      return (-1);
    }
  is >> a_numLevels; // number of levels dumped

  a_coveredCells.resize(a_numLevels);
  a_multiValuedCells.resize(a_numLevels);
  for(int ilev = 0; ilev < a_numLevels; ilev++)
    {
      int ivecsize;
      is >>  ivecsize; // number of covered boxes
      Box coveredBox;
      IntVectSet& coveredCells = a_coveredCells[ilev];
      coveredCells.makeEmpty();
      for(int ibox = 0; ibox < ivecsize; ibox++)
        {
          is >>  coveredBox; //input each covered box
          assert(!coveredBox.isEmpty());
          coveredCells |= coveredBox;
        }

      IntVectSet& multiIVS = a_multiValuedCells[ilev];
      multiIVS.makeEmpty();
      is >> ivecsize ; //number of multi-valued cells
      for(int ivec = 0; ivec < ivecsize; ivec++)
        {
          IntVect iv;
          is >> iv;
          multiIVS |= iv;
        }
    }

  a_refRatio.resize(a_numLevels);
  a_vectGrids.resize(a_numLevels);
  a_vectData.resize(a_numLevels);
  for(int irat = 0; irat < a_numLevels-1; irat++)
    is >> a_refRatio[irat];          //refinement ratios

  is >> a_domain;                //coarse level domain
  for (int ilevel = 0; ilevel < a_numLevels; ilevel++)
    {
      int itemp;
      is  >> itemp;  //level number
      if(itemp != ilevel)
        {
          MayDay::Warning("ReadAMRHierarchyASCII: Input file wrong");
          return (-1);
        }

      DisjointBoxLayout& dbl = a_vectGrids[ilevel];
      int nboxes;
      is >> nboxes;
      for(int ibox = 0; ibox < nboxes; ibox++)
        {
          Box btemp;
          is >> btemp;
          dbl.addBox(btemp, 0);
        }
      dbl.close();
      LevelData<FArrayBox> mf_lev(dbl,nvar,a_ghostVector);
      DataIterator dit = mf_lev.dataIterator();
      // read actual data
      for(dit.reset(); dit.ok(); ++dit)
        {
          FArrayBox datnoGhost;
          //fab was written out as a farraybox.  no need
          //for special read function
          readAFabASCII(is, datnoGhost);
          FArrayBox& dat = mf_lev[dit()];
          dat.copy(datnoGhost);
        }
      a_vectData[ilevel] = new LevelData<FArrayBox>();
      a_vectData[ilevel]->define(mf_lev);
    }
  is.close();
  return (0);
}
